unit CBase;

interface

uses
  Kernel, Objects, Lists, Streaming, Files, Configs, Platforms;

const

  sAppName = 'Vide';
  sFileTypesCfg = 'FileTypes.xml';

  sVideProject = 'Vide Project';
  sVideProjectExt = 'videproj';
  sVideSolution = 'Vide Solution';
  sVideSolutionExt = 'videsol';

type

  {Type of Vide project:
    ptGuiApp      - project is GUI executable application
    ptConsoleApp  - project is Console executable application
    ptLibrary     - project is dynamicaly link library
    ptCustom      - unknown type, will not be compiled
    ptPackage     - project is Delphi package}
  TProjectType = (ptGuiApp, ptConsoleApp, ptLibrary, ptCustom, ptPackage);

  {Type of file:
    ffText    - file is source/text file
    ffBinary  - file is binary file}
  TFileFormat = (ffText, ffBinary);

  {Type of text file (used for syntax highlight):
    ttNone         - no syntax higlight
    ttPascal       - use Pascal/Delphi syntax
    ttCpp          - use C/C++ syntax
    ttAsm          - use X86 assembly syntax
    ttJava         - use Java syntax
    ttCSharp       - use C# syntax
    ttBasic        - use Basic/VB syntax
    ttDelphiForm   - use syntax for Delphi/Kylix form files
    ttXML          - use XML syntax
    ttHTML         - use HTML syntax
    ttIni          - use syntax for Windows ini files}
  TTextType = (ttNone, ttPascal, ttCpp, ttAsm, ttJava, ttCSharp, ttBasic,
    ttDelphiForm, ttXML, ttHTML, ttINI);

  {Type of binary file (some binary files may be opened or edited):
    btUnknown      - unknown type of binary file
    btResource     - file is resource file
    btIcon         - file is icon image
    btImage        - file is image
    btExecutable   - file is Exe/Dll/Bpl/... executable file}
  TBinaryType = (btUnknown, btResource, btIcon, btImage, btExecutable);

  TVidePersistent = class(TNPersistent)
  end;

  TVideSingleton = class(TNSingleton)
    procedure Initialize; virtual;
    procedure Finalize; virtual;
  end;

  TVideFileType = class(TVidePersistent)
  protected
    FFormat: TFileFormat;
    FName: NString;
    FExtensions: TNStringList;
    FTextType: TTextType;
    FBinaryType: TBinaryType;
    FCanOpenToEditor: Boolean;
    FFixed: Boolean;
    function GetFilter: NString;
    procedure SetExtensions(const Value: TNStringList);
  public
    constructor Create; override;
    destructor Destroy; override;
    function Clone: TVideFileType;
    procedure AssignTo(Dest: TNPersistent); override;
  published
    property Format: TFileFormat read FFormat write FFormat;
    property Name: NString read FName write FName;
    property Extensions: TNStringList read FExtensions write SetExtensions;
    property TextType: TTextType read FTextType write FTextType;
    property BinaryType: TBinaryType read FBinaryType write FBinaryType;
    property CanOpenToEditor: Boolean read FCanOpenToEditor write
      FCanOpenToEditor;
    property Fixed: Boolean read FFixed write FFixed;
    property Filter: NString read GetFilter;
  end;

  TVideFileTypeList = class(TVidePersistent)
  protected
    FList: TNObjectList;
    function GetCount: Long;
    function GetType(Index: Long): TVideFileType;
    function GetTypeExt(const Ext: NString): TVideFileType;
    function GetTypeName(const AName: NString): TVideFileType;
  public
    constructor Create; override;
    destructor Destroy; override;
    procedure LoadTypes;
    procedure SaveTypes;
    procedure Add(NewType: TVideFileType);
    procedure Delete(Index: Long);
    procedure Clear;
    function IndexOfType(FileType: TVideFileType): Long;
    property Count: Long read GetCount;
    property Types[Index: Long]: TVideFileType read GetType;
    property TypesExt[const Ext: NString]: TVideFileType read GetTypeExt;
    default;
    property TypesName[const AName: NString]: TVideFileType read GetTypeName;
  end;

  {Basic file item.}
  TVideItem = class(TVidePersistent)
  protected
    FName: NString;
    FFileName: NString;
    FDirty: Boolean;
    FModified: Boolean;
    procedure SetFileName(const Value: NString); virtual;
  public
    {This renames item (path and ext is the same, only filename
     is changed) and deletes old file.}
    procedure Rename(const NewName: NString); virtual;
    {Saves item to file and changes FileName (leaving old file intact).}
    procedure SaveToFile(const AFileName: NString); virtual;
  published
    {Items's name.}
    property Name: NString read FName write FName;
    {File name of this item.}
    property FileName: NString read FFileName write SetFileName;
    {True if this item was never saved (e.g. File->New).}
    property Dirty: Boolean read FDirty write FDirty;
    {True if item was changed from its original state (e.g.
     writting for files, add/remove for projects).}
    property Modified: Boolean read Fmodified write FModified;
  end;

  TVidePropItemList = class;

  {Basic item which conatins more items.}
  TVideMultiItem = class(TVideItem)
  protected
    FList: TVidePropItemList;
    FActiveIndex: Long;
    function GetCount: Long;
    function GetItem(Index: Long): TVideItem;
    function GetItemByName(const AFileName: NString): TVideItem;
    procedure SetActiveIndex(Value: Long); virtual;
    procedure SetFileName(const Value: NString); override;
  public
    constructor Create; override;
    destructor Destroy; override;
    {Adds item to multi items. Use this rather than accessing FList.Add.}
    procedure Add(Item: TVideItem); virtual;
    {Deletes all items.}
    procedure Clear;
    {Returns index of item defined by FileName. Returns -1 if
     item was not found.}
    function IndexOfFileName(const AFileName: NString): Long;
    {Deletes item defined by index form list and frees it.}
    procedure Delete(Index: Long); virtual;
    {Number of items in this multi item.}
    property Count: Long read GetCount;
    {Access to individual items by index.}
    property Items[Index: Long]: TVideItem read GetItem;
    {Access to individual items by Name.}
    property ItemsByName[const AFileName: NString]: TVideItem read
    GetItemByName; default;
    property ActiveIndex: Long read FActiveIndex write SetActiveIndex;
  end;

  TVidePropItemList = class(TNCustomPropObjectList)
  protected
    function GetItem(Index: Long): TVideItem; reintroduce;
    procedure PutItem(Index: Long; Value: TVideItem); reintroduce;
  public
    property Items[Index: Long]: TVideItem read GetItem write PutItem;
  end;

function ExpandFileTo(const FileName: NString; const BasePath: NString): NString;

implementation

uses
  SysUtils, CSingletons;

function ExpandFileTo(const FileName: NString; const BasePath: NString): NString;
var
 OldPath: NString;
begin
  OldPath:= GetCurrentDir;
  try
   if SysUtils.DirectoryExists(BasePath) then
   begin
     ChDir(BasePath);
     Result:= ExpandFileName(FileName);
   end
   else
    Result:=FileName;
  finally
   ChDir(OldPath);
  end;
end;


{ TVideSingleton }

procedure TVideSingleton.Finalize;
begin
end;

procedure TVideSingleton.Initialize;
begin
end;

{ TFileType }

procedure TVideFileType.AssignTo(Dest: TNPersistent);
begin
  if Dest is TVideFileType then
    with TVideFileType(Dest) do
    begin
      Name := Self.Name;
      Format := Self.Format;
      TextType := Self.TextType;
      BinaryType := Self.BinaryType;
      CanOpenToEditor := Self.CanOpenToEditor;
      Fixed := Self.Fixed;
      Extensions.Text := Self.Extensions.Text;
    end
  else
    inherited AssignTo(Dest);
end;

function TVideFileType.Clone: TVideFileType;
begin
  Result := TVideFileType.Create;
  Result.Assign(Self);
end;

constructor TVideFileType.Create;
begin
  inherited Create;
  FExtensions := TNStringList.Create;
end;

destructor TVideFileType.Destroy;
begin
  FExtensions.Free;
  inherited Destroy;
end;

function TVideFileType.GetFilter: NString;
var
  J: Long;
begin
  Result := '';
  Result := Result + FName + ' (';
  for J := 0 to FExtensions.Count - 1 do
    Result := Result + '*.' + FExtensions[J] + Iff(J < FExtensions.Count -
      1, ',', ')|');
  for J := 0 to FExtensions.Count - 1 do
    Result := Result + '*.' + FExtensions[J] + Iff(J < FExtensions.Count -
      1, ';', '|');
end;

procedure TVideFileType.SetExtensions(const Value: TNStringList);
begin
  FExtensions.Clear;
  FExtensions.AddStrings(Value);
end;

{ TVideItem }

procedure TVideItem.Rename(const NewName: NString);
begin
  if FDirty then
  begin
    FileName := NewName;
    Solution.Changed := True;
  end
  else
    if Vfs.FileExists(FFileName) then
    begin
      Vfs.DeleteFile(FFileName);
      SaveToFile(NewName);
      Solution.Changed := True;
    end;
end;

procedure TVideItem.SaveToFile(const AFileName: NString);
begin
  FDirty := False;
  FModified := False;
  FileName := AFileName;
end;

procedure TVideItem.SetFileName(const Value: NString);
begin
  FFileName := Value;
  FName := ChangeFileExt(ExtractFileName(FFileName), '');
  Solution.Changed := True;
end;


{*************************************************************************}
{                         TVideMultiItem methods                          }
{*************************************************************************}


procedure TVideMultiItem.Add(Item: TVideItem);
begin
  FList.Add('FileName', Item);
end;

procedure TVideMultiItem.Clear;
begin
  while FList.Count > 0 do
    Delete(0);
end;

constructor TVideMultiItem.Create;
begin
  inherited Create;
  FList := TVidePropItemList.Create;
  FActiveIndex := -1;
end;

procedure TVideMultiItem.Delete(Index: Long);
begin
  if (Index > -1) and (Index < FList.Count) then
  begin
    TVideItem(FList.Items[Index]).Free;
    FList.Delete(Index);
  end;
  if (FActiveIndex = Index) and (FActiveIndex = FList.Count) then
    FActiveIndex := FList.Count - 1;
end;

destructor TVideMultiItem.Destroy;
begin
  Clear;
  FList.Free;
  inherited Destroy;
end;

function TVideMultiItem.GetCount: Long;
begin
  Result := FList.Count;
end;

function TVideMultiItem.GetItem(Index: Long): TVideItem;
begin
  Result := FList.Items[Index];
end;

function TVideMultiItem.GetItemByName(const AFileName: NString): TVideItem;
begin
  Result := FList.Items[FList.IndexOf(AFileName)];
end;

function TVideMultiItem.IndexOfFileName(const AFileName: string): Long;
begin
  Result := FList.IndexOf(AFileName);
end;

procedure TVideMultiItem.SetActiveIndex(Value: Long);
begin
  FActiveIndex := Value;
  Modified := True;
end;

procedure TVideMultiItem.SetFileName(const Value: NString);
begin
  FFileName := Value;
  FName := ChangeFileExt(ExtractFileName(FFileName), '');
end;

{ TVidePropItemList }

function TVidePropItemList.GetItem(Index: Long): TVideItem;
begin
  if (Index > -1) and (Index < FCount) then
    Result := PRefPtrArray(FData)[Index].Value
  else
    Result := nil;
end;

procedure TVidePropItemList.PutItem(Index: Long; Value: TVideItem);
begin
  if (Index > -1) and (Index < FCount) then
  begin
    Notify(Index, lnValueDeleted);
    PRefPtrArray(FData)[Index].Value := Value;
    Notify(Index, lnValueAdded);
  end;
end;

{ TVideFileTypeList }

procedure TVideFileTypeList.Add(NewType: TVideFileType);
begin
  if (FList.IndexOf(NewType) < 0) and (NewType <> nil) then
    FList.Add(NewType);
end;

procedure TVideFileTypeList.Clear;
begin
  while Count > 0 do
    Delete(0);
end;

constructor TVideFileTypeList.Create;
begin
  inherited Create;
  FList := TNObjectList.Create;
end;

procedure TVideFileTypeList.Delete(Index: Long);
var
  FT: TVideFileType;
begin
  FT := GetType(Index);
  if FT <> nil then
  begin
    FT.Free;
    FList.Delete(Index);
  end;
end;

destructor TVideFileTypeList.Destroy;
begin
  while FList.Count > 0 do
    Delete(0);
  FList.Free;
  inherited Destroy;
end;

function TVideFileTypeList.GetCount: Long;
begin
  Result := FList.Count;
end;

function TVideFileTypeList.GetType(Index: Long): TVideFileType;
begin
  Result := TVideFileType(FList.Objects[Index]);
end;

function TVideFileTypeList.GetTypeExt(const Ext: NString): TVideFileType;
var
  I, J: Long;
begin
  Result := nil;
  for I := 0 to FList.Count - 1 do
    for J := 0 to GetType(I).FExtensions.Count - 1 do
      if SameStr(GetType(I).FExtensions[J], Ext) then
      begin
        Result := GetType(I);
        Exit;
      end;
end;

function TVideFileTypeList.GetTypeName(const AName: NString): TVideFileType;
var
  I: Long;
begin
  Result := nil;
  for I := 0 to FList.Count - 1 do
    if SameStr(GetType(I).Name, AName) then
    begin
      Result := GetType(I);
      Exit;
    end;
end;

function TVideFileTypeList.IndexOfType(FileType: TVideFileType): Long;
begin
  Result := FList.IndexOf(FileType);
end;

procedure TVideFileTypeList.LoadTypes;
var
  I, Count: Long;
  FType: TVideFileType;
  F: TNBinaryFile;
  Reader: TNXMLReader;
begin
  F := Vfs.Open(SettingsDir + PathDelim + sFileTypesCfg, fdRead);
  if F <> nil then
  try
    Reader := TNXMLReader.Create(F.Stream);
    Reader.Flush;
    Count := Reader.ReadInt;
    for I := 0 to Count - 1 do
    begin
      Reader.ReadInstance(TNPersistent(FType));
      Add(FType);
    end;
  except
    on E: Exception do
      ShowError(ErrorStr(Self, 'LoadTypes', 'Error when loading File Types. ' +
        sExceptMsg + E.Message), E.ClassName);
  end;
  F.Free;

  // provisorious
{ FType := TVideFileType.Create;
  FType.Name := 'Pascal Source File';
  FType.Format := ffText;
  FType.TextType := ttPascal;
  FType.CanOpenToEditor := True;
  FType.Extensions.Add('pas');
  FType.Extensions.Add('pp');
  FType.Extensions.Add('dpr');
  Add(FType);

  FType := TVideFileType.Create;
  FType.Name := 'XML Document';
  FType.Format := ffText;
  FType.TextType := ttXML;
  FType.CanOpenToEditor := True;
  FType.Extensions.Add('xml');
  FType.Extensions.Add('xsl');
  FType.Extensions.Add('xsd');
  Add(FType);

  FType := TVideFileType.Create;
  FType.Name := 'C/C++ Source File';
  FType.Format := ffText;
  FType.TextType := ttCpp;
  FType.CanOpenToEditor := True;
  FType.Extensions.Add('cpp');
  FType.Extensions.Add('c');
  FType.Extensions.Add('cc');
  Add(FType);   }

  FType := TVideFileType.Create;
  FType.Name := sVideProject;
  FType.Format := ffText;
  FType.TextType := ttXML;
  FType.CanOpenToEditor := True;
  FType.Fixed := True;
  FType.Extensions.Add(sVideProjectExt);
  Add(FType);

  FType := TVideFileType.Create;
  FType.Name := sVideSolution;
  FType.Format := ffText;
  FType.TextType := ttXML;
  FType.CanOpenToEditor := True;
  FType.Fixed := True;
  FType.Extensions.Add(sVideSolutionExt);
  Add(FType);
end;

procedure TVideFileTypeList.SaveTypes;
var
  I, Count: Long;
  F: TNBinaryFile;
  Writer: TNXMLWriter;
begin
  Count := 0;
  for I := 0 to FList.Count - 1 do
    if not Types[I].FFixed then
      Inc(Count);

  F := Vfs.Open(SettingsDir + PathDelim + sFileTypesCfg, fdCreate);
  if F <> nil then
  try
    Writer := TNXMLWriter.Create(F.Stream);
    Writer.WriteInt(Count);
    for I := 0 to FList.Count - 1 do
      if not Types[I].FFixed then
        Writer.WriteInstance(Types[I]);
    Writer.Flush;
    Writer.Free;
  except
    on E: Exception do
      ShowError(ErrorStr(Self, 'SaveTypes', 'Error when saving File Types. ' +
        sExceptMsg + E.Message), E.ClassName);
  end;
  F.Free;
end;

initialization

  Objects.RegisterClasses([TVideFileType]);

end.

