unit CEditor;

interface

uses
  SysUtils, Kernel, Objects, Lists, Files, Configs, CBase, CProjects;

type

  {Item for source/text files.}
  TVideEditorFile = class(TVideFile)
  private
    procedure SetFileInProject(const Value: TVideFile);
  protected
    FBuffer: TNStringList;
    FCaretX: Long;
    FCaretY: Long;
    FTopLine: Long;
    FFileInProject: TVideFile;
    FTag: TObject;
    procedure SetFileName(const Value: NString); override;
  public
    constructor Create(const AFileName: NString; AType: TVideFileType;
      ADirty: Boolean); override;
    destructor Destroy; override;
    procedure SaveToFile(const AFileName: NString); override;
    {Buffer containing lines of text source/file.}
    property Buffer: TNStringList read FBuffer;
    property CaretX: Long read FCaretX write FCaretX;
    property CaretY: Long read FCaretY write FCaretY;
    property TopLine: Long read FTopLine write FTopLine;
    property FileInProject: TVideFile read FFileInProject write
      SetFileInProject;
    property Tag: TObject read FTag write FTag;
  end;

  TVideEditor = class(TVideMultiItem)
  protected
    function GetFile(Index: Long): TVideEditorFile;
    function GetFileByName(const Name: NString): TVideEditorFile;
    function GetActiveFile: TVideEditorFile;
  public
    constructor Create; override;
    destructor Destroy; override;

    function AddFile(const FileName: NString; FileType: TVideFileType):
      TVideEditorFile; overload;
    function AddFile(AFile: TVideEditorFile): TVideEditorFile; overload;

    procedure Exchange(Index1, Index2: Long);

    property ActiveFile: TVideEditorFile read GetActiveFile;
    {Access to individual editor's files by index.}
    property Files[Index: Long]: TVideEditorFile read GetFile;
    {Access to individual editor's files by Name.}
    property FilesByName[const Name: NString]: TVideEditorFile read
    GetFileByName;
  end;


implementation

uses
  CSingletons;


{*************************************************************************}
{                         TVideEditorFile methods                         }
{*************************************************************************}


constructor TVideEditorFile.Create(const AFileName: NString; AType:
  TVideFileType; ADirty: Boolean);
var
  I, J: Long;
begin
  inherited Create(AFileName, AType, ADirty);
  FBuffer := TNStringList.Create;
  if not FDirty then
  begin
    try
      FBuffer.LoadFromFile(AFileName);
    except
      // file was not loaded -> not exists on real path
    end;
  end;
  with Solution do
    for I := 0 to Count - 1 do
      for J := 0 to Projects[I].Count - 1 do
        if SameStr(Projects[I].Files[J].FileName, FFileName) then
          FileInProject := Projects[I].Files[J];
end;

destructor TVideEditorFile.Destroy;
begin
  FBuffer.Free;
  if FTag <> nil then
    FTag.Free;
  inherited Destroy;
end;

procedure TVideEditorFile.SetFileName(const Value: NString);
begin
  if FFileInProject <> nil then
    FFileInProject.FileName := Value;
  inherited SetFileName(Value);
end;

procedure TVideEditorFile.SetFileInProject(const Value: TVideFile);
begin
  FFileInProject := Value;
  if Value <> nil then
    FProject := Value.Project
  else
    FProject := nil;
end;

procedure TVideEditorFile.SaveToFile(const AFileName: NString);
begin
  inherited SaveToFile(AFileName);
  if FFileInProject <> nil then
  begin
    FFileInProject.Dirty := False;
  end;
  Buffer.SaveToFile(AFileName);
end;

{ TVideEditorFiles }

function TVideEditor.AddFile(const FileName: NString;
  FileType: TVideFileType): TVideEditorFile;
begin
  Result := nil;
  if FList.IndexOf(FileName) < 0 then
  begin
    Result := TVideEditorFile.Create(FileName, FileType, False);
    Add(Result);
  end;
end;

function TVideEditor.AddFile(AFile: TVideEditorFile): TVideEditorFile;
begin
  Result := AFile;
  if (FList.IndexOf(AFile.FileName) < 0) and (AFile <> nil) then
    Add(Result);
end;

constructor TVideEditor.Create;
begin
  inherited Create;
  Log.Add('Editor created.');
  FList.Duplicates := dupIgnore;
end;

destructor TVideEditor.Destroy;
begin
  inherited Destroy;
end;

procedure TVideEditor.Exchange(Index1, Index2: Long);
begin
  FList.Exchange(Index1, Index2);
end;

function TVideEditor.GetActiveFile: TVideEditorFile;
begin
  Result := TVideEditorFile(FList.Items[FActiveIndex]);
end;

function TVideEditor.GetFile(Index: Long): TVideEditorFile;
begin
  Result := TVideEditorFile(FList.Items[Index]);
end;

function TVideEditor.GetFileByName(const Name: NString): TVideEditorFile;
begin
  Result := TVideEditorFile(FList.Items[FList.IndexOf(Name)]);
end;


end.

