unit CProjects;

interface

uses
  SysUtils, Kernel, Objects, Lists, Files, Configs, CBase;

type

  TVideProject = class;
  TVideCompilerSettings = class;

  TVideCompiler = class(TVidePersistent)
  protected
    FName: NString;
    FPath: NString;
  public
    function Compile(Project: TVideProject; Build:Boolean = False):Boolean; virtual; abstract;
    procedure ApplyDefaults(Settings: TVideCompilerSettings); virtual; abstract;
  published
    property Name: NString read FName write FName;
    property Path: NString read FPath write FPath;
  end;

  TTargetCPU = (tc386, tcP5, tcP6);
  TCompilerMode = (cmDefault, cmDelphi, cmObjFPC, cmTP, cmGPC);

  TVideCompilerSettings = class(TVidePersistent)
  protected
    FCompiler: TVideCompiler;
    FCmdLine: NString;
    FMinEnumSize: Long;
    FRecordAlign: Long;
    FTargetCPU: TTargetCPU;
    FMode: TCompilerMode;
    FBooleanEval: Boolean;
    FTypedPointers: Boolean;
    FExtSyntax: Boolean;
    FAnsiStrings: Boolean;
    FCppExtensions: Boolean;
    FDebugInfo: Boolean;
    FAssertions: Boolean;
    FWarnings: Boolean;
    FHints: Boolean;
    FRangeCheck: Boolean;
    FIOCheck: Boolean;
    FOverflowCheck: Boolean;
  public
    constructor Create(ACompiler: TVideCompiler); reintroduce;
    procedure LoadFromConfig(Cfg:TNConfig);
    procedure SaveToConfig(Cfg:TNConfig);
  published
    property Compiler: TVideCompiler read FCompiler write FCompiler;
    property CmdLine: NString read FCmdLine write FCmdLine;
    property MinEnumSize: Long read FMinEnumSize write FMinEnumSize;
    property RecordAlign: Long read FRecordAlign write FRecordAlign;
    property TargetCPU: TTargetCPU read FTargetCPU write FTargetCPU;
    property Mode: TCompilerMode read FMode write FMode;
    property BooleanEval: Boolean read FBooleanEval write FBooleanEval;
    property TypedPointers: Boolean read FTypedPointers write FTypedPointers;
    property ExtSyntax: Boolean read FExtSyntax write FExtSyntax;
    property AnsiStrings: Boolean read FAnsiStrings write FAnsiStrings;
    property CppExtensions: Boolean read FCppExtensions write FCppExtensions;
    property DebugInfo: Boolean read FDebugInfo write FDebugInfo;
    property Assertions: Boolean read FAssertions write FAssertions;
    property Warnings: Boolean read FWarnings write FWarnings;
    property Hints: Boolean read FHints write FHints;
    property RangeCheck: Boolean read FRangeCheck write FRangeCheck;
    property IOCheck: Boolean read FIOCheck write FIOCheck;
    property OverflowCheck: Boolean read FOverflowCheck write FOverflowCheck;
  end;

  {General item for individual file.}
  TVideFile = class(TVideItem)
  protected
    FFileType: TVideFileType;
    FProject: TVideProject;
  public
    constructor Create(const AFileName: NString; AType: TVideFileType;
      ADirty: Boolean); reintroduce; virtual;
    destructor Destroy; override;
    procedure Rename(const NewName: NString); override;
    {Describes type of this file.}
    property FileType: TVideFileType read FFileType;
    {Parent Project (can be nil for some editor files)}
    property Project: TVideProject read FProject write FProject;
  end;

  {Item representing single project. Project is set of
   various files and it has individual compiler and other settings.}
  TVideProject = class(TVideMultiItem)
  private
    function GetActiveFile: TVideFile;
  protected
    FCompilerSettings: TVideCompilerSettings;
    FProjectType: TProjectType;
    FOutputDir: NString;
    FPaths: TNStringList;
    FDescription: NString;
    function GetFile(Index: Long): TVideFile;
    function GetFileByName(const AName: NString): TVideFile;
  public
    constructor Create(ACompiler:TVideCompiler); reintroduce;
    destructor Destroy; override;

    procedure LoadFromFile(const AFileName: NString);
    procedure SaveToFile(const AFileName: NString); override;

    procedure RemoveFile(const AFileName: NString);

    {Adds existing file on disk to project.}
    function AddFile(const AFileName: NString): TVideFile;
    {Adds new file to project.}
    function AddNewFile(const AFileName: NString): TVideFile;
    {Type of this project.}
    property ProjectType: TProjectType read FProjectType write FProjectType;
    {Compiler settings of this project.}
    property CompilerSettings: TVideCompilerSettings read FCompilerSettings;
    property OutputDir: NString read FOutputDir write FOutputDir;
    property Paths:TNStringList read FPaths;
    property Description: NString read FDescription write FDescription;
    property ActiveFile:TVideFile read GetActiveFile;
    {Access to individual project's files by index.}
    property Files[Index: Long]: TVideFile read GetFile; default;
    {Access to individual project's files by Name.}
    property FilesByName[const AName: NString]: TVideFile read GetFileByName;
  end;

  TOnSolutionChange = procedure of object;
  TOnIndexChange = procedure(NewIndex: Long) of object;
  TOnSolutionSaving = procedure(Cfg: TNConfig) of object;
  TOnSolutionLoading = procedure(Cfg: TNConfig) of object;
  TOnGetCompiler = function(const Name:NString):TVideCompiler of object;

  {Item representing Solution - container for projects and
   other files. }
  TVideSolution = class(TVideMultiItem)
  protected
    FDefaultCompiler:TVideCompiler;
    FChanged: Boolean;
    FOnChange: TOnSolutionChange;
    FOnLoading: TOnSolutionLoading;
    FOnSaving: TOnSolutionSaving;
    FOnIndexChange: TOnIndexChange;
    FOnGetCompiler: TOnGetCompiler;
    function GetProject(Index: Long): TVideProject;
    function GetProjectByName(const AName: NString): TVideProject;
    function GetActiveProject: TVideProject;
    procedure SetActiveIndex(Value: Long); override;
    procedure SetChanged(Value: Boolean);
  public
    constructor Create; override;
    destructor Destroy; override;
    procedure NewSolution;

    function AddNewProject(const AName: NString): TVideProject;
    function AddProject(const AFileName: NString): TVideProject;

    procedure RemoveProject(const AFileName: NString);

    procedure LoadFromFile(const AFileName: NString);
    procedure SaveToFile(const AFileName: NString); override;

    property ActiveProject: TVideProject read GetActiveProject;

    {Access to individual projects by index.}
    property Projects[Index: Long]: TVideProject read GetProject; default;
    {Access to individual projects by Name.}
    property ProjectsByName[const AName: NString]: TVideProject read
    GetProjectByName;

    property DefaultCompiler:TVideCompiler read FDefaultCompiler write FDefaultCompiler; 
    property Changed: Boolean read FChanged write SetChanged;
    property OnChange: TOnSolutionChange read FOnChange write FOnChange;
    property OnIndexChange: TOnIndexChange read FOnIndexChange write
      FOnIndexChange;
    property OnLoading: TOnSolutionLoading read FOnLoading write FOnLoading;
    property OnSaving: TOnSolutionSaving read FOnSaving write FOnSaving;
    property OnGetCompiler: TOnGetCompiler read FOnGetCompiler write FOnGetCompiler;
  end;


implementation

uses
  CSingletons, CEditor;

{ TVideCompilerSettings }

constructor TVideCompilerSettings.Create(ACompiler: TVideCompiler);
begin
  inherited Create;
  FCompiler := ACompiler;
  if FCompiler<>nil then
    FCompiler.ApplyDefaults(Self);
end;

procedure TVideCompilerSettings.LoadFromConfig(Cfg: TNConfig);
begin
  FAnsiStrings:=Cfg.ReadBool('AnsiStrings');
  FAssertions:=Cfg.ReadBool('Assertions');
  FBooleanEval:=Cfg.ReadBool('BooleanEval');
  FCmdLine:=Cfg.ReadString('CmdLine');
  FCppExtensions:=Cfg.ReadBool('CppExtensions');
  FDebugInfo:=Cfg.ReadBool('DebugInfo');
  FExtSyntax:=Cfg.ReadBool('ExtSyntax');
  FHints:=Cfg.ReadBool('Hints');
  FIOCheck:=Cfg.ReadBool('IOCheck');
  FMinEnumSize:=Cfg.ReadInt('MinEnumSize');
  FMode:=TCompilerMode(Cfg.ReadInt('Mode'));
  FOverflowCheck:=Cfg.ReadBool('OverflowCheck');
  FRangeCheck:=Cfg.ReadBool('RangeCheck');
  FRecordAlign:=Cfg.ReadInt('RecordAlign');
  FTargetCPU:=TTargetCPU(Cfg.ReadInt('TargetCPU'));
  FTypedPointers:=Cfg.ReadBool('TypedPointers');
  FWarnings:=Cfg.ReadBool('Warnings');
end;

procedure TVideCompilerSettings.SaveToConfig(Cfg: TNConfig);
begin
  Cfg.WriteBool('AnsiStrings', FAnsiStrings);
  Cfg.WriteBool('Assertions', FAssertions);
  Cfg.WriteBool('BooleanEval', FBooleanEval);
  Cfg.WriteString('CmdLine', FCmdLine);
  Cfg.WriteBool('CppExtensions', FCppExtensions);
  Cfg.WriteBool('DebugInfo', FDebugInfo);
  Cfg.WriteBool('ExtSyntax', FExtSyntax);
  Cfg.WriteBool('Hints', FHints);
  Cfg.WriteBool('IOCheck', FIOCheck);
  Cfg.WriteInt('MinEnumSize', FMinEnumSize);
  Cfg.WriteInt('Mode', Long(FMode));
  Cfg.WriteBool('OverflowCheck', FOverflowCheck);
  Cfg.WriteBool('RangeCheck', FRangeCheck);
  Cfg.WriteInt('RecordAlign', FRecordAlign);
  Cfg.WriteInt('TargetCPU', Long(FTargetCPU));
  Cfg.WriteBool('TypedPointers', FTypedPointers);
  Cfg.WriteBool('Warnings', FWarnings);
end;


{*************************************************************************}
{                         TVideFile methods                              }
{*************************************************************************}


constructor TVideFile.Create(const AFileName: NString; AType: TVideFileType;
  ADirty: Boolean);
begin
  inherited Create;
  FileName := AFileName;
  FFileType := AType;
  Dirty := ADirty or (not Vfs.FileExistsOutsideVfs(AFileName));
  FProject := nil;
end;

destructor TVideFile.Destroy;
begin
  inherited Destroy;
end;

procedure TVideFile.Rename(const NewName: NString);
var
  F1, F2: TNBinaryFile;
begin
  if FDirty then
    inherited Rename(NewName)
  else
    if Vfs.FileExists(FFileName) then
    begin
      F1 := Vfs.Open(FFileName, fdRead);
      F2 := Vfs.Open(NewName, fdCreate);
      F2.Stream.CopyFrom(F1.Stream, F1.Size);
      Vfs.Close(F1);
      Vfs.Close(F2);
      Vfs.DeleteFile(FFileName);
      SaveToFile(NewName);
      Solution.Changed := True;
    end;
end;


{*************************************************************************}
{                         TVideProject methods                            }
{*************************************************************************}


constructor TVideProject.Create(ACompiler:TVideCompiler);
begin
  inherited Create;
  FCompilerSettings:=TVideCompilerSettings.Create(ACompiler);
  FPaths := TNStringList.Create;
  FOutputDir := '';
//  FList.Sorted := True;
  FList.Duplicates := dupIgnore;
  Dirty := True;
end;

destructor TVideProject.Destroy;
begin
  FPaths.Free;
  FCompilerSettings.Free;
  inherited Destroy;
end;

function TVideProject.GetFile(Index: Long): TVideFile;
begin
  Result := TVideFile(FList.Items[Index]);
end;

function TVideProject.GetFileByName(const AName: NString): TVideFile;
begin
  Result := TVideFile(FList.Items[FList.IndexOf(AName)]);
end;

function TVideProject.GetActiveFile: TVideFile;
begin
  Result := TVideFile(FList.Items[FActiveIndex]);
end;

procedure TVideProject.LoadFromFile(const AFileName: NString);
var
  Cfg: TNConfig;
  I, Num: Long;
  S, TName:NString;
begin
  Clear;
  TName:=Vfs.GetFullFileName(AFileName, True);
  Log.Add('Loading Project from file: %s', [TName]);
  Cfg := TNConfig.Create(TName, csXML);
  Name := Cfg.ReadString('Name');
  Description := Cfg.ReadString('Description');
  Num := Cfg.ReadInt('Files.Count');
  for I := 0 to Num - 1 do
    AddFile(ExpandFileTo(Cfg.ReadString('Files.Item' + IntToStr(I)),
    ExtractFileDir(TName)));

  OutputDir := Cfg.ReadString('Settings.OutputDir');
  ProjectType:=TProjectType(Cfg.ReadInt('Settings.Type'));
  Num := Cfg.ReadInt('Settings.Paths.Count');
  for I := 0 to Num - 1 do
    FPaths.Add(Cfg.ReadString('Settings.Paths.Item' + IntToStr(I)));

  S:=Cfg.ReadString('Settings.Compiler.UsedCompiler');
  if Solution.OnGetCompiler(S) <> nil then
  begin
    FCompilerSettings.FCompiler:=Solution.OnGetCompiler(S);
    FCompilerSettings.LoadFromConfig(Cfg.FindSubNode('Settings.Compiler'));
  end;

  ActiveIndex:=Cfg.ReadInt('ActiveIndex');
  if ActiveIndex>=Count then
    ActiveIndex := Iff(Count > 0, 0, -1);

  FileName := TName;
  Cfg.Free;
  Solution.Changed := True;
  Modified := False;
  Dirty := False;
end;

procedure TVideProject.SaveToFile(const AFileName: NString);
var
  Cfg: TNConfig;
  I: Long;
begin
  inherited SaveToFile(AFileName);
  Log.Add('Saving Project "%s" to file: %s', [Name, AFileName]);
  Cfg := TNConfig.Create(csXML);
  Cfg.WriteString('Name', Name);
  Cfg.WriteString('Description', FDescription);
  Cfg.WriteInt('ActiveIndex', ActiveIndex);
  Cfg.WriteInt('Files.Count', Count);
  for I := 0 to Count - 1 do
    Cfg.WriteString('Files.Item' + IntToStr(I), ExtractRelativePath(AFileName,
      GetFile(I).FileName));

  Cfg.WriteString('Settings.OutputDir', FOutputDir);
  Cfg.WriteInt('Settings.Type',Long(ProjectType));
  Cfg.WriteInt('Settings.Paths.Count', FPaths.Count);
  for I := 0 to FPaths.Count - 1 do
    Cfg.WriteString('Settings.Paths.Item' + IntToStr(I), FPaths[I]);

  if FCompilerSettings.FCompiler<>nil then
  begin
    Cfg.WriteString('Settings.Compiler.UsedCompiler', FCompilerSettings.FCompiler.Name);
    FCompilerSettings.SaveToConfig(Cfg.FindSubNode('Settings.Compiler'));
  end;

  Cfg.SaveToFile(AFileName);
  Solution.Changed := True;
  Cfg.Free;
end;

procedure TVideProject.RemoveFile(const AFileName: NString);
begin
  if FilesByName[AFileName] <> nil then
  begin
    Delete(IndexOfFileName(AFileName));
    Solution.Changed := True;
    Modified := True;
  end;
end;

function TVideProject.AddFile(const AFileName: NString): TVideFile;
begin
  Result := FilesByName[AFileName];
  if Vfs.FileExists(AFileName) and (Result = nil) then
  begin
    Result := TVideFile.Create(AFileName,
      FileTypes.TypesExt[GetFileExt(AFileName)], False);
    Result.Project := Self;
    Add(Result);
    if Count = 1 then
      ActiveIndex := 0;
    Solution.Changed := True;
    Modified := True;
  end;
end;

function TVideProject.AddNewFile(const AFileName: NString): TVideFile;
begin
  Result := FilesByName[AFileName];
  if Result = nil then
  begin
    Result := TVideFile.Create(AFileName,
      FileTypes.TypesExt[GetFileExt(AFileName)], True);
    Result.Project := Self;
    Add(Result);
    if Count = 1 then
      ActiveIndex := 0;
    Solution.Changed := True;
    Modified := True;
  end;
end;


{*************************************************************************}
{                         TVideSolution methods                              }
{*************************************************************************}


constructor TVideSolution.Create;
begin
  inherited Create;
  Log.Add('Solution created.');
  NewSolution;
end;

destructor TVideSolution.Destroy;
begin
  inherited Destroy;
end;

function TVideSolution.GetProject(Index: Long): TVideProject;
begin
  Result := TVideProject(FList.Items[Index]);
end;

function TVideSolution.GetProjectByName(const AName: NString): TVideProject;
begin
  Result := TVideProject(FList.Items[FList.IndexOf(AName)]);
end;

procedure TVideSolution.NewSolution;
begin
  Clear;
  FileName := 'Solution1.' + sVideSolutionExt;
  Log.Add('Blank Solution created: %s', [Name]);
  ActiveIndex := -1;
//  AddNewProject('Project1.'+sVideProjectExt);
  Dirty := True;
  Changed := True;
  Modified := False;
end;

function TVideSolution.AddNewProject(const AName: NString): TVideProject;
begin
  Result := TVideProject.Create(FDefaultCompiler);
  Result.FileName := AName;
  Result.Dirty := True;
  Add(Result);
  Log.Add('New Project created: %s', [Result.Name]);
  ActiveIndex := Count - 1;
  Changed := True;
  Modified := True;
end;

function TVideSolution.AddProject(const AFileName: NString): TVideProject;
begin
  Result := ProjectsByName[AFileName];
  if Vfs.FileExists(AFileName) and (Result = nil) then
  begin
    Result := TVideProject.Create(nil);
    Result.LoadFromFile(AFileName);
    Add(Result);
    ActiveIndex := Count - 1;
    Changed := True;
    Modified := True;
  end;
end;

procedure TVideSolution.LoadFromFile(const AFileName: NString);
var
  Cfg: TNConfig;
  I, Num: Long;
  Prj: TVideProject;
  PrjName, TName: NString;
begin
  Clear;
  TName:=Vfs.GetFullFileName(AFileName, True);
  Log.Add('Loading Solution from file: %s', [TName]);
  Cfg := TNConfig.Create(TName, csXML);
  Name := Cfg.ReadString('Name');
  Num := Cfg.ReadInt('Projects.Count');
  for I := 0 to Num - 1 do
  begin
    PrjName := ExpandFileTo(Cfg.ReadString('Projects.Item' + IntToStr(I) +
      '.Path'), ExtractFileDir(TName));
    if Vfs.FileExists(PrjName) then
    begin
      Prj := TVideProject.Create(nil);
      Prj.LoadFromFile(PrjName);
      Add(Prj);
    end;
  end;
  ActiveIndex:=Cfg.ReadInt('ActiveIndex');
  if ActiveIndex>=Count then
    ActiveIndex := Iff(Count > 0, 0, -1);
  if Assigned(FOnLoading) then
    FOnLoading(Cfg);
  FileName := TName;
  Modified := False;
  Cfg.Free;
  Changed := True;
  Dirty := False;
end;

procedure TVideSolution.SaveToFile(const AFileName: NString);
var
  Cfg: TNConfig;
  I: Long;
begin
  inherited SaveToFile(AFileName);
  Log.Add('Saving Solution "%s" to file: %s', [Name, AFileName]);
  Cfg := TNConfig.Create(csXML);
  Cfg.WriteString('Name', Name);
  Cfg.WriteInt('ActiveIndex', ActiveIndex);
  Cfg.WriteInt('Projects.Count', Count);
  for I := 0 to Count - 1 do
  begin
    Cfg.WriteString(Format('Projects.Item%d.Name', [I]), GetProject(I).Name);
    Cfg.WriteString(Format('Projects.Item%d.Path', [I]),
      ExtractRelativePath(AFileName, GetProject(I).FileName));
  end;
  if Assigned(FOnSaving) then
    FOnSaving(Cfg);
  Cfg.SaveToFile(AFileName);
  Changed:=True;
  Cfg.Free;
end;

function TVideSolution.GetActiveProject: TVideProject;
begin
  Result := TVideProject(FList.Items[FActiveIndex]);
end;

procedure TVideSolution.SetActiveIndex(Value: Long);
begin
  inherited SetActiveIndex(Value);
  if Assigned(FOnIndexChange) then
    FOnIndexChange(Value);
end;

procedure TVideSolution.SetChanged(Value: Boolean);
begin
  FChanged := Value;
  if Assigned(FOnChange) then
    FOnChange;
end;

procedure TVideSolution.RemoveProject(const AFileName: NString);
begin
  if ProjectsByName[AFileName] <> nil then
  begin
    Delete(IndexOfFileName(AFileName));
    Changed := True;
    Modified := True;
  end;
end;

initialization

  RegisterClasses([TVideCompilerSettings]);

end.

