{*************************************************************************}
{*   TarArchive                                                          *}
{*   Copyright (C) 2003 Marek Mauder, Vampyre SoftWorks                  *}
{*   Engine\Files\ArchiveFormats\Tar\TarArchive.pas                      *}
{*************************************************************************}
{                                                                         }
{ This program is free software; you can redistribute it and/or modify it }
{ under the terms of the GNU General Public License as published by the   }
{ Free Software Foundation; either version 2 of the License, or (at your  }
{ option) any later version. This program is distributed in the hope that }
{ it will be useful, but WITHOUT ANY WARRANTY; without even the implied   }
{ warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See   }
{ the GNU General Public License for more details.                        }
{                                                                         }
{ You should have received a copy of the GNU General Public License       }
{ along with this program; if not, write to the Free Software Foundation, }
{ Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.         }
{                                                                         }
{*************************************************************************}
{  Created On: 12nd November 2003  (12.11.2003)                           }
{*************************************************************************}

{This unit contains class for reading from and writing to Tar archives.}

unit TarArchive;

interface

uses
  SysUtils, Kernel, Objects;

type

  {TNTarArchive can read/write from/to Tar archives. But it can't
   overwrite existing files.}
  TNTarArchive = class(TNBaseObject)
  protected
    FStream: TNFileStream;
    FFiles: TNSortedStringPointers;
    FFileName: NString;
    procedure UpdateDirectory;
  public
    constructor Create(const AFileName: NString); reintroduce;
    destructor Destroy; override;
    procedure GetFileList(Strings: TNStringList);
    function ReadFile(const AFileName: NString; var Data: Pointer;
      var Size: Long):Boolean;
    function WriteFile(const AFileName: NString; Data: Pointer; Size: Long):Boolean;
    function WriteDir(const ADirName: NString):Boolean;
    property Stream: TNFileStream read FStream;
    property Files: TNSortedStringPointers read FFiles;
    property FileName: NString read FFileName;
  end;

const
  RecordSize = 512;
  NameSize = 100;
  TunmLen = 32;
  TgnmLen = 32;
  ChkBlanks = #32#32#32#32#32#32#32#32;
  TarMagic = 'ustar';

type
  TTarHeader = packed record
    Name: array[0..NameSize - 1] of Char;
    Mode: array[0..7] of Char;
    UID: array[0..7] of Char;
    GID: array[0..7] of Char;
    Size: array[0..11] of Char;
    MTime: array[0..11] of Char;
    ChkSum: array[0..7] of Char;
    LinkFlag: Char;
    LinkName: array[0..NameSize - 1] of Char;
    Magic: array[0..7] of Char;
    UName: array[0..TunmLen - 1] of Char;
    GName: array[0..TgnmLen - 1] of Char;
    DevMajor: array[0..7] of Char;
    DevMinor: array[0..7] of Char;
  end;

  TFileEntry = record
    Name: NString;
    Size: Long;
    Offset: Long;
  end;
  PFileEntry = ^TFileEntry;


{*************************************************************************}
{                             IMPLEMEMTATION                              }
{*************************************************************************}


implementation

function Records(Bytes: Int64): Int64;
begin
  Result := Bytes div RecordSize;
  if Bytes mod RecordSize > 0 then
    Inc(Result);
end;

function ExtractNumber64(P: PChar; MaxLen: Long): Int64;
var
  PP: PChar;
  S0: array[0..255] of Char;
  Strg: NString;
begin
  StrLCopy(S0, P, MaxLen);
  Strg := Trim(StrPas(S0));
  PP := PChar(Strg);
  Result := 0;
  while (PP^ <> #32) and (PP^ <> #0) do
  begin
    Result := (Ord(PP^) - Ord('0')) or (Result shl 3);
    Inc(PP);
  end;
end;

// Makes a NString of octal digits
procedure Octal(N: Long; P: PChar; Len: Long);
var
  I: Long;
begin
  for I := Len - 2 downto 0 do
  begin
    (P + I)^ := Chr(Ord('0') + Ord(N and $07));
    N := N shr 3;
  end;
  for I := 0 to Len - 3 do
    if (P + I)^ = '0' then
      (P + I)^ := #32
    else
      Break;
  (P + Len - 1)^ := #32;
end;

// Makes a NString of octal digits
procedure Octal64(N: Int64; P: PChar; Len: Long);
var
  I: Long;
begin
  for I := Len - 2 downto 0 do
  begin
    (P + I)^ := Chr(Ord('0') + Ord(N and $07));
    N := N shr 3;
  end;
  for I := 0 to Len - 3 do
    if (P + I)^ = '0' then
      (P + I)^ := #32
    else
      Break;
  (P + Len - 1)^ := #32;
end;

procedure OctalN(N: Long; P: PChar; Len: Long);
begin
  Octal(N, P, Len - 1);
  (P + Len - 1)^ := #0;
end;


{*************************************************************************}
{                            TTarArchive                                  }
{*************************************************************************}


constructor TNTarArchive.Create(const AFileName: NString);
var
  Rec: array[0..RecordSize - 1] of Char;
begin
  inherited Create;
  FStream := nil;
  FFileName := AFileName;
  FFiles := TNSortedStringPointers.Create;
  try
    if FileExists(AFileName) then
      FStream := TNFileStream.Create(AFileName, fmOpenReadWrite)
    else
    begin
      ZeroMemory(@Rec, RecordSize);
      FStream := TNFileStream.Create(AFileName, fmCreate);
      FStream.WriteBuffer(Rec, RecordSize);
    end;
  except
    on E: Exception do
      ShowError(ErrorStr(Self, 'Create', 'Can not open tar archive "'
        + AFileName + '"!'), E.ClassName);
  end;
  UpdateDirectory;
end;

destructor TNTarArchive.Destroy;
var
  I: Long;
begin
  for I := 0 to FFiles.Count - 1 do
    Dispose(PFileEntry(FFiles.Pointers[I]));
  FStream.Free;
  FFiles.Free;
  inherited Destroy;
end;

procedure TNTarArchive.UpdateDirectory;
var
  Hdr: TTarHeader;
  Rec: array[0..RecordSize - 1] of Char;
  FSize: Long;

  procedure AddFile;
  var
    Entry: PFileEntry;
  begin
    if Hdr.LinkFlag in [#0, '0', '5'] then
    begin
      New(Entry);
      Entry^.Name := Hdr.Name;
      Entry^.Size := FSize;
      Entry^.Offset := FStream.Position;
      FFiles.Add(Entry^.Name, Entry);
    end;
  end;

begin
  if FStream = nil then
    Exit;
  FStream.Position := 0;
  while FStream.Position + RecordSize < FStream.Size do
  begin
    FStream.Read(Rec, RecordSize);
    Move(Rec, Hdr, SizeOf(Hdr));
    if Hdr.Name[0] <> #0 then
    begin
      FSize := ExtractNumber64(@Hdr.Size, 12);
      AddFile;
      FStream.Seek(Records(FSize) * RecordSize, soFromCurrent);
    end;
  end;
end;

procedure TNTarArchive.GetFileList(Strings: TNStringList);
begin
  Strings.Clear;
  Strings.AddStrings(FFiles);
end;

function TNTarArchive.ReadFile(const AFileName: NString; var Data: Pointer;
  var Size: Long):Boolean;
var
  I: Long;
  Entry: PFileEntry;
begin
  Result:=False;
  if FFiles.Find(AFileName, I) then
  begin
    Entry := PFileEntry(FFiles.Pointers[I]);
    Size := Entry^.Size;
    GetMem(Data, Size);
    FStream.Position := Entry^.Offset;
    FStream.ReadBuffer(Data^, Size);
    Result:=True;
  end;
end;

function TNTarArchive.WriteDir(const ADirName: NString):Boolean;
var
  Rec: array[0..RecordSize - 1] of Char;
  I : Long;
  Name:NString;
  Hdr: TTarHeader absolute Rec;
  NullDate, DateTime: TDateTime;
  CheckSum: ULong;
begin
  Result:=False;
  Name:=ADirName;
  if Name[Length(Name)]<>'/' then Name:=Name+'/';
  if FFiles.Find(Name,I) then  Exit;

  ZeroMemory(@Rec, RecordSize);
  StrLCopy(Hdr.Name, PChar(Name), NameSize);
  Hdr.LinkFlag := '5';
  // $1FF - access for everyone
  OctalN($1FF, @Hdr.Mode, 8);
  OctalN(0, @Hdr.UID, 8);
  OctalN(0, @Hdr.GID, 8);
  Octal64(0, @Hdr.Size, 12);
  StrLCopy(Hdr.Magic, PChar(TarMagic + ChkBlanks), 8);

  NullDate := EncodeDate(1970, 1, 1);
  DateTime := GetTime;
  if DateTime >= NullDate then
    Octal(Trunc((DateTime - NullDate) * 86400.0), @Hdr.MTime, 12)
  else
    Octal(Trunc(NullDate * 86400.0), @Hdr.MTime, 12);

  StrMove(Hdr.ChkSum, ChkBlanks, 8);
  CheckSum := 0;
  for I := 0 to SizeOf(TTarHeader) - 1 do
    Inc(CheckSum, Long(Ord(Rec[I])));
  OctalN(CheckSum, @Hdr.ChkSum, 8);

  // rewrite end-of-archive header with new header
  FStream.Position := FStream.Size - RecordSize;
  FStream.Write(Rec, RecordSize);
  // write end-of-archive marker
  ZeroMemory(@Rec, RecordSize);
  FStream.Write(Rec, RecordSize);
  Result:=True;
end;

function TNTarArchive.WriteFile(const AFileName: NString; Data: Pointer;
  Size: Long):Boolean;
var
  Rec: array[0..RecordSize - 1] of Char;
  BytesToRead: Int64;
  BlockSize: Int64;
  I, Pos: Long;
  Hdr: TTarHeader absolute Rec;
  NullDate, DateTime: TDateTime;
  CheckSum: ULong;
begin
  Result:=False;
  // sorry, but I don't know how to rewrite archive without using
  // temporary files (which I hate) or loading
  // almost whole archive to memory (which I dont like)
  if FFiles.Find(AFileName,I) then  Exit;

  ZeroMemory(@Rec, RecordSize);
  StrLCopy(Hdr.Name, PChar(AFileName), NameSize);
  Hdr.LinkFlag := '0';
  // $1FF - access for everyone
  OctalN($1FF, @Hdr.Mode, 8);
  OctalN(0, @Hdr.UID, 8);
  OctalN(0, @Hdr.GID, 8);
  Octal64(Size, @Hdr.Size, 12);
  StrLCopy(Hdr.Magic, PChar(TarMagic + ChkBlanks), 8);

  NullDate := EncodeDate(1970, 1, 1);
  DateTime := GetTime;
  if DateTime >= NullDate then
    Octal(Trunc((DateTime - NullDate) * 86400.0), @Hdr.MTime, 12)
  else
    Octal(Trunc(NullDate * 86400.0), @Hdr.MTime, 12);

  StrMove(Hdr.ChkSum, ChkBlanks, 8);
  CheckSum := 0;
  for I := 0 to SizeOf(TTarHeader) - 1 do
    Inc(CheckSum, Long(Ord(Rec[I])));
  OctalN(CheckSum, @Hdr.ChkSum, 8);
  // rewrite end-of-archive header with new header
  FStream.Position := FStream.Size - RecordSize;
  FStream.Write(Rec, RecordSize);
  // file must be written in 512-byte blocks
  BytesToRead := Size;
  Pos := 0;
  while BytesToRead > 0 do
  begin
    BlockSize := BytesToRead;
    if BlockSize > RecordSize then BlockSize := RecordSize;
    if Pos + BlockSize > Size then BlockSize := Size - Pos;
    ZeroMemory(@Rec, RecordSize);
    Move(Pointer(Long(Data) + Pos)^, Rec, BlockSize);
    FStream.Write(Rec, RecordSize);
    Dec(BytesToRead, BlockSize);
    Inc(Pos, BlockSize);
  end;
  // write end-of-archive marker
  ZeroMemory(@Rec, RecordSize);
  FStream.Write(Rec, RecordSize);
  Result:=True;
end;

end.

