{*************************************************************************}
{*   TarFormat                                                           *}
{*   Copyright (C) 2003 Marek Mauder, Vampyre SoftWorks                  *}
{*   Engine\Files\ArchiveFormats\Tar\TarFormat.pas                       *}
{*************************************************************************}
{                                                                         }
{ This program is free software; you can redistribute it and/or modify it }
{ under the terms of the GNU General Public License as published by the   }
{ Free Software Foundation; either version 2 of the License, or (at your  }
{ option) any later version. This program is distributed in the hope that }
{ it will be useful, but WITHOUT ANY WARRANTY; without even the implied   }
{ warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See   }
{ the GNU General Public License for more details.                        }
{                                                                         }
{ You should have received a copy of the GNU General Public License       }
{ along with this program; if not, write to the Free Software Foundation, }
{ Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.         }
{                                                                         }
{*************************************************************************}
{  Created On: 5th November 2003  (05.11.2003)                            }
{*************************************************************************}

{This unit contains class for accessing TAR file archives.}

unit TarFormat;

{$I Compiler.inc}

interface

uses
  SysUtils, Kernel, Objects, Platforms, Files, TarArchive;

type

  {This archive format supports extracting files from Tar archives.}
  TNTarFormat = class(TNArchiveFormat)
  protected
    function DoOpen(const FileName: NString): TNArchive; override;
    procedure DoClose(Archive: TNArchive); override;
    procedure DoFillFileList(Archive: TNArchive); override;
    function DoReadFile(Archive: TNArchive; const FileName: NString; var Data:
      Pointer; var Size: Long): Boolean; override;
    function DoAddFile(Archive: TNArchive; const FileName: NString; Data:
      Pointer; Size: Long): Boolean; override;
    function DoAddDirectory(Archive: TNArchive; const DirName: NString): Boolean;
      override;
  public
    constructor Create; override;
    function TestFormat(Stream: TNStream): Boolean; override;
  end;

const

  sTarExtensions = 'tar';
  sTarFormatName = 'Tar Archive';
  TarCaps: TArchiveCaps = [acReadFile, acAddFile, acAddDir];

var

  TarInstance: TNTarFormat = nil;


{*************************************************************************}
{                             IMPLEMEMTATION                              }
{*************************************************************************}


implementation


{*************************************************************************}
{                        TNTarFormat methods                              }
{*************************************************************************}


constructor TNTarFormat.Create;
begin
  inherited Create;
  FExtensions := sTarExtensions;
  FFormatName := sTarFormatName;
  FCaps := TarCaps;
end;

function TNTarFormat.DoOpen(const FileName: NString): TNArchive;
var
  Arc: TNTarArchive;
begin
  Result := nil;
  Arc := TNTarArchive.Create(FileName);
  if Arc = nil then
    Exit;
  Result := TNArchive.Create(Self, Arc);
end;

procedure TNTarFormat.DoClose(Archive: TNArchive);
begin
  TNTarArchive(Archive.Handle).Free;
end;

procedure TNTarFormat.DoFillFileList(Archive: TNArchive);
begin
  inherited DoFillFileList(Archive);
  TNTarArchive(Archive.Handle).GetFileList(Archive.Files);
end;

function TNTarFormat.DoReadFile(Archive: TNArchive; const FileName: NString;
 var Data: Pointer; var Size: Long): Boolean;
begin
  Result := TNTarArchive(Archive.Handle).ReadFile(SwapPathDelims(FileName, '/'),
    Data, Size);
end;

function TNTarFormat.DoAddFile(Archive: TNArchive; const FileName: NString;
  Data: Pointer; Size: Long): Boolean;
var
  S: NString;
begin
  // if file is in subdirectory, add subdirectories too. It works
  // without it, but may be somewhere somehow ...
  S := ExtractFileDir(SwapPathDelims(FileName, PathDelim));
  if S <> '' then DoAddDirectory(Archive, S);
  Result := TNTarArchive(Archive.Handle).WriteFile(SwapPathDelims(FileName,
    '/'), Data, Size);
end;

function TNTarFormat.DoAddDirectory(Archive: TNArchive;
  const DirName: NString): Boolean;
var
  S: NString;
begin
  Result:=False;
  // if dir contains subdirectories, add subdirectories too. It works
  // without it, but may be somewhere somehow ...
  S := SwapPathDelims(DirName, PathDelim);
  if S <> '' then
    if S[Length(S)] = PathDelim then S := Copy(S, 1, Length(S) - 1);
  while S <> '' do
  begin
    Result:=TNTarArchive(Archive.Handle).WriteDir(SwapPathDelims(S, '/'))
     and Result;
    S := ExtractFileDir(S);
  end;
end;

function TNTarFormat.TestFormat(Stream: TNStream): Boolean;
var
  OldPos: Int64;
  Magic: Char;
begin
  Result := False;
  if Stream = nil then Exit;
  OldPos := Stream.Position;
  if Stream.Size < 512 then Exit;
  Stream.Read(Magic, SizeOf(Magic));
  Result := Ord(Magic) > 31;
  Stream.Position := OldPos;
end;


{*************************************************************************}
{                         Internal Stuff                                  }
{*************************************************************************}

initialization

  TarInstance := TNTarFormat.Create;

finalization

  FreeAndNil(TarInstance);

end.

