{*************************************************************************}
{*   ZipFormat                                                           *}
{*   Copyright (C) 2003 Marek Mauder, Vampyre SoftWorks                  *}
{*   Engine\Files\ArchiveFormats\Zip\ZipFormat.pas                       *}
{*************************************************************************}
{                                                                         }
{ This program is free software; you can redistribute it and/or modify it }
{ under the terms of the GNU General Public License as published by the   }
{ Free Software Foundation; either version 2 of the License, or (at your  }
{ option) any later version. This program is distributed in the hope that }
{ it will be useful, but WITHOUT ANY WARRANTY; without even the implied   }
{ warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See   }
{ the GNU General Public License for more details.                        }
{                                                                         }
{ You should have received a copy of the GNU General Public License       }
{ along with this program; if not, write to the Free Software Foundation, }
{ Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.         }
{                                                                         }
{*************************************************************************}
{  Created On: 5th November 2003  (05.11.2003)                            }
{*************************************************************************}

{This unit contains class for accessing Zip file archives.}

unit ZipFormat;

{$I Compiler.inc}

interface

uses
  SysUtils, Kernel, Objects, Platforms, Files, Unzip;

type

  {This archive format supports extracting files from Zip archives.}
  TNZipFormat = class(TNArchiveFormat)
  protected
    function DoOpen(const FileName: NString): TNArchive; override;
    procedure DoClose(Archive: TNArchive); override;
    procedure DoFillFileList(Archive: TNArchive); override;
    function DoReadFile(Archive: TNArchive; const FileName: NString; var Data:
      Pointer; var Size: LongInt): Boolean; override;
  public
    constructor Create; override;
    function TestFormat(Stream: TNStream): Boolean; override;
  end;

const

  sZipExtensions = 'zip,pk3,jar,vzf';
  sZipFormatName = 'Zip Archive';
  ZipCaps: TArchiveCaps = [acReadFile];

var

  ZipInstance: TNZipFormat = nil;


{*************************************************************************}
{                             IMPLEMEMTATION                              }
{*************************************************************************}


implementation


{*************************************************************************}
{                        TNZipFormat methods                              }
{*************************************************************************}


constructor TNZipFormat.Create;
begin
  inherited Create;
  FExtensions := sZipExtensions;
  FFormatName := sZipFormatName;
  FCaps := ZipCaps;
end;

function TNZipFormat.DoOpen(const FileName: NString): TNArchive;
var
  UFile: Pointer;
begin
  Result := nil;
  UFile := Unzip.unzOpen(PChar(FileName));
  if UFile = nil then
    Exit;
  Result := TNArchive.Create(Self, UFile);
end;

procedure TNZipFormat.DoClose(Archive: TNArchive);
begin
  if UnzClose(Archive.Handle) <> UNZ_OK then
    raise Exception.Create('Error when closing Zip archive!');
end;

procedure TNZipFormat.DoFillFileList(Archive: TNArchive);
var
  Status: Long;
  FileName: array[0..511] of Char;
begin
  Status := Unzip.UnzGoToFirstFile(Archive.Handle);
  while (Status = UNZ_OK) do
  begin
    if UnzGetCurrentFileInfo(Archive.Handle, nil, FileName, 512, nil, 0,
      nil, 0) = UNZ_OK then
      Archive.Files.Add(FileName);
    Status := UnzGoToNextFile(Archive.Handle);
  end;
end;

function TNZipFormat.DoReadFile(Archive: TNArchive; const FileName: NString;
  var Data: Pointer; var Size: LongInt): Boolean;
var
  I: Long;
  LocRes: Long;
  Found: Boolean;
  Info: unz_file_info;
  OnlyName: NString;
begin
  Result := False;
  LocRes := UnzLocateFile(Archive.Handle, PChar(SwapPathDelims(FileName, '/')),
    0);
  if (LocRes <> UNZ_OK) then
  begin
    Found := False;
    OnlyName := ExtractFileName(FileName);
    for I := 0 to Archive.Files.Count - 1 do
      if PosNoCase(OnlyName, Archive.Files[I], 1) > 0 then
        if UnzLocateFile(Archive.Handle, PChar(Archive.Files[I]), 0) <>
          UNZ_OK then
          Continue
        else
        begin
          Found := True;
          Break;
        end;
    if not Found then
      Exit;
  end;
  if UnzGetCurrentFileInfo(Archive.Handle, @Info, nil, 0, nil, 0, nil, 0) <>
    UNZ_OK then
    Exit;
  if UnzOpenCurrentFile(Archive.Handle) <> UNZ_OK then
    Exit;
  Size := Info.uncompressed_size;
  GetMem(Data, Size);
  if UnzReadCurrentFile(Archive.Handle, Data, Size) <>
    Long(Info.uncompressed_size) then
    Exit;
  if UnzCloseCurrentFile(Archive.Handle) <> UNZ_OK then
    Exit;
end;

function TNZipFormat.TestFormat(Stream: TNStream): Boolean;
var
  OldPos: Int64;
  Magic: array[0..1] of Char;
begin
  Result := False;
  if Stream = nil then
    Exit;
  OldPos := Stream.Position;
  if Stream is TNFileStream then
    Result := Unzip.unzlocal_SearchCentralDir(TNFileStream(Stream)) <> 0
  else
  begin
    ZeroMemory(@Magic, SizeOf(Magic));
    Stream.Read(Magic, SizeOf(Magic));
    Result := (Magic[0] = 'P') and (Magic[1] = 'K')
  end;
  Stream.Position := OldPos;
end;


{*************************************************************************}
{                         Internal Stuff                                  }
{*************************************************************************}


initialization

  ZipInstance := TNZipFormat.Create;

finalization

  FreeAndNil(ZipInstance);

end.

