{*************************************************************************}
{*   RleFormat                                                           *}
{*   Copyright (C) 2003 Marek Mauder, Vampyre SoftWorks                  *}
{*   Engine\Files\Compressions\Rle\RleFormat.pas                         *}
{*************************************************************************}
{                                                                         }
{ This program is free software; you can redistribute it and/or modify it }
{ under the terms of the GNU General Public License as published by the   }
{ Free Software Foundation; either version 2 of the License, or (at your  }
{ option) any later version. This program is distributed in the hope that }
{ it will be useful, but WITHOUT ANY WARRANTY; without even the implied   }
{ warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See   }
{ the GNU General Public License for more details.                        }
{                                                                         }
{ You should have received a copy of the GNU General Public License       }
{ along with this program; if not, write to the Free Software Foundation, }
{ Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.         }
{                                                                         }
{*************************************************************************}
{  Created On: 7th November 2003  (07.11.2003)                            }
{*************************************************************************}

{This unit contains class for compress/decompress data using simple
 Run Length Encoding (RLE).}

unit RleFormat;

{$I 'Compiler.inc'}

interface

uses
  Kernel, Objects, Platforms, Files;

type

  {This compression format provides RLE compression/decompression.}
  TNRleFormat = class(TNCompression)
  protected
    procedure DoCompress(InData: Pointer; InSize: Long; var OutData: Pointer; var
      OutSize: Long); override;
    procedure DoDecompress(InData: Pointer; InSize: Long; var OutData: Pointer;
      var OutSize: Long); override;
  public
    constructor Create; override;
  end;

const

  sRleExtensions = 'rle,r,tr';
  sRleFormatName = 'Run Length Encoding compression/decompression';

var

  RleInstance: TNRleFormat = nil;


{*************************************************************************}
{                         IMPLEMEMTATION                                  }
{*************************************************************************}


implementation


{*************************************************************************}
{                        TNRleFormat methods                              }
{*************************************************************************}


constructor TNRleFormat.Create;
begin
  inherited Create;
  FExtensions := sRleExtensions;
  FFormatName := sRleFormatName;
  FCanLoad := True;
  FCanSave := True;
end;

procedure TNRleFormat.DoCompress(InData: Pointer; InSize: Long;
  var OutData: Pointer; var OutSize: Long);
var
  Rle, B: Byte;
  Pos, D, S, CByte, Step: Long;

  function CountSame: Long;
  var
    I: Long;
  begin
    I := CByte;
    Result := 1;
    B := PByte(Long(InData) + I)^;
    while I < InSize - 1 do
    begin
      I := I + 1;
      if (B = PByte(Long(InData) + I)^) and
        // we can have max 127 count
      (Result < 127) then
        Result := Result + 1
      else
        Break;
    end;
  end;

  function CountDiff: Long;
  var
    I: Long;
  begin
    I := CByte;
    Result := 0;
    while I < InSize do
    begin
      B := PByte(Long(InData) + I)^;
      I := I + 1;
      // Adding  or(I=InSize) is very important. Otherwise it adds extra byte
      // on decompression because CountDiff returns 0 but it should be 1
      if ((B <> PByte(Long(InData) + I)^) or (I = InSize)) and
      // we can have max 127 count
      (Result < 127) then
        Result := Result + 1
      else
        Break;
    end;
  end;

begin
  CByte := 0;
  Step := InSize shr 3;
  OutSize := Step;
  Pos := 0;
  GetMem(OutData, OutSize);
  while CByte < InSize do
  begin
    D := CountDiff;
    S := CountSame;
    if D > 0 then
    begin
      Rle := D - 1;
      if Pos + D + 1 >= OutSize then
      begin
        OutSize := OutSize + Step;
        ReallocMem(OutData, OutSize);
      end;
      PByte(Long(OutData) + Pos)^ := Rle;
      Pos := Pos + 1;
      Move(Pointer(Long(InData) + CByte)^, Pointer(Long(OutData) + Pos)^, D);
      Pos := Pos + D;
      CByte := CByte + D;
    end
    else
      if S > 1 then
      begin
        Rle := S + 127;
        if Pos + 2 >= OutSize then
        begin
          OutSize := OutSize + Step;
          ReallocMem(OutData, OutSize);
        end;
        PByte(Long(OutData) + Pos)^ := Rle;
        Pos := Pos + 1;
        PByte(Long(OutData) + Pos)^ := PByte(Long(InData) + CByte)^;
        Pos := Pos + 1;
        CByte := CByte + S;
      end;
  end;
  OutSize := Pos;
  ReallocMem(OutData, OutSize);
end;

procedure TNRleFormat.DoDecompress(InData: Pointer; InSize: Long;
  var OutData: Pointer; var OutSize: Long);
var
  Step, I, CByte, Pos: Long;
  B, Rle: Byte;
begin
  Step := InSize shr 2;
  OutSize := Step;
  GetMem(OutData, OutSize);
  Pos := 0;
  CByte := 0;
  while CByte < InSize do
  begin
    Rle := PByte(Long(InData) + CByte)^;
    CByte := CByte + 1;
    if Rle < 128 then
    begin
      Rle := Rle + 1;
      if Pos + Rle >= OutSize then
      begin
        OutSize := OutSize + Step;
        ReallocMem(OutData, OutSize);
      end;
      Move(PByte(Long(InData) + CByte)^, PByte(Long(OutData) + Pos)^, Rle);
      CByte := CByte + Rle;
      Pos := Pos + Rle;
    end
    else
    begin
      Rle := Rle - 127;
      B := PByte(Long(InData) + CByte)^;
      CByte := CByte + 1;
      if Pos + Rle >= OutSize then
      begin
        OutSize := OutSize + Step;
        ReallocMem(OutData, OutSize);
      end;
      for I := 0 to Rle - 1 do
      begin
        PByte(Long(OutData) + Pos)^ := B;
        Pos := Pos + 1;
      end;
    end;
  end;
  OutSize := Pos;
  ReallocMem(OutData, OutSize);
end;


{*************************************************************************}
{                         Internal Stuff                                  }
{*************************************************************************}


initialization

  RleInstance := TNRleFormat.Create;

finalization

  FreeAndNil(RleInstance);

end.

