{*************************************************************************}
{*   Configs                                                             *}
{*   Copyright (C) 2004 Marek Mauder, Vampyre SoftWorks                  *}
{*   Engine\Files\Configs.pas                                            *}
{*************************************************************************}
{                                                                         }
{ This program is free software; you can redistribute it and/or modify it }
{ under the terms of the GNU General Public License as published by the   }
{ Free Software Foundation; either version 2 of the License, or (at your  }
{ option) any later version. This program is distributed in the hope that }
{ it will be useful, but WITHOUT ANY WARRANTY; without even the implied   }
{ warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See   }
{ the GNU General Public License for more details.                        }
{                                                                         }
{ You should have received a copy of the GNU General Public License       }
{ along with this program; if not, write to the Free Software Foundation, }
{ Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.         }
{                                                                         }
{*************************************************************************}
{  Created On: 1st March 2004  (01.03.2004)                               }
{*************************************************************************}

{This unit contains classes for storage and management of configuration data.}

unit Configs;

{ TODOS:
  * Config links as in CS's cfg ...  System.VFS = vfs.cfg.
  - Comments stay in saved configs.
  * Implement XML config driver.
  - It is persistent so add streaming.
}

{$I Compiler.inc}

interface

uses
  SysUtils, Kernel, Objects, Platforms, Files, Lists, Dom;

const
  {Char used as delimeter of keys in config path. Ex: Video.Resolution.Width
   (where '.' is sConfigPathDelim).}
  sConfigPathDelim = '.';
  {If this is found as value when loading config form file, %s is taken
   as name of another config file which is then loaded and assigned
   to key which has CfgLoad as value. Example: You can have config with line
   Root.Key = CfgLoad(another.cfg) - another.cfg will be loaded and keys in
   it will become subkeys of Root.Key.}
  sCfgLoad = 'CfgLoad';
  {Format used when parsing CfgLoad command.}
  sCfgLoadFmt = sCfgLoad + '(%s)';

  {Default extension of text config files.}
  sTextConfigExt = 'cfg';
  {Char used in text config files to separate keys from values. Ex:
   Video.Resolution.Width=800 (where '=' is delim).}
  sTextConfigValueDelim = '=';
  {Chars used to mark line in text config file as comment.}
  TextConfigCommnets = ['/', ';', '{'];

  {Default extension of XML config files.}
  sXMLConfigExt = 'xml';
  {XML document version used by XML config driver.}
  sXVersion = '1.0';
  {XML document encoding used by XML config driver.}
  sXEncoding = 'utf-8';
  {Root node tag used by XML config driver.}
  sXConfig = 'Config';
  {Name of tag in which vaules are stored by XML config driver.}
  sXValue = 'Value';

type

  {Defines format in which TNConfigFile will be saved:
    csText - config will be saved as text file.
    csXML  - config will be saved as XML document.}
  TConfigStorage = (csText, csXML);

  {Mode in which TNConfigDriver works.}
  TDriverMode = (dmRead, dmWrite);

  {Abstract class used by TNConfigFile to do real saving/loading
   of config data.}
  TNAbstractConfigDriver = class(TNSingleton)
  protected
    FStream: TNStream;
    FMode: TDriverMode;
    procedure Initialize; virtual; abstract;
  public
    {Creates new config driver using AStream in AMode mode.}
    constructor Create(AStream: TNStream; AMode: TDriverMode); reintroduce;
      overload;
    destructor Destroy; override;
    {Starts new section called NodeName.}
    procedure BeginNode(const NodeName: NString); virtual; abstract;
    {Ends current section.}
    procedure EndNode; virtual; abstract;
    {Reads next key and value from stream and moves position in
     source data to next key+value.}
    procedure ReadValue(var Key, Value: NString); virtual; abstract;
    {Writes Key and Value to actual position in stream.}
    procedure WriteValue(const Key, Value: NString); virtual; abstract;
    {Returns True if there is no further config data in stream.}
    function IsEnd: Boolean; virtual; abstract;
    {In dmRead mode this is called in constructor, in dmWrite mode
     it is called in destructor. It can be used to preload/writeall
     data from/to stream.}
    procedure Flush; virtual; abstract;
  end;

  {Class refernce for TNAbstractConfigDriver.}
  TNConfigDriverClass = class of TNAbstractConfigDriver;

  {Config driver operating on text files. Files must be in
   'Key1.Key2.Key3.Key4=Value' format.
   Comments are allowed: they must start with one of TextConfigCommnets
   chars or simply do not containt '=' character.
   Line is broken to Key and Value at position of first '=' char
   so Value can contain additional '=' chars.
   All leading and trailing spaces and control characters are trimmed from
   string (by SysUtils.Trim()).
   Warning: input stream is read from actual pos to end in dmRead mode
   so this driver is not "stream safe".
   Example:
     Video.Resolution.Width=1024
     // this is config comment
     Games.RGP.Best = Daggerfall
     Character.Name=Tiber Septim}
  TNTextConfigDriver = class(TNAbstractConfigDriver)
  protected
    FParent: NString;
    FBuffer: NString;
    procedure Initialize; override;
  public
    procedure BeginNode(const NodeName: NString); override;
    procedure EndNode; override;
    procedure ReadValue(var Key, Value: NString); override;
    procedure WriteValue(const Key, Value: NString); override;
    function IsEnd: Boolean; override;
    procedure Flush; override;
  end;

  {Config driver operating on XML documents. Values are stored in UTF8
   encoding.}
  TNXMLConfigDriver = class(TNAbstractConfigDriver)
  protected
    FDoc: TXMLDocument;
    FLocal: TDOMElement;
    FNode: TDOMElement;
    FText: TDOMText;
    FParent: NString;
    procedure Initialize; override;
  public
    destructor Destroy; override;
    procedure BeginNode(const NodeName: NString); override;
    procedure EndNode; override;
    procedure ReadValue(var Key, Value: NString); override;
    procedure WriteValue(const Key, Value: NString); override;
    function IsEnd: Boolean; override;
    procedure Flush; override;
  end;

  {Class for storage configuration data. Each TNConfig can have
   other TNConfigs as children so large hierarchies can be
   created. All values are stored as strings here.
   If you want to access some value in config file you must use
   string called Key. Each Key can contain more subkeys separated
   by sConfigPathDelim (usualy '.'). Each subkey refers to
   TNConfigFile deeper in hierarchy. Last subkey in Key is
   name of value owned by TNConfig refered by previous subkey.
   So if you have Key 'Raiders.Germans.Vizigoths' then
   'Raiders' is child (node) of current config. 'Raiders' has
   child called 'Germans' and 'Germans' has value called 'Vizigoths'.}
  TNConfig = class(TNNamedPersistent)
  protected
    FNodes: TNSortedStringObjects;
    FValues: TNSortedDictionary;
    FStorage: TConfigStorage;
    function GetNode(Index: Long): TNConfig;
    function GetNodeCount: Long;
    function GetValueCount: Long;
    function GetValue(Index: Long): NString;
    procedure SetValue(Index: Long; const Value: NString);
    function GetValueKey(Index: Long): NString;
    procedure SetStorage(Value: TConfigStorage);
    {Checks if Value string contains reference to another config file.
     If True, name of config extracted from Value is returned in CfgFile.}
    function CheckForCfgLoad(const Value: NString; var CfgFile: NString):
      Boolean;
    {Adds new node called NodeName (and creates necessary
     subnodes in path). Returns created node or node found
     in NodePath.}
    function AddNode(const NodeName: NString): TNConfig;
    {Saves this node using given config driver.}
    procedure SaveAsNode(Driver: TNAbstractConfigDriver);
  public
    constructor Create; overload; override;
    {Calls Create() and sets Storage to AStorage.}
    constructor Create(AStorage: TConfigStorage); reintroduce; overload;
    {Calls Create() constructor and then calls LoadFromFile(AFileName).
     AStorage identifies in which format is AFileName config (used if
     AFileName's extension is not one of known).}
    constructor Create(const AFileName: NString;
      AStorage: TConfigStorage = csText); reintroduce; overload;
    destructor Destroy; override;

    {Reads string value and if successful returns it. Otherwise returns Default.}
    function ReadString(const Key: NString; const Default: NString = ''):
      NString;
    {Reads boolean value and if successful returns it. Value can be stored
     in config either as number (0,1) or string (True, False). On failure returns
     Default.}
    function ReadBool(const Key: NString; Default: Boolean = False): Boolean;
    {Reads integer value and if successful returns it. Otherwise returns Default.}
    function ReadInt(const Key: NString; const Default: Int64 = 0): Int64;
    {Reads float value and if successful returns it. Otherwise returns Default.}
    function ReadFloat(const Key: NString; const Default: Float = 0.0): Float;

    {Writes string value to config.}
    procedure WriteString(const Key, Value: NString);
    {Writes boolean value to config.}
    procedure WriteBool(const Key: NString; Value: Boolean);
    {Writes integer value to config.}
    procedure WriteInt(const Key: NString; const Value: Int64);
    {Writes float value to config.}
    procedure WriteFloat(const Key: NString; const Value: Float);

    {Adds given Config as node of current TNConfigFile. Config's
     name (path used in Key) is replaced by NewName.
     If node with NewName already exists it is merged with Config
     and then deleted.
     Also Config's storage is changed to reflect Self's Storage.
     This file adds reference to Config to nodes it does not create
     copy of Config (so do not Destroy Config after Add).
     If you want to copy contents of Config use AddConfig(Config.Clone)
     instead.}
    procedure AddConfig(const NewName: NString; Config: TNConfig); overload;
    {Same as above but config is loaded from file. If some error occurs
     nothing is added. AStorage is used when Config's extension is
     not one of known.}
    procedure AddConfig(const NewName: NString; const Config: NString;
      AStorage: TConfigStorage = csText); overload;
    {Copies all values and nodes from Config (all nodes are cloned).}
    procedure CopyFromConfig(Config: TNConfig); overload;
    {Same as above, but config which is being copied is loaded from file.
     AStorage is used when Config's extension is not one of known.}
    procedure CopyFromConfig(const Config: NString;
      AStorage: TConfigStorage = csText); overload;
    {Creates new instance of TNConfigFile and copies all values and nodes
     to it.}
    function Clone: TNConfig;
    {Deletes all configuration data of this node.}
    procedure Clear;
    {Returns subnode in node hierarchy defined by NodeName
     (it can contain subkeys). If node is not found on given path nil
     is returned.}
    function FindSubNode(const NodeName: NString): TNConfig;

    {Loads new configuration data from file. Current data are deleted.
     If FileName's extension is one of known appropriate config driver
     is used but Storage is set to original value after loading.}
    procedure LoadFromFile(const FileName: NString);
    {Loads new configuration data from stream. Current data are deleted.}
    procedure LoadFromStream(Stream: TNStream);
    {Saves actual configuration data to file.}
    procedure SaveToFile(const FileName: NString);
    {Saves actual configuration data to stream.}
    procedure SaveToStream(Stream: TNStream);

    {Returns number of this config's nodes.}
    property NodeCount: Long read GetNodeCount;
    {Returns node at given index.}
    property Nodes[Index: Long]: TNConfig read GetNode;
    {Returns number of valus of this config.}
    property ValueCount: Long read GetValueCount;
    {Returns Key of value at given index.}
    property ValueKeys[Index: Long]: NString read GetValueKey;
    {Returns value at given index as NString (all values as internally
     stored as strings).}
    property Values[Index: Long]: NString read GetValue write SetValue;
    {Defines how this config saves/loads its data. Note that
     if you load config saved with different Storage it will fail or crash.
     If you set Storage it is set also for subnodes.}
    property Storage: TConfigStorage read FStorage write SetStorage;
  end;

  {For backward compatibility.}
  TNConfigFile = TNConfig;

{Returns True if given extension is supported by one of config drivers.}
function IsSupportedExt(const Ext: NString): Boolean;
{Returns TConfigStorage type according to given extension (without dot).}
function GetStorageByExt(const Ext: NString): TConfigStorage;
{Returns class reference of config driver defined by given Storage.}
function GetDriverClass(Storage: TConfigStorage): TNConfigDriverClass;


{*************************************************************************}
{                             IMPLEMEMTATION                              }
{*************************************************************************}


implementation


function IsSupportedExt(const Ext: NString): Boolean;
begin
  Result := SameStr(Ext, sTextConfigExt) or SameStr(Ext, sXMLConfigExt);
end;

function GetStorageByExt(const Ext: NString): TConfigStorage;
begin
  if SameStr(Ext, sXMLConfigExt) then
    Result := csXML
  else
    Result := csText;
end;

function GetDriverClass(Storage: TConfigStorage): TNConfigDriverClass;
begin
  Result := nil;
  case Storage of
    csText: Result := TNTextConfigDriver;
    csXML: Result := TNXMLConfigDriver;
  end;
end;


{*************************************************************************}
{                          TNConfigDriver methods                         }
{*************************************************************************}


constructor TNAbstractConfigDriver.Create(AStream: TNStream; AMode:
  TDriverMode);
begin
  Create;
  FMode := AMode;
  FStream := AStream;
  Initialize;
end;

destructor TNAbstractConfigDriver.Destroy;
begin
  inherited Destroy;
end;


{*************************************************************************}
{                          TNTextConfigDriver methods                     }
{*************************************************************************}


procedure TNTextConfigDriver.Initialize;
begin
  FBuffer := '';
  FParent := '';
end;

procedure TNTextConfigDriver.BeginNode(const NodeName: NString);
begin
  if (NodeName <> '') and (FParent <> '') then
    FParent := FParent + sConfigPathDelim + NodeName
  else
    FParent := NodeName;
end;

procedure TNTextConfigDriver.EndNode;
begin
  StrTokenEnd(FParent, sConfigPathDelim);
end;

procedure TNTextConfigDriver.ReadValue(var Key, Value: NString);
var
  Start, P: PChar;
  I: Long;
  S: NString;
begin
  Key := '';
  Value := '';
  P := Pointer(FBuffer);
  if P <> nil then
  begin
    Start := P;
    // inc P to get to end of line
    while not (P^ in [#0, #10, #13]) do
      Inc(P);
    SetString(S, Start, P - Start);
    // inc P to get to start of new line
    while (P^ in [#10, #13]) do
      Inc(P);
    Delete(FBuffer, 1, P - Start);
    // check if S is valid line of type Key = Value
    if S <> '' then
    begin
      I := PosNoCase(sTextConfigValueDelim, S);
      if not ((S[1] in TextConfigCommnets) or (I = 0)) then
      begin
        Key := Trim(Copy(S, 1, I - 1));
        Value := Trim(Copy(S, I + 1, MaxLong));
      end;
    end;
  end;
end;

procedure TNTextConfigDriver.WriteValue(const Key, Value: NString);
var
  S: NString;
begin
  if FParent <> '' then
    S := FParent + sConfigPathDelim + Key
  else
    S := Key;
  FBuffer := FBuffer + S + sTextConfigValueDelim + Value + sLineBreak;
end;

function TNTextConfigDriver.IsEnd: Boolean;
begin
  Result := FBuffer = '';
end;

procedure TNTextConfigDriver.Flush;
var
  Size: Long;
begin
  case FMode of
    dmWrite:
      begin
        FStream.Write(Pointer(FBuffer)^, Length(FBuffer));
        FBuffer := '';
      end;
    dmRead:
      begin
        Size := FStream.Size - FStream.Position;
        SetString(FBuffer, nil, Size);
        FStream.Read(Pointer(FBuffer)^, Size);
      end;
  end;
end;


{*************************************************************************}
{                          TNXMLConfigDriver methods                      }
{*************************************************************************}


procedure TNXMLConfigDriver.Initialize;
begin
  FParent := '';
  if FMode = dmWrite then
  begin
    FDoc := TXMLDocument.Create;
    FDoc.XMLVersion := sXVersion;
    FDoc.Encoding := sXEncoding;
    FDoc.WriteHeaders := True;
    FLocal := FDoc.CreateElement(sXConfig);
    FDoc.AppendChild(FLocal);
  end;
end;

destructor TNXMLConfigDriver.Destroy;
begin
  FLocal := nil;
  FNode := nil;
  FDoc.Free;
  inherited Destroy;
end;

procedure TNXMLConfigDriver.BeginNode(const NodeName: NString);
begin
  if (NodeName <> '') then
  begin
    FNode := FDoc.CreateElement(NodeName);
    FLocal.AppendChild(FNode);
    FLocal := TDOMElement(FNode);
  end;
end;

procedure TNXMLConfigDriver.EndNode;
begin
  if FLocal.ParentNode <> nil then
    FLocal := TDOMElement(FLocal.ParentNode);
end;

procedure TNXMLConfigDriver.ReadValue(var Key, Value: NString);

  function GetNextNode(Node: TDOMNode): TDOMElement;
  begin
    Result := TDOMElement(Node);
    if Node <> nil then
    begin
      // check if local node has siblings, if it has
      // set next sibling as result
      if Node.NextSibling <> nil then
        Result := TDOMElement(Node.NextSibling)
      else
      begin
        // check if we are not at the end of document
        if Node <> FDoc.DocumentElement then
        begin
          // check for new local node at parent level
          Result := GetNextNode(Node.ParentNode);
          StrTokenEnd(FParent, sConfigPathDelim);
        end;
      end;
    end;
  end;

  function GetKeyPath: NString;
  begin
    if FParent <> '' then
      Result := FParent + sConfigPathDelim + FLocal.TagName
    else
      Result := FLocal.TagName;
  end;

begin
  Key := '';
  Value := '';
  while not IsEnd do
  begin
    if (FLocal.FirstChild <> nil) then
    begin
      // local node has children
      if (FLocal.FirstChild is TDOMText) then
      begin
        // local node's child is Text Node so
        // return its contents as Value and local's tag as Key
        Key := GetKeyPath;
        Value := UTF8ToAnsi(FLocal.FirstChild.NodeValue);
        // set next local node
        FLocal := GetNextNode(FLocal);
        Break;
      end
      else
      begin
        // local node has children but not Text Nodes, so
        // set first child as new local node
        FParent := GetKeyPath;
        FLocal := TDOMElement(FLocal.FirstChild);
      end;
    end
    else
      // local node has no children, so look for siblings or
      // parent's siblings
      FLocal := GetNextNode(FLocal);
  end;
end;

procedure TNXMLConfigDriver.WriteValue(const Key, Value: NString);
begin
  // values are stored as Text Nodes of elements with tag names
  // from Keys
  FNode := FDoc.CreateElement(Key);
  FText := FDoc.CreateTextNode(AnsiToUTF8(Value));
  FNode.AppendChild(FText);
  FLocal.AppendChild(FNode);
end;

function TNXMLConfigDriver.IsEnd: Boolean;
begin
  Result := (FLocal = FDoc.DocumentElement);
end;

procedure TNXMLConfigDriver.Flush;
begin
  case FMode of
    dmWrite: WriteXMLFile(FDoc, FStream);
    dmRead:
      begin
        ReadXMLFile(FDoc, FStream);
        if FDoc.DocumentElement.FirstChild <> nil then
          FLocal := TDOMElement(FDoc.DocumentElement.FirstChild)
        else
          FLocal := TDOMElement(FDoc.DocumentElement);
      end;
  end;
end;


{*************************************************************************}
{                           TNConfigFile methods                          }
{*************************************************************************}


constructor TNConfig.Create;
begin
  inherited Create;
  FNodes := TNSortedStringObjects.Create;
  FValues := TNSortedDictionary.Create;
  FName := '';
end;

constructor TNConfig.Create(AStorage: TConfigStorage);
begin
  Create;
  FStorage := AStorage;
end;

constructor TNConfig.Create(const AFileName: NString; AStorage: TConfigStorage);
begin
  Create(AStorage);
  LoadFromFile(AFileName);
end;

destructor TNConfig.Destroy;
begin
  Clear;
  FNodes.Free;
  FValues.Free;
  inherited Destroy;
end;

function TNConfig.GetNodeCount: Long;
begin
  Result := FNodes.Count;
end;

function TNConfig.GetNode(Index: Long): TNConfig;
begin
  Result := TNConfig(FNodes.Objects[Index]);
end;

function TNConfig.GetValueCount: Long;
begin
  Result := FValues.Count;
end;

function TNConfig.GetValue(Index: Long): NString;
begin
  Result := FValues.Words[Index];
end;

procedure TNConfig.SetValue(Index: Long; const Value: NString);
begin
  FValues.Words[Index] := Value;
end;

function TNConfig.GetValueKey(Index: Long): NString;
begin
  Result := FValues.Keys[Index];
end;

procedure TNConfig.SetStorage(Value: TConfigStorage);
var
  I: Long;
begin
  FStorage := Value;
  for I := 0 to FNodes.Count - 1 do
    TNConfig(FNodes.Objects[I]).Storage := Value;
end;

function TNConfig.CheckForCfgLoad(const Value: NString;
  var CfgFile: NString): Boolean;
var
  I, J: Long;
begin
  CfgFile := '';
  I := PosNoCase(sCfgLoad, Value);
  Result := I > 0;
  if Result then
  begin
    I := PosNoCase('(', Value);
    J := PosNoCase(')', Value, I + 1);
    if (I = 0) or (J = 0) then
    begin
      Result := False;
      Exit;
    end;
    CfgFile := Trim(Copy(Value, I + 1, J - I - 1));
  end;
end;

function TNConfig.AddNode(const NodeName: NString): TNConfig;
var
  I: Long;
  SubName, SubNode: NString;
  Node: TNConfig;
begin
  Result := FindSubNode(NodeName);
  if Result = nil then
  begin
    I := PosNoCase(sConfigPathDelim, NodeName);
    if I > 0 then
    begin
      // NodeName contains subnodes (error)
      SubName := Copy(NodeName, 1, I - 1);
      SubNode := Copy(NodeName, I + 1, MaxLong);
      Node := FindSubNode(SubName);
      if Node = nil then
      begin
        Node := TNConfig.Create;
        Node.Name := SubName;
        FNodes.Add(SubName, Node);
      end;
      Result := Node.AddNode(SubNode);
    end
    else
    begin
      // NodeName will belong to Self
      Result := TNConfig.Create;
      Result.Name := NodeName;
      FNodes.Add(NodeName, Result);
    end;
  end;
end;

procedure TNConfig.SaveAsNode(Driver: TNAbstractConfigDriver);
var
  I: Long;
begin
  Driver.BeginNode(Name);
  for I := 0 to FValues.Count - 1 do
    Driver.WriteValue(FValues.Keys[I], FValues.Words[I]);
  for I := 0 to FNodes.Count - 1 do
    TNConfig(FNodes.Objects[I]).SaveAsNode(Driver);
  Driver.EndNode;
end;

function TNConfig.ReadString(const Key, Default: NString): NString;
var
  I: Long;
  SubKey, NodeName: NString;
begin
  Result := Default;
  I := PosNoCase(sConfigPathDelim, Key);
  if I > 0 then
  begin
    // search for value in subnodes
    NodeName := Copy(Key, 1, I - 1);
    SubKey := Copy(Key, I + 1, MaxLong);
    if FNodes.Find(NodeName, I) then
      Result := TNConfig(FNodes.Objects[I]).ReadString(SubKey, Default);
  end
  else
    // value should be in this TNConfig instance
    if FValues.Find(Key, I) then
      Result := FValues.Words[I];
end;

function TNConfig.ReadBool(const Key: NString; Default: Boolean): Boolean;
var
  S: NString;
  I: Long;
begin
  S := ReadString(Key);
  Result := TryStrToInt(S, I);
  if Result then
    Result := I <> 0
  else
  begin
    Result := SameStr(S, sTrue);
    if not Result then
    begin
      Result := not SameStr(S, sFalse);
      if Result then
        Result := Default;
    end;
  end;
end;

function TNConfig.ReadInt(const Key: NString; const Default: Int64): Int64;
var
  S: NString;
begin
  S := ReadString(Key);
  if (Length(S) > 2) and (S[1] = '0') and ((S[2] = 'X') or (S[2] = 'x')) then
    S := '$' + Copy(S, 3, MaxLong);
  Result := StrToInt64Def(S, Default);
end;

function TNConfig.ReadFloat(const Key: NString; const Default: Float): Float;
begin
  Result := StrToFloatDef(ReadString(Key), Default);
end;

procedure TNConfig.WriteString(const Key, Value: NString);
var
  I: Long;
  SubKey, NodeName: NString;
begin
  if Key <> '' then
  begin
    I := PosNoCase(sConfigPathDelim, Key);
    if I > 0 then
    begin
      // Value belongs to some subnode
      NodeName := Copy(Key, 1, I - 1);
      SubKey := Copy(Key, I + 1, MaxLong);
      if FNodes.Find(NodeName, I) then
      begin
        // write NString to existing subnode
        TNConfig(FNodes.Objects[I]).WriteString(SubKey, Value);
      end
      else
      begin
        // node in key path does not exist so create one and write
        AddNode(NodeName);
        if FNodes.Find(NodeName, I) then
          TNConfig(FNodes.Objects[I]).WriteString(SubKey, Value);
      end;
    end
    else
    begin
      // Value belongs to this TNConfig instance
      if FValues.Find(Key, I) then
        // value already exists so overwrite it
        FValues.Words[I] := Value
      else
        // value does not exist so create it
        FValues.Add(Key, Value);
    end;
  end;
end;

procedure TNConfig.WriteBool(const Key: NString; Value: Boolean);
begin
  WriteString(Key, Iff(Value, sTrue, sFalse));
end;

procedure TNConfig.WriteInt(const Key: NString; const Value: Int64);
begin
  WriteString(Key, IntToStr(Value));
end;

procedure TNConfig.WriteFloat(const Key: NString; const Value: Float);
begin
{$IFDEF FLOATSINGLE}
  WriteString(Key, Format('%.4f', [Value]));
{$ELSE}
  WriteString(Key, Format('%.10f', [Value]));
{$ENDIF}
end;

procedure TNConfig.AddConfig(const NewName: NString; Config: TNConfig);
var
  Node: TNConfig;
begin
  if Config <> nil then
  begin
    Config.Storage := Storage;
    Config.Name := NewName;
    if Config.Name = '' then
      Config.Name := IntToStr(Long(Config));
    // if node with name same as NewName exists then
    // merge it with Config and delete it
    Node := FindSubNode(Config.Name);
    if Node <> nil then
    begin
      FNodes.Delete(FNodes.IndexOf(Config.Name));
      Config.CopyFromConfig(Node);
      Node.Free;
    end;
    // adds Config to nodes so do not Free it now!
    FNodes.Add(Config.Name, Config);
  end;
end;

procedure TNConfig.AddConfig(const NewName, Config: NString;
  AStorage: TConfigStorage);
begin
  AddConfig(NewName, TNConfig.Create(Config, AStorage));
end;

procedure TNConfig.CopyFromConfig(Config: TNConfig);
var
  I: Long;
begin
  if Config <> nil then
  begin
    // adds all Config's values
    for I := 0 to Config.FValues.Count - 1 do
      FValues.Add(Config.FValues.Keys[I], Config.FValues.Words[I]);
    // adds all Config's cloned nodes (recursively)
    for I := 0 to Config.FNodes.Count - 1 do
      AddConfig(TNConfig(Config.FNodes.Objects[I]).Name,
        TNConfig(Config.FNodes.Objects[I]).Clone);
  end;
end;

procedure TNConfig.CopyFromConfig(const Config: NString;
  AStorage: TConfigStorage);
var
  Cfg: TNConfig;
begin
  Cfg := TNConfig.Create(Config, AStorage);
  CopyFromConfig(Cfg);
  Cfg.Free;
end;

function TNConfig.Clone: TNConfig;
var
  I: Long;
begin
  Result := TNConfig.Create;
  Result.Name := Self.Name;
  for I := 0 to FValues.Count - 1 do
    Result.FValues.Add(FValues.Keys[I], FValues.Words[I]);
  for I := 0 to FNodes.Count - 1 do
    Result.AddConfig(TNConfig(FNodes.Objects[I]).Name,
      TNConfig(FNodes.Objects[I]).Clone);
end;

procedure TNConfig.Clear;
var
  I: Long;
begin
  for I := 0 to FNodes.Count - 1 do
  begin
    FNodes.Objects[I].Free;
    FNodes.Objects[I] := nil;
  end;
  FNodes.Clear;
  FValues.Clear;
end;

function TNConfig.FindSubNode(const NodeName: NString): TNConfig;
var
  I: Long;
  SubName, SubNode: NString;
begin
  Result := nil;
  if NodeName = '' then
    Result := Self
  else
  begin
    I := PosNoCase(sConfigPathDelim, NodeName);
    if I > 0 then
    begin
      // NodeName is subnodes
      SubName := Copy(NodeName, 1, I - 1);
      SubNode := Copy(NodeName, I + 1, MaxLong);
      if FNodes.Find(SubName, I) then
        Result := TNConfig(FNodes.Objects[I]).FindSubNode(SubNode);
    end
    else
      // NodeName is current node's child
      if FNodes.Find(NodeName, I) then
        Result := TNConfig(FNodes.Objects[I]);
  end;
end;

procedure TNConfig.LoadFromFile(const FileName: NString);
var
  F: TNBinaryFile;
  OldStorage, NewStorage: TConfigStorage;
  Ext: NString;
begin
  try
    F := Vfs.Open(FileName, fdRead);
    OldStorage := Storage;
    Ext := ExtractFileExt(FileName);
    if Ext <> '' then
      Delete(Ext, 1, 1);
    // set new storage if supported
    if IsSupportedExt(Ext) then
    begin
      NewStorage := GetStorageByExt(Ext);
      if FStorage <> NewStorage then
        Storage := NewStorage;
    end;
  except
    on E: Exception do
    begin
      ShowError(ErrorStr(Self, 'LoadFromFile', E.Message), E.ClassName);
      Exit;
    end;
  end;
  if F <> nil then
  begin
    LoadFromStream(F.Stream);
    F.Free;
  end;
  // set original storage
  if FStorage <> OldStorage then
    Storage := OldStorage;
end;

procedure TNConfig.LoadFromStream(Stream: TNStream);
var
  Driver: TNAbstractConfigDriver;
  Key, Value, CfgFile: NString;
  Node: TNConfig;
begin
  Clear;
  try
    Driver := GetDriverClass(Storage).Create(Stream, dmRead);
    Driver.Flush;
    // read and add all values until end of stream (or something)
    while not Driver.IsEnd do
    begin
      Driver.ReadValue(Key, Value);
      if CheckForCfgLoad(Value, CfgFile) then
      begin
        Node := AddNode(Key);
        if Node <> nil then
          Node.CopyFromConfig(CfgFile);
      end
      else
        WriteString(Key, Value);
    end;
    Driver.Free;
  except
    on E: Exception do
      ShowError(ErrorStr(Self, 'LoadFromStream', 'Error when loading ' +
        'configuration data from stream.' + sLineBreak + sExceptMsg +
        E.Message), E.ClassName);
  end;
end;

procedure TNConfig.SaveToFile(const FileName: NString);
var
  F: TNBinaryFile;
begin
  try
    F := Vfs.Open(FileName, fdCreate);
  except
    on E: Exception do
    begin
      ShowError(ErrorStr(Self, 'SaveToFile', E.Message), E.ClassName);
      Exit;
    end;
  end;
  if F <> nil then
  begin
    SaveToStream(F.Stream);
    F.Free;
  end;
end;

procedure TNConfig.SaveToStream(Stream: TNStream);
var
  Driver: TNAbstractConfigDriver;
begin
  try
    Driver := GetDriverClass(Storage).Create(Stream, dmWrite);
    SaveAsNode(Driver);
    Driver.Flush;
    Driver.Free;
  except
    on E: Exception do
      ShowError(ErrorStr(Self, 'SaveToStream', 'Error when saving ' +
        'configuration data to stream.' + sLineBreak + sExceptMsg + E.Message),
        E.ClassName);
  end;
end;

end.

