{*************************************************************************}
{*   Files                                                               *}
{*   Copyright (C) 2003 Marek Mauder, Vampyre SoftWorks                  *}
{*   Engine\Files\Files.pas                                              *}
{*************************************************************************}
{                                                                         }
{ This program is free software; you can redistribute it and/or modify it }
{ under the terms of the GNU General Public License as published by the   }
{ Free Software Foundation; either version 2 of the License, or (at your  }
{ option) any later version. This program is distributed in the hope that }
{ it will be useful, but WITHOUT ANY WARRANTY; without even the implied   }
{ warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See   }
{ the GNU General Public License for more details.                        }
{                                                                         }
{ You should have received a copy of the GNU General Public License       }
{ along with this program; if not, write to the Free Software Foundation, }
{ Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.         }
{                                                                         }
{*************************************************************************}
{  Created On: 5th November 2003  (05.11.2003)                            }
{*************************************************************************}

{This unit contains classes and functions working with files,
 file archives, compression, ...}

unit Files;

{ TODOS:
  - Test BuildFileList on Linux.
  - Test VFS on Linux heavily.
  - Vfs.DirectoryExists('') returns True, solve
}

{$I Compiler.inc}

interface

uses
  SysUtils, Kernel, Objects, Platforms;


{*************************************************************************}
{                         Consts, Types                                   }
{*************************************************************************}


const

  {Identifier used to differ Compression module library
   from another module libraries.}
  CmpLibraryID = 'cmp';
  {Identifier used to differ ArchiveFormat module library
   from another module libraries.}
  ArcLibraryID = 'arc';

  {Path variables used by VFS. If encoutered in filename, they are
   replaced by according path.}
  {This var is replaced by ExeDir() string.}
  PathVarExeDir = 'EXEDIR';
  {This var is replaced by root dir - ExtractFileDir(ExeDir).}
  PathVarRoot = 'ROOT';

  {Some defalut directories.}
  {Directory for stroring various app settings.}
  SettingsDir = 'Settings';
  {Directory for stroring app data (RootDir\DataDir).}
  DataDir = 'Data';
  {Directory for stroring app graphic data.}
  GraphicDir = DataDir + PathDelim + 'Graphic';
  {Directory for stroring app music data.}
  MusicDir = DataDir + PathDelim + 'Music';
  {Directory for stroring app sound data.}
  SoundDir = DataDir + PathDelim + 'Sound';


{*************************************************************************}
{                        File Module classes                              }
{*************************************************************************}


type

  {This module compresses and decompresses input data.}
  TNCompression = class(TNFileModule)
  protected
    {These methods are called by Compress/Decompress methods to
     do real compression/decompression. Override these in descendants.}
    procedure DoCompress(InData: Pointer; InSize: Long; var OutData: Pointer; var
      OutSize: Long); virtual;
    procedure DoDecompress(InData: Pointer; InSize: Long; var OutData: Pointer;
      var OutSize: Long); virtual;
  public
    {Compresses input data to output data.}
    function Compress(InData: Pointer; InSize: Long; var OutData: Pointer; var
      OutSize: Long): Boolean;
    {Decompresses input data to output data.}
    function Decompress(InData: Pointer; InSize: Long; var OutData: Pointer;
      var OutSize: Long): Boolean;
  end;

  {Class refernce for TNCompression.}
  TNCompressionClass = class of TNCompression;

  {Enum describing capabilities of archive format:
     acReadFile      - files can be read from archive
     acAddFile       - new non-existing files can be written to archive
     acDeleteFile    - existing file can be deleted
     acAddDir        - new empty directory can be add to archive
     acRenameDir     - existing directory can be renamed (and all files
                         in dir in archives where dir is part of filename)
     acDeleteDir     - existing directory and all files in it can be deleted}
  TArchiveCap = (acReadFile, acAddFile, acDeleteFile,
    acAddDir, acRenameDir, acDeleteDir);
  {Set of archive caps.}
  TArchiveCaps = set of TArchiveCap;

  TNArchive = class;

  {This module is abstract wrapper to APIs that control access to
   various archive file formats like zip, tar, ...}
  TNArchiveFormat = class(TNFileModule)
  protected
    FCaps: TArchiveCaps;
    {These methods are called by public methods to
     do real things. Override these in descendants.}
    function DoOpen(const FileName: NString): TNArchive; virtual;
    procedure DoClose(Archive: TNArchive); virtual;
    procedure DoFillFileList(Archive: TNArchive); virtual;
    function DoReadFile(Archive: TNArchive; const FileName: NString; var Data:
      Pointer; var Size: Long): Boolean; virtual;
    function DoAddFile(Archive: TNArchive; const FileName: NString; Data:
      Pointer; Size: Long): Boolean; virtual;
    function DoDeleteFile(Archive: TNArchive; const FileName: NString): Boolean;
      virtual;
    function DoAddDirectory(Archive: TNArchive; const DirName: NString):
      Boolean;
      virtual;
    function DoRenameDirectory(Archive: TNArchive; const OldName,
      NewName: NString): Boolean; virtual;
    function DoDeleteDirectory(Archive: TNArchive; const DirName: NString):
      Boolean; virtual;
  public
    {Opens archive of given name and creates new instance
     of TNArchive for it. If opening fail, nil is returned.}
    function Open(const FileName: NString): TNArchive;
    {Closes given archive. It is called in destructor of TNArchive.
     It does not free archive instance. Rather call Archive.Free.}
    procedure Close(Archive: TNArchive);
    {Fills archive's file list with actual contents of archive
     (files and directories, dirs end with PathDelim).}
    procedure FillFileList(Archive: TNArchive);
    {Returns True if given file or directory is in Archive.}
    function IsInArchive(Archive: TNArchive; const AName: NString):
      Boolean; virtual;
    {Reads file (in Archive) of given name to Data buffer. Returns True if
     successful. acReadFile must be in Caps.}
    function ReadFile(Archive: TNArchive; const FileName: NString; var Data:
      Pointer; var Size: Long): Boolean;
    {Writes file (to Archive) of given name contained in Data buffer. Returns
     True if successful. acAddFile must be in Caps.
     If file already exists in archive:
       - acDeleteFile IS IN Caps - old file is deleted and new is written,
                                     True is returned
       - acDeleteFile IS NOT IN Caps - nothing is written, False is returned}
    function AddFile(Archive: TNArchive; const FileName: NString; Data:
      Pointer; Size: Long): Boolean;
    {Deletes given file and returns True if successful.
     acDeleteFile must be in FCaps.}
    function DeleteFile(Archive: TNArchive; const FileName: NString): Boolean;
    {Adds new directory to archive. acAddDir must be in Caps.}
    function AddDirectory(Archive: TNArchive; const DirName: NString): Boolean;
    {Renames given directory in archive. acRenameDir must be in Caps.}
    function RenameDirectory(Archive: TNArchive; const OldName,
      NewName: NString): Boolean;
    {Deletes given directory and all files in it. acDeleteDir must be in Caps}
    function DeleteDirectory(Archive: TNArchive; const DirName: NString):
      Boolean;
    {Archive format capabilities}
    property Caps: TArchiveCaps read FCaps;
  end;

  {Class refernce for TNArchiveFormat.}
  TNArchiveFormatClass = class of TNArchiveFormat;

  {This class represents single file archive of any type. Internal format
   of archive (zip, tar, ...) is accessed trough Format property.}
  TNArchive = class(TNNamedPersistent)
  protected
    FFiles: TNSortedStrings;
    FFormat: TNArchiveFormat;
    FHandle: Pointer;
    FPath: NString;
  public
    constructor Create(AFormat: TNArchiveFormat; AHandle: Pointer); reintroduce;
    destructor Destroy; override;
    {Returns True if given file or directory is in archive.}
    function IsInArchive(const AName: NString): Boolean;
    {Reads file of given name to Data buffer. Returns True if successful.}
    function ReadFile(const FileName: NString; var Data: Pointer; var Size:
      LongInt): Boolean; overload;
    {Writes file of given name contained in Data buffer. Returns True if successful.}
    function AddFile(const FileName: NString; Data: Pointer; Size: LongInt):
      Boolean;
    {Deletes given file from archive and returns True if successful.}
    function DeleteFile(const FileName: NString): Boolean;
    {Adds new directory to archive and returns True if successful.}
    function AddDirectory(const DirName: NString): Boolean;
    {Renames given directory in archive. acRenameDir must be in Caps.}
    function RenameDirectory(Archive: TNArchive; const OldName,
      NewName: NString): Boolean;
    {Deletes given directory and all files in it. acDeleteDir must be in Caps}
    function DeleteDirectory(Archive: TNArchive; const DirName: NString):
      Boolean;
    {Archive low-level handle used by archive formats.}
    property Handle: Pointer read FHandle;
    {File format of this archive (zip,tar,...).}
    property Format: TNArchiveFormat read FFormat;
  published
    {Contains list of all files in archive.}
    property Files: TNSortedStrings read FFiles write FFiles;
    {Full path to archive file.}
    property Path: NString read FPath write FPath;
  end;


{*************************************************************************}
{                            Virtual File System                          }
{*************************************************************************}


{$IFNDEF MODLIBRARY}

  {Type of VFS file:
    ftInvalid     - Invalid type
    ftDiskFile    - File is raw file on disk
    ftArchiveFile - File is stored in some archive.}
  TFileType = (ftInvalid, ftDiskFile, ftArchiveFile);
  {Type of VFS directory:
    dtInvalid     - Invalid type
    dtDiskDir     - Dir is raw directory on disk
    dtArchiveDir  - Dir is stored in some archive.}
  TDirType = (dtInvalid, dtDiskDir, dtArchiveDir);
  {Mode of VFS file:
    fdRead      - File is opened only for reading (if file does not exist NOTHING
                  is created)
    fdWrite     - File is opened inly for writing (if file does not exist NEW one
                  is created)
    fdReadWrite - File is opened for reading and writing (if file does not exist
                  NEW one is created)
    fdCreate    - If file exists it is overwritten, otherwise new file is
                  created. After creation/opening mode is set to fmReadWrite!}
  TFileMode = (fdRead = 0, fdWrite = 1, fdReadWrite = 2, fdCreate = 3);

  {Base class for binary files used by VFS. Do not create instances
   of this class - it is abstract.}
  TNBinaryFile = class(TNBaseObject)
  protected
    FFileName: NString;
    FFileType: TFileType;
    FMode: TFileMode;
    function GetSize: Long;
    function GetPosition: Long;
    function GetStream: TNStream; virtual; abstract;
    procedure SetPosition(Value: Long);
  public
    {Creates instance and opens/creates file AFileName in AMode mode.}
    constructor Create(const AFileName: NString; AMode: TFileMode); reintroduce;
      virtual;
    {Reads Size bytes from file to Data buffer (at least Size big).
     Returns number of bytes really read. Works only if Mode is not fmWrite.}
    function Read(Data: Pointer; Size: Long): Long;
    {Writes Size bytes from Data buffer to file.
     Returns number of bytes really written. Works only if Mode is not fmRead.}
    function Write(Data: Pointer; Size: Long): Long;
    {This function reads whole file to newlly allocated buffer which is returned
     (which is GetSize bytes large. Result must be freed after usage.
      If error occurrs during reading, nil is returned. Stream position is not
      changed by successful call to this method.}
    function ReadWholeFile: Pointer;
    {Reads whole file to one string. Stream position is not changed.}
    function ReadToString: NString;
    {Name of this file. It can be absolute or relative path.}
    property FileName: NString read FFileName;
    {Size of this file in bytes.}
    property Size: Long read GetSize;
    {Actual posotion in file.}
    property Position: Long read GetPosition write SetPosition;
    {Type of this file.}
    property FileType: TFileType read FFileType;
    {Mode of this file.}
    property Mode: TFileMode read FMode;
    {Underlying stream. All IO operations are done trough this stream.}
    property Stream: TNStream read GetStream;
  end;

  {This represents binary file stored on disk. It uses standard TNFileStream
   to do IO operations.}
  TNDiskFile = class(TNBinaryFile)
  protected
    FStream: TNFileStream;
    function GetStream: TNStream; override;
  public
    constructor Create(const AFileName: NString; AMode: TFileMode); override;
    destructor Destroy; override;
  end;

  {This represents binary file stored in some supported archive.
   It uses TNMemoryStream for IO. After creation entire file
   is read form archive to stream (if file exists, otherwise stream is empty).
   File is opened in write mode only if archive format supports saving files
   (otherwise fmRead is set automaticaly) and then file is written to
   archive by Flush procedure (called in destructor).}
  TNArchiveFile = class(TNBinaryFile)
  protected
    FStream: TNMemoryStream;
    FArchive: TNArchive;
    function GetStream: TNStream; override;
  public
    constructor Create(const AFileName: NString; AMode: TFileMode; AArchive:
      TNArchive); reintroduce;
    destructor Destroy; override;
    {Writes file into archive (if mode is not fmRead.)}
    procedure Flush;
  end;

  TNVirtualFileSystem = class;

  {Class used internaly by VFS to representation of virtual directories.
   You can assign more real on-disk directories or archives to
   single virtual directory.}
  TNVFSNode = class(TNBaseObject)
  protected
    FOwner: TNVirtualFileSystem;
    FVirtualPath: NString;
    FRealPaths: TNSortedStrings;
    FLastArchive: TNArchive;
    FLastType: TFileType;
    FLastDirType: TDirType;
    FLastReal: NString;
    function GetRelName(const AName: NString): NString;
    function DeleteFile(const AFileName: NString): Boolean;
    {Returns real path of file located in virtual dir
     (only for disk files).}
    function GetRealFileName(const AFileName: NString): NString;
    {Returns real path of directory located in virtual dir
     (only for disk dirs.}
    function GetRealDirName(const ADirName: NString): NString;
    {Returns real name of file that can be created in this virtual directory
     (that means all required subdirectories exist) (only for disk files).}
    function GetNewFileName(const AFileName: NString): NString;
    {Returns real name of directory that can be created in this virtual directory
     (that means all required subdirectories exist) (only for disk dirs).}
    function GetNewDirName(const ADirName: NString): NString;
    {Returns index to firts of real path thta is disk directory. Returns
      -1 if all real paths are archives.}
    function GetFirstDiskRealPath: Long;
  public
    {Creates new node with AVirtualPath path.}
    constructor Create(AOwner: TNVirtualFileSystem; const AVirtualPath:
      NString);
      reintroduce;
    destructor Destroy; override;
    {Assigns new real path or archive with virtual path.}
    procedure AddRealPath(const ARealPath: NString);
    {Returns Type of file in virtual dir or if file is not in dir
     returns ftInvalid.}
    function FileType(const AFileName: NString): TFileType;
    {Returns Type of ADirName directory in virtual dir or if ADirName
     is not in dir returns dtInvalid.}
    function DirType(const ADirName: NString): TDirType;
    {Fills List with names of all files which are located in this
     virtual directory.}
    procedure GetFileList(List: TNStringList);
    {Virtual directory path.}
    property VirtualPath: NString read FVirtualPath;
    {Real paths (and archives) assigned with this virtual directory.}
    property RealPaths: TNSortedStrings read FRealPaths;
  end;

  {Cross-platform virtual file system class which is intended to
   ease access to files. Some features:
   - More real-on-disk directories as well as file archives can be
       mapped to one virtual directory.
   - Path delimiter independent - you can
       call VFS.Open('\home/galfar\\projects/Files.pas');
   - You can use path variables in path which will be translated to
       adequate values - "$(EXEDIR)\Logs\App.log" will translate to
       ExeDir+"Logs\App.log" >> (D:\Project\Bin\Logs\App.log).
       Path vars (translated) do not contain PathDelim at the end!}
  TNVirtualFileSystem = class(TNSingleton)
  protected
    FNodes: TNSortedStringObjects;
    FArchives: TNSortedStringObjects;
    function GetRealPathsStr(const AVirtualPath: NString): TNStringList;
    function GetRealPathsIdx(Index: Long): TNStringList;
    function GetVirtualDir(Index: Long): NString;
    function GetVirtualDirCount: Long;
    function FindFileNode(const AFileName: NString): TNVFSNode;
    function FindDirNode(const ADirName: NString): TNVFSNode;
    function ConvertFName(const AName: NString): NString;
    function ConvertDName(const AName: NString): NString;
    function ConvertRPath(const APath: NString): NString;
    function ConvertVPath(const APath: NString): NString;
    function CreateFile(const AFileName: NString; AMode: TFileMode):
      TNBinaryFile;
    {Translates PathVar constants in paths to actual paths.}
    function TranslateVars(const APath: NString): NString;
    procedure DeleteEmptyNodes;
    procedure DeleteNode(Index: Long);
    {This procedure mount some default directories (Bin, Modules, Data, ...)}
    procedure AddDefaultDirectories;
  public
    constructor Create; override;
    destructor Destroy; override;

    {* File operations *}

    {Opens file AFileName and returns:
       - TNDiskFile instance    - if AFileName is file on disk
       - TNArchiveFile instance - if AFileName is file in archive.
     If AFileName was not found new file is created (unless AMode is
     fmRead) in adequate virtual directory. If adequate virtual directory
     can not be resolved AFileName is used directly as name of file.}
    function Open(const AFileName: NString; AMode: TFileMode = fdReadWrite):
      TNBinaryFile;
    {This procedure is equivalent to FreeAndNil(BinaryFile).}
    procedure Close(var BinaryFile);
    {Deletes given file. Returns True if file was found and deleted.}
    function DeleteFile(const AFileName: NString): Boolean;
    {Returns True if file exists on disk or in directory
      or archive assigned to one of virtual dirs.}
    function FileExists(const AFileName: NString): Boolean;
    {Returns True if file exists on disk. Search does not
     include virtual dirs or archives (This is for function which does not
     use Vfs for opening files (e.g. Lists, orsomething in Kernel)).}
    function FileExistsOutsideVfs(const AFileName: NString): Boolean;
    {Returns True if directory exists on disk or in directory
      or archive assigned to one of virtual dirs.}
    function DirectoryExists(const ADirName: NString): Boolean;
    {Returns True if directory exists on disk. Search does not
     include virtual dirs or archives (This is for function which does not
     use Vfs for opening files (e.g. Lists, orsomething in Kernel))}
    function DirectoryExistsOutsideVfs(const ADirName: NString): Boolean;
    {Returns full real path to file (ONLY FOR files on DISK).
     Useful for modules which take only file names (e.g. DirectShow
     or lists in Kernel, ...).
     If OnlyExisting is True function returns empty string if file can not
     be found. If OnlyExisting is False it may return name of non-existing
     file (for saving) based on virtual dir, real path and file name.}
    function GetFullFileName(const AFileName: NString; OnlyExisting: Boolean =
      False): NString;
    {Returns full real path to directory (ONLY FOR dirs on DISK).
     If dir can not be found empty string is returned.
     if ADirName is name of one of virtual directories then
     first realpath (which is disk directory) of this virtual dir is returned.}
    function GetFullDirName(const ADirName: NString; OnlyExisting: Boolean =
      False): NString;

    {* Virtual directory control *}

    {Assigns new real directory or archive to virtual directory.}
    procedure Mount(const AVirtualPath, ARealPath: NString);
    {Removes assignment of real directory or archive to virtual directory.}
    procedure Unmount(const AVirtualPath, ARealPath: NString);
    {This fills List with names of all files in given virtual dir
     (it is recursive). It can take a while.}
    procedure VirtualDirFileList(const AVirtualPath: NString; List:
      TNStringList);
      overload;
    {Same as above but virtual directory is defined by Index (virtual
     directories are sorted in alphabetical order).}
    procedure VirtualDirFileList(Index: Long; List: TNStringList); overload;

    {* Properties *}

    {Number of virtual directories.}
    property VirtualDirCount: Long read GetVirtualDirCount;
    {Name of given virtual directory.}
    property VirtualDir[Index: Long]: NString read GetVirtualDir;
    {Real paths and archives assigned to given virtual directory. StringList
     is reference to node's list so DO NOT DESTROY IT.}
    property VirtualDirRealPathsStr[const AVirtualPath: NString]: TNStringList
    read GetRealPathsStr;
    {Same as above but virtual directory is defined by Index (virtual
     directories are sorted in alphabetical order).}
    property VirtualDirRealPathsIdx[Index: Long]: TNStringList read
    GetRealPathsIdx;
  end;

{$ENDIF}


{*************************************************************************}
{                           Misc functions                                }
{*************************************************************************}


{$IFNDEF MODLIBRARY}

{* TNCompression registration/other *}

{Registers new Compression instance so they can be used by system.}
procedure RegisterCompression(Format: TNCompression);
{Registers new Compression instances.}
procedure RegisterCompressions(Formats: array of TNCompression);
{UnRegisters Compression instance.}
procedure UnRegisterCompression(Format: TNCompression);
{UnRegisters new Compression instance.}
procedure UnRegisterCompressions(Formats: array of TNCompression);
{Returns number of registered compression modules.}
function RegisteredCompressionsCount: Long;
{Returns Compression registered with given extension. If not found
 it returns nil.}
function FindCompression(const Ext: NString): TNCompression; overload;
{Returns Compression at given Index. If index is invalid it returns nil.}
function FindCompression(Index: Long): TNCompression; overload;

{* TNArchiveFormat registration/other *}

{Registers new Archive File Format instance so they can be used by system.}
procedure RegisterArchiveFormat(Format: TNArchiveFormat);
{Registers new Archive File Format instances.}
procedure RegisterArchiveFormats(Formats: array of TNArchiveFormat);
{UnRegisters Archive File Format instance.}
procedure UnRegisterArchiveFormat(Format: TNArchiveFormat);
{UnRegisters new Archive File Format instance.}
procedure UnRegisterArchiveFormats(Formats: array of TNArchiveFormat);
{Returns Image file format registered with given extension. If not found
 it returns nil.}
function FindArchiveFormat(const Ext: NString): TNArchiveFormat;

{* Other functions *}

{Returns True if file is one of supported archive formats
 (file extension check).}
function FileIsArchive(const FileName: NString): Boolean;

{$ENDIF}

type
  {Options for BuildFileList function:
  flFullNames - file names in result will have full path names
                (ExtractFileDir(Path)+FileName)
  flRelNames  - file names in result will have names relative to
                ExtractFileDir(Path) dir
  flRecursive - adds files in subdirectories foun in Path.}
  TFileListOption = (flFullNames, flRelNames, flRecursive);
  TFileListOptions = set of TFileListOption;

{This function fills Files string list with names of files found
 with FindFirst/FindNext functions (See details on Path/Atrr here).
 - BuildFileList('c:\*.*',faAnyFile, List, [flRecursive]) returns
   list of all files (only name.ext - no path) on C drive
 - BuildFileList('d:\*.*',faDirectory, List, [flFullNames]) returns
   list of all directories (d:\dirxxx) in root of D drive.}
function BuildFileList(Path: NString; Attr: Long; Files: TNStringList;
  Options: TFileListOptions = []): Boolean;


{*************************************************************************}
{                           Singleton access                              }
{*************************************************************************}


{$IFNDEF MODLIBRARY}

{This function provides access to TNVirtualFileSystem singleton instance.
 I rather use function than global variable becaude it can check
 if instance is valid.}
function Vfs: TNVirtualFileSystem;

{$ENDIF}


{*************************************************************************}
{                             Implementation                              }
{*************************************************************************}


implementation

{$IFNDEF MODLIBRARY}
uses
  {Default archive file formats.}
  ZipFormat, TarFormat,
  {Default archive file formats.}
  ZLibFormat, RleFormat;

var
  {Global instance of VFS file accessible trough Vfs() function
   (it check if instance is created).}
  VfsInstance: TNVirtualFileSystem = nil;

  {List for holding reistered compressions formats. Accessed trough
   RegisteredCompressions function.}
  RegCompressions: TNFileModuleList = nil;

  {List for holding reistered archive file formats. Accessed trough
   RegisteredArchiveFormat function.}
  RegArchiveFormats: TNFileModuleList = nil;
{$ENDIF}


{*************************************************************************}
{                           Singleton access                              }
{*************************************************************************}


{$IFNDEF MODLIBRARY}

function Vfs: TNVirtualFileSystem;
begin
  if VfsInstance = nil then
    VfsInstance := TNVirtualFileSystem.Create;
  Result := VfsInstance;
end;

{$ENDIF}

{*************************************************************************}
{                          TNCompression methods                          }
{*************************************************************************}


procedure TNCompression.DoCompress(InData: Pointer; InSize: Long;
  var OutData: Pointer; var OutSize: Long);
begin
end;

procedure TNCompression.DoDecompress(InData: Pointer; InSize: Long;
  var OutData: Pointer; var OutSize: Long);
begin
end;

function TNCompression.Compress(InData: Pointer; InSize: Long;
  var OutData: Pointer; var OutSize: Long): Boolean;
begin
  Result := False;
  if FCanSave then
  try
    DoCompress(InData, InSize, OutData, OutSize);
    Result := True;
  except
    on E: Exception do
    begin
      ShowError(ErrorStr(Self, 'Compress',
        'Error when compressing data! Data can be corrupted.'
        + sLineBreak + sExceptMsg + E.Message), E.ClassName);
      Result := False;
    end;
  end;
end;

function TNCompression.Decompress(InData: Pointer; InSize: Long;
  var OutData: Pointer; var OutSize: Long): Boolean;
begin
  Result := False;
  if FCanLoad then
  try
    DoDecompress(InData, InSize, OutData, OutSize);
    Result := True;
  except
    on E: Exception do
    begin
      ShowError(ErrorStr(Self, 'Decompress',
        'Error when decompressing data! Data can be corrupted.'
        + sLineBreak + sExceptMsg + E.Message), E.ClassName);
      Result := False;
    end;
  end;
end;


{*************************************************************************}
{                        TNArchiveFormat methods                          }
{*************************************************************************}


function TNArchiveFormat.DoOpen(const FileName: NString): TNArchive;
begin
  Result := nil;
end;

procedure TNArchiveFormat.DoClose(Archive: TNArchive);
begin
end;

procedure TNArchiveFormat.DoFillFileList(Archive: TNArchive);
begin
end;

function TNArchiveFormat.DoReadFile(Archive: TNArchive;
  const FileName: NString; var Data: Pointer; var Size: Long): Boolean;
begin
  Result := False;
end;

function TNArchiveFormat.DoAddFile(Archive: TNArchive;
  const FileName: NString; Data: Pointer; Size: Long): Boolean;
begin
  Result := False;
end;

function TNArchiveFormat.DoDeleteFile(Archive: TNArchive; const FileName:
  NString): Boolean;
begin
  Result := False;
end;

function TNArchiveFormat.DoAddDirectory(Archive: TNArchive; const DirName:
  NString): Boolean;
begin
  Result := False;
end;

function TNArchiveFormat.DoRenameDirectory(Archive: TNArchive; const OldName,
  NewName: NString): Boolean;
begin
  Result := False;
end;

function TNArchiveFormat.DoDeleteDirectory(Archive: TNArchive; const DirName:
  NString): Boolean;
begin
  Result := False;
end;

function TNArchiveFormat.Open(const FileName: NString): TNArchive;
begin
  try
    if not FileExists(FileName) then
      raise Exception.Create('File can not be found!');
    Result := DoOpen(FileName);
    if Result.Path = '' then
      Result.Path := FileName;
    if Result.Name = '' then
      Result.Name := ExtractFileName(FileName);
    FillFileList(Result);
{$IFDEF DEBUG}
    Log.Add('Archive opened: %s (%d files)', [Result.Name, Result.Files.Count]);
{$ENDIF}
  except
    on E: Exception do
    begin
      ShowError(ErrorStr(Self, 'Open', 'Error when opening archive "'
        + FileName + '" !' + sLineBreak + sExceptMsg + E.Message), E.ClassName);
      Result := nil;
    end;
  end;
end;

procedure TNArchiveFormat.Close(Archive: TNArchive);
begin
  if Archive <> nil then
  try
    DoClose(Archive);
{$IFDEF DEBUG}
    Log.Add('Archive closed: ' + Archive.Name);
{$ENDIF}
  except
    on E: Exception do
      ShowError(ErrorStr(Self, 'Close', 'Error when closing archive "'
        + Archive.Name + '" !' + sLineBreak + sExceptMsg + E.Message),
        E.ClassName);
  end;
end;

procedure TNArchiveFormat.FillFileList(Archive: TNArchive);
var
  I: Long;
begin
  if Archive <> nil then
  try
    Archive.Files.Sorted := False;
    Archive.Files.Clear;
    DoFillFileList(Archive);
    // change path delimiters to match current platform delimiters
    for I := 0 to Archive.Files.Count - 1 do
      Archive.Files[I] := SwapPathDelims(Archive.Files[I], PathDelim);
    Archive.Files.Sorted := True;
  except
    on E: Exception do
      ShowError(ErrorStr(Self, 'FillFileList', 'Error when filling file ' +
        'list for archive "' + Archive.Name + '" !' + sLineBreak + sExceptMsg +
        E.Message), E.ClassName);
  end;
end;

function TNArchiveFormat.IsInArchive(Archive: TNArchive; const AName:
  NString): Boolean;
begin
  Result :=
    (Archive.Files.IndexOf(SwapPathDelims(AName, PathDelim)) > -1) or
  // in the case that AName is directory without PathDelim at the end
  (Archive.Files.IndexOf(SwapPathDelims(AName, PathDelim) + PathDelim) > -1);
end;

function TNArchiveFormat.ReadFile(Archive: TNArchive;
  const FileName: NString; var Data: Pointer; var Size: Long): Boolean;
begin
  Result := False;
  Data := nil;
  Size := 0;
  if (acReadFile in FCaps) and (Archive <> nil) then
    if IsInArchive(Archive, FileName) then
    try
      Result := DoReadFile(Archive, FileName, Data, Size);
    except
      on E: Exception do
      begin
        ShowError(ErrorStr(Self, 'ReadFile',
          'Error while reading file "' + FileName + '" from archive "'
          + Archive.Name + '"!.' + sLineBreak + sExceptMsg + E.Message),
          E.ClassName);
        Result := False;
      end;
    end;
end;

function TNArchiveFormat.AddFile(Archive: TNArchive;
  const FileName: NString; Data: Pointer; Size: Long): Boolean;
begin
  Result := False;
  if (acAddFile in FCaps) and (Archive <> nil) then
  try
    // if file is already in archive, check if we can overwrite it (delete+add)
    // and if not then jump to exit
    if IsInArchive(Archive, FileName) then
    begin
      if (acDeleteFile in FCaps) then
      begin
        if not DeleteFile(Archive, FileName) then
          Exit;
      end
      else
        Exit;
    end;
    Result := DoAddFile(Archive, FileName, Data, Size);
    if Result then
      FillFileList(Archive);
  except
    on E: Exception do
    begin
      ShowError(ErrorStr(Self, 'AddFile',
        'Error while writing file "' + FileName + '" to archive "'
        + Archive.Name + '"!.' + sLineBreak + sExceptMsg + E.Message),
        E.ClassName);
      Result := False;
    end;
  end;
end;

function TNArchiveFormat.DeleteFile(Archive: TNArchive; const FileName:
  NString): Boolean;
begin
  Result := False;
  if (acDeleteFile in FCaps) and (Archive <> nil) then
    if IsInArchive(Archive, FileName) then
    try
      Result := DoDeleteFile(Archive, FileName);
      if Result then
        FillFileList(Archive);
    except
      on E: Exception do
      begin
        ShowError(ErrorStr(Self, 'DeleteFile',
          'Error while deleting file "' + FileName + '" from archive "'
          + Archive.Name + '"!.' + sLineBreak + sExceptMsg + E.Message),
          E.ClassName);
        Result := False;
      end;
    end;
end;

function TNArchiveFormat.AddDirectory(Archive: TNArchive;
  const DirName: NString): Boolean;
begin
  Result := False;
  if (acAddDir in FCaps) and (Archive <> nil) then
    if not IsInArchive(Archive, DirName) then
    try
      Result := DoAddDirectory(Archive, DirName);
      if Result then
        FillFileList(Archive);
    except
      on E: Exception do
      begin
        ShowError(ErrorStr(Self, 'AddDirectory',
          'Error while adding directory "' + DirName + '" to archive "'
          + Archive.Name + '"!.' + sLineBreak + sExceptMsg + E.Message),
          E.ClassName);
        Result := False;
      end;
    end;
end;

function TNArchiveFormat.RenameDirectory(Archive: TNArchive; const OldName,
  NewName: NString): Boolean;
begin
  Result := False;
  if (acRenameDir in FCaps) and (Archive <> nil) then
    if IsInArchive(Archive, OldName) then
    try
      Result := DoRenameDirectory(Archive, OldName, NewName);
      if Result then
        FillFileList(Archive);
    except
      on E: Exception do
      begin
        ShowError(ErrorStr(Self, 'RenameDirectory',
          'Error while renaming directory "' + OldName +
          '" to "' + NewName + '" in archive "' + Archive.Name +
          '"!.' + sLineBreak + sExceptMsg + E.Message), E.ClassName);
        Result := False;
      end;
    end;
end;

function TNArchiveFormat.DeleteDirectory(Archive: TNArchive; const DirName:
  NString): Boolean;
begin
  Result := False;
  if (acDeleteFile in FCaps) and (Archive <> nil) then
    if IsInArchive(Archive, DirName) then
    try
      Result := DoDeleteDirectory(Archive, DirName);
      if Result then
        FillFileList(Archive);
    except
      on E: Exception do
      begin
        ShowError(ErrorStr(Self, 'DeleteDirectory',
          'Error while deleting directory "' + DirName + '" from archive "'
          + Archive.Name + '"!.' + sLineBreak + sExceptMsg + E.Message),
          E.ClassName);
        Result := False;
      end;
    end;
end;


{*************************************************************************}
{                        TNArchive methods                                }
{*************************************************************************}


constructor TNArchive.Create(AFormat: TNArchiveFormat; AHandle: Pointer);
begin
  inherited Create;
  FHandle := AHandle;
  FFormat := AFormat;
  FFiles := TNSortedStrings.Create;
  if FFormat = nil then
    ShowError(ErrorStr(Self, 'Create', 'Format of this archive is nil!' +
      'If you do anything else with this archive, Access Violations will arise!'));
end;

destructor TNArchive.Destroy;
begin
  FFormat.Close(Self);
  FFormat := nil;
  FFiles.Free;
  inherited Destroy;
end;

function TNArchive.IsInArchive(const AName: NString): Boolean;
begin
  Result := FFormat.IsInArchive(Self, AName);
end;

function TNArchive.ReadFile(const FileName: NString; var Data: Pointer; var
  Size:
  LongInt): Boolean;
begin
  Result := FFormat.ReadFile(Self, FileName, Data, Size);
end;

function TNArchive.AddFile(const FileName: NString; Data: Pointer; Size:
  LongInt): Boolean;
begin
  Result := FFormat.AddFile(Self, FileName, Data, Size);
end;

function TNArchive.DeleteFile(const FileName: NString): Boolean;
begin
  Result := FFormat.DeleteFile(Self, FileName);
end;

function TNArchive.AddDirectory(const DirName: NString): Boolean;
begin
  Result := FFormat.AddDirectory(Self, DirName);
end;

function TNArchive.RenameDirectory(Archive: TNArchive; const OldName,
  NewName: NString): Boolean;
begin
  Result := FFormat.RenameDirectory(Self, OldName, NewName);
end;

function TNArchive.DeleteDirectory(Archive: TNArchive; const DirName: NString):
  Boolean;
begin
  Result := FFormat.DeleteDirectory(Self, DirName);
end;

{$IFNDEF MODLIBRARY}


{*************************************************************************}
{                         TNBinaryFile methods                            }
{*************************************************************************}


constructor TNBinaryFile.Create(const AFileName: NString; AMode: TFileMode);
begin
  inherited Create;
  FFileName := AFileName;
  FMode := AMode;
  FFileType := ftInvalid;
end;

function TNBinaryFile.GetSize: Long;
begin
  Result := GetStream.Size;
end;

function TNBinaryFile.GetPosition: Long;
begin
  Result := GetStream.Position;
end;

procedure TNBinaryFile.SetPosition(Value: Long);
begin
  GetStream.Position := Value;
end;

function TNBinaryFile.Read(Data: Pointer; Size: Long): Long;
begin
  Result := 0;
  if FMode <> fdWrite then
  try
    Result := GetStream.Read(Data^, Size);
  except
    on E: Exception do
      ShowError(ErrorStr(Self, 'Read', 'Error when reading form file "' +
        FFileName + '"!.' + sLineBreak + sExceptMsg + E.Message), E.ClassName);
  end;
end;

function TNBinaryFile.Write(Data: Pointer; Size: Long): Long;
begin
  Result := 0;
  if FMode <> fdRead then
  try
    Result := GetStream.Write(Data^, Size);
  except
    on E: Exception do
      ShowError(ErrorStr(Self, 'Write', 'Error when writing to file "' +
        FFileName + '"!.' + sLineBreak + sExceptMsg + E.Message), E.ClassName);
  end;
end;

function TNBinaryFile.ReadWholeFile: Pointer;
var
  OldPos: Long;
begin
  OldPos := GetStream.Position;
  GetMem(Result, GetSize);
  if Read(Result, GetSize) <> GetSize then
    FreeMemNil(Result);
  GetStream.Position := OldPos;
end;

function TNBinaryFile.ReadToString: NString;
var
  OldPos: Long;
begin
  OldPos := GetStream.Position;
  SetString(Result, nil, GetSize);
  Read(Pointer(Result), GetSize);
  GetStream.Position := OldPos;
end;


{*************************************************************************}
{                         TNDiskFile methods                              }
{*************************************************************************}


constructor TNDiskFile.Create(const AFileName: NString; AMode: TFileMode);
begin
  inherited Create(AFileName, AMode);
  FFileType := ftDiskFile;
  try
    if SysUtils.FileExists(AFileName) and (FMode <> fdCreate) then
      FStream := TNFileStream.Create(AFileName, Word(FMode))
    else
    begin
      FStream := TNFileStream.Create(AFileName, fmCreate);
      FMode := fdReadWrite;
    end;
  except
    on E: Exception do
      ShowError(ErrorStr(Self, 'Create', 'Error when opening disk file "' +
        AFileName + '"!.' + sLineBreak + sExceptMsg + E.Message), E.ClassName);
  end;
end;

destructor TNDiskFile.Destroy;
begin
  FStream.Free;
  inherited Destroy;
end;

function TNDiskFile.GetStream: TNStream;
begin
  Result := FStream;
end;


{*************************************************************************}
{                         TNArchiveFile methods                           }
{*************************************************************************}


constructor TNArchiveFile.Create(const AFileName: NString; AMode: TFileMode;
  AArchive: TNArchive);
var
  Data: Pointer;
  Size: Long;
begin
  inherited Create(AFileName, AMode);
  FFileType := ftArchiveFile;
  FStream := TNMemoryStream.Create;
  FArchive := AArchive;
  if FArchive = nil then
    ShowError(ErrorStr(Self, 'Create', 'Archive for file "' +
      AFileName + '" is nil!. Working with this file will cause errors!'));
  if FMode = fdCreate then
    FMode := fdReadWrite;
  if not (acAddFile in FArchive.Format.FCaps) then
    FMode := fdRead;
  if FArchive.IsInArchive(AFileName) then
  try
    FArchive.ReadFile(AFileName, Data, Size);
    FStream.Write(Data^, Size);
    FStream.Position := 0;
    FreeMem(Data);
  except
    on E: Exception do
      ShowError(ErrorStr(Self, 'Create', 'Error when reading file "' +
        AFileName + '" form arhive "' + FArchive.Name + '"!' + sLineBreak +
        sExceptMsg + E.Message), E.ClassName);
  end;
end;

destructor TNArchiveFile.Destroy;
begin
  Flush;
  FStream.Free;
  inherited Destroy;
end;

function TNArchiveFile.GetStream: TNStream;
begin
  Result := FStream;
end;

procedure TNArchiveFile.Flush;
begin
  if FMode <> fdRead then
  try
    FArchive.AddFile(FFileName, FStream.Memory, FStream.Size);
  except
    on E: Exception do
      ShowError(ErrorStr(Self, 'Flush', 'Error when writing file "' +
        FFileName + '" to arhive "' + FArchive.Name + '"!' + sLineBreak +
        sExceptMsg + E.Message), E.ClassName);
  end;
end;


{*************************************************************************}
{                            TNVFSNode methods                            }
{*************************************************************************}


constructor TNVFSNode.Create(AOwner: TNVirtualFileSystem; const AVirtualPath:
  NString);
begin
  inherited Create;
  FOwner := AOwner;
  FVirtualPath := AVirtualPath;
  FRealPaths := TNSortedStrings.Create;
  FRealPaths.Duplicates := dupIgnore;
end;

destructor TNVFSNode.Destroy;
begin
  FRealPaths.Free;
  inherited Destroy;
end;

function TNVFSNode.GetRelName(const AName: NString): NString;
begin
  if PosNoCase(FVirtualPath, AName) = 1 then
    Result := ExtractRelativePath(FVirtualPath, AName)
  else
    Result := AName;
end;

function TNVFSNode.DeleteFile(const AFileName: NString): Boolean;
var
  Name: NString;
begin
  // this must be called after FileType
  Result := False;
  Name := GetRelName(AFileName);
  case FLastType of
    ftDiskFile: Result := SysUtils.DeleteFile(FLastReal + Name);
    ftArchiveFile: Result := FLastArchive.DeleteFile(Name);
  end;
end;

function TNVFSNode.GetRealFileName(const AFileName: NString): NString;
var
  FName: NString;
begin
  // this func must be called after FileType and it works only for ftDiskFile
  FName := GetRelName(AFileName);
  Result := FName;
  if FLastType = ftDiskFile then
    Result := FLastReal + FName;
end;

function TNVFSNode.GetRealDirName(const ADirName: NString): NString;
var
  FName: NString;
begin
  // this func must be called after DirType and it works only for dtDiskDir
  FName := GetRelName(ADirName);
  Result := FName;
  if FLastDirType = dtDiskDir then
    Result := FLastReal + FName;
end;

function TNVFSNode.GetNewFileName(const AFileName: NString): NString;
var
  DName, AName, FName: NString;
  I, L: Long;
begin
  AName := GetRelName(AFileName);
  Result := '';
  DName := ExtractFileDir(AName);
  L := Length(DName);
  if (L > 0) and (DName[L] <> PathDelim) then
    DName := DName + PathDelim;
  FName := ExtractFileName(AName);
  // search for realpath which contain required subdirectories
  for I := 0 to FRealPaths.Count - 1 do
    if SysUtils.DirectoryExists(FRealPaths[I] + DName) then
    begin
      Result := FRealPaths[I] + DName + PathDelim + FName;
      Break;
    end;
  // no realpath with subdirs found so siply put file to first disk directory
  if Result = '' then
    Result := FRealPaths[GetFirstDiskRealPath] + FName;
end;

function TNVFSNode.GetNewDirName(const ADirName: NString): NString;
var
  DName, AName, FName: NString;
  I, L: Long;
begin
  AName := GetRelName(ADirName);
  Result := '';
  L := Length(AName);
  if (L > 0) and (AName[L] = PathDelim) then
    Delete(AName, Length(AName), 1);
  DName := ExtractFileDir(AName);
  L := Length(DName);
  if (L > 0) and (DName[L] <> PathDelim) then
    DName := DName + PathDelim;
  FName := ExtractFileName(AName);
  // search for realpath which contain required subdirectories
  for I := 0 to FRealPaths.Count - 1 do
    if SysUtils.DirectoryExists(FRealPaths[I] + DName) then
    begin
      Result := FRealPaths[I] + DName + PathDelim + FName + PathDelim;
      Break;
    end;
  // no realpath with subdirs found so siply put dir to first disk directory
  if Result = '' then
    Result := FRealPaths[GetFirstDiskRealPath] + FName + PathDelim;
end;

function TNVFSNode.GetFirstDiskRealPath: Long;
begin
  for Result := 0 to FRealPaths.Count - 1 do
    if FRealPaths[Result][Length(FRealPaths[Result])] = PathDelim then
      Exit;
  Result := -1;
end;

procedure TNVFSNode.AddRealPath(const ARealPath: NString);
begin
  FRealPaths.Add(ARealPath);
end;

function TNVFSNode.FileType(const AFileName: NString): TFileType;
var
  Name: NString;
  I: Long;
  Arc: TNArchive;
begin
  Result := ftInvalid;
  Name := GetRelName(AFileName);
  for I := 0 to FRealPaths.Count - 1 do
    if FileIsArchive(FRealPaths[I]) then
    begin
      // this real path is archive
      Arc := TNArchive(FOwner.FArchives.FindValue(FRealPaths[I]));
      if Arc.IsInArchive(Name) then
      begin
        // file is in archive
        FLastArchive := Arc;
        Result := ftArchiveFile;
        Break;
      end;
    end
    else
      if SysUtils.FileExists(FRealPaths[I] + Name) then
      begin
        // file exists on disk
        Result := ftDiskFile;
        FLastReal := FRealPaths[I];
        Break;
      end;
  FLastType := Result;
end;

function TNVFSNode.DirType(const ADirName: NString): TDirType;
var
  Name: NString;
  I: Long;
  Arc: TNArchive;
begin
  Result := dtInvalid;
  Name := GetRelName(ADirName);
  for I := 0 to FRealPaths.Count - 1 do
    if FileIsArchive(FRealPaths[I]) then
    begin
      // this real path is archive
      Arc := TNArchive(FOwner.FArchives.FindValue(FRealPaths[I]));
      if Arc.IsInArchive(Name) then
      begin
        // dir is in archive
        FLastArchive := Arc;
        Result := dtArchiveDir;
        Break;
      end;
    end
    else
      if SysUtils.DirectoryExists(FRealPaths[I] + Name) then
      begin
        // dir exists on disk
        Result := dtDiskDir;
        FLastReal := FRealPaths[I];
        Break;
      end;
  FLastDirType := Result;
end;

procedure TNVFSNode.GetFileList(List: TNStringList);
var
  I, J: Long;
  Arc: TNArchive;
  TmpList: TNStringList;
begin
  List.Clear;
  TmpList := TNStringList.Create;
  for I := 0 to FRealPaths.Count - 1 do
    if FileIsArchive(FRealPaths[I]) then
    begin
      // add all file names in archive
      Arc := TNArchive(FOwner.FArchives.FindValue(FRealPaths[I]));
      if Arc = nil then
        Continue;
      for J := 0 to Arc.Files.Count - 1 do
        if Arc.Files[J][Length(Arc.Files[J])] <> PathDelim then
          List.Add(Arc.Files[J]);
    end
    else
    begin
      // build recursive list
      BuildFileList(FRealPaths[I] + '*.*', faAnyFile, TmpList, [flRelNames,
        flRecursive]);
      List.AddStrings(TmpList);
    end;
  TmpList.Free;
end;


{*************************************************************************}
{                      TNVirtualFileSystem methods                        }
{*************************************************************************}


constructor TNVirtualFileSystem.Create;
begin
  inherited Create;
  Name := 'Virtual File System';
  Description := 'Cross-platform virtual file system.';
  FNodes := TNSortedStringObjects.Create;
  FArchives := TNSortedStringObjects.Create;
  FNodes.Duplicates := dupIgnore;
  FArchives.Duplicates := dupIgnore;
  AddDefaultDirectories;
  Log.Add(Name + ' created.');
end;

destructor TNVirtualFileSystem.Destroy;
var
  I: Long;
begin
  for I := 0 to FNodes.Count - 1 do
    FNodes.Objects[I].Free;
  FNodes.Free;
  for I := 0 to FArchives.Count - 1 do
    FArchives.Objects[I].Free;
  FArchives.Free;
  Log.Add(Name + ' destroyed.');
  inherited Destroy;
end;

function TNVirtualFileSystem.GetRealPathsStr(const AVirtualPath: NString):
  TNStringList;
var
  I: Long;
begin
  if FNodes.Find(ConvertVPath(AVirtualPath), I) then
    Result := TNVFSNode(FNodes.Objects[I]).FRealPaths
  else
    Result := nil;
end;

function TNVirtualFileSystem.GetRealPathsIdx(Index: Long): TNStringList;
begin
  Result := TNVFSNode(FNodes.Objects[Index]).FRealPaths;
end;

function TNVirtualFileSystem.GetVirtualDir(Index: Long): NString;
begin
  Result := FNodes.Keys[Index];
end;

function TNVirtualFileSystem.GetVirtualDirCount: Long;
begin
  Result := FNodes.Count;
end;

function TNVirtualFileSystem.FindFileNode(const AFileName: NString): TNVFSNode;
var
  I: Long;
begin
  Result := nil;
  for I := 0 to FNodes.Count - 1 do
    if TNVFSNode(FNodes.Objects[I]).FileType(AFileName) <> ftInvalid then
    begin
      Result := TNVFSNode(FNodes.Objects[I]);
      Exit;
    end;
end;

function TNVirtualFileSystem.FindDirNode(const ADirName: NString): TNVFSNode;
var
  I: Long;
begin
  Result := nil;
  for I := 0 to FNodes.Count - 1 do
    if TNVFSNode(FNodes.Objects[I]).DirType(ADirName) <> dtInvalid then
    begin
      Result := TNVFSNode(FNodes.Objects[I]);
      Exit;
    end;
end;

function TNVirtualFileSystem.ConvertFName(const AName: NString): NString;
begin
  Result := Kernel.SwapPathDelims(AName, PathDelim);
  Result := TranslateVars(Result);
  Result := StrReplace(Result, PathDelim + PathDelim, PathDelim);
end;

function TNVirtualFileSystem.ConvertDName(const AName: NString): NString;
var
  Len: Long;
begin
  Result := Kernel.SwapPathDelims(AName, PathDelim);
  Result := TranslateVars(Result);
  Result := StrReplace(Result, PathDelim + PathDelim, PathDelim);
  Len := Length(Result);
  if (Len > 0) and (Result[Len] <> PathDelim) then
    Result := Result + PathDelim;
end;

function TNVirtualFileSystem.ConvertRPath(const APath: NString): NString;
var
  Len: Long;
begin
  Result := Kernel.SwapPathDelims(APath, PathDelim);
  Result := TranslateVars(Result);
  if not FileIsArchive(Result) then
  begin
    Len := Length(Result);
    if (Len > 0) and (Result[Len] <> PathDelim) then
      Result := Result + PathDelim;
  end;
  Result := StrReplace(Result, PathDelim + PathDelim, PathDelim);
end;

function TNVirtualFileSystem.ConvertVPath(const APath: NString): NString;
var
  Len: Long;
begin
  Result := Kernel.SwapPathDelims(APath, PathDelim);
  if not FileIsArchive(Result) then
  begin
    Len := Length(Result);
    if (Len > 0) and (Result[Len] <> PathDelim) then
      Result := Result + PathDelim;
  end;
  Result := StrReplace(Result, PathDelim + PathDelim, PathDelim);
end;

function TNVirtualFileSystem.CreateFile(const AFileName: NString;
  AMode: TFileMode): TNBinaryFile;
var
  I, J: Long;
  Node: TNVFSNode;
begin
  Result := nil;
  if AMode = fdRead then
    Exit;
  for I := 0 to FNodes.Count - 1 do
  begin
    Node := TNVFSNode(FNodes.Objects[I]);
    if PosNoCase(Node.FVirtualPath, AFileName, 1) = 1 then
    begin
      // AFileName contains Node.VirtualPath
      for J := 0 to Node.FRealPaths.Count - 1 do
        // try to find real directory rather than archive
        if Node.FRealPaths[J][Length(Node.FRealPaths[J])] = PathDelim then
        begin
          // real directory found - create disk file
          Result := TNDiskFile.Create(Node.FRealPaths[J] + StrReplace(AFileName,
            Node.FVirtualPath, ''), AMode);
          Exit;
        end
        else
        begin
          // no real directory found - we must add file to archive
          Result := TNArchiveFile.Create(Node.GetRelName(AFileName), AMode,
            TNArchive(FArchives.FindValue(Node.FRealPaths[J])));
          Exit;
        end;
    end;
  end;
  // AFileName is not in one of virtual directories - create disk file
  Result := TNDiskFile.Create(AFileName, AMode);
end;

function TNVirtualFileSystem.TranslateVars(const APath: NString): NString;
var
  Ex: NString;
begin
  Result := APath;
  Ex := ExeDir;
  Ex := Copy(Ex, 1, Length(Ex) - 1);
  if PosNoCase(PathVarExeDir, Result) > 0 then
    Result := StrReplace(Result, '$(' + PathVarExeDir + ')', Ex);
  if PosNoCase(PathVarRoot, Result) > 0 then
    Result := StrReplace(Result, '$(' + PathVarRoot + ')', ExtractFileDir(Ex));
end;

procedure TNVirtualFileSystem.DeleteEmptyNodes;
var
  I: Long;
begin
  I := 0;
  while I < FNodes.Count do
    if TNVFSNode(FNodes.Objects[I]).FRealPaths.Count = 0 then
      DeleteNode(I)
    else
      Inc(I);
end;

procedure TNVirtualFileSystem.DeleteNode(Index: Long);
begin
  if (Index > -1) and (Index < FNodes.Count) then
  begin
    FNodes.Objects[Index].Free;
    FNodes.Delete(Index);
  end;
end;

procedure TNVirtualFileSystem.AddDefaultDirectories;
begin
  Mount(BinDir, '$(EXEDIR)');
  Mount(LogDir, '$(EXEDIR)' + PathDelim + LogDir);
  Mount(ModuleDir, '$(EXEDIR)' + PathDelim + ModuleDir);
  Mount(RootDir, '$(ROOT)');
  Mount(SettingsDir, '$(ROOT)' + PathDelim + SettingsDir);
  Mount(DataDir, '$(ROOT)' + PathDelim + DataDir);
  Mount(GraphicDir, '$(ROOT)' + PathDelim + GraphicDir);
  Mount(MusicDir, '$(ROOT)' + PathDelim + MusicDir);
  Mount(SoundDir, '$(ROOT)' + PathDelim + SoundDir);
end;

function TNVirtualFileSystem.Open(const AFileName: NString;
  AMode: TFileMode): TNBinaryFile;
var
  Node: TNVFSNode;
  AName: NString;
begin
  Result := nil;
  AName := ConvertFName(AFileName);
  if SysUtils.FileExists(AName) then
    // file exists directly on disk
    Result := TNDiskFile.Create(AName, AMode)
  else
  begin
    // file is not directly on disk - search in virtual directories
    Node := FindFileNode(AName);
    if Node <> nil then
      case Node.FLastType of
        // file is disk file
        ftDiskFile: Result := TNDiskFile.Create(Node.GetRealFileName(AName),
            AMode);
        // file is in archive
        ftArchiveFile: Result := TNArchiveFile.Create(Node.GetRelName(AName),
            AMode, Node.FLastArchive);
      end
    else
      // file was not found on disk nor in virtual dirs - create new file
      Result := CreateFile(AName, AMode);
  end;
end;

procedure TNVirtualFileSystem.Close(var BinaryFile);
begin
  try
    FreeAndNil(BinaryFile);
  except
    // maybe BinaryFile is not class instance at all, is it?
  end;
end;

function TNVirtualFileSystem.DeleteFile(const AFileName: NString): Boolean;
var
  AName: NString;
  Node: TNVFSNode;
begin
  Result := False;
  AName := ConvertFName(AFileName);
  if SysUtils.FileExists(AName) then
    Result := SysUtils.DeleteFile(AName)
  else
  begin
    Node := FindFileNode(AName);
    if Node <> nil then
      Result := Node.DeleteFile(AName);
  end;
end;

function TNVirtualFileSystem.FileExists(const AFileName: NString): Boolean;
var
  AName: NString;
begin
  AName := ConvertFName(AFileName);
  Result := SysUtils.FileExists(AName);
  if not Result then
    Result := FindFileNode(AName) <> nil;
end;

function TNVirtualFileSystem.FileExistsOutsideVfs(const AFileName: NString):
  Boolean;
begin
  Result := SysUtils.FileExists(AFileName);
end;

function TNVirtualFileSystem.DirectoryExists(const ADirName: NString): Boolean;
var
  AName: NString;
begin
  AName := ConvertDName(ADirName);
  Result := SysUtils.DirectoryExists(AName);
  if not Result then
    Result := FindDirNode(AName) <> nil;
end;

function TNVirtualFileSystem.DirectoryExistsOutsideVfs(const ADirName: NString):
  Boolean;
begin
  Result := SysUtils.DirectoryExists(ADirName);
end;

function TNVirtualFileSystem.GetFullFileName(const AFileName: NString;
  OnlyExisting: Boolean): NString;
var
  Node: TNVFSNode;
  AName, VDir: NString;
  I: Long;
begin
  Result := '';
  AName := ConvertFName(AFileName);
  // if translated file really exists, return it
  if SysUtils.FileExists(AName) then
    Result := AName
  else
  begin
    // try to find file in one of virtual directories
    Node := FindFileNode(AName);
    if Node <> nil then
      Result := Node.GetRealFileName(AName);
    if (Result = '') and (not OnlyExisting) then
    begin
      // return new file name
      VDir := ConvertVPath(StrToken(AName, PathDelim));
      if FNodes.Find(VDir, I) then
        Result := ConvertFName(TNVFSNode(FNodes.Objects[I]).GetNewFileName(VDir
          + AName));
    end;
  end;
end;

function TNVirtualFileSystem.GetFullDirName(const ADirName: NString;
  OnlyExisting: Boolean): NString;
var
  Node: TNVFSNode;
  AName, VDir: NString;
  I: Long;
begin
  Result := '';
  AName := ConvertDName(ADirName);
  // if translated directory really exists, return it
  if SysUtils.DirectoryExists(AName) then
    Result := AName
  else
    // check if AName is directly name of some virtual directory
    if FNodes.Find(AName, I) then
      Result :=
        TNVFSNode(FNodes.Objects[I]).FRealPaths[TNVFSNode(FNodes.Objects[I]).GetFirstDiskRealPath]
    else
    begin
      // then try to find existing dir in one of virtual directories
      Node := FindDirNode(AName);
      if Node <> nil then
        Result := Node.GetRealDirName(AName);
      if (Result = '') and (not OnlyExisting) then
      begin
        // return new directory name
        VDir := ConvertVPath(StrToken(AName, PathDelim));
        if FNodes.Find(VDir, I) then
          Result := ConvertDName(TNVFSNode(FNodes.Objects[I]).GetNewDirName(VDir
            + AName));
      end;
    end;
end;

procedure TNVirtualFileSystem.Mount(const AVirtualPath, ARealPath: NString);
var
  Node: TNVFSNode;
  Arc: TNArchive;
  VPath, RPath: NString;
begin
  VPath := ConvertVPath(AVirtualPath);
  Node := TNVFSNode(FNodes.FindValue(VPath));
  if Node = nil then
  begin
    Node := TNVFSNode.Create(Self, VPath);
    FNodes.Add(Node.VirtualPath, Node);
  end;
  RPath := ConvertRPath(ARealPath);
  if FileIsArchive(RPath) then
  begin
    // real path is archive
    if (TNArchive(FArchives.FindValue(RPath)) = nil) and
      SysUtils.FileExists(RPath) then
    begin
      // archive is not opened - open it and add to cache
      Arc := FindArchiveFormat(Copy(ExtractFileExt(RPath), 2,
        MaxLong)).Open(RPath);
      FArchives.Add(RPath, Arc);
    end;
    Node.AddRealPath(RPath);
{$IFDEF DEBUG}
    Log.Add('VFS Mount: virtual("%s") real("%s")', [VPath, RPath]);
{$ENDIF}
  end
  else
    if DirectoryExists(RPath) then
    begin
      Node.AddRealPath(RPath);
{$IFDEF DEBUG}
      Log.Add('VFS Mount: virtual("%s") real("%s")', [VPath, RPath]);
{$ENDIF}
    end;
  // delete empty nodes
  DeleteEmptyNodes;
end;

procedure TNVirtualFileSystem.Unmount(const AVirtualPath,
  ARealPath: NString);
var
  Node: TNVFSNode;
begin
  Node := TNVFSNode(FNodes.FindValue(ConvertVPath(AVirtualPath)));
  if Node <> nil then
  begin
    Node.FRealPaths.Delete(Node.FRealPaths.IndexOf(ConvertRPath(ARealPath)));
{$IFDEF DEBUG}
    Log.Add('VFS Unmount: virtual("%s") real("%s")',
      [ConvertVPath(AVirtualPath), ConvertRPath(ARealPath)]);
{$ENDIF}
  end;
  // delete empty nodes
  DeleteEmptyNodes;
end;

procedure TNVirtualFileSystem.VirtualDirFileList(const AVirtualPath: NString;
  List: TNStringList);
var
  I: Long;
begin
  if FNodes.Find(AVirtualPath, I) then
    TNVFSNode(FNodes.Objects[I]).GetFileList(List);
end;

procedure TNVirtualFileSystem.VirtualDirFileList(Index: Long;
  List: TNStringList);
begin
  if (Index > -1) and (Index < FNodes.Count) then
    TNVFSNode(FNodes.Objects[Index]).GetFileList(List)
end;

{$ENDIF}


{*************************************************************************}
{                           Misc functions                                }
{*************************************************************************}


{$IFNDEF MODLIBRARY}

function RegisteredCompressions: TNFileModuleList;
begin
  if RegCompressions = nil then
    RegCompressions := TNFileModuleList.Create;
  Result := RegCompressions;
end;

procedure RegisterCompression(Format: TNCompression);
begin
  RegisteredCompressions.Add(Format);
end;

procedure RegisterCompressions(Formats: array of TNCompression);
var
  I: Long;
begin
  for I := Low(Formats) to High(Formats) do
    RegisterCompression(Formats[I]);
end;

procedure UnRegisterCompression(Format: TNCompression);
begin
  RegisteredCompressions.Remove(Format);
end;

procedure UnRegisterCompressions(Formats: array of TNCompression);
var
  I: Long;
begin
  for I := Low(Formats) to High(Formats) do
    UnRegisterCompression(Formats[I]);
end;

function RegisteredCompressionsCount: Long;
begin
  Result := RegisteredCompressions.Modules.Count;
end;

function FindCompression(const Ext: NString): TNCompression;
begin
  Result := TNCompression(RegisteredCompressions.Find(Ext));
end;

function FindCompression(Index: Long): TNCompression; overload;
begin
  Result := TNCompression(RegisteredCompressions.Modules[Index]);
end;

function RegisteredArchiveFormats: TNFileModuleList;
begin
  if RegArchiveFormats = nil then
    RegArchiveFormats := TNFileModuleList.Create;
  Result := RegArchiveFormats;
end;

procedure RegisterArchiveFormat(Format: TNArchiveFormat);
begin
  RegisteredArchiveFormats.Add(Format);
end;

procedure RegisterArchiveFormats(Formats: array of TNArchiveFormat);
var
  I: Long;
begin
  for I := Low(Formats) to High(Formats) do
    RegisterArchiveFormat(Formats[I]);
end;

procedure UnRegisterArchiveFormat(Format: TNArchiveFormat);
begin
  RegisteredArchiveFormats.Remove(Format);
end;

procedure UnRegisterArchiveFormats(Formats: array of TNArchiveFormat);
var
  I: Long;
begin
  for I := Low(Formats) to High(Formats) do
    UnRegisterArchiveFormat(Formats[I]);
end;

function FindArchiveFormat(const Ext: NString): TNArchiveFormat;
begin
  Result := TNArchiveFormat(RegisteredArchiveFormats.Find(Ext));
end;

function FileIsArchive(const FileName: NString): Boolean;
var
  Ext: NString;
begin
  Result := False;
  Ext := ExtractFileExt(FileName);
  if Ext <> '' then
    Result := FindArchiveFormat(Copy(Ext, 2, MaxLong)) <> nil;
end;

{$ENDIF}

function BuildFileList(Path: NString; Attr: Long;
  Files: TNStringList; Options: TFileListOptions): Boolean;
var
  FileMask: NString;
  RootDir: NString;
  Folders: TNStringList;
  CurrentItem: Long;
  Counter: Long;
  LocAttr: Long;

  procedure BuildFolderList;
  var
    FindInfo: TSearchRec;
    Rslt: Long;
  begin
    Counter := Folders.Count - 1;
    CurrentItem := 0;
    while CurrentItem <= Counter do
    begin
      // searching for subfolders
      Rslt := FindFirst(Folders[CurrentItem] + '*', faDirectory, FindInfo);
      try
        while Rslt = 0 do
        begin
          if (FindInfo.Name <> '.') and (FindInfo.Name <> '..') and
            (FindInfo.Attr and faDirectory = faDirectory) then
            Folders.Add(Folders[CurrentItem] + FindInfo.Name + PathDelim);
          Rslt := FindNext(FindInfo);
        end;
      finally
        FindClose(FindInfo);
      end;
      Counter := Folders.Count - 1;
      Inc(CurrentItem);
    end;
  end;

  procedure FillFileList(CurrentCounter: Long);
  var
    FindInfo: TSearchRec;
    Rslt: Long;
    CurrentFolder: NString;
  begin
    CurrentFolder := Folders[CurrentCounter];
    Rslt := FindFirst(CurrentFolder + FileMask, LocAttr, FindInfo);
    if flRelNames in Options then
      CurrentFolder := ExtractRelativePath(RootDir, CurrentFolder);
    try
      while Rslt = 0 do
      begin
        if (FindInfo.Name <> '.') and (FindInfo.Name <> '..') then
        begin
          if (flFullNames in Options) or (flRelNames in Options) then
            Files.Add(CurrentFolder + FindInfo.Name)
          else
            Files.Add(FindInfo.Name);
        end;
        Rslt := FindNext(FindInfo);
      end;
    finally
      FindClose(FindInfo);
    end;
  end;

begin
  FileMask := ExtractFileName(Path);
  RootDir := ExtractFilePath(Path);
  Folders := TNStringList.Create;
  Folders.Add(RootDir);
  Files.Clear;
{$WARN SYMBOL_PLATFORM OFF}
  if Attr = faAnyFile then
    LocAttr := faSysFile or faHidden or faArchive or faReadOnly
  else
    LocAttr := Attr;
{$WARN SYMBOL_PLATFORM ON}
  // here's the recursive search for nested folders
  if flRecursive in Options then
    BuildFolderList;
  if Attr <> faDirectory then
    for Counter := 0 to Folders.Count - 1 do
      FillFileList(Counter)
  else
    Files.AddStrings(Folders);
  Folders.Free;
  Result := True;
end;


{*************************************************************************}
{                         Internall Stuff                                 }
{*************************************************************************}


{$IFNDEF MODLIBRARY}

procedure LoadedCmp(Entry: PModuleEntry; var Done: Boolean);
var
  GetModule: TGetModuleProc;
  I, J: Long;
  Cls: TNCompressionClass;
  Inst: TNCompression;
  Check: Boolean;
begin
  if Entry^.ID = CmpLibraryID then
  try
    @GetModule := GetProcAddress(Entry^.Handle, sGetModuleProc);
    for I := 0 to Entry^.Count - 1 do
    begin
      GetModule(I, Pointer(Cls));
      Check := True;
      for J := 0 to RegisteredCompressions.Modules.Count - 1 do
        if TNCompression(
          RegisteredCompressions.Modules[J]).ClassNameIs(Cls.ClassName) then
        begin
          Check := False;
          Break;
        end;
      if Check then
      begin
        Inst := Cls.Create;
        if Inst <> nil then
          RegisterCompressions(Inst);
      end;
    end;
  except
    raise Exception.Create('Error when creating Compression from ' +
      'module library!');
  end;
end;

procedure UnLoadedCmp(Entry: PModuleEntry; var Done: Boolean);
var
  GetModule: TGetModuleProc;
  I, J: Long;
  Cls: TNCompressionClass;
  Inst: TNCompression;
begin
  if Entry^.ID = CmpLibraryID then
  try
    @GetModule := GetProcAddress(Entry^.Handle, sGetModuleProc);
    for I := 0 to Entry^.Count - 1 do
    begin
      GetModule(I, Pointer(Cls));
      for J := 0 to RegisteredCompressions.Modules.Count - 1 do
        if TNCompression(
          RegisteredCompressions.Modules[J]).ClassNameIs(Cls.ClassName) then
        begin
          Inst := TNCompression(RegisteredCompressions.Modules[J]);
          UnRegisterCompression(Inst);
          Inst.Free;
          Break;
        end;
    end;
  except
    raise Exception.Create('Error when unloading Compression from ' +
      'module library!');
  end;
end;

procedure LoadedArc(Entry: PModuleEntry; var Done: Boolean);
var
  GetModule: TGetModuleProc;
  I, J: Long;
  Cls: TNArchiveFormatClass;
  Inst: TNArchiveFormat;
  Check: Boolean;
begin
  if Entry^.ID = ArcLibraryID then
  try
    @GetModule := GetProcAddress(Entry^.Handle, sGetModuleProc);
    for I := 0 to Entry^.Count - 1 do
    begin
      GetModule(I, Pointer(Cls));
      Check := True;
      for J := 0 to RegisteredArchiveFormats.Modules.Count - 1 do
        if TNArchiveFormat(
          RegisteredArchiveFormats.Modules[J]).ClassNameIs(Cls.ClassName) then
        begin
          Check := False;
          Break;
        end;
      if Check then
      begin
        Inst := Cls.Create;
        if Inst <> nil then
          RegisterArchiveFormat(Inst);
      end;
    end;
  except
    raise Exception.Create('Error when creating ArchiveFormat from ' +
      'module library!');
  end;
end;

procedure UnLoadedArc(Entry: PModuleEntry; var Done: Boolean);
var
  GetModule: TGetModuleProc;
  I, J: Long;
  Cls: TNArchiveFormatClass;
  Inst: TNArchiveFormat;
begin
  if Entry^.ID = ArcLibraryID then
  try
    @GetModule := GetProcAddress(Entry^.Handle, sGetModuleProc);
    for I := 0 to Entry^.Count - 1 do
    begin
      GetModule(I, Pointer(Cls));
      for J := 0 to RegisteredArchiveFormats.Modules.Count - 1 do
        if TNArchiveFormat(
          RegisteredArchiveFormats.Modules[J]).ClassNameIs(Cls.ClassName) then
        begin
          Inst := TNArchiveFormat(RegisteredArchiveFormats.Modules[J]);
          UnRegisterArchiveFormat(Inst);
          Inst.Free;
          Break;
        end;
    end;
  except
    raise Exception.Create('Error when unloading ArchiveFormat from ' +
      'module library!');
  end;
end;

{$ENDIF}

initialization

{$IFNDEF MODLIBRARY}
  {Adds load/unload callbacks for ModuleRegister.}
  ModuleRegister.AddLoadCallBack(LoadedCmp);
  ModuleRegister.AddUnLoadCallBack(UnLoadedCmp);
  ModuleRegister.AddLoadCallBack(LoadedArc);
  ModuleRegister.AddUnLoadCallBack(UnLoadedArc);

  {Registration of default compression formats is put here instead of
   their original units to be sure that they will be registered in
   this sequence (first will be chosen as default format).}
  RegisterCompression(ZLibFormat.ZLibInstance);
  RegisterCompression(RleFormat.RleInstance);

  {Registration of default archive formats is put here instead of
   their original units to be sure that they will be registered in
   this sequence (first will be chosen as default format).}
  RegisterArchiveFormat(ZipFormat.ZIPInstance);
  RegisterArchiveFormat(TarFormat.TARInstance);
{$ENDIF}

finalization

{$IFNDEF MODLIBRARY}
  FreeAndNil(VfsInstance);
  FreeAndNil(RegCompressions);
  FreeAndNil(RegArchiveFormats);
{$ENDIF}

end.

