{*************************************************************************}
{*   Drawings                                                            *}
{*   Copyright (C) 2003 Marek Mauder, Vampyre SoftWorks                  *}
{*   Engine\Graphic\Drawings.pas                                         *}
{*************************************************************************}
{                                                                         }
{ This program is free software; you can redistribute it and/or modify it }
{ under the terms of the GNU General Public License as published by the   }
{ Free Software Foundation; either version 2 of the License, or (at your  }
{ option) any later version. This program is distributed in the hope that }
{ it will be useful, but WITHOUT ANY WARRANTY; without even the implied   }
{ warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See   }
{ the GNU General Public License for more details.                        }
{                                                                         }
{ You should have received a copy of the GNU General Public License       }
{ along with this program; if not, write to the Free Software Foundation, }
{ Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.         }
{                                                                         }
{*************************************************************************}
{  Created On: 11st September 2003  (11.09.2003)                          }
{*************************************************************************}

{This unit contains amount of image drawing and effect functions.
 Drawing functions are provided only for ImageDataFormats which
 has field SupportsDrawing set to True.}
{Note of speed: Drawing is always FASTEST in fixed pixel format images
 like A8R8G8B8 and R8G8B8.
 When copying from one image to another and both images are
 in same format then it is a LOT FASTER.}
{Note of parameters: If color parameter name is ARGB it means
 that color will be translated to real value stored in image first
 (pixel format conversions in PF formats, search in palette in indexed
 formats, conversion from RGB in intensity formats, ...).
 If color parameter name is Color it means that this value will
 be directly written to image bits.}

unit Drawings;

{TODO -oGalfar: Add base drawing/copy functions:
Fill,CopyRect,Circle,Line,Box,Pixels,< StretchCopy >, HorzLine/VertLine}
{TODO -oGalfar: Speed up CopyRect with diff formats.}
{DONE -oGalfar: Add functions: Negative, Mirror)}
{TODO -oGalfar: Add image drawing: DrawAdditive,DrawAlpha, DrawStretch, ...
RotateLeft,RotateRight}
{TODO -oGalfar: Add support for color-keying in drawing functions.}

{$I Compiler.inc}

interface

uses
  Kernel, Objects, Platforms, Graphic, Images;


{*************************************************************************}
{                         Pixel/Color functions                           }
{*************************************************************************}


{This function translates ARGB color to Format specific color.
It uses Format's TransARGBFunc to conversion. If this func is
nil returns unchanged input.}
function TranslateARGB(Format: PImageDataFormat; ARGB: TARGB;
  Pal: PPalArray = nil): ULong;
{This function translates Format specific color to ARGB color.
It uses Format's TransColorFunc to conversion. If this func is
nil returns unchanged input.}
function TranslateColor(Format: PImageDataFormat; Color: ULong;
  Pal: PPalArray =  nil): TARGB;

{This returns number of unique colors (not with alpha) in image.}
function CountColors(var Level: TImageLevel; Format: PImageDataFormat): ULong;

{This returns pixel color at [X,Y] translated to ARGB format.
Pal is needed only when translating color in indexed mode images.}
function GetPixelARGB(const Level: TImageLevel; Format: PImageDataFormat;
  X, Y: Long; Pal: PPalArray = nil): TARGB;
{This returns pixel color at [X,Y] in raw Format specific format.}
function GetPixelColor(const Level: TImageLevel; Format: PImageDataFormat;
  X, Y: Long): ULong;
{This sets pixel color at [X,Y]. It translates ARGB input to
Format specific color.
Pal is needed only when translating color in indexed mode images.}
procedure SetPixelARGB(var Level: TImageLevel; Format: PImageDataFormat;
  X, Y: Long; ARGB: TARGB; Pal: PPalArray = nil);
{This sets pixel color at [X,Y] in raw Format specific format.}
procedure SetPixelColor(var Level: TImageLevel; Format: PImageDataFormat;
  X, Y: Long; Color: ULong);


{*************************************************************************}
{                   Fill/Modify Image functions                           }
{*************************************************************************}


{Fills all levels of image with ARGB color.}
procedure FillARGB(var Image: TImageData; ARGB: TARGB);
{Fills all levels of image with Image DataFormat's specific color.}
procedure FillColor(var Image: TImageData; Color: ULong);
{Fills image level of image with Format's specific color.
if FillSubLevels is True all Level's sublevels are filled too.}
procedure FillLevelColor(var Level: TImageLevel; Format: PImageDataFormat;
  Color: ULong; FillSubLevels: Boolean = True);
{Fills rectangle with ARGB color.}
procedure FillRectARGB(var Level: TImageLevel; Format: PImageDataFormat; X, Y,
  Width, Height: Long; ARGB: TARGB; Pal: PPalArray = nil);
{Fills rectangle with Format specific color.}
procedure FillRectColor(var Level: TImageLevel; Format: PImageDataFormat; X, Y,
  Width, Height: Long; Color: ULong);

{Negates colors (no alpha) of given image. In indexed modes
it negates palette entires. In other modes it calls NegateLevel() for
each Level.}
procedure NegativeImage(var Image: TImageData);
{Negates colors (no alpha) of given level.}
procedure NegativeLevel(var Level: TImageLevel; NegateSubLevels:Boolean =True);


{*************************************************************************}
{                      Shapes drawing functions                           }
{*************************************************************************}


{Draws circle with ARGB color.}
procedure CircleARGB(var Level: TImageLevel; Format: PImageDataFormat; X, Y,
  Diameter: Long; ARGB: TARGB; Pal: PPalArray = nil);
{Draws circle with Format specific color.}
procedure CircleColor(var Level: TImageLevel; Format: PImageDataFormat; X, Y,
  Diameter: Long; Color: ULong);
{Draws ellipse with ARGB color.}
procedure EllipseARGB(var Level: TImageLevel; Format: PImageDataFormat; X, Y,
  Width, Height: Long; ARGB: TARGB; Pal: PPalArray = nil);
{Draws ellipse with Format specific color.}
procedure EllipseColor(var Level: TImageLevel; Format: PImageDataFormat; X, Y,
  Width, Height: Long; Color: ULong);

{Draws line from [X1,Y1] to [X2,Y2] with ARGB color.}
procedure LineARGB(var Level: TImageLevel; Format: PImageDataFormat; X1, Y1,
  X2, Y2: Long; ARGB: TARGB; Pal: PPalArray = nil);
{Draws line from [X1,Y1] to [X2,Y2] with fomat specific color.}
procedure LineColor(var Level: TImageLevel; Format: PImageDataFormat; X1, Y1,
  X2, Y2: Long; Color: ULong);
{Draws horizontal line from [X1,Y] to [X2,Y] with ARGB color.
 Its is 30x faster than using Line to draw horiz line.}
procedure HLineARGB(var Level: TImageLevel; Format: PImageDataFormat; X1, X2,
  Y: Long; ARGB: TARGB; Pal: PPalArray = nil);
{Draws horizontal line from [X1,Y] to [X2,Y] with fomat specific color.
 Its is 30x faster than using Line to draw horiz line.}
procedure HLineColor(var Level: TImageLevel; Format: PImageDataFormat; X1, X2,
  Y: Long; Color: ULong);
{Draws vertical line from [X,Y1] to [X,Y2] with ARGB color.
 Its is 20% faster than using Line to draw vert line.}
procedure VLineARGB(var Level: TImageLevel; Format: PImageDataFormat; X, Y1,
  Y2: Long; ARGB: TARGB; Pal: PPalArray = nil);
{Draws vertical line from [X,Y1] to [X,Y2] with Format specific color.
 Its is 20% faster than using Line to draw vert line.}
procedure VLineColor(var Level: TImageLevel; Format: PImageDataFormat; X, Y1,
  Y2: Long; Color: ULong);
{Draws polyline defined by Points with ARGB color.}
procedure PolylineARGB(var Level: TImageLevel; Format: PImageDataFormat;
  Points: array of TPoint; ARGB: TARGB; Pal: PPalArray = nil);
{Draws polyline defined by Points with Format specific color.}
procedure PolylineColor(var Level: TImageLevel; Format: PImageDataFormat;
  Points: array of TPoint; Color: ULong);

{Draws rectangle with ARGB color.}
procedure RectangleARGB(var Level: TImageLevel; Format: PImageDataFormat; X, Y,
  Width, Height: Long; ARGB: TARGB; Pal: PPalArray = nil);
{Draws rectangle with Format specific color.}
procedure RectangleColor(var Level: TImageLevel; Format: PImageDataFormat; X, Y,
  Width, Height: Long; Color: ULong);


{*************************************************************************}
{                   Copy/Draw Image functions                             }
{*************************************************************************}


{Copies rect from one image level (Src) to another (Dest). [SX, SY] is
 position in source level, [Width, Height] is size of rect, [DX,DY]
 is position in destination level. Palettes are used when
 one or both of images are in indexed mode. When copying
 from one format to another ARGB >> color conversiorns are done.
 It copies raw color data - no blending, colorkeying, ... is done.}
{Note: copying is 30x faster when Dest and Src Formats are same.}
procedure CopyRectARGB(const Src: TImageLevel; var Dest: TImageLevel; SrcFormat,
  DestFormat: PImageDataFormat; SX, SY, Width, Height, DX, DY: Long;
  SrcPal: PPalArray = nil; DestPal: PPalArray = nil);
{Copies rect from one image level (Src) to another (Dest). [SX, SY] is
 position in source level, [Width, Height] is size of rect, [DX,DY]
 is position in destination level. When copying
 from one format to another NO ARGB >> color conversiorns are done.
 It copies raw color data - no blending, colorkeying, ... is done.}
{Note: copying is 20x faster when Dest and Src BitCounts are same.}
procedure CopyRectColor(const Src: TImageLevel; var Dest: TImageLevel;
  SrcFormat, DestFormat: PImageDataFormat; SX, SY, Width, Height, DX, DY: Long);


{*************************************************************************}
{                             IMPLEMEMTATION                              }
{*************************************************************************}


implementation


{*************************************************************************}
{                         Pixel/Color functions                           }
{*************************************************************************}


function TranslateARGB(Format: PImageDataFormat; ARGB: TARGB; Pal: PPalArray):
  ULong;
begin
  if Assigned(Format^.TransARGBFunc) then
    Result := Format^.TransARGBFunc(Format, ARGB, Pal)
  else
    Result := ARGB;
end;

function TranslateColor(Format: PImageDataFormat; Color: ULong; Pal: PPalArray):
  TARGB;
begin
  if Assigned(Format^.TransColorFunc) then
    Result := Format^.TransColorFunc(Format, Color, Pal)
  else
    Result := Color;
end;

function CountColors(var Level: TImageLevel; Format: PImageDataFormat): ULong;

  function Count8: Long;
  var
    I: Long;
    PB, PC: PByte;
    Check: array[Byte] of Byte;
  begin
    Result := 0;
    ZeroMemory(@Check, SizeOf(Check));
    PB := Level.Bits;
    for I := 0 to Level.Width * Level.Height - 1 do
    begin
      PC := @Check[PB^];
      if PC^ = 0 then
      begin
        Inc(Result);
        PC^ := 1;
      end;
      if Result = 256 then Exit;
      Inc(PB);
    end;
  end;

  function Count16: Long;
  var
    PW: PWord;
    PC: PByte;
    I: Long;
    Check: array[Word] of Byte;
  begin
    Result := 0;
    ZeroMemory(@Check, SizeOf(Check));
    PW := Level.Bits;
    for I := 0 to Level.Width * Level.Height - 1 do
    begin
      PC := @Check[PW^];
      if PC^ = 0 then
      begin
        Inc(Result);
        PC^ := 1;
      end;
      Inc(PW);
    end;
  end;

  function Count24: Long;
  type
    PCheck = ^TCheck;
    TCheck = array[Byte, Byte, 0..31] of Byte;
  var
    PB: PByte;
    PC: PRGBRec;
    Check: PCheck;
    I, C: Long;
  begin
    Result := 0;
    New(Check);
    ZeroMemory(Check, SizeOf(TCheck));
    PC := Level.Bits;
    for I := 0 to Level.Width * Level.Height - 1 do
    begin
      PB := @Check[PC.R, PC.G, PC.B shr 3];
      C := 1 shl (PC.B and 7);
      if (C and PB^) = 0 then
      begin
        Inc(Result);
        PB^ := PB^ or C;
      end;
      Inc(PC);
    end;
    Dispose(Check);
  end;

  function Count32: Long;
  type
    PCheck = ^TCheck;
    TCheck = array[Byte, Byte, 0..31] of Byte;
  var
    PB: PByte;
    PC: PARGBRec;
    I, C: Long;
    Check: PCheck;
  begin
    Result := 0;
    New(Check);
    ZeroMemory(Check, SizeOf(TCheck));
    PC := Level.Bits;
    for I := 0 to Level.Width * Level.Height - 1 do
    begin
      PB := @Check[PC.R, PC.G, PC.B shr 3];
      C := 1 shl (PC.B and 7);
      if (C and PB^) = 0 then
      begin
        Inc(Result);
        PB^ := PB^ or C;
      end;
      Inc(PC);
    end;
    Dispose(Check);
  end;

begin
  case Format^.BitCount of
    8: Result := Count8;
    16: Result := Count16;
    24: Result := Count24;
    32: Result := Count32
    else
      Result := 0;
  end;
end;

function GetPixelARGB(const Level: TImageLevel; Format: PImageDataFormat;
  X, Y: Long; Pal: PPalArray): TARGB;
begin
  Result := GetPixelColor(Level, Format, X, Y);
  if Assigned(Format^.TransColorFunc) then
    Result := Format^.TransColorFunc(Format, Result, Pal);
end;

function GetPixelColor(const Level: TImageLevel; Format: PImageDataFormat;
  X, Y: Long): ULong;
begin
  Result := 0;
  if (X >= 0) and (X < Level.Width) and (Y >= 0) and (Y < Level.Height) then
    with Level do
      case Format^.BitCount of
        8: Result := PByteArray(Long(Bits) + Y * (Width))[X];
        16: Result := PWordArray(Long(Bits) + Y * (Width shl 1))[X];
        24: with PRGBRecArray(Long(Bits) + Y * Width * 3)[X] do
            Result := (R shl 16) or (G shl 8) or B;
        32: Result := PULongArray(Long(Bits) + Y * (Width shl 2))[X];
      end;
end;

procedure SetPixelARGB(var Level: TImageLevel; Format: PImageDataFormat;
  X, Y: Long; ARGB: TARGB; Pal: PPalArray);
begin
  if Assigned(Format^.TransARGBFunc) then
    ARGB := Format^.TransARGBFunc(Format, ARGB, Pal);
  SetPixelColor(Level, Format, X, Y, ARGB);
end;

procedure SetPixelColor(var Level: TImageLevel; Format: PImageDataFormat;
  X, Y: Long; Color: ULong);
begin
  if (X >= 0) and (X < Level.Width) and (Y >= 0) and (Y < Level.Height) then
    with Level do
      case Format^.BitCount of
        8: PByteArray(Long(Bits) + Y * Width)[X] := Color;
        16: PWordArray(Long(Bits) + Y * (Width shl 1))[X] := Color;
        24: PRGBRecArray(Long(Bits) + Y * Width * 3)[X] := PRGBRec(@Color)^;
        32: PULongArray(Long(Bits) + Y * (Width shl 2))[X] := Color;
      end;
end;


{*************************************************************************}
{                         Fill/Modify Image functions                     }
{*************************************************************************}


procedure FillARGB(var Image: TImageData; ARGB: TARGB);
begin
  FillColor(Image, TranslateARGB(Image.DataFormat, ARGB, Image.Palette));
end;

procedure FillColor(var Image: TImageData; Color: ULong);
var
  I: Long;
begin
  for I := 0 to Image.Count - 1 do
    FillLevelColor(Image.Levels[I], Image.DataFormat, Color, True);
end;

procedure FillLevelColor(var Level: TImageLevel; Format: PImageDataFormat;
  Color: ULong; FillSubLevels: Boolean);
var
  I, Incr: Long;
  C: PRGBRec;
begin
  case Format^.BitCount of
    8: FillMemory(Level.Bits, Level.Size, Color);
    16: FillMemoryWord(Level.Bits, Level.Size, Color);
    24:
      with Level do
      begin
        C := Pointer(Bits);
        Incr := Width * 3;
        for I := 0 to Width - 1 do
        begin
          C^ := PRGBRec(@Color)^;
          Inc(C);
        end;
        for I := 1 to Height - 1 do
          Move(Bits^, Pointer(Long(Bits) + I * Incr)^, Incr);
      end;
    32: FillMemoryULong(Level.Bits, Level.Size, Color);
  end;
  if FillSubLevels then
    for I := 0 to Level.Count - 1 do
      FillLevelColor(Level, Format, Color, FillSubLevels);
end;

procedure FillRectARGB(var Level: TImageLevel; Format: PImageDataFormat; X, Y,
  Width, Height: Long; ARGB: TARGB; Pal: PPalArray);
begin
  FillRectColor(Level, Format, X, Y, Width, Height,
    TranslateARGB(Format, ARGB, Pal));
end;

procedure FillRectColor(var Level: TImageLevel; Format: PImageDataFormat; X, Y,
  Width, Height: Long; Color: ULong);
var
  I, LWidth, Bpp: Long;
  C: Long;
  Cr: PRGBREc;
begin
  if X < 0 then X := 0;
  if X + Width >= Level.Width then Width := Level.Width - X;
  if Y < 0 then Y := 0;
  if Y + Height >= Level.Height then Height := Level.Height - Y;

  Bpp := Format^.BitCount shr 3;
  Width := Width * Bpp;
  LWidth := Level.Width * Bpp;
  X := X * Bpp;
  C := Long(Pointer(Long(Level.Bits) + Y * LWidth)) + X;
  case Format^.BitCount of
    8:
      for I := 0 to Height - 1 do
      begin
        FillMemory(Ptr(C), Width, Color);
        Inc(C, LWidth);
      end;
    16:
      for I := 0 to Height - 1 do
      begin
        FillMemoryWord(Ptr(C), Width, Color);
        Inc(C, LWidth);
      end;
    24:
      begin
        Cr := Ptr(C);
        for I := 0 to Width div 3 - 1 do
        begin
          Cr^ := PRGBRec(@Color)^;
          Inc(Cr);
        end;
        Inc(C, LWidth);
        for I := 1 to Height - 1 do
        begin
          Move(Pointer(Long(Cr) - Width)^, Ptr(C)^, Width);
          Inc(C, LWidth);
        end;
      end;
    32:
      for I := 0 to Height - 1 do
      begin
        FillMemoryULong(Ptr(C), Width, Color);
        Inc(C, LWidth);
      end;
  end;
end;

procedure NegativeImage(var Image: TImageData);
var
  I: Long;
begin
  if Image.DataFormat^.Indexed then
    for I := 0 to Image.DataFormat^.PalSize - 1 do
    begin
      Image.Palette[I].R := 255 - Image.Palette[I].R;
      Image.Palette[I].G := 255 - Image.Palette[I].G;
      Image.Palette[I].B := 255 - Image.Palette[I].B;
    end
  else
    for I := 0 to Image.Count - 1 do
      NegativeLevel(Image.Levels[I],True);
end;

procedure NegativeLevel(var Level: TImageLevel;NegateSubLevels:Boolean);
var
  I: Long;
begin
  {$IFDEF ASMVERSIONS}
  asm
    MOV    ECX, [EAX].TImageLevel.Size
    MOV    EAX, [EAX].TImageLevel.Bits
    MOV    EDX, ECX
  @Qwskip:
    SHR    ECX, 2
    JZ     @Dwskip
    DEC    ECX
  @Dwloop:
    NOT    DWORD PTR [EAX + ECX * 4]
    DEC    ECX
    JNL    @Dwloop
    MOV    ECX, EDX
    SHR    ECX, 2
  @Dwskip:
    MOV    ECX, EDX
    AND    ECX, 3
    JZ     @Byskip
    DEC    ECX
  @Loopby:
    NOT    BYTE PTR [EAX + ECX]
    DEC    ECX
    JNL    @Loopby
  @Byskip:
  end;
  {$ELSE}
  for I := 0 to Level.Size - 1 do
    PByteArray(Level.Bits)[I] := not PByteArray(Level.Bits)[I];
  {$ENDIF}
  if NegateSubLevels then
  for I := 0 to Level.Count - 1 do
    NegativeLevel(Level,NegateSubLevels);
end;


{*************************************************************************}
{                         Shapes drawing functions                        }
{*************************************************************************}


procedure CircleARGB(var Level: TImageLevel; Format: PImageDataFormat; X, Y,
  Diameter: Long; ARGB: TARGB; Pal: PPalArray);
begin
  EllipseColor(Level, Format, X, Y, Diameter, Diameter,
    TranslateARGB(Format, ARGB, Pal));
end;

procedure CircleColor(var Level: TImageLevel; Format: PImageDataFormat; X, Y,
  Diameter: Long; Color: ULong);
begin
  EllipseColor(Level, Format, X, Y, Diameter, Diameter, Color);
end;

procedure EllipseARGB(var Level: TImageLevel; Format: PImageDataFormat; X, Y,
  Width, Height: Long; ARGB: TARGB; Pal: PPalArray);
begin
  EllipseColor(Level, Format, X, Y, Width, Height,
    TranslateARGB(Format, ARGB, Pal));
end;

procedure EllipseColor(var Level: TImageLevel; Format: PImageDataFormat; X, Y,
  Width, Height: Long; Color: ULong);
var
  XX, Mx1, Mx2, My1, My2: Long;
  Aq, Bq, Dx, Dy, R, Rx, Ry: Long;
begin
  SetPixelColor(Level, Format, X + Width, Y, Color);
  SetPixelColor(Level, Format, X - Width, Y, Color);
  Mx1 := X - Width;
  Mx2 := X + Width;
  My1 := Y;
  My2 := Y;
  Aq := Width * Width;
  Bq := Height * Height;
  Dx := Aq shl 1;
  Dy := Bq shl 1;
  R := Width * Bq;
  Rx := R shl 1;
  Ry := 0;
  Xx := Width;
  while Xx > 0 do
  begin
    if R > 0 then
    begin
      Inc(My1);
      Dec(My2);
      Inc(Ry, Dx);
      Dec(R, Ry);
    end;
    if R <= 0 then
    begin
      Dec(Xx);
      Inc(Mx1);
      Dec(Mx2);
      Dec(Rx, Dy);
      Inc(R, RX);
    end;
    SetPixelColor(Level, Format, Mx1, My1, Color);
    SetPixelColor(Level, Format, Mx1, My2, Color);
    SetPixelColor(Level, Format, Mx2, My1, Color);
    SetPixelColor(Level, Format, Mx2, My2, Color);
  end;
end;

procedure LineARGB(var Level: TImageLevel; Format: PImageDataFormat; X1, Y1,
  X2, Y2: Long; ARGB: TARGB; Pal: PPalArray);
begin
  LineColor(Level, Format, X1, Y1, X2, Y2, TranslateARGB(Format, ARGB, Pal));
end;

procedure LineColor(var Level: TImageLevel; Format: PImageDataFormat; X1, Y1,
  X2, Y2: Long; Color: ULong);
var
  I, S, D1X, D1Y, D2X, D2Y, U, V, M, N: Long;
begin
  U := X2 - X1;
  V := Y2 - Y1;
  D1X := Sign(u);
  D1Y := Sign(v);
  D2X := Sign(u);
  D2Y := 0;
  M := Abs(U);
  N := Abs(V);
  if not (M > N) then
  begin
    D2X := 0;
    D2Y := Sign(V);
    M := Abs(V);
    N := Abs(U);
  end;
  S := M shr 1;
  for I := 0 to M do
  begin
    SetPixelColor(Level, Format, X1, Y1, Color);
    S := S + N;
    if not (S < M) then
    begin
      S := S - M;
      X1 := X1 + D1X;
      Y1 := Y1 + D1Y;
    end
    else
    begin
      X1 := X1 + D2X;
      Y1 := Y1 + D2Y;
    end;
  end;
end;

procedure HLineARGB(var Level: TImageLevel; Format: PImageDataFormat; X1, X2,
  Y: Long; ARGB: TARGB; Pal: PPalArray);
begin
  HLineColor(Level, Format, X1, X2, Y, TranslateARGB(Format, ARGB, Pal));
end;

procedure HLineColor(var Level: TImageLevel; Format: PImageDataFormat; X1, X2,
  Y: Long; Color: ULong);
begin
  if X2 < X1 then Swap(X1, X2);
  FillRectColor(Level, Format, X1, Y, X2 - X1, 1, Color);
end;

procedure VLineARGB(var Level: TImageLevel; Format: PImageDataFormat; X, Y1,
  Y2: Long; ARGB: TARGB; Pal: PPalArray);
begin
  VLineColor(Level, Format, X, Y1, Y2, TranslateARGB(Format, ARGB, Pal));
end;

procedure VLineColor(var Level: TImageLevel; Format: PImageDataFormat; X, Y1,
  Y2: Long; Color: ULong);
begin
  if Y2 < Y1 then Swap(Y1, Y2);
  FillRectColor(Level, Format, X, Y1, 1, Y2 - Y1, Color);
end;

procedure PolylineARGB(var Level: TImageLevel; Format: PImageDataFormat;
  Points: array of TPoint; ARGB: TARGB; Pal: PPalArray);
var
  N, I: Long;
  Color: ULong;
begin
  N := High(Points) + 1;
  Color := TranslateARGB(Format, ARGB, Pal);
  for I := 0 to N - 1 do
    LineColor(Level, Format, Points[I].X, Points[I].Y, Points[(I + 1) mod N].X,
      Points[(I + 1) mod N].Y, Color);
end;

procedure PolylineColor(var Level: TImageLevel; Format: PImageDataFormat;
  Points: array of TPoint; Color: ULong);
var
  N, I: Long;
begin
  N := High(Points) + 1;
  for I := 0 to N - 1 do
    LineColor(Level, Format, Points[I].X, Points[I].Y, Points[(I + 1) mod N].X,
      Points[(I + 1) mod N].Y, Color);
end;

procedure RectangleARGB(var Level: TImageLevel; Format: PImageDataFormat; X, Y,
  Width, Height: Long; ARGB: TARGB; Pal: PPalArray = nil);
begin
  PolyLineColor(Level, Format, [Point(X, Y), Point(X + Width, Y), Point(X +
      Width, Y + Height), Point(X, Y + Height), Point(X, Y)],
    TranslateARGB(Format, ARGB, Pal));
end;

procedure RectangleColor(var Level: TImageLevel; Format: PImageDataFormat; X, Y,
  Width, Height: Long; Color: ULong);
begin
  PolyLineColor(Level, Format, [Point(X, Y), Point(X + Width, Y), Point(X +
      Width, Y + Height), Point(X, Y + Height), Point(X, Y)], Color);
end;


{*************************************************************************}
{                         Copy/Draw Image functions                       }
{*************************************************************************}


procedure CopyRectARGB(const Src: TImageLevel; var Dest: TImageLevel; SrcFormat,
  DestFormat: PImageDataFormat; SX, SY, Width, Height, DX, DY: Long;
  SrcPal: PPalArray; DestPal: PPalArray);
var
  SC, SWidth, DC, DWidth, I, Bpp, X, Y: Long;
begin
  if SX < 0 then SX := 0;
  if SY < 0 then SY := 0;
  if DX < 0 then DX := 0;
  if DY < 0 then DY := 0;
  if SX + Width > Src.Width then Width := Src.Width - SX;
  if SY + Height > Src.Height then Height := Src.Height - SY;
  if DX + Width > Dest.Width then Width := Dest.Width - DX;
  if DY + Height > Dest.Height then Height := Dest.Height - DY;

  if DestFormat = SrcFormat then
  begin
    Bpp := DestFormat^.BitCount shr 3;
    SWidth := Src.Width * Bpp;
    DWidth := Dest.Width * Bpp;
    Width := Width * Bpp;
    SX := SX * Bpp;
    DX := DX * Bpp;
    SC := Long(Pointer(Long(Src.Bits) + SY * SWidth)) + SX;
    DC := Long(Pointer(Long(Dest.Bits) + DY * DWidth)) + DX;
    for I := 0 to Height - 1 do
    begin
      Move(Ptr(SC)^, Ptr(DC)^, Width);
      Inc(SC, SWidth);
      Inc(DC, DWidth);
    end;
  end
  else
  begin
    for X := 0 to Width - 1 do
      for Y := 0 to Height - 1 do
        SetPixelARGB(Dest, DestFormat, X + DX, Y + DY, GetPixelARGB(Src,
          SrcFormat, X + SX, Y + SY, SrcPal), DestPal);
  end;
end;

procedure CopyRectColor(const Src: TImageLevel; var Dest: TImageLevel;
  SrcFormat,
  DestFormat: PImageDataFormat; SX, SY, Width, Height, DX, DY: Long);
var
  SC, SWidth, DC, DWidth, I, Bpp, X, Y: Long;
begin
  if SX < 0 then SX := 0;
  if SY < 0 then SY := 0;
  if DX < 0 then DX := 0;
  if DY < 0 then DY := 0;
  if SX + Width > Src.Width then Width := Src.Width - SX;
  if SY + Height > Src.Height then Height := Src.Height - SY;
  if DX + Width > Dest.Width then Width := Dest.Width - DX;
  if DY + Height > Dest.Height then Height := Dest.Height - DY;

  if DestFormat^.BitCount = SrcFormat^.BitCount then
  begin
    Bpp := DestFormat^.BitCount shr 3;
    SWidth := Src.Width * Bpp;
    DWidth := Dest.Width * Bpp;
    Width := Width * Bpp;
    SX := SX * Bpp;
    DX := DX * Bpp;
    SC := Long(Pointer(Long(Src.Bits) + SY * SWidth)) + SX;
    DC := Long(Pointer(Long(Dest.Bits) + DY * DWidth)) + DX;
    for I := 0 to Height - 1 do
    begin
      Move(Ptr(SC)^, Ptr(DC)^, Width);
      Inc(SC, SWidth);
      Inc(DC, DWidth);
    end;
  end
  else
  begin
    for X := 0 to Width - 1 do
      for Y := 0 to Height - 1 do
        SetPixelColor(Dest, DestFormat, X + DX, Y + DY, GetPixelColor(Src,
          SrcFormat, X + SX, Y + SY));
  end;
end;


end.

