{*************************************************************************}
{*   Graphic                                                             *}
{*   Copyright (C) 2003 Marek Mauder, Vampyre SoftWorks                  *}
{*   Engine\Graphic\Graphic.pas                                          *}
{*************************************************************************}
{                                                                         }
{ This program is free software; you can redistribute it and/or modify it }
{ under the terms of the GNU General Public License as published by the   }
{ Free Software Foundation; either version 2 of the License, or (at your  }
{ option) any later version. This program is distributed in the hope that }
{ it will be useful, but WITHOUT ANY WARRANTY; without even the implied   }
{ warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See   }
{ the GNU General Public License for more details.                        }
{                                                                         }
{ You should have received a copy of the GNU General Public License       }
{ along with this program; if not, write to the Free Software Foundation, }
{ Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.         }
{                                                                         }
{*************************************************************************}
{  Created On: 11st September 2003  (11.09.2003)                          }
{*************************************************************************}

{This unit contains base graphic classes, types and constants
and functions.}

unit Graphic;

{DONE -oGalfar: Add PixelFormat support. Test it first!}
{DONE -oGalfar: Add support for 64/48 bit and foating point images.}

{$I Compiler.inc}

interface

uses
  SysUtils, Kernel, Objects, Platforms;


{*************************************************************************}
{                         Color and other types/consts                    }
{*************************************************************************}


type

  {Base color type. Native color format is ARGB (or BGRA somewhere,
   it depends from which side you count bytes)). This color type
   supports max 32bit images with channels max 8 bit wide.
   Channel:           |A|R|G|B|
   Byte:      highest |3|2|1|0| lowest}
  TARGB = ULong;

  {RGB Color type. Highest byte is ignored.}
  TRGB = ULong;

  {BGR Color type. Highest byte is ignored. This format is
   used in Windows as color value (TColor from VCL is this type).}
  TBGR = ULong;

  {Record for easy access to color channels of 32bit ARGB Color.}
  TARGBRec = record
    case Integer of
      0: (Color: TARGB);
      1: (B, G, R, A: Byte);
  end;

  {Record for easy access to color channels of 64bit ARGB Color.}
  TARGB64Rec = record
    case Integer of
      0: (BG, RA: ULong);
      1: (B, G, R, A: Word);
  end;

  {Record for easy access to color channels of
   128bit floating-point ARGB Color.}
  TARGBFRec = record
    B, G, R, A: Single;
  end;

  {Record for representing 24bit RGB color.}
  TRGBRec = record
    B, G, R: Byte;
  end;

  {Record for representing 48bit RGB color.}
  TRGB48Rec = record
    B, G, R: Word;
  end;

  {Record for representing 24bit BGR color.}
  TBGRRec = record
    R, G, B: Byte;
  end;

  PBGR = ^TBGR;
  PARGB = ^TARGB;
  PARGBRec = ^TARGBRec;
  PARGB64Rec = ^TARGB64Rec;
  PRGB48Rec = ^TRGB48Rec;
  PARGBFRec = ^TARGBFRec;
  PRGBRec = ^TRGBRec;
  PBGRRec = ^TBGRRec;

  {Array of ARGB colors. Used mainly as scanline when reading/writing images.}
  TARGBRecArray = array[0..MaxLongArraySize - 1] of TARGBRec;
  PARGBRecArray = ^TARGBRecArray;

  {Array of RGB colors. Used mainly as scanline when reading/writing images.}
  TRGBRecArray = array[0..MaxLongArraySize - 1] of TRGBRec;
  PRGBRecArray = ^TRGBRecArray;

  {Pallette for indexed mode images.
   Alpha will be ignored in the most cases, but ...}
  TPalette = array[0..255] of TARGBRec;
  PPalette = ^TPalette;

  {This custom size palette can be used for some indexed images
   with weird bitcounts or other.}
  TPalArray = array[0..MaxLongArraySize - 1] of TARGBRec;
  PPalArray = ^TPalArray;

  {Pallette for indexed mode images.
   This palette is used for reading/writting from indexed image formats}
  TPalRGB = array[0..255] of TRGBRec;
  PPalRGB = ^TPalRGB;

  {This custom size palette can be used for some indexed images
   with weird bitcounts or other.}
  TPalRGBArray = array[0..MaxLongArraySize - 1] of TRGBRec;
  PPalRGBArray = ^TPalRGBArray;

  {Pixel format is used with bitcounts where channel sizes
   are not whole bytes (like 16bit or 27bit).
   or custom channel sizes are used.
   !!Supports channel sizes max 8bits wide.}
  TPixelFormat = record
    ABitCount, RBitCount, GBitCount, BBitCount: Byte; // bits per channel
    AInvBitCount, RInvBitCount, GInvBitCount, BInvBitCount: Byte; // 8 - BitCount
    ABitMask, RBitMask, GBitMask, BBitMask: ULong; // channel mask
    AShift, RShift, GShift, BShift: Byte; // channel shift in bits
  end;
  PPixelFormat = ^TPixelFormat;

const

  {Constants for some default colors. Uses colXXX prefix to avoid
   possible conflicts with Graphics.pas.}
  colNone: TARGB = $00000000;
  colBlack: TARGB = $FF000000;
  colWhite: TARGB = $FFFFFFFF;
  colRed: TARGB = $FFFF0000;
  colLime: TARGB = $FF00FF00;
  colBlue: TARGB = $FF0000FF;
  colMaroon: TARGB = $FF800000;
  colGreen: TARGB = $FF008000;
  colNavy: TARGB = $FF000080;
  colYellow: TARGB = $FFFFFF00;
  colOlive: TARGB = $FF808000;
  colFuchsia: TARGB = $FFFF00FF;
  colPurple: TARGB = $FF800080;
  colAqua: TARGB = $FF00FFFF;
  colTeal: TARGB = $FF008080;
  colSilver: TARGB = $FFC0C0C0;
  colGray: TARGB = $FF808080;
  colSkyBlue: TARGB = $FFA6CAF0;
  colMoneyGreen: TARGB = $FFC0DC0;
  colCream: TARGB = $FFFFFBF0;


{*************************************************************************}
{                     Color/Palette/other functions                       }
{*************************************************************************}


{Color functions parameters:
 ARGB : TARGB - This is color value in fixed ARGB format=>PixelFormat(8,8,8,8).
  One channel has one byte
 Color : ULong - This is color value can be in custom pixel format or palette
  index .

 A, R, G, B: Byte - each color channel intensity. Can be 0..255 (0.$FF)
}

{* ARGB Color manipulation *}

{Creates ARGB color value from color channel bytes.}
function ARGB(A, R, G, B: Byte): TARGB;
{Returns color channel bytes from ARGB value.}
procedure GetARGB(ARGB: TARGB; var A, R, G, B: Byte);
{Creates RGB color value from color channel bytes.}
function RGB(R, G, B: Byte): TRGB;
{Creates BGR color value from color channel bytes.}
function BGR(R, G, B: Byte): TBGR;
{Returns alpha channel of ARGB color.}
function GetA(ARGB: TARGB): byte;
{Returns red channel of ARGB color.}
function GetR(ARGB: TARGB): Byte;
{Returns green channel of ARGB color.}
function GetG(ARGB: TARGB): Byte;
{Returns blue channel of ARGB color.}
function GetB(ARGB: TARGB): Byte;

{Creates ARGB record from ARGB color.}
function ARGBRec(ARGB: TARGB): TARGBRec; overload;
{Creates ARGB record from A R G B bytes.}
function ARGBRec(A, R, G, B: Byte): TARGBRec; overload;
{Creates RGB record from ARGB color.}
function RGBRec(ARGB: TARGB): TRGBRec; overload;
{Creates RGB record from ARGB color.}
function RGBRec(R, G, B: Byte): TRGBRec; overload;
{Creates BGR record from ARGB color.}
function BGRRec(ARGB: TARGB): TBGRRec;

{Sets alpha channel of ARGB color.}
function SetA(ARGB: TARGB; A: Byte): TARGB;
{Sets red channel of ARGB color.}
function SetR(ARGB: TARGB; R: Byte): TARGB;
{Sets green channel of ARGB color.}
function SetG(ARGB: TARGB; G: Byte): TARGB;
{Sets blue channel of ARGB color.}
function SetB(ARGB: TARGB; B: Byte): TARGB;


type

  {Types used in color/string conversions.}
  TARGBStrFmt = (afHex, afInt, afBytes, afDescBytes, afFullDesc);

{Returns string representation of Color in hex format.}
function ARGBToStr(ARGB: TARGB): NString;
{Returns string representation of Color in selected format.}
function ARGBToStrEx(ARGB: TARGB; Fmt: TARGBStrFmt): NString;

{Linear interpolation of colors C1 and  C2. Factor must be 0..1.}
function ARGBLerp(C1, C2: TARGB; Factor: Float): TARGB;


{** Palette operations **}

{This function allocates and sets grey scale palette.
 Entries are numer of color entries (256 for 8bit).}
function GreyScalePal(Entries: Long; Alpha: Byte = $FF): PPalArray;
{Creates palette with given bitcount for each channel.}
function HalftonePal(Entries, R, G, B: Long; Alpha: Byte = $FF): PPalArray;

{This function converts RGB pal to ARGB pal (which allocates).}
function RGBPalToARGBPal(PalRGB: PPalRGBArray; Entries: Long = 256; Alpha:
  Byte = $FF): PPalArray; overload;
function RGBPalToARGBPal(const PalRGB: TPalRGB; Alpha: Byte = $FF): TPalette;
overload;

{This function converts ARGB pal to RGB pal.}
function ARGBPalToRGBPal(PalARGB: PPalArray; Entries: Long = 256):
  PPalRGBArray; overload;
function ARGBPalToRGBPal(const PalARGB: TPalette): TPalRGB; overload;

{This function swaps R and B vaules of palette.}
procedure PalSwapChannels(PalARGB: PPalArray; Entries: Long = 256);
{Returns color from Pal that is the closest color to Col}
function ClosestColorInPal(Pal: PPalArray; Entries: Long; ARGB: TARGB): Long;

{** Pixel format oprations **}

{Returns pixel format based on given channel sizes.}
function PixelFormat(ABitCount, RBitCount, GBitCount, BBitCount: Long):
  TPixelFormat;
{Returns pixel format based on given channel masks.}
function PixelFormatMask(ABitMask, RBitMask, GBitMask, BBitMask: Long):
  TPixelFormat;

{Returns Color from given bytes encoded by given PF pixel format.
 Note: returned color may not be in ARGB format. It depends on pixel format.}
function PFSetARGB(const PF: TPixelFormat; A, R, G, B: Byte): TARGB; overload;
{Returns A,R,G,B color bytes from Color in PF pixel format.}
procedure PFGetARGB(const PF: TPixelFormat; Color: ULong; var A, R, G, B: Byte);
overload;
{Returns Color from given ARGB color encoded by given PF pixel format.
 Note: returned color may not be in ARGB format. It depends on pixel format.}
function PFSetARGB(const PF: TPixelFormat; ARGB: TARGB): TARGB; overload;
{Returns ARGB color value from Color in PF pixel format.}
function PFGetARGB(const PF: TPixelFormat; Color: ULong): TARGB; overload;

{Converts Pixel format to string info}
function PFToStr(const PF: TPixelFormat): NString;


{*************************************************************************}
{                             IMPLEMEMTATION                              }
{*************************************************************************}


implementation


{*************************************************************************}
{                      Color/Palette/other functions                      }
{*************************************************************************}


function ARGB(A, R, G, B: Byte): TARGB;
begin
  Result := (A shl 24) or (R shl 16) or (G shl 8) or B;
end;

function RGB(R, G, B: Byte): TRGB;
begin
  Result := (R shl 16) or (G shl 8) or B;
end;

function BGR(R, G, B: Byte): TBGR;
begin
  Result := (B shl 16) or (G shl 8) or R;
end;

function GetA(ARGB: TARGB): Byte;
begin
  Result := ARGB shr 24;
end;

function GetR(ARGB: TARGB): Byte;
begin
  Result := ARGB shr 16;
end;

function GetG(ARGB: TARGB): Byte;
begin
  Result := ARGB shr 8;
end;

function GetB(ARGB: TARGB): Byte;
begin
  Result := ARGB;
end;

function ARGBRec(ARGB: TARGB): TARGBRec;
begin
  Result.Color := ARGB;
end;

function ARGBRec(A, R, G, B: Byte): TARGBRec;
begin
  Result.A := A;
  Result.R := R;
  Result.G := G;
  Result.B := B;
end;

function RGBRec(ARGB: TARGB): TRGBRec;
begin
  Result.B := ARGB;
  Result.G := ARGB shr 8;
  Result.R := ARGB shr 16;
end;

function RGBRec(R, G, B: Byte): TRGBRec;
begin
  Result.R := R;
  Result.G := G;
  Result.B := B;
end;

function BGRRec(ARGB: TARGB): TBGRRec;
begin
  Result.B := ARGB;
  Result.G := ARGB shr 8;
  Result.R := ARGB shr 16;
end;

procedure GetARGB(ARGB: TARGB; var A, R, G, B: Byte);
begin
  A := ARGB shr 24;
  R := ARGB shr 16;
  G := ARGB shr 8;
  B := ARGB;
end;

function SetA(ARGB: TARGB; A: Byte): TARGB;
begin
  Result := (A shl 24) or (ARGB and $00FFFFFF);
end;

function SetR(ARGB: TARGB; R: Byte): TARGB;
begin
  Result := (R shl 16) or (ARGB and $FF00FFFF);
end;

function SetG(ARGB: TARGB; G: Byte): TARGB;
begin
  Result := (G shl 8) or (ARGB and $FFFF00FF);
end;

function SetB(ARGB: TARGB; B: Byte): TARGB;
begin
  Result := B or (ARGB and $FFFFFF00);
end;

function ARGBToStr(ARGB: TARGB): NString;
begin
  Result := ARGBToStrEx(ARGB, afHex);
end;

function ARGBToStrEx(ARGB: TARGB; Fmt: TARGBStrFmt): NString;
begin
  Result := '';
  case Fmt of
    afHex: Result := Format('$%.8x', [ARGB]);
    afInt: Result := IntToStr(ARGB);
    afBytes: Result := SysUtils.Format('%d %d %d %d', [GetA(ARGB), GetR(ARGB),
        GetG(ARGB), GetB(ARGB)]);
    afDescBytes: Result := SysUtils.Format('A:%d R:%d G:%d B:%d', [GetA(ARGB),
        GetR(ARGB), GetG(ARGB), GetB(ARGB)]);
    afFullDesc: Result :=
      SysUtils.Format('Alpha: %d  Red: %d  Green: %d  Blue: %d', [GetA(ARGB),
        GetR(ARGB), GetG(ARGB), GetB(ARGB)]);
  end;
end;

function ARGBLerp(C1, C2: TARGB; Factor: Float): TARGB;
var
  A1, R1, G1, B1: Byte;
  A2, R2, G2, B2: Byte;
begin
  GetARGB(C1, A1, R1, G1, B1);
  GetARGB(C2, A2, R2, G2, B2);
  with PARGBREc(@Result)^ do
  begin
    A := Round(A1 + Factor * (A2 - A1));
    R := Round(R1 + Factor * (R2 - R1));
    G := Round(G1 + Factor * (G2 - G1));
    B := Round(B1 + Factor * (B2 - B1));
  end;
end;

function GreyScalePal(Entries: Long; Alpha: Byte): PPalArray;
var
  Pal: PPalArray;
  I: Long;
begin
  GetMem(Pal, Entries * SizeOf(TARGBRec));
  for I := 0 to Entries - 1 do
    with Pal[I] do
    begin
      A := Alpha;
      R := I;
      G := I;
      B := I;
    end;
  Result := Pal;
end;

function HalftonePal(Entries, R, G, B: Long; Alpha: Byte): PPalArray;
var
  Pal: PPalArray;
  I: Long;
begin
  GetMem(Pal, Entries * SizeOf(TARGBRec));
  for I := 0 to Entries - 1 do
  begin
    Pal[I].A := Alpha;
    Pal[I].R := ((I shr (G + B - 1)) and (1 shl R - 1)) * 255 div (1 shl R - 1);
    Pal[I].G := ((I shr (B - 1)) and (1 shl G - 1)) * 255 div (1 shl G - 1);
    Pal[I].B := ((I shr 0) and (1 shl B - 1)) * 255 div (1 shl B - 1);
  end;
  Result := Pal;
end;

function RGBPalToARGBPal(PalRGB: PPalRGBArray; Entries: Long; Alpha:
  Byte): PPalArray;
var
  I: Long;
begin
  Result := nil;
  if PalRGB = nil then Exit;
  GetMem(Result, Entries * SizeOf(TARGBRec));
  for I := 0 to Entries - 1 do
    with Result[I] do
    begin
      A := Alpha;
      R := PalRGB[I].R;
      G := PalRGB[I].G;
      B := PalRGB[I].B;
    end;
end;

function ARGBPalToRGBPal(PalARGB: PPalArray; Entries: Long): PPalRGBArray;
var
  I: Long;
begin
  Result := nil;
  if PalARGB = nil then Exit;
  GetMem(Result, Entries * SizeOf(TRGBRec));
  for I := 0 to Entries - 1 do
    with Result[I] do
    begin
      R := PalARGB[I].R;
      G := PalARGB[I].G;
      B := PalARGB[I].B;
    end;
end;

function RGBPalToARGBPal(const PalRGB: TPalRGB; Alpha: Byte): TPalette;
var
  I: Long;
begin
  for I := Low(TPalette) to High(TPalette) do
    with Result[I] do
    begin
      A := Alpha;
      R := PalRGB[I].R;
      G := PalRGB[I].G;
      B := PalRGB[I].B;
    end;
end;

function ARGBPalToRGBPal(const PalARGB: TPalette): TPalRGB;
var
  I: Long;
begin
  for I := Low(TPalRGB) to High(TPalRGB) do
    with Result[I] do
    begin
      R := PalARGB[I].R;
      G := PalARGB[I].G;
      B := PalARGB[I].B;
    end;
end;

procedure PalSwapChannels(PalARGB: PPalArray; Entries: Long);
var
  I: Long;
  T: Byte;
begin
  if PalARGB = nil then Exit;
  for I := 0 to Entries - 1 do
  begin
    T := PalARGB[I].R;
    PalARGB[I].R := PalARGB[I].B;
    PalARGB[I].B := T;
  end;
end;

function ClosestColorInPal(Pal: PPalArray; Entries: Long; ARGB: TARGB): Long;
var
  N: Long;
  PC: PARGBRec;
  C: TARGBRec;
  I, X, D: Long;
begin
  Result := 0;
  if Pal = nil then Exit;
  C := TARGBRec(ARGB);
  X := 765;
  N := 0;
  PC := PARGBRec(Pal);
  for I := 0 to Entries - 1 do
  begin
    if PC.A > C.A then
      D := PC.A - C.A
    else
      D := C.A - PC.A;
    if PC.R > C.R then
      Inc(D, PC.R - C.R)
    else
      Inc(D, C.R - PC.R);
    if PC.G > C.G then
      Inc(D, PC.G - C.G)
    else
      Inc(D, C.G - PC.G);
    if PC.B > C.B then
      Inc(D, PC.B - C.B)
    else
      Inc(D, C.B - PC.B);
    if D < X then
    begin
      X := D;
      N := I;
    end;
    Inc(PC);
  end;
  Result := N;
end;

function PixelFormat(ABitCount, RBitCount, GBitCount, BBitCount: Long):
  TPixelFormat;
begin
  Result.ABitMask := ((1 shl ABitCount) - 1) shl (RBitCount + GBitCount +
    BBitCount);
  Result.RBitMask := ((1 shl RBitCount) - 1) shl (GBitCount + BBitCount);
  Result.GBitMask := ((1 shl GBitCount) - 1) shl (BBitCount);
  Result.BBitMask := (1 shl BBitCount) - 1;
  Result.ABitCount := ABitCount;
  Result.RBitCount := RBitCount;
  Result.GBitCount := GBitCount;
  Result.BBitCount := BBitCount;
  Result.AInvBitCount := 8 - ABitCount;
  Result.RInvBitCount := 8 - RBitCount;
  Result.GInvBitCount := 8 - GBitCount;
  Result.BInvBitCount := 8 - BBitCount;
  Result.AShift := RBitCount + GBitCount + BBitCount;
  Result.RShift := GBitCount + BBitCount;
  Result.GShift := BBitCount;
  Result.BShift := 0;
end;

function PixelFormatMask(ABitMask, RBitMask, GBitMask, BBitMask: Long):
  TPixelFormat;

  function GetBitCount(B: Long): Long;
  var
    I: Long;
  begin
    I := 0;
    while (I < 31) and (((1 shl I) and B) = 0) do
      Inc(I);
    Result := 0;
    while ((1 shl I) and B) <> 0 do
    begin
      Inc(I);
      Inc(Result);
    end;
  end;

begin
  Result := PixelFormat(GetBitCount(ABitMask), GetBitCount(RBitMask),
    GetBitCount(GBitMask), GetBitCount(BBitMask));
end;

function PFSetARGB(const PF: TPixelFormat; A, R, G, B: Byte): TARGB;
begin
  with PF do
    Result :=
      (A shl ABitCount shr 8 shl AShift) or
      (R shl RBitCount shr 8 shl RShift) or
      (G shl GBitCount shr 8 shl GShift) or
      (B shl BBitCount shr 8 shl BShift);
end;

procedure PFGetARGB(const PF: TPixelFormat; Color: ULong; var A, R, G, B: Byte);
begin
  with PF do
  begin
    A := Color and ABitMask shr AShift shl AInvBitCount;
    R := Color and RBitMask shr RShift shl RInvBitCount;
    G := Color and GBitMask shr GShift shl GInvBitCount;
    B := Color and BBitMask shl BShift shl BInvBitCount;
  end;
end;

function PFSetARGB(const PF: TPixelFormat; ARGB: TARGB): TARGB;
begin
  with PF do
    Result :=
      (Byte(ARGB shr 24) shl ABitCount shr 8 shl AShift) or
      (Byte(ARGB shr 16) shl RBitCount shr 8 shl RShift) or
      (Byte(ARGB shr 8) shl GBitCount shr 8 shl GShift) or
      (Byte(ARGB) shl BBitCount shr 8 shl BShift);
end;

function PFGetARGB(const PF: TPixelFormat; Color: ULong): TARGB;
begin
  with PF do
    Result :=
      (Color and ABitMask shr AShift shl AInvBitCount shl 24) or
      (Color and RBitMask shr RShift shl RInvBitCount shl 16) or
      (Color and GBitMask shr GShift shl GInvBitCount shl 8) or
      (Color and BBitMask shl BShift shl BInvBitCount);
end;

function PFToStr(const PF: TPixelFormat): NString;
begin
  Result := Format('AlphaBits: %d, RedBits: %d, GreenBits: %d, BlueBits: %d',
    [PF.ABitCount, PF.RBitCount, PF.GBitCount, PF.BBitCount]);
end;

end.

