{*************************************************************************}
{*   ColorQuantization                                                   *}
{*   Copyright (C) 2003 Marek Mauder, Vampyre SoftWorks                  *}
{*   Engine\Graphic\ImageDataFormats\ColorQuantization.pas               *}
{*************************************************************************}
{                                                                         }
{ This program is free software; you can redistribute it and/or modify it }
{ under the terms of the GNU General Public License as published by the   }
{ Free Software Foundation; either version 2 of the License, or (at your  }
{ option) any later version. This program is distributed in the hope that }
{ it will be useful, but WITHOUT ANY WARRANTY; without even the implied   }
{ warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See   }
{ the GNU General Public License for more details.                        }
{                                                                         }
{ You should have received a copy of the GNU General Public License       }
{ along with this program; if not, write to the Free Software Foundation, }
{ Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.         }
{                                                                         }
{*************************************************************************}
{  Created On: 8th November 2003  (08.11.2003)                            }
{*************************************************************************}

{This unit contains class that converts true-color images to paletted
 ones with optimized palette (uses octree color quantization).}

unit ColorQuantization;

{$I Compiler.inc}

interface

uses
  Kernel, Objects, Platforms, Graphic, Images;

type

  TNOctreeNode = class;

  {Node's children.}
  TReducibleNodes = array[0..7] of TNOctreeNode;

  {Node for octree.}
  TNOctreeNode = class(TNBaseObject)
  protected
    IsLeaf: Boolean;
    PixelCount: Long;
    RedSum: Long;
    GreenSum: Long;
    BlueSum: Long;
    Level: Long;
    ColorIndex: Long;
    Next: TNOctreeNode;
    Child: TReducibleNodes;
  public
    constructor Create(ALevel, AColorBits: Long; var LeafCount: Long;
      var ReducibleNodes: TReducibleNodes); reintroduce;
    destructor Destroy; override;
  end;

  {Class which does color quantization using octree. Works only
   for 24/32bit Images.}
  TNColorQuantizer = class(TNBaseObject)
  protected
    FTree: TNOctreeNode;
    FLeafCount: Long;
    FReducibleNodes: TReducibleNodes;
    FMaxColors: Long;
    FColorBits: Long;
    procedure AddColor(var Node: TNOctreeNode; R, G, B: Byte;
      ColorBits, Level: Long; var LeafCount: Long;
      var ReducibleNodes: TReducibleNodes);
    procedure DeleteTree(var Node: TNOctreeNode);
    procedure GetPaletteColors(Node: TNOctreeNode; var Pal: PPalArray;
      var Index: Long);
    procedure ReduceTree(ColorBits: Long; var LeafCount: Long;
      var ReducibleNodes: TReducibleNodes);
    function FindInNode(Node: TNOctreeNode; R, G, B: Byte): Long;
  public
    {Creates new quantizer. MaxColors is max number of colors in output image,
     ColorBits controls creation of octree (use max 7), it has nothing to do
     with BitCounts or so.}
    constructor Create(MaxColors, ColorBits: Long); reintroduce;
    destructor Destroy; override;
    {Returns palette created from octree.
     Pal must be allocated GetMem(Pal,MaxColors * SizeOf(TARGBRec))}
    procedure GetPalette(var Pal: PPalArray);
    {This processes single image and updates octree.
     You can process several images by this function if you want single
     optimized palette for multiple images. It must be 24 RGB bit image
     or 32bit image (but alpha is ignored). Returns True if successful.}
    function UpdateOctree(Width, Height, BitCount: Long; Bits: Pointer):
      Boolean;
    {Returns closest color (index to palette returned by GetPalette) in octree.}
    function Find(Color: TARGB): Long; overload;
    function Find(R, G, B: Byte): Long; overload;
    {Converts input image (24/32bit) to 8bit using actual octree
     (so UpdateOctree had to be called at least once). OutBits are supposed
     to be nil when calling.}
    function ConvertImage(Width, Height, BitCount: Long; Bits: Pointer;
      var OutBits: Pointer): Boolean;
    {Number of leaves in octree.}
    property ColorCount: Long read FLeafCount;
  end;


{*************************************************************************}
{                             IMPLEMEMTATION                              }
{*************************************************************************}


implementation

const
  Mask: array[0..7] of Byte = ($80, $40, $20, $10, $08, $04, $02, $01);


{*************************************************************************}
{                           TNOctreeNode methods                          }
{*************************************************************************}


constructor TNOctreeNode.Create(ALevel, AColorBits: Long; var LeafCount: Long;
  var ReducibleNodes: TReducibleNodes);
var
  I: Long;
begin
  PixelCount := 0;
  RedSum := 0;
  GreenSum := 0;
  BlueSum := 0;
  Level := ALevel;
  for I := Low(Child) to High(Child) do
    Child[I] := nil;
  IsLeaf := (ALevel = AColorBits);
  if IsLeaf then
  begin
    Next := nil;
    Inc(LeafCount);
  end
  else
  begin
    Next := ReducibleNodes[Level];
    ReducibleNodes[Level] := Self;
  end
end;

destructor TNOctreeNode.Destroy;
var
  I: Long;
begin
  for I := Low(Child) to High(Child) do
    Child[I].Free
end;


{*************************************************************************}
{                           TNColorQuantizer methods                          }
{*************************************************************************}


constructor TNColorQuantizer.Create(MaxColors, ColorBits: Long);
var
  I: Long;
begin
  FTree := nil;
  FLeafCount := 0;
  // Initialize all nodes even though only ColorBits+1 of them are needed
  for I := Low(FReducibleNodes) to High(FReducibleNodes) do
    FReducibleNodes[I] := nil;
  FMaxColors := MaxColors;
  FColorBits := ColorBits
end;

destructor TNColorQuantizer.Destroy;
begin
  if FTree <> nil then
    DeleteTree(FTree)
end;

procedure TNColorQuantizer.AddColor(var Node: TNOctreeNode; R, G, B: Byte;
  ColorBits, Level: Long; var LeafCount: Long;
  var ReducibleNodes: TReducibleNodes);
var
  Index: Long;
  Shift: Long;
begin
  // If the node doesn't exist, create it.
  if Node = nil then
    Node := TNOctreeNode.Create(Level, ColorBits, LeafCount, ReducibleNodes);

  if Node.IsLeaf then
  begin
    Inc(Node.PixelCount);
    Inc(Node.RedSum, R);
    Inc(Node.GreenSum, G);
    Inc(Node.BlueSum, B)
  end
  else
  begin
    // Recurse a level deeper if the node is not a leaf.
    Shift := 7 - Level;
    Index := (((R and Mask[Level]) shr Shift) shl 2) or
      (((G and Mask[Level]) shr Shift) shl 1) or
      ((B and Mask[Level]) shr Shift);
    AddColor(Node.Child[Index], R, G, B, ColorBits, Level + 1,
      LeafCount, ReducibleNodes)
  end;
end;

procedure TNColorQuantizer.DeleteTree(var Node: TNOctreeNode);
var
  I: Long;
begin
  for I := Low(TReducibleNodes) to High(TReducibleNodes) do
  begin
    if Node.Child[I] <> nil then DeleteTree(Node.Child[I]);
  end;
  Node.Free;
  Node := nil;
end;

procedure TNColorQuantizer.GetPaletteColors(Node: TNOctreeNode;
  var Pal: PPalArray; var Index: Long);
var
  I: Long;
begin
  if Node.IsLeaf then
  begin
    with Pal[Index] do
    begin
      try
        R := Byte(Node.RedSum div Node.PixelCount);
        G := Byte(Node.GreenSum div Node.PixelCount);
        B := Byte(Node.BlueSum div Node.PixelCount)
      except
        R := 0;
        G := 0;
        B := 0
      end;
      A := $FF;
    end;
    Inc(Index);
  end
  else
  begin
    for I := Low(Node.Child) to High(Node.Child) do
    begin
      if Node.Child[I] <> nil then
        GetPaletteColors(Node.Child[I], Pal, Index)
    end
  end
end;

procedure TNColorQuantizer.ReduceTree(ColorBits: Long; var LeafCount: Long;
  var ReducibleNodes: TReducibleNodes);
var
  BlueSum: Long;
  Children: Long;
  GreenSum: Long;
  I: Long;
  Node: TNOctreeNode;
  RedSum: Long;
begin
  // Find the deepest level containing at least one reducible node
  I := Colorbits - 1;
  while (I > 0) and (ReducibleNodes[I] = nil) do
    Dec(I);

  // Reduce the node most recently added to the list at level i.
  Node := ReducibleNodes[I];
  ReducibleNodes[I] := Node.Next;
  RedSum := 0;
  GreenSum := 0;
  BlueSum := 0;
  Children := 0;
  for I := Low(ReducibleNodes) to High(ReducibleNodes) do
  begin
    if Node.Child[I] <> nil then
    begin
      Inc(RedSum, Node.Child[I].RedSum);
      Inc(GreenSum, Node.Child[I].GreenSum);
      Inc(BlueSum, Node.Child[I].BlueSum);
      Inc(Node.PixelCount, Node.Child[I].PixelCount);
      Node.Child[I].Free;
      Node.Child[I] := nil;
      Inc(Children)
    end
  end;
  Node.IsLeaf := True;
  Node.RedSum := RedSum;
  Node.GreenSum := GreenSum;
  Node.BlueSum := BlueSum;
  Dec(LeafCount, Children - 1)
end;

function TNColorQuantizer.FindInNode(Node: TNOctreeNode; R, G, B: Byte): Long;
var
  Shift: Long;
  Index: Long;
begin
  if Node.IsLeaf then
    Result := Node.ColorIndex
  else
  begin
    Shift := 7 - Node.Level;
    Index := (((R and Mask[Node.Level]) shr Shift) shl 2) or
      (((G and Mask[Node.Level]) shr Shift) shl 1) or
      ((B and Mask[Node.Level]) shr Shift);
    Result := FindInNode(Node.Child[Index], R, G, B);
  end;
end;

procedure TNColorQuantizer.GetPalette(var Pal: PPalArray);
var
  Index: Long;
begin
  Index := 0;
  GetPaletteColors(FTree, Pal, Index)
end;

function TNColorQuantizer.UpdateOctree(Width, Height, BitCount: Long;
  Bits: Pointer): Boolean;
var
  I, J: Long;
  ScanLine: PRGBRecArray;
  ScanLineA: PARGBRecArray;

  procedure SetIndices(Node: TNOctreeNode; var Index: Long);
  var
    I: Long;
  begin
    if Node.IsLeaf then
    begin
      Node.ColorIndex := Index;
      Inc(Index);
    end
    else
      for I := Low(Node.Child) to High(Node.Child) do
        if Node.Child[I] <> nil then
          SetIndices(Node.Child[I], Index)
  end;

begin
  Result := False;
  if Bits <> nil then
  try
    case BitCount of
      24:
        begin
          Scanline := PRGBRecArray(Bits);
          for J := 0 to Height - 1 do
          begin
            for I := 0 to Width - 1 do
            begin
              with Scanline[I] do
                AddColor(FTree, R, G, B,
                  FColorBits, 0, FLeafCount, FReducibleNodes);
              while FLeafCount > FMaxColors do
                ReduceTree(FColorBits, FLeafCount, FReducibleNodes)
            end;
            ScanLine := PRGBRecArray(Long(Scanline) + Width * SizeOf(TRGBRec));
          end;
        end;
      32:
        begin
          ScanlineA := PARGBRecArray(Bits);
          for J := 0 to Height - 1 do
          begin
            for I := 0 to Width - 1 do
            begin
              with ScanlineA[I] do
                AddColor(FTree, R, G, B,
                  FColorBits, 0, FLeafCount, FReducibleNodes);
              while FLeafCount > FMaxColors do
                ReduceTree(FColorBits, FLeafCount, FReducibleNodes)
            end;
            ScanLineA := PARGBRecArray(Long(ScanlineA) + Width *
              SizeOf(TARGBRec));
          end;
        end;
    end;
    I := 0;
    SetIndices(FTree, I);
    Result := True;
  except
    Result := False;
  end;
end;

function TNColorQuantizer.Find(Color: TARGB): Long;
var
  Shift: Long;
  Index: Long;
begin
  if FTree.IsLeaf then
    Result := FTree.ColorIndex
  else
  begin
    Shift := 7 - FTree.Level;
    Index := (((TARGBRec(Color).R and Mask[FTree.Level]) shr Shift) shl 2) or
      (((TARGBRec(Color).G and Mask[FTree.Level]) shr Shift) shl 1) or
      ((TARGBRec(Color).B and Mask[FTree.Level]) shr Shift);
    Result := FindInNode(FTree.Child[Index], TARGBRec(Color).R,
      TARGBRec(Color).G, TARGBRec(Color).B);
  end;
end;

function TNColorQuantizer.Find(R, G, B: Byte): Long;
var
  Shift: Long;
  Index: Long;
begin
  if FTree.IsLeaf then
  begin
    Result := FTree.ColorIndex;
  end
  else
  begin
    Shift := 7 - FTree.Level;
    Index := (((R and Mask[FTree.Level]) shr Shift) shl 2) or
      (((G and Mask[FTree.Level]) shr Shift) shl 1) or
      ((B and Mask[FTree.Level]) shr Shift);
    Result := FindInNode(FTree.Child[Index], R, G, B);
  end;
end;

function TNColorQuantizer.ConvertImage(Width, Height, BitCount: Long;
  Bits: Pointer; var OutBits: Pointer): Boolean;
var
  X, Y: Long;
  B: PByte;
  C: PRGBRec;
  CA: PARGBRec;
begin
  Result := False;
  if Bits <> nil then
  try
    GetMem(OutBits, Width * Height * SizeOf(Byte));
    B := OutBits;
    C := Bits;
    CA := Bits;
    for X := 0 to Width - 1 do
      for Y := 0 to Height - 1 do
      begin
        case BitCount of
          24:
            begin
              B^ := Find(C.R, C.G, C.B);
              Inc(C);
            end;
          32:
            begin
              B^ := Find(CA.R, CA.G, CA.B);
              Inc(CA);
            end;
        end;
        Inc(B);
      end;
    Result := True;
  except
    Result := False;
  end;
end;

end.

