{*************************************************************************}
{*   BmpFormat                                                           *}
{*   Copyright (C) 2003 Marek Mauder, Vampyre SoftWorks                  *}
{*   Graphic\ImageFileFormats\Bmp\BmpFormat.pas                          *}
{*************************************************************************}
{                                                                         }
{ This program is free software; you can redistribute it and/or modify it }
{ under the terms of the GNU General Public License as published by the   }
{ Free Software Foundation; either version 2 of the License, or (at your  }
{ option) any later version. This program is distributed in the hope that }
{ it will be useful, but WITHOUT ANY WARRANTY; without even the implied   }
{ warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See   }
{ the GNU General Public License for more details.                        }
{                                                                         }
{ You should have received a copy of the GNU General Public License       }
{ along with this program; if not, write to the Free Software Foundation, }
{ Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.         }
{                                                                         }
{*************************************************************************}
{  Created On: 12st September 2003  (12.09.2003)                          }
{*************************************************************************}

{This unit contains image file format loader/saver for Bitmap files.}

unit BmpFormat;

{$I Compiler.inc}

interface

uses
  SysUtils, Kernel, Objects, Platforms, Graphic, Images;

type

  {Image encoder for loading/saving bitmaps in BMP/DIB format.
  Supports: 1,4,8,15,16,24,32 bit bitmaps loading/saving (OS2 format loading)
  Compression: RLE4, RLE8 loading/saving.}
  TNBMPFormat = class(TNImageFileFormat)
  protected
    procedure PrepareImage(const Image: TImageData; var Clone: TImageData);
      override;
    procedure DoLoadImage(var Image: TImageData; Stream: TNStream); override;
    procedure DoSaveImage(var Image: TImageData; Stream: TNStream;
      SaveCompressed: Boolean); override;
  public
    constructor Create; override;
    function TestFormat(Stream: TNStream): Boolean; override;
  end;

const

  sBMPExtensions = 'bmp,dib';
  sBMPFormatName = 'Windows/OS2 Bitmap';

var

  BMPInstance: TNBMPFormat = nil;


{*************************************************************************}
{                             IMPLEMEMTATION                              }
{*************************************************************************}


implementation

const

  {BMP file identifier}
  BMPMagic: array[0..1] of Char = 'BM';

  { constants for the TBitmapInfoHeader.Compression field }
  BI_RGB = 0;
  BI_RLE8 = 1;
  BI_RLE4 = 2;
  BI_BITFIELDS = 3;

type

  {File Header for Windows/OS2 bitmap file.}
  TBitMapFileHeader = packed record
    ID: Word; // Is always 19778 : 'BM'
    Size: ULong; // Filesize
    Reserved1: Word;
    Reserved2: Word;
    Offset: ULong; // Offset of image data
  end;

  {Info Header for Windows bitmap file.}
  TBitMapInfoHeader = packed record
    Size: ULong;
    Width: Long;
    Height: Long;
    Planes: Word;
    BitCount: Word;
    Compression: ULong;
    SizeImage: ULong;
    XPelsPerMeter: Long;
    YPelsPerMeter: Long;
    ClrUsed: ULong;
    ClrImportant: ULong;
  end;

  {Info Header for OS2 bitmaps}
  TBitmapCoreHeader = packed record
    Size: ULong;
    Width: Word;
    Height: Word;
    Planes: Word;
    BitCount: Word;
  end;

  {Used with BitmapInfo.Compression = BI_BITFIELDS}
  TLocalPixelFormat = packed record
    RBitMask, GBitMask, BBitMask: ULong;
  end;

const

  MinSize = SizeOf(TBitmapFileHeader) + SizeOf(TBitmapCoreHeader);


procedure Convert1To8(var Level: TImageLevel);
const
  Mask1: array[0..7] of Byte = ($80, $40, $20, $10, $08, $04, $02, $01);
  Shift1: array[0..7] of Byte = (7, 6, 5, 4, 3, 2, 1, 0);
var
  Data: Pointer;
  X, Y: Long;
  WB: Long;
begin
  with Level do
  begin
    GetMem(Data, Width * Height);
    WB := ((Width + 31) shr 5) * 4;
    try
      for X := 0 to Width - 1 do
        for Y := 0 to Height - 1 do
          PByte(Long(Data) + Y * Width + X)^ :=
            (PByteArray(Integer(Bits) + Y * WB)[X shr 3] and Mask1[X and 7])
            shr Shift1[X and 7];
    finally
      FreeMemNil(Bits);
      Bits := Data;
      Size := Width * Height;
    end;
  end;
end;

procedure Convert4To8(var Level: TImageLevel);
const
  Mask4: array[0..1] of Byte = ($F0, $0F);
  Shift4: array[0..1] of Byte = (4, 0);
var
  Data: Pointer;
  X, Y: Long;
  WB: Long;
begin
  with Level do
  begin
    GetMem(Data, Width * Height);
    WB := (((Width * 4) + 31) shr 5) * 4;
    try
      for X := 0 to Width - 1 do
        for Y := 0 to Height - 1 do
          PByte(Long(Data) + Y * Width + X)^ :=
            (PByteArray(Integer(Bits) + Y * WB)[X shr 1] and Mask4[X and 1])
            shr Shift4[X and 1];
    finally
      FreeMemNil(Bits);
      Bits := Data;
      Size := Width * Height;
    end;
  end;
end;

procedure AddPadBytes(const Level: TImageLevel; BitCount: Long; var Data:
  Pointer; var Size:Long);
var
  I,W, WB: Long;
begin
  WB := (((Level.Width * BitCount) + 31) shr 5) * 4;
  W := Level.Width * BitCount shr 3;
  Size := Level.Height * WB;
  GetMem(Data, Size);
  for I := 0 to Level.Height - 1 do
    Move(Pointer(Long(Level.Bits) + I * W)^, Pointer(Long(Data) + I * WB)^, W);
end;

procedure RemovePadBytes(var Level: TImageLevel; BitCount: Long);
var
  I: Long;
  Data: Pointer;
  W, WB: Long;
begin
  with Level do
  begin
    W := Width * (BitCount shr 3);
    Size := Height * W;
    WB := (((Width * BitCount) + 31) shr 5) * 4;
    GetMem(Data, Size);
    try
      for I := 0 to Height - 1 do
        Move(Pointer(Long(Bits) + I * WB)^, Pointer(Long(Data) + I * W)^, W);
    finally
      FreeMemNil(Bits);
      Bits := Data;
    end;
  end;
end;


{*************************************************************************}
{                        TNBMPFormat methods                              }
{*************************************************************************}


constructor TNBMPFormat.Create;
begin
  inherited Create;
  FExtensions := sBMPExtensions;
  FFormatName := sBMPFormatName;
  FCanLoad := True;
  FCanSave := True;
end;

procedure TNBMPFormat.PrepareImage(const Image: TImageData;
  var Clone: TImageData);
begin
  inherited PrepareImage(Image, Clone);
  with Clone.DataFormat^ do
  begin
    // BMP stores images as flipped with dword aligned scanlines
    // 16bit images can be saved -> it supports custom pixel formats
    // in all 8bit images must be palette present
    // A4R4G4B4 and A8R8G8B8 images can be properly saved
    // but many image viewers/editors do not support them
    // (A8R8G8B8 are treat like R8G8B8 and A4R4G4B4 are mostly
    // displayed with wrong pixel format.
    if OneChannelFormat then
      ConvertFunc(Clone, FindImageDataFormat(idfR8G8B8))
    else
      if FloatFormat then
        ConvertFunc(Clone, FindImageDataFormat(idfA8R8G8B8))
      else
        if Indexed and (BitCount > 8) then
          ConvertFunc(Clone, FindImageDataFormat(idfIndex8))
        else
          if not FixedPixelFormat then
          begin
            if HasAlpha then
              ConvertFunc(Clone, FindImageDataFormat(idfA4R4G4B4))
            else
              ConvertFunc(Clone, FindImageDataFormat(idfR5G5B5))
          end
          else
            if (OneChannelSize > 8) then
            begin
              if HasAlpha then
                ConvertFunc(Clone, FindImageDataFormat(idfA8R8G8B8))
              else
                ConvertFunc(Clone, FindImageDataFormat(idfR8G8B8))
            end;
    // final check
    if (Clone.DataFormat <> FindImageDataFormat(idfA8R8G8B8)) and
      (Clone.DataFormat <> FindImageDataFormat(idfR8G8B8)) and
      (Clone.DataFormat <> FindImageDataFormat(idfR5G5B5)) and
      (Clone.DataFormat <> FindImageDataFormat(idfA4R4G4B4)) and
      (Clone.DataFormat <> FindImageDataFormat(idfIndex8)) then
      ConvertFunc(Clone, FindImageDataFormat(idfR8G8B8));
    FlipYImage(Clone);
  end;
end;

procedure TNBMPFormat.DoLoadImage(var Image: TImageData; Stream: TNStream);
var
  BF: TBitmapFileHeader;
  BI: TBitmapInfoHeader;
  BC: TBitmapCoreHeader;
  IsOS2: boolean;
  StartPos: Int64;
  LocalPF: TLocalPixelFormat;
  PalRGB: PPalRGBArray;
  I: Long;
  FPalSize, FBitCount, WidthBytes: Long;

  procedure LoadRGB;
  var
    I: Long;
  begin
    with Image.Levels[0] do
    begin
      GetMem(Bits, Size);
      // if Height is <0 then image data are stored non-flipped (like in OGL+D3D)
      // but default in windows is flipped so if Height is positive we must flip it
      if BI.Height < 0 then
        Stream.Read(Bits^, Size)
      else
        for I := Height - 1 downto 0 do
          Stream.Read(Pointer(Long(Bits) + I * WidthBytes)^,
            WidthBytes);
    end;
  end;

  procedure LoadRLE4;
  var
    B1, B2, C: Byte;
    Data, Dest, Src, P: PByte;
    X, Y, I: Long;
    BW: Long;
  begin
    with Image.Levels[0] do
    begin
      BW := (((Width * 4) + 31) shr 5) * 4;
      Size := Height * BW;
      GetMem(Data, BI.SizeImage);
      GetMem(Bits, Size);
      try
        Stream.Read(Data^, BI.SizeImage);
        X := 0;
        Y := 0;
        Src := Data;
        while True do
        begin
          B1 := Src^;
          Inc(Src);
          B2 := Src^;
          Inc(Src);
          if B1 = 0 then
          begin
            case B2 of
              0:
                begin //  End of line
                  X := 0;
                  Inc(Y);
                end;
              1: Break; //  End of bitmap
              2:
                begin //  Difference of coordinates
                  Inc(X, B1);
                  Inc(Y, B2);
                  Inc(Src, 2);
                end;
              else
                Dest := Pointer(Long(Bits) + Y * BW);
                C := 0;
                for i := 0 to B2 - 1 do
                begin
                  if i and 1 = 0 then
                  begin
                    C := Src^;
                    Inc(Src);
                  end
                  else
                  begin
                    C := C shl 4;
                  end;
                  P := Pointer(Long(Dest) + X shr 1);
                  if X and 1 = 0 then
                    P^ := (P^ and $0F) or (C and $F0)
                  else
                    P^ := (P^ and $F0) or ((C and $F0) shr 4);
                  Inc(X);
                end;
            end;
          end
          else
          begin
            Dest := Pointer(Long(Bits) + Y * BW);
            for I := 0 to B1 - 1 do
            begin
              P := Pointer(Long(Dest) + X shr 1);
              if X and 1 = 0 then
                P^ := (P^ and $0F) or (B2 and $F0)
              else
                P^ := (P^ and $F0) or ((B2 and $F0) shr 4);
              Inc(X);
              // Swap nibble
              B2 := (B2 shr 4) or (B2 shl 4);
            end;
          end;
          //  Word arrangement
          Inc(Src, Long(Src) and 1);
        end;
      finally
        FreeMemNil(Data);
      end;
    end;
  end;

  procedure LoadRLE8;
  var
    B1, B2: Byte;
    Data, Dest, Src: PByte;
    X, Y, BW: Long;
  begin
    with Image.Levels[0] do
    begin
      BW := (((Width * 8) + 31) shr 5) * 4;
      Size := Height * BW;
      GetMem(Data, BI.SizeImage);
      GetMem(Bits, Size);
      try
        Stream.Read(Data^, BI.SizeImage);
        Dest := Bits;
        Src := Data;
        X := 0;
        Y := 0;
        while True do
        begin
          B1 := Src^;
          Inc(Src);
          B2 := Src^;
          Inc(Src);
          if B1 = 0 then
          begin
            case B2 of
              0:
                begin //  End of line
                  X := 0;
                  Inc(Y);
                  Dest := Pointer(Long(Bits) + Y * BW + X);
                end;
              1: Break; //  End of bitmap
              2:
                begin //  Difference of coordinates
                  Inc(X, B1);
                  Inc(Y, B2);
                  Inc(Src, 2);
                  Dest := Pointer(Long(Bits) + Y * BW + X);
                end;
              else
                Move(Src^, Dest^, B2);
                Inc(Dest, B2);
                Inc(Src, B2);
            end;
          end
          else
          begin
            FillChar(Dest^, B1, B2);
            Inc(Dest, B1);
          end;
          //  Word arrangement
          Inc(Src, Longint(Src) and 1);
        end;
      finally
        FreeMemNil(Data);
      end;
    end;
  end;

begin
  StartPos := Stream.Position;
  AddImageLevel(Image);
  with Stream, Image, Image.Levels[0] do
  begin
    Read(BF, SizeOf(BF));
    Read(BI.Size, SizeOf(BI.Size));
    IsOS2 := BI.Size = SizeOf(TBitmapCoreHeader);

    // Bitmap Info reading
    if IsOS2 then
    begin
      // OS/2 type bitmap
      // Reads info header without 4 already read bytes
      Read(Pointer(Long(@BC) + 4)^, SizeOf(TBitmapCoreHeader) - 4);
      with BI do
      begin
        ClrUsed := 0;
        Compression := BI_RGB;
        BitCount := BC.BitCount;
        Height := BC.Height;
        Width := BC.Width;
      end;
    end
    else
      //  Windows type bitmap
      Read(Pointer(Long(@BI) + 4)^, SizeOf(TBitmapInfoHeader) - 4);

    Image.Levels[0].Width := BI.Width;
    Image.Levels[0].Height := Abs(BI.Height);
    FBitCount := BI.BitCount;
    Image.Levels[0].Size := BI.SizeImage;
    WidthBytes := (((Image.Levels[0].Width * FBitCount) + 31) shr 5) * 4;
    if Image.Levels[0].Size = 0 then
      Image.Levels[0].Size := Image.Levels[0].Height * WidthBytes;

    //  Bit mask reading.
    if BI.Compression = BI_BITFIELDS then
      Read(LocalPF, SizeOf(LocalPF));

    case FBitCount of
      1, 4, 8: Image.DataFormat := FindImageDataFormat(idfIndex8);
      16:
        if LocalPF.RBitMask = 3480 then
          Image.DataFormat := FindImageDataFormat(idfA4R4G4B4)
        else
          Image.DataFormat := FindImageDataFormat(idfR5G5B5);
      24: Image.DataFormat := FindImageDataFormat(idfR8G8B8);
      32: Image.DataFormat := FindImageDataFormat(idfA8R8G8B8);
    end;

    //  Palette settings and reading
    if FBitCount <= 8 then
      if IsOS2 then
      begin
        //  OS/2 type
        FPalSize := 0;
        case FBitCount of
          1: FPalSize := 2;
          4: FPalSize := 16;
          8: FPalSize := 256;
        end;
        if FBitCount = 1 then Inc(FPalSize);
        GetMem(Palette, FPalSize * SizeOf(TARGBRec));
        GetMem(PalRGB, FPalSize * SizeOf(TRGBRec));
        Read(PalRGB^, FPalSize * SizeOf(TRGBRec));
        for I := 0 to FPalSize - 1 do
          with PalRGB[I] do
            Palette[I] := ARGBRec($FF, R, G, B);
        FreeMemNil(PalRGB);
      end
      else
      begin
        //  Windows type
        FPalSize := BI.ClrUsed;
        GetMem(Palette, FPalSize * SizeOf(TARGBRec));
        Read(Palette^, FPalSize * SizeOf(TARGBRec));
        for I := 0 to FPalSize - 1 do
          Palette[I].A := $FF;
      end;

    case BI.Compression of
      BI_RGB: LoadRGB;
      BI_RLE4: LoadRLE4;
      BI_RLE8: LoadRLE8;
      BI_BITFIELDS: LoadRGB;
    end;

    // 1 and 4 bpp images are supported only for loading which is now
    // so we now convert them to 8bpp.
    // must be called before RemovePadBytes or remove will fail
    if FBitCount < 8 then
    begin
      case FBitCount of
        1: Convert1To8(Image.Levels[0]);
        4: Convert4To8(Image.Levels[0]);
      end;
      // enlarge palette
      ReallocMem(Palette, 256 * SizeOf(TARGBRec));
    end;

    // BMP contains pad bytes which we do not want
    RemovePadBytes(Image.Levels[0], Image.DataFormat^.BitCount);

    // images were not flipped when decoding
    if BI.Compression in [BI_RLE4, BI_RLE8] then
      if BI.Height > 0 then
        FlipYImage(Image);

    Position := StartPos + BF.Size;
  end;
end;

procedure TNBMPFormat.DoSaveImage(var Image: TImageData; Stream: TNStream;
  SaveCompressed: Boolean);
var
  BF: TBitmapFileHeader;
  BI: TBitmapInfoHeader;
  StartPos: Int64;
  LocalPF: TLocalPixelFormat;
  FSize, Off: Long;
  WidthBytes: Long;
  Data: Pointer;
  DSize, FBitCount: Long;
  Clone: TImageData;
  Level:PImageLevel;

  procedure SaveRGB;
  begin
    Stream.Write(Data^, DSize);
  end;

  procedure SaveRLE8;
  var
    Size: Long;
    B1, B2: Byte;
    PB1, PB2: Long;
    Src, Buff: PByte;
    X, Y: Long;

    function AllocByte: PByte;
    begin
      if Size mod 4096 = 0 then
        ReAllocMem(Buff, Size + 4096);
      Result := Pointer(Long(Buff) + Size);
      Size := Size + 1;
    end;

  begin
    // not necessary
    GetMem(Buff, 1024);
    try
      Size := 0;
      for Y := 0 to Level.Height - 1 do
      begin
        X := 0;
        Src := Pointer(Long(Data) + Y * WidthBytes);
        while X < Level.Width do
        begin
          if (Level.Width - X > 2) and (Src^ = PByte(Long(Src) + 1)^) then
          begin
            B1 := 2;
            B2 := Src^;
            Inc(X, 2);
            Inc(Src, 2);
            while (X < Level.Width) and (Src^ = B2) and (B1 < 255) do
            begin
              Inc(B1);
              Inc(X);
              Inc(Src);
            end;
            AllocByte^ := B1;
            AllocByte^ := B2;
          end
          else
            if (Level.Width - X > 2) and (Src^ <> PByte(Long(Src) + 1)^) and
              (PByte(Long(Src) + 1)^ = PByte(Long(Src) + 2)^) then
            begin
              AllocByte^ := 1;
              AllocByte^ := Src^;
              Inc(Src);
              Inc(X);
            end
            else
            begin
              if (Level.Width - X < 4) then
              begin
                if Level.Width - X = 2 then
                begin
                  AllocByte^ := 1;
                  AllocByte^ := Src^;
                  Inc(Src);
                  AllocByte^ := 1;
                  AllocByte^ := Src^;
                  Inc(Src);
                  Inc(X, 2);
                end
                else
                begin
                  AllocByte^ := 1;
                  AllocByte^ := Src^;
                  Inc(Src);
                  Inc(X);
                end;
              end
              else
              begin
                PB1 := Size;
                AllocByte;
                PB2 := Size;
                AllocByte;
                B1 := 0;
                B2 := 3;
                Inc(X, 3);
                AllocByte^ := Src^;
                Inc(Src);
                AllocByte^ := Src^;
                Inc(Src);
                AllocByte^ := Src^;
                Inc(Src);
                while (X < Clone.Levels[0].Width) and (B2 < 255) do
                begin
                  if (Clone.Levels[0].Width - X > 3) and (Src^ = PByte(Long(Src)
                    +
                    1)^) and
                    (Src^ = PByte(Long(Src) + 2)^) and (Src^ = PByte(Long(Src) +
                    3)^) then
                    Break;
                  AllocByte^ := Src^;
                  Inc(Src);
                  Inc(B2);
                  Inc(X);
                end;
                PByte(Long(Buff) + PB1)^ := B1;
                PByte(Long(Buff) + PB2)^ := B2;
              end;
            end;
          if Size and 1 = 1 then
            AllocByte;
        end;
        AllocByte^ := 0;
        AllocByte^ := 0;
      end;
      AllocByte^ := 0;
      AllocByte^ := 1;
    finally
      Stream.WriteBuffer(Buff^, Size);
      FreeMemNil(Buff);
    end;
  end;

begin
  StartPos := Stream.Position;
  PrepareImage(Image, Clone);
  Level:=FindFirstValidLevel(Clone);
  if Level=nil then
    Level:=@Clone.Levels[0];
  with Stream, Clone, Level^ do
  try
    FBitCount := DataFormat^.BitCount;
    WidthBytes := (((Width * FBitCount) + 31) shr 5) * 4;
    // other fields will be filled later - we don't know all values now
    ZeroMemory(@BF, SizeOf(BF));
    BF.ID := 19778;
    Write(BF, SizeOF(BF));
    // other fields will be filled later - we don't know all values now
    ZeroMemory(@BI, SizeOf(BI));
    BI.Size := SizeOf(BI);
    BI.Width := Width;
    BI.Height := Height;
    BI.Planes := 1;
    BI.BitCount := FBitCount;

    if (FBitCount = 8) and SaveCompressed then
      BI.Compression := BI_RLE8
    else
      if (FBitCount = 16) then
        BI.Compression := BI_BITFIELDS
      else
        BI.Compression := BI_RGB;

    if Clone.Palette <> nil then
    begin
      BI.ClrUsed := DataFormat^.PalSize;
      BI.ClrImportant := DataFormat^.PalSize;
    end;
    Write(BI, SizeOf(BI));

    // write mask info
    if BI.Compression = BI_BITFIELDS then
      with DataFormat^.PixelFormat do
      begin
        LocalPF.RBitMask := RBitMask;
        LocalPF.GBitMask := GBitMask;
        LocalPF.BBitMask := BBitMask;
        Write(LocalPF, SizeOf(LocalPF));
      end;

    // write palette
    if Palette <> nil then
      Write(Palette^, DataFormat^.PalSize * SizeOf(TARGBRec));

    Off := Position;

    AddPadBytes(Level^, FBitCount, Data, DSize);
    case BI.Compression of
      BI_RGB, BI_BITFIELDS: SaveRGB;
      BI_RLE8: SaveRLE8;
    end;
    FreeMemNil(Data);

    FSize := Position - StartPos;
    // rewrite header with new values
    Position := StartPos;
    BF.Size := FSize;
    BF.Offset := Off;
    BI.SizeImage := FSize - Off;
    Write(BF, SizeOf(BF));
    Write(BI, SizeOf(BI));
    Position := StartPos + FSize;
  finally
    FreeImage(Clone);
  end;
end;

function TNBMPFormat.TestFormat(Stream: TNStream): Boolean;
var
  ID: array[0..1] of Char;
  OldPos: Int64;
begin
  Result := False;
  if Stream = nil then Exit;
  if Stream.Size < MinSize then Exit;
  OldPos := Stream.Position;
  Stream.Read(ID, SizeOf(ID));
  Stream.Position := OldPos;
  if CompareMem(@ID, @BMPMagic, SizeOf(ID)) then
    Result := True;
end;


{*************************************************************************}
{                         Internal Stuff                                  }
{*************************************************************************}


initialization

  BMPInstance := TNBMPFormat.Create;

finalization

  FreeAndNil(BMPInstance);

end.

