{*************************************************************************}
{*   JpgFormat                                                           *}
{*   Copyright (C) 2003 Marek Mauder, Vampyre SoftWorks                  *}
{*   Graphic\ImageFileFormats\Jpg\JpgFormat.pas                          *}
{*************************************************************************}
{                                                                         }
{ This program is free software; you can redistribute it and/or modify it }
{ under the terms of the GNU General Public License as published by the   }
{ Free Software Foundation; either version 2 of the License, or (at your  }
{ option) any later version. This program is distributed in the hope that }
{ it will be useful, but WITHOUT ANY WARRANTY; without even the implied   }
{ warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See   }
{ the GNU General Public License for more details.                        }
{                                                                         }
{ You should have received a copy of the GNU General Public License       }
{ along with this program; if not, write to the Free Software Foundation, }
{ Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.         }
{                                                                         }
{*************************************************************************}
{  Created On: 12st September 2003  (12.09.2003)                          }
{*************************************************************************}

{This unit contains image file format loader/saver for JPEG files.}

unit JpgFormat;

{$I Compiler.inc}

interface

uses
  SysUtils, Kernel, Objects, Platforms, Graphic, Images,
  {Uses modified JPEG library shipped with Delphi}
  JpegLib;

type

  {Image encoder for loading/saving JPEG file format.
   Supports: all types loading/saving.}
  TNJPGFormat = class(TNImageFileFormat)
  protected
    FGreyScale: Boolean;
    FProgressive: Boolean;
    FSaveQuality: Long;

    procedure PrepareImage(const Image: TImageData; var Clone: TImageData);
      override;
    procedure DoLoadImage(var Image: TImageData; Stream: TNStream); override;
    procedure DoSaveImage(var Image: TImageData; Stream: TNStream;
      SaveCompressed: Boolean = True); override;
  public
    constructor Create; override;
    function TestFormat(Stream: TNStream): Boolean; override;

    {JPEG save quality  1(small/ugly file) .. 100(large/nice file) default = 90}
    property SaveQuality: Long read FSaveQuality write FSaveQuality;
    {If True JPEG is saved as progressive image. Default = False}
    property Progressive: Boolean read FProgressive write FProgressive;
  end;

const

  sJPGExtensions = 'jpg,jpeg,jif,jfif';
  sJPGFormatName = 'Joint Photographic Experts Group';

var

  JPGInstance: TNJPGFormat = nil;


{*************************************************************************}
{                             IMPLEMEMTATION                              }
{*************************************************************************}


implementation

const
  {JPG file identifier}
  JPGMagic: array[0..2] of Char = #$FF#$D8#$FF;

procedure JpegError(cinfo: j_common_ptr);
{$IFDEF LINUX} cdecl; export;
{$ENDIF}
begin
end;

procedure EmitMessage(cinfo: j_common_ptr; msg_level: Integer);
{$IFDEF LINUX} cdecl; export;
{$ENDIF}
begin
end;

procedure OutputMessage(cinfo: j_common_ptr);
{$IFDEF LINUX} cdecl; export;
{$ENDIF}
begin
end;

procedure FormatMessage(cinfo: j_common_ptr; buffer: PChar);
{$IFDEF LINUX} cdecl; export;
{$ENDIF}
begin
end;

procedure ResetErrorMgr(cinfo: j_common_ptr);
{$IFDEF LINUX} cdecl; export;
{$ENDIF}
begin
  cinfo^.err^.num_warnings := 0;
  cinfo^.err^.msg_code := 0;
end;

const
  jpeg_std_error: jpeg_error_mgr = (
    error_exit: JpegError;
    emit_message: EmitMessage;
    output_message: OutputMessage;
    format_message: FormatMessage;
    reset_error_mgr: ResetErrorMgr);

procedure ReleaseContext(var jc: TJPEGContext);
begin
  if jc.common.err = nil then Exit;
  jpeg_destroy(jc.common);
  jc.common.err := nil;
end;

procedure InitDecompressor(Stream: TNStream; var jc: TJPEGContext);
begin
  ZeroMemory(@jc, sizeof(jc));
  jc.err := jpeg_std_error;
  jc.common.err := @jc.err;
  jpeg_CreateDecompress(jc.d, JPEG_LIB_VERSION, sizeof(jc.d));
  try
{$IFDEF MSWINDOWS}
    jpeg_stdio_src(jc.d, Stream);
{$ENDIF}
{$IFDEF LINUX}
    JPEGStdioSrc(jc.d, Stream);
{$ENDIF}
    jpeg_read_header(jc.d, True);
    jc.d.scale_num := 1;
    jc.d.scale_denom := 1;
    jc.d.do_block_smoothing := true;
    if jc.d.out_color_space = JCS_GRAYSCALE then
    begin
      jc.d.quantize_colors := True;
      jc.d.desired_number_of_colors := 256;
    end;
    jc.FinalDCT := jc.d.dct_method;
    jc.FinalTwoPassQuant := jc.d.two_pass_quantize;
    jc.FinalDitherMode := jc.d.dither_mode;
    if jpeg_has_multiple_scans(jc.d) then
    begin
      jc.d.enable_2pass_quant := jc.d.two_pass_quantize;
      jc.d.dct_method := JDCT_IFAST;
      jc.d.two_pass_quantize := False;
      jc.d.dither_mode := JDITHER_ORDERED;
      jc.d.buffered_image := True;
    end;
  except
    ReleaseContext(jc);
    raise Exception.Create('Error when initializing JPEG decompressor!');
  end;
end;


{*************************************************************************}
{                        TNJPGFormat methods                              }
{*************************************************************************}


constructor TNJPGFormat.Create;
begin
  inherited Create;
  FExtensions := sJPGExtensions;
  FFormatName := sJPGFormatName;
  FCanLoad := True;
  FCanSave := True;
  FSaveQuality := 90;
  FProgressive := False;
  FGreyScale := False;
end;

procedure TNJPGFormat.PrepareImage(const Image: TImageData;
  var Clone: TImageData);
begin
  inherited PrepareImage(Image, Clone);
  with Clone.DataFormat^ do
  begin
    FGreyScale := OneChannelFormat;
    ConvertFunc(Clone, FindImageDataFormat(idfR8G8B8));
    {$IFDEF LINUX}
    Images.SwapChannels(Clone);
    {$ENDIF}
  end;
end;

procedure TNJPGFormat.DoLoadImage(var Image: TImageData; Stream: TNStream);
var
  LinesPerCall, LinesRead: Long;
  Dest: PChar;
{$IFDEF LINUX}
  Swap: Char;
  CurPixel: PChar;
{$ENDIF}
  PtrInc: Long;
  jc: TJPEGContext;
  FBitCount: Long;

begin
  AddImageLevel(Image);
  InitDecompressor(Stream, jc);
  with Stream, Image, Image.Levels[0] do
  try
    if (jc.d.out_color_space = JCS_GRAYSCALE) then
      FBitCount := 8
    else
      FBitCount := 24;

    jpeg_start_decompress(jc.d);

    case FBitCount of
      8: DataFormat := FindImageDataFormat(idfIntensity8);
      24: DataFormat := FindImageDataFormat(idfR8G8B8);
    end;
    Width := jc.d.output_width;
    Height := jc.d.output_height;
    Size := Width * Height * (FBitCount shr 3);
    GetMem(Bits, Size);
    Dest := Bits;
    PtrInc := Width * (FBitCount shr 3);
    LinesPerCall := jc.d.rec_outbuf_height;

    // support for progressive images
    if jc.d.buffered_image then
    begin
      while jpeg_consume_input(jc.d) <> JPEG_REACHED_EOI do
      begin
        jpeg_start_output(jc.d, jc.d.input_scan_number);
        Dest := Bits;
        while (jc.d.output_scanline < jc.d.output_height) do
        begin
          LinesRead := jpeg_read_scanlines(jc.d, @Dest, LinesPerCall);
{$IFDEF LINUX}
          if FBitCount = 24 then
          begin
            CurPixel := Dest;
            while (CurPixel - Dest) < (Width * 3) do
            begin
              Swap := CurPixel[0];
              CurPixel[0] := CurPixel[2];
              CurPixel[2] := Swap;
              Inc(CurPixel, 3);
            end;
          end;
{$ENDIF}
          Inc(Long(Dest), PtrInc * LinesRead);
        end;
        jpeg_finish_output(jc.d);
      end;
        // reset options for final pass at requested quality
      jc.d.dct_method := jc.FinalDCT;
      jc.d.dither_mode := jc.FinalDitherMode;
      if jc.FinalTwoPassQuant then
      begin
        jc.d.two_pass_quantize := True;
        jc.d.colormap := nil;
      end;
      jpeg_start_output(jc.d, jc.d.input_scan_number);
      Dest := Bits;
    end;

      // final image pass for progressive, first and only pass for baseline
    while (jc.d.output_scanline < jc.d.output_height) do
    begin
      LinesRead := jpeg_read_scanlines(jc.d, @Dest, LinesPerCall);
{$IFDEF LINUX}
      if FBitCount = 24 then
      begin
        CurPixel := Dest;
        while (CurPixel - Dest) < (Width * 3) do
        begin
          Swap := CurPixel[0];
          CurPixel[0] := CurPixel[2];
          CurPixel[2] := Swap;
          Inc(CurPixel, 3);
        end;
      end;
{$ENDIF}
      Inc(Long(Dest), PtrInc * LinesRead);
    end;

    if jc.d.buffered_image then
      jpeg_finish_output(jc.d);
  finally
    jpeg_finish_decompress(jc.d);
    ReleaseContext(jc);
  end;
end;

procedure TNJPGFormat.DoSaveImage(var Image: TImageData; Stream: TNStream;
  SaveCompressed: Boolean);
var
  Clone: TImageData;
  LinesWritten, LinesPerCall: Long;
  Src: Pointer;
  PtrInc: Long;
  jc: TJPEGContext;
  Level: PImageLevel;
begin
  PrepareImage(Image, Clone);
  Level := FindFirstValidLevel(Clone);
  if Level = nil then
    Level := @Clone.Levels[0];
  with Stream, Clone, Level^ do
  try
    ZeroMemory(@jc, SizeOf(jc));
    jc.err := jpeg_std_error;
    jc.common.err := @jc.err;
    jpeg_CreateCompress(jc.c, JPEG_LIB_VERSION, sizeof(jc.c));

{$IFDEF MSWINDOWS}
    jpeg_stdio_dest(jc.c, Stream);
{$ENDIF}
{$IFDEF LINUX}
    JPEGStdioDest(jc.c, Stream);
{$ENDIF}

    jc.c.image_width := Width;
    jc.c.image_height := Height;
    jc.c.input_components := 3; // JPEG requires 24bit RGB input
    jc.c.in_color_space := JCS_RGB;

    jpeg_set_defaults(jc.c);
    jpeg_set_quality(jc.c, FSaveQuality, True);

    if FGreyScale then
      jpeg_set_colorspace(jc.c, JCS_GRAYSCALE);

    if FProgressive then
      jpeg_simple_progression(jc.c);

    Src := Bits;
    PtrInc := Width * (DataFormat^.BitCount shr 3);
    LinesPerCall := 1;

    jpeg_start_compress(jc.c, true);

    while (jc.c.next_scanline < jc.c.image_height) do
    begin
      LinesWritten := jpeg_write_scanlines(jc.c, @Src,
        LinesPerCall);
      Inc(Long(Src), PtrInc * LinesWritten);
    end;

    jpeg_finish_compress(jc.c);
  finally
    ReleaseContext(jc);
    FreeImage(Clone);
  end;
end;

function TNJPGFormat.TestFormat(Stream: TNStream): Boolean;
var
  OldPos: Int64;
  ID: array[0..2] of Char;
begin
  Result := False;
  if Stream = nil then Exit;
  OldPos := Stream.Position;
  ZeroMemory(@ID, SizeOf(ID));
  Stream.Read(ID, SizeOf(ID));
  Stream.Position := OldPos;
  if CompareMem(@ID, @JPGMagic, SizeOf(ID)) then
    Result := True;
end;


{*************************************************************************}
{                         Internal Stuff                                  }
{*************************************************************************}


initialization

  JPGInstance := TNJPGFormat.Create;

finalization

  FreeAndNil(JPGInstance);

end.

