{*************************************************************************}
{*   PngFormat                                                           *}
{*   Copyright (C) 2003 Marek Mauder, Vampyre SoftWorks                  *}
{*   Graphic\ImageFileFormats\Png\PngFormat.pas                          *}
{*************************************************************************}
{                                                                         }
{ This program is free software; you can redistribute it and/or modify it }
{ under the terms of the GNU General Public License as published by the   }
{ Free Software Foundation; either version 2 of the License, or (at your  }
{ option) any later version. This program is distributed in the hope that }
{ it will be useful, but WITHOUT ANY WARRANTY; without even the implied   }
{ warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See   }
{ the GNU General Public License for more details.                        }
{                                                                         }
{ You should have received a copy of the GNU General Public License       }
{ along with this program; if not, write to the Free Software Foundation, }
{ Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.         }
{                                                                         }
{*************************************************************************}
{  Created On: 12st September 2003  (12.09.2003)                          }
{*************************************************************************}

{This unit contains image file format loader/saver for PNG files.}

unit PngFormat;

{$I Compiler.inc}

interface

uses
  SysUtils, Kernel, Objects, Platforms, Graphic, Images,
  {Uses ZLib for compression/decompression.}
  ZLibEx;

type

  {Image encoder for loading/saving png file format.
   Supports: 1,4,8,16,24,32bpp images(filtered/noninterlaced/interlaced) loading
             8,16,24,32bpp noninterlaced nonfiltered saving.
   Compression: ZLIB.}
  TNPNGFormat = class(TNImageFileFormat)
  protected
    FStream: TZStreamRec;

    procedure PrepareImage(const Image: TImageData; var Clone: TImageData);
      override;
    procedure DoLoadImage(var Image: TImageData; Stream: TNStream); override;
    procedure DoSaveImage(var Image: TImageData; Stream: TNStream;
      SaveCompressed: Boolean); override;
  public
    constructor Create; override;
    function TestFormat(Stream: TNStream): Boolean; override;
  end;

const

  sPNGExtensions = 'png';
  sPNGFormatName = 'Portable Network Graphics';

var

  PNGInstance: TNPNGFormat = nil;


{*************************************************************************}
{                             IMPLEMEMTATION                              }
{*************************************************************************}


implementation

type

  {Chunk header}
  TChunkHdr = packed record
    DataSize: ULong;
    ChunkID: ULong;
  end;

  {(*}

  {IHDR chunk format}
  TIHDR = packed record
    ImgWidth: ULong;
    ImgHeight: ULong;
    BitDepth: Byte;    // bits per pixel  or bits per sample (for truecolor)
    ColorType: Byte;   // 0 = greyscale, 2 = truecolor, 3 = palette,
                       // 4 = grey+alpha, 6 = truecolor + alpha
    Compress: Byte;    // compression type  00 = ZLib
    Filter: Byte;      // used precompress filter
    Interlaced: Byte;  // 0 = no int; 1 = Adam7 interlacing
  end;

  {*)}

const

  {Constants for chunk identifiers and signature identifiers.
   They are already in big-endian format.}
  IHDRChunk = $49484452;
  GAMAChunk = $67414D41;
  PLTEChunk = $504C5445;
  BKGDChunk = $624B4744;
  TRNSChunk = $74524E53;
  IDATChunk = $49444154;
  IENDChunk = $49454E44;
  CheckSig: array[0..7] of Byte = ($89, $50, $4E, $47, $0D, $0A, $1A, $0A);

  { Interlace start and offsets}
  RowStart: array[0..6] of Long = (0, 0, 4, 0, 2, 0, 1);
  ColumnStart: array[0..6] of Long = (0, 4, 0, 2, 0, 1, 0);
  RowIncrement: array[0..6] of Long = (8, 8, 8, 4, 4, 2, 2);
  ColumnIncrement: array[0..6] of Long = (8, 8, 4, 4, 2, 2, 1);
  PassMask: array[0..6] of Byte = ($80, $08, $88, $22, $AA, $55, $FF);


function PaethPredictor(A, B, C: Byte): Byte;
var
  P, PA, PB, PC: Long;

begin
  P := A + B - C;
  PA := Abs(P - A);
  PB := Abs(P - B);
  PC := Abs(P - C);
  if (PA <= PB) and (PA <= PC) then
    Result := A
  else
    if PB <= PC then
      Result := B
    else
      Result := C;
end;

procedure ApplyFilter(Filter: Byte; Line, PrevLine, Target: PByte; BPP,
  BytesPerRow: Long);
var
  I: Long;
  Raw,
    Decoded,
    Prior,
    PriorDecoded,
    TargetRun: PByte;
begin
  case Filter of
    0: // no filter, just copy data
      Move(Line^, Target^, BytesPerRow);
    1: // subtraction filter
      begin
        Raw := Line;
        TargetRun := Target;
        // Transfer BPP bytes without filtering. This mimics the effect of bytes left to the
        // scanline being zero.
        Move(Raw^, TargetRun^, BPP);
        // now do rest of the line
        Decoded := TargetRun;
        Inc(Raw, BPP);
        Inc(TargetRun, BPP);
        Dec(BytesPerRow, BPP);
        while BytesPerRow > 0 do
        begin
          TargetRun^ := Byte(Raw^ + Decoded^);
          Inc(Raw);
          Inc(Decoded);
          Inc(TargetRun);
          Dec(BytesPerRow);
        end;
      end;
    2: // Up filter
      begin
        Raw := Line;
        Prior := PrevLine;
        TargetRun := Target;
        while BytesPerRow > 0 do
        begin
          TargetRun^ := Byte(Raw^ + Prior^);
          Inc(Raw);
          Inc(Prior);
          Inc(TargetRun);
          Dec(BytesPerRow);
        end;
      end;
    3: // average filter
      begin
        // first handle BPP virtual pixels to the left
        Raw := Line;
        Decoded := Line;
        Prior := PrevLine;
        TargetRun := Target;
        for I := 0 to BPP - 1 do
        begin
          TargetRun^ := Byte(Raw^ + Floor(Prior^ / 2));
          Inc(Raw);
          Inc(Prior);
          Inc(TargetRun);
        end;
        Dec(BytesPerRow, BPP);
        // now do rest of line
        while BytesPerRow > 0 do
        begin
          TargetRun^ := Byte(Raw^ + Floor((Decoded^ + Prior^) / 2));
          Inc(Raw);
          Inc(Decoded);
          Inc(Prior);
          Inc(TargetRun);
          Dec(BytesPerRow);
        end;
      end;
    4: // paeth prediction
      begin
       // again, start with first BPP pixel which would refer to non-existing pixels to the left
        Raw := Line;
        Decoded := Target;
        Prior := PrevLine;
        PriorDecoded := PrevLine;
        TargetRun := Target;
        for I := 0 to BPP - 1 do
        begin
          TargetRun^ := Byte(Raw^ + PaethPredictor(0, Prior^, 0));
          Inc(Raw);
          Inc(Prior);
          Inc(TargetRun);
        end;
        Dec(BytesPerRow, BPP);
        // finally do rest of line
        while BytesPerRow > 0 do
        begin
          TargetRun^ := Byte(Raw^ + PaethPredictor(Decoded^, Prior^,
            PriorDecoded^));
          Inc(Raw);
          Inc(Decoded);
          Inc(Prior);
          Inc(PriorDecoded);
          Inc(TargetRun);
          Dec(BytesPerRow);
        end;
      end;
  end;
end;

procedure Convert1To8(var Level: TImageLevel);
const
  Mask1: array[0..7] of Byte = ($80, $40, $20, $10, $08, $04, $02, $01);
  Shift1: array[0..7] of Byte = (7, 6, 5, 4, 3, 2, 1, 0);
var
  Data: Pointer;
  X, Y: Long;
  WB: Long;
begin
  with Level do
  begin
    GetMem(Data, Width * Height);
    WB := ((Width + 31) shr 5) * 4;
    try
      for X := 0 to Width - 1 do
        for Y := 0 to Height - 1 do
          PByte(Long(Data) + Y * Width + X)^ :=
            (PByteArray(Integer(Bits) + Y * WB)[X shr 3] and Mask1[X and 7])
            shr Shift1[X and 7];
    finally
      FreeMemNil(Bits);
      Bits := Data;
      Size := Width * Height;
    end;
  end;
end;

procedure Convert4To8(var Level: TImageLevel);
const
  Mask4: array[0..1] of Byte = ($F0, $0F);
  Shift4: array[0..1] of Byte = (4, 0);
var
  Data: Pointer;
  X, Y: Long;
  WB: Long;
begin
  with Level do
  begin
    GetMem(Data, Width * Height);
    WB := (((Width * 4) + 31) shr 5) * 4;
    try
      for X := 0 to Width - 1 do
        for Y := 0 to Height - 1 do
          PByte(Long(Data) + Y * Width + X)^ :=
            (PByteArray(Integer(Bits) + Y * WB)[X shr 1] and Mask4[X and 1])
            shr Shift4[X and 1];
    finally
      FreeMemNil(Bits);
      Bits := Data;
      Size := Width * Height;
    end;
  end;
end;


{*************************************************************************}
{                        TNPNGFormat methods                              }
{*************************************************************************}


constructor TNPNGFormat.Create;
begin
  inherited Create;
  FExtensions := sPNGExtensions;
  FFormatName := sPNGFormatName;
  FCanLoad := True;
  FCanSave := True;
end;

procedure TNPNGFormat.PrepareImage(const Image: TImageData;
  var Clone: TImageData);
begin
  inherited PrepareImage(Image, Clone);
  with Clone.DataFormat^ do
  begin
    if OneChannelFormat and (OneChannelSize > 16) then
      ConvertFunc(Clone, FindImageDataFormat(idfIntensity16))
    else
      if FloatFormat then
        ConvertFunc(Clone, FindImageDataFormat(idfA8R8G8B8))
      else
        if Indexed and (BitCount > 8) then
          ConvertFunc(Clone, FindImageDataFormat(idfIndex8))
        else
          if not FixedPixelFormat then
          begin
            if HasAlpha then
              ConvertFunc(Clone, FindImageDataFormat(idfA8R8G8B8))
            else
              ConvertFunc(Clone, FindImageDataFormat(idfR8G8B8))
          end;

    // final check
    if (Clone.DataFormat <> FindImageDataFormat(idfA8R8G8B8)) and
      (Clone.DataFormat <> FindImageDataFormat(idfR8G8B8)) and
      (Clone.DataFormat <> FindImageDataFormat(idfA16R16G16B16)) and
      (Clone.DataFormat <> FindImageDataFormat(idfR16G16B16)) and
      (Clone.DataFormat <> FindImageDataFormat(idfIntensity8)) and
      (Clone.DataFormat <> FindImageDataFormat(idfA8Intensity8)) and
      (Clone.DataFormat <> FindImageDataFormat(idfIntensity16)) and
      (Clone.DataFormat <> FindImageDataFormat(idfIndex8)) then
      ConvertFunc(Clone, FindImageDataFormat(idfR8G8B8));

    if not Indexed then
      SwapChannels(Clone);
  end;
end;

procedure TNPNGFormat.DoLoadImage(var Image: TImageData; Stream: TNStream);
var
  ID: Int64;
  Buff: Pointer;
  ChunkHdr: TChunkHdr;
  IHDR: TIHDR;
  CRC: ULong;
  FRawData: Pointer;
  FCompData: Pointer;
  FBitCount, FPalSize: Long;

  procedure DecodeInit;
  begin
    FillChar(FStream, sizeof(FStream), 0);
    FStream.zalloc := zcAlloc;
    FStream.zfree := zcFree;
    InflateInit_(FStream, zlib_version, sizeof(FStream));
  end;

  procedure Decode(var InData, OutData: Pointer; PackedSize,
    UnpackedSize: Long);
  begin
    FStream.next_in := InData;
    FStream.avail_in := PackedSize;
    FStream.next_out := OutData;
    FStream.avail_out := UnpackedSize;
    Inflate(FStream, Z_PARTIAL_FLUSH);
    // advance pointers so used input can be calculated
    InData := FStream.Next_In;
    OutData := FStream.Next_Out;
  end;

  procedure DecodeEnd;
  begin
    InflateEnd(FStream);
  end;

  procedure ReadRow(RowBuffer: Pointer; BytesPerRow: Integer);
  var
    LocalBuffer: Pointer;
    PendingOutput: Integer;
    OPos: Long;

  begin
    LocalBuffer := RowBuffer;
    PendingOutput := BytesPerRow;
    repeat
      if ChunkHdr.DataSize + ULong(FRawData) = ULong(FCompData) then
      begin
        // read next IDAT chunk
        OPos := Stream.Position;
        Stream.Read(CRC, 4);
        Stream.Read(ChunkHdr, SizeOf(ChunkHdr));
        SwapLong(@ChunkHdr, 2);
        if ChunkHdr.ChunkID = IDATChunk then
        begin
          FreeMemNil(FRawData);
          GetMem(FRawData, ChunkHdr.DataSize);
          Stream.Read(FRawData^, ChunkHdr.DataSize);
          FCompData := FRawData;
        end
        else
        begin
          Stream.Position := OPos;
          Exit;
        end;
      end;
      // this decode call will advance InData and OutData accordingly
      Decode(FCompData, LocalBuffer,
        ChunkHdr.DataSize - (ULong(FCompData) - ULong(FRawData)),
        PendingOutput);
      PendingOutput := BytesPerRow - (Long(LocalBuffer) - Long(RowBuffer));
    until PendingOutput = 0;
  end;

  procedure LoadIDAT;
  var
    I: Long;
    RowBuffer: array[Boolean] of PChar;
    EvenRow: boolean;
    Pass: Long;
    BytesPerRow,
      InterlaceRowBytes,
      InterlaceWidth: Long;

    procedure DecodeAdam7;
    const
      BitTable: array[1..8] of Integer = ($1, $3, 0, $F, 0, 0, 0, $FF);
      StartBit: array[1..8] of Integer = (7, 0, 0, 4, 0, 0, 0, 0);
    var
      Dst2: PChar;
      Src, Dst: PByte;
      CurBit, Col: Long;
    begin
      Src := PByte(Long(RowBuffer[EvenRow]) + 1);
      Col := ColumnStart[Pass];
      with Image do
        case FBitCount of
          1, 4, 8:
            begin
              Dst := Pointer(Long(Image.Levels[0].Bits) + I * (BytesPerRow -
                1));
              repeat
                CurBit := StartBit[FBitCount];
                repeat
                  Dst2 := PChar(Long(Dst) + (FBitCount * Col) shr 3);
                  Byte(Dst2^) := Byte(Dst2^) or
                    (((Byte(Src^) shr CurBit) and BitTable[FBitCount])
                    shl (StartBit[FBitCount] - (Col * FBitCount mod 8)));

                  Inc(Col, ColumnIncrement[Pass]);
                  Dec(CurBit, FBitCount);
                until CurBit < 0;
                Inc(Src);
              until Col >= Levels[0].Width;
            end;
          16:
            begin
              Dst := Pointer(Long(Image.Levels[0].Bits) + I * (BytesPerRow - 1)
                + Col * 2);
              repeat
                Byte(Dst^) := PByte(Long(Src) + 0)^;
                Inc(Dst);
                Byte(Dst^) := PByte(Long(Src) + 1)^;
                Inc(Dst);
                Inc(Src, 2);
                Inc(Dst, ColumnIncrement[Pass] * 2 - 2);
                Inc(Col, ColumnIncrement[Pass]);
              until Col >= Levels[0].Width;
            end;
          24:
            begin
              Dst := Pointer(Long(Image.Levels[0].Bits) + I * (BytesPerRow - 1)
                + Col * 3);
              repeat
                Byte(Dst^) := PByte(Long(Src) + 0)^;
                Inc(Dst);
                Byte(Dst^) := PByte(Long(Src) + 1)^;
                Inc(Dst);
                Byte(Dst^) := PByte(Long(Src) + 2)^;
                Inc(Dst);
                Inc(Src, 3);
                Inc(Dst, ColumnIncrement[Pass] * 3 - 3);
                Inc(Col, ColumnIncrement[Pass]);
              until Col >= Levels[0].Width;
            end;
          32:
            begin
              Dst := Pointer(Long(Image.Levels[0].Bits) + I * (BytesPerRow - 1)
                + Col * 4);
              repeat
                Byte(Dst^) := PByte(Long(Src) + 0)^;
                Inc(Dst);
                Byte(Dst^) := PByte(Long(Src) + 1)^;
                Inc(Dst);
                Byte(Dst^) := PByte(Long(Src) + 2)^;
                Inc(Dst);
                Byte(Dst^) := PByte(Long(Src) + 3)^;
                Inc(Dst);
                Inc(Src, 4);
                Inc(Dst, ColumnIncrement[Pass] * 4 - 4);
                Inc(Col, ColumnIncrement[Pass]);
              until Col >= Levels[0].Width;
            end;
        end;
    end;

  begin
    RowBuffer[False] := nil;
    RowBuffer[True] := nil;
    with Image, Image.Levels[0] do
    try
      // maximum number of bytes per row and consider there's one filter byte
      // at the start of each row
      // +7 is here fot 1/4 bit images support
      BytesPerRow := ((Width * FBitCount + 7) shr 3) + 1;

      GetMem(FRawData, ChunkHdr.DataSize);
      Stream.Read(FRawData^, ChunkHdr.DataSize);
      FCompData := FRawData;
      Size := Height * (BytesPerRow - 1);
      GetMem(Bits, Size);
      GetMem(RowBuffer[True], BytesPerRow);
      GetMem(RowBuffer[False], BytesPerRow);
      // there can be more than one IDAT chunk in the file but then they must directly
      // follow each other (handled in ReadRow)
      EvenRow := True;
      DecodeInit;
      // prepare interlaced images
      if IHDR.Interlaced = 1 then
      begin
        for Pass := 0 to 6 do
        begin
          // prepare next interlace run
          if Width <= ColumnStart[Pass] then
            Continue;
          InterlaceWidth := (Width + ColumnIncrement[Pass] - 1 -
            ColumnStart[Pass])
            div ColumnIncrement[Pass];
          InterlaceRowBytes := ((InterlaceWidth * FBitCount + 7) shr 3) + 1;
          I := RowStart[Pass];
          while I < Height do
          begin
            ReadRow(RowBuffer[EvenRow], InterlaceRowBytes);
            ApplyFilter(Byte(RowBuffer[EvenRow]^),
              Pointer(RowBuffer[EvenRow] + 1),
              Pointer(RowBuffer[not EvenRow] + 1),
              Pointer(RowBuffer[EvenRow] + 1),
              (FBitCount + 7) shr 3,
              InterlaceRowBytes - 1);
            DecodeAdam7;
            EvenRow := not EvenRow;
            // continue with next row in interlaced order
            Inc(I, RowIncrement[Pass]);
          end;
        end;
      end
      else
      begin
        for I := 0 to Height - 1 do
        begin
          ReadRow(RowBuffer[EvenRow], BytesPerRow);
          ApplyFilter(Byte(RowBuffer[EvenRow]^),
            Pointer(RowBuffer[EvenRow] + 1),
            Pointer(RowBuffer[not EvenRow] + 1),
            Pointer(RowBuffer[EvenRow] + 1),
            (FBitCount + 7) shr 3,
            BytesPerRow - 1);
          Move(Pointer(RowBuffer[EvenRow] + 1)^,
            Pointer(Long(Bits) + I * (BytesPerRow - 1))^, BytesPerRow - 1);
          EvenRow := not EvenRow;
        end;
      end;
    finally
      FreeMemNil(FRawData);
      FreeMemNil(RowBuffer[True]);
      FreeMemNil(RowBuffer[False]);
      DecodeEnd;
    end;
  end;

  procedure LoadPLTE;
  var
    Pal: PPalRGBArray;

    function ConvPal(Pal: PPalRGBArray; Entries: Long): PPalArray;
    var
      I: Long;
    begin
      GetMemFill(Result, Entries * SizeOf(TARGB));
      for I := 0 to Entries - 1 do
      begin
        Result[I].A := $FF;
        Result[I].R := Pal[I].R;
        Result[I].G := Pal[I].G;
        Result[I].B := Pal[I].B;
      end;
    end;

  begin
    if IHDR.ColorType <> 3 then Exit;
    with Image do
    try
      GetMemFill(Pal, ChunkHdr.DataSize);
      Stream.Read(Pal^, ChunkHdr.DataSize);
      FPalSize := ChunkHdr.DataSize div 3;
      Palette := ConvPal(Pal, FPalSize);
    finally
      FreeMemNil(Pal);
    end;
  end;

begin
  AddImageLevel(Image);
  with Stream, Image, Image.Levels[0] do
  begin
    // read signature
    Stream.Read(ID, 8);
    // read IHDR
    Stream.Read(ChunkHdr, SizeOf(ChunkHdr));
    Stream.Read(IHDR, SizeOf(IHDR));
    Stream.Read(CRC, SizeOf(CRC));
    // swap width and height, next values are single bytes
    SwapLong(@IHDR, 2);
    Width := IHDR.ImgWidth;
    Height := IHDR.ImgHeight;
    case IHDR.ColorType of
      0: case IHDR.BitDepth of
          1..8:
            begin
              FBitCount := IHDR.BitDepth;
              DataFormat := FindImageDataFormat(idfIntensity8);
            end;
          16:
            begin
              FBitCount := 16;
              DataFormat := FindImageDataFormat(idfIntensity16);
            end;
        end;
      2: case IHDR.BitDepth of
          8:
            begin
              FBitCount := 24;
              DataFormat := FindImageDataFormat(idfR8G8B8);
            end;
          16:
            begin
              FBitCount := 48;
              DataFormat := FindImageDataFormat(idfR16G16B16);
            end;
        end;
      3:
        begin
          case IHDR.BitDepth of
            1: FBitCount := 1;
            4: FBitCount := 4;
            8: FBitCount := 8;
          end;
          // all indexed modes will be converted to 8bit
          DataFormat := FindImageDataFormat(idfIndex8);
        end;
      // greyscale + alpha hybrid format
      // only A8I8 is supported, others are treat like simple intensity formats.
      4:
        begin
          FBitCount := IHDR.BitDepth * 2;
          case IHDR.BitDepth of
            1..4: DataFormat := FindImageDataFormat(idfIntensity8);
            8: DataFormat := FindImageDataFormat(idfA8Intensity8);
            16: DataFormat := FindImageDataFormat(idfIntensity32);
          end;
        end;
      6: case IHDR.BitDepth of
          8:
            begin
              FBitCount := 32;
              DataFormat := FindImageDataFormat(idfA8R8G8B8);
            end;
          16:
            begin
              FBitCount := 64;
              DataFormat := FindImageDataFormat(idfA16R16G16B16);
            end;
        end;
    end;

    Stream.Read(ChunkHdr, SizeOf(ChunkHdr));
    SwapLong(@ChunkHdr, 2);
    // read chunks until IEND chunk is reached
    while ChunkHdr.ChunkID <> IENDCHUNK do
    begin
      case ChunkHdr.ChunkID of
        IDATChunk: LoadIDAT;
        PLTEChunk: LoadPLTE;
        else
          // skip unsupported chunks
          if Stream.Size - Stream.Position > ChunkHdr.DataSize then
          begin
            GetMem(Buff, ChunkHdr.DataSize);
            Stream.ReadBuffer(Buff^, ChunkHdr.DataSize);
            FreeMemNil(Buff);
          end
          else
            // some programs do not add IEND !! (ACDSee)
            Break;
      end;
      Stream.Read(CRC, SizeOf(CRC));
      Stream.Read(ChunkHdr, SizeOf(ChunkHdr));
      SwapLong(@ChunkHdr, 2);
    end;

    case FBitCount of
      1: Convert1To8(Image.Levels[0]);
      4: Convert4To8(Image.Levels[0]);
    end;
    SwapChannels(Image);
  end;
end;

procedure TNPNGFormat.DoSaveImage(var Image: TImageData; Stream: TNStream;
  SaveCompressed: Boolean);
var
  ChunkHdr: TChunkHdr;
  IHDR: TIHDR;
  CRC: ULong;
  Clone: TImageData;
  Level: PImageLevel;

  procedure SaveIDAT;
  var
    BytesPerRow,
      OutSize,
      OPos, OPos2,
      I: Long;
    Buff, InData,
      OutData: Pointer;
  begin
    with Clone, Level^ do
    try
      BytesPerRow := Width * (DataFormat^.BitCount shr 3);
      OPos := Stream.Position;
      ChunkHdr.DataSize := 0;
      ChunkHdr.ChunkID := SwapLong(IDATChunk);
      Stream.Write(ChunkHdr, SizeOf(ChunkHdr));
      CRC := $FFFFFFFF;
      CalcCRC32(Pointer(Long(@ChunkHdr) + 4), 4, CRC);
      GetMem(InData, BytesPerRow + 1);
      GetMem(Buff, Size + Height);
      for I := 0 to Height - 1 do
      begin
        Move(Pointer(Long(Bits) + I * BytesPerRow)^, Pointer(Long(InData) + 1)^,
          BytesPerRow);
        PByte(InData)^ := 0;
        Move(InData^, Pointer(Long(Buff) + I * (BytesPerRow + 1))^,
          BytesPerRow + 1);
      end;
      CompressBuffer(Buff, Size + Height, OutData, OutSize);
      Stream.Write(OutData^, OutSize);
      OPos2 := Stream.Position;
      ChunkHdr.DataSize := SwapLong(OPos2 - Opos - SizeOf(ChunkHdr));
      Stream.Position := OPos;
      Stream.Write(ChunkHdr, SizeOf(ChunkHdr));
      Stream.Position := OPos2;
      CalcCRC32(OutData, OutSize, CRC);
      CRC := SwapLong(CRC xor $FFFFFFFF);
      Stream.Write(CRC, 4);
    finally
      FreeMemNil(InData);
      FreeMemNil(OutData);
      FreeMemNil(Buff);
    end;
  end;

  procedure SavePLTE;
  var
    Pal: TPalRGB;
    I: Long;
  begin
    with Clone do
      for I := 0 to DataFormat^.PalSize - 1 do
      begin
        Pal[I].B := Palette[I].R;
        Pal[I].G := Palette[I].G;
        Pal[I].R := Palette[I].B;
      end;
    ChunkHdr.ChunkID := PLTEChunk;
    ChunkHdr.DataSize := SizeOf(Pal);
    SwapLong(@ChunkHdr, 2);
    Stream.Write(ChunkHdr, SizeOf(ChunkHdr));
    Stream.Write(Pal, SizeOf(Pal));
    CRC := $FFFFFFFF;
    CalcCRC32(Pointer(Long(@ChunkHdr) + 4), 4, CRC);
    CalcCRC32(@Pal, SizeOf(Pal), CRC);
    CRC := SwapLong(CRC xor $FFFFFFFF);
    Stream.Write(CRC, 4);
  end;

begin
  PrepareImage(Image, Clone);
  Level := FindFirstValidLevel(Clone);
  if Level = nil then
    Level := @Clone.Levels[0];
  with Stream, Clone, Level^ do
  try
    Stream.Write(CheckSig, 8);
    ChunkHdr.DataSize := SizeOf(IHDR);
    ChunkHdr.ChunkID := IHDRChunk;
    SwapLong(@ChunkHdr, 2);
    IHDR.ImgWidth := Width;
    IHDR.ImgHeight := Height;
    IHDR.Compress := 0;
    IHDR.Filter := 0;
    IHDR.Interlaced := 0;
    IHDR.BitDepth := DataFormat^.OneChannelSize;

    if DataFormat^.OneChannelFormat then
    begin
      IHDR.ColorType := 0;
      if DataFormat^.HasAlpha then IHDR.ColorType := 4;
    end
    else
      if DataFormat^.Indexed then
      begin
        IHDR.ColorType := 3;
      end
      else
        if DataFormat^.FixedPixelFormat then
        begin
          IHDR.ColorType := 2;
          if DataFormat^.HasAlpha then IHDR.ColorType := 6;
        end;

    SwapLong(@IHDR, 2);
    Stream.Write(ChunkHdr, SizeOf(ChunkHdr));
    Stream.Write(IHDR, SizeOf(IHDR));
    CRC := $FFFFFFFF;
    CalcCRC32(Pointer(Long(@ChunkHdr) + 4), 4, CRC);
    CalcCRC32(@IHDR, SizeOf(IHDR), CRC);
    CRC := SwapLong(CRC xor $FFFFFFFF);
    Stream.Write(CRC, 4);

    if IHDR.ColorType = 3 then SavePLTE;
    SaveIDAT;

    ChunkHdr.DataSize := 0;
    ChunkHdr.ChunkID := SwapLong(IENDChunk);
    CRC := $FFFFFFFF;
    CalcCRC32(Pointer(Long(@ChunkHdr) + 4), 4, CRC);
    CRC := SwapLong(CRC xor $FFFFFFFF);
    Stream.Write(ChunkHdr, SizeOf(ChunkHdr));
    Stream.Write(CRC, 4);
  finally
    FreeImage(Clone);
  end;
end;

function TNPNGFormat.TestFormat(Stream: TNStream): Boolean;
var
  OldPos: Int64;
  C: array[0..7] of Char;
begin
  Result := False;
  if Stream = nil then Exit;
  OldPos := Stream.Position;
  Stream.Read(C, SizeOf(C));
  Stream.Position := OldPos;
  if CompareMem(@C, @CheckSig, SizeOf(C)) then Result := True;
end;


{*************************************************************************}
{                         Internal Stuff                                  }
{*************************************************************************}


initialization

  PNGInstance := TNPNGFormat.Create;

finalization

  FreeAndNil(PNGInstance);

end.

