{*************************************************************************}
{*   TgaFormat                                                           *}
{*   Copyright (C) 2003 Marek Mauder, Vampyre SoftWorks                  *}
{*   Graphic\ImageFileFormats\Tga\TgaFormat.pas                          *}
{*************************************************************************}
{                                                                         }
{ This program is free software; you can redistribute it and/or modify it }
{ under the terms of the GNU General Public License as published by the   }
{ Free Software Foundation; either version 2 of the License, or (at your  }
{ option) any later version. This program is distributed in the hope that }
{ it will be useful, but WITHOUT ANY WARRANTY; without even the implied   }
{ warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See   }
{ the GNU General Public License for more details.                        }
{                                                                         }
{ You should have received a copy of the GNU General Public License       }
{ along with this program; if not, write to the Free Software Foundation, }
{ Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.         }
{                                                                         }
{*************************************************************************}
{  Created On: 12st September 2003  (12.09.2003)                          }
{*************************************************************************}

{This unit contains image file format loader/saver for Tga files.}

unit TgaFormat;

{$I Compiler.inc}

interface

uses
  SysUtils, Kernel, Objects, Platforms, Graphic, Images;

type

  {Image encoder for loading/saving Targa file format.
  Supports: 1,8i/bw,16,24,32 bit bitmaps loading/saving.
  Compression/Decompression: None, RLE.}
  TNTGAFormat = class(TNImageFileFormat)
  protected
    procedure PrepareImage(const Image: TImageData; var Clone: TImageData);
      override;
    procedure DoLoadImage(var Image: TImageData; Stream: TNStream); override;
    procedure DoSaveImage(var Image: TImageData; Stream: TNStream;
      SaveCompressed: Boolean); override;
  public
    constructor Create; override;
    function TestFormat(Stream: TNStream): Boolean; override;
  end;

const

  sTGAExtensions = 'tga';
  sTGAFormatName = 'Truevision Targa';

var

  TGAInstance: TNTGAFormat = nil;


{*************************************************************************}
{                             IMPLEMEMTATION                              }
{*************************************************************************}


implementation

const

  sTGASignature = 'TRUEVISION-XFILE';

type

  {Targa file header}
  TTGAHeader = packed record
    IDLength: Byte;
    ColorMapType: Byte;
    ImageType: Byte;
    ColorMapOff: Word;
    ColorMapLength: Word;
    ColorEntrySize: Byte;
    XOrg: Small;
    YOrg: Small;
    Width: Small;
    Height: Small;
    PixelSize: Byte;
    Desc: Byte;
  end;

  {Footer at the end of TGA file}
  TTGAFooter = packed record
    ExtOff: ULong; // Extension Area Offset
    DevDirOff: ULong; // Developer Directory Offset
    Signature: array[0..15] of Char; // TRUEVISION-XFILE
    Reserved: Byte; // ASCII period '.'
    NullChar: Byte; // 0
  end;


{*************************************************************************}
{                        TNTGAFormat methods                              }
{*************************************************************************}


constructor TNTGAFormat.Create;
begin
  inherited Create;
  FExtensions := sTGAExtensions;
  FFormatName := sTGAFormatName;
  FCanLoad := True;
  FCanSave := True;
end;

procedure TNTGAFormat.PrepareImage(const Image: TImageData;
  var Clone: TImageData);
begin
  inherited PrepareImage(Image, Clone);
  with Clone.DataFormat^ do
  begin
    if OneChannelFormat and (BitCount > 8) then
      ConvertFunc(Clone, FindImageDataFormat(idfIntensity8))
    else
      if FloatFormat then
        ConvertFunc(Clone, FindImageDataFormat(idfA8R8G8B8))
      else
        if Indexed and (BitCount > 8) then
          ConvertFunc(Clone, FindImageDataFormat(idfIndex8))
        else
          if not FixedPixelFormat then
          begin
            if HasAlpha then
              ConvertFunc(Clone, FindImageDataFormat(idfA8R8G8B8))
            else
              ConvertFunc(Clone, FindImageDataFormat(idfR5G5B5))
          end
          else
            if (OneChannelSize > 8) then
            begin
              if HasAlpha then
                ConvertFunc(Clone, FindImageDataFormat(idfA8R8G8B8))
              else
                ConvertFunc(Clone, FindImageDataFormat(idfR8G8B8))
            end;
    // final check
    if (Clone.DataFormat <> FindImageDataFormat(idfA8R8G8B8)) and
      (Clone.DataFormat <> FindImageDataFormat(idfR8G8B8)) and
      (Clone.DataFormat <> FindImageDataFormat(idfR5G5B5)) and
      (Clone.DataFormat <> FindImageDataFormat(idfIndex8)) and
      (Clone.DataFormat <> FindImageDataFormat(idfIntensity8)) then
      ConvertFunc(Clone, FindImageDataFormat(idfR8G8B8));
  end;
end;

procedure TNTGAFormat.DoLoadImage(var Image: TImageData; Stream: TNStream);
var
  StartPos: Int64;
  Hdr: TTGAHeader;
  Foo: TTGAFooter;
  I, PSize, FSize, FBitCount, FPalSize: Long;
  Pal: Pointer;
  W: Word;

  procedure LoadRGB;
  begin
    Stream.ReadBuffer(Image.Levels[0].Bits^, Image.Levels[0].Size);
  end;

  procedure LoadRLE;
  var
    I, CPixel, Cnt: Long;
    B, Rle: Byte;
    Pix: PULong;
    Buffer: Pointer;
    BufPos, BufSize, OStreamPos: Long;

    procedure CheckBuffer(Size: Long);
    var
      OBufSize: Long;
    begin
      if BufPos + Size > BufSize then
      begin
        OBufSize := BufSize;
        BufSize := BufSize + Image.Levels[0].Width * Image.Levels[0].Height *
          FBitCount shr 6;
        ReallocMem(Buffer, BufSize);
        BufSize := OBufSize + Stream.Read(Pointer(Long(Buffer) + OBufSize)^,
          BufSize - OBufSize);
      end;
    end;

  begin
    with Stream, Image, Image.Levels[0] do
    begin
      BufSize := Width * Height * FBitCount shr 4;
      BufPos := 0;
      GetMem(Buffer, BufSize);
      OStreamPos := Stream.Position;
      BufSize := Stream.Read(Buffer^, BufSize);
      Cnt := Width * Height;
      B := FBitCount shr 3;
      GetMem(Pix, B);
      CPixel := 1;
      while CPixel <= Cnt do
      begin
        Rle := 0;
        CheckBuffer(1);
        Move(Pointer(Long(Buffer) + BufPos)^, Rle, 1);
        Inc(BufPos, 1);
        if Rle < 128 then
        begin
          // uncompressed pixel
          Rle := Rle + 1;
          CheckBuffer(Rle * B);
          for I := 0 to Rle - 1 do
          begin
            // read pixel
            Move(Pointer(Long(Buffer) + BufPos)^, Pix^, B);
            Inc(BufPos, B);
            // write pixel
            // !! these lines are much faster than using Move procedure
            //!! Move(Pix^,Pointer(Long(Bits) + Pad+(CPixel-1) * B)^,B);
            PByte(Long(Bits) + (CPixel - 1) * B)^ := PByte(Long(Pix) + 0)^;
            if B > 1 then
              PByte(Long(Bits) + (CPixel - 1) * B + 1)^ := PByte(Long(Pix)
                + 1)^;
            if B > 2 then
              PByte(Long(Bits) + (CPixel - 1) * B + 2)^ := PByte(Long(Pix)
                + 2)^;
            if B > 3 then
              PByte(Long(Bits) + (CPixel - 1) * B + 3)^ := PByte(Long(Pix)
                + 3)^;
            CPixel := CPixel + 1;
          end;
        end
        else
        begin
          // compressed pixels
          Rle := Rle - 127;
          // read pixel
          CheckBuffer(B);
          Move(Pointer(Long(Buffer) + BufPos)^, Pix^, B);
          Inc(BufPos, B);
          for I := 0 to Rle - 1 do
          begin
            // write pixel
            PByte(Long(Bits) + (CPixel - 1) * B)^ := PByte(Long(Pix) + 0)^;
            if B > 1 then
              PByte(Long(Bits) + (CPixel - 1) * B + 1)^ := PByte(Long(Pix) +
                1)^;
            if B > 2 then
              PByte(Long(Bits) + (CPixel - 1) * B + 2)^ := PByte(Long(Pix) +
                2)^;
            if B > 3 then
              PByte(Long(Bits) + (CPixel - 1) * B + 3)^ := PByte(Long(Pix) +
                3)^;
            CPixel := CPixel + 1;
          end;
        end;
      end;
      FreeMemNil(Pix);
      Stream.Position := OStreamPos + BufPos;
      FreeMemNil(Buffer);
    end;
  end;

begin
  FSize := SizeOf(Hdr);
  StartPos := Stream.Position;
  AddImageLevel(Image);
  with Stream, Image, Image.Levels[0] do
  begin
    Read(Hdr, SizeOf(Hdr));
    Width := Hdr.Width;
    Height := Hdr.Height;
    FBitCount := Hdr.PixelSize;
    Size := Width * Height * (FBitCount shr 3);
    // skip image ID info
    Position := Position + Hdr.IDLength;
    Inc(FSize, Hdr.IDLength);
    if Hdr.ColorMapType = 1 then
    begin
      // read palette
      PSize := Hdr.ColorMapLength * (Hdr.ColorEntrySize shr 3);
      Inc(FSize, PSize);
      GetMem(Pal, PSize);
      Read(Pal^, PSize);
      // process palette
      FPalSize := Hdr.ColorMapLength;
      GetMem(Palette, FPalSize * SizeOf(TARGBRec));
      for I := 0 to FPalSize - 1 do
        case Hdr.ColorEntrySize of
          24: Palette[I] := ARGBRec($FF, PByte(Long(Pal) + I * 3 + 2)^,
              PByte(Long(Pal) + I * 3 + 1)^, PByte(Long(Pal) + I * 3)^);
            // I've never seen tga with these palettes
          15, 16, 32: ;
        end;
      FreeMemNil(Pal);
    end;

    // reading image data
    GetMem(Bits, Size);
    case Hdr.ImageType of
      0, 1, 2, 3:
        //load uncompressed mode images
        LoadRGB;
      9, 10, 11:
        //load rle compressed mode images
        LoadRLE;
    end;
    Inc(FSize, Position - StartPos - FSize);

    case FBitCount of
      8: DataFormat := FindImageDataFormat(idfIndex8);
      16: DataFormat := FindImageDataFormat(idfR5G5B5);
      24: DataFormat := FindImageDataFormat(idfR8G8B8);
      32: DataFormat := FindImageDataFormat(idfA8R8G8B8);
    end;

    // set format for Black-White targas
    if Hdr.ImageType in [3, 11] then
      case FBitCount of
        8: DataFormat := FindImageDataFormat(idfIntensity8);
        16: DataFormat := FindImageDataFormat(idfIntensity16);
      end;

    // we must find true end of file and set Input.Position to it
    // paint programs appends extra info at the end of TGAs
    if Position + 2 < Stream.Size then
    begin
      Read(W, 2);
      // 495 = size of Extension Area
      if W = 495 then
      begin
        Inc(FSize, 495);
        Position := StartPos + FSize;
      end
      else
        Position := Position - 2;

      if Position + SizeOf(Foo) <= Stream.Size then
      begin
        Read(Foo, SizeOf(Foo));
        if Foo.Signature = sTGASignature then
        begin
          Inc(FSize, SizeOf(Foo));
          Position := StartPos + FSize;
        end
        else
          Position := Position - SizeOf(Foo);
      end;
    end;

    // some editors save tgas flipped
    if Hdr.Desc < 31 then
      FlipYImage(Image);

    Position := StartPos + FSize;
  end;
end;

procedure TNTGAFormat.DoSaveImage(var Image: TImageData; Stream: TNStream;
  SaveCompressed: Boolean);
var
  Hdr: TTGAHeader;
  I: Long;
  Pal: PPalRGBArray;
  Clone: TImageData;
  Level: PImageLevel;

  procedure SaveRGB;
  begin
    Stream.WriteBuffer(Level.Bits^, Level.Size);
  end;

  procedure SaveRLE;
  var
    APix, Pix: PULong;
    D, S, Cnt, CPixel: Long;
    B, Rle: Byte;
    Buffer: Pointer;
    BufPos, BufSize: Long;

    function CountSame: Long;
    var
      I: Long;
    begin
      I := CPixel;
      Result := 1;
      Move(Pointer(Long(Level.Bits) + I * B)^, Pix^, B);
      while I < Cnt do
      begin
        I := I + 1;
        Move(Pointer(Long(Level.Bits) + I * B)^, APix^, B);
        if (CompareMem(APix, Pix, B)) and
          (Result < 127) then
          Result := Result + 1
        else
          Break;
      end;
    end;

    function CountDiff: Long;
    var
      I: Long;
    begin
      I := CPixel;
      Result := 0;
      while I < Cnt do
      begin
        Move(Pointer(Long(Level.Bits) + I * B)^, Pix^, B);
        I := I + 1;
        Move(Pointer(Long(Level.Bits) + I * B)^, APix^, B);
        if (not CompareMem(APix, Pix, B)) and
          (Result < 127) then
          Result := Result + 1
        else
          Break;
      end;
    end;

    procedure CheckBuffer(Size: Long);
    begin
      if BufPos + Size > BufSize then
      begin
        BufSize := BufSize + Level.Width * Level.Height *
          Clone.DataFormat^.BitCount shr 6;
        ReallocMem(Buffer, BufSize);
      end;
    end;

  begin
    with Stream, Clone, Level^ do
    begin
      BufPos := 0;
      BufSize := Width * Height * DataFormat^.BitCount shr 3;
      GetMem(Buffer, BufSize);
      B := DataFormat^.BitCount shr 3;
      Cnt := Width * Height;
      CPixel := 0;
      GetMem(APix, 4);
      GetMem(Pix, 4);
      CPixel := 0;
      while CPixel < Cnt do
      begin
        D := CountDiff;
        S := CountSame;
        if D > 0 then
        begin
          Rle := D - 1;
          CheckBuffer(1 + D * B);
          Move(Rle, Pointer(Long(Buffer) + BufPos)^, 1);
          Inc(BufPos, 1);
          Move(Pointer(Long(Bits) + CPixel * B)^, Pointer(Long(Buffer) +
            BufPos)^,
            D * B);
          Inc(BufPos, D * B);
          CPixel := CPixel + D;
        end;
        if S > 1 then
        begin
          Rle := S + 127;
          CheckBuffer(1 + B);
          Move(Rle, Pointer(Long(Buffer) + BufPos)^, 1);
          Inc(BufPos, 1);
          Move(Pointer(Long(Bits) + CPixel * B)^, Pointer(Long(Buffer) +
            BufPos)^,
            B);
          Inc(BufPos, B);
          CPixel := CPixel + S;
        end;
      end;
      FreeMemNil(APix);
      FreeMemNil(Pix);
      Stream.WriteBuffer(Buffer^, BufPos);
      FreeMemNil(Buffer);
    end;
  end;

begin
  PrepareImage(Image, Clone);
  Level := FindFirstValidLevel(Clone);
  if Level = nil then Level := @Clone.Levels[0];
  with Stream, Clone, Level^ do
  try
    ZeroMemory(@Hdr, SizeOf(Hdr));
    Hdr.IDLength := 0;
    Hdr.ColorMapType := Iff(DataFormat^.PalSize > 0, 1, 0);
    Hdr.Width := Width;
    Hdr.Height := Height;
    Hdr.PixelSize := DataFormat^.BitCount;
    Hdr.ColorMapLength := DataFormat^.PalSize;
    Hdr.ColorEntrySize := Iff(DataFormat^.PalSize > 0, 24, 0);
    Hdr.ColorMapOff := 0;
    // this indicates that targa is stored in top-left format
    // as our images -> no flipping is needed.
    Hdr.Desc := 32;

    case DataFormat^.BitCount of
      8:
        if DataFormat^.Indexed then
          Hdr.ImageType := Iff(SaveCompressed, 9, 1)
        else
          Hdr.ImageType := Iff(SaveCompressed, 11, 3);
      else
        Hdr.ImageType := Iff(SaveCompressed, 10, 2);
    end;

    Write(Hdr, SizeOf(Hdr));

    // writting palette
    if DataFormat^.PalSize > 0 then
    begin
      GetMem(Pal, DataFormat^.PalSize * SizeOf(TRGBRec));
      for I := 0 to DataFormat^.PalSize - 1 do
        Pal[I] := RGBRec(Palette[I].R, Palette[I].G, Palette[I].B);
      Write(Pal^, DataFormat^.PalSize * SizeOf(TRGBRec));
      FreeMemNil(Pal);
    end;

    case Hdr.ImageType of
      1, 2, 3:
        //save uncompressed mode images
        SaveRGB;
      9, 10, 11:
        //save rle compressed mode images
        SaveRLE;
    end;
  finally
    FreeImage(Clone);
  end;
end;

function TNTGAFormat.TestFormat(Stream: TNStream): Boolean;
var
  Hdr: TTGAHeader;
  OldPos: Int64;
begin
  Result := False;
  if Stream = nil then Exit;
  if Stream.Size < SizeOf(Hdr) then Exit;
  OldPos := Stream.Position;
  Stream.ReadBuffer(Hdr, SizeOf(Hdr));
  Stream.Position := OldPos;
  if not (Hdr.ImageType in [0, 1, 2, 3, 9, 10, 11]) then Exit;
  if not (Hdr.PixelSize in [1, 4, 8, 15, 16, 24, 32]) then Exit;
  Result := True;
end;


{*************************************************************************}
{                         Internal Stuff                                  }
{*************************************************************************}


initialization

  TGAInstance := TNTGAFormat.Create;

finalization

  FreeAndNil(TGAInstance);

end.

