{*************************************************************************}
{*   Images                                                              *}
{*   Copyright (C) 2003 Marek Mauder, Vampyre SoftWorks                  *}
{*   Engine\Graphic\Images.pas                                           *}
{*************************************************************************}
{                                                                         }
{ This program is free software; you can redistribute it and/or modify it }
{ under the terms of the GNU General Public License as published by the   }
{ Free Software Foundation; either version 2 of the License, or (at your  }
{ option) any later version. This program is distributed in the hope that }
{ it will be useful, but WITHOUT ANY WARRANTY; without even the implied   }
{ warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See   }
{ the GNU General Public License for more details.                        }
{                                                                         }
{ You should have received a copy of the GNU General Public License       }
{ along with this program; if not, write to the Free Software Foundation, }
{ Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.         }
{                                                                         }
{*************************************************************************}
{  Created On: 11st September 2003  (11.09.2003)                          }
{*************************************************************************}

{This unit contains image classes, types and constants, support for
 loading/saving of different image file formats, image conversions, ...}

unit Images;

{TODO -oGalfar: Saving multilevel images with singlelevel modules.}
{TODO -oGalfar: Solve problem with extensions - dag textures, use filters?}

{$I Compiler.inc}

interface

uses
  SysUtils, Kernel, Objects, Platforms, Graphic;


{*************************************************************************}
{                          Image Types                                    }
{*************************************************************************}


type

  PImageLevelsArray = ^TImageLevelsArray;

  {(*}

  {One level of image data. Can contain additional levels -
   actually it is tree.
   Data in Bits are stored non-fliped, ordered ARGB ARGB (in multi channel
   images) ..., and no pad bytes are present.}
  TImageLevel = record
    Width: Long;                 // Width in pixels of level
    Height: Long;                // Height in pixels of level
    Size: Long;                  // Size of Bits in bytes
    Tag: Long;                   // User data
    Count: Long;                 // Number of sublevels
    Levels: PImageLevelsArray;   // Sublevels array
    Bits: Pointer;               // Actual image data of level
  end;
  PImageLevel = ^TImageLevel;

  {Array of TImageLevel records. It can contain max 76M of levels. It must
   be enough for every one}
  TImageLevelsArray = array[0..MaxByteArraySize div SizeOf(TImageLevel) - 1] of
    TImageLevel;

  PImageDataFormat = ^TImageDataFormat;
  PImageData = ^TImageData;

  {Low level image representation. Data are stored in the
   tree of TImageLevel records. Levels can be used
   as frames in animations, as mipmaps, ...}
  TImageData = record
    DataFormat: PImageDataFormat;   // Data format of image
    Tag: Long;                      // User data
    Count: Long;                    // Number of levels
    Levels: PImageLevelsArray;      // Sublevels array
    Palette: PPalArray;             // Image's palette
  end;

  {Function used by ImageDataFormat to convert image to another format.}
  TIDFConvertFunc = function(var Image: TImageData; DestFormat: PImageDataFormat):
    Boolean;
  {Function used by ImageDataFormat to convert 32bit ARGB color to its
  specific format. It is supported only for formats which have
  SupportsDrawing field set to True. Palette parameter is needed only for
  indexed mode images.}
  TIDFTranslateARGBFunc = function(Format:PImageDataFormat; ARGB:TARGB; Pal:PPalArray = nil):ULong;
  {Function used by ImageDataFormat to convert its specific format
  to 32bit ARGB color. It is supported only for formats which have
  SupportsDrawing field set to True. Palette parameter is needed only for
  indexed mode images.}
  TIDFTranslateColorFunc = function(Format:PImageDataFormat; Color:ULong; Pal:PPalArray = nil):TARGB;

  {This record describes how are data stored in image. This means
   bitcount, palette size, pixel format, info and encoding/decoding functions.
   Some default fotmats (A8R8G8B8,R8G8B8,Index8,Intensity8) are defined
   in StdImageDataFormats unit.
   Although some information (BitCount, Has Alpha, OneChannel, ...)can be derived
   from pixel format they are present so you should not mess with pixel format. }
  {Note: if TransARGBFunc and TransColorFunc are nil then no color
   conversions are done (they are used by drawing functions in Drawings.pas).
   However, they are nil either for some supported formats (A8R8G8B8)
   because their specific color type is identical with 32bit ARGB, so
   calling color conv functions would be only slowdown here.}
  TImageDataFormat = record
    Name: array[0..31] of Char;     // sSring id of format (in the form of
                                    //'idfXXXX' where XXX is AxRxGxBx for
                                    // ARGB formats, IndexX for Indexed and
                                    // IntensityX for intensity formats)
    BitCount: Small;                // Number of bytes per pixel
    PalSize: ULong;                 // Number of palette entries
    Indexed: Boolean;               // Is format indexed?
    HasAlpha: Boolean;              // Has format alpha channel?
    FixedPixelFormat: Boolean;      // Is channel size other than 8bits?
    FloatFormat: Boolean;           // Is image in floating point format?
    OneChannelFormat: Boolean;      // Has image only one channel?
    OneChannelSize: Byte;           // Size in bits of one channel. (define if
                                    // All (used) channels have same size)
    SupportsDrawing:Boolean;        // Drawing functions can draw to/from this
                                    // image without decoding
    PixelFormat: TPixelFormat;      // Pixel format of image
    ConvertFunc: TIDFConvertFunc;   // Function for converting images
    TransARGBFunc: TIDFTranslateARGBFunc;     //  ARGB color conv function
    TransColorFunc: TIDFTranslateColorFunc;   //  Color conv function
  end;

const

  {Constants for ImageDataFormats' names. You can use them
   in FindDataFormat() call to get apropiate format record.
   Declarations of formats are in StdImageDataFormats unit.
   They are automatically registered at startup.}
  idfA8R8G8B8 = 'idfA8R8G8B8';          // 32bit ARGB
  idfR8G8B8 = 'idfR8G8B8';              // 24bit RGB
  idfIndex8 = 'idfIndex8';              // 8bit indexed mode
  idfIntensity8 = 'idfIntensity8';      // 8bit one channel (greyscale)
  idfIntensity16 = 'idfIntensity16';    // 16bit one channel (greyscale)
  idfIntensity32 = 'idfIntensity32';    // 32bit one-channel (greyscale)
  idfIntensity32F = 'idfIntensity32F';  // 32bit floating-point one-channel
  idfA8Intensity8 = 'idfA8Intensity8';  // 16bit (8b Alpha, 8b Intensity)
  idfR5G5B5 = 'idfR5G5B5';              // 16bit RGB mode (5b per channel, no A)
  idfA1R5G5B5 = 'idfA1R5G5B5';          // 16bit ARGB mode (1bit alpha)
  idfA4R4G4B4 = 'idfA4R4G4B4';          // 16 bit ARGB mode (4b per channel)
  idfR3G3B2 = 'idfR3G3B2';              // 8bit RGB mode
  idfR1G1B1 = 'idfR1G1B1';              // 8/3bit RGB ugly mode
  idfA16R16G16B16 = 'idfA16R16G16B16';  // 64bit ARGB mode
  idfR16G16B16 = 'idfR16G16B16';        // 48bit RGB mode
  idfA32R32G32B32F = 'idfA32R32G32B32F';// 128bit floating-point ARGB
  {*)}

  {Identifier used to differ ImageFileFormat module library
   from another module libraries.}
  IffLibraryID = 'iff';


{*************************************************************************}
{                          Class Types                                    }
{*************************************************************************}


type

  {This class is abstract image file loader/saver. Create descendant
   for each wanted format (TGA, JPG, PNG, ...)}
  TNImageFileFormat = class(TNFileModule)
  protected
    {Returns pointer to first level in level tree which contain
     data (Size>0). It is helper for modules that save to single-image format
     - this will quickly find requested level for saving.}
    function FindFirstValidLevel(const Image: TImageData): PImageLevel;
    {Call this procedure in SaveXXX functions. It creates clone of
     Image. If encoder can not save image in Image's original
     format then it is converted (e.g. Converted to 8bpp for Daggerfall modules,
     FlipY for BMP, Swap-Channels for PNG, ... ).
     Clone must be freed after usage!!}
    procedure PrepareImage(const Image: TImageData; var Clone: TImageData);
      virtual;
    {These methods are called by LoadImage/SaveImage methods to
     do real loading/saving. Override these in descendants.}
    procedure DoLoadImage(var Image: TImageData; Stream: TNStream); virtual;
    procedure DoSaveImage(var Image: TImageData; Stream: TNStream;
      SaveCompressed: Boolean = True); virtual;
  public
    {Loads image from stream. If image is not in supported stream or
     error occurs False is returned. Note sure that Stream.Size
     might not be size of image! After loading set stream's position
     right after image data so next image can be read.}
    function LoadImage(var Image: TImageData; Stream: TNStream): Boolean;
      overload;
    {Saves image to stream. If error occurs False is returned. If SaveCompressed
    is True image is saved in compressed format (if format supports
    comp and uncomp mode (TGA, BMP), otherwise it is ignored (PNG, JPG)).}
    function SaveImage(var Image: TImageData; Stream: TNStream;
      SaveCompressed: Boolean = True): Boolean; overload;

    {Loads image from file.}
    function LoadImage(var Image: TImageData; const FileName: NString): Boolean;
      overload;
    {Saves image to file.}
    function SaveImage(var Image: TImageData; const FileName: NString;
      SaveCompressed: Boolean = True): Boolean; overload;
  end;

  {Class refernce for TNImageFileFormat.}
  TNImageFileFormatClass = class of TNImageFileFormat;

{$IFNDEF MODLIBRARY}
  {High-level wrapper to low-level TImageData structure.
   It is TNPersistent so it supports streaming (It actually
   calls Save/Load Image methods of selected TNImageFileFormat).
   Supports many conversion/drawing/effect functions.}
  {Note: all drawing functions in this class (including Set/Get Pixel)
   works only for Images with DataFormats that allow drawing.
   These are 8 - 32bpp no floating-point formats.}
  TNImage = class(TNPersistent)
  protected
    FActiveLevel: Long;
    FParentLevel: PImageLevel;
    function GetBitCount: Long;
    function GetPixelARGB(X, Y: Long): TARGB;
    function GetPixelDir(X, Y: Long): ULong;
    function GetActiveLevelPtr: PImageLevel;
    procedure SetBitCount(Value: Long);
    procedure SetPixelARGB(X, Y: Long; Value: TARGB);
    procedure SetPixelDir(X, Y: Long; Value: ULong);
    procedure SetActiveLevel(Value: Long);
    procedure SetParentLevel(Value: PImageLevel);
    {Initializes underlying TImageData struct.}
    procedure CreateImageData;
    {Finalizes underlying TImageData struct.}
    procedure FreeImageData;
    procedure AssignTo(Dest: TNPersistent); override;
  public
    {ImageData structure containing whole image.
     It is public so you can acccess it by image manipulation functions directly.
     But be carefull!}
    ImageData: TImageData;
    {Creates default 32x32x32 one level image.}
    constructor Create; override;
    {Creates TImage from existing TImageData struct.
     It copies all memory data from TImageData so you can
     safelly Free source ImageData. In destructor it frees all image data.}
    constructor CreateFromData(const Image: TImageData);
    {Frees all image data if TNImage was NOT created by CreateDirect().}
    destructor Destroy; override;
    {Clears all image data and creates new default image
    1 levels 32x32x32 resolution.}
    procedure Clear;
    {Bits per one pixel of image.}
    property BitCount: Long read GetBitCount write SetBitCount;
    {Use for setting/getting colors of pixels of ActiveLevel. This
     returns/sets color in ARGB format of pixel at X,Y.
     It returns/sets same value as PixelsDirect for 24bit and 32bit images.
     For 16bit images it uses pixelformats and for 8bit it uses palette/pf.}
    property PixelsARGB[X, Y: Long]: TARGB read GetPixelARGB write SetPixelARGB;
    {Use for direct access to pixels of ActiveLevel. This
     returns/sets raw value stored in image at position of X,Y.}
    property PixelsDirect[X, Y: Long]: ULong read GetPixelDir write SetPixelDir;
    {Indicates active working level. All drawing/modifying methods
     work with this level (However, there are some which
     have parameter AllLevels, they affect whole image).}
    property ActiveLevel: Long read FActiveLevel write SetActiveLevel;
    {Pointer to active image level. It can be changed by ActiveLevel and
     ParentLevel properties.ActiveLevelPtr:=ParentLevel.Levels[ActiveLevel].}
    property ActiveLevelPtr: PImageLevel read GetActiveLevelPtr;
    {Defines Parent level. ActiveLevel always points to
     array of levels which is owned by ParentLevel.
     If Parentlevel is nil (default state) ActiveLevel points to
     level in ImageData.Levels array.}
    property ParentLevel: PImageLevel read FParentLevel write SetParentLevel;
  end;
{$ENDIF}


{*************************************************************************}
{                       Image manipulation functions                      }
{*************************************************************************}


{* Alloc/Free/Clone/... images or levels}

{Frees current data and sets all image pointers to nil and all fileds to 0.}
procedure InitImage(var Image: TImageData);
{Creates new image with given attribs. If format is nil then default
 is chosen.}
procedure NewImage(var Image: TImageData; Width, Height: Long; Format:
  PImageDataFormat = nil);
{Frees all image pointers and sets all fileds to 0 or nil.
 Checks if image is valid before freeing pointers.}
procedure FreeImage(var Image: TImageData);
{This proc creates identical clone of image. All data are copied.}
procedure CloneImage(const Image: TImageData; var Clone: TImageData);

{Adds new level to image. if Owner is nil level is add to main image.}
procedure AddImageLevel(var Image: TImageData; Owner: PImageLevel = nil);
{Removes level from image.}
procedure RemoveImageLevel(var Image: TImageData; Level: Long; Owner:
  PImageLevel = nil);
{This proc creates identical clone of image level. All data are copied.
 if CloneSubLevels is True then all Level's levels are cloned too.}
procedure CloneImageLevel(var Level: TImageLevel; var Clone: TImageLevel;
  CloneSubLevels: Boolean = True);

{Tests if given Image structure is valid.}
function TestImage(const Image: TImageData): Boolean;
{Returns string information about ImageData fields.}
function ImageToStr(const Image: TImageData): NString;

{* Modify/Get/... image}

{Flips scanlines of image data (whole sublevel tree).}
procedure FlipYImage(var Image: TImageData);
{Mirrors image (whole sublevel tree).}
procedure FlipXImage(var Image: TImageData);
{Swaps Red and Blue channel of image data. Default value of Level indicates
 that whole sublevels tree will be flipped.}
procedure SwapChannels(var Image: TImageData);
{Performs pixel resize of whole image. No filtering is done.}
procedure ResizeImage(var Image: TImageData; NewWidth, NewHeight: Long);

{Flips scanlines of image level.}
procedure FlipYImageLevel(var Level: TImageLevel; Format: PImageDataFormat;
  FlipSubLevels: Boolean = True);
{Mirrors image level.}
procedure FlipXImageLevel(var Level: TImageLevel; Format: PImageDataFormat;
  FlipSubLevels: Boolean = True);
{Swaps Red and Blue channel of image level.}
procedure SwapChannelsLevel(var Level: TImageLevel; Format: PImageDataFormat;
  SwapSubLevels: Boolean = True);
{Performs pixel resize of image level. No filtering is done.}
procedure ResizeImageLevel(var Level: TImageLevel; Format: PImageDataFormat;
  NewWidth, NewHeight: Long; ResizeSubLevels: Boolean = True);
{Returns pointer to given scanline of image level.}
function GetScanline(const Level: TImageLevel; Y: Long; Format:
  PImageDataFormat): Pointer;



{*************************************************************************}
{                              Misc functions                             }
{*************************************************************************}


{* TImageDataFormat registration *}

{Registers new Image Data Format so it can be used by system.}
procedure RegisterImageDataFormat(Format: PImageDataFormat);
{Registers new Image Data Formats.}
procedure RegisterImageDataFormats(Formats: array of PImageDataFormat);
{UnRegisters Image Data Format so it NOT can be used by system.}
procedure UnRegisterImageDataFormat(Format: PImageDataFormat);
{UnRegisters Image Data Formats.}
procedure UnRegisterImageDataFormats(Formats: array of PImageDataFormat);
{Returns True if given Format is registered.}
function IsImageDataFormatRegistered(Format: PImageDataFormat): Boolean;
{Returns number of registered Image Data Formats.}
function RegisteredImageDataFormatsCount:Long;
{Returns Image Data Format of given name if registered, otherwise
returns nil.}
function FindImageDataFormat(const Name: NString): PImageDataFormat;overload;
{Returns Image Data Format at given Index, if index is wrong returns nil.}
function FindImageDataFormat(Index:Long): PImageDataFormat;overload;
{Returns current default format.}
function GetDefaultImageDataFormat: PImageDataFormat;


{$IFNDEF MODLIBRARY}

{* TNImageFileFormat registration *}

{Returns pointer to instance of module list storing
 registered TNImageFileFormat instances}
function RegisteredImageFileFormats: TNFileModuleList;
{Registers new Image File Format instance so they can be used by system.}
procedure RegisterImageFileFormat(Format: TNImageFileFormat);
{Registers new Image File Format instances.}
procedure RegisterImageFileFormats(Formats: array of TNImageFileFormat);
{UnRegisters Image File Format instance.}
procedure UnRegisterImageFileFormat(Format: TNImageFileFormat);
{UnRegisters new Image File Format instance.}
procedure UnRegisterImageFileFormats(Formats: array of TNImageFileFormat);
{Returns Image file format registered with given extension. If not found
 it returns nil.}
function FindImageFileFormat(const Ext: NString): TNImageFileFormat;

{$ENDIF}


{*************************************************************************}
{                             IMPLEMEMTATION                              }
{*************************************************************************}


implementation

uses
  {Default image data formats - registered at startup}
  StdImageDataFormats
{$IFNDEF MODLIBRARY}
  {Default image file formats - registered at startup}
  , TgaFormat, VgiFormat, JpgFormat, PngFormat, BmpFormat,
  {Image drawing library - used by TNImage}
  Drawings
{$ENDIF};

var
  {List for holding reistered image file formats. Accessed trough
   RegisteredImageFileFormats function.}
  RegFileFormats: TNFileModuleList = nil;


{*************************************************************************}
{                       TNImageDataFormats methods                        }
{*************************************************************************}


type

  {This class manages registered PImageDataFormats.}
  TNImageDataFormats = class(TNBaseObject)
  protected
    FFormats: TNSortedStringPointers;
    FDefaultFormat: PImageDataFormat;
    function GetDefault: PImageDataFormat;
  public
    constructor Create; override;
    destructor Destroy; override;
    procedure Add(Format: PImageDataFormat);
    procedure Remove(Format: PImageDataFormat);
    function Find(Format: PImageDataFormat): Boolean; overload;
    function Find(const Name: NString): PImageDataFormat; overload;
    property DefaultFormat: PImageDataFormat read GetDefault write
      FDefaultFormat;
  end;

var
  {List for holding registred image data formats.}
  RegDataFormats: TNImageDataFormats = nil;


constructor TNImageDataFormats.Create;
begin
  inherited Create;
  FFormats := TNSortedStringPointers.Create;
end;

destructor TNImageDataFormats.Destroy;
begin
  FFormats.Free;
  inherited Destroy;
end;

function TNImageDataFormats.GetDefault: PImageDataFormat;
begin
  Result := @StdImageDataFormats.DefFormat;
end;

procedure TNImageDataFormats.Add(Format: PImageDataFormat);
begin
  if Format <> nil then FFormats.Add(Format^.Name, Format);
end;

procedure TNImageDataFormats.Remove(Format: PImageDataFormat);
var
  I: Long;
begin
  if Format <> nil then
  begin
    I := FFormats.IndexOfValue(Format);
    if I > -1 then
      FFormats.Delete(I);
  end;
end;

function TNImageDataFormats.Find(Format: PImageDataFormat): Boolean;
begin
  Result := FFormats.IndexOfValue(Format) > -1;
end;

function TNImageDataFormats.Find(const Name: NString): PImageDataFormat;
var
  I: Long;
begin
  if FFormats.Find(Name, I) then
    Result := PImageDataFormat(FFormats.Pointers[I])
  else
    Result := nil;
end;


{*************************************************************************}
{                      TNImageFileFormat methods                          }
{*************************************************************************}


function TNImageFileFormat.FindFirstValidLevel(const Image: TImageData):
  PImageLevel;
var
  I: Long;

  // There must be var, otherwise bad address is here
  procedure CheckLevels(var Level: TImageLevel; var Res: PImageLevel);
  var
    I: Long;
  begin
    if Level.Size > 0 then
    begin
      Res := @Level;
      Exit;
    end;
    for I := 0 to Level.Count - 1 do
      if Res = nil then
        CheckLevels(Level.Levels[I], Res)
      else
        Exit;
  end;

begin
  Result := nil;
  for I := 0 to Image.Count - 1 do
    if Result = nil then
      CheckLevels(Image.Levels[I], Result)
    else
      Exit;
end;

procedure TNImageFileFormat.PrepareImage(const Image: TImageData;
  var Clone: TImageData);
begin
  CloneImage(Image, Clone);
  if (not IsImageDataFormatRegistered(Clone.DataFormat))
    and (FindImageDataFormat(Clone.DataFormat^.Name) <> nil) then
    Clone.DataFormat := FindImageDataFormat(Clone.DataFormat^.Name);
end;

procedure TNImageFileFormat.DoLoadImage(var Image: TImageData; Stream:
  TNStream);
begin
end;

procedure TNImageFileFormat.DoSaveImage(var Image: TImageData; Stream: TNStream;
  SaveCompressed: Boolean);
begin
end;

function TNImageFileFormat.LoadImage(var Image: TImageData;
  Stream: TNStream): Boolean;
var
  StartPos: Int64;
begin
  Result := False;
  if FCanLoad then
  begin
    StartPos := Stream.Position;
    FreeImage(Image);
    if not TestFormat(Stream) then
    begin
      ShowError(ErrorStr(Self, 'LoadImage', 'Stream image data format ' +
        'not supported by this module.'));
      Exit;
    end;
    try
      DoLoadImage(Image, Stream);
      // this is here because if module is in external library it has
      // its own list of data formats - they are same but has different addresses
      if (not IsImageDataFormatRegistered(Image.DataFormat))
        and (FindImageDataFormat(Image.DataFormat^.Name) <> nil) then
        Image.DataFormat := FindImageDataFormat(Image.DataFormat^.Name);
      Result := True;
    except
      on E: Exception do
      begin
        ShowError(ErrorStr(Self, 'LoadImage',
          'Error while loading image data! Image can be corrupted.' + sLineBreak
          + E.Message), E.ClassName);
        FreeImage(Image);
        Result := False;
        Stream.Position := StartPos;
      end;
    end;

  end;
end;

function TNImageFileFormat.SaveImage(var Image: TImageData; Stream: TNStream;
  SaveCompressed: Boolean): Boolean;
var
  StartPos: Int64;
begin
  Result := False;
  if FCanSave then
  begin
    if Stream = nil then
    begin
      ShowError(ErrorStr(Self, 'SaveImage', 'Output stream is nil!'));
      Exit;
    end;
    StartPos := Stream.Position;
    if not TestImage(Image) then
    begin
      ShowError(ErrorStr(Self, 'SaveImage',
        'Image data information is invalid! Image parameter: ' + sLineBreak +
        ImageToStr(Image)));
      Exit;
    end;
    try
      DoSaveImage(Image, Stream, SaveCompressed);
      Result := True;
    except
      on E: Exception do
      begin
        ShowError(ErrorStr(Self, 'SaveImage',
          'Error while saving image data! Image can be corrupted.' + sLineBreak
          + sExceptMsg + E.Message), E.ClassName);
        Stream.Position := StartPos;
      end;
    end;
  end;
end;

function TNImageFileFormat.LoadImage(var Image: TImageData;
  const FileName: NString): Boolean;
var
  Stream: TNFileStream;
begin
  Result := False;
  try
    Stream := TNFileStream.Create(FileName, fmOpenRead);
  except
    on E: Exception do
    begin
      ShowError(ErrorStr(Self, 'LoadImage', E.Message), E.ClassName);
      Exit;
    end;
  end;
  try
    Result := LoadImage(Image, Stream);
  except
    on E: Exception do
      ShowError(ErrorStr(Self, 'LoadImage', 'Error when loading image file "'
        + FileName + '" !' + sLineBreak + sExceptMsg + E.Message), E.ClassName);
  end;
  Stream.Free;
end;

function TNImageFileFormat.SaveImage(var Image: TImageData;
  const FileName: NString; SaveCompressed: Boolean): Boolean;
var
  Stream: TNFileStream;
begin
  Result := False;
  try
    Stream := TNFileStream.Create(FileName, fmCreate);
  except
    on E: Exception do
    begin
      ShowError(ErrorStr(Self, 'SaveImage', E.Message), E.ClassName);
      Exit;
    end;
  end;
  try
    Result := SaveImage(Image, Stream, SaveCompressed);
  except
    on E: Exception do
      ShowError(ErrorStr(Self, 'SaveImage', 'Error when saving image file "'
        + FileName + '" !' + sLineBreak + sExceptMsg + E.Message), E.ClassName);
  end;
  Stream.Free;
end;


{*************************************************************************}
{                              TNImage methods                            }
{*************************************************************************}


{$IFNDEF MODLIBRARY}

constructor TNImage.Create;
begin
  inherited Create;
  if not TestImage(ImageData) then
    CreateImageData;
  FActiveLevel := 0;
  FParentLevel := nil;
end;

constructor TNImage.CreateFromData(const Image: TImageData);
begin
  CloneImage(Image, ImageData);
  Create;
end;

destructor TNImage.Destroy;
begin
  FreeImageData;
  inherited Destroy;
end;

function TNImage.GetBitCount: Long;
begin
  Result := ImageData.DataFormat^.BitCount;
end;

function TNImage.GetPixelARGB(X, Y: Long): TARGB;
var
  Level: PImageLevel;
begin
  Result := 0;
  try
    if FParentLevel = nil then
      Level := @ImageData.Levels[FActiveLevel]
    else
      Level := @FParentLevel^.Levels[FActiveLevel];
    Result := Drawings.GetPixelARGB(Level^, ImageData.DataFormat, X, Y,
      ImageData.Palette);
  except
    on E: Exception do
      ShowError(ErrorStr(Self, 'GetPixelARGB', 'Error when getting pixel color!'
        + sLineBreak + sExceptMsg + E.Message), E.ClassName);
  end;
end;

function TNImage.GetPixelDir(X, Y: Long): ULong;
begin
  Result := 0;
  try
    Result := Drawings.GetPixelColor(GetActiveLevelPtr^, ImageData.DataFormat,
      X, Y);
  except
    on E: Exception do
      ShowError(ErrorStr(Self, 'GetPixelDir', 'Error when getting pixel color!'
        + sLineBreak + sExceptMsg + E.Message), E.ClassName);
  end;
end;

function TNImage.GetActiveLevelPtr: PImageLevel;
begin
  if FParentLevel = nil then
    Result := @ImageData.Levels[FActiveLevel]
  else
    Result := @FParentLevel^.Levels[FActiveLevel];
end;

procedure TNImage.SetBitCount(Value: Long);
begin
  with ImageData.DataFormat^ do
  try
    case Value of
      8: ConvertFunc(ImageData, FindImageDataFormat(idfIndex8));
      16: ConvertFunc(ImageData, FindImageDataFormat(idfA1R5G5B5));
      24: ConvertFunc(ImageData, FindImageDataFormat(idfR8G8B8));
      32: ConvertFunc(ImageData, FindImageDataFormat(idfA8R8G8B8));
      48: ConvertFunc(ImageData, FindImageDataFormat(idfR16G16B16));
      64: ConvertFunc(ImageData, FindImageDataFormat(idfA16R16G16B16));
      128: ConvertFunc(ImageData, FindImageDataFormat(idfA32R32G32B32F));
    end;
  except
    on E: Exception do
      ShowError(ErrorStr(Self, 'SetBitCount', 'Error when changing data format!'
        + sLineBreak + sExceptMsg + E.Message), E.ClassName);
  end;
end;

procedure TNImage.SetPixelARGB(X, Y: Long; Value: TARGB);
begin
  try
    Drawings.SetPixelARGB(GetActiveLevelPtr^, ImageData.DataFormat, X, Y, Value,
      ImageData.Palette);
  except
    on E: Exception do
      ShowError(ErrorStr(Self, 'SetPixelARGB', 'Error when setting pixel color!'
        + sLineBreak + sExceptMsg + E.Message), E.ClassName);
  end;
end;

procedure TNImage.SetPixelDir(X, Y: Long; Value: ULong);
begin
  try
    Drawings.SetPixelColor(GetActiveLevelPtr^, ImageData.DataFormat, X, Y,
      Value);
  except
    on E: Exception do
      ShowError(ErrorStr(Self, 'SetPixelDir', 'Error when setting pixel color!'
        + sLineBreak + sExceptMsg + E.Message), E.ClassName);
  end;
end;

procedure TNImage.SetActiveLevel(Value: Long);
begin
  if FParentLevel = nil then
  begin
    if Value < ImageData.Count then FActiveLevel := Value;
  end
  else
    if Value < FParentLevel^.Count then FActiveLevel := Value;
end;

procedure TNImage.SetParentLevel(Value: PImageLevel);
var
  OldLevel: PImageLevel;
begin
  OldLevel := FParentLevel;
  if Value <> nil then
    if Value^.Count > 0 then
    try
      if FActiveLevel >= Value^.Count then
        FActiveLevel := Value^.Count - 1;
      FParentLevel := Value;
    except
      FParentLevel := OldLevel;
    end;
end;

procedure TNImage.CreateImageData;
begin
  try
    FreeImageData;
    // create default starting image
    with ImageData do
    begin
      DataFormat := GetDefaultImageDataFormat;
      AddImageLevel(ImageData);
      Levels[0].Width := 32;
      Levels[0].Height := 32;
      Levels[0].Size := 32 * 32 * DataFormat^.BitCount shr 3;
      GetMemFill(Levels[0].Bits, 32 * 32 * DataFormat^.BitCount shr 3, $FF);
    end;
  except
    on E: Exception do
      ShowError(ErrorStr(Self, 'CreateImageData',
        'Unknown error when creating default image! This should never happen!' +
        sLineBreak + sExceptMsg + E.Message), E.ClassName);
  end;
end;

procedure TNImage.FreeImageData;
begin
  try
    FreeImage(ImageData);
  except
    on E: Exception do
      ShowError(ErrorStr(Self, 'FreeImageData',
        'Unknown error when freeing image data! This should never happen!' +
        sLineBreak + sExceptMsg + E.Message), E.ClassName);
  end;
end;

procedure TNImage.AssignTo(Dest: TNPersistent);
begin
  if Dest is TNImage then
  try
    FreeImage(TNImage(Dest).ImageData);
    CloneImage(Self.ImageData, TNImage(Dest).ImageData);
  except
    on E: Exception do
      ShowError(ErrorStr(Self, 'AssignTo',
        'Error when assigning to another image!' +
        sLineBreak + sExceptMsg + E.Message), E.ClassName);
  end
  else
    inherited AssignTo(Dest);
end;

procedure TNImage.Clear;
begin
  FreeImageData;
  // at least one image level must be active
  CreateImageData;
end;

{$ENDIF}


{*************************************************************************}
{                       Image manipulation functions                      }
{*************************************************************************}


procedure InitImage(var Image: TImageData);
begin
  ZeroMemory(@Image, SizeOf(Image));
end;

procedure NewImage(var Image: TImageData; Width, Height: Long; Format:
  PImageDataFormat);
begin
  if Format = nil then Format := GetDefaultImageDataFormat;
  if Format <> nil then
  begin
    FreeImage(Image);
    AddImageLevel(Image);
    Image.DataFormat := Format;
    if Format^.Indexed then
      Image.Palette := GreyScalePal(Format^.PalSize, $FF);
    Image.Levels[0].Width := Width;
    Image.Levels[0].Height := Height;
    Image.Levels[0].Size := Width * Height * (Format^.BitCount shr 3);
    GetMemFill(Image.Levels[0].Bits, Image.Levels[0].Size);
  end;
end;

procedure FreeImage(var Image: TImageData);
begin
  with Image do
    if TestImage(Image) then
    try
      if DataFormat^.PalSize > 0 then
        FreeMemNil(Palette);
      // important - Image.Count is changed in RemoveLevel() - can not use for
      while Count > 0 do
        RemoveImageLevel(Image, 0);
    except
      // pointers are not nil, but they are not valid
    end;
  ZeroMemory(@Image, SizeOf(Image));
end;

procedure CloneImage(const Image: TImageData; var Clone: TImageData);
var
  I: Long;
begin
  InitImage(Clone);
  if TestImage(Image) then
    with Image do
    begin
      Clone.DataFormat := DataFormat;
      Clone.Count := Count;
      Clone.Tag := Tag;
      if DataFormat^.PalSize > 0 then
      begin
        GetMem(Clone.Palette, DataFormat^.PalSize * SizeOf(TARGBRec));
        Move(Palette^, Clone.Palette^, DataFormat^.PalSize * SizeOf(TARGBRec));
      end;
      GetMem(Clone.Levels, Count * SizeOf(TImageLevel));
      for I := 0 to Count - 1 do
        CloneImageLevel(Levels[I], Clone.Levels[I]);
    end;
end;

procedure AddImageLevel(var Image: TImageData; Owner: PImageLevel);
begin
  if Owner = nil then
  begin
    Inc(Image.Count);
    ReallocMem(Image.Levels, Image.Count * SizeOf(TImageLevel));
    ZeroMemory(@Image.Levels[Image.Count - 1], SizeOf(TImageLevel));
  end
  else
  begin
    Inc(Owner^.Count);
    ReallocMem(Owner^.Levels, Owner^.Count * SizeOf(TImageLevel));
    ZeroMemory(@Owner^.Levels[Owner^.Count - 1], SizeOf(TImageLevel));
  end
end;

procedure RemoveImageLevel(var Image: TImageData; Level: Long; Owner:
  PImageLevel);
var
  I: Long;
begin
  if Owner = nil then
  begin
    if Level >= Image.Count then Exit;
    FreeMemNil(Image.Levels[Level].Bits);
    while Image.Levels[Level].Count > 0 do
      RemoveImageLevel(Image, 0, @Image.Levels[Level]);
    FreeMemNil(Image.Levels[Level].Levels);
    for I := Level to Image.Count - 2 do
      Image.Levels[I] := Image.Levels[I + 1];
    Dec(Image.Count);
    ReallocMem(Image.Levels, Image.Count * SizeOf(TImageLevel));
  end
  else
  begin
    if Level >= Owner^.Count then Exit;
    FreeMemNil(Owner^.Levels[Level].Bits);
    while Owner^.Levels[Level].Count > 0 do
      RemoveImageLevel(Image, 0, @Owner^.Levels[Level]);
    FreeMemNil(Owner^.Levels[Level].Levels);
    for I := Level to Owner^.Count - 2 do
      Owner^.Levels[I] := Owner^.Levels[I + 1];
    Dec(Owner^.Count);
    ReallocMem(Owner^.Levels, Owner^.Count * SizeOf(TImageLevel));
  end;
end;

procedure CloneImageLevel(var Level: TImageLevel; var Clone: TImageLevel;
  CloneSubLevels: Boolean = True);
var
  I: Long;
begin
  with Level do
  begin
    Clone.Width := Width;
    Clone.Height := Height;
    Clone.Size := Size;
    Clone.Tag := Tag;
    Clone.Count := Count;
    GetMem(Clone.Bits, Clone.Size);
    Move(Bits^, Clone.Bits^, Clone.Size);
    GetMem(Clone.Levels, Count * SizeOf(TImageLevel));
    if CloneSubLevels then
      for I := 0 to Count - 1 do
        CloneImageLevel(Levels[I], Clone.Levels[I]);
  end;
end;

function TestImage(const Image: TImageData): boolean;
begin
  Result := False;
  with Image do
  begin
    if ((Long(DataFormat) > 1) and (Long(DataFormat) < $7FFFF)) or
      ((Long(Palette) > 1) and (Long(Palette) < $7FFFF)) or
      ((Long(Levels) > 1) and (Long(Levels) < $7FFFF)) then Exit;
    if Image.Count > 0 then
      with Image.Levels[0] do
      begin
        if DataFormat = nil then Exit;
        if FindImageDataFormat(DataFormat.Name) = nil then Exit;
        if (DataFormat^.PalSize > 0) and (Palette = nil) then Exit;
        if (Width * Height * (DataFormat^.BitCount shr 3) <> Size) and
          (DataFormat^.BitCount >= 8) then Exit;
        Result := True;
      end;
  end;
end;

function ImageToStr(const Image: TImageData): NString;
var
  S: NString;
begin
  with Image do
  begin
    if Image.DataFormat <> nil then
      S := Format('Image Format: %s (BitCount: %d, PalSize: %d), Levels: %d, ' +
        'Palette: ', [Image.DataFormat^.Name, Image.DataFormat^.BitCount,
        Image.DataFormat^.PalSize, Count])
    else
      S := Format('Image Format: %s (BitCount: %d, PalSize: %d), Levels: %d, ' +
        'Palette: ', [sUnknown, 0, 0, Count]);
    if Palette = nil then
      S := S + sNil + ', '
    else
      S := S + Format('$%p', [Palette]);
    if Count > 0 then
    begin
      S := S + sLineBreak + Format('Level 0 (W: %d  H: %d  Size: %d  Bits: ',
        [Levels[0].Width, Levels[0].Height, Levels[0].Size]);
      if Levels[0].Bits = nil then
        S := S + sNil + ')  '
      else
        S := S + Format('$%p)', [Levels[0].Bits]);
    end
    else
      S := S + sLineBreak + 'Level 0: Not present';
  end;
  Result := S;
end;

procedure FlipYImage(var Image: TImageData);
var
  I: Long;
begin
  for I := 0 to Image.Count - 1 do
    FlipYImageLevel(Image.Levels[I], Image.DataFormat, True);
end;

procedure FlipXImage(var Image: TImageData);
var
  I: Long;
begin
  for I := 0 to Image.Count - 1 do
    FlipXImageLevel(Image.Levels[I], Image.DataFormat, True);
end;

procedure SwapChannels(var Image: TImageData);
var
  I: Long;
begin
  if Image.DataFormat^.OneChannelFormat then
    Exit
  else
    if Image.DataFormat^.Indexed then
      PalSwapChannels(Image.Palette, Image.DataFormat^.PalSize)
    else
      for I := 0 to Image.Count - 1 do
        SwapChannelsLevel(Image.Levels[I], Image.DataFormat, True);
end;

procedure ResizeImage(var Image: TImageData; NewWidth, NewHeight: Long);
var
  I: Long;
begin
  for I := 0 to Image.Count - 1 do
    ResizeImageLevel(Image.Levels[I], Image.DataFormat, NewWidth, NewHeight,
      True);
end;

procedure FlipYImageLevel(var Level: TImageLevel; Format: PImageDataFormat;
  FlipSubLevels: Boolean);
var
  P1, P2, Buff: Pointer;
  Sz, I: Long;
begin
  with Level do
  begin
    Sz := Width * (Format^.BitCount shr 3);
    GetMem(Buff, Sz);
    try
      for I := 0 to Height shr 1 - 1 do
      begin
        P1 := Pointer(Long(Bits) + I * Sz);
        P2 := Pointer(Long(Bits) + (Height - I - 1) * Sz);
        Move(P1^, Buff^, Sz);
        Move(P2^, P1^, Sz);
        Move(Buff^, P2^, Sz);
      end;
    finally
      FreeMemNil(Buff);
    end;
    if FlipSubLevels then
      for I := 0 to Count - 1 do
        FlipYImageLevel(Levels[I], Format, FlipSubLevels);
  end;
end;

procedure FlipXImageLevel(var Level: TImageLevel; Format: PImageDataFormat;
  FlipSubLevels: Boolean);
var
  P1, P2, Buff: Pointer;
  I, Bpp, Y, X, Width2, Sz: Long;
begin
  with Level do
  begin
    Bpp := Format^.BitCount shr 3;
    Width2 := Width shr 1;
    Sz := Width * (Format^.BitCount shr 3);
    GetMem(Buff, Bpp);
    try
      for Y := 0 to Height - 1 do
      begin
        P1 := Pointer(Long(Bits) + Y * Sz);
        P2 := Pointer(Long(P1) + (Width - 1) * Bpp);
        for X := 0 to Width2 - 1 do
        begin
          Move(P1^, Buff^, Bpp);
          Move(P2^, P1^, Bpp);
          Move(Buff^, P2^, Bpp);
          P1 := Pointer(Long(P1) + Bpp);
          P2 := Pointer(Long(P2) - Bpp);
        end;
      end;
    finally
      FreeMemNil(Buff);
    end;
    if FlipSubLevels then
      for I := 0 to Count - 1 do
        FlipYImageLevel(Levels[I], Format, FlipSubLevels);
  end;
end;

procedure SwapChannelsLevel(var Level: TImageLevel; Format: PImageDataFormat;
  SwapSubLevels: Boolean);
var
  I: Long;

  procedure SwapPFLevel;
  var
    X, Y: Long;
    C: TARGB;
    PB: PByte;
    PW: PWord;
    PU: PULong;
  begin
    with Level, Format^ do
    begin
      PB := Bits;
      for X := 0 to Width - 1 do
        for Y := 0 to Height - 1 do
        begin
          case BitCount of
            8:
              begin
                C := PFGetARGB(PixelFormat, PB^);
                C := (C and $FF00FF00) or (Byte(C) shl 16) or (C shr 16);
                PB^ := PFSetARGB(PixelFormat, C);
              end;
            16:
              begin
                PW := PWord(PB);
                C := PFGetARGB(PixelFormat, PW^);
                C := (C and $FF00FF00) or (Byte(C) shl 16) or (C shr 16);
                PW^ := PFSetARGB(PixelFormat, C);
              end;
            32:
              begin
                PU := PULong(PB);
                C := PFGetARGB(PixelFormat, PU^);
                C := (C and $FF00FF00) or (Byte(C) shl 16) or (C shr 16);
                PU^ := PFSetARGB(PixelFormat, C);
              end;
          end;
          PB := Pointer(Long(PB) + BitCount shr 3);
        end;
    end;
  end;

  procedure SwapRGBLevel;
  var
    X, Y: Long;
    TB: Byte;
    TW: Word;
    TU: ULong;
    PB: PByteArray;
    PW: PWordArray;
    PU: PLongArray;
  begin
    with Level, Format^ do
    begin
      PB := Bits;
      for X := 0 to Width - 1 do
        for Y := 0 to Height - 1 do
        begin
          case BitCount of
            24, 32:
              begin
                TB := PB[0];
                PB[0] := PB[2];
                PB[2] := TB;
              end;
            48, 64:
              begin
                PW := PWordArray(PB);
                TW := PW[0];
                PW[0] := PW[2];
                PW[2] := TW;
              end;
            96, 128:
              begin
                PU := PLongArray(PB);
                TU := PU[0];
                PU[0] := PU[2];
                PU[2] := TU;
              end;
          end;
          PB := Pointer(Long(PB) + BitCount shr 3);
        end;
    end;
  end;

begin
  with Level, Format^ do
  begin
    if not FixedPixelFormat then
      SwapPFLevel
    else
      SwapRGBLevel;
    if SwapSubLevels then
      for I := 0 to Count - 1 do
        SwapChannelsLevel(Levels[I], Format, SwapSubLevels);
  end;
end;

procedure ResizeImageLevel(var Level: TImageLevel; Format: PImageDataFormat;
  NewWidth, NewHeight: Long; ResizeSubLevels: Boolean);
var
  X, Y, Xp, Yp, Sx, Sy: Long;
  BytesPP: Byte;
  BytesPL: Long;
  I: Long;
  PARGBF: PARGBFRec;
  Data, Line: Pointer;
begin
  BytesPP := Format^.BitCount shr 3;
  BytesPL := Level.Width * BytesPP;
  GetMem(Data, NewWidth * NewHeight * BytesPP);
  Sx := (Level.Width shl 16) div NewWidth;
  Sy := (Level.Height shl 16) div NewHeight;
  Yp := 0;
  PARGBF := Data;
  for Y := 0 to NewHeight - 1 do
  begin
    Xp := 0;
    Line := Pointer(Long(Level.Bits) + (Yp shr 16) * BytesPL);
    for X := 0 to NewWidth - 1 do
    begin
      case BytesPP of
        1: PByte(PARGBF)^ := PByte(Long(Line) + (Xp shr 16) * BytesPP)^;
        2: PWord(PARGBF)^ := PWord(Long(Line) + (Xp shr 16) * BytesPP)^;
        3: PRGBRec(PARGBF)^ := PRGBRec(Long(Line) + (Xp shr 16) * BytesPP)^;
        4: PULong(PARGBF)^ := PULong(Long(Line) + (Xp shr 16) * BytesPP)^;
        6: PRGB48Rec(PARGBF)^ := PRGB48Rec(Long(Line) + (Xp shr 16) * BytesPP)^;
        8: PInt64(PARGBF)^ := PInt64(Long(Line) + (Xp shr 16) * BytesPP)^;
        16: PARGBF^ := PARGBFRec(Long(Line) + (Xp shr 16) * BytesPP)^;
      end;
      PARGBF := Pointer(Long(PARGBF) + BytesPP);
      Inc(Xp, Sx);
    end;
    Inc(Yp, Sy);
  end;
  Level.Width := NewWidth;
  Level.Height := NewHeight;
  Level.Size := NewWidth * NewHeight * BytesPP;
  FreeMem(Level.Bits);
  Level.Bits := Data;
  if ResizeSubLevels then
    for I := 0 to Level.Count - 1 do
      ResizeImageLevel(Level.Levels[I], Format, NewWidth, NewHeight,
        ResizeSubLevels);
end;

function GetScanline(const Level: TImageLevel; Y: Long; Format:
  PImageDataFormat): Pointer;
begin
  if (Y >= 0) and (Y < Level.Height) then
    Result := Pointer(Long(Level.Bits) + Y *
      Level.Width * (Format^.BitCount shr 3))
  else
    Result := nil;
end;


{*************************************************************************}
{                              Misc functions                             }
{*************************************************************************}


function GetRegDataFormats: TNImageDataFormats;
begin
  if RegDataFormats = nil then RegDataFormats := TNImageDataFormats.Create;
  Result := RegDataFormats;
end;

procedure RegisterImageDataFormat(Format: PImageDataFormat);
begin
  GetRegDataFormats.Add(Format);
end;

procedure RegisterImageDataFormats(Formats: array of PImageDataFormat);
var
  I: Long;
begin
  for I := Low(Formats) to High(Formats) do
    RegisterImageDataFormat(Formats[I]);
end;

procedure UnRegisterImageDataFormat(Format: PImageDataFormat);
begin
  GetRegDataFormats.Remove(Format);
end;

procedure UnRegisterImageDataFormats(Formats: array of PImageDataFormat);
var
  I: Long;
begin
  for I := Low(Formats) to High(Formats) do
    UnRegisterImageDataFormat(Formats[I]);
end;

function IsImageDataFormatRegistered(Format: PImageDataFormat): Boolean;
begin
  Result := GetRegDataFormats.Find(Format);
end;

function RegisteredImageDataFormatsCount:Long;
begin
  Result:=GetRegDataFormats.FFormats.Count;
end;

function FindImageDataFormat(const Name: NString): PImageDataFormat;
begin
  Result := GetRegDataFormats.Find(Name);
end;

function FindImageDataFormat(Index:Long): PImageDataFormat;
begin
  Result := PImageDataFormat(GetRegDataFormats.FFormats.Pointers[Index]);
end;

function GetDefaultImageDataFormat: PImageDataFormat;
begin
  Result := GetRegDataFormats.DefaultFormat;
end;

{$IFNDEF MODLIBRARY}

function RegisteredImageFileFormats: TNFileModuleList;
begin
  if RegFileFormats = nil then RegFileFormats := TNFileModuleList.Create;
  Result := RegFileFormats;
end;

procedure RegisterImageFileFormat(Format: TNImageFileFormat);
begin
  RegisteredImageFileFormats.Add(Format);
end;

procedure RegisterImageFileFormats(Formats: array of TNImageFileFormat);
var
  I: Long;
begin
  for I := Low(Formats) to High(Formats) do
    RegisterImageFileFormat(Formats[I]);
end;

procedure UnRegisterImageFileFormat(Format: TNImageFileFormat);
begin
  RegisteredImageFileFormats.Remove(Format);
end;

procedure UnRegisterImageFileFormats(Formats: array of TNImageFileFormat);
var
  I: Long;
begin
  for I := Low(Formats) to High(Formats) do
    UnRegisterImageFileFormat(Formats[I]);
end;

function FindImageFileFormat(const Ext: NString): TNImageFileFormat;
begin
  Result := TNImageFileFormat(RegisteredImageFileFormats.Find(Ext));
end;

{$ENDIF}


{*************************************************************************}
{                         Internall Stuff                                 }
{*************************************************************************}


{$IFNDEF MODLIBRARY}

procedure LoadedIff(Entry: PModuleEntry; var Done: Boolean);
var
  GetModule: TGetModuleProc;
  I, J: Long;
  Cls: TNImageFileFormatClass;
  Inst: TNImageFileFormat;
  Check: Boolean;
begin
  if Entry^.ID = IffLibraryID then
  try
    @GetModule := GetProcAddress(Entry^.Handle, sGetModuleProc);
    for I := 0 to Entry^.Count - 1 do
    begin
      GetModule(I, Pointer(Cls));
      Check := True;
      for J := 0 to RegisteredImageFileFormats.Modules.Count - 1 do
        if TNImageFileFormat(
          RegisteredImageFileFormats.Modules[J]).ClassNameIs(Cls.ClassName) then
        begin
          Check := False;
          Break;
        end;
      if Check then
      begin
        Inst := Cls.Create;
        if Inst <> nil then
          RegisterImageFileFormat(Inst);
      end;
    end;
  except
    raise Exception.Create('Error when creating ImageFileFormat from ' +
      'module library!');
  end;
end;

procedure UnLoadedIff(Entry: PModuleEntry; var Done: Boolean);
var
  GetModule: TGetModuleProc;
  I, J: Long;
  Cls: TNImageFileFormatClass;
  Inst: TNImageFileFormat;
begin
  if Entry^.ID = IffLibraryID then
  try
    @GetModule := GetProcAddress(Entry^.Handle, sGetModuleProc);
    for I := 0 to Entry^.Count - 1 do
    begin
      GetModule(I, Pointer(Cls));
      for J := 0 to RegisteredImageFileFormats.Modules.Count - 1 do
        if TNImageFileFormat(
          RegisteredImageFileFormats.Modules[J]).ClassNameIs(Cls.ClassName) then
        begin
          Inst := TNImageFileFormat(RegisteredImageFileFormats.Modules[J]);
          UnRegisterImageFileFormat(Inst);
          Inst.Free;
          Break;
        end;
    end;
  except
    raise Exception.Create('Error when unloading ImageFileFormat from ' +
      'module library!');
  end;
end;

{$ENDIF}


initialization

{$IFNDEF MODLIBRARY}
  {Adds load/unload callbacks for ModuleRegister.}
  ModuleRegister.AddLoadCallBack(LoadedIff);
  ModuleRegister.AddUnLoadCallBack(UnLoadedIff);

  {Registration of default image formats is put here instead of
   their original units to be sure that they will be registered in
   this sequence (first will be chosen as default format for saving).}
  RegisterImageFileFormat(TgaFormat.TGAInstance);
  RegisterImageFileFormat(VgiFormat.VGIInstance);
  RegisterImageFileFormat(JpgFormat.JPGInstance);
  RegisterImageFileFormat(PngFormat.PNGInstance);
  RegisterImageFileFormat(BmpFormat.BMPInstance);
{$ENDIF}

finalization

  FreeAndNil(RegDataFormats);
  FreeAndNil(RegFileFormats);

end.

