{*************************************************************************}
{*   Kernel                                                              *}
{*   Copyright (C) 2003 Marek Mauder, Vampyre SoftWorks                  *}
{*   Engine\Kernel\Kernel.pas                                            *}
{*************************************************************************}
{                                                                         }
{ This program is free software; you can redistribute it and/or modify it }
{ under the terms of the GNU General Public License as published by the   }
{ Free Software Foundation; either version 2 of the License, or (at your  }
{ option) any later version. This program is distributed in the hope that }
{ it will be useful, but WITHOUT ANY WARRANTY; without even the implied   }
{ warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See   }
{ the GNU General Public License for more details.                        }
{                                                                         }
{ You should have received a copy of the GNU General Public License       }
{ along with this program; if not, write to the Free Software Foundation, }
{ Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.         }
{                                                                         }
{*************************************************************************}
{  Created On: 12th August 2003  (12.08.2003)                             }
{*************************************************************************}

{This is base system unit of entire engine. It contains basic
 types, constants and functions.}

unit Kernel;

{ TODOS:
  -Add ExpandFileTo from Vide

  - Look at PasDoc, Doc-O-Matic and try to be compatible.
  - Better exceptions and errors.
  - Look at Integer const idents (TCursor,TColor).
  - Create NEW DEVELOPER TOOLS (with Create Source Size report in HTML,
      store these reports).
  - Test ASM and NOASM versions.
  - Test under Linux if PIC code can be in std executables.
  - Add assertations, Assert(ErrorStr())
}

{$I Compiler.inc}

interface

uses
{$IFDEF USEMEMMGR}
  {MemMgrs must be first initialized unit.}
  MemMgrs,
{$ENDIF}
  SysUtils;


{*************************************************************************}
{                           Basic Types                                   }
{*************************************************************************}


type

  {Some integer types.}
  Long = LongInt;
  ULong = LongWord;
  Small = SmallInt;
  Integer = LongInt;
  Short = ShortInt;

  {Some pointer types.}
  PLong = ^Long;
  PShort = ^Short;
  PSmall = ^Small;
  PULong = ^ULong;

{$IFDEF FLOATSINGLE}
  Float = Single;
{$ELSE}
  Float = Double;
{$ENDIF}
  PFloat = ^Float;

  {Some other types.}
  THandle = ULong;
  PHandle = ^THandle;

  {String types.
   When writing string manipulation functions use NString as paramter type
   only when function will work with all possible string types (Ansi, Wide,
   Short). In other cases (e.g. assesmbler functions) use paramter type
   for which function works.}
  NString = AnsiString;
  PNString = ^AnsiString;

  {Character types.}
  Char = AnsiChar;
  PChar = PAnsiChar;

const

  {Maximum values of int types.}
  MaxShort = High(Short);
  MaxByte = High(Byte);
  MaxSmall = High(Small);
  MaxWord = High(Word);
  MaxLong = High(Long);
  MaxULong = High(ULong);
  MaxInt64 = High(Int64);

  {Minimum values of int types.}
  MinShort = Low(Short);
  MinByte = Low(Byte);
  MinSmall = Low(Small);
  MinWord = Low(Word);
  MinLong = Low(Long);
  MinULong = Low(ULong);
  MinInt64 = Low(Int64);

  {Maximum values of float types.}
  MaxSingle = 3.4E+38;
  MaxDouble = 1.7E+308;
  MaxExtended = 1.1E+4932;
{$IFDEF FLOATSINGLE}
  MaxFloat = MaxSingle;
{$ELSE}
  MaxFloat = MaxDouble;
{$ENDIF}

  {Minimum values of float types.}
  MinSingle = 1.5E-45;
  MinDouble = 5.0E-324;
  MinExtended = 3.4E-4932;
{$IFDEF FLOATSINGLE}
  MinFloat = MinSingle;
{$ELSE}
  MinFloat = MinDouble;
{$ENDIF}

  {Max size for some arrays.}
  MaxByteArraySize = MaxLong;
  MaxWordArraySize = MaxLong div SizeOf(Word);
  MaxLongArraySize = MaxLong div SizeOf(Long);
  MaxInt64ArraySize = MaxLong div SizeOf(Int64);
  MaxSingleArraySize = MaxLong div SizeOf(Single);
  MaxDoubleArraySize = MaxLong div SizeOf(Double);
  MaxFloatArraySize = MaxLong div SizeOf(Float);

type

  {some general arrays.}
  TByteArray = array[0..MaxByteArraySize - 1] of Byte;
  TWordArray = array[0..MaxWordArraySize - 1] of Word;
  TLongArray = array[0..MaxLongArraySize - 1] of Long;
  TULongArray = array[0..MaxLongArraySize - 1] of Ulong;
  TInt64Array = array[0..MaxInt64ArraySize - 1] of Int64;
  TPCharArray = array[0..MaxLongArraySize - 1] of PChar;
  TStringArray = array[0..MaxLongArraySize - 1] of NString;
  TPointerArray = array[0..MaxLongArraySize - 1] of Pointer;

  PByteArray = ^TByteArray;
  PWordArray = ^TWordArray;
  PLongArray = ^TLongArray;
  PULongArray = ^TULongArray;
  PInt64Array = ^TInt64Array;
  PPCharArray = ^TPCharArray;
  PStringArray = ^TStringArray;
  PPointerArray = ^TPointerArray;

  {(*}    // do not process by SourceFormatter

  {Variant record which can be used to get
   things like 'high part of', 'low byte of' and so.
   Note that all array types are unsigned.}
  TInt64Rec = packed record
    case Integer of
      0: (Int64Value: Int64);
      // Array variants - Index 0 means lowest significant byte (word, ...)
      1: (ULongs: array[0..1] of ULong);
      2: (Words: array[0..3] of Word);
      3: (Bytes: array[0..7] of Byte);
  end;

  {Variant record which can be used to get
   things like 'high part of', 'low byte of' and so.
   Note that all array types are unsigned.}
  TLongRec = packed record
    case Integer of
      0: (LongValue: Long);
      // Array variants - Index 0 means lowest significant byte (word, ...)
      1: (Words: array[0..1] of Word);
      2: (Bytes: array[0..3] of Byte);
  end;

  {Variant record which can be used to get
   things like 'high part of', 'low byte of' and so.
   Note that all array types are unsigned.}
  TWordRec = packed record
    case Integer of
      0: (WordValue: Long);
      // Array variants - Index 0 means lowest significant byte (word, ...)
      1: (Bytes: array[0..1] of Byte);
  end;

  {Record defining 2D integer point.}
  TPoint = record
    X: Long;          // X coordinate
    Y: Long;          // Y coordinate
  end;

  {Version info record.}
  TVersionInfo = packed record
    case Integer of
      0: (Major: Byte;
          Minor: Byte;
          Build: Word; );
      1: (Version: ULong);
  end;


  {Message structure used for comunication between objects and other
   purposes.}
  TMsgRec = packed record
    MsgID: Word;             // Message ID constant. According to this ID
                             // procedures which performs this message
                             // access other memers of structure.
                             // E.g. Object receives SM_FreeSender so
                             // Sender can not be nil !
    Sender: TObject;         // Identifies object that send this message
    case Long of
      0: (Data: Pointer);    // Custom pointer data
      1: (ULParam: ULong);   // Custom unsigned Long value
      2: (LParam: Long);     // Custom Long value
      3: (Words: array[0..1] of Word);  // Custom word values
      4: (Bytes: array[0..3] of Byte);  // Custom byte values
      5: (Chars: array[0..3] of Char);  // Custom chars
  end;

  {*)}


{*************************************************************************}
{                         Some constants                                  }
{*************************************************************************}


const

  {Name of compiler this unit was compiled in.}
  CompilerName =
{$IFDEF DELPHI7}
  'Delphi 7';
{$ENDIF}
{$IFDEF DELPHI6}
  'Delphi 6';
{$ENDIF}
{$IFDEF DELPHI5}
  'Delphi 5';
{$ENDIF}
{$IFDEF DELPHI5}
  'Delphi 4';
{$ENDIF}
{$IFDEF BCB}
  'C++ Builder';
{$ENDIF}
{$IFDEF KYLIX}
  'Kylix';
{$ENDIF}
{$IFDEF FPC}
  'Free Pascal ' + {$I %FPCVERSION%};
{$ENDIF}
{$IFDEF UNKCOMPILER}
  'Some Pascal Compiler';
{$ENDIF}

  {Actual engine version. Read-only.}
  EngineVersion: TVersionInfo = (
    Major: 00;
    Minor: 02;
    // days from 12.08.2003
    Build: 0264);

  {Tolerance when comparing floats}
{$IFDEF FLOATSINGLE}
  FloatTolerance: Float = 0.00001;
{$ELSE}
  FloatTolerance: Float = 0.0000000001;
{$ENDIF}

  {(*}    // do not process by SourceFormatter

  {Kernel messages (0 - 511)
   All messages defined in Kernel directory must begin with KM prefix.
   Put them all here so they are at one place - less mess}
  KM_First         = 0000;

  // .. add something really important at indexes 1..7

  KM_GetID         = 0008; // Receiverer should send its ID to
                           // sender in TMsgRec.LParam
  KM_FreeSelf      = 0009; // Receiver calls Self.Free;
  KM_FreeSender    = 0010; // Receiver calls Sender.Free;

   // ...

  KM_Last         = 0511;

  {*)}

const

  {Some useful string constants.}
  sTrue = 'True';
  sFalse = 'False';
  sNil = 'Nil';
  sNull = 'Null';
  sError = 'Error';
  sWarning = 'Warning';
  sUnknown = 'Unknown';
  sRelease = 'Release';
  sDebug = 'Debug';
  sName = 'Name';
  sType = 'Type';
  sClass = 'Class';

const

  {Some defalut directories.}
  {Prefered Bin (for exe/dll/bpl/so) files directory (ExeDir).}
  BinDir = 'Bin';
  {Main engine directory (ExtractFileDir(Bin))}
  RootDir = 'Root';
  {Prefered Log files directory (Bin\LogDir).}
  LogDir = 'Logs';
  {Default subdirectory where module libraries will be stored
   (Bin\ModuleDir).}
  ModuleDir = 'Modules';

type
  TByteCharArray = array[Byte] of Char;

var

  {Char lookup tables which speeds up string ops.}
  UpperChars, LowerChars: TByteCharArray;


{*************************************************************************}
{                         Misc functions                                  }
{*************************************************************************}


{Switches Boolean value.}
procedure Switch(var Value: Boolean);

{Swaps two values.}
procedure Swap(var A, B: Long); overload;
procedure Swap(var A, B: Float); overload;
procedure Swap(var A, B: NString); overload;
procedure Swap(var A, B: Pointer); overload;
procedure Swap(var A, B: Boolean); overload;

{Creates TPoint record from coordinates.}
function Point(X, Y: Long): TPoint; overload;
procedure Point(var P: TPoint; X, Y: Long); overload;

{This clamps Value to Byte bounds 0..255. Used mainly
in some drawing functions to clamp image color channels.}
function Set255(Value: Long): Long;

{Equivalents to C "? :" operator.}
function Iff(const Condition: Boolean; const TruePart, FalsePart: NString):
  NString; overload;
function Iff(const Condition: Boolean; const TruePart, FalsePart: Char):
  Char; overload;
function Iff(const Condition: Boolean; const TruePart, FalsePart: Byte):
  Byte; overload;
function Iff(const Condition: Boolean; const TruePart, FalsePart: Long):
  Long; overload;
function Iff(const Condition: Boolean; const TruePart, FalsePart: ULong):
  ULong; overload;
function Iff(const Condition: Boolean; const TruePart, FalsePart: Float):
  Float; overload;
function Iff(const Condition: Boolean; const TruePart, FalsePart: Boolean):
  Boolean; overload;
function Iff(const Condition: Boolean; const TruePart, FalsePart: Pointer):
  Pointer; overload;
function Iff(const Condition: Boolean; const TruePart, FalsePart: Int64):
  Int64; overload;

{Swaps high and low bytes of 32 bit values. Count means number of
 Longs in P.}
procedure SwapLong(P: PLong; Count: Long); overload;
{Swaps high and low bytes of 32 bit value and returns swapped
 value.}
function SwapLong(Value: Long): Long; overload;


{*************************************************************************}
{                      Coding/Encoding/CRC  functions                     }
{*************************************************************************}


{This function calculates CRC32 value for given string.
 It calculates case insensitive.}
function CalcCRC32NoCase(const Str: AnsiString): ULong; overload;
{This function calculates CRC32 value for given buffer.
 It calculates case insensitive.}
function CalcCRC32NoCase(Data: Pointer; Size: Long): ULong; overload;
{This procedure calculates CRC32 value for given buffer. Starting CRC32 value
 can be set here.It calculates case insensitive.}
procedure CalcCRC32NoCase(Data: Pointer; Size: Long; var CRC32: ULong);
overload;
{This function calculates CRC32 value for given string.
 It calculates case sensitive.}
function CalcCRC32(const Str: AnsiString): ULong; overload;
{This function calculates CRC32 value for given buffer.
 It calculates case sensitive.}
function CalcCRC32(Data: Pointer; Size: Long): ULong; overload;
{This procedure calculates CRC32 value for given buffer. Starting CRC32 value
 can be set here.It calculates case sensitive.}
procedure CalcCRC32(Data: Pointer; Size: Long; var CRC32: ULong); overload;

{This encodes Data using Base64 and returns encoded string.}
function Base64Encode(Data: Pointer; Size: Long): AnsiString; overload;
{This encodes S string using Base64 and returns encoded string.}
function Base64Encode(const S: AnsiString): AnsiString; overload;
{This decodes string S encoded by Base64 and returns Data and Size of Data.}
procedure Base64Decode(const S: AnsiString; var Data: Pointer; var Size: Long);
overload;
{This decodes string S encoded by Base64 and returns decoded string.}
function Base64Decode(const S: AnsiString): AnsiString; overload;


{*************************************************************************}
{                   Memory/Pointer Management                             }
{*************************************************************************}


{Frees object and sets reference to nil. Supports only TObject descendants!!}
procedure FreeAndNil(var Obj);
{Frees memory pointer and sets reference to nil. P must be
 valid pointer!!!}
procedure FreeMemNil(var P);
{Calls GetMem and then fills memory with Value byte.}
procedure GetMemFill(var P; Size: Long; Value: Byte = 0);

{These functions return non-zero values only when USEMEMMGR
symbol is defined.}
{Indicates how many times was GetMem (and procs that use
 it: New, ...) called}
function GetMemCount: Long;
{Indicates how many times was FreeMem (and procs that use
 it: Dispose, ...) called}
function FreeMemCount: Long;
{Indicates how many times was ReallocMem called}
function ReallocMemCount: Long;

{Fills given buffer with zeroes. Size is size of buffer in bytes.}
procedure ZeroMemory(Data: Pointer; Size: Long);
{Fills given buffer with given byte value. Size is size of buffer in bytes.}
procedure FillMemory(Data: Pointer; Size: Long; Value: Byte);
{Fills given buffer with given word value. Size is size of buffer in bytes.}
procedure FillMemoryWord(Data: Pointer; Size: Long; Value: Word);
{Fills given buffer with given ulong value. Size is size of buffer in bytes.}
procedure FillMemoryULong(Data: Pointer; Size: Long; Value: ULong);


{*************************************************************************}
{                             Base Math                                   }
{*************************************************************************}


{Returns -1 if Value is negative, otherwise returns 1 }
function Sign(const Value: Long): Short; overload;
function Sign(const Value: Float): Short; overload;

{Returns Min of two values}
function Min(const A, B: Long): Long; overload;
function Min(const A, B: Float): Float; overload;

{Returns Max of two values}
function Max(const A, B: Long): Long; overload;
function Max(const A, B: Float): Float; overload;

{Clamps value by Min and Max}
function Clamp(const Value: Float; const Min: Float = 0.0;
  const Max: Float = 1.0): Float; overload;
function Clamp(const Value: Long; const Min: Long = 0; const Max: Long = 1):
  Long; overload;

{Compares two floats, returns True if A and B are (almost) identical.
 Uses FloatTolerance constant in comparsion.}
function CompFloat(const A, B: Float): Boolean;
{Is float zero? Uses FloatTolerance constant in comparsion.}
function IsZero(const Value: Float): Boolean;

{Returns float   0 <= X < 1}
function Rand: Float; overload;
{Returns long 0 <= X < Max}
function Rand(const Max: Long): Long; overload;
function Rand(const Max: Float): Float; overload;

{Returns True if Num is power of 2.}
function IsPow2(Num: Long): Boolean;
{Returns next power of 2 (if Num itself is power of 2, then
retuns Num).}
function NextPow2(Num: Long): Long;

{Returns smallest integer >= Val (for 5.1 returns 6).}
function Ceil(Value: Float): Long;
{Returns largest integer <= Val (for 5.9 returns 5).}
function Floor(Value: Float): Long;


{*************************************************************************}
{                            Some strings                                 }
{*************************************************************************}


{Returns string representatin of TVersionInfo.}
function VersionToStr(const Version: TVersionInfo): NString;

{Compares two strings case-insensitive and returns <0 if
 S1<S2, >0 if S1>S2 or 0 if S1=S2. Contrary to
 other comp functions (except AnsiXXXX of course)
 it supports even characters like ,,,,... which I use
 in my Czech language.}
function CompStr(const S1, S2: AnsiString): Long;
{Compares two strings case-sensitive. Return is same as in Comp() }
function CompStrCase(const S1, S2: AnsiString): Long;
{Compares two strings case-insensitive. Supports chars like CompStr().}
function SameStr(const S1, S2: AnsiString): Boolean;
{Compares two strings case-sensitive. This function
 is even faster than 'S1 = S2'.}
function SameStrCase(const S1, S2: AnsiString): Boolean;

{Converts string to lower case. Supports chars like CompStr().}
function StrLowerCase(const S: AnsiString): AnsiString;
{Converts string to upper case. Supports chars like CompStr().}
function StrUpperCase(const S: AnsiString): AnsiString;

type
  TReplaceFlags = set of (srReplaceAll, srIgnoreCase);

{StrReplace replaces occurances of <oldpattern> with <newpattern> in a
 given string.}
function StrReplace(const S, OldPattern, NewPattern: AnsiString;
  Flags: TReplaceFlags = [srReplaceAll, srIgnoreCase]): AnsiString;

{Returns position of SubStr in S from position StartPos case sensitively.
 Returns 0 if SubStr was not found in S.}
function PosCase(const SubStr, S: AnsiString; StartPos: Long = 1): Long;
{Returns position of SubStr in S from position StartPos case insensitively.
 Returns 0 if SubStr was not found in S.}
function PosNoCase(const SubStr, S: AnsiString; StartPos: Long = 1): Long;

{Returns a sub-string from S which is followed by
 Sep separator and deletes the sub-string from S including the separator.}
function StrToken(var S: NString; Sep: Char): NString; overload;
{Returns a sub-string from S which is followed by one of
 Seps separators and deletes the sub-string from S including the separator.}
function StrToken(var S: NString; const Seps: array of Char): NString; overload;
{Same as StrToken but searches from the end of S string.}
function StrTokenEnd(var S: NString; Sep: Char): NString; overload;
{Same as StrToken but searches from the end of S string.}
function StrTokenEnd(var S: NString; const Seps: array of Char): NString;
overload;

{This changes all path delimiters in FileName to NewDelim. I wrote this
 when I needed to use SysUtils.ExtractFileDir in Windows but with
 Unix filename, but ExtractXXX used active platform's delimiter.}
function SwapPathDelims(const FileName: NString; NewDelim: Char = PathDelim):
  NString;
{This works like SysUtils.ExtractFileExt but it does not return horrible
 dot at position 1.}
function GetFileExt(const FileName:NString):NString;


{*************************************************************************}
{                           Error functions                               }
{*************************************************************************}


{This function generates formatted error string from input parameters
 and adds additional info.}
function ErrorStr(Caller: TObject; const FuncName, Message: NString): NString;

{Use this procedure as exception handler in try...except clauses.
 It shows error message with continue/terminate app option and logs
 error message.}
procedure ShowError(const Message: NString; const ExClassName: NString = '');


{*************************************************************************}
{                           Library/File functions                        }
{*************************************************************************}


{Returns full path to given module.}
function GetModuleFileName(Handle: THandle): NString;
{Returns full path to main executable.}
function ExeName: NString;
{Returns directory where exe file is located.}
function ExeDir: NString;

{Returns True if application was compiled with runtime packages.}
function IsCompiledWithPackages: Boolean;
{Returns True if library defined by Handle is Delphi package.}
function IsLibPackage(Handle: THandle): Boolean;

{Loads given library (can be any Win32/Linux executable) and returns Handle.
 if loaded library is Delphi package then InitializePackage is called.
 If operation was not successful Handle is 0.}
function LoadLibrary(const FileName: NString): THandle;
{Unloads given library.
 If library is Delphi package then FinalizaPackage is called. Returns true if successful}
function UnloadLibrary(Handle: THandle): Boolean;
{Returns address of given symbol in given module. Returns nil if unsuccessful.}
function GetProcAddress(Handle: THandle; const Name: NString): Pointer;


{*************************************************************************}
{                           Messaging functions                           }
{*************************************************************************}


{Creates valid  TMsgRec from given parameters.}
procedure CreateMsg(MsgID: Word; var Msg: TMsgRec); overload;
{Creates valid  TMsgRec from given parameters.}
procedure CreateMsg(MsgID: Word; Sender: TObject; var Msg: TMsgRec); overload;
{Creates valid  TMsgRec from given parameters.}
procedure CreateMsg(MsgID: Word; Sender: TObject; Data: Pointer;
  var Msg: TMsgRec); overload;
{Creates valid  TMsgRec from given parameters.}
procedure CreateMsg(MsgID: Word; Sender: TObject; ULongParam: ULong;
  var Msg: TMsgRec); overload;


{*************************************************************************}
{                             Implementation                              }
{*************************************************************************}


implementation

uses
  Objects, Platforms;

type

  {Type for table used in CRC32 calculations.}
  TCRC32Table = array[Byte] of ULong;

const
  {Symbols used when encoding/decoding to/from Base64.}
  Base64Chars = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ' +
    'abcdefghijklmnopqrstuvwxyz0123456789+/';

var

  {Table used in CRC32 calculations.}
  CRC32Table: TCRC32Table;


{*************************************************************************}
{                         Misc functions                                  }
{*************************************************************************}


procedure Switch(var Value: Boolean);
begin
  Value := not Value;
end;

procedure Swap(var A, B: Long);
var
  T: Long;
begin
  T := A;
  A := B;
  B := T;
end;

procedure Swap(var A, B: Float);
var
  T: Float;
begin
  T := A;
  A := B;
  B := T;
end;

procedure Swap(var A, B: NString);
var
  T: NString;
begin
  T := A;
  A := B;
  B := T;
end;

procedure Swap(var A, B: Pointer);
var
  T: Pointer;
begin
  T := A;
  A := B;
  B := T;
end;

procedure Swap(var A, B: Boolean);
var
  T: Boolean;
begin
  T := A;
  A := B;
  B := T;
end;

function Point(X, Y: Long): TPoint;
begin
  Result.X := X;
  Result.Y := Y;
end;

procedure Point(var P: TPoint; X, Y: Long);
begin
  P.X := X;
  P.Y := Y;
end;

function Set255(Value: Long): Long;
{$IFDEF ASMVERSIONS}
asm
  CMP    EAX, 254
  JG     @SetHi
  CMP    EAX, 1
  JL     @SetLo
  RET
@SetHi:
  MOV    EAX, 255
  RET
@SetLo:
  MOV    EAX, 0
end;
{$ELSE}
begin
  Result := Value;
  if Result > 255 then
    Result := 255;
  if Result < 0 then
    Result := 0;
end;
{$ENDIF}

function Iff(const Condition: Boolean; const TruePart, FalsePart: NString):
  NString;
begin
  if Condition then
    Result := TruePart
  else
    Result := FalsePart;
end;

function Iff(const Condition: Boolean; const TruePart, FalsePart: Char): Char;
begin
  if Condition then
    Result := TruePart
  else
    Result := FalsePart;
end;

function Iff(const Condition: Boolean; const TruePart, FalsePart: Byte): Byte;
begin
  if Condition then
    Result := TruePart
  else
    Result := FalsePart;
end;

function Iff(const Condition: Boolean; const TruePart, FalsePart: Long): Long;
begin
  if Condition then
    Result := TruePart
  else
    Result := FalsePart;
end;

function Iff(const Condition: Boolean; const TruePart, FalsePart: ULong): ULong;
begin
  if Condition then
    Result := TruePart
  else
    Result := FalsePart;
end;

function Iff(const Condition: Boolean; const TruePart, FalsePart: Float): Float;
begin
  if Condition then
    Result := TruePart
  else
    Result := FalsePart;
end;

function Iff(const Condition: Boolean; const TruePart, FalsePart: Boolean):
  Boolean;
begin
  if Condition then
    Result := TruePart
  else
    Result := FalsePart;
end;

function Iff(const Condition: Boolean; const TruePart, FalsePart: Pointer):
  Pointer;
begin
  if Condition then
    Result := TruePart
  else
    Result := FalsePart;
end;

function Iff(const Condition: Boolean; const TruePart, FalsePart: Int64): Int64;
begin
  if Condition then
    Result := TruePart
  else
    Result := FalsePart;
end;

procedure SwapLong(P: PLong; Count: Long);
{$IFDEF ASMVERSIONS}
asm
@Loop:
  MOV    ECX, [EAX]
  BSWAP  ECX
  MOV    [EAX], ECX
  ADD    EAX, 4
  DEC    EDX
  JNZ    @Loop
end;
{$ELSE}
var
  I: Long;
  Temp: Long;
begin
  for I := 0 to Count - 1 do
  begin
    Temp := PLongArray(P)[I];
    TLongRec(PLongArray(P)[I]).Bytes[0] := TLongRec(Temp).Bytes[3];
    TLongRec(PLongArray(P)[I]).Bytes[1] := TLongRec(Temp).Bytes[2];
    TLongRec(PLongArray(P)[I]).Bytes[2] := TLongRec(Temp).Bytes[1];
    TLongRec(PLongArray(P)[I]).Bytes[3] := TLongRec(Temp).Bytes[0];
  end;
end;
{$ENDIF}

function SwapLong(Value: Long): Long;
{$IFDEF ASMVERSIONS}
asm
  BSWAP   EAX
end;
{$ELSE}
begin
  TLongRec(Result).Bytes[0] := TLongRec(Value).Bytes[3];
  TLongRec(Result).Bytes[1] := TLongRec(Value).Bytes[2];
  TLongRec(Result).Bytes[2] := TLongRec(Value).Bytes[1];
  TLongRec(Result).Bytes[3] := TLongRec(Value).Bytes[0];
end;
{$ENDIF}


{*************************************************************************}
{                     Encoding/Decoding/CRC functions                     }
{*************************************************************************}


function CalcCRC32NoCase(const Str: AnsiString): ULong;
begin
  Result := CalcCRC32NoCase(PChar(Str), Length(Str));
end;

function CalcCRC32NoCase(Data: Pointer; Size: Long): ULong;
var
  I: Long;
  PB: PByte;
  B: Byte;
begin
  Result := $FFFFFFFF;
  PB := Data;
  for I := 0 to Size - 1 do
  begin
    B := PB^;
    if Char(B) in ['A'..'Z'] then
      B := B or 32;
    Result := (Result shr 8) xor
      CRC32Table[B xor Byte(Result)];
    Inc(PB);
  end
end;

procedure CalcCRC32NoCase(Data: Pointer; Size: Long; var CRC32: ULong);
var
  I: Long;
  PB: PByte;
  B: Byte;
begin
  PB := Data;
  for I := 0 to Size - 1 do
  begin
    B := PB^;
    if Char(B) in ['A'..'Z'] then
      B := B or 32;
    CRC32 := (CRC32 shr 8) xor
      CRC32Table[B xor Byte(CRC32)];
    Inc(PB);
  end
end;

function CalcCRC32(const Str: AnsiString): ULong;
begin
  Result := CalcCRC32(PChar(Str), Length(Str));
end;

function CalcCRC32(Data: Pointer; Size: Long): ULong;
var
  I: Long;
  B: PByte;
begin
  Result := $FFFFFFFF;
  B := Data;
  for I := 0 to Size - 1 do
  begin
    Result := (Result shr 8) xor
      CRC32Table[B^ xor Byte(Result)];
    Inc(B);
  end
end;

procedure CalcCRC32(Data: Pointer; Size: Long; var CRC32: ULong);
var
  I: Long;
  B: PByte;
begin
  B := Data;
  for I := 0 to Size - 1 do
  begin
    CRC32 := (CRC32 shr 8) xor
      CRC32Table[B^ xor Byte(CRC32)];
    Inc(B);
  end
end;

function Base64Encode(const S: AnsiString): AnsiString;
var
  R, C: Byte;
  F, L, M, N, U: Long;
  P: PChar;
begin
  L := Length(S);
  if L = 0 then
    Exit;
  M := L mod 3;
  N := (L div 3) * 4 + M;
  if M > 0 then
    Inc(N);

  U := N mod 4;
  if U > 0 then
  begin
    U := 4 - U;
    Inc(N, U);
  end;
  SetLength(Result, N);
  P := Pointer(Result);
  R := 0;
  for F := 0 to L - 1 do
  begin
    C := Byte(S[F + 1]);
    case F mod 3 of
      0:
        begin
          P^ := Base64Chars[C shr 2 + 1];
          Inc(P);
          R := (C and 3) shl 4;
        end;
      1:
        begin
          P^ := Base64Chars[C shr 4 + R + 1];
          Inc(P);
          R := (C and $0F) shl 2;
        end;
      2:
        begin
          P^ := Base64Chars[C shr 6 + R + 1];
          Inc(P);
          P^ := Base64Chars[C and $3F + 1];
          Inc(P);
        end;
    end;
  end;
  if M > 0 then
  begin
    P^ := Base64Chars[R + 1];
    Inc(P);
  end;
  for F := 1 to U do
  begin
    P^ := '=';
    Inc(P);
  end;
end;

function Base64Decode(const S: AnsiString): AnsiString;
var
  F, L, M, P: Long;
  B, OutPos: Byte;
  OutB: array[1..3] of Byte;
  Lookup: array[Char] of Byte;
  R: PChar;
begin
  L := Length(S);
  P := 0;
  while (L - P > 0) and (S[L - P] = '=') do
    Inc(P);
  M := L - P;
  if M = 0 then
    Exit;
  SetLength(Result, (M * 3) div 4);
  FillChar(Lookup, SizeOf(Lookup), #0);
  for F := 0 to 63 do
    Lookup[Base64Chars[F + 1]] := F;
  R := Pointer(Result);
  OutPos := 0;
  for F := 1 to L - P do
  begin
    B := Lookup[S[F]];
    case OutPos of
      0: OutB[1] := B shl 2;
      1:
        begin
          OutB[1] := OutB[1] or (B shr 4);
          R^ := Char(OutB[1]);
          Inc(R);
          OutB[2] := (B shl 4) and $FF;
        end;
      2:
        begin
          OutB[2] := OutB[2] or (B shr 2);
          R^ := Char(OutB[2]);
          Inc(R);
          OutB[3] := (B shl 6) and $FF;
        end;
      3:
        begin
          OutB[3] := OutB[3] or B;
          R^ := Char(OutB[3]);
          Inc(R);
        end;
    end;
    OutPos := (OutPos + 1) mod 4;
  end;
  if (OutPos > 0) and (P = 0) then
    if OutB[OutPos] <> 0 then
      Result := Result + Char(OutB[OutPos]);
end;

function Base64Encode(Data: Pointer; Size: Long): AnsiString;
var
  S: AnsiString;
begin
  SetString(S, nil, Size);
  Move(Data^, Pointer(S)^, Size);
  Result := Base64Encode(S);
end;

procedure Base64Decode(const S: AnsiString; var Data: Pointer; var Size: Long);
var
  Buff: AnsiString;
begin
  Buff := Base64Decode(S);
  Size := Length(Buff);
  GetMem(Data, Size);
  Move(Pointer(Buff)^, Data^, Size);
end;


{*************************************************************************}
{                   Memory/Pointer Management                             }
{*************************************************************************}


procedure FreeAndNil(var Obj);
var
  Temp: TObject;
begin
  Temp := TObject(Obj);
  Pointer(Obj) := nil;
  Temp.Free;
end;

procedure FreeMemNil(var P);
begin
  FreeMem(Pointer(P));
  Pointer(P) := nil;
end;

procedure GetMemFill(var P; Size: Long; Value: Byte);
begin
  GetMem(Pointer(P), Size);
  FillMemory(Pointer(P), Size, Value);
end;

function GetMemCount: Long;
begin
  Result := {$IFDEF USEMEMMGR}MemMgrs.GetMemCount{$ELSE}0{$ENDIF};
end;

function FreeMemCount: Long;
begin
  Result := {$IFDEF USEMEMMGR}MemMgrs.FreeMemCount{$ELSE}0{$ENDIF};
end;

function ReallocMemCount: Long;
begin
  Result := {$IFDEF USEMEMMGR}MemMgrs.ReallocMemCount{$ELSE}0{$ENDIF};
end;

procedure ZeroMemory(Data: Pointer; Size: Long);
begin
  FillMemory(Data, Size, 0);
end;

procedure FillMemory(Data: Pointer; Size: Long; Value: Byte);
{$IFDEF ASMVERSIONS}
asm
  PUSH   EDI
  MOV    EDI, EAX
  MOV    EAX, ECX
  MOV    AH, AL
  MOV    CX, AX
  SHL    EAX, 16
  MOV    AX, CX
  MOV    ECX, EDX
  SAR    ECX, 2
  JS     @Exit
  REP    STOSD
  MOV    ECX, EDX
  AND    ECX, 3
  REP    STOSB
  POP    EDI
@Exit:
end;
{$ELSE}
begin
  FillChar(Data^, Size, Value);
end;
{$ENDIF}

procedure FillMemoryWord(Data: Pointer; Size: Long; Value: Word);
{$IFDEF ASMVERSIONS}
asm
  PUSH   EDI
  PUSH   EBX
  MOV    EBX, EDX
  MOV    EDI, EAX
  MOV    EAX, ECX
  MOV    CX, AX
  SHL    EAX, 16
  MOV    AX, CX
  MOV    ECX, EDX
  SHR    ECX, 2
  JZ     @Word
  REP    STOSD
@Word:
  MOV    ECX, EBX
  AND    ECX, 2
  JZ     @Byte
  MOV    [EDI], AX
  ADD    EDI, 2
@Byte:
  MOV    ECX, EBX
  AND    ECX, 1
  JZ     @Exit
  MOV    [EDI], AL
@Exit:
  POP    EBX
  POP    EDI
end;
{$ELSE}
var
  I, V: Long;
begin
  V := Value * $10000 + Value;
  for I := 0 to (Size shr 2) - 1 do
    PULongArray(Data)[I] := V;
  case Size mod 4 of
    1: PByteArray(Data)[Size - 1] := Lo(Value);
    2: PWordArray(Data)[Size div 2] := Value;
    3:
      begin
        PWordArray(Data)[Size div 2 - 1] := Value;
        PByteArray(Data)[Size - 1] := Lo(Value);
      end;
  end;
end;
{$ENDIF}

procedure FillMemoryULong(Data: Pointer; Size: Long; Value: ULong);
{$IFDEF ASMVERSIONS}
asm
  PUSH   EDI
  PUSH   EBX
  MOV    EBX, EDX
  MOV    EDI, EAX
  MOV    EAX, ECX
  MOV    ECX, EDX
  SHR    ECX, 2
  JZ     @Word
  REP    STOSD
@Word:
  MOV    ECX, EBX
  AND    ECX, 2
  JZ     @Byte
  MOV    [EDI], AX
  ADD    EDI, 2
@Byte:
  MOV    ECX, EBX
  AND    ECX, 1
  JZ     @Exit
  MOV    [EDI], AL
@Exit:
  POP    EBX
  POP    EDI
end;
{$ELSE}
var
  I: Long;
begin
  for I := 0 to Size shr 2 - 1 do
    PULongArray(Data)[I] := Value;
  case Size mod 4 of
    1: PByteArray(Data)[Size - 1] := TLongRec(Value).Bytes[0];
    2: PWordArray(Data)[Size div 2] := TLongRec(Value).Words[0];
    3:
      begin
        PWordArray(Data)[Size div 2 - 1] := TLongRec(Value).Words[0];
        PByteArray(Data)[Size - 1] := TLongRec(Value).Bytes[0];
      end;
  end;
end;
{$ENDIF}


{*************************************************************************}
{                             Base Math                                   }
{*************************************************************************}


function Sign(const Value: Long): Short;
begin
  if Value < 0 then
    Result := -1
  else
    Result := 1;
end;

function Sign(const Value: Float): Short;
begin
  if Value < 0 then
    Result := -1
  else
    Result := 1;
end;

function Min(const A, B: Long): Long;
begin
  if A < B then
    Result := A
  else
    Result := B;
end;

function Min(const A, B: Float): Float;
begin
  if A < B then
    Result := A
  else
    Result := B;
end;

function Max(const A, B: Long): Long;
begin
  if A > B then
    Result := A
  else
    Result := B;
end;

function Max(const A, B: Float): Float;
begin
  if A > B then
    Result := A
  else
    Result := B;
end;

function Clamp(const Value: Float; const Min: Float; const Max: Float):
  Float;
begin
  if Value < Min then
    Result := Min
  else
    if Value > Max then
      Result := Max
    else
      Result := Value;
end;

function Clamp(const Value: Long; const Min: Long; const Max: Long):
  Long;
begin
  if Value < Min then
    Result := Min
  else
    if Value > Max then
      Result := Max
    else
      Result := Value;
end;

function CompFloat(const A, B: Float): Boolean;
begin
  Result := (A = B) or (Abs(B - A) < FloatTolerance);
end;

function IsZero(const Value: Float): Boolean;
begin
  Result := Abs(Value) < FloatTolerance;
end;

function Rand: Float;
begin
  Result := Random;
end;

function Rand(const Max: Long): Long;
begin
  Result := Random(Max);
end;

function Rand(const Max: Float): Float;
begin
  Result := Random(Round(Max));
end;

function IsPow2(Num: Long): Boolean;
begin
  Result := (Num and -Num) = Num;
end;

function NextPow2(Num: Long): Long;
begin
  Result := Num and -Num;
  while (Result < Num) do
    Result := Result shl 1;
end;

function Ceil(Value: Float): Long;
begin
  Result := Trunc(Value);
  if Frac(Value) > 0 then
    Inc(Result);
end;

function Floor(Value: Float): Long;
begin
  Result := Trunc(Value);
  if Frac(Value) < 0 then
    Dec(Result);
end;


{*************************************************************************}
{                         Some strings                                    }
{*************************************************************************}


function VersionToStr(const Version: TVersionInfo): NString;
begin
  Result := Format('%.2d.%.2d.%.4d', [Version.Major, Version.Minor,
    Version.Build]);
end;

function CompStr(const S1, S2: AnsiString): Long;
{$IFDEF ASMVERSIONS}
{$IFNDEF PIC}
asm
  TEST   EAX, EAX
  JE     @@2
  TEST   EDX, EDX
  JE     @@3
  PUSH   ESI
  PUSH   EDI
  MOV    ESI, EAX
  MOV    EDI, EDX
  JMP    @@1
@@0:
  TEST   AL, AL
  JE     @@4
  INC    ESI
  INC    EDI
@@1:
  MOVZX  EAX, BYTE PTR [ESI]
  MOVZX  EDX, BYTE PTR [EDI]
  CMP    AL, DL
  JE     @@0
  MOV    AL, BYTE PTR [EAX + UpperChars]
  MOV    DL, BYTE PTR [EDX + UpperChars]
  CMP    AL, DL
  JE     @@0
  MOVZX  EAX, AL
  MOVZX  EDX, DL
  SUB    EAX, EDX
  POP    EDI
  POP    ESI
  RET
@@2:
  TEST   EDX, EDX
  JE     @@7
  MOV    CH, BYTE PTR [EDX]
  TEST   CH, CH
  JE     @@7
  NOT    EAX
  RET
@@3:
  MOV    CL, BYTE PTR [EAX]
  TEST   CL, CL
  JE     @@5
  MOV    EAX, 1
  RET
@@4:
  POP    EDI
  POP    ESI
@@5:
  XOR    EAX, EAX
@@7:
end;
{$ELSE}
begin
  Result := SysUtils.CompareText(S1, S2)
end;
{$ENDIF}
{$ELSE}
begin
  Result := SysUtils.CompareText(S1, S2)
end;
{$ENDIF}

function CompStrCase(const S1, S2: AnsiString): Long;
{$IFDEF ASMVERSIONS}
asm
  PUSH   ESI
  PUSH   EDI
  MOV    ESI, EAX
  MOV    EDI, EDX
  OR     EAX, EAX
  JE     @@1
  MOV    EAX, [EAX - 4]
@@1:
  OR     EDX, EDX
  JE     @@2
  MOV    EDX, [EDX - 4]
@@2:
  MOV    ECX, EAX
  CMP    ECX, EDX
  JBE    @@3
  MOV    ECX, EDX
@@3:
  CMP    ECX, ECX
  REPE   CMPSB
  JE     @@4
  MOVZX  EAX, BYTE PTR [ESI - 1]
  MOVZX  EDX, BYTE PTR [EDI - 1]
@@4:
  SUB    EAX, EDX
  POP    EDI
  POP    ESI
end;
{$ELSE}
begin
  Result := SysUtils.CompareStr(S1, S2)
end;
{$ENDIF}

function SameStr(const S1, S2: AnsiString): Boolean;
{$IFDEF ASMVERSIONS}
{$IFNDEF PIC}
asm
  CMP    EAX, EDX
  JE     @@08
  TEST   EAX, EAX
  JE     @@09
  TEST   EDX, EDX
  JE     @@07
  MOV    ECX, [EAX - 4]
  CMP    ECX, [EDX - 4]
  JE     @@I
  XOR    EAX, EAX
  RET
@@I:
  TEST   ECX, ECX
  JE     @@07
  PUSH   ESI
  PUSH   EDI
  MOV    ESI, EAX
  MOV    EDI, EDX
@@00:
  DEC    ECX
  JS     @@Q
@@01:
  MOVZX  EAX, BYTE PTR [ESI + ECX]
  MOVZX  EDX, BYTE PTR [EDI + ECX]
  CMP    AL, DL
  JE     @@00
  MOV    AL, BYTE PTR [EAX + UpperChars]
  XOR    AL, BYTE PTR [EDX + UpperChars]
  JE     @@00
  POP    EDI
  POP    ESI
@@07:
  XOR    EAX, EAX
@@09:
  RET
@@Q:
  POP    EDI
  POP    ESI
@@08:
  MOV    EAX, 1
end;
{$ELSE}
begin
  Result := StrLowerCase(S1) = StrLowerCase(S2);
end;
{$ENDIF}
{$ELSE}
begin
  Result := StrLowerCase(S1) = StrLowerCase(S2);
end;
{$ENDIF}

function SameStrCase(const S1, S2: AnsiString): Boolean;
{$IFDEF ASMVERSIONS}
asm
  CMP    EAX, EDX
  JE     @@08
  TEST   EAX, EAX
  JE     @@qt2
  TEST   EDX, EDX
  JE     @@qt1
  MOV    ECX, [EAX - 4]
  CMP    ECX, [EDX - 4]
  JE     @@01
@@qt1:
  XOR    EAX, EAX
@@qt2:
  RET
@@01:
  PUSH   EBX
  SUB    ECX, 8
  JS     @@nx
@@lp:
  MOV    EBX, DWORD PTR [EAX + ECX]
  CMP    EBX, DWORD PTR [EDX + ECX]
  JNE    @@zq
  MOV    EBX, DWORD PTR [EAX + ECX + 4]
  CMP    EBX, DWORD PTR [EDX + ECX + 4]
  JNE    @@zq
  SUB    ECX, 8
  JNS    @@lp
@@nx:
  JMP    DWORD PTR @@tV[ECX * 4 + 32]
@@tV:
  DD     @@eq,@@t1,@@t2,@@t3
  DD     @@t4,@@t5,@@t6,@@t7
@@t7:
  MOV    BL, BYTE PTR [EAX + 6]
  XOR    BL, BYTE PTR [EDX + 6]
  JNE    @@zq
@@t6:
  MOV    BL, BYTE PTR [EAX + 5]
  XOR    BL, BYTE PTR [EDX + 5]
  JNE    @@zq
@@t5:
  MOV    BL, BYTE PTR [EAX + 4]
  XOR    BL, BYTE PTR [EDX + 4]
  JNE    @@zq
@@t4:
  MOV    EBX, DWORD PTR [EAX]
  CMP    EBX, DWORD PTR [EDX]
  JNE    @@zq
@@eq:
  POP    EBX
@@08:
  MOV    EAX, 1
  RET
@@t3:
  MOV    BL, BYTE PTR [EAX + 2]
  XOR    BL, BYTE PTR [EDX + 2]
  JNE    @@zq
@@t2:
  MOV    BL, BYTE PTR [EAX + 1]
  XOR    BL, BYTE PTR [EDX + 1]
  JNE    @@zq
@@t1:
  MOV    BL, BYTE PTR [EAX]
  XOR    BL, BYTE PTR [EDX]
  JNE    @@zq
  POP    EBX
  MOV    EAX, 1
  RET
@@zq:
  POP    EBX
@@07:
  XOR    EAX, EAX
end;
{$ELSE}
begin
  Result := S1 = S2;
end;
{$ENDIF}

{$IFDEF ASMVERSIONS}
{$IFNDEF PIC}
procedure StrUpperMove(const Source: AnsiString; var Dest: AnsiString; MaxL: ULong);
asm
  MOV    EDX, [EDX]
  TEST   EAX, EAX
  JE     @@x2
  PUSH   EDI
  MOV    EDI, [EAX - 4]
  TEST   EDI, EDI
  JE     @@x4
  CMP    ECX, EDI
  JB     @@x0
  MOV    ECX, EDI
@@x0:
  MOV    [EDX - 4], ECX
  MOV    BYTE PTR [ECX + EDX], $00
  DEC    ECX
  JS     @@qt
  MOV    EDI, EDX
@@1:
  MOVZX  EDX, BYTE PTR [EAX + ECX]
  MOV    DL, BYTE PTR [EDX + UpperChars]
  MOV    BYTE PTR [EDI + ECX], DL
  DEC    ECX
  JNS    @@1
@@qt:
  POP    EDI
  RET
@@x4:
  POP    EDI
  XOR    EAX, EAX
@@x2:
  TEST   EDX, EDX
  JE     @@x3
  MOV    [EDX - 4], EAX
  MOV    BYTE PTR [EDX], AL
@@x3:
end;

procedure StrLowerMove(const Source: AnsiString; var Dest: AnsiString; MaxL: ULong);
asm
  MOV    EDX, [EDX]
  TEST   EAX, EAX
  JE     @@x2
  PUSH   EDI
  MOV    EDI, [EAX - 4]
  TEST   EDI, EDI
  JE     @@x4
  CMP    ECX, EDI
  JB     @@x0
  MOV    ECX, EDI
@@x0:
  MOV    [EDX - 4], ECX
  MOV    BYTE PTR [ECX + EDX], $00
  DEC    ECX
  JS     @@qt
  MOV    EDI, EDX
@@1:
  MOVZX  EDX, BYTE PTR [EAX + ECX]
  MOV    DL, BYTE PTR [EDX + LowerChars]
  MOV    BYTE PTR [EDI + ECX], DL
  DEC    ECX
  JNS    @@1
@@qt:
  POP    EDI
  RET
@@x4:
  POP    EDI
  XOR    EAX, EAX
@@x2:
  TEST   EDX, EDX
  JE     @@x3
  MOV    [EDX - 4], EAX
  MOV    BYTE PTR [EDX], AL
@@x3:
end;
{$ENDIF}
{$ENDIF}

function StrUpperCase(const S: AnsiString): AnsiString;
{$IFDEF ASMVERSIONS}
{$IFNDEF PIC}
var
  L: Long;
begin
  L := Length(S);
  SetLength(Result, L);
  StrUpperMove(S, Result, L);
end;
{$ELSE}
begin
  Result := SysUtils.UpperCase(S);
end;
{$ENDIF}
{$ELSE}
begin
  Result := SysUtils.UpperCase(S);
end;
{$ENDIF}

function StrLowerCase(const S: AnsiString): AnsiString;
{$IFDEF ASMVERSIONS}
{$IFNDEF PIC}
var
  L: Long;
begin
  L := Length(S);
  SetLength(Result, L);
  StrLowerMove(S, Result, L);
end;
{$ELSE}
begin
  Result := SysUtils.LowerCase(S);
end;
{$ENDIF}
{$ELSE}
begin
  Result := SysUtils.LowerCase(S);
end;
{$ENDIF}

function StrReplace(const S, OldPattern, NewPattern: AnsiString;
  Flags: TReplaceFlags): AnsiString;
var
  SearchStr, Patt, NewStr: AnsiString;
  Offset: Long;
begin
  if srIgnoreCase in Flags then
  begin
    SearchStr := StrUpperCase(S);
    Patt := StrUpperCase(OldPattern);
  end
  else
  begin
    SearchStr := S;
    Patt := OldPattern;
  end;
  NewStr := S;
  Result := '';
  while SearchStr <> '' do
  begin
    Offset := PosCase(Patt, SearchStr, 1);
    if Offset = 0 then
    begin
      Result := Result + NewStr;
      Break;
    end;
    Result := Result + Copy(NewStr, 1, Offset - 1) + NewPattern;
    NewStr := Copy(NewStr, Offset + Length(OldPattern), MaxLong);
    if not (srReplaceAll in Flags) then
    begin
      Result := Result + NewStr;
      Break;
    end;
    SearchStr := Copy(SearchStr, Offset + Length(Patt), MaxLong);
  end;
end;

function PosCase(const SubStr, S: AnsiString; StartPos: Long): Long;
{$IFDEF ASMVERSIONS}
asm
  PUSH   ESI
  PUSH   EDI
  PUSH   EBX
  PUSH   EDX
  TEST   EAX, EAX
  JE     @@qt
  TEST   EDX, EDX
  JE     @@qt0
  MOV    ESI, EAX
  MOV    EDI, EDX
  MOV    EAX, [EAX - 4]
  MOV    EDX, [EDX - 4]
  DEC    EAX
  SUB    EDX, EAX
  DEC    ECX
  SUB    EDX, ECX
  JNG    @@qt0
  XCHG   EAX, EDX
  ADD    EDI, ECX
  MOV    ECX, EAX
  JMP    @@nx
@@fr:
  INC    EDI
  DEC    ECX
  JE     @@qt0
@@nx:
  MOV    EBX, EDX
  MOV    AL, BYTE PTR [ESI]
@@lp1:
  CMP    AL, BYTE PTR [EDI]
  JE     @@uu
  INC    EDI
  DEC    ECX
  JE     @@qt0
  CMP    AL, BYTE PTR [EDI]
  JE     @@uu
  INC    EDI
  DEC    ECX
  JE     @@qt0
  CMP    AL, BYTE PTR [EDI]
  JE     @@uu
  INC    EDI
  DEC    ECX
  JE     @@qt0
  CMP    AL, BYTE PTR [EDI]
  JE     @@uu
  INC    EDI
  DEC    ECX
  JNE    @@lp1
@@qt0:
  XOR    EAX, EAX
@@qt:
  POP    ECX
  POP    EBX
  POP    EDI
  POP    ESI
  RET
@@uu:
  TEST   EDX, EDX
  JE     @@fd
@@lp2:
  MOV    AL, BYTE PTR [ESI + EBX]
  CMP    AL, BYTE PTR [EDI + EBX]
  JNE    @@fr
  DEC    EBX
  JE     @@fd
  MOV    AL, BYTE PTR [ESI + EBX]
  CMP    AL, BYTE PTR [EDI + EBX]
  JNE    @@fr
  DEC    EBX
  JE     @@fd
  MOV    AL, BYTE PTR [ESI + EBX]
  CMP    AL, BYTE PTR [EDI + EBX]
  JNE    @@fr
  DEC    EBX
  JE     @@fd
  MOV    AL, BYTE PTR [ESI + EBX]
  CMP    AL, BYTE PTR [EDI + EBX]
  JNE    @@fr
  DEC    EBX
  JNE    @@lp2
@@fd:
  LEA    EAX,[EDI + 1]
  SUB    EAX,[ESP]
  POP    ECX
  POP    EBX
  POP    EDI
  POP    ESI
end;
{$ELSE}
var
  I, X: Integer;
  Len, LenSubStr: Integer;
begin
  if StartPos = 1 then
    Result := Pos(SubStr, S)
  else
  begin
    I := StartPos;
    LenSubStr := Length(SubStr);
    Len := Length(S) - LenSubStr + 1;
    while I <= Len do
    begin
      if S[I] = SubStr[1] then
      begin
        X := 1;
        while (X < LenSubStr) and (S[I + X] = SubStr[X + 1]) do
          Inc(X);
        if (X = LenSubStr) then
        begin
          Result := I;
          Exit;
        end;
      end;
      Inc(I);
    end;
    Result := 0;
  end;
end;
{$ENDIF}

function PosNoCase(const SubStr, S: AnsiString; StartPos: Long): Long;
{$IFDEF ASMVERSIONS}
{$IFNDEF PIC}
asm
  PUSH   ESI
  PUSH   EDI
  PUSH   EBX
  TEST   EAX, EAX
  JE     @@qt
  TEST   EDX, EDX
  JE     @@qt0
  MOV    ESI, EAX
  MOV    EDI, EDX
  PUSH   EDX
  MOV    EAX,[EAX - 4]
  MOV    EDX,[EDX - 4]
  DEC    EAX
  SUB    EDX, EAX
  DEC    ECX
  PUSH   EAX
  SUB    EDX, ECX
  JNG    @@qtx
  ADD    EDI, ECX
  MOV    ECX, EDX
  MOV    EDX, EAX
  MOVZX  EBX, BYTE PTR [ESI]
  MOV    AL, BYTE PTR [EBX + UpperChars]
@@lp1:
  MOVZX  EBX, BYTE PTR [EDI]
  CMP    AL, BYTE PTR [EBX + UpperChars]
  JE     @@uu
@@fr:
  INC    EDI
  DEC    ECX
  JE     @@qtx
  MOVZX  EBX, BYTE PTR [EDI]
  CMP    AL, BYTE PTR [EBX + UpperChars]
  JE     @@uu
  INC    EDI
  DEC    ECX
  JE     @@qtx
  MOVZX  EBX, BYTE PTR [EDI]
  CMP    AL, BYTE PTR [EBX + UpperChars]
  JE     @@uu
  INC    EDI
  DEC    ECX
  JE     @@qtx
  MOVZX  EBX, BYTE PTR [EDI]
  CMP    AL, BYTE PTR [EBX + UpperChars]
  JE     @@uu
  INC    EDI
  DEC    ECX
  JNE    @@lp1
@@qtx:
  ADD    ESP, $08
@@qt0:
  XOR    EAX, EAX
@@qt:
  POP    EBX
  POP    EDI
  POP    ESI
  RET
@@ms:
  MOVZX  EBX, BYTE PTR [ESI]
  MOV    AL, BYTE PTR [EBX + UpperChars]
  MOV    EDX, [ESP]
  JMP    @@fr
@@uu:
  TEST   EDX,EDX
  JE     @@fd
@@lp2:
  MOV    BL, BYTE PTR [ESI + EDX]
  MOV    AH, BYTE PTR [EDI + EDX]
  CMP    BL, AH
  JE     @@eq
  MOV    AL, BYTE PTR [EBX + UpperChars]
  MOVZX  EBX, AH
  CMP    AL, BYTE PTR [EBX + UpperChars]
  JNE    @@ms
@@eq:
  DEC    EDX
  JNZ    @@lp2
@@fd:
  LEA    EAX,[EDI + 1]
  POP    ECX
  SUB    EAX,[ESP]
  POP    ECX
  POP    EBX
  POP    EDI
  POP    ESI
end;
{$ELSE}
begin
  Result := PosCase(StrLowerCase(SubStr), StrLowerCase(S), StartPos);
end;
{$ENDIF}
{$ELSE}
begin
  Result := PosCase(StrLowerCase(SubStr), StrLowerCase(S), StartPos);
end;
{$ENDIF}

function StrToken(var S: NString; Sep: Char): NString;
var
  I: Long;
begin
  I := PosNoCase(Sep, S, 1);
  if I <> 0 then
  begin
    Result := Copy(S, 1, I - 1);
    Delete(S, 1, I);
  end
  else
  begin
    Result := S;
    S := '';
  end;
end;

function StrToken(var S: NString; const Seps: array of Char): NString;
var
  I, J, K: Long;
begin
  J := Length(S) + 1;
  for I := Low(Seps) to High(Seps) do
  begin
    K := PosNoCase(Seps[I], S, 1);
    if (K < J) and (K > 0) then
      J := K;
  end;
  if J <> Length(S) + 1 then
  begin
    Result := Copy(S, 1, J - 1);
    Delete(S, 1, J);
  end
  else
  begin
    Result := S;
    S := '';
  end;
end;

function StrTokenEnd(var S: NString; Sep: Char): NString;
var
  I, J: Long;
begin
  J := 0;
  I := PosNoCase(Sep, S, 1);
  while I <> 0 do
  begin
    J := I;
    I := PosNoCase(Sep, S, J + 1);
  end;
  if J <> 0 then
  begin
    Result := Copy(S, J + 1, MaxLong);
    Delete(S, J, MaxLong);
  end
  else
  begin
    Result := S;
    S := '';
  end;
end;

function StrTokenEnd(var S: NString; const Seps: array of Char): NString;
var
  I, J: Long;

  function PosFirst(Start: Long): Long;
  var
    I, J, K: Long;
  begin
    J := 0;
    for I := Low(Seps) to High(Seps) do
    begin
      K := PosNoCase(Seps[I], S, Start);
      if K > J then
        J := K;
    end;
    Result := J;
  end;

begin
  J := 0;
  I := PosFirst(1);
  while I <> 0 do
  begin
    J := I;
    I := PosFirst(J + 1);
  end;
  if J <> 0 then
  begin
    Result := Copy(S, J + 1, MaxLong);
    Delete(S, J, MaxLong);
  end
  else
  begin
    Result := S;
    S := '';
  end;
end;

function SwapPathDelims(const FileName: NString; NewDelim: Char): NString;
begin
  Result := FileName;
  Result := StrReplace(Result, '\', NewDelim, [srReplaceAll]);
  Result := StrReplace(Result, '/', NewDelim, [srReplaceAll]);
end;

function GetFileExt(const FileName:NString):NString;
begin
  Result:=ExtractFileExt(FileName);
  if Length(Result) > 1 then
    Delete(Result, 1, 1);
end;


{*************************************************************************}
{                           Error functions                               }
{*************************************************************************}


function ErrorStr(Caller: TObject; const FuncName, Message: NString):
  NString;
var
  S: NString;
begin
  if Caller = nil then
    S := sNil;
  if Caller is TObject then
    S := Format('Class =  %s, Address =  $%p', [Caller.ClassName,
      Pointer(Caller)]);
  if Caller is TNBaseObject then
    S := S + ', ID =  ' + IntToStr(TNBaseObject(Caller).ID);
  if Caller is TNSingleton then
    S := S + ', Name =  ''' + TNSingleton(Caller).Name + '''';
  Result := 'Error occurred!' + sLineBreak + 'Caller Object: ' + S
    + sLineBreak + 'Function: ' + FuncName + sLineBreak +
    Format('Platform: %s %s', [Platform.Name,
    VersionToStr(Platform.Version)]) + sLineBreak +
    'Message:' + sLineBreak + Message + sLineBreak + sLineBreak;
end;

procedure ShowError(const Message: NString; const ExClassName: NString);
var
  P: Long;
  Msg, EMsg: NString;
begin
  Msg := Message;
  // checks if err msg was created by Kernel.ErrorStr
  P := Pos('Message:', Msg);
  EMsg := Msg;
  if P > 0 then
    // delete long error message header
    Delete(EMsg, 1, P + 9);
  // replace all line breaks with spaces
  EMsg := StrReplace(EMsg, sLineBreak, #32);
  Log.AddError(EMsg);
  // insert exception class name
  if (P > 0) and (ExClassName <> '') then
    Insert('Exception Class: ' + ExClassName + sLineBreak, Msg, P);
  case Platform.ErrorMsg(Msg) of
    emTerminate: Halt(1);
    emContinue: ;
  end;
end;


{*************************************************************************}
{                         Library/File functions                          }
{*************************************************************************}


function GetModuleFileName(Handle: THandle): NString;
begin
  Result := Platform.GetModuleFileName(Handle);
end;

function ExeName: NString;
begin
  Result := GetModuleFileName(MainInstance);
end;

function ExeDir: NString;
begin
  Result := ExtractFileDir(ExeName) + PathDelim;
end;

function IsCompiledWithPackages: Boolean;
begin
  Result := FindClassHInstance(System.TObject) <> HInstance;
end;

function IsLibPackage(Handle: THandle): Boolean;
begin
  Result := False;
  if Handle <> 0 then
    Result := (GetProcAddress(Handle, 'Initialize') <> nil) and
      (GetProcAddress(Handle, 'Finalize') <> nil);
end;

function LoadLibrary(const FileName: NString): THandle;
begin
  Result := Platform.LoadLibrary(FileName);
  if IsLibPackage(Result) then
    InitializePackage(Result);
end;

function UnloadLibrary(Handle: THandle): Boolean;
begin
  Result := False;
  if (Handle = 0) or (Handle = HInstance) or (Handle = MainInstance) then
    Exit;
  if IsLibPackage(Handle) then
    FinalizePackage(Handle);
  Result := Platform.UnloadLibrary(Handle);
end;

function GetProcAddress(Handle: THandle; const Name: NString): Pointer;
begin
  Result := Platform.GetProcAddress(Handle, PChar(Name));
end;


{*************************************************************************}
{                           Messaging functions                           }
{*************************************************************************}


procedure CreateMsg(MsgID: Word; var Msg: TMsgRec);
begin
  Msg.MsgID := MsgID;
  Msg.Sender := nil;
  Msg.ULParam := 0;
end;

procedure CreateMsg(MsgID: Word; Sender: TObject; var Msg: TMsgRec);
begin
  Msg.MsgID := MsgID;
  Msg.Sender := Sender;
  Msg.ULParam := 0;
end;

procedure CreateMsg(MsgID: Word; Sender: TObject; Data: Pointer;
  var Msg: TMsgRec);
begin
  Msg.MsgID := MsgID;
  Msg.Sender := Sender;
  Msg.Data := Data;
end;

procedure CreateMsg(MsgID: Word; Sender: TObject; ULongParam: ULong;
  var Msg: TMsgRec);
begin
  Msg.MsgID := MsgID;
  Msg.Sender := Sender;
  Msg.ULParam := ULongParam;
end;


{*************************************************************************}
{                         Internal stuff                                  }
{*************************************************************************}


procedure InitCRC32Table;
const
  Polynom = $EDB88320;
var
  I, J: Long;
  C: ULong;
begin
  for I := 0 to 255 do
  begin
    C := I;
    for J := 0 to 7 do
    begin
      if (C and $01) <> 0 then
        C := Polynom xor (C shr 1)
      else
        C := C shr 1;
    end;
    CRC32Table[I] := C;
  end;
end;

procedure InitUpperChars(var Data: array of Char);
var
  I: Long;
begin
  for I := Low(Data) to High(Data) do
    Data[I] := Char(I);
  for I := $61 to $7A do
    Data[I] := Char(I - $20);
  Data[$83] := #$81;
  Data[$90] := #$80;
  Data[$9A] := #$8A;
  Data[$9C] := #$8C;
  Data[$9D] := #$8D;
  Data[$9E] := #$8E;
  Data[$9F] := #$8F;
  Data[$A2] := #$A1;
  Data[$B3] := #$B2;
  Data[$B4] := #$A5;
  Data[$B8] := #$A8;
  Data[$BA] := #$AA;
  Data[$BC] := #$A3;
  Data[$BE] := #$BD;
  Data[$BF] := #$AF;
  for I := $E0 to $FF do
    Data[I] := Char(I - $20);
end;

procedure InitLowerChars(var Data: array of Char);
var
  I: Long;
begin
  for I := Low(Data) to High(Data) do
    Data[I] := Char(I);
  for I := $41 to $5A do
    Data[I] := Char(I + $20);
  Data[$81] := #$83;
  Data[$80] := #$90;
  Data[$8A] := #$9A;
  Data[$8C] := #$9C;
  Data[$8D] := #$9D;
  Data[$8E] := #$9E;
  Data[$8F] := #$9F;
  Data[$A1] := #$A2;
  Data[$B2] := #$B3;
  Data[$A5] := #$B4;
  Data[$A8] := #$B8;
  Data[$AA] := #$BA;
  Data[$A3] := #$BC;
  Data[$BD] := #$BE;
  Data[$AF] := #$BF;
  for I := $C0 to $DF do
    Data[I] := Char(I + $20);
end;

procedure AssertHandler(const Message, Filename: NString;
  LineNumber: Long; ErrorAddr: Pointer);
var
  P: Long;
  Msg: NString;
begin
  Msg := Message;
  P := Pos('Message', Msg);
  if P > 0 then
    Insert('Source module: ' + FileName + sLineBreak +
      Format('Source line: %d' + sLineBreak + 'Error Address: $%.8x' +
      sLineBreak, [LineNumber, Cardinal(ErrorAddr)]), Msg, P)
  else
    Msg := Message + ' (' + FileName + ', line ' + IntToStr(LineNumber) + ')';
  ShowError(Msg, EAssertionFailed.ClassName);
end;

procedure InitSingletons;
begin
  Platform;
  Log;
{$IFNDEF MODLIBRARY}
  ModuleRegister;
{$ENDIF}
end;

var
  Msg: TMsgRec;

initialization

  {I always forget call this before working with random numbers, so
   I put it here.}
  Randomize;

  InitCRC32Table;
  InitUpperChars(UpperChars);
  InitLowerChars(LowerChars);

  {This is very important though it does not look.
   Text streaming systems (like XML) stores float values
   in the form of FloatToStr() which uses locale dependent
   separators. My locale is CZ so my decimal separator
   is ','. So if someone with different locale loads this
   float-to-string and converts it back - error will occur!!}
  SysUtils.DecimalSeparator := '.';
  SysUtils.ThousandSeparator := ',';

  {Sets AssertErrorProc variable from System.pas to
   new handler (default is in SysUtils.pas - it causes exception
   and invoke debugger if running in Delphi IDE but
   shows ugly windows Send Error Message and Runtime Error
   dialogs if running off IDE (and off VCL)}
  AssertErrorProc := @AssertHandler;

  {Call singleton's get functions so they are by them created now.}
  InitSingletons;

  {Register some TNPersistent classes for streaming. It must be
   here insted of Objects.pas because in Objects' init section
   UpperChars are not yet initialized bu they are used.}
  RegisterClasses([TNStringList, TNStringObjectList, TNStringPointerList,
    TNPointerList]);

finalization

  CreateMsg(KM_FreeSelf, Msg);
{$IFNDEF MODLIBRARY}
  ModuleRegister.Dispatch(Msg);
{$ENDIF}
  Log.Dispatch(Msg);
  Platform.Dispatch(Msg);

end.

