{*************************************************************************}
{*   Lists                                                               *}
{*   Copyright (C) 2004 Marek Mauder, Vampyre SoftWorks                  *}
{*   Engine\Kernel\Lists.pas                                             *}
{*************************************************************************}
{                                                                         }
{ This program is free software; you can redistribute it and/or modify it }
{ under the terms of the GNU General Public License as published by the   }
{ Free Software Foundation; either version 2 of the License, or (at your  }
{ option) any later version. This program is distributed in the hope that }
{ it will be useful, but WITHOUT ANY WARRANTY; without even the implied   }
{ warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See   }
{ the GNU General Public License for more details.                        }
{                                                                         }
{ You should have received a copy of the GNU General Public License       }
{ along with this program; if not, write to the Free Software Foundation, }
{ Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.         }
{                                                                         }
{*************************************************************************}
{  Created On: 12nd January 2004  (12.01.2004)                            }
{*************************************************************************}

{This unit contains some additional List classes, descendants of
 TNAbstractList in Objects.pas.}

unit Lists;

{$I Compiler.inc}

interface

uses
  SysUtils, Kernel, Objects, Platforms, TypInfo;
  
type


{*************************************************************************}
{                      TNCustomStringKeys descendants                     }
{*************************************************************************}


  TStrValueRecord = record
    Key: NString;
    Value: Pointer;
  end;
  PStrValueRecord = ^TStrValueRecord;

  TStrValueArray = array[0..MaxListSize - 1] of TStrValueRecord;
  PStrValueArray = ^TStrValueArray;

  {Base list for list which store some Value (number, record, ...)
   associated with string Key.}
  TNCustomStringValueList = class(TNCustomStringKeys)
  protected
    FValueSize: Long;
    function GetKey(Index: Long): NString; override;
    procedure PutKey(Index: Long; const Value: NString); override;
    procedure InsertInternal(Index: Long; const AKey: NString; const Item);
    procedure FinalizeItem(Index: Long); override;
    function Compare(Index1, Index2: Long): Long; override;
    function Compare(Index: Long; const S: NString): Long; override;
    function GetMemoryUsage: Long; override;
    function GetLine(Index: Long): NString; override;
    procedure AddFromLine(const S: NString); override;
    function ValueToString(const Value): NString; virtual;
    procedure StringToValue(const S: NString; var Value); virtual;
  public
    constructor Create(AValueSize: Long); overload;
    {Adds new Key and associated Value.}
    function Add(const AKey: NString; const Value): Long;
    {Inserts new Key and associated Value.}
    procedure Insert(Index: Long; const AKey: NString; const Value);
    {Returns index of given Value. Returns -1 if item is not in list.}
    function IndexOfValue(const Value): Long;
    {Size of Value associated with each Key in Bytes.}
    property ValueSize: Long read FValueSize;
  end;

  {List for storing Int64 numbers associated to string Keys.}
  TNStringInt64List = class(TNCustomStringValueList)
  protected
    function GetNumber(Index: Long): Int64;
    procedure PutNumber(Index: Long; const Value: Int64);
    function ValueToString(const Value): NString; override;
    procedure StringToValue(const S: NString; var Value); override;
    procedure ReadData(Reader: TNAbstractReader); override;
    procedure WriteData(Writer: TNAbstractWriter); override;
  public
    constructor Create; override;
    {Adds new Key and associated Int64 Value.}
    function Add(const AKey: NString; const Value: Int64): Long;
    {Inserts new Key and associated Int64 Value.}
    procedure Insert(Index: Long; const AKey: NString; const Value: Int64);
    {Returns index of given Int64 Value. Returns -1 if item is not in list.}
    function IndexOfValue(const Value: Int64): Long;
    property Numbers[Index: Long]: Int64 read GetNumber write PutNumber;
  end;

  TDictRecord = record
    Key: NString;
    Word: NString;
  end;
  PDictRecord = ^TDictRecord;

  TDictArray = array[0..MaxListSize - 1] of TDictRecord;
  PDictArray = ^TDictArray;

  {This list works like simple dictionary. Each string Key has
   associated another string Word. Sorted is set to True for
   TNDictionary, duplicates are accepted.}
  TNDictionary = class(TNCustomStringKeys)
  protected
    function GetKey(Index: Long): NString; override;
    procedure PutKey(Index: Long; const Value: NString); override;
    function GetWord(Index: Long): NString; virtual;
    procedure PutWord(Index: Long; const Value: NString); virtual;
    procedure InsertInternal(Index: Long; const S: NString; const Value: NString);
    procedure FinalizeItem(Index: Long); override;
    function Compare(Index1, Index2: Long): Long; override;
    function Compare(Index: Long; const S: NString): Long; override;
    function GetMemoryUsage: Long; override;
    function GetLine(Index: Long): NString; override;
    procedure AddFromLine(const S: NString); override;
    procedure WriteData(Writer: TNAbstractWriter); override;
    procedure ReadData(Reader: TNAbstractReader); override;
  public
    constructor Create; override;
    {Adds new Key and Word strings to list.}
    function Add(const AKey, AWord: NString): Long;
    {Inserts new Key and Word strings to list}
    procedure Insert(Index: Long; const AKey, AWord: NString);
    {Returns index of given Word. If not found, -1 is returned.}
    function IndexOfWord(const AWord: NString): Long;
    property Words[Index: Long]: NString read GetWord write PutWord;
  end;

  {Standard TNDictionary which is set to Sorted in constructor.
   Duplicate items are not added to list.}
  TNSortedDictionary = class(TNDictionary)
  public
    constructor Create; override;
  end;


{*************************************************************************}
{                       TNCustomRefPointerList descendants                }
{*************************************************************************}


  TRefPtrRecord = record
    Ref: Pointer;
    Value: Pointer;
  end;
  PRefPtrRecord = ^TRefPtrRecord;

  TRefPtrArray = array[0..MaxListSize - 1] of TRefPtrRecord;
  PRefPtrArray = ^TRefPtrArray;

  {Base list for list where Key (Ref reference here)
   is not string directly but for example it is pointer to string.
   Loading of items is not supported.}
  TNCustomRefPointerList = class(TNCustomStringKeys)
  protected
    function GetRef(Index: Long): Pointer; virtual;
    procedure PutRef(Index: Long; Value: Pointer); virtual;
    function GetPointer(Index: Long): Pointer; virtual;
    procedure PutPointer(Index: Long; Value: Pointer); virtual;
  public
    constructor Create; override;
    {Overriden to do nothing.}
    procedure LoadFromStream(Stream: TNStream); override;
  end;

  {Base list for lists where Key is pointer to string and Value
   is Pointer.}
  TNCustomStringRefPointerList = class(TNCustomRefPointerList)
  protected
    function GetKey(Index: Long): NString; override;
    procedure PutKey(Index: Long; const Value: NString); override;
    function GetRef(Index: Long): PString; reintroduce;
    procedure PutRef(Index: Long; Value: PString); reintroduce;
    function Compare(Index1, Index2: Long): Long; override;
    function Compare(Index: Long; const S: NString): Long; override;
    function GetLine(Index: Long): NString; override;
  public
    {Adds new Ref and Value to list.}
    function Add(ARef: PString; Value: Pointer): Long;
    {Inserts new Ref and Value to list.}
    procedure Insert(Index: Long; ARef: PString; Value: Pointer);
    {Returns index of given Value. Returns -1 if item is not in list.}
    function IndexOfValue(Value: Pointer): Long;
    property Refs[Index: Long]: PString read GetRef write PutRef;
  end;

  {List where Ref is pointer to string and Value is Pointer.}
  TNStringRefPointerList = class(TNCustomStringRefPointerList)
  public
    property Pointers[Index: Long]: Pointer read GetPointer write PutPointer;
  end;

  {List where Ref is pointer to string and Value is TObject.}
  TNStringRefObjectList = class(TNCustomStringRefPointerList)
  protected
    function GetObject(Index: Long): TObject;
    procedure PutObject(Index: Long; Value: TObject);
  public
    property Objects[Index: Long]: TObject read GetObject write PutObject;
  end;

  {This is useful list whose Keys are values of objects' properties.
   Example: you have list of named objects. You can use something
   like Classes.TStringList or Objects.TNStringObjectList for
   faster access by object's name - List.Add(O.name, O).
   But what if object's name changes? You must manually
   update List's items - OR - use TNCustomPropObjectList.
   This list stores names of properties by which you want to
   access objects and uses actual values of these properties
   in searches. So you can solve above example
   by calling PropList.Add('Name',O) and when you call
   PropList.Find/IndexOf actual value of property named 'Name'
   will be used.
   Important: used properties must have RTTI (object is
    descendant of Object.TNPersistent or Classes.TPersistent). }
  TNCustomPropObjectList = class(TNCustomRefPointerList)
  protected
    function GetKey(Index: Long): NString; override;
    procedure PutKey(Index: Long; const Value: NString); override;
    function GetName(Index: Long): NString;
    procedure PutName(Index: Long; const Value: NString);
    function GetObject(Index: Long): TObject;
    procedure PutObject(Index: Long; Value: TObject);
    function Compare(Index1, Index2: Long): Long; override;
    function Compare(Index: Long; const S: NString): Long; override;
    function GetLine(Index: Long): NString; override;
  public
    {Adds new property name and object to list.}
    function Add(const PropName: NString; AObject: TObject): Long;
    {Inserts new property name and object to list.}
    procedure Insert(Index: Long; const PropName: NString; AObject: TObject);
    {Returns index of given object. Returns -1 if item is not in list.}
    function IndexOfObject(AObject: TObject): Long;
    property PropNames[Index: Long]: NString read GetName write PutName;
  end;

  {PropObjectList for TObject instances.}
  TNPropObjectList = class(TNCustomPropObjectList)
  public
    property Objects[Index: Long]: TObject read GetObject write PutObject;
  end;


{*************************************************************************}
{                    TNCustomStringPointerList descendants                }
{*************************************************************************}


  {List for storing Persistents associated to string Keys.}
  TNStringPersistentList = class(TNCustomStringPointerList)
  protected
    function GetObject(Index: Long): TNPersistent; virtual;
    procedure PutObject(Index: Long; Value: TNPersistent); virtual;
  public
    property Objects[Index: Long]: TNPersistent read GetObject write PutObject;
  end;


{*************************************************************************}
{                      TNCustomValueList descendants                      }
{*************************************************************************}


  {Value list for storing Int64 numbers.}
  TNInt64List = class(TNCustomValueList)
  protected
    function GetNumber(Index: Long): Int64;
    procedure PutNumber(Index: Long; const Value: Int64);
    {Writes items to nicer form than binary one.}
    procedure WriteData(Writer: TNAbstractWriter); override;
    {Reads items from nicer form than binary one.}
    procedure ReadData(Reader: TNAbstractReader); override;
  public
    constructor Create; override;
    {Adds new item to list.}
    function Add(const Item: Int64): Long;
    {Inserts new item to list.}
    procedure Insert(Index: Long; const Item: Int64);
    {Returns index of given number. Returns -1 if item is not in list.}
    function IndexOf(const Item: Int64): Long;
    property Items[Index: Long]: Int64 read GetNumber write PutNumber;
    default;
  end;

  {List for storing TObject instances.}
  TNObjectList = class(TNCustomPointerList)
  protected
    function GetObject(Index: Long): TObject;
    procedure PutObject(Index: Long; Value: TObject);
  public
    property Objects[Index: Long]: TObject read GetObject write PutObject;default;
  end;


{*************************************************************************}
{                      TNCustomOrderedList descendants                    }
{*************************************************************************}


  {Stack for storing TObject instances.}
  TNObjectStack = class(TNPointerStack)
  public
    function Pop: TObject;
    function Peek: TObject;
  end;

  {Queue for storing TObject instances.}
  TNObjectQueue = class(TNPointerQueue)
  public
    function Pop: TObject;
    function Peek: TObject;
  end;

  {Stack for storing strings.}
  TNStringStack = class(TNCustomStack)
  public
    constructor Create; override;
    function Pop: NString;
    function Peek: NString;
    procedure Push(const Item: NString);
  end;

  {Queue for storing strings.}
  TNStringQueue = class(TNCustomQueue)
  public
    constructor Create; override;
    function Pop: NString;
    function Peek: NString;
    procedure Push(const Item: NString);
  end;


{*************************************************************************}
{                             IMPLEMEMTATION                              }
{*************************************************************************}


implementation


{*************************************************************************}
{                      TNCustomStringValueList methods                    }
{*************************************************************************}


constructor TNCustomStringValueList.Create(AValueSize: Long);
begin
  inherited Create(SizeOf(TStrValueRecord));
  FValueSize := AValueSize;
end;

function TNCustomStringValueList.GetKey(Index: Long): NString;
begin
  if (Index > -1) and (Index < FCount) then
    Result := PStrValueArray(FData)[Index].Key
  else
    Result := '';
end;

procedure TNCustomStringValueList.PutKey(Index: Long; const Value: NString);
begin
  if (Index > -1) and (Index < FCount) and (not FSorted) then
  begin
    Notify(Index, lnKeyDeleted);
    PStrPtrArray(FData)[Index].Key := Value;
    Notify(Index, lnKeyAdded);
  end;
end;

procedure TNCustomStringValueList.InsertInternal(Index: Long; const AKey:
  NString; const Item);
var
  Rec: TStrValueRecord;
begin
  InsertItem(Index, Rec);
  with PStrValueArray(FData)[Index] do
  begin
    GetMem(Value, FValueSize);
    Move(Item, Value^, FValueSize);
    Key := AKey;
  end;
end;

procedure TNCustomStringValueList.FinalizeItem(Index: Long);
begin
  Finalize(PStrValueArray(FData)[Index], 1);
  FreeMem(PStrValueArray(FData)[Index].Value);
end;

function TNCustomStringValueList.Compare(Index1, Index2: Long): Long;
begin
  if FCaseSensitive then
    Result := Kernel.CompStrCase(PStrValueArray(FData)[Index1].Key,
      PStrValueArray(FData)[Index2].Key)
  else
    Result := Kernel.CompStr(PStrValueArray(FData)[Index1].Key,
      PStrValueArray(FData)[Index2].Key);
end;

function TNCustomStringValueList.Compare(Index: Long; const S: NString): Long;
begin
  if FCaseSensitive then
    Result := Kernel.CompStrCase(PStrValueArray(FData)[Index].Key, S)
  else
    Result := Kernel.CompStr(PStrValueArray(FData)[Index].Key, S);
end;

function TNCustomStringValueList.GetMemoryUsage: Long;
var
  I: Long;
begin
  Result := inherited GetMemoryUsage;
  for I := 0 to FCount - 1 do
    Result := Result + Length(PStrPtrArray(FData)[I].Key) * SizeOf(Char) +
      FValueSize;
end;

function TNCustomStringValueList.GetLine(Index: Long): NString;
begin
  Result := PStrValueArray(FData)[Index].Key + #09 +
    ValueToString(PStrValueArray(FData)[Index].Value^) + sLineBreak;
end;

procedure TNCustomStringValueList.AddFromLine(const S: NString);
var
  Key: NString;
  Value: NString;
  P: Long;
  ValItem: Pointer;
begin
  P := PosCase(#09, S);
  Key := Copy(S, 1, P - 1);
  Value := Copy(S, P + 1, Length(S) - P);
  GetMem(ValItem, FValueSize);
  StringToValue(Value, ValItem^);
  Add(Key, ValItem^);
  FreeMem(ValItem);
end;

function TNCustomStringValueList.ValueToString(const Value): NString;
begin
  Result := '';
end;

procedure TNCustomStringValueList.StringToValue(const S: NString; var Value);
begin
  ZeroMemory(@Value, FValueSize);
end;

function TNCustomStringValueList.Add(const AKey: NString; const Value): Long;
begin
  if not FSorted then
    Result := FCount
  else
    if Find(AKey, Result) then
      if FDuplicates = dupIgnore then Exit;
  InsertInternal(Result, AKey, Value);
  Notify(Result, lnAdded);
end;

procedure TNCustomStringValueList.Insert(Index: Long; const AKey: NString;
  const Value);
begin
  if (Index > -1) and (Index <= FCount) then
  begin
    if not FSorted then
      InsertInternal(Index, AKey, Value)
    else
      Add(AKey, Value);
  end;
end;

function TNCustomStringValueList.IndexOfValue(const Value): Long;
begin
  for Result := 0 to FCount - 1 do
    if CompareMem(@Value, PStrValueArray(FData)[Result].Value, FItemSize) then
      Exit;
  Result := -1;
end;


{*************************************************************************}
{                        TNStringInt64List methods                        }
{*************************************************************************}


constructor TNStringInt64List.Create;
begin
  inherited Create(SizeOf(Int64));
end;

function TNStringInt64List.GetNumber(Index: Long): Int64;
begin
  if (Index > -1) and (Index < FCount) then
    Result:=Int64(PStrValueArray(FData)[Index].Value^)
  else
    Result := 0;
end;

procedure TNStringInt64List.PutNumber(Index: Long; const Value: Int64);
begin
  if (Index > -1) and (Index < FCount) then
  begin
    Notify(Index, lnValueDeleted);
    Int64(PStrValueArray(FData)[Index].Value^):=Value;
    Notify(Index, lnValueAdded);
  end;
end;

function TNStringInt64List.ValueToString(const Value): NString;
begin
  Result := IntToStr(Int64(Value));
end;

procedure TNStringInt64List.StringToValue(const S: NString; var Value);
begin
  Int64(Value) := StrToInt(S);
end;

procedure TNStringInt64List.ReadData(Reader: TNAbstractReader);
var
  Number, I: Long;
  Val: Int64;
  K: NString;
begin
  Number := Reader.ReadInt;
  for I := 0 to Number - 1 do
  begin
    K := Reader.ReadIdent;
    Val := Reader.ReadInt;
    Add(K, Val);
  end;
end;

procedure TNStringInt64List.WriteData(Writer: TNAbstractWriter);
var
  I: Long;
begin
  Writer.WriteInt(FCount);
  for I := 0 to FCount - 1 do
  begin
    Writer.WriteIdent(GetKey(I));
    Writer.WriteInt(GetNumber(I));
  end;
end;

function TNStringInt64List.Add(const AKey: NString; const Value: Int64): Long;
begin
  Result := inherited Add(AKey, Value);
end;

function TNStringInt64List.IndexOfValue(const Value: Int64): Long;
begin
  for Result:=0 to FCount-1 do
    if Value = Int64(PStrValueArray(FData)[Result].Value^) then Exit;
  Result := -1;
end;

procedure TNStringInt64List.Insert(Index: Long; const AKey: NString; const Value:
  Int64);
begin
  inherited Insert(Index, AKey, Value);
end;


{*************************************************************************}
{                         TDictionary methods                             }
{*************************************************************************}


constructor TNDictionary.Create;
begin
  inherited Create(SizeOf(TDictRecord));
  FSorted := True;
  FDuplicates := dupAccept;
end;

function TNDictionary.GetKey(Index: Long): NString;
begin
  if (Index > -1) and (Index < FCount) then
    Result := PDictArray(FData)[Index].Key
  else
    Result := '';
end;

procedure TNDictionary.PutKey(Index: Long; const Value: NString);
begin
  if (Index > -1) and (Index < FCount) and (not FSorted) then
  begin
    Notify(Index, lnKeyDeleted);
    PDictArray(FData)[Index].Key := Value;
    Notify(Index, lnKeyAdded);
  end;
end;

function TNDictionary.GetWord(Index: Long): NString;
begin
  if (Index > -1) and (Index < FCount) then
    Result := PDictArray(FData)[Index].Word
  else
    Result := '';
end;

procedure TNDictionary.PutWord(Index: Long; const Value: NString);
begin
  if (Index > -1) and (Index < FCount) then
  begin
    Notify(Index, lnValueDeleted);
    PDictArray(FData)[Index].Word := Value;
    Notify(Index, lnValueAdded);
  end;
end;

procedure TNDictionary.InsertInternal(Index: Long; const S, Value: NString);
var
  Rec: TDictRecord;
begin
  InsertItem(Index, Rec);
  with PDictArray(FData)[Index] do
  begin
    Word := Value;
    Key := S;
  end;
end;

procedure TNDictionary.FinalizeItem(Index: Long);
begin
  Finalize(PDictArray(FData)[Index], 1);
end;

function TNDictionary.Compare(Index1, Index2: Long): Long;
begin
  if FCaseSensitive then
    Result := Kernel.CompStrCase(PDictArray(FData)[Index1].Key,
      PDictArray(FData)[Index2].Key)
  else
    Result := Kernel.CompStr(PDictArray(FData)[Index1].Key,
      PDictArray(FData)[Index2].Key);
end;

function TNDictionary.Compare(Index: Long; const S: NString): Long;
begin
  if FCaseSensitive then
    Result := Kernel.CompStrCase(PDictArray(FData)[Index].Key, S)
  else
    Result := Kernel.CompStr(PDictArray(FData)[Index].Key, S);
end;

function TNDictionary.GetMemoryUsage: Long;
var
  I: Long;
begin
  Result := inherited GetMemoryUsage;
  for I := 0 to FCount - 1 do
    Result := Result + Length(PDictArray(FData)[I].Key) * SizeOf(Char)
      + Length(PDictArray(FData)[I].Word) * SizeOf(Char);
end;

function TNDictionary.GetLine(Index: Long): NString;
begin
  Result := PDictArray(FData)[Index].Key + #09 + PDictArray(FData)[Index].Word +
    sLineBreak;
end;

procedure TNDictionary.AddFromLine(const S: NString);
var
  Key: NString;
  Word: NString;
  P: Long;
begin
  P := PosCase(#09, S);
  if P > 0 then
  begin
    Key := Copy(S, 1, P - 1);
    Word := Copy(S, P + 1, Length(S) - P);
    Add(Key, Word);
  end
  else
    Add(S, '');
end;

procedure TNDictionary.ReadData(Reader: TNAbstractReader);
var
  Number, I: Long;
  K, W: NString;
begin
  Number := Reader.ReadInt;
  for I := 0 to Number - 1 do
  begin
    K := Reader.ReadIdent;
    W := Reader.ReadString;
    Add(K, W);
  end;
end;

procedure TNDictionary.WriteData(Writer: TNAbstractWriter);
var
  I: Long;
begin
  Writer.WriteInt(FCount);
  for I := 0 to FCount - 1 do
  begin
    Writer.WriteIdent(GetKey(I));
    Writer.WriteString(GetWord(I));
  end;
end;

function TNDictionary.Add(const AKey, AWord: NString): Long;
begin
  if not FSorted then
    Result := FCount
  else
    if Find(AKey, Result) then
      if FDuplicates = dupIgnore then Exit;
  InsertInternal(Result, AKey, AWord);
  Notify(Result, lnAdded);
end;

procedure TNDictionary.Insert(Index: Long; const AKey, AWord: NString);
begin
  if (Index > -1) and (Index <= FCount) then
  begin
    if not FSorted then
      InsertInternal(Index, AKey, AWord)
    else
      Add(AKey, AWord);
  end;
end;

function TNDictionary.IndexOfWord(const AWord: NString): Long;
begin
  if FCaseSensitive then
  begin
    for Result := 0 to FCount - 1 do
      if Kernel.CompStrCase(AWord, PDictArray(FData)[Result].Word) = 0 then
        Exit;
  end
  else
    for Result := 0 to FCount - 1 do
      if Kernel.CompStr(AWord, PDictArray(FData)[Result].Word) = 0 then Exit;
  Result := -1;
end;


{*************************************************************************}
{                        TNSortedDictionary methods                   }
{*************************************************************************}


constructor TNSortedDictionary.Create;
begin
  inherited Create;
  FSorted := True;
  FDuplicates := dupIgnore;
  FCaseSensitive := False;
end;


{*************************************************************************}
{                        TNCustomRefPointerList methods                   }
{*************************************************************************}


constructor TNCustomRefPointerList.Create;
begin
  inherited Create(SizeOf(TRefPtrRecord));
end;

function TNCustomRefPointerList.GetRef(Index: Long): Pointer;
begin
  if (Index > -1) and (Index < FCount) then
    Result := PRefPtrArray(FData)[Index].Ref
  else
    Result := nil;
end;

procedure TNCustomRefPointerList.PutRef(Index: Long; Value: Pointer);
begin
  if (Index > -1) and (Index < FCount) and (not FSorted) then
  begin
    Notify(Index, lnKeyDeleted);
    PRefPtrArray(FData)[Index].Ref := Value;
    Notify(Index, lnKeyAdded);
  end;
end;

function TNCustomRefPointerList.GetPointer(Index: Long): Pointer;
begin
  if (Index > -1) and (Index < FCount) then
    Result := PRefPtrArray(FData)[Index].Value
  else
    Result := nil;
end;

procedure TNCustomRefPointerList.PutPointer(Index: Long; Value: Pointer);
begin
  if (Index > -1) and (Index < FCount) then
  begin
    Notify(Index, lnKeyDeleted);
    PRefPtrArray(FData)[Index].Value := Value;
    Notify(Index, lnKeyAdded);
  end;
end;

procedure TNCustomRefPointerList.LoadFromStream(Stream: TNStream);
begin
end;


{*************************************************************************}
{                    TNCustomStringRefObjectList methods                  }
{*************************************************************************}


function TNCustomStringRefPointerList.GetKey(Index: Long): NString;
begin
  if (Index > -1) and (Index < FCount) then
    Result := PString(PRefPtrArray(FData)[Index].Ref)^
  else
    Result := '';
end;

procedure TNCustomStringRefPointerList.PutKey(Index: Long; const Value: NString);
begin
  if (Index > -1) and (Index < FCount) and (not FSorted) then
  begin
    Notify(Index, lnKeyDeleted);
    PString(PRefPtrArray(FData)[Index].Ref)^ := Value;
    Notify(Index, lnKeyAdded);
  end;
end;

function TNCustomStringRefPointerList.GetRef(Index: Long): PString;
begin
  if (Index > -1) and (Index < FCount) then
    Result := PRefPtrArray(FData)[Index].Ref
  else
    Result := nil;
end;

procedure TNCustomStringRefPointerList.PutRef(Index: Long; Value: PString);
begin
  if (Index > -1) and (Index < FCount) then
  begin
    Notify(Index, lnKeyDeleted);
    PRefPtrArray(FData)[Index].Ref := Value;
    Notify(Index, lnKeyAdded);
  end;
end;

function TNCustomStringRefPointerList.Compare(Index1, Index2: Long): Long;
begin
  if FCaseSensitive then
    Result := Kernel.CompStrCase(PString(PRefPtrArray(FData)[Index1].Ref)^,
      PString(PRefPtrArray(FData)[Index2].Ref)^)
  else
    Result := Kernel.CompStr(PString(PRefPtrArray(FData)[Index1].Ref)^,
      PString(PRefPtrArray(FData)[Index2].Ref)^);
end;

function TNCustomStringRefPointerList.Compare(Index: Long;
  const S: NString): Long;
begin
  if FCaseSensitive then
    Result := Kernel.CompStrCase(PString(PRefPtrArray(FData)[Index].Ref)^, S)
  else
    Result := Kernel.CompStr(PString(PRefPtrArray(FData)[Index].Ref)^, S);
end;

function TNCustomStringRefPointerList.GetLine(Index: Long): NString;
begin
  Result := PString(PRefPtrArray(FData)[Index].Ref)^ + sLineBreak;
end;

function TNCustomStringRefPointerList.Add(ARef: PString; Value: Pointer): Long;
var
  Rec: TRefPtrRecord;
begin
  if not FSorted then
    Result := FCount
  else
    if Find(PString(ARef)^, Result) then
      if FDuplicates = dupIgnore then Exit;
  Rec.Ref := ARef;
  Rec.Value := Value;
  InsertItem(Result, Rec);
  Notify(Result, lnAdded);
end;

procedure TNCustomStringRefPointerList.Insert(Index: Long; ARef: PString;
  Value: Pointer);
var
  Rec: TRefPtrRecord;
begin
  if not FSorted then
  begin
    Rec.Ref := ARef;
    Rec.Value := Value;
    InsertItem(Index, Rec);
  end
  else
    Add(ARef, Value);
end;

function TNCustomStringRefPointerList.IndexOfValue(Value: Pointer): Long;
begin
  for Result := 0 to FCount - 1 do
    if PRefPtrArray(FData)[Result].Value = Value then Exit;
  Result := -1;
end;


{*************************************************************************}
{                       TNStringRefObjectList methods                     }
{*************************************************************************}


function TNStringRefObjectList.GetObject(Index: Long): TObject;
begin
  if (Index > -1) and (Index < FCount) then
    Result := PRefPtrArray(FData)[Index].Value
  else
    Result := nil;
end;

procedure TNStringRefObjectList.PutObject(Index: Long; Value: TObject);
begin
  if (Index > -1) and (Index < FCount) then
  begin
    Notify(Index, lnValueDeleted);
    PRefPtrArray(FData)[Index].Value := Value;
    Notify(Index, lnValueAdded);
  end;
end;


{*************************************************************************}
{                       TNCustomPropObjectList methods                    }
{*************************************************************************}


function TNCustomPropObjectList.GetKey(Index: Long): NString;
begin
  if (Index > -1) and (Index < FCount) then
    Result := GetStrProp(TObject(PRefPtrArray(FData)[Index].Value),
      PRefPtrArray(FData)[Index].Ref)
  else
    Result := '';
end;

procedure TNCustomPropObjectList.PutKey(Index: Long; const Value: NString);
begin
  if (Index > -1) and (Index < FCount) and (not FSorted)
    and (PPropInfo(PRefPtrArray(FData)[Index].Ref).SetProc <> nil) then
  begin
    Notify(Index, lnKeyDeleted);
    SetStrProp(TObject(PRefPtrArray(FData)[Index].Value),
      PRefPtrArray(FData)[Index].Ref, Value);
    Notify(Index, lnKeyAdded);
  end;
end;

function TNCustomPropObjectList.GetName(Index: Long): NString;
begin
  if (Index > -1) and (Index < FCount) then
    Result := PPropInfo(PRefPtrArray(FData)[Index].Ref).Name
  else
    Result := '';
end;

procedure TNCustomPropObjectList.PutName(Index: Long; const Value: NString);
begin
  if (Index > -1) and (Index < FCount)
  and (GetPropInfo(TObject(PRefPtrArray(FData)[Index].Value), Value) <> nil)then
  begin
    Notify(Index, lnKeyDeleted);
    PRefPtrArray(FData)[Index].Ref :=
      GetPropInfo(TObject(PRefPtrArray(FData)[Index].Value), Value);
    Notify(Index, lnKeyAdded);
  end;
end;

function TNCustomPropObjectList.GetObject(Index: Long): TObject;
begin
  if (Index > -1) and (Index < FCount) then
    Result := PRefPtrArray(FData)[Index].Value
  else
    Result := nil;
end;

procedure TNCustomPropObjectList.PutObject(Index: Long; Value: TObject);
begin
  if (Index > -1) and (Index < FCount) then
  begin
    Notify(Index, lnValueDeleted);
    PRefPtrArray(FData)[Index].Value := Value;
    Notify(Index, lnValueAdded);
  end;
end;

function TNCustomPropObjectList.Compare(Index1, Index2: Long): Long;
begin
  if FCaseSensitive then
    Result := Kernel.CompStrCase(
      GetStrProp(TObject(PRefPtrArray(FData)[Index1].Value),
      PRefPtrArray(FData)[Index1].Ref),
      GetStrProp(TObject(PRefPtrArray(FData)[Index2].Value),
      PRefPtrArray(FData)[Index2].Ref))
  else
    Result := Kernel.CompStr(
      GetStrProp(TObject(PRefPtrArray(FData)[Index1].Value),
      PRefPtrArray(FData)[Index1].Ref),
      GetStrProp(TObject(PRefPtrArray(FData)[Index2].Value),
      PRefPtrArray(FData)[Index2].Ref));
end;

function TNCustomPropObjectList.Compare(Index: Long; const S: NString): Long;
begin
  if FCaseSensitive then
    Result :=
      Kernel.CompStrCase(GetStrProp(TObject(PRefPtrArray(FData)[Index].Value),
      PRefPtrArray(FData)[Index].Ref), S)
  else
    Result :=
      Kernel.CompStr(GetStrProp(TObject(PRefPtrArray(FData)[Index].Value),
      PRefPtrArray(FData)[Index].Ref), S);
end;


function TNCustomPropObjectList.GetLine(Index: Long): NString;
begin
  Result := PPropInfo(PRefPtrArray(FData)[Index].Ref).Name + #9 +
    GetStrProp(TObject(PRefPtrArray(FData)[Index].Value),
    PRefPtrArray(FData)[Index].Ref) + sLineBreak;
end;

function TNCustomPropObjectList.Add(const PropName: NString; AObject: TObject):
  Long;
var
  Rec: TRefPtrRecord;
begin
  if not FSorted then
    Result := FCount
  else
    if Find(GetStrProp(AObject, PropName), Result) then
      if FDuplicates = dupIgnore then Exit;
  Rec.Ref := GetPropInfo(AObject, PropName);
  Rec.Value := AObject;
  if Rec.Ref <> nil then
  begin
    InsertItem(Result, Rec);
    Notify(Result, lnAdded);
  end;
end;

procedure TNCustomPropObjectList.Insert(Index: Long; const PropName: NString;
  AObject: TObject);
var
  Rec: TRefPtrRecord;
begin
  if not FSorted then
  begin
    Rec.Ref := GetPropInfo(AObject, PropName);
    Rec.Value := AObject;
    if Rec.Ref <> nil then
      InsertItem(Index, Rec);
  end
  else
    Add(PropName, AObject);
end;

function TNCustomPropObjectList.IndexOfObject(AObject: TObject): Long;
begin
  for Result := 0 to FCount - 1 do
    if PRefPtrArray(FData)[Result].Value = AObject then Exit;
  Result := -1;
end;


{*************************************************************************}
{                     TNStringPersistentList methods                      }
{*************************************************************************}


function TNStringPersistentList.GetObject(Index: Long): TNPersistent;
begin
  if (Index > -1) and (Index < FCount) then
    Result := PStrPtrArray(FData)[Index].Value
  else
    Result := nil;
end;

procedure TNStringPersistentList.PutObject(Index: Long;
  Value: TNPersistent);
begin
  if (Index > -1) and (Index < FCount) then
  begin
    Notify(Index, lnValueDeleted);
    PStrPtrArray(FData)[Index].Value := Value;
    Notify(Index, lnValueAdded);
  end;
end;


{*************************************************************************}
{                         TNInt64List methods                             }
{*************************************************************************}


constructor TNInt64List.Create;
begin
  inherited Create(SizeOf(Int64));
end;

function TNInt64List.GetNumber(Index: Long): Int64;
begin
  if (Index > -1) and (Index < FCount) then
    Result := PInt64Array(FData)[Index]
  else
    Result := 0;
end;

procedure TNInt64List.PutNumber(Index: Long; const Value: Int64);
begin
  if (Index > -1) and (Index < FCount) then
  begin
    Notify(Index, lnDeleted);
    PInt64Array(FData)[Index] := Value;
    Notify(Index, lnAdded);
  end;
end;

procedure TNInt64List.ReadData(Reader: TNAbstractReader);
var
  I, Number: Long;
begin
  Number := Reader.ReadInt;
  for I := 0 to Number - 1 do
    Add(Reader.ReadInt);
end;

procedure TNInt64List.WriteData(Writer: TNAbstractWriter);
var
  I: Long;
begin
  Writer.WriteInt(FCount);
  for I := 0 to FCount - 1 do
    Writer.WriteInt(PInt64Array(FData)[I]);
end;

function TNInt64List.Add(const Item: Int64): Long;
begin
  Result := inherited Add(Item);
end;

procedure TNInt64List.Insert(Index: Long; const Item: Int64);
begin
  inherited Insert(Index, Item);
end;

function TNInt64List.IndexOf(const Item: Int64): Long;
begin
  for Result:=0 to FCount-1 do
    if Item = PInt64Array(FData)[Result] then Exit;
  Result := -1;
end;


{*************************************************************************}
{                         TNObjectList methods                            }
{*************************************************************************}


function TNObjectList.GetObject(Index: Long): TObject;
begin
   if (Index > -1) and (Index < FCount) then
    Result := TObject(PPointerArray(FData)[Index])
  else
    Result := nil;
end;

procedure TNObjectList.PutObject(Index: Long; Value: TObject);
begin
  if (Index > -1) and (Index < FCount) then
  begin
    Notify(Index, lnDeleted);
    PPointerArray(FData)[Index] := Value;
    Notify(Index, lnAdded);
  end;
end;


{*************************************************************************}
{                        TNObjectStack methods                            }
{*************************************************************************}


function TNObjectStack.Pop: TObject;
begin
  PopItem(Result);
end;

function TNObjectStack.Peek: TObject;
begin
  PeekItem(Result);
end;


{*************************************************************************}
{                        TNObjectQueue methods                            }
{*************************************************************************}


function TNObjectQueue.Pop: TObject;
begin
  PopItem(Result);
end;

function TNObjectQueue.Peek: TObject;
begin
  PeekItem(Result);
end;


{*************************************************************************}
{                         TNStringStack methods                           }
{*************************************************************************}


constructor TNStringStack.Create;
begin
  inherited Create(SizeOf(TStringRecord));
end;

procedure TNStringStack.Push(const Item: NString);
var
  Rec: TStringRecord;
begin
  FList.Insert(FList.Count, Rec);
  PStrRecArray(FList.Data)[FList.Count - 1].Key := Item;
end;

function TNStringStack.Pop: NString;
var
  Rec: TStringRecord;
begin
  PopItem(Rec);
  Result := Rec.Key;
end;

function TNStringStack.Peek: NString;
var
  Rec: TStringRecord;
begin
  PeekItem(Rec);
  Result := Rec.Key;
end;


{*************************************************************************}
{                         TNStringQueue methods                           }
{*************************************************************************}


constructor TNStringQueue.Create;
begin
  inherited Create(SizeOf(TStringRecord));
end;

procedure TNStringQueue.Push(const Item: NString);
var
  Rec: TStringRecord;
begin
  FList.Insert(0, Rec);
  PStrRecArray(FList.Data)[0].Key := Item;
end;

function TNStringQueue.Pop: NString;
var
  Rec: TStringRecord;
begin
  PopItem(Rec);
  Result := Rec.Key;
end;

function TNStringQueue.Peek: NString;
var
  Rec: TStringRecord;
begin
  PeekItem(Rec);
  Result := Rec.Key;
end;


{*************************************************************************}
{                         Internal stuff                                  }
{*************************************************************************}


initialization

  RegisterClasses([TNDictionary, TNInt64List, TNObjectList,
   TNStringInt64List]);

end.

