{*************************************************************************}
{*   Maths                                                               *}
{*   Copyright (C) 2004 Marek Mauder, Vampyre SoftWorks                  *}
{*   Engine\Kernel\Maths.pas                                             *}
{*************************************************************************}
{                                                                         }
{ This program is free software; you can redistribute it and/or modify it }
{ under the terms of the GNU General Public License as published by the   }
{ Free Software Foundation; either version 2 of the License, or (at your  }
{ option) any later version. This program is distributed in the hope that }
{ it will be useful, but WITHOUT ANY WARRANTY; without even the implied   }
{ warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See   }
{ the GNU General Public License for more details.                        }
{                                                                         }
{ You should have received a copy of the GNU General Public License       }
{ along with this program; if not, write to the Free Software Foundation, }
{ Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.         }
{                                                                         }
{*************************************************************************}
{  Created On: 22th January 2004  (22.01.2004)                            }
{*************************************************************************}

{This unit contains some mathematical functions and classes for working
 with numbers, verctors, matrices, quaternions, ...}

unit Maths;

{TODO -oGalfar : Add Matrix2x2, plane, quaternion things.}

interface

{$I Compiler.inc}

uses
  Kernel, Objects;


{*************************************************************************}
{                         Types and constants                             }
{*************************************************************************}


type

  {(*}      // do not proccess by Source Formatter

  {Record representing 2-dimensional float vector.}
  TVector2 = record
    case Long of
      0: (X: Float;
          Y: Float);
      1: (Arr: array[0..1] of Float);
  end;
  PVector2 = ^TVector2;

  {Record representing 3-dimensional float vector.}
  TVector3 = record
    case Long of
      0: (X: Float;
          Y: Float;
          Z: Float);
      1: (Arr: array[0..2] of Float);
  end;
  PVector3 = ^TVector3;

  {Record representing 4-dimensional float vector.}
  TVector4 = record
    case Long of
      0: (X: Float;
          Y: Float;
          Z: Float;
          W: Float);
      1: (Arr: array[0..3] of Float);
  end;
  PVector4 = TVector4;

  {Record representing 3x3 float matrix.}
  TMatrix3x3 = record
    case Long of
      0: (Item11, Item12, Item13: Float;
          Item21, Item22, Item23: Float;
          Item31, Item32, Item33: Float);
      1: (Arr: array[0..2, 0..2] of Float);
      2: (Floats: array[0..8] of Float);
  end;
  PMatrix3x3 = ^TMatrix3x3;

  {Record representing 4x4 float matrix.}
  TMatrix4x4 = record
    case Long of
      0: (Item11, Item12, Item13, Item14: Float;
          Item21, Item22, Item23, Item24: Float;
          Item31, Item32, Item33, Item34: Float;
          Item41, Item42, Item43, Item44: Float);
      1: (Arr: array[0..3, 0..3] of Float);
      2: (Floats: array[0..15] of Float);
  end;
  PMatrix4x4 = ^TMatrix4x4;

  {Record representing 3-dimensional float plane.}
  TPlane = record
    case Long of
      0: (A: Float;
          B: Float;
          C: Float;
          D: Float);
      1: (Arr: array[0..3] of Float);
  end;
  PPlane = ^TPlane;

  {Record representing quaternion.}
  TQuaternion = record
    case Long of
      0: (X: Float;
          Y: Float;
          Z: Float;
          W: Float);
      1: (Imag: TVector3;
          Real: Float);
      2: (Vector: TVector4);
      3: (Arr: array[0..3] of Float);
  end;
  PQuaternion = ^TQuaternion;

const

  {Various mathematical constants}
  PI:        Float = 3.1415926535897932384626433832795; // PI
  PIdiv2:    Float = 1.5707963267948966192313216916398; // PI / 2
  PIdiv3:    Float = 1.0471975511965977461542144610932; // PI / 3
  PIdiv4:    Float = 0.7853981633974483096156608458198; // PI / 4
  PIdiv180:  Float = 0.0174532925199432957692369076848; // PI / 180
  D180divPI: Float = 57.295779513082320876798154814105; // 180 / PI
  Sqrt2:     Float = 1.4142135623730950488016887242097; // Sqrt(2)
  Sqrt3:     Float = 1.7320508075688772935274463415059; // Sqrt(3)
  Sqrt5:     Float = 2.2360679774997896964091736687313; // Sqrt(5)
  Sqrt10:    Float = 3.1622776601683793319988935444327; // Sqrt(10)
  SqrtPI:    Float = 1.7724538509055160272981674833411; // Sqrt(PI)
  Sqrt2PI:   Float = 2.5066282746310005024157652848110; // Sqrt(2 * PI)
  PI2:       Float = 6.2831853071795864769252867665590; // 2 * PI
  PI3:       Float = 9.4247779607693797153879301498385; // 3 * PI
  Ln2:       Float = 0.6931471805599453094172321214581; // Ln(2)
  Ln10:      Float = 2.3025850929940456840179914546844; // Ln(10)
  LnPI:      Float = 1.1447298858494001741434273513531; // Ln(PI)
  LogPI:     Float = 0.4971498726941338543512682882909; // Log10(PI)
  LogE:      Float = 0.4342944819032518276511289189166; // Log10(E)
  E:         Float = 2.7182818284590452353602874713527; // Natural constant
  One:       Float = 1.0000000000000000000000000000000; // Sqrt(11-10)
  Intfinity: Float = 1 / 0;                             // Infinity
  Nan:       Float = 0 / 0;                             // Not a number

  {*)}

  {Empty 2-D vector.}
  ZeroVector2: TVector2 = (X: 0; Y: 0);
  {Empty 3-D vector.}
  ZeroVector3: TVector3 = (X: 0; Y: 0; Z: 0);
  {Empty 4-D vector.}
  ZeroVector4: TVector4 = (X: 0; Y: 0; Z: 0; W: 0);
  {Empty 3x3 matrix.}
  ZeroMatrix3x3: TMatrix3x3 = (Floats: (0, 0, 0, 0, 0, 0, 0, 0, 0));
  {Empty 4x4 matrix.}
  ZeroMatrix4x4: TMatrix4x4 = (Floats: (0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0));
  {Empty 4-D vector.}
  ZeroQuaternion: TQuaternion = (X: 0; Y: 0; Z: 0; W: 0);

  {Origin of 2-D coordinate system.}
  OriginVector2: TVector2 = (X: 0; Y: 0);
  {X Axis 2-D unit vector.}
  XVector2: TVector2 = (X: 1; Y: 0);
  {Y Axis 2-D unit vector.}
  YVector2: TVector2 = (X: 0; Y: 1);

  {Origin of 3-D coordinate system.}
  OriginVector3: TVector3 = (X: 0; Y: 0; Z: 0);
  {X Axis 3-D unit vector.}
  XVector3: TVector3 = (X: 1; Y: 0; Z: 0);
  {Y Axis 3-D unit vector.}
  YVector3: TVector3 = (X: 0; Y: 1; Z: 0);
  {Z Axis 3-D unit vector.}
  ZVector3: TVector3 = (X: 0; Y: 0; Z: 1);

  {Origin of 4-D coordinate system.}
  OriginVector4: TVector4 = (X: 0; Y: 0; Z: 0; W: 0);
  {X Axis 4-D vector.}
  XVector4: TVector4 = (X: 1; Y: 0; Z: 0; W: 0);
  {Y Axis 4-D vector.}
  YVector4: TVector4 = (X: 0; Y: 1; Z: 0; W: 0);
  {Z Axis 4-D vector.}
  ZVector4: TVector4 = (X: 0; Y: 0; Z: 1; W: 0);
  {W 4-D vector.}
  WVector4: TVector4 = (X: 0; Y: 0; Z: 0; W: 1);

  {3x3 identity matrix.}
  Identity3x3: TMatrix3x3 = (Arr: ((1, 0, 0), (0, 1, 0), (0, 0, 1)));
  {4x4 identity matrix.}
  Identity4x4: TMatrix4x4 = (Arr: ((1, 0, 0, 0), (0, 1, 0, 0), (0, 0, 1, 0),
    (0, 0, 0, 1)));
  {Identity quaternion.}
  IdentityQuaternion: TQuaternion = (Imag: (Arr: (0, 0, 0)); Real: 1);


{*************************************************************************}
{                         Base functions                                  }
{*************************************************************************}


{Returns Min value from array.}
function MinFrom(const Values: array of Long): Long; overload;
function MinFrom(const Values: array of Float): Float; overload;
{Returns Max value from array.}
function MaxFrom(const Values: array of Long): Long; overload;
function MaxFrom(const Values: array of Float): Float; overload;

{Returns True if value is in range (inclusive).}
function InRange(const Value, Min, Max: Long): Boolean; overload;
function InRange(const Value, Min, Max: Float): Boolean; overload;

{Returns sum of values.}
function SumI(const Values: array of Long): Long;
function SumF(const Values: array of Float): Float;
{Returns sum of squares of values.}
function SumSquaresI(const Values: array of Long): Long;
function SumSquaresF(const Values: array of Float): Float;

{Returns float division remainder.}
function ModF(const A, B: Float): Float;
{Linearly interpolates between two float values A and B at time T.}
function Lerp(const A, B, T: Float): Float;


{*************************************************************************}
{                      Random number functions                            }
{*************************************************************************}


{Returns long Min <= X <= Max}
function RandRange(const Min, Max: Long): Long; overload;
function RandRange(const Min, Max: Float): Float; overload;
{Returns value with random sign + / -}
function RandSign(const Value: Long): Long; overload;
function RandSign(const Value: Float): Float; overload;
{Returns random value from array}
function RandFromI(const Values: array of Long): Long;
function RandFromF(const Values: array of Float): Float;


{*************************************************************************}
{                Logarithmical/Exponential functions                      }
{*************************************************************************}


{Returns Log base 2 of X.}
function Log2(const X: Float): Float;
{Returns Log base 10 of X.}
function Log10(const X: Float): Float;
{Returns Log base Base of X.}
function LogN(const X, Base: Float): Float;

{Raise two to Exp.}
function Power2(const Exponent: Float): Float;
{Raise ten to Exp.}
function Power10(const Exponent: Float): Float;
{Raise base to any power.}
function Power(const Base, Exponent: Float): Float;


{*************************************************************************}
{                    Trigonometric/Angle functions                        }
{*************************************************************************}


{Adjusts angle to fit in [-PI, PI].}
function AdjustAngle(const Angle: Float): Float;

{Converts degrees to radians.}
function DegToRad(const Degrees: Float): Float;
{Converts radians to degrees.}
function RadToDeg(const Radians: Float): Float;

{Returns sine and cosine of theta angle (in radians)}
procedure SinCos(const Angle: Float; var SinF, CosF: Double); overload;
procedure SinCos(const Angle: Float; var SinF, CosF: Single); overload;

{Returns tangent of angle.}
function Tan(const Angle: Float): Float;
{Returns cotangent of angle.}
function CoTan(const Angle: Float): Float;

{Returns inverse sine (in radians) of Value (between -1 and 1).}
function ArcSin(const Value: Float): Float;
{Returns inverse cosine (in radians) of Value (between -1 and 1).}
function ArcCos(const Value: Float): Float;
{Returns inverse tangent (in radians) of Value.}
function ArcTan(const Value: Float): Float;
{Returns inverse cotangent (in radians) of Value.}
function ArcCoTan(const Value: Float): Float;


{*************************************************************************}
{                            Vector functions                             }
{*************************************************************************}


{Vector creation procedures.}
procedure Vector2(const X, Y: Float; var Vec: TVector2); overload;
procedure Vector3(const X, Y, Z: Float; var Vec: TVector3); overload;
procedure Vector4(const X, Y, Z, W: Float; var Vec: TVector4); overload;
{Vector creation functions.}
function Vector2(const X, Y: Float): TVector2; overload;
function Vector3(const X, Y, Z: Float): TVector3; overload;
function Vector4(const X, Y, Z, W: Float): TVector4; overload;

{* 2D Vector functions *}

{Adds vector V1 to V2.}
function Vec2Add(const V1, V2: TVector2): TVector2;
{Subtracts vector V2 from V1.}
function Vec2Subtract(const V1, V2: TVector2): TVector2;
{Multiplies vector V1 with V2.}
function Vec2Multiply(const V1, V2: TVector2): TVector2;
{Divides vector V1 by V2.}
function Vec2Divide(const V1, V2: TVector2): TVector2;
{Returns True if vectors have the same components.}
function Vec2Comp(const V1, V2: TVector2): Boolean;

{Returns dot product of two vectors V1 and V2.}
function Vec2Dot(const V1, V2: TVector2): Float;
{Returns length of vector V.}
function Vec2Length(const V: TVector2): Float;
{Returns distance between vectors V1 and V2.}
function Vec2Dist(const V1, V2: TVector2): Float;
{Calculates linear interpolation between vectors V1 and V2 at poin T (0.0 - 1.0)}
function Vec2Lerp(const V1, V2: TVector2; const T: Float): TVector2;
{Returns vector that is made up of the largest components of vectors V1 and V2.}
function Vec2Maximize(const V1, V2: TVector2): TVector2;
{Returns vector that is made up of the smallest components of vectors V1 and V2.}
function Vec2Minimize(const V1, V2: TVector2): TVector2;

{Scales vector V by scalar ScaleBy.}
procedure Vec2ScaleP(var V: TVector2; const ScaleBy: Float);
function Vec2Scale(const V: TVector2; const ScaleBy: Float): TVector2;
{Negates given vector V (Returns (-1, 2, -3) from input (1, -2, 3)). }
procedure Vec2NegateP(var V: TVector2);
function Vec2Negate(const V: TVector2): TVector2;
{Normalizes given vector V (Result will be unit length vector).}
procedure Vec2NormalizeP(var V: TVector2);
function Vec2Normalize(const V: TVector2): TVector2;

{* 3D Vector functions *}

{Adds vector V1 to V2.}
function Vec3Add(const V1, V2: TVector3): TVector3;
{Subtracts vector V2 from V1.}
function Vec3Subtract(const V1, V2: TVector3): TVector3;
{Multiplies vector V1 with V2.}
function Vec3Multiply(const V1, V2: TVector3): TVector3;
{Divides vector V1 by V2.}
function Vec3Divide(const V1, V2: TVector3): TVector3;
{Returns True if vectors have the same components.}
function Vec3Comp(const V1, V2: TVector3): Boolean;

{Returns dot product of two vectors V1 and V2.}
function Vec3Dot(const V1, V2: TVector3): Float;
{Returns cross product of two vectors V1 and V2.}
function Vec3Cross(const V1, V2: TVector3): TVector3;
{Returns length of vector V.}
function Vec3Length(const V: TVector3): Float;
{Returns distance between vectors V1 and V2.}
function Vec3Dist(const V1, V2: TVector3): Float;
{Calculates linear interpolation between vectors V1 and V2 at poin T (0.0 - 1.0)}
function Vec3Lerp(const V1, V2: TVector3; const T: Float): TVector3;
{Returns vector that is made up of the largest components of vectors V1 and V2.}
function Vec3Maximize(const V1, V2: TVector3): TVector3;
{Returns vector that is made up of the smallest components of vectors V1 and V2.}
function Vec3Minimize(const V1, V2: TVector3): TVector3;

{Scales vector V by scalar ScaleBy.}
procedure Vec3ScaleP(var V: TVector3; const ScaleBy: Float);
function Vec3Scale(const V: TVector3; const ScaleBy: Float): TVector3;
{Negates given vector V (Returns (-1, 2, -3) from input (1, -2, 3)). }
procedure Vec3NegateP(var V: TVector3);
function Vec3Negate(const V: TVector3): TVector3;
{Normalizes given vector V (Result will be unit length vector).}
procedure Vec3NormalizeP(var V: TVector3);
function Vec3Normalize(const V: TVector3): TVector3;

{* 4D Vector functions *}

{Adds vector V1 to V2.}
function Vec4Add(const V1, V2: TVector4): TVector4;
{Subtracts vector V2 from V1.}
function Vec4Subtract(const V1, V2: TVector4): TVector4;
{Multiplies vector V1 with V2.}
function Vec4Multiply(const V1, V2: TVector4): TVector4;
{Divides vector V1 by V2.}
function Vec4Divide(const V1, V2: TVector4): TVector4;
{Returns True if vectors have the same components.}
function Vec4Comp(const V1, V2: TVector4): Boolean;

{Returns dot product of two vectors V1 and V2.}
function Vec4Dot(const V1, V2: TVector4): Float;
{Returns cross product of two vectors V1 and V2.}
function Vec4Cross(const V1, V2: TVector4): TVector4;
{Returns length of vector V.}
function Vec4Length(const V: TVector4): Float;
{Returns distance between vectors V1 and V2.}
function Vec4Dist(const V1, V2: TVector4): Float;
{Calculates linear interpolation between vectors V1 and V2 at poin T (0.0 - 1.0)}
function Vec4Lerp(const V1, V2: TVector4; const T: Float): TVector4;
{Returns vector that is made up of the largest components of vectors V1 and V2.}
function Vec4Maximize(const V1, V2: TVector4): TVector4;
{Returns vector that is made up of the smallest components of vectors V1 and V2.}
function Vec4Minimize(const V1, V2: TVector4): TVector4;

{Scales vector V by scalar ScaleBy.}
procedure Vec4ScaleP(var V: TVector4; const ScaleBy: Float);
function Vec4Scale(const V: TVector4; const ScaleBy: Float): TVector4;
{Negates given vector V (Returns (-1, 2, -3) from input (1, -2, 3)). }
procedure Vec4NegateP(var V: TVector4);
function Vec4Negate(const V: TVector4): TVector4;
{Normalizes given vector V (Result will be unit length vector).}
procedure Vec4NormalizeP(var V: TVector4);
function Vec4Normalize(const V: TVector4): TVector4;


{*************************************************************************}
{                            Matrix functions                             }
{*************************************************************************}


{Matrix creation procedures.}
procedure Matrix3x3(const Floats: array of Float; var M: TMatrix3x3); overload;
procedure Matrix3x3(const Row1: array of Float; const Row2: array of Float;
  const Row3: array of Float; var M: TMatrix3x3); overload;
procedure Matrix4x4(const Floats: array of Float; var M: TMatrix4x4); overload;
procedure Matrix4x4(const Row1: array of Float; const Row2: array of Float;
  const Row3: array of Float; const Row4: array of Float; var M: TMatrix4x4);
overload;
{Matrix creation functions.}
function Matrix3x3(const Floats: array of Float): TMatrix3x3; overload;
function Matrix3x3(const Row1: array of Float; const Row2: array of Float;
  const Row3: array of Float): TMatrix3x3; overload;
function Matrix4x4(const Floats: array of Float): TMatrix4x4; overload;
function Matrix4x4(const Row1: array of Float; const Row2: array of Float;
  const Row3: array of Float; const Row4: array of Float): TMatrix4x4; overload;

{* 3x3 Matrix functions *}

{Adds 3x3 matrix M1 to M2.}
function Mat3Add(const M1, M2: TMatrix3x3): TMatrix3x3;
{Subtracts 3x3 matrix M2 form M1.}
function Mat3Subtract(const M1, M2: TMatrix3x3): TMatrix3x3;
{Multiplies 3x3 matrices M1 and M2.}
function Mat3Multiply(const M1, M2: TMatrix3x3): TMatrix3x3;
{Scales all components of 3x3 matrix M by F.}
function Mat3ScalarMultiply(const M: TMatrix3x3; const F: Float): TMatrix3x3;
{Returns True if 3x3 matrices M1 and M2 have the same components.}
function Mat3Comp(const M1, M2: TMatrix3x3): Boolean;

{Transposes 3x3 matrix M.}
function Mat3Transpose(const M: TMatrix3x3): TMatrix3x3;
{Returns determinant of 3x3 matrix M.}
function Mat3Determinant(const M: TMatrix3x3): Float;
{Returns inverted 3x3 matrix M.}
function Mat3Inverse(const M: TMatrix3x3): TMatrix3x3;

{* 4x4 Matrix functions *}

{Adds 4x4 matrix M1 to M2.}
function Mat4Add(const M1, M2: TMatrix4x4): TMatrix4x4;
{Subtracts 4x4 matrix M2 form M1.}
function Mat4Subtract(const M1, M2: TMatrix4x4): TMatrix4x4;
{Multiplies 4x4 matrices M1 and M2.}
function Mat4Multiply(const M1, M2: TMatrix4x4): TMatrix4x4;
{Scales all components of 4x4 matrix M by F.}
function Mat4ScalarMultiply(const M: TMatrix4x4; const F: Float): TMatrix4x4;
{Returns True if 4x4 matrices M1 and M2 have the same components.}
function Mat4Comp(const M1, M2: TMatrix4x4): Boolean;

{Transposes 4x4 matrix M.}
function Mat4Transpose(const M: TMatrix4x4): TMatrix4x4;
{Returns determinant of 4x4 matrix M.}
function Mat4Determinant(const M: TMatrix4x4): Float;
{Returns inverted 4x4 matrix M.}
function Mat4Inverse(const M: TMatrix4x4): TMatrix4x4;


{*************************************************************************}
{                              Classes                                    }
{*************************************************************************}


type

  {Stack for storing 3x3 matrices.}
  TNMatrix3x3Stack = class(TNCustomStack)
  public
    constructor Create; override;
    procedure Push(const Mat: TMatrix3x3);
    procedure Pop(var Mat: TMatrix3x3);
    procedure Peek(var Mat: TMatrix3x3);
  end;

  {Stack for storing 4x4 matrices.}
  TNMatrix4x4Stack = class(TNCustomStack)
  public
    constructor Create; override;
    procedure Push(const Mat: TMatrix4x4);
    procedure Pop(var Mat: TMatrix4x4);
    procedure Peek(var Mat: TMatrix4x4);
  end;


{*************************************************************************}
{                             IMPLEMEMTATION                              }
{*************************************************************************}


implementation


{*************************************************************************}
{                         Base functions                                  }
{*************************************************************************}


function MinFrom(const Values: array of Long): Long;
var
  I: Long;
begin
  Result := Values[Low(Values)];
  for I := Low(Values) + 1 to High(Values) do
    if Result > Values[I] then
      Result := Values[I];
end;

function MinFrom(const Values: array of Float): Float;
var
  I: Long;
begin
  Result := Values[Low(Values)];
  for I := Low(Values) + 1 to High(Values) do
    if Result > Values[I] then
      Result := Values[I];
end;

function MaxFrom(const Values: array of Long): Long;
var
  I: Long;
begin
  Result := Values[Low(Values)];
  for I := Low(Values) + 1 to High(Values) do
    if Result < Values[I] then
      Result := Values[I];
end;

function MaxFrom(const Values: array of Float): Float;
var
  I: Long;
begin
  Result := Values[Low(Values)];
  for I := Low(Values) + 1 to High(Values) do
    if Result < Values[I] then
      Result := Values[I];
end;

function InRange(const Value, Min, Max: Long): Boolean;
begin
  Result := (Value >= Min) and (Value <= Max);
end;

function InRange(const Value, Min, Max: FLoat): Boolean;
begin
  Result := (Value >= Min) and (Value <= Max);
end;

function SumI(const Values: array of Long): Long;
var
  I: Long;
begin
  Result := 0;
  for I := Low(Values) to High(Values) do
    Result := Result + Values[I];
end;

function SumF(const Values: array of Float): Float;
var
  I: Long;
begin
  Result := 0.0;
  for I := Low(Values) to High(Values) do
    Result := Result + Values[I];
end;

function SumSquaresI(const Values: array of Long): Long;
var
  I: Long;
begin
  Result := 0;
  for I := Low(Values) to High(Values) do
    Result := Result + Sqr(Values[I]);
end;

function SumSquaresF(const Values: array of Float): Float;
var
  I: Long;
begin
  Result := 0.0;
  for I := Low(Values) to High(Values) do
    Result := Result + Sqr(Values[I]);
end;

function ModF(const A, B: Float): Float;
var
  Z: Float;
begin
  Result := A / B;
  Z := Int(Result);
  if Result < 0.0 then
    Z := Z - 1.0;
  Result := A - Z * B;
end;

function Lerp(const A, B, T: Float): Float;
begin
  Result := A + (B - A) * T;
end;


{*************************************************************************}
{                      Random number functions                            }
{*************************************************************************}


function RandRange(const Min, Max: Long): Long;
begin
  if Min > Max then
    Result := Random(Min - Max + 1) + Max
  else
    Result := Random(Max - Min + 1) + Min;
end;

function RandRange(const Min, Max: Float): Float;
begin
  if Min > Max then
    Result := Random(Round(Min - Max) + 1) + Max
  else
    Result := Random(Round(Max - Min) + 1) + Min;
end;

function RandSign(const Value: Long): Long;
begin
  if Random(2) = 0 then
    Result := Value
  else
    Result := -Value;
end;

function RandSign(const Value: Float): Float;
begin
  if Random(2) = 0 then
    Result := Value
  else
    Result := -Value;
end;

function RandFromI(const Values: array of Long): Long;
begin
  Result := Values[RandRange(Low(Values), High(Values))];
end;

function RandFromF(const Values: array of Float): Float;
begin
  Result := Values[RandRange(Low(Values), High(Values))];
end;


{*************************************************************************}
{                   Logarithmical/Exponential functions                   }
{*************************************************************************}


function Log2(const X: Float): Float;
begin
  Result := Ln(X) / Ln2;
end;

function Log10(const X: Float): Float;
begin
  Result := Ln(X) / Ln10;
end;

function LogN(const X, Base: Float): Float;
begin
  Result := Ln(X) / Ln(Base);
end;

function Power2(const Exponent: Float): Float;
begin
  if Exponent = 0.0 then
    Result := 1.0
  else
    Result := Exp(Exponent * Ln2);
end;

function Power10(const Exponent: Float): Float;
begin
  if Exponent = 0.0 then
    Result := 1.0
  else
    Result := Exp(Exponent * Ln10);
end;

function Power(const Base, Exponent: Float): Float;
begin
  if Exponent = 0.0 then
    Result := 1.0
  else
    if (Base = 0.0) and (Exponent > 0.0) then
      Result := 0.0
    else
      Result := Exp(Exponent * Ln(Base))
end;


{*************************************************************************}
{                      Trigonometric/Angle functions                      }
{*************************************************************************}


function AdjustAngle(const Angle: Float): Float;
begin
  Result := Angle;
  repeat
    if Result <= -PI then
      Result := Result + 2 * PI;
    if Result > PI then
      Result := Result - 2 * PI;
  until (Result > -PI) and (Result <= PI);
end;

function DegToRad(const Degrees: Float): Float;
{$IFDEF ASMVERSIONS}
{$IFNDEF PIC}
asm
  FLD    Degrees
  FMUL   PIdiv180
  FWAIT
end;
{$ELSE}
begin
  Result := Degrees * PIdiv180;
end;
{$ENDIF}
{$ELSE}
begin
  Result := Degrees * PIdiv180;
end;
{$ENDIF}

function RadToDeg(const Radians: Float): Float;
{$IFDEF ASMVERSIONS}
{$IFNDEF PIC}
asm
  FLD    Radians
  FMUL   D180divPi
  FWAIT
end;
{$ELSE}
begin
  Result := Radians * D180divPI;
end;
{$ENDIF}
{$ELSE}
begin
  Result := Radians * D180divPI;
end;
{$ENDIF}

procedure SinCos(const Angle: Float; var SinF, CosF: Double);
{$IFDEF ASMVERSIONS}
asm
  FLD    Angle
  FSINCOS
  FSTP   QWORD PTR [EDX]
  FSTP   QWORD PTR [EAX]
end;
{$ELSE}
begin
  SinF := System.Sin(Angle);
  CosF := System.Cos(Angle);
end;
{$ENDIF}

procedure SinCos(const Angle: Float; var SinF, CosF: Single);
{$IFDEF ASMVERSIONS}
asm
  FLD    Angle
  FSINCOS
  FSTP   DWORD PTR [EDX]
  FSTP   DWORD PTR [EAX]
end;
{$ELSE}
begin
  SinF := System.Sin(Angle);
  CosF := System.Cos(Angle);
end;
{$ENDIF}

function Tan(const Angle: Float): Float;
{$IFDEF ASMVERSIONS}
asm
  FLD    Angle
  FPTAN
  FSTP   ST(0)
end;
{$ELSE}
begin
  Result := Sin(Angle) / Cos(Angle);
end;
{$ENDIF}

function CoTan(const Angle: Float): Float;
{$IFDEF ASMVERSIONS}
asm
  FLD    Angle
  FPTAN
  FDIVRP
end;
{$ELSE}
begin
  Result := Cos(Angle) / Sin(Angle);
end;
{$ENDIF}

function ArcSin(const Value: Float): Float;
{$IFDEF ASMVERSIONS}
{$IFNDEF PIC}
asm
  PUSH   EBX
  FLD    Value
  FLD    ST
  FMUL   ST, ST
  FSUBR  One
  FSQRT
  FPATAN
  POP    EBX
end;
{$ELSE}
begin
  if Value < 0.0 then
    Result := -ArcSin(-Value)
  else
    if Value = 1.0 then
      Result := PI / 2
    else
      Result := ArcTan(Value / Sqrt(1 - Value * Value));
end;
{$ENDIF}
{$ELSE}
begin
  if Value < 0.0 then
    Result := -ArcSin(-Value)
  else
    if Value = 1.0 then
      Result := PI / 2
    else
      Result := ArcTan(Value / Sqrt(1 - Value * Value));
end;
{$ENDIF}

function ArcCos(const Value: Float): Float;
{$IFDEF ASMVERSIONS}
{$IFNDEF PIC}
asm
  FLD    Value
  FMUL   ST, ST
  FSUBR  One
  FSQRT
  FLD    Value
  FPATAN
end;
{$ELSE}
begin
  if Value < 0.0 then
    Result := PI - ArcCos(-Value)
  else
    if Value = 0.0 then
      Result := PI / 2
    else
      Result := ArcTan(Sqrt(1 - Value * Value) / Value);
end;
{$ENDIF}
{$ELSE}
begin
  if Value < 0.0 then
    Result := PI - ArcCos(-Value)
  else
    if Value = 0.0 then
      Result := PI / 2
    else
      Result := ArcTan(Sqrt(1 - Value * Value) / Value);
end;
{$ENDIF}

function ArcTan(const Value: Float): Float;
{$IFDEF ASMVERSIONS}
asm
  FLD    Value
  FLD1
  FPATAN
  FWAIT
end;
{$ELSE}
begin
  Result:=System.ArcTan(Value);
end;
{$ENDIF}

function ArcCoTan(const Value: Float): Float;
{$IFDEF ASMVERSIONS}
asm
  FLD1
  FLD    Value
  FPATAN
  FWAIT
end;
{$ELSE}
begin
 Result := ArcTan(1 / Value);
end;
{$ENDIF}


{*************************************************************************}
{                            Vector functions                             }
{*************************************************************************}


procedure Vector2(const X, Y: Float; var Vec: TVector2);
begin
  Vec.X := X;
  Vec.Y := Y;
end;

procedure Vector3(const X, Y, Z: Float; var Vec: TVector3);
begin
  Vec.X := X;
  Vec.Y := Y;
  Vec.Z := Z;
end;

procedure Vector4(const X, Y, Z, W: Float; var Vec: TVector4);
begin
  Vec.X := X;
  Vec.Y := Y;
  Vec.Z := Z;
  Vec.W := W;
end;

function Vector2(const X, Y: Float): TVector2;
begin
  Result.X := X;
  Result.Y := Y;
end;

function Vector3(const X, Y, Z: Float): TVector3;
begin
  Result.X := X;
  Result.Y := Y;
  Result.Z := Z;
end;

function Vector4(const X, Y, Z, W: Float): TVector4;
begin
  Result.X := X;
  Result.Y := Y;
  Result.Z := Z;
  Result.W := W;
end;

{* 2D Vector functions *}

function Vec2Add(const V1, V2: TVector2): TVector2;
begin
  Result.X := V1.X + V2.X;
  Result.Y := V1.Y + V2.Y;
end;

function Vec2Subtract(const V1, V2: TVector2): TVector2;
begin
  Result.X := V1.X - V2.X;
  Result.Y := V1.Y - V2.Y;
end;

function Vec2Multiply(const V1, V2: TVector2): TVector2;
begin
  Result.X := V1.X * V2.X;
  Result.Y := V1.Y * V2.Y;
end;

{(*}

function Vec2Divide(const V1, V2: TVector2): TVector2;
begin
  if V2.X <> 0 then Result.X := V1.X / V2.X else Result.X := 0;
  if V2.Y <> 0 then Result.Y := V1.Y / V2.Y else Result.Y := 0;
end;

{*)}

function Vec2Comp(const V1, V2: TVector2): Boolean;
begin
  Result :=
    (Abs(V1.X - V2.X) < FloatTolerance) and
    (Abs(V1.Y - V2.Y) < FloatTolerance);
end;

function Vec2Dot(const V1, V2: TVector2): Float;
{$IFDEF ASMVERSIONS}
asm
  FLD    V1.X
  FMUL   V2.X
  FLD    V1.Y
  FMUL   V2.Y
  FADDP
end;
{$ELSE}
begin
  Result := V1.X * V2.X + V1.Y * V2.Y;
end;
{$ENDIF}

function Vec2Length(const V: TVector2): Float;
begin
  Result := Sqrt(V.X * V.X + V.Y * V.Y);
end;

function Vec2Dist(const V1, V2: TVector2): Float;
{$IFDEF ASMVERSIONS}
asm
  FLD    V1.X
  FSUB   V2.X
  FMUL   ST, ST
  FLD    V1.Y
  FSUB   V2.Y
  FMUL   ST, ST
  FADD
  FSQRT
end;
{$ELSE}
begin
  Result := Sqrt((V1.X - V2.X) * (V1.X - V2.X) + (V1.Y - V2.Y) * (V1.Y - V2.Y));
end;
{$ENDIF}

function Vec2Lerp(const V1, V2: TVector2; const T: Float): TVector2;
begin
  Result.X := V1.X + T * (V2.X - V1.X);
  Result.Y := V1.Y + T * (V2.Y - V1.Y);
end;

function Vec2Maximize(const V1, V2: TVector2): TVector2;
begin
  Result := V1;
  if V2.X > Result.X then
    Result.X := V2.X;
  if V2.Y > Result.Y then
    Result.Y := V2.Y;
end;

function Vec2Minimize(const V1, V2: TVector2): TVector2;
begin
  Result := V1;
  if V2.X < Result.X then
    Result.X := V2.X;
  if V2.Y < Result.Y then
    Result.Y := V2.Y;
end;

procedure Vec2ScaleP(var V: TVector2; const ScaleBy: Float);
begin
  V.X := V.X * ScaleBy;
  V.Y := V.Y * ScaleBy;
end;

function Vec2Scale(const V: TVector2; const ScaleBy: Float): TVector2;
begin
  Result.X := V.X * ScaleBy;
  Result.Y := V.Y * ScaleBy;
end;

procedure Vec2NegateP(var V: TVector2);
begin
  V.X := -V.X;
  V.Y := -V.Y;
end;

function Vec2Negate(const V: TVector2): TVector2;
begin
  Result.X := -V.X;
  Result.Y := -V.Y;
end;

procedure Vec2NormalizeP(var V: TVector2);
{$IFDEF ASMVERSIONS}
asm
  FLD    V.X
  FMUL   ST, ST
  FLD    V.Y
  FMUL   ST, ST
  FADD
  FSQRT
  FLD1
  FDIVR
  FLD    ST
  FMUL   V.X
  FSTP   V.X
  FMUL   V.Y
  FSTP   V.Y
end;
{$ELSE}
var
  Len: Float;
  InvLen: Float;
begin
  Len := V.X * V.X + V.Y * V.Y;
  if Len <> 0 then
  begin
    InvLen := 1 / Sqrt(Len);
    V.X := V.X * InvLen;
    V.Y := V.Y * InvLen;
  end
  else
    V := ZeroVector2;
end;
{$ENDIF}

function Vec2Normalize(const V: TVector2): TVector2;
{$IFDEF ASMVERSIONS}
asm
  FLD    V.X
  FMUL   ST, ST
  FLD    V.Y
  FMUL   ST, ST
  FADD
  FSQRT
  FLD1
  FDIVR
  FLD    ST
  FMUL   V.X
  FSTP   Result.X
  FMUL   V.Y
  FSTP   Result.Y
end;
{$ELSE}
var
  Len: Float;
  InvLen: Float;
begin
  Len := V.X * V.X + V.Y * V.Y;
  if Len <> 0 then
  begin
    InvLen := 1 / Sqrt(Len);
    Result.X := V.X * InvLen;
    Result.Y := V.Y * InvLen;
  end
  else
    Result := ZeroVector2;
end;
{$ENDIF}

{* 3D Vector functions}

function Vec3Add(const V1, V2: TVector3): TVector3;
begin
  Result.X := V1.X + V2.X;
  Result.Y := V1.Y + V2.Y;
  Result.Z := V1.Z + V2.Z;
end;

function Vec3Subtract(const V1, V2: TVector3): TVector3;
begin
  Result.X := V1.X - V2.X;
  Result.Y := V1.Y - V2.Y;
  Result.Z := V1.Z - V2.Z;
end;

function Vec3Multiply(const V1, V2: TVector3): TVector3;
begin
  Result.X := V1.X * V2.X;
  Result.Y := V1.Y * V2.Y;
  Result.Z := V1.Z * V2.Z;
end;

{(*}

function Vec3Divide(const V1, V2: TVector3): TVector3;
begin
  if V2.X <> 0 then Result.X := V1.X / V2.X else Result.X := 0;
  if V2.Y <> 0 then Result.Y := V1.Y / V2.Y else Result.Y := 0;
  if V2.Z <> 0 then Result.Z := V1.Z / V2.Z else Result.Z := 0;
end;

{*)}

function Vec3Comp(const V1, V2: TVector3): Boolean;
begin
  Result :=
    (Abs(V1.X - V2.X) < FloatTolerance) and
    (Abs(V1.Y - V2.Y) < FloatTolerance) and
    (Abs(V1.Z - V2.Z) < FloatTolerance);
end;

function Vec3Dot(const V1, V2: TVector3): Float;
{$IFDEF ASMVERSIONS}
asm
  FLD    V1.X
  FMUL   V2.X
  FLD    V1.Y
  FMUL   V2.Y
  FADDP
  FLD    V1.Z
  FMUL   V2.Z
  FADDP
end;
{$ELSE}
begin
  Result := V1.X * V2.X + V1.Y * V2.Y + V1.Z * V2.Z;
end;
{$ENDIF}

function Vec3Cross(const V1, V2: TVector3): TVector3;
begin
  Result.X := V1.Y * V2.Z - V1.Z * V2.Y;
  Result.Y := V1.Z * V2.X - V1.X * V2.Z;
  Result.Z := V1.X * V2.Y - V1.Y * V2.X;
end;

function Vec3Length(const V: TVector3): Float;
begin
  Result := Sqrt(V.X * V.X + V.Y * V.Y + V.Z * V.Z);
end;

function Vec3Dist(const V1, V2: TVector3): Float;
{$IFDEF ASMVERSIONS}
asm
  FLD    V1.X
  FSUB   V2.X
  FMUL   ST, ST
  FLD    V1.Y
  FSUB   V2.Y
  FMUL   ST, ST
  FADD
  FLD    V1.Z
  FSUB   V2.Z
  FMUL   ST, ST
  FADD
  FSQRT
end;
{$ELSE}
begin
  Result := Sqrt((V1.X - V2.X) * (V1.X - V2.X) + (V1.Y - V2.Y) * (V1.Y - V2.Y) +
    (V1.Z - V2.Z) * (V1.Z - V2.Z));
end;
{$ENDIF}

function Vec3Lerp(const V1, V2: TVector3; const T: Float): TVector3;
begin
  Result.X := V1.X + T * (V2.X - V1.X);
  Result.Y := V1.Y + T * (V2.Y - V1.Y);
  Result.Z := V1.Z + T * (V2.Z - V1.Z);
end;

function Vec3Maximize(const V1, V2: TVector3): TVector3;
begin
  Result := V1;
  if V2.X > Result.X then
    Result.X := V2.X;
  if V2.Y > Result.Y then
    Result.Y := V2.Y;
  if V2.Z > Result.Z then
    Result.Z := V2.Z;
end;

function Vec3Minimize(const V1, V2: TVector3): TVector3;
begin
  Result := V1;
  if V2.X < Result.X then
    Result.X := V2.X;
  if V2.Y < Result.Y then
    Result.Y := V2.Y;
  if V2.Z < Result.Z then
    Result.Z := V2.Z;
end;

procedure Vec3ScaleP(var V: TVector3; const ScaleBy: Float);
begin
  V.X := V.X * ScaleBy;
  V.Y := V.Y * ScaleBy;
  V.Z := V.Z * ScaleBy;
end;

function Vec3Scale(const V: TVector3; const ScaleBy: Float): TVector3;
begin
  Result.X := V.X * ScaleBy;
  Result.Y := V.Y * ScaleBy;
  Result.Z := V.Z * ScaleBy;
end;

procedure Vec3NegateP(var V: TVector3);
begin
  V.X := -V.X;
  V.Y := -V.Y;
  V.Z := -V.Z;
end;

function Vec3Negate(const V: TVector3): TVector3;
begin
  Result.X := -V.X;
  Result.Y := -V.Y;
  Result.Z := -V.Z;
end;

procedure Vec3NormalizeP(var V: TVector3);
{$IFDEF ASMVERSIONS}
asm
  FLD    V.X
  FMUL   ST, ST
  FLD    V.Y
  FMUL   ST, ST
  FADD
  FLD    V.Z
  FMUL   ST, ST
  FADD
  FSQRT
  FLD1
  FDIVR
  FLD    ST
  FMUL   V.X
  FSTP   V.X
  FLD    ST
  FMUL   V.Y
  FSTP   V.Y
  FMUL   V.Z
  FSTP   V.Z
end;
{$ELSE}
var
  Len: Float;
  InvLen: Float;
begin
  Len := V.X * V.X + V.Y * V.Y + V.Z * V.Z;
  if Len <> 0 then
  begin
    InvLen := 1 / Sqrt(Len);
    V.X := V.X * InvLen;
    V.Y := V.Y * InvLen;
    V.Z := V.Z * InvLen;
  end
  else
    V := ZeroVector3;
end;
{$ENDIF}

function Vec3Normalize(const V: TVector3): TVector3;
{$IFDEF ASMVERSIONS}
asm
  FLD    V.X
  FMUL   ST, ST
  FLD    V.Y
  FMUL   ST, ST
  FADD
  FLD    V.Z
  FMUL   ST, ST
  FADD
  FSQRT
  FLD1
  FDIVR
  FLD    ST
  FMUL   V.X
  FSTP   Result.X
  FLD    ST
  FMUL   V.Y
  FSTP   Result.Y
  FMUL   V.Z
  FSTP   Result.Z
end;
{$ELSE}
var
  Len: Float;
  InvLen: Float;
begin
  Len := V.X * V.X + V.Y * V.Y + V.Z * V.Z;
  if Len <> 0 then
  begin
    InvLen := 1 / Sqrt(Len);
    Result.X := V.X * InvLen;
    Result.Y := V.Y * InvLen;
    Result.Z := V.Z * InvLen;
  end
  else
    Result := ZeroVector3;
end;
{$ENDIF}

{* 4D Vector functions *}

function Vec4Add(const V1, V2: TVector4): TVector4;
begin
  Result.X := V1.X + V2.X;
  Result.Y := V1.Y + V2.Y;
  Result.Z := V1.Z + V2.Z;
  Result.W := V1.W + V2.W;
end;

function Vec4Subtract(const V1, V2: TVector4): TVector4;
begin
  Result.X := V1.X - V2.X;
  Result.Y := V1.Y - V2.Y;
  Result.Z := V1.Z - V2.Z;
  Result.W := V1.W - V2.W;
end;

function Vec4Multiply(const V1, V2: TVector4): TVector4;
begin
  Result.X := V1.X * V2.X;
  Result.Y := V1.Y * V2.Y;
  Result.Z := V1.Z * V2.Z;
  Result.W := V1.W * V2.W;
end;

{(*}

function Vec4Divide(const V1, V2: TVector4): TVector4;
begin
  if V2.X <> 0 then Result.X := V1.X / V2.X else Result.X := 0;
  if V2.Y <> 0 then Result.Y := V1.Y / V2.Y else Result.Y := 0;
  if V2.Z <> 0 then Result.Z := V1.Z / V2.Z else Result.Z := 0;
  if V2.W <> 0 then Result.W := V1.W / V2.W else Result.W := 0;
end;

{*)}

function Vec4Comp(const V1, V2: TVector4): Boolean;
begin
  Result :=
    (Abs(V1.X - V2.X) < FloatTolerance) and
    (Abs(V1.Y - V2.Y) < FloatTolerance) and
    (Abs(V1.Z - V2.Z) < FloatTolerance) and
    (Abs(V1.W - V2.W) < FloatTolerance);
end;

function Vec4Dot(const V1, V2: TVector4): Float;
{$IFDEF ASMVERSIONS}
asm
  FLD    V1.X
  FMUL   V2.X
  FLD    V1.Y
  FMUL   V2.Y
  FADDP
  FLD    V1.Z
  FMUL   V2.Z
  FADDP
  FLD    V1.W
  FMUL   V2.W
  FADDP
end;
{$ELSE}
begin
  Result := V1.X * V2.X + V1.Y * V2.Y + V1.Z * V2.Z + V1.W * V2.W;
end;
{$ENDIF}

function Vec4Cross(const V1, V2: TVector4): TVector4;
begin
  Result.X := V1.Y * V2.Z - V1.Z * V2.Y;
  Result.Y := V1.Z * V2.X - V1.X * V2.Z;
  Result.Z := V1.X * V2.Y - V1.Y * V2.X;
  Result.W := 1;
end;

function Vec4Length(const V: TVector4): Float;
begin
  Result := Sqrt(V.X * V.X + V.Y * V.Y + V.Z * V.Z + V.W * V.W);
end;

function Vec4Dist(const V1, V2: TVector4): Float;
{$IFDEF ASMVERSIONS}
asm
  FLD    V1.X
  FSUB   V2.X
  FMUL   ST, ST
  FLD    V1.Y
  FSUB   V2.Y
  FMUL   ST, ST
  FADD
  FLD    V1.Z
  FSUB   V2.Z
  FMUL   ST, ST
  FADD
  FLD    V1.W
  FSUB   V2.W
  FMUL   ST, ST
  FADD
  FSQRT
end;
{$ELSE}
begin
  Result := Sqrt((V1.X - V2.X) * (V1.X - V2.X) + (V1.Y - V2.Y) * (V1.Y - V2.Y) +
    (V1.Z - V2.Z) * (V1.Z - V2.Z) + (V1.W - V2.W) * (V1.W - V2.W));
end;
{$ENDIF}

function Vec4Lerp(const V1, V2: TVector4; const T: Float): TVector4;
begin
  Result.X := V1.X + T * (V2.X - V1.X);
  Result.Y := V1.Y + T * (V2.Y - V1.Y);
  Result.Z := V1.Z + T * (V2.Z - V1.Z);
  Result.W := V1.W + T * (V2.W - V1.W);
end;

function Vec4Maximize(const V1, V2: TVector4): TVector4;
begin
  Result := V1;
  if V2.X > Result.X then
    Result.X := V2.X;
  if V2.Y > Result.Y then
    Result.Y := V2.Y;
  if V2.Z > Result.Z then
    Result.Z := V2.Z;
  if V2.W > Result.W then
    Result.W := V2.W;
end;

function Vec4Minimize(const V1, V2: TVector4): TVector4;
begin
  Result := V1;
  if V2.X < Result.X then
    Result.X := V2.X;
  if V2.Y < Result.Y then
    Result.Y := V2.Y;
  if V2.Z < Result.Z then
    Result.Z := V2.Z;
  if V2.W < Result.W then
    Result.W := V2.W;
end;

procedure Vec4ScaleP(var V: TVector4; const ScaleBy: Float);
begin
  V.X := V.X * ScaleBy;
  V.Y := V.Y * ScaleBy;
  V.Z := V.Z * ScaleBy;
  V.W := V.W * ScaleBy;
end;

function Vec4Scale(const V: TVector4; const ScaleBy: Float): TVector4;
begin
  Result.X := V.X * ScaleBy;
  Result.Y := V.Y * ScaleBy;
  Result.Z := V.Z * ScaleBy;
  Result.W := V.W * ScaleBy;
end;

procedure Vec4NegateP(var V: TVector4);
begin
  V.X := -V.X;
  V.Y := -V.Y;
  V.Z := -V.Z;
  V.W := -V.W;
end;

function Vec4Negate(const V: TVector4): TVector4;
begin
  Result.X := -V.X;
  Result.Y := -V.Y;
  Result.Z := -V.Z;
  Result.W := -V.W;
end;

procedure Vec4NormalizeP(var V: TVector4);
{$IFDEF ASMVERSIONS}
asm
  FLD    V.X
  FMUL   ST, ST
  FLD    V.Y
  FMUL   ST, ST
  FADD
  FLD    V.Z
  FMUL   ST, ST
  FADD
  FLD    V.W
  FMUL   ST, ST
  FADD
  FSQRT
  FLD1
  FDIVR
  FLD    ST
  FMUL   V.X
  FSTP   V.X
  FLD    ST
  FMUL   V.Y
  FSTP   V.Y
  FLD    ST
  FMUL   V.Z
  FSTP   V.Z
  FMUL   V.W
  FSTP   V.W
end;
{$ELSE}
var
  Len: Float;
  InvLen: Float;
begin
  Len := V.X * V.X + V.Y * V.Y + V.Z * V.Z + V.W * V.W;
  if Len <> 0 then
  begin
    InvLen := 1 / Sqrt(Len);
    V.X := V.X * InvLen;
    V.Y := V.Y * InvLen;
    V.Z := V.Z * InvLen;
    V.W := V.W * InvLen;
  end
  else
    V := ZeroVector4;
end;
{$ENDIF}

function Vec4Normalize(const V: TVector4): TVector4;
{$IFDEF ASMVERSIONS}
asm
  FLD    V.X
  FMUL   ST, ST
  FLD    V.Y
  FMUL   ST, ST
  FADD
  FLD    V.Z
  FMUL   ST, ST
  FADD
  FLD    V.W
  FMUL   ST, ST
  FADD
  FSQRT
  FLD1
  FDIVR
  FLD    ST
  FMUL   V.X
  FSTP   Result.X
  FLD    ST
  FMUL   V.Y
  FSTP   Result.Y
  FLD    ST
  FMUL   V.Z
  FSTP   Result.Z
  FMUL   V.W
  FSTP   Result.W
end;
{$ELSE}
var
  Len: Float;
  InvLen: Float;
begin
  Len := V.X * V.X + V.Y * V.Y + V.Z * V.Z + V.W * V.W;
  if Len <> 0 then
  begin
    InvLen := 1 / Sqrt(Len);
    Result.X := V.X * InvLen;
    Result.Y := V.Y * InvLen;
    Result.Z := V.Z * InvLen;
    Result.W := V.W * InvLen;
  end
  else
    Result := ZeroVector4;
end;
{$ENDIF}


{*************************************************************************}
{                            Matrix functions                             }
{*************************************************************************}


procedure Matrix3x3(const Floats: array of Float; var M: TMatrix3x3);
begin
  M.Floats[0] := Floats[0];
  M.Floats[1] := Floats[1];
  M.Floats[2] := Floats[2];
  M.Floats[3] := Floats[3];
  M.Floats[4] := Floats[4];
  M.Floats[5] := Floats[5];
  M.Floats[6] := Floats[6];
  M.Floats[7] := Floats[7];
  M.Floats[8] := Floats[8];
end;

procedure Matrix3x3(const Row1: array of Float; const Row2: array of Float;
  const Row3: array of Float; var M: TMatrix3x3);
begin
  M.Arr[0, 0] := Row1[0];
  M.Arr[0, 1] := Row1[1];
  M.Arr[0, 2] := Row1[2];
  M.Arr[1, 0] := Row2[0];
  M.Arr[1, 1] := Row2[1];
  M.Arr[1, 2] := Row2[2];
  M.Arr[2, 0] := Row3[0];
  M.Arr[2, 1] := Row3[1];
  M.Arr[2, 2] := Row3[2];
end;

procedure Matrix4x4(const Floats: array of Float; var M: TMatrix4x4);
begin
  M.Floats[0] := Floats[0];
  M.Floats[1] := Floats[1];
  M.Floats[2] := Floats[2];
  M.Floats[3] := Floats[3];
  M.Floats[4] := Floats[4];
  M.Floats[5] := Floats[5];
  M.Floats[6] := Floats[6];
  M.Floats[7] := Floats[7];
  M.Floats[8] := Floats[8];
  M.Floats[9] := Floats[9];
  M.Floats[10] := Floats[10];
  M.Floats[11] := Floats[11];
  M.Floats[12] := Floats[12];
  M.Floats[13] := Floats[13];
  M.Floats[14] := Floats[14];
  M.Floats[15] := Floats[15];
end;

procedure Matrix4x4(const Row1: array of Float; const Row2: array of Float;
  const Row3: array of Float; const Row4: array of Float; var M: TMatrix4x4);
begin
  M.Arr[0, 0] := Row1[0];
  M.Arr[0, 1] := Row1[1];
  M.Arr[0, 2] := Row1[2];
  M.Arr[0, 3] := Row1[3];
  M.Arr[1, 0] := Row2[0];
  M.Arr[1, 1] := Row2[1];
  M.Arr[1, 2] := Row2[2];
  M.Arr[1, 3] := Row2[3];
  M.Arr[2, 0] := Row3[0];
  M.Arr[2, 1] := Row3[1];
  M.Arr[2, 2] := Row3[2];
  M.Arr[2, 3] := Row3[3];
  M.Arr[3, 0] := Row4[0];
  M.Arr[3, 1] := Row4[1];
  M.Arr[3, 2] := Row4[2];
  M.Arr[3, 3] := Row4[3];
end;

function Matrix3x3(const Floats: array of Float): TMatrix3x3;
begin
  Result.Floats[0] := Floats[0];
  Result.Floats[1] := Floats[1];
  Result.Floats[2] := Floats[2];
  Result.Floats[3] := Floats[3];
  Result.Floats[4] := Floats[4];
  Result.Floats[5] := Floats[5];
  Result.Floats[6] := Floats[6];
  Result.Floats[7] := Floats[7];
  Result.Floats[8] := Floats[8];
end;

function Matrix3x3(const Row1: array of Float; const Row2: array of Float;
  const Row3: array of Float): TMatrix3x3;
begin
  Result.Arr[0, 0] := Row1[0];
  Result.Arr[0, 1] := Row1[1];
  Result.Arr[0, 2] := Row1[2];
  Result.Arr[1, 0] := Row2[0];
  Result.Arr[1, 1] := Row2[1];
  Result.Arr[1, 2] := Row2[2];
  Result.Arr[2, 0] := Row3[0];
  Result.Arr[2, 1] := Row3[1];
  Result.Arr[2, 2] := Row3[2];
end;

function Matrix4x4(const Floats: array of Float): TMatrix4x4;
begin
  Result.Floats[0] := Floats[0];
  Result.Floats[1] := Floats[1];
  Result.Floats[2] := Floats[2];
  Result.Floats[3] := Floats[3];
  Result.Floats[4] := Floats[4];
  Result.Floats[5] := Floats[5];
  Result.Floats[6] := Floats[6];
  Result.Floats[7] := Floats[7];
  Result.Floats[8] := Floats[8];
  Result.Floats[9] := Floats[9];
  Result.Floats[10] := Floats[10];
  Result.Floats[11] := Floats[11];
  Result.Floats[12] := Floats[12];
  Result.Floats[13] := Floats[13];
  Result.Floats[14] := Floats[14];
  Result.Floats[15] := Floats[15];
end;

function Matrix4x4(const Row1: array of Float; const Row2: array of Float;
  const Row3: array of Float; const Row4: array of Float): TMatrix4x4;
begin
  Result.Arr[0, 0] := Row1[0];
  Result.Arr[0, 1] := Row1[1];
  Result.Arr[0, 2] := Row1[2];
  Result.Arr[0, 3] := Row1[3];
  Result.Arr[1, 0] := Row2[0];
  Result.Arr[1, 1] := Row2[1];
  Result.Arr[1, 2] := Row2[2];
  Result.Arr[1, 3] := Row2[3];
  Result.Arr[2, 0] := Row3[0];
  Result.Arr[2, 1] := Row3[1];
  Result.Arr[2, 2] := Row3[2];
  Result.Arr[2, 3] := Row3[3];
  Result.Arr[3, 0] := Row4[0];
  Result.Arr[3, 1] := Row4[1];
  Result.Arr[3, 2] := Row4[2];
  Result.Arr[3, 3] := Row4[3];
end;

{* 3x3 Matrix functions *}

function Mat3Add(const M1, M2: TMatrix3x3): TMatrix3x3;
begin
  Result.Floats[0] := M1.Floats[0] + M2.Floats[0];
  Result.Floats[1] := M1.Floats[1] + M2.Floats[1];
  Result.Floats[2] := M1.Floats[2] + M2.Floats[2];
  Result.Floats[3] := M1.Floats[3] + M2.Floats[3];
  Result.Floats[4] := M1.Floats[4] + M2.Floats[4];
  Result.Floats[5] := M1.Floats[5] + M2.Floats[5];
  Result.Floats[6] := M1.Floats[6] + M2.Floats[6];
  Result.Floats[7] := M1.Floats[7] + M2.Floats[7];
  Result.Floats[8] := M1.Floats[8] + M2.Floats[8];
end;

function Mat3Subtract(const M1, M2: TMatrix3x3): TMatrix3x3;
begin
  Result.Floats[0] := M1.Floats[0] - M2.Floats[0];
  Result.Floats[1] := M1.Floats[1] - M2.Floats[1];
  Result.Floats[2] := M1.Floats[2] - M2.Floats[2];
  Result.Floats[3] := M1.Floats[3] - M2.Floats[3];
  Result.Floats[4] := M1.Floats[4] - M2.Floats[4];
  Result.Floats[5] := M1.Floats[5] - M2.Floats[5];
  Result.Floats[6] := M1.Floats[6] - M2.Floats[6];
  Result.Floats[7] := M1.Floats[7] - M2.Floats[7];
  Result.Floats[8] := M1.Floats[8] - M2.Floats[8];
end;

function Mat3Multiply(const M1, M2: TMatrix3x3): TMatrix3x3;
begin
  Result.Arr[0, 0] := M1.Arr[0, 0] * M2.Arr[0, 0] + M1.Arr[0, 1] * M2.Arr[1, 0]
    + M1.Arr[0, 2] * M2.Arr[2, 0];
  Result.Arr[0, 1] := M1.Arr[0, 0] * M2.Arr[0, 1] + M1.Arr[0, 1] * M2.Arr[1, 1]
    + M1.Arr[0, 2] * M2.Arr[2, 1];
  Result.Arr[0, 2] := M1.Arr[0, 0] * M2.Arr[0, 2] + M1.Arr[0, 1] * M2.Arr[1, 2]
    + M1.Arr[0, 2] * M2.Arr[2, 2];
  Result.Arr[1, 0] := M1.Arr[1, 0] * M2.Arr[0, 0] + M1.Arr[1, 1] * M2.Arr[1, 0]
    + M1.Arr[1, 2] * M2.Arr[2, 0];
  Result.Arr[1, 1] := M1.Arr[1, 0] * M2.Arr[0, 1] + M1.Arr[1, 1] * M2.Arr[1, 1]
    + M1.Arr[1, 2] * M2.Arr[2, 1];
  Result.Arr[1, 2] := M1.Arr[1, 0] * M2.Arr[0, 2] + M1.Arr[1, 1] * M2.Arr[1, 2]
    + M1.Arr[1, 2] * M2.Arr[2, 2];
  Result.Arr[2, 0] := M1.Arr[2, 0] * M2.Arr[0, 0] + M1.Arr[2, 1] * M2.Arr[1, 0]
    + M1.Arr[2, 2] * M2.Arr[2, 0];
  Result.Arr[2, 1] := M1.Arr[2, 0] * M2.Arr[0, 1] + M1.Arr[2, 1] * M2.Arr[1, 1]
    + M1.Arr[2, 2] * M2.Arr[2, 1];
  Result.Arr[2, 2] := M1.Arr[2, 0] * M2.Arr[0, 2] + M1.Arr[2, 1] * M2.Arr[1, 2]
    + M1.Arr[2, 2] * M2.Arr[2, 2];
end;

function Mat3ScalarMultiply(const M: TMatrix3x3; const F: Float): TMatrix3x3;
begin
  Result.Floats[0] := M.Floats[0] * F;
  Result.Floats[1] := M.Floats[1] * F;
  Result.Floats[2] := M.Floats[2] * F;
  Result.Floats[3] := M.Floats[3] * F;
  Result.Floats[4] := M.Floats[4] * F;
  Result.Floats[5] := M.Floats[5] * F;
  Result.Floats[6] := M.Floats[6] * F;
  Result.Floats[7] := M.Floats[7] * F;
  Result.Floats[8] := M.Floats[8] * F;
end;

function Mat3Comp(const M1, M2: TMatrix3x3): Boolean;
begin
  Result :=
    (Abs(M1.Floats[0] - M2.Floats[0]) < FloatTolerance) and
    (Abs(M1.Floats[1] - M2.Floats[1]) < FloatTolerance) and
    (Abs(M1.Floats[2] - M2.Floats[2]) < FloatTolerance) and
    (Abs(M1.Floats[3] - M2.Floats[3]) < FloatTolerance) and
    (Abs(M1.Floats[4] - M2.Floats[4]) < FloatTolerance) and
    (Abs(M1.Floats[5] - M2.Floats[5]) < FloatTolerance) and
    (Abs(M1.Floats[6] - M2.Floats[6]) < FloatTolerance) and
    (Abs(M1.Floats[7] - M2.Floats[7]) < FloatTolerance) and
    (Abs(M1.Floats[8] - M2.Floats[8]) < FloatTolerance);
end;

function Mat3Transpose(const M: TMatrix3x3): TMatrix3x3;
begin
  Result.Arr[0, 0] := M.Arr[0, 0];
  Result.Arr[0, 1] := M.Arr[1, 0];
  Result.Arr[0, 2] := M.Arr[2, 0];
  Result.Arr[1, 0] := M.Arr[0, 1];
  Result.Arr[1, 1] := M.Arr[1, 1];
  Result.Arr[1, 2] := M.Arr[2, 1];
  Result.Arr[2, 0] := M.Arr[0, 2];
  Result.Arr[2, 1] := M.Arr[1, 2];
  Result.Arr[2, 2] := M.Arr[2, 2];
end;

function Mat3Determinant(const M: TMatrix3x3): Float;
begin
  Result :=
    M.Arr[0, 0] * (M.Arr[1, 1] * M.Arr[2, 2] - M.Arr[2, 1] * M.Arr[1, 2])
    - M.Arr[0, 1] * (M.Arr[1, 0] * M.Arr[2, 2] - M.Arr[2, 0] * M.Arr[1, 2])
    + M.Arr[0, 2] * (M.Arr[1, 0] * M.Arr[2, 1] - M.Arr[2, 0] * M.Arr[1, 1]);
end;

function Mat3Inverse(const M: TMatrix3x3): TMatrix3x3;
var
  Det: Float;
begin
  Det := Mat3Determinant(M);
  if Abs(Det) < FloatTolerance then
  begin
    Result := Identity3x3;
    Exit;
  end;
  {(*}
  Result.Arr[0, 0] := M.Arr[1, 1] * M.Arr[2, 2] - M.Arr[1, 2] * M.Arr[2, 1] / Det;
  Result.Arr[0, 1] := -(M.Arr[0, 1] * M.Arr[2, 2] - M.Arr[2, 1] * M.Arr[0, 2]) / Det;
  Result.Arr[0, 2] := M.Arr[0, 1] * M.Arr[1, 2] - M.Arr[1, 1] * M.Arr[0, 2] / Det;
  Result.Arr[1, 0] := -(M.Arr[1, 0] * M.Arr[2, 2] - M.Arr[1, 2] * M.Arr[2, 0]) / Det;
  Result.Arr[1, 1] := M.Arr[0, 0] * M.Arr[2, 2] - M.Arr[2, 0] * M.Arr[0, 2] / Det;
  Result.Arr[1, 2] := -(M.Arr[0, 0] * M.Arr[1, 2] - M.Arr[1, 0] * M.Arr[0, 2]) / Det;
  Result.Arr[2, 0] := M.Arr[1, 0] * M.Arr[2, 1] - M.Arr[2, 0] * M.Arr[1, 1] / Det;
  Result.Arr[2, 1] := -(M.Arr[0, 0] * M.Arr[2, 1] - M.Arr[2, 0] * M.Arr[0, 1]) / Det;
  Result.Arr[2, 2] := M.Arr[0, 0] * M.Arr[1, 1] - M.Arr[0, 1] * M.Arr[1, 0] / Det;
  {*)}
end;

{* 4x4 Matrix functions *}

function Mat4Add(const M1, M2: TMatrix4x4): TMatrix4x4;
begin
  Result.Floats[0] := M1.Floats[0] + M2.Floats[0];
  Result.Floats[1] := M1.Floats[1] + M2.Floats[1];
  Result.Floats[2] := M1.Floats[2] + M2.Floats[2];
  Result.Floats[3] := M1.Floats[3] + M2.Floats[3];
  Result.Floats[4] := M1.Floats[4] + M2.Floats[4];
  Result.Floats[5] := M1.Floats[5] + M2.Floats[5];
  Result.Floats[6] := M1.Floats[6] + M2.Floats[6];
  Result.Floats[7] := M1.Floats[7] + M2.Floats[7];
  Result.Floats[8] := M1.Floats[8] + M2.Floats[8];
  Result.Floats[9] := M1.Floats[9] + M2.Floats[9];
  Result.Floats[10] := M1.Floats[10] + M2.Floats[10];
  Result.Floats[11] := M1.Floats[11] + M2.Floats[11];
  Result.Floats[12] := M1.Floats[12] + M2.Floats[12];
  Result.Floats[13] := M1.Floats[13] + M2.Floats[13];
  Result.Floats[14] := M1.Floats[14] + M2.Floats[14];
  Result.Floats[15] := M1.Floats[15] + M2.Floats[15];
end;

function Mat4Subtract(const M1, M2: TMatrix4x4): TMatrix4x4;
begin
  Result.Floats[0] := M1.Floats[0] - M2.Floats[0];
  Result.Floats[1] := M1.Floats[1] - M2.Floats[1];
  Result.Floats[2] := M1.Floats[2] - M2.Floats[2];
  Result.Floats[3] := M1.Floats[3] - M2.Floats[3];
  Result.Floats[4] := M1.Floats[4] - M2.Floats[4];
  Result.Floats[5] := M1.Floats[5] - M2.Floats[5];
  Result.Floats[6] := M1.Floats[6] - M2.Floats[6];
  Result.Floats[7] := M1.Floats[7] - M2.Floats[7];
  Result.Floats[8] := M1.Floats[8] - M2.Floats[8];
  Result.Floats[9] := M1.Floats[9] - M2.Floats[9];
  Result.Floats[10] := M1.Floats[10] - M2.Floats[10];
  Result.Floats[11] := M1.Floats[11] - M2.Floats[11];
  Result.Floats[12] := M1.Floats[12] - M2.Floats[12];
  Result.Floats[13] := M1.Floats[13] - M2.Floats[13];
  Result.Floats[14] := M1.Floats[14] - M2.Floats[14];
  Result.Floats[15] := M1.Floats[15] - M2.Floats[15];
end;

function Mat4Multiply(const M1, M2: TMatrix4x4): TMatrix4x4;
begin
  Result.Arr[0, 0] := M1.Arr[0, 0] * M2.Arr[0, 0] + M1.Arr[0, 1] * M2.Arr[1, 0]
    + M1.Arr[0, 2] * M2.Arr[2, 0] + M1.Arr[0, 3] * M2.Arr[3, 0];
  Result.Arr[0, 1] := M1.Arr[0, 0] * M2.Arr[0, 1] + M1.Arr[0, 1] * M2.Arr[1, 1]
    + M1.Arr[0, 2] * M2.Arr[2, 1] + M1.Arr[0, 3] * M2.Arr[3, 1];
  Result.Arr[0, 2] := M1.Arr[0, 0] * M2.Arr[0, 2] + M1.Arr[0, 1] * M2.Arr[1, 2]
    + M1.Arr[0, 2] * M2.Arr[2, 2] + M1.Arr[0, 3] * M2.Arr[3, 2];
  Result.Arr[0, 3] := M1.Arr[0, 0] * M2.Arr[0, 3] + M1.Arr[0, 1] * M2.Arr[1, 3]
    + M1.Arr[0, 2] * M2.Arr[2, 3] + M1.Arr[0, 3] * M2.Arr[3, 3];
  Result.Arr[1, 0] := M1.Arr[1, 0] * M2.Arr[0, 0] + M1.Arr[1, 1] * M2.Arr[1, 0]
    + M1.Arr[1, 2] * M2.Arr[2, 0] + M1.Arr[1, 3] * M2.Arr[3, 0];
  Result.Arr[1, 1] := M1.Arr[1, 0] * M2.Arr[0, 1] + M1.Arr[1, 1] * M2.Arr[1, 1]
    + M1.Arr[1, 2] * M2.Arr[2, 1] + M1.Arr[1, 3] * M2.Arr[3, 1];
  Result.Arr[1, 2] := M1.Arr[1, 0] * M2.Arr[0, 2] + M1.Arr[1, 1] * M2.Arr[1, 2]
    + M1.Arr[1, 2] * M2.Arr[2, 2] + M1.Arr[1, 3] * M2.Arr[3, 2];
  Result.Arr[1, 3] := M1.Arr[1, 0] * M2.Arr[0, 3] + M1.Arr[1, 1] * M2.Arr[1, 3]
    + M1.Arr[1, 2] * M2.Arr[2, 3] + M1.Arr[1, 3] * M2.Arr[3, 3];
  Result.Arr[2, 0] := M1.Arr[2, 0] * M2.Arr[0, 0] + M1.Arr[2, 1] * M2.Arr[1, 0]
    + M1.Arr[2, 2] * M2.Arr[2, 0] + M1.Arr[2, 3] * M2.Arr[3, 0];
  Result.Arr[2, 1] := M1.Arr[2, 0] * M2.Arr[0, 1] + M1.Arr[2, 1] * M2.Arr[1, 1]
    + M1.Arr[2, 2] * M2.Arr[2, 1] + M1.Arr[2, 3] * M2.Arr[3, 1];
  Result.Arr[2, 2] := M1.Arr[2, 0] * M2.Arr[0, 2] + M1.Arr[2, 1] * M2.Arr[1, 2]
    + M1.Arr[2, 2] * M2.Arr[2, 2] + M1.Arr[2, 3] * M2.Arr[3, 2];
  Result.Arr[2, 3] := M1.Arr[2, 0] * M2.Arr[0, 3] + M1.Arr[2, 1] * M2.Arr[1, 3]
    + M1.Arr[2, 2] * M2.Arr[2, 3] + M1.Arr[2, 3] * M2.Arr[3, 3];
  Result.Arr[3, 0] := M1.Arr[3, 0] * M2.Arr[0, 0] + M1.Arr[3, 1] * M2.Arr[1, 0]
    + M1.Arr[3, 2] * M2.Arr[2, 0] + M1.Arr[3, 3] * M2.Arr[3, 0];
  Result.Arr[3, 1] := M1.Arr[3, 0] * M2.Arr[0, 1] + M1.Arr[3, 1] * M2.Arr[1, 1]
    + M1.Arr[3, 2] * M2.Arr[2, 1] + M1.Arr[3, 3] * M2.Arr[3, 1];
  Result.Arr[3, 2] := M1.Arr[3, 0] * M2.Arr[0, 2] + M1.Arr[3, 1] * M2.Arr[1, 2]
    + M1.Arr[3, 2] * M2.Arr[2, 2] + M1.Arr[3, 3] * M2.Arr[3, 2];
  Result.Arr[3, 3] := M1.Arr[3, 0] * M2.Arr[0, 3] + M1.Arr[3, 1] * M2.Arr[1, 3]
    + M1.Arr[3, 2] * M2.Arr[2, 3] + M1.Arr[3, 3] * M2.Arr[3, 3];
end;

function Mat4ScalarMultiply(const M: TMatrix4x4; const F: Float): TMatrix4x4;
begin
  Result.Floats[0] := M.Floats[0] * F;
  Result.Floats[1] := M.Floats[1] * F;
  Result.Floats[2] := M.Floats[2] * F;
  Result.Floats[3] := M.Floats[3] * F;
  Result.Floats[4] := M.Floats[4] * F;
  Result.Floats[5] := M.Floats[5] * F;
  Result.Floats[6] := M.Floats[6] * F;
  Result.Floats[7] := M.Floats[7] * F;
  Result.Floats[8] := M.Floats[8] * F;
  Result.Floats[9] := M.Floats[9] * F;
  Result.Floats[10] := M.Floats[10] * F;
  Result.Floats[11] := M.Floats[11] * F;
  Result.Floats[12] := M.Floats[12] * F;
  Result.Floats[13] := M.Floats[13] * F;
  Result.Floats[14] := M.Floats[14] * F;
  Result.Floats[15] := M.Floats[15] * F;
end;

function Mat4Comp(const M1, M2: TMatrix4x4): Boolean;
begin
  Result :=
    (Abs(M1.Floats[0] - M2.Floats[0]) < FloatTolerance) and
    (Abs(M1.Floats[1] - M2.Floats[1]) < FloatTolerance) and
    (Abs(M1.Floats[2] - M2.Floats[2]) < FloatTolerance) and
    (Abs(M1.Floats[3] - M2.Floats[3]) < FloatTolerance) and
    (Abs(M1.Floats[4] - M2.Floats[4]) < FloatTolerance) and
    (Abs(M1.Floats[5] - M2.Floats[5]) < FloatTolerance) and
    (Abs(M1.Floats[6] - M2.Floats[6]) < FloatTolerance) and
    (Abs(M1.Floats[7] - M2.Floats[7]) < FloatTolerance) and
    (Abs(M1.Floats[9] - M2.Floats[9]) < FloatTolerance) and
    (Abs(M1.Floats[10] - M2.Floats[10]) < FloatTolerance) and
    (Abs(M1.Floats[11] - M2.Floats[11]) < FloatTolerance) and
    (Abs(M1.Floats[12] - M2.Floats[12]) < FloatTolerance) and
    (Abs(M1.Floats[13] - M2.Floats[13]) < FloatTolerance) and
    (Abs(M1.Floats[14] - M2.Floats[14]) < FloatTolerance) and
    (Abs(M1.Floats[15] - M2.Floats[15]) < FloatTolerance);
end;

function Mat4Transpose(const M: TMatrix4x4): TMatrix4x4;
begin
  Result.Arr[0, 0] := M.Arr[0, 0];
  Result.Arr[0, 1] := M.Arr[1, 0];
  Result.Arr[0, 2] := M.Arr[2, 0];
  Result.Arr[0, 3] := M.Arr[3, 0];
  Result.Arr[1, 0] := M.Arr[0, 1];
  Result.Arr[1, 1] := M.Arr[1, 1];
  Result.Arr[1, 2] := M.Arr[2, 1];
  Result.Arr[1, 3] := M.Arr[3, 1];
  Result.Arr[2, 0] := M.Arr[0, 2];
  Result.Arr[2, 1] := M.Arr[1, 2];
  Result.Arr[2, 2] := M.Arr[2, 2];
  Result.Arr[2, 3] := M.Arr[3, 2];
  Result.Arr[3, 0] := M.Arr[0, 3];
  Result.Arr[3, 1] := M.Arr[1, 3];
  Result.Arr[3, 2] := M.Arr[2, 3];
  Result.Arr[3, 3] := M.Arr[3, 3];
end;

function MatDetInternal(const A1, A2, A3, B1, B2, B3, C1, C2, C3: Float): Float;
begin
  Result := A1 * (B2 * C3 - B3 * C2)
    - B1 * (A2 * C3 - A3 * C2)
    + C1 * (A2 * B3 - A3 * B2);
end;

{(*}

function Mat4Determinant(const M: TMatrix4x4): Float;
begin
  Result := M.Arr[0, 0] *
    MatDetInternal(M.Arr[1, 1], M.Arr[2, 1], M.Arr[3, 1],
                   M.Arr[1, 2], M.Arr[2, 2], M.Arr[3, 2],
                   M.Arr[1, 3], M.Arr[2, 3], M.Arr[3, 3])
    - M.Arr[0, 1] *
    MatDetInternal(M.Arr[1, 0], M.Arr[2, 0], M.Arr[3, 0],
                   M.Arr[1, 2], M.Arr[2, 2], M.Arr[3, 2],
                   M.Arr[1, 3], M.Arr[2, 3], M.Arr[3, 3])
    + M.Arr[0, 2] *
    MatDetInternal(M.Arr[1, 0], M.Arr[2, 0], M.Arr[3, 0],
                   M.Arr[1, 1], M.Arr[2, 1], M.Arr[3, 1],
                   M.Arr[1, 3], M.Arr[2, 3], M.Arr[3, 3])
    - M.Arr[0, 3] *
    MatDetInternal(M.Arr[1, 0], M.Arr[2, 0], M.Arr[3, 0],
                   M.Arr[1, 1], M.Arr[2, 1], M.Arr[3, 1],
                   M.Arr[1, 2], M.Arr[2, 2], M.Arr[3, 2]);
end;

{*)}

function MatGet3x3SubInternal(const M: TMatrix4x4; R, C: Long): TMatrix3x3;
var
  DI, DJ, SI, SJ: Long;
begin
  for DI := 0 to 2 do
  begin
    for DJ := 0 to 2 do
    begin
      if DI >= R then
        SI := DI + 1
      else
        SI := DI;
      if DJ >= C then
        SJ := DJ + 1
      else
        SJ := DJ;
      Result.Arr[DI, DJ] := M.Arr[SI, SJ];
    end;
  end;
end;

function Mat4Inverse(const M: TMatrix4x4): TMatrix4x4;
var
  Det: Float;
  Tmp: TMatrix3x3;
  I, J, Sign: Long;
begin
  Det := Mat4Determinant(M);
  if Abs(Det) < FloatTolerance then
  begin
    Result := Identity4x4;
    Exit;
  end;
  for I := 0 to 3 do
    for J := 0 to 3 do
    begin
      Sign := 1 - ((I + J) mod 2) * 2;
      Tmp := MatGet3x3SubInternal(M, I, J);
      Result.Arr[J, I] := Mat3Determinant(Tmp) * Sign / Det;
    end;
end;


{*************************************************************************}
{                      TNMatrix3x3Stack methods                           }
{*************************************************************************}


constructor TNMatrix3x3Stack.Create;
begin
  inherited Create(SizeOf(TMatrix3x3));
end;

procedure TNMatrix3x3Stack.Push(const Mat: TMatrix3x3);
begin
  PushItem(Mat);
end;

procedure TNMatrix3x3Stack.Pop(var Mat: TMatrix3x3);
begin
  PopItem(Mat);
end;

procedure TNMatrix3x3Stack.Peek(var Mat: TMatrix3x3);
begin
  PeekItem(Mat);
end;


{*************************************************************************}
{                      TNMatrix4x4Stack methods                           }
{*************************************************************************}


constructor TNMatrix4x4Stack.Create;
begin
  inherited Create(SizeOf(TMatrix4x4));
end;

procedure TNMatrix4x4Stack.Push(const Mat: TMatrix4x4);
begin
  PushItem(Mat);
end;

procedure TNMatrix4x4Stack.Pop(var Mat: TMatrix4x4);
begin
  PopItem(Mat);
end;

procedure TNMatrix4x4Stack.Peek(var Mat: TMatrix4x4);
begin
  PeekItem(Mat);
end;

end.

