{*************************************************************************}
{*   MemMgrs                                                             *}
{*   Copyright (C) 2003 Marek Mauder, Vampyre SoftWorks                  *}
{*   Engine\Kernel\MemMgrs.pas                                           *}
{*************************************************************************}
{                                                                         }
{ This program is free software; you can redistribute it and/or modify it }
{ under the terms of the GNU General Public License as published by the   }
{ Free Software Foundation; either version 2 of the License, or (at your  }
{ option) any later version. This program is distributed in the hope that }
{ it will be useful, but WITHOUT ANY WARRANTY; without even the implied   }
{ warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See   }
{ the GNU General Public License for more details.                        }
{                                                                         }
{ You should have received a copy of the GNU General Public License       }
{ along with this program; if not, write to the Free Software Foundation, }
{ Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.         }
{                                                                         }
{*************************************************************************}
{  Created On: 22th September 2003  (22.09.2003)                          }
{*************************************************************************}

{Provisional memory manager settings until we have our own
 global crossplatform mem manager.}

unit MemMgrs;

{ TODOS
  - Test if ShareMem works under FPC.
  - Do do not write debug log to file on end only, but
     after _x_ lines added.
}     

{$I Compiler.inc}

interface

uses
{$IFDEF MSWINDOWS}
  Win32MemMgrs;
{$ENDIF}
{$IFDEF LINUX}
LinuxMemMgrs;
{$ENDIF}

var
  StdMemoryManager: TMemoryManager;
  NewMemoryManager: TMemoryManager;
  UsedMemoryManager: TMemoryManager;
  DebugMemoryManager: TMemoryManager;

  {Memory usage/alloc stats variables}
  GetMemCount: LongInt = 0;
  FreeMemCount: LongInt = 0;
  ReallocMemCount: LongInt = 0;


{*************************************************************************}
{                             IMPLEMEMTATION                              }
{*************************************************************************}


implementation

uses
  SysUtils, Kernel;


{*************************************************************************}
{                               DEBUG things                              }
{*************************************************************************}


{$IFDEF DEBUG}

const
  MaxBlocks = 65536 * 1024;

type
  TBlockInfo = record
    Address: Pointer;
    Size: Long;
    Reallocated: Long;
    AllocOrder: Long;
  end;
  PBlockInfo = ^TBlockInfo;

  TBlocks = array[0..MaxBlocks - 1] of TBlockInfo;
  PBlocks = ^TBlocks;

  TMsgs = array[0..1024 * 1024 - 1] of ShortString;
  PMsgs = ^TMsgs;

  TBlockList = object
  private
    FBlocks: PBlocks;
    FCapacity: Long;
    FCount: Long;
    function GetBlock(Index: Long): PBlockInfo;
  public
    constructor Create;
    destructor Destroy;
    procedure Add(const Block: TBlockInfo);
    procedure Delete(Adr: Pointer);
    function Find(Adr: Pointer; var Index: Long): Boolean;
    procedure Grow;
    property Blocks[Index: Long]: PBlockInfo read GetBlock;
  end;

  TDebugLog = object
  private
    FLogFile: TextFile;
    FFileName: ShortString;
    FBuf: array[0..255] of Char;
    FMsgs: PMsgs;
    FCapacity: Long;
    FCount: Long;
  public
    constructor Create;
    destructor Destroy;
    procedure Grow;
    procedure AddMsg(const S: ShortString);
    procedure ReportOnStartUp;
    procedure ReportOnShutDown;
    procedure AddLB;
    procedure AddAllocMsg(const Block: TBlockInfo);
    procedure AddFreeMsg(const Block: TBlockInfo);
    procedure AddReallocMsg(const Block: TBlockInfo);
    procedure Flush;
  end;

var
  Allocations: Long = 0;
  DoDebug: Boolean = True;
  List: TBlockList;
  Block: TBlockInfo;
  DebugLog: TDebugLog;


{*************************************************************************}
{                         TBlockList methods                              }
{*************************************************************************}


constructor TBlockList.Create;
begin
  FCount := 0;
  FCapacity := 0;
  FBlocks := nil;
end;

destructor TBlockList.Destroy;
begin
  DebugMemoryManager.FreeMem(FBlocks);
end;

function TBlockList.GetBlock(Index: Long): PBlockInfo;
begin
  if (Index > -1) and (Index < FCount) then
    Result := @FBlocks[Index]
  else
    Result := nil;
end;

procedure TBlockList.Add(const Block: TBlockInfo);
var
  I, Size: Long;
begin
  Find(Block.Address, I);
  if FCount = FCapacity then Grow;
  if I < FCount then
  begin
    Size := SizeOf(TBlockInfo);
    Move(Pointer(Long(FBlocks) + I * Size)^,
      Pointer(Long(FBlocks) + (I + 1) * Size)^,
      (FCount - I) * Size);
  end;
  FBlocks[I] := Block;
  Inc(FCount);
end;

procedure TBlockList.Delete(Adr: Pointer);
var
  Index, Size: Long;
begin
  if Find(Adr, Index) then
  begin
    Size := SizeOf(TBlockInfo);
    Dec(FCount);
    if Index < FCount then
      Move(Pointer(Long(FBlocks) + (Index + 1) * Size)^,
        Pointer(Long(FBlocks) + Index * Size)^,
        (FCount - Index) * Size);
    if FCount < FCapacity shr 1 then
    begin
      FCapacity := FCount + 1;
      FBlocks := DebugMemoryManager.ReallocMem(FBlocks, FCapacity *
        SizeOf(TBlockInfo));
    end;
  end;
end;

function TBlockList.Find(Adr: Pointer; var Index: Long): Boolean;
var
  L, H, I, C: Long;
begin
  Result := False;
  L := 0;
  H := FCount - 1;
  while L <= H do
  begin
    I := (L + H) shr 1;
    C := Long(FBlocks[I].Address) - Long(Adr);
    if C < 0 then
      L := I + 1
    else
    begin
      H := I - 1;
      if C = 0 then Result := True;
    end;
  end;
  Index := L;
end;

procedure TBlockList.Grow;
begin
  FCapacity := FCapacity + FCapacity shr 2 + 1;
  FBlocks := DebugMemoryManager.ReallocMem(FBlocks, FCapacity *
    SizeOf(TBlockInfo));
end;

{*************************************************************************}
{                         TDebugLog methods                               }
{*************************************************************************}


constructor TDebugLog.Create;
begin
  FCapacity := 0;
  FCount := 0;
  FMsgs := nil;
  ReportOnStartUp;
{$IFDEF MSDOS}
  Watch out for long file name
{$ENDIF}
  FFileName := ExeDir + LogDir + PathDelim + ExtractFileName(
    ChangeFileExt(ExeName, '.Memory.log'));
  ForceDirectories(ExtractFileDir(FFileName));
  AssignFile(FLogFile, FFileName);
  Rewrite(FLogFile);
end;

destructor TDebugLog.Destroy;
begin
  ReportOnShutDown;
  Flush;
  CloseFile(FLogFile);
  DebugMemoryManager.FreeMem(FMsgs);
end;

procedure TDebugLog.Grow;
begin
  FCapacity := FCapacity + FCapacity shr 2 + 1;
  FMsgs := DebugMemoryManager.ReallocMem(FMsgs, FCapacity *
    SizeOf(ShortString));
end;

procedure TDebugLog.AddMsg(const S: ShortString);
begin
  if FCount = FCapacity then Grow;
  Inc(FCount);
  FMsgs[FCount - 1] := S;
end;

procedure TDebugLog.ReportOnStartUp;
begin
  DoDebug := False;
  AddLB;
  AddMsg('               Debug Memory Log File');
  AddMsg('                      StartUp       ');
  AddLB;
  DoDebug := True;
end;

procedure TDebugLog.ReportOnShutDown;
var
  I: Long;
begin
  AddLB;
  AddMsg('                    ShutDown       ');
  if List.FCount > 0 then
  begin
    AddMsg(Format('There are still %d allocated pointers!',
      [List.FCount]));
    for I := 0 to List.FCount - 1 do
      with List.FBlocks[I] do
        AddMsg(Format('Address: %p, Size: %.0n B, AllocOrder: %d',
          [Address, Size / 1, AllocOrder]));
  end;
  AddLB;
  AddMsg(Format('Pointers allocated: %d', [GetMemCount]));
  AddMsg(Format('Pointers reallocated: %d', [ReallocMemCount]));
  AddMsg(Format('Pointers freed: %d', [FreeMemCount]));
  AddLB;
end;

procedure TDebugLog.AddLB;
begin
  AddMsg('----------------------------------------------------');
end;

procedure TDebugLog.AddAllocMsg(const Block: TBlockInfo);
begin
  with Block do
    DebugLog.AddMsg(
      StrFmt(FBuf,
        'Pointer allocated: Address: %p, Size: %.0n B, AllocOrder: %d',
      [Address, Size / 1, AllocOrder]));
end;

procedure TDebugLog.AddFreeMsg(const Block: TBlockInfo);
begin
  with Block do
    DebugLog.AddMsg(
      StrFmt(FBuf, 'Pointer freed: Address: %p, Size: %.0n B, AllocOrder: %d',
      [Address, Size / 1, AllocOrder]));
end;

procedure TDebugLog.AddReallocMsg(const Block: TBlockInfo);
begin
  with Block do
    DebugLog.AddMsg(
      StrFmt(FBuf, 'Pointer reallocated: Address: %p, Size: %.0n B,' +
      ' AllocOrder: %d, Reallocated: %d',
      [Address, Size / 1, AllocOrder, Reallocated]));
end;

procedure TDebugLog.Flush;
var
  I: Long;
begin
  for I := 0 to FCount - 1 do
    WriteLn(FLogFile, FMsgs[I]);
end;

{$ENDIF}


{*************************************************************************}
{                    New Get/Realloc/Free Mem functions                   }
{*************************************************************************}


function NewGetMem(Size: Integer): Pointer;
begin
  Result := UsedMemoryManager.GetMem(Size);
{$IFDEF DEBUG}
  if DoDebug and (Result <> nil) then
  begin
    Inc(Allocations);
    Inc(GetMemCount);
    Block.Address := Result;
    Block.Size := Size;
    Block.AllocOrder := Allocations;
    Block.Reallocated := 0;
    List.Add(Block);
{$IFDEF LOGTOFILES}
    DebugLog.AddAllocMsg(Block);
{$ENDIF}
  end;
{$ELSE}
  Inc(GetMemCount);
{$ENDIF}
end;

function NewFreeMem(P: Pointer): Integer;
{$IFDEF DEBUG}
var
  I: Long;
{$ENDIF}
begin
  Result := -1;
  if P <> nil then
  begin
    Result := UsedMemoryManager.FreeMem(P);
{$IFDEF DEBUG}
    if DoDebug and List.Find(P, I) then
    begin
{$IFDEF LOGTOFILES}
      DebugLog.AddFreeMsg(List.FBlocks[I]);
{$ENDIF}
      Inc(FreeMemCount);
      List.Delete(P);
    end;
{$ELSE}
    Inc(FreeMemCount);
{$ENDIF}
  end;
end;

function NewReallocMem(P: Pointer; Size: Integer): Pointer;
{$IFDEF DEBUG}
var
  I: Long;
{$ENDIF}
begin
  Result := UsedMemoryManager.ReallocMem(P, Size);
{$IFDEF DEBUG}
  if DoDebug and List.Find(P, I) then
  begin
    if P <> Result then
    begin
      if Result = nil then
      begin
{$IFDEF LOGTOFILES}
        DebugLog.AddFreeMsg(List.FBlocks[I]);
{$ENDIF}
        List.Delete(P);
      end
      else
      begin
        Block := List.FBlocks[I];
        List.Delete(P);
        Block.Address := Result;
        Block.Size := Size;
        Inc(Block.Reallocated);
        List.Add(Block);
{$IFDEF LOGTOFILES}
        DebugLog.AddReallocMsg(Block);
{$ENDIF}
      end;
    end
    else
    begin
      List.Find(P, I);
      Inc(ReallocMemCount);
      List.FBlocks[I].Size := Size;
      Inc(List.FBlocks[I].Reallocated);
{$IFDEF LOGTOFILES}
      DebugLog.AddReallocMsg(List.FBlocks[I]);
{$ENDIF}
    end;
  end;
{$ELSE}
  Inc(ReallocMemCount);
{$ENDIF}
end;

procedure SetMemMgr;
begin
  GetMemoryManager(StdMemoryManager);
{$IFDEF MSWINDOWS}
  UsedMemoryManager := Win32MemMgrs.ActiveMemoryManager^;
  DebugMemoryManager := Win32MemMgrs.DebugMemoryManager^;
{$ENDIF}
{$IFDEF LINUX}
  UsedMemoryManager := LinuxMemMgrs.ActiveMemoryManager^;
  DebugMemoryManager := LinuxMemMgrs.DebugMemoryManager^;
{$ENDIF}
  // FPC's MemMgr contains more than Get/Free/Realloc so we keep additional
  // functions in new manager too
  NewMemoryManager := StdMemoryManager;
  NewMemoryManager.GetMem := NewGetMem;
  NewMemoryManager.FreeMem := NewFreeMem;
  NewMemoryManager.ReallocMem := NewReallocMem;
  SetMemoryManager(NewMemoryManager);
end;

procedure UnSetMemMgr;
begin
  SetMemoryManager(StdMemoryManager);
end;


{*************************************************************************}
{                         Internal stuff                                  }
{*************************************************************************}


initialization

  SetMemMgr;
{$IFDEF DEBUG}
  List.Create;
{$IFDEF LOGTOFILES}
{$IFNDEF MODLIBRARY}
  try
    DebugLog.Create;
  except
    // do not show anything
  end;
{$ENDIF}
{$ENDIF}
  DoDebug := True;
{$ENDIF}

finalization

{$IFDEF DEBUG}
  DoDebug := False;
{$IFDEF LOGTOFILES}
{$IFNDEF MODLIBRARY}
  try
    DebugLog.Destroy;
  except
    // do not show anything
  end;
{$ENDIF}
{$ENDIF}
  List.Destroy;
{$ENDIF}
  UnSetMemMgr;

end.

