{*************************************************************************}
{*   Objects                                                             *}
{*   Copyright (C) 2003 Marek Mauder, Vampyre SoftWorks                  *}
{*   Engine\Kernel\Objects.pas                                           *}
{*************************************************************************}
{                                                                         }
{ This program is free software; you can redistribute it and/or modify it }
{ under the terms of the GNU General Public License as published by the   }
{ Free Software Foundation; either version 2 of the License, or (at your  }
{ option) any later version. This program is distributed in the hope that }
{ it will be useful, but WITHOUT ANY WARRANTY; without even the implied   }
{ warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See   }
{ the GNU General Public License for more details.                        }
{                                                                         }
{ You should have received a copy of the GNU General Public License       }
{ along with this program; if not, write to the Free Software Foundation, }
{ Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.         }
{                                                                         }
{*************************************************************************}
{  Created On: 12nd August 2003  (12.08.2003)                             }
{*************************************************************************}

{This unit contains base, moreless abstract classes declarations.
 There are three types of classes in Numalian Engine:
  1) Singletons - there is always one instance of this class. Used mainly
      as plugins, engines, ...
  2) Stored - like TPersistent's descendants in VCL. Their published properties
      can be saved. Useful for GUI forms, Images, other  data objects ...
  3) Other - like lists, streams and other.

 There are also redefinitions of some std Delphi's classes with our names.
 You also do not need to add 'Classes, Objects' but only 'Objects' to uses.}

unit Objects;

{ TODOS:
    - Add BeginUpdate/Changing to lists.
    - Implement Assign in lists.
    - Add OwnObjects to object lists
    - Add FindInstanceOf(AClass:TClass) to Object lists,
        add TNCustomObjectLists, add TNClassList
    - Remove buffer from TNLog, use fmemorylog instead,
        solve "Umajo ; Umajo (Added 3 times).
    - Try creating fake CDROM in NERO and put Exe.
        on it. Add some donotstoresetttingsandlogs option for this case.
    - Do not forget to use message handler procedures.
    * Move ShowError to Kernel
    - Add comments to implementation
}

{$I Compiler.inc}

interface

uses
  SysUtils, Classes, TypInfo, Kernel;


{*************************************************************************}
{                         Types/Constants                                 }
{*************************************************************************}


const

  {Some usefull strings.}
  sExceptMsg = 'Exception Message: ';
  sCustom = 'Custom';
  sAbstract = 'Abstract';
  {Default capacity of Lists.}
  DefaultListCapacity = 16;
  {Maximal number of List's items.}
  MaxListSize = MaxLong div 16;

  {Stream Types/Consts}
  fmCreate = Classes.fmCreate;
  fmOpenRead = SysUtils.fmOpenRead;
  fmOpenWrite = SysUtils.fmOpenWrite;
  fmOpenReadWrite = SysUtils.fmOpenReadWrite;
  soFromBeginning = Classes.soFromBeginning;
  soFromCurrent = Classes.soFromCurrent;
  soFromEnd = Classes.soFromEnd;

type

  {In some unit this were only things used from SysUtils. Until now...}
  Exception = SysUtils.Exception;
  ExceptClass = SysUtils.ExceptClass;


{*************************************************************************}
{                         Base Classes                                    }
{*************************************************************************}


  {Basic class from which all other shalt arise.
   It also supports interfaces.}
  TNBaseObject = class(TInterfacedObject)
  protected
    FID: ULong;
    function GetID: ULong; virtual;
    {Each object can receive messages and this procedure
     can process them. You can use message methods
     (procedure Umajo(var Msg);message KM_Umajo;)
     as used in VCL but it is needless to write new procs for short message
     handlers - so put them in one (this) procedure.
     Do not call directly !! - use Dispatch method always.
     Be careful with sending messages like KM_Free, ...}
    procedure Perform(var Message: TMsgRec); virtual;
  public
    constructor Create; virtual;
    destructor Destroy; override;
    {Overrides basic TObject's message dispatching method so
     it calls Perform or associated message methods.}
    procedure DefaultHandler(var Message); override;
    {Objects Unique identifier.}
    property ID: ULong read GetID;
  end;

  {Class refernce for TNBaseObject.}
  TNBaseObjectClass = class of TNBaseObject;

  {Base class for all singletons in engine.}
  TNSingleton = class(TNBaseObject)
  protected
    FName: NString;
    FDescription: NString;
    function GetName: NString; virtual;
    function GetDescription: NString; virtual;
    procedure SetName(const Value: NString); virtual;
    procedure SetDescription(const Value: NString); virtual;
  public
    constructor Create; override;
    destructor Destroy; override;
    {Singleton's name. Should be unique in descendants.}
    property Name: NString read GetName write SetName;
    {Short description of singleton's function.}
    property Description: NString read GetDescription write SetDescription;
  end;

  {Class refernce for TNSingleton.}
  TNSingletonClass = class of TNSingleton;

  TNFiler = class;
  TNAbstractReader = class;
  TNAbstractWriter = class;

{$M+} // Runtime type info: on

  {Base class for all 'stored' objects. Published members
   of these classes can be streamed similarly as TPersistent
   descendants in VCL.}
  TNPersistent = class(TNBaseObject)
  protected
    {Assigns self to destination persistent object.}
    procedure AssignTo(Dest: TNPersistent); virtual;
  public
    {Assigns from source persistent object}
    procedure Assign(Source: TNPersistent); virtual;
    {Defines new property for streaming. Use
     Filer.DefineProperty() in overrided methods.}
    procedure DefineProperties(Filer: TNFiler); virtual;
  end;

{$M-} // Runtime type info: off

  {Class refernce for TNPersistent.}
  TNPersistentClass = class of TNPersistent;

  {Persistent with Name.}
  TNNamedPersistent = class(TNPersistent)
  protected
    FName: NString;
    procedure SetName(const Value: NString); virtual;
  published
    property Name: NString read FName write SetName;
  end;


{*************************************************************************}
{                         Streaming support                               }
{*************************************************************************}


  {Just to keep 'N' in class names.}
  TNStream = Classes.TStream;
  TNCustomMemoryStream = Classes.TCustomMemoryStream;
  TNMemoryStream = Classes.TMemoryStream;
  TNHandleStream = Classes.THandleStream;
  TNFileStream = Classes.TFileStream;
  TNResourceStream = Classes.TResourceStream;
  TNStringStream = Classes.TStringStream;

  {Procedure used by TNFilers to read custom properties.}
  TReadDataProc = procedure(Reader: TNAbstractReader) of object;
  {Procedure used by TNFilers to write custom properties.}
  TWriteDataProc = procedure(Writer: TNAbstractWriter) of object;

  TNPointerStack = class;

  {Base class for streaming persistent objects and other variables.
   Descendants are TNReader and TNWriter in Streaming.pas.}
  TNFiler = class(TNBaseObject)
  protected
    FStream: TNStream;
    FRoot: TNPersistent;
    FLocalRoot: TNPersistent;
    FRootStack: TNPointerStack;
    procedure SetRoot(Value: TNPersistent);
  public
    constructor Create(AStream: TStream); reintroduce; virtual;
    destructor Destroy; override;
    {This method updates filer's buffer or stream.
     In TNWriter it writes its buffer to stream. In TNReader reads
     buffer from stream. Call this method before/after reading/writing
     to be sure that buffers and streams are up to date.}
    procedure Flush; virtual; abstract;
    {Defines new property to be read/written when streaming persistent
     objects. Look at overrided methods in TNReader and TNWriter.
     It is called by TNPersistent.DefineProperties (in descendants).}
    procedure DefineProperty(const Name: NString; ReadData: TReadDataProc;
      WriteData: TWriteDataProc); virtual; abstract;
    {Used when loading nested objects from stream.}
    property Root: TNPersistent read FRoot write SetRoot;
    {Used when loading nested objects from stream.}
    property LocalRoot: TNPersistent read FLocalRoot;
  end;

  {Pure abstract class which contains only public methods of TNReaders.
   It exists because of TReadDataProc which must take filer
   with reading capabilities as prameter but it can not be
   directly TNReader because it is in another unit.}
  TNAbstractReader = class(TNFiler)
  public
    {Returns true if next value in stream is list of values of
     given name. If so begins reading of list. Must be accompanied
     with EndList() call}
    function BeginList(const Name: NString): Boolean; virtual; abstract;
    {Returns true if end of list is next and ends reading of this list.}
    function EndList: Boolean; virtual; abstract;
    {ReadXXX functions read data from the form
     specific for given TNReader descendant.}
    function ReadString: NString; virtual; abstract;
    function ReadIdent: NString; virtual; abstract;
    function ReadChar: Char; virtual; abstract;
    function ReadInt: Int64; virtual; abstract;
    function ReadBool: Boolean; virtual; abstract;
    function ReadFloat: Double; overload; virtual; abstract;
    {Returns nil when fails, otherwise returns newly allocated pointer}
    procedure ReadBinary(var Data: Pointer; var Count: Long); virtual; abstract;
    {Reads persistent object with all its subobjects and properties.
     Returns nil when fails, otherwise returns newly created instance of
     read class.}
    procedure ReadInstance(var Instance: TNPersistent); virtual; abstract;
  end;

  {Pure abstract class which contains only public methods of TNWriters.
   It exists because of TWriteDataProc which must take filer
   with writing capabilities as prameter but it can not be
   directly TNWriter because it is in another unit.}
  TNAbstractWriter = class(TNFiler)
  public
    {Begins new list of values. Next calls to WriteXXX will
     add data to active list (if any). E.g. in TNXMLWriter
     it creates new node. List must be ended with EndList() call.}
    procedure BeginList(const Name: NString); virtual; abstract;
    {Ends active list.}
    procedure EndList; virtual; abstract;
    {WriteXXX procedures write their parameters in the form
     specific for given TNWriter descendant. But it always
     writes also info about parameter's type or size.}
    procedure WriteString(const Value: NString); virtual; abstract;
    procedure WriteIdent(const Value: NString); virtual; abstract;
    procedure WriteChar(const Value: Char); virtual; abstract;
    procedure WriteInt(const Value: Int64); overload; virtual; abstract;
    procedure WriteBool(const Value: Boolean); virtual; abstract;
    procedure WriteFloat(const Value: Double); overload; virtual; abstract;
    procedure WriteBinary(Data: Pointer; Count: Long); virtual; abstract;
    {Wrtites object to stream with all its subobjects and properties.}
    procedure WriteInstance(Instance: TNPersistent); virtual; abstract;
  end;


{*************************************************************************}
{                            Lists                                        }
{*************************************************************************}


  {Notifications of Lists' actions. Used in TNCustomList.Notify.}
  TListNotification = (lnAdded, lnExtracted, lnDeleted, lnKeyAdded,
    lnValueAdded, lnKeyDeleted, lnValueDeleted);

  {Behaviour of List when adding new items which already reside
   in List. If dupAccept item is accepted to list, otherwise
   it is ignored.}
  TDuplicates = (dupAccept, dupIgnore);

  TNAbstractList = class(TNPersistent)
  end;

  {Base class for all Lists. It can store items of custom size.
   Do not use directly. Duplicates are accepted as default.}
  TNCustomList = class(TNAbstractList)
  protected
    FData: Pointer;
    FCount: Long;
    FItemSize: Long;
    FCapacity: Long;
    FDuplicates: TDuplicates;
    procedure GetItem(Index: Long; var Item); virtual;
    procedure PutItem(Index: Long; const Item); virtual;
    function AddItem(const Item): Long; virtual;
    procedure InsertItem(Index: Long; const Item); virtual;
    function IndexOfItem(const Item): Long; virtual;
    procedure SetCount(Value: Long); virtual;
    procedure SetCapacity(Value: Long); virtual;
    {This is called for every item added/deleted to/from list.
     You can use it for example for freeing some memory contained
     in items.}
    procedure Notify(Index: Long; Action: TListNotification); virtual;
    function GetMemoryUsage: Long; virtual;
    {Support for streaming.}
    procedure ReadData(Reader: TNAbstractReader); virtual;
    {Support for streaming.}
    procedure WriteData(Writer: TNAbstractWriter); virtual;
  public
    {Creates new list with items of size AItemSize.}
    constructor Create(AItemSize: Long); reintroduce; overload;
    destructor Destroy; override;
    {Deletes all items in list.}
    procedure Clear; virtual;
    {Exchanges items at given positions.}
    procedure Exchange(Index1, Index2: Long); virtual;
    {Deletes item at given index.}
    procedure Delete(Index: Long); virtual;
    {Makes more space for items. It is called automatically
     in Add/Insert functions if list is full.}
    procedure Grow; virtual;
    {Saves contents of list to file (uses SaveToStream for writing itself).
     Functionality changes in each descendant.}
    procedure SaveToFile(const FileName: NString);
    {Saves contents of list to stream. Functionality changes in
     each descendant.}
    procedure SaveToStream(Stream: TNStream); virtual;
    {Loads contents of list from file (uses LoadFromStream for reading itself).
    Functionality changes in each descendant.}
    procedure LoadFromFile(const FileName: NString);
    {Loads contents of list from stream. Functionality changes in
     each descendant.}
    procedure LoadFromStream(Stream: TNStream); virtual;
    {Defines functions for streaming.}
    procedure DefineProperties(Filer: TNFiler); override;
    {Number of items currently in list.}
    property Count: Long read FCount write SetCount;
    {Capacity of list in items.}
    property Capacity: Long read FCapacity write SetCapacity;
    {Controls if list accepts items which are already in list
     (only in Add/Insert methods, not in List.Items[I]:=F.}
    property Duplicates: TDuplicates read FDuplicates write FDuplicates;
    {Memory used by all list's items in bytes.}
    property MemoryUsage: Long read GetMemoryUsage;
    {Pointer which stores all items. Be careful with this.}
    property Data: Pointer read FData;
  end;

  {Base list for list storing single value items (numbers, records,
   pointers, ...). It can save/load items from files/streams and
   supports streaming with TNReader/TNWriter.
   Duplicates are always checked in this list (in Add/Insert).}
  TNCustomValueList = class(TNCustomList)
  protected
    {Support for streaming.}
    procedure ReadData(Reader: TNAbstractReader); override;
    {Support for streaming.}
    procedure WriteData(Writer: TNAbstractWriter); override;
  public
    {Adds new item to list.}
    function Add(const Item): Long;
    {Inserts new item to list.}
    procedure Insert(Index: Long; const Item);
    {Returns index of given item. Returns -1 if item is not in list.}
    function IndexOf(const Item): Long;
    {Saves contents of list to stream in binary format.}
    procedure SaveToStream(Stream: TNStream); override;
    {Loads contents of list from stream in binary format.}
    procedure LoadFromStream(Stream: TNStream); override;
  end;

  {Base list for lists which store pointers (objects).}
  TNCustomPointerList = class(TNCustomValueList)
  protected
    function GetPointer(Index: Long): Pointer;
    procedure PutPointer(Index: Long; Value: Pointer);
  public
    constructor Create; override;
    {Adds new pointer to list.}
    function Add(Item: Pointer): Long;
    {Inserts new pointer to list.}
    procedure Insert(Index: Long; Item: Pointer);
    {Returns index of given item. Returns -1 if item is not in list.}
    function IndexOf(Item: Pointer): Long;
  end;

  {List for storing Pointers (as TList in Classes.pas).}
  TNPointerList = class(TNCustomPointerList)
  public
    property Items[Index: Long]: Pointer read GetPointer write PutPointer;
    default;
  end;

  {This is base class for lists where some value is associated
   with string Key (e.g. TStringList in Classes.pas).
   This base class supports saving to files and support
   for TNWriter, but reading must be implemented by descendants.
   There are some things you should now:
    - Duplicates are checked only if Sorted is True
    - PutKey works only if Sorted is False
    - if Sorted then Insert calls Add instead
    - Exchange calls Sort (after exchange) if Sorted is True.
    - LoadFromStream reads from Stream.Position to end of stream.}
  TNCustomStringKeys = class(TNCustomList)
  protected
    FSorted: Boolean;
    FCaseSensitive: Boolean;
    function GetKey(Index: Long): NString; virtual; abstract;
    procedure PutKey(Index: Long; const Value: NString); virtual; abstract;
    procedure Notify(Index: Long; Action: TListNotification); override;
    procedure SetSorted(Value: Boolean);
    procedure SetCaseSensitive(const Value: Boolean);
    {AnsiStrings must be finalized on deletion. Called by
     Notify(Idx, lnDeleted).}
    procedure FinalizeItem(Index: Long); virtual;
    {Returns string created from contents of one item.}
    function GetLine(Index: Long): NString; virtual;
    {Adds item to list created from S (created by GetLine).}
    procedure AddFromLine(const S: NString); virtual;
    {Functions for comparing string which use GetKey() to
     access Keys which is not very fast. In descendants they are
     mostly overriden with ones which access to Keys directly
     (this is not possible here because we dont know how Keys are stored).}
    function Compare(Index1, Index2: Long): Long; overload; virtual;
    function Compare(Index: Long; const S: NString): Long; overload; virtual;
    procedure WriteData(Writer: TNAbstractWriter); override;
    function GetText: NString; virtual;
    procedure SetText(const Value: NString); virtual;
  public
    {Returns index of item with given Key. Returns -1 if item is not in list.}
    function IndexOf(const AKey: NString): Long; virtual;
    {Finds index of item with given Key and returns its index
     in Index. If item is found True is returned.}
    function Find(const AKey: NString; var Index: Long): Boolean; virtual;
    {Sorts items in list.}
    procedure Sort; virtual;
    {Exchanges two items and if list is sorted then list
     is resorted after exchange.}
    procedure Exchange(Index1, Index2: Long); override;
    {Saves items to stream. Uses GetLine to get strings to save.}
    procedure SaveToStream(Stream: TNStream); override;
    {Loads items from stream. Uses AddFromLine to add item from read strings.
     Reads until end of stream is encountered!}
    procedure LoadFromStream(Stream: TNStream); override;
    {Array of Keys in oreder as they are stored in list.}
    property Keys[Index: Long]: NString read GetKey write PutKey; default;
    {If list is sorted then Find and IndexOf functions are much faster
     than in non-sorted list but adding items to list is slower and
     Keys[Index]:=S does not work at all. }
    property Sorted: Boolean read FSorted write SetSorted;
    {If CaseSensitive then comparing of keys (in Sort/Find/IndexOf) is
     CasesSensitive otherwise not.}
    property CaseSensitive: Boolean read FCaseSensitive write SetCaseSensitive;
    {Represents all strings and values in list as a single string.
     Uses GetLine and AddFromLine to create/load resulting string.}
    property Text: NString read GetText write SetText;
  end;

  TStringRecord = record
    Key: NString;
  end;
  PStringRecord = ^TStringRecord;

  TStrRecArray = array[0..MaxListSize - 1] of TStringRecord;
  PStrRecArray = ^TStrRecArray;

  {This is list for storing only string Keys with no
   associated values. After construction this list is not sorted
   and is case insensitive.}
  TNStringList = class(TNCustomStringKeys)
  protected
    function GetKey(Index: Long): NString; override;
    procedure PutKey(Index: Long; const Value: NString); override;
    procedure InsertInternal(Index: Long; const S: NString);
    procedure FinalizeItem(Index: Long); override;
    function Compare(Index1, Index2: Long): Long; override;
    function Compare(Index: Long; const S: NString): Long; override;
    function GetMemoryUsage: Long; override;
    function GetLine(Index: Long): NString; override;
    procedure AddFromLine(const S: NString); override;
    procedure ReadData(Reader: TNAbstractReader); override;
    procedure WriteData(Writer: TNAbstractWriter); override;
  public
    {Creates list  with default item size of 4 B.}
    constructor Create; override;
    {Adds new string to list.}
    function Add(const AKey: NString): Long;
    {Adds string at desired position.}
    procedure Insert(Index: Long; const AKey: NString);
    {Adds Keys from List to TNStringList.}
    procedure AddStrings(List: TNCustomStringKeys);
  end;

  TStrPtrRecord = record
    Key: NString;
    Value: Pointer;
  end;
  PStrPtrRecord = ^TStrPtrRecord;

  TStrPtrArray = array[0..MaxListSize - 1] of TStrPtrRecord;
  PStrPtrArray = ^TStrPtrArray;

  {Base list for list which associate pointers (and objects) to
   string Keys. Only Key strings are saved/streamed
   in SaveTo/WriteData and when loading only Key strings
   are loaded - Add(LoadedKey, nil). }
  TNCustomStringPointerList = class(TNCustomStringKeys)
  protected
    function GetKey(Index: Long): NString; override;
    procedure PutKey(Index: Long; const Value: NString); override;
    function GetPointer(Index: Long): Pointer;
    procedure PutPointer(Index: Long; const Value: Pointer);
    procedure InsertInternal(Index: Long; const AKey: NString; Item: Pointer);
    procedure FinalizeItem(Index: Long); override;
    function Compare(Index1, Index2: Long): Long; override;
    function Compare(Index: Long; const S: NString): Long; override;
    function GetMemoryUsage: Long; override;
    function GetLine(Index: Long): NString; override;
    procedure AddFromLine(const S: NString); override;
    procedure ReadData(Reader: TNAbstractReader); override;
  public
    constructor Create; override;
    {Returns index of given Value. Returns -1 if item is not in list.}
    function IndexOfValue(Value: Pointer): Long;
    {Adds new Key to list with associated Pointer. }
    function Add(const AKey: NString; Value: Pointer): Long;
    {Inserts new Key to list with associated Pointer at given position.}
    procedure Insert(Index: Long; const AKey: NString; Value: Pointer);
    {Returns pointer Value associtaed with Key string. If Key is not
     found, nil is returned.}
    function FindValue(const AKey: NString): Pointer;
  end;

  {List for storing Pointers associated to string Keys.}
  TNStringPointerList = class(TNCustomStringPointerList)
  public
    property Pointers[Index: Long]: Pointer read GetPointer write PutPointer;
  end;

  {List for storing Objects associated to string Keys.
   This works as TStringList from Classes.pas.}
  TNStringObjectList = class(TNCustomStringPointerList)
  protected
    function GetObject(Index: Long): TObject;
    procedure PutObject(Index: Long; Value: TObject);
  public
    property Objects[Index: Long]: TObject read GetObject write PutObject;
  end;

  {Standard TNStringList which is set to Sorted in constructor.
   Duplicate items are not added to list.}
  TNSortedStrings = class(TNStringList)
  public
    constructor Create; override;
  end;

  {Standard TNStringPointerList which is set to Sorted in constructor.
   Duplicate items are not added to list.}
  TNSortedStringPointers = class(TNStringPointerList)
  public
    constructor Create; override;
  end;

  {Standard TNStringObjectList which is set to Sorted in constructor.
   Duplicate items are not added to list}
  TNSortedStringObjects = class(TNStringObjectList)
  public
    constructor Create; override;
  end;

  {Custom ordered list which can be used for implementations of
   stacks and queus.}
  TNCustomOrderedList = class(TNAbstractList)
  protected
    FList: TNCustomValueList;
    {Adds item to list at position defined by implementation.}
    procedure PushItem(const Item); virtual; abstract;
    {Returns last item in list and deletes it from list.}
    procedure PopItem(var Item);
    {Returns last item in list.}
    procedure PeekItem(var Item);
  public
    {Creates new list with items of size AItemSize.}
    constructor Create(AItemSize: Long); reintroduce; overload;
    destructor Destroy; override;
    {Deletes all items in list.}
    procedure Clear;
    {Returns number of items in list.}
    function Count: Long;
  end;

  {Custom LIFO stack implementation.}
  TNCustomStack = class(TNCustomOrderedList)
  protected
    {Adds item at the end of the list.}
    procedure PushItem(const Item); override;
  end;

  {Custom queue implementation.}
  TNCustomQueue = class(TNCustomOrderedList)
  protected
    {Adds item at the beginning of the list.}
    procedure PushItem(const Item); override;
  end;

  {Stack for storing Pointers.}
  TNPointerStack = class(TNCustomStack)
  public
    constructor Create; override;
    procedure Push(Item: Pointer);
    function Pop: Pointer;
    function Peek: Pointer;
  end;

  {Queue for storing Pointers.}
  TNPointerQueue = class(TNCustomQueue)
  public
    constructor Create; override;
    procedure Push(Item: Pointer);
    function Pop: Pointer;
    function Peek: Pointer;
  end;


{*************************************************************************}
{                          Other classes                                  }
{*************************************************************************}


const

  {Max number of lines in TNLog's message buffer.}
  MaxLogBuffer = {$IFNDEF DEBUG}4{$ELSE}1{$ENDIF};
  {Max number of characters per line. Messages are wrapped after this position.}
  MaxLogLine = 80;
  {Extension of TNLog file.}
  sLogFileExt = 'log';
  {Char used to create linebreaks in log file.}
  sLogLBChar = '-';
  {Log file Ident increase/decrease.}
  sLogIdent = #32#32;

type

  {Callback procedure called when new line is added to TNLog.}
  TOnAddMsg = procedure(const Msg: NString);

  {Class for buffered storing messages during program flow.
   Supports nested and repeated messages handling. Wraps messages
   when they exceed MaxLogLine length.
   It adds messages to FBuffer and when it is 'full' writes messages
   to file (with name AFileName+sLogFileExt).}
  TNLog = class(TNSingleton)
  protected
    FMemoryLog: TNStringList;
    FBuffer: TNStringList;
    FBufMax: Long;
    FStream: TNFileStream;
    FIdent: NString;
    FLastMsg: NString;
    FStartTime: Int64;
    FRepeats: Long;
    FCheckRepeats: Boolean;
    FWriteToFile: Boolean;
    FLineLength: Long;
    FFileName: NString;
    FOnAddMsg: TOnAddMsg;
    procedure AddStartup; virtual;
    procedure AddShutdown; virtual;
    procedure IncIdent;
    procedure DecIdent;
    procedure WriteToBuffer(const S: NString);
  public
    {Creates new log file. If file can not be created logging is done
     into memory (stringlist). FileName extension is always overrided
     by sLogExt constant.}
    constructor Create(const AFileName: NString); reintroduce; virtual;
    destructor Destroy; override;
    {Writes buffer to file (if enabled) and clears it (always).}
    procedure Flush;
    {Adds simple string message. When buffer is full calls Flush().}
    procedure Add(const Msg: NString); overload;
    {Adds formatted string message.}
    procedure Add(const MsgFmt: NString; Args: array of const); overload;
    {Adds message with current time and microseconds from program startup.}
    procedure AddTime(const Msg: NString);
    {Adds error message.}
    procedure AddError(const Msg: NString);
    {Adds warning message.}
    procedure AddWarning(const Msg: NString);
    {Adds Msg then increases Ident. Accompany it with EndSubList() call.}
    procedure BeginSubList(const Msg: NString);
    {Decreases Ident and adds Msg.}
    procedure EndSubList(const Msg: NString);
    {Adds (MaxLogLine-Length(Ident)) long line break of sLBChar.}
    procedure AddLineBreak;
    {Current Ident.}
    property Ident: NString read FIdent;
    {If True repeating messages are written only once with number of
     repeats added.}
    property CheckRepeats: Boolean read FCheckRepeats write FCheckRepeats;
    {Name of file to which this log instance is assigned. Set by constructor.}
    property FileName: NString read FFileName;
    {Copy of whole log file in memory. Writting to this list is done when
     writting to buffer.}
    property MemoryLog: TNStringList read FMemoryLog;
    {You can assign your own procedure that is called when new msg is written
     to log (can be used e.g. for writting log messages on screen).}
    property OnAddMsg: TOnAddMsg read FOnAddMsg write FOnAddMsg;
  end;

  {Descendant of TNLog. Some stratup and shutdown info is added.
   Accessible trough Log() function. Put Kernel related messages here.}
  TNKernelLog = class(TNLog)
  protected
    procedure AddStartup; override;
    procedure AddShutdown; override;
  public
    constructor Create(const AFileName: NString); override;
    destructor Destroy; override;
  end;


{*************************************************************************}
{                      Plugin/Module Support                              }
{*************************************************************************}


const

  {Name prefix for module libraries. E.g. ModDirect3D.ren, ModGif.iff,...}
  sModulePrefix = 'Mod';
  {Name of TGetIDProc in module library.}
  sGetIDProc = 'GetID';
  {Name of TGetCountProc in module library.}
  sGetCountProc = 'GetCount';
  {Name of TGetModuleProc in module library.}
  sGetModuleProc = 'GetModule';

type

  {Module identifier. Used to determine which type of module is
   in library. Actuall values are defined in units which
   declare some new modules. E.g. in Images.pas it is IffLibraryID = 'iff'.
   You can change library file extension to  its ID to be more transparent.}
  TModuleID = array[0..2] of Char;

  {(*}

  {Record with information about module library. These records are
   stored by TModuleRegister.}
  TModuleEntry = record
    Name: NString;             // Name of module e.g. ModMng.iff
    FileName: NString;         // Full path to module e.g. D:\project\ModMng.iff
    Handle: THandle;          // Handle to loaded library
    ID: TModuleID;            // Module identifier
    Count: Long;              // Number of unique modules in library
  end;
  PModuleEntry = ^TModuleEntry;

  {*)}

  {This procedure must be in all valid module libraries. It returns
   identifier for contained modules.}
  TGetIDProc = procedure(var ID: TModuleID); cdecl;

  {This procedure must be in all valid module libraries. It returns
   number of unique modules in library.}
  TGetCountProc = procedure(var Count: Long); cdecl;

  {This procedure must be in all valid module libraries. It returns
   pointer to module identified by Index (0 .. GetCountProc-1).
   It can point to function, class, record, ... }
  TGetModuleProc = procedure(Index: Long; var Module: Pointer); cdecl;

  {This callback (list of them stored in TNModuleRegister) is
   called when new module is loaded form library. Based on
   module's ID callback performs some things (e.g. registers
   new ImageFileFormat so it can be used by image loading/saving).}
  TModLoadCallBack = procedure(Entry: PModuleEntry; var Done: Boolean);
  {This callback (list of them stored in TNModuleRegister) is
   called when module (from library) is unloaded. Based on
   module's ID callback performs some things (e.g. unregisters
   new ImageFileFormat so it can NOT be used by image loading/saving any more).}
  TModUnLoadCallBack = procedure(Entry: PModuleEntry; var Done: Boolean);

  {Base class for modules.}
  TNModule = class(TNBaseObject)
  end;

  {Class reference for TNModule.}
  TNModuleClass = class of TNModule;

  {Base class for file formats loading/saving modules.}
  TNFileModule = class(TNModule)
  protected
    FExtensions: NString;
    FFormatName: NString;
    FCanLoad: Boolean;
    FCanSave: Boolean;
  public
    {This function tests if input data are in supported format. It must then
     set stream's position to original pos.}
    function TestFormat(Stream: TNStream): Boolean; virtual;
    {File name extensions used by this format.
     Separate them by semicolon or comma.
     Example:= 'jpg,jpeg,jfif'}
    property Extensions: NString read FExtensions;
    {Name of file format (e.g. 'Portable Network Graphics').}
    property FormatName: NString read FFormatName;
    {Indicates that class can load files.}
    property CanLoad: Boolean read FCanLoad;
    {Indicates that class can save files.}
    property CanSave: Boolean read FCanSave;
  end;

  {List for storing file module instances. For example in Images.pas is
   TNImageFileFormats to hold registered image file formats.
   They are mostly declared in implementation part of units}
  TNFileModuleList = class(TNBaseObject)
  protected
    FModules: TNPointerList;
    FDefault: TNFileModule;
    function GetDefault: TNFileModule;
  public
    constructor Create; override;
    destructor Destroy; override;
    {Adds new module to list.}
    procedure Add(Module: TNFileModule);
    {Removes module to list.}
    procedure Remove(Module: TNFileModule);
    {Returns True if given module is in list.}
    function Find(Module: TNFileModule): Boolean; overload;
    {Finds module which supports given file extension. If not
     fount nil is returned.}
    function Find(const Ext: NString): TNFileModule; overload;
    {List of all registered modules.}
    property Modules: TNPointerList read FModules;
    {Default module - by default it is set to module at index 0
     can can be changed as you want. Defaults are for example
     TGA file format for images and ZIP archive format fo archives.}
    property Default: TNFileModule read GetDefault write FDefault;
  end;

{$IFNDEF MODLIBRARY}

  {This class manages modules from external libraries. It stores
   TModuleEntry record for each loaded module and calls
   callback functions when module is loaded/unloaded.}
  TNModuleRegister = class(TNSingleton)
  protected
    FList: TNSortedStringPointers;
    FLoadCalls: TNPointerList;
    FUnLoadCalls: TNPointerList;
    FDefaultDirectory: NString;
    {Unloads all modules.}
    procedure UnLoadAll;
    {Runs all registered module load callbacks.}
    procedure RunLoadCallBacks(Entry: PModuleEntry);
    {Runs all registered module unload callbacks.}
    procedure RunUnLoadCallBacks(Entry: PModuleEntry);
  public
    constructor Create; override;
    destructor Destroy; override;
    {Loads new module from FileName and runs load callbacks.}
    procedure Load(const FileName: NString);
    {Loads all modules from Dir.}
    procedure LoadFromDirectory(const Dir: NString);
    {UnLoads module with given FileName and runs unload callbacks.}
    procedure UnLoad(const FileName: NString);
    {Adds new load callback.}
    procedure AddLoadCallBack(Call: TModLoadCallBack);
    {Adds new unload callback.}
    procedure AddUnLoadCallBack(Call: TModUnLoadCallBack);
    {Returns True if given module is loaded.}
    function IsLoaded(const FileName: NString): Boolean;
    {Returns True if given module is valid module library.}
    function IsValidModuleLibrary(const FileName: NString): Boolean;
    {Default directory where ModuleRegister looks for
     modules with non-fullpath FileName.}
    property DefaultDirectory: NString read FDefaultDirectory;
  end;

{$ENDIF}


{*************************************************************************}
{                         Class/Object functions                          }
{*************************************************************************}


{This function generates new IDs. It is very simple but why complicate it?.
 However it will repeat numbers after creation of 4294k objects.}
function GetNewID: ULong;

{Number of created TNBaseObject descendants from engine startup.}
function GetCreatedObjects: Long;
{Number of free TNBaseObject descendants from engine startup.}
function GetFreedObjects: Long;
{Number of actually "living" TNBaseObject descendants.}
function GetLivingObjects: Long;

{Registers class so it can be used by streaming system.
 Supports classes with same name from different units.}
procedure RegisterClass(AClass: TNPersistentClass);
{Registers multiple classes at once.}
procedure RegisterClasses(AClasses: array of TNPersistentClass);
{Unregisters registered class.
 Supports classes with same name from different units.}
procedure UnRegisterClass(AClass: TNPersistentClass);
{Unegisters multiple classes at once.}
procedure UnRegisterClasses(AClasses: array of TNPersistentClass);
{Returns registered class reference of given classname.
 If not found it returns nil.Supports classes with same name from different
 units.}
function GetClass(AClassName: NString): TNPersistentClass;


{*************************************************************************}
{                           Singleton access                              }
{*************************************************************************}


{This function provides access to TNLog singleton instance.
 I rather use function than global variable becaude it can check
 if instance is valid.}
function Log: TNLog;

{$IFNDEF MODLIBRARY}

{This function provides access to TNModuleRegister singleton instance.
 I rather use function than global variable becaude it can check
 if instance is valid.}
function ModuleRegister: TNModuleRegister;

{$ENDIF}


{*************************************************************************}
{                             Implementation                              }
{*************************************************************************}


implementation

uses
  Platforms;

var

  {Object statistics}
  NowID: ULong = 1;
  ObjCreated: Long = 0;
  ObjFreed: Long = 0;

  {List constaining registered classes.
   Each class used by streaming system must be first registered by
   RegisterClass() call.}
  RegClasses: TNSortedStringObjects = nil;

  {Global instance of Log file accessible trough Log() function
   (it check if instance is created).}
  LogInstance: TNLog = nil;

{$IFNDEF MODLIBRARY}

  {Global instance of ModuleRegister accessible trough ModuleRegister() function
   (it check if instance is created).}
  ModRegInstance: TNModuleRegister = nil;

{$ENDIF}


{*************************************************************************}
{                         Class/Object functions                          }
{*************************************************************************}


function GetNewID: ULong;
begin
  Result := NowID;
  Inc(NowID);
end;

function GetCreatedObjects: Long;
begin
  Result := ObjCreated;
end;

function GetFreedObjects: Long;
begin
  Result := ObjFreed;
end;

function GetLivingObjects: Long;
begin
  Result := Abs(ObjCreated - ObjFreed);
end;

function GetRegClasses: TNSortedStringObjects;
begin
  if RegClasses = nil then
    RegClasses := TNSortedStringObjects.Create;
  Result := RegClasses;
end;

procedure RegisterClass(AClass: TNPersistentClass);
var
  Name: NString;
begin
  if AClass = nil then
    Exit;
  Name := GetTypeData(PTypeInfo(AClass.ClassInfo)).UnitName + '.' +
    AClass.ClassName;
  while GetRegClasses.IndexOf(Name) < 0 do
  begin
    if (not SameStr(sCustom, Copy(Name, PosNoCase('.', Name) + 3, 6)))
      and (not SameStr(sAbstract, Copy(Name, PosNoCase('.', Name) + 3, 8))) then
    begin
      GetRegClasses.Add(Name, TObject(AClass));
      Log.Add('Class registered: ' + Name);
    end;
    if AClass = TNPersistent then
      Break;
    AClass := TNPersistentClass(AClass.ClassParent);
    Name := GetTypeData(PTypeInfo(AClass.ClassInfo)).UnitName + '.' +
      AClass.ClassName;
  end;
end;

procedure RegisterClasses(AClasses: array of TNPersistentClass);
var
  I: Long;
begin
  for I := Low(AClasses) to High(AClasses) do
    RegisterClass(AClasses[I]);
end;

procedure UnRegisterClass(AClass: TNPersistentClass);
var
  FullName: NString;
  I: Long;
begin
  if AClass = nil then
    Exit;
  FullName := GetTypeData(PTypeInfo(AClass.ClassInfo)).UnitName + '.' +
    AClass.ClassName;
  if GetRegClasses.Find(FullName, I) then
  begin
    GetRegClasses.Delete(I);
    Log.Add('Class unregistered: ' + FullName);
  end;
end;

procedure UnRegisterClasses(AClasses: array of TNPersistentClass);
var
  I: Long;
begin
  for I := Low(AClasses) to High(AClasses) do
    RegisterClass(AClasses[I]);
end;

function GetClass(AClassName: NString): TNPersistentClass;
var
  I, P: Long;
  S: NString;
begin
  // first check incomming classname and if not found
  // check classname without its unit name
  Result := nil;
  if AClassName = '' then
    Exit;
  I := -1;
  if not GetRegClasses.Find(AClassName, I) then
  begin
    P := Pos('.', AClassName);
    if P > 0 then
    begin
      S := Copy(AClassName, P + 1, Length(AClassName) - P);
      if not GetRegClasses.Find(AClassName, I) then
        I := -1;
    end;
  end;
  if I >= 0 then
    Result := TNPersistentClass(GetRegClasses.Objects[I]);
end;


{*************************************************************************}
{                           Singleton access                              }
{*************************************************************************}


function Log: TNLog;
begin
  if LogInstance = nil then
    LogInstance := TNKernelLog.Create(ExeName);
  Result := LogInstance;
end;

{$IFNDEF MODLIBRARY}

function ModuleRegister: TNModuleRegister;
begin
  if ModRegInstance = nil then
    ModRegInstance := TNModuleRegister.Create;
  Result := ModRegInstance;
end;

{$ENDIF}


{*************************************************************************}
{                         TNBaseObject Methods                            }
{*************************************************************************}


constructor TNBaseObject.Create;
begin
  FID := GetNewID;
  Inc(ObjCreated);
end;

destructor TNBaseObject.Destroy;
begin
  Inc(ObjFreed);
  inherited Destroy;
end;

function TNBaseObject.GetID: ULong;
begin
  Result := FID;
end;

procedure TNBaseObject.DefaultHandler(var Message);
begin
  inherited DefaultHandler(Message);
  Perform(TMsgRec(Message));
end;

procedure TNBaseObject.Perform(var Message: TMsgRec);
begin
  case Message.MsgID of
    KM_GetID: Message.LParam := ID;
    KM_FreeSelf: Self.Free;
    KM_FreeSender: Message.Sender.Free;
  else
    // in descendants put here this line
    // inherited Perform(..);
    // to perform messages not processed by this handler
  end;
end;


{*************************************************************************}
{                         TNSingleton methods                             }
{*************************************************************************}


constructor TNSingleton.Create;
begin
  inherited Create;
  FName := 'Unknown Singleton';
  FDescription := 'Singleton with unknown capabilities';
end;

destructor TNSingleton.Destroy;
begin
  inherited Destroy;
end;

function TNSingleton.GetName: NString;
begin
  Result := FName;
end;

function TNSingleton.GetDescription: NString;
begin
  Result := FDescription;
end;

procedure TNSingleton.SetName(const Value: NString);
begin
  FName := Value;
end;

procedure TNSingleton.SetDescription(const Value: NString);
begin
  FDescription := Value;
end;


{*************************************************************************}
{                             TNPersistent methods                        }
{*************************************************************************}


procedure TNPersistent.AssignTo(Dest: TNPersistent);
begin
end;

procedure TNPersistent.Assign(Source: TNPersistent);
begin
  if Source <> nil then
    Source.AssignTo(Self);
end;

procedure TNPersistent.DefineProperties(Filer: TNFiler);
begin
end;


{*************************************************************************}
{                          TNNamedPersistent methods                      }
{*************************************************************************}


procedure TNNamedPersistent.SetName(const Value: NString);
begin
  FName := Value;
end;


{*************************************************************************}
{                         TNFiler methods                                 }
{*************************************************************************}


constructor TNFiler.Create(AStream: TStream);
begin
  inherited Create;
  FStream := AStream;
  FRootStack := TNPointerStack.Create;
end;

destructor TNFiler.Destroy;
begin
  FRootStack.Free;
  inherited Destroy;
end;

procedure TNFiler.SetRoot(Value: TNPersistent);
begin
  FRootStack.Clear;
  FRoot := Value;
  FLocalRoot := Value;
end;


{*************************************************************************}
{                         TNCustomList methods                            }
{*************************************************************************}


constructor TNCustomList.Create(AItemSize: Long);
begin
  inherited Create;
  FCount := 0;
  FItemSize := AItemSize;
  SetCapacity(DefaultListCapacity);
  FDuplicates := dupAccept;
end;

destructor TNCustomList.Destroy;
begin
  Clear;
  inherited Destroy;
end;

procedure TNCustomList.GetItem(Index: Long; var Item);
begin
  if (Index > -1) and (Index < FCount) then
    Move(Pointer(Long(FData) + Index * FItemSize)^, Item, FItemSize)
  else
    ZeroMemory(@Item, FItemSize);
end;

procedure TNCustomList.PutItem(Index: Long; const Item);
begin
  if (Index > -1) and (Index < FCount) then
  begin
    Notify(Index, lnDeleted);
    Move(Item, Pointer(Long(FData) + Index * FItemSize)^, FItemSize);
    Notify(Index, lnAdded);
  end;
end;

function TNCustomList.AddItem(const Item): Long;
begin
  Result := FCount;
  if Result = FCapacity then
    Grow;
  Move(Item, Pointer(Long(FData) + Result * FItemSize)^, FItemSize);
  Inc(FCount);
  Notify(Result, lnAdded);
end;

procedure TNCustomList.InsertItem(Index: Long; const Item);
begin
  if (Index > -1) and (Index <= FCount) then
  begin
    if FCount = FCapacity then
      Grow;
    if Index < FCount then
      Move(Pointer(Long(FData) + Index * FItemSize)^,
        Pointer(Long(FData) + (Index + 1) * FItemSize)^,
        (FCount - Index) * FItemSize);
    ZeroMemory(Pointer(Long(FData) + Index * FItemSize), FItemSize);
    Move(Item, Pointer(Long(FData) + Index * FItemSize)^, FItemSize);
    Inc(FCount);
    Notify(Index, lnAdded);
  end;
end;

function TNCustomList.IndexOfItem(const Item): Long;
begin
  for Result := 0 to FCount - 1 do
    if CompareMem(@Item, Pointer(Long(FData) + Result * FItemSize), FItemSize)
      then
      Exit;
  Result := -1;
end;

procedure TNCustomList.SetCount(Value: Long);
var
  I: Integer;
begin
  if (Value > -1) and (Value < MaxListSize) then
  begin
    if Value > FCapacity then
      SetCapacity(Value);
    if Value > FCount then
      ZeroMemory(Pointer(Long(FData) + FCount * FItemSize), (Value - FCount) *
        FItemSize)
    else
      for I := FCount - 1 downto Value do
        Delete(I);
    FCount := Value;
  end;
end;

procedure TNCustomList.SetCapacity(Value: Long);
begin
  if Value <> FCapacity then
  begin
    ReallocMem(FData, Value * FItemSize);
    FCapacity := Value;
  end;
end;

procedure TNCustomList.Notify(Index: Long; Action: TListNotification);
begin
end;

function TNCustomList.GetMemoryUsage: Long;
begin
  Result := FCount * FItemSize;
end;

procedure TNCustomList.ReadData(Reader: TNAbstractReader);
begin
end;

procedure TNCustomList.WriteData(Writer: TNAbstractWriter);
begin
end;

procedure TNCustomList.Clear;
begin
  SetCount(0);
  SetCapacity(0);
end;

procedure TNCustomList.Exchange(Index1, Index2: Long);
var
  Item1: Pointer;
  Item2: Pointer;
begin
  if (Index1 < 0) and (Index1 >= FCount) then
    Exit;
  if (Index2 < 0) and (Index2 >= FCount) then
    Exit;
  GetMem(Item1, FItemSize);
  GetMem(Item2, FItemSize);
  Move(Pointer(Long(FData) + Index1 * FItemSize)^, Item1^, FItemSize);
  Move(Pointer(Long(FData) + Index2 * FItemSize)^, Item2^, FItemSize);
  Move(Item1^, Pointer(Long(FData) + Index2 * FItemSize)^, FItemSize);
  Move(Item2^, Pointer(Long(FData) + Index1 * FItemSize)^, FItemSize);
  FreeMem(Item1);
  FreeMem(Item2);
end;

procedure TNCustomList.Delete(Index: Long);
begin
  if (Index > -1) and (Index < FCount) then
  begin
    Notify(Index, lnDeleted);
    Dec(FCount);
    if Index < FCount then
      System.Move(Pointer(Long(FData) + (Index + 1) * FItemSize)^,
        Pointer(Long(FData) + Index * FItemSize)^,
        (FCount - Index) * FItemSize);
  end;
end;

procedure TNCustomList.Grow;
var
  Delta: Long;
begin
  if FCapacity > 64 then
    Delta := FCapacity shr 2
  else
    Delta := 8;
  SetCapacity(FCapacity + Delta);
end;

procedure TNCustomList.SaveToFile(const FileName: NString);
var
  Stream: TNStream;
begin
  Stream := TNFileStream.Create(FileName, fmCreate);
  try
    SaveToStream(Stream);
  finally
    Stream.Free;
  end;
end;

procedure TNCustomList.SaveToStream(Stream: TNStream);
begin
end;

procedure TNCustomList.LoadFromFile(const FileName: NString);
var
  Stream: TNStream;
begin
  Stream := TNFileStream.Create(FileName, fmOpenRead or fmShareDenyWrite);
  try
    LoadFromStream(Stream);
  finally
    Stream.Free;
  end;
end;

procedure TNCustomList.LoadFromStream(Stream: TNStream);
begin
end;

procedure TNCustomList.DefineProperties(Filer: TNFiler);
begin
  Filer.DefineProperty('Items', ReadData, WriteData);
end;


{*************************************************************************}
{                       TNCustomValueList methods                         }
{*************************************************************************}


procedure TNCustomValueList.ReadData(Reader: TNAbstractReader);
var
  I, Size, Number: Long;
  P: Pointer;
begin
  Number := Reader.ReadInt;
  for I := 0 to Number - 1 do
  begin
    Reader.ReadBinary(P, Size);
    Add(P^);
    FreeMem(P);
  end;
end;

procedure TNCustomValueList.WriteData(Writer: TNAbstractWriter);
var
  I: Long;
  P: Pointer;
begin
  Writer.WriteInt(FCount);
  GetMem(P, FItemSize);
  for I := 0 to FCount - 1 do
  begin
    GetItem(I, P^);
    Writer.WriteBinary(P, FItemSize);
  end;
  FreeMem(P);
end;

function TNCustomValueList.Add(const Item): Long;
begin
  if FDuplicates = dupAccept then
    Result := AddItem(Item)
  else
  begin
    if IndexOfItem(Item) < 0 then
      Result := AddItem(Item)
    else
      Result := -1;
  end;
end;

procedure TNCustomValueList.Insert(Index: Long; const Item);
begin
  if FDuplicates = dupAccept then
    InsertItem(Index, Item)
  else
    if IndexOfItem(Item) < 0 then
      InsertItem(Index, Item);
end;

function TNCustomValueList.IndexOf(const Item): Long;
begin
  Result := IndexOfItem(Item);
end;

procedure TNCustomValueList.SaveToStream(Stream: TNStream);
begin
  Stream.WriteBuffer(FCount, SizeOf(FCount));
  Stream.WriteBuffer(FData^, FCount * FItemSize);
end;

procedure TNCustomValueList.LoadFromStream(Stream: TNStream);
begin
  Clear;
  Stream.ReadBuffer(FCount, SizeOf(FCount));
  SetCapacity(FCount);
  Stream.ReadBuffer(FData^, FCount * FItemSize);
end;


{*************************************************************************}
{                        TNCustomPointerList methods                      }
{*************************************************************************}


constructor TNCustomPointerList.Create;
begin
  inherited Create(SizeOf(Pointer));
end;

function TNCustomPointerList.GetPointer(Index: Long): Pointer;
begin
  if (Index > -1) and (Index < FCount) then
    Result := PPointerArray(FData)[Index]
  else
    Result := nil;
end;

procedure TNCustomPointerList.PutPointer(Index: Long; Value: Pointer);
begin
  if (Index > -1) and (Index < FCount) then
  begin
    Notify(Index, lnDeleted);
    PPointerArray(FData)[Index] := Value;
    Notify(Index, lnAdded);
  end;
end;

function TNCustomPointerList.Add(Item: Pointer): Long;
var
  I: ULong;
begin
  I := ULong(Item);
  Result := inherited Add(I);
end;

procedure TNCustomPointerList.Insert(Index: Long; Item: Pointer);
begin
  inherited Insert(Index, Item);
end;

function TNCustomPointerList.IndexOf(Item: Pointer): Long;
begin
  for Result := 0 to FCount - 1 do
    if Item = PPointerArray(FData)[Result] then
      Exit;
  Result := -1;
end;


{*************************************************************************}
{                     TNCustomStringKeys methods                          }
{*************************************************************************}


procedure TNCustomStringKeys.Notify(Index: Long; Action: TListNotification);
begin
  if Action = lnDeleted then
    FinalizeItem(Index);
end;

procedure TNCustomStringKeys.SetSorted(Value: Boolean);
begin
  if FSorted <> Value then
  begin
    if Value then
      Sort;
    FSorted := Value;
  end;
end;

procedure TNCustomStringKeys.SetCaseSensitive(const Value: Boolean);
begin
  if Value <> FCaseSensitive then
  begin
    FCaseSensitive := Value;
    if FSorted then
      Sort;
  end;
end;

procedure TNCustomStringKeys.FinalizeItem(Index: Long);
begin
end;

function TNCustomStringKeys.GetLine(Index: Long): NString;
begin
  Result := '';
end;

procedure TNCustomStringKeys.AddFromLine(const S: NString);
begin
end;

function TNCustomStringKeys.Compare(Index1, Index2: Long): Long;
begin
  if FCaseSensitive then
    Result := Kernel.CompStrCase(GetKey(Index1), GetKey(Index2))
  else
    Result := Kernel.CompStr(GetKey(Index1), GetKey(Index2));
end;

function TNCustomStringKeys.Compare(Index: Long; const S: NString): Long;
begin
  if FCaseSensitive then
    Result := Kernel.CompStrCase(GetKey(Index), S)
  else
    Result := Kernel.CompStr(GetKey(Index), S);
end;

procedure TNCustomStringKeys.WriteData(Writer: TNAbstractWriter);
var
  I: Long;
begin
  Writer.WriteInt(FCount);
  for I := 0 to FCount - 1 do
    Writer.WriteString(GetKey(I));
end;

function TNCustomStringKeys.GetText: NString;
var
  I, L, Size: Long;
  S: NString;
  List: TNStringList;
  P: PChar;
begin
  List := TNStringList.Create;
  for I := 0 to FCount - 1 do
    List.InsertInternal(List.FCount, GetLine(I));
  Size := 0;
  for I := 0 to FCount - 1 do
    Size := Size + Length(List[I]);
  SetString(Result, nil, Size);
  P := Pointer(Result);
  for I := 0 to FCount - 1 do
  begin
    S := PStrRecArray(List.FData)[I].Key;
    L := Length(S);
    if L <> 0 then
    begin
      System.Move(Pointer(S)^, P^, L);
      Inc(P, L);
    end;
  end;
  List.Free;
end;

procedure TNCustomStringKeys.SetText(const Value: NString);
var
  Start, P: PChar;
  S: NString;
begin
  Clear;
  P := Pointer(Value);
  if P <> nil then
    while P^ <> #0 do
    begin
      Start := P;
      while not (P^ in [#0, #10, #13]) do
        Inc(P);
      SetString(S, Start, P - Start);
      AddFromLine(S);
      if P^ = #13 then
        Inc(P);
      if P^ = #10 then
        Inc(P);
    end;
end;

function TNCustomStringKeys.IndexOf(const AKey: NString): Long;
begin
  if not FSorted then
  begin
    for Result := 0 to FCount - 1 do
      if Compare(Result, AKey) = 0 then
        Exit;
    Result := -1;
  end
  else
    if not Find(AKey, Result) then
      Result := -1;
end;

function TNCustomStringKeys.Find(const AKey: NString; var Index: Long): Boolean;
var
  L, H, I, C: Long;
begin
  Result := False;
  L := 0;
  H := FCount - 1;
  while L <= H do
  begin
    I := (L + H) shr 1;
    C := Compare(I, AKey);
    if C < 0 then
      L := I + 1
    else
    begin
      H := I - 1;
      if C = 0 then
      begin
        Result := True;
        if FDuplicates <> dupAccept then
          L := I;
      end;
    end;
  end;
  Index := L;
end;

procedure TNCustomStringKeys.Sort;

  procedure QuickSort(L, R: Long);
  var
    I, J, P: Long;
  begin
    repeat
      I := L;
      J := R;
      P := (L + R) shr 1;
      repeat
        while Compare(I, P) < 0 do
          Inc(I);
        while Compare(J, P) > 0 do
          Dec(J);
        if I <= J then
        begin
          inherited Exchange(I, J);
          if P = I then
            P := J
          else
            if P = J then
              P := I;
          Inc(I);
          Dec(J);
        end;
      until I > J;
      if L < J then
        QuickSort(L, J);
      L := I;
    until I >= R;
  end;

begin
  if FCount > 1 then
    QuickSort(0, FCount - 1)
end;

procedure TNCustomStringKeys.Exchange(Index1, Index2: Long);
begin
  inherited Exchange(Index1, Index2);
  if FSorted then
    Sort;
end;

procedure TNCustomStringKeys.SaveToStream(Stream: TNStream);
var
  S: NString;
begin
  S := GetText;
  Stream.Write(Pointer(S)^, Length(S));
end;

procedure TNCustomStringKeys.LoadFromStream(Stream: TNStream);
var
  Size: Long;
  Big: NString;
begin
  Size := Stream.Size - Stream.Position;
  SetString(Big, nil, Size);
  Stream.Read(Pointer(Big)^, Size);
  SetText(Big);
end;


{*************************************************************************}
{                         TNStringList methods                            }
{*************************************************************************}


constructor TNStringList.Create;
begin
  inherited Create(SizeOf(TStringRecord));
end;

function TNStringList.GetKey(Index: Long): NString;
begin
  if (Index > -1) and (Index < FCount) then
    Result := PStrRecArray(FData)[Index].Key
  else
    Result := '';
end;

procedure TNStringList.PutKey(Index: Long; const Value: NString);
begin
  if (Index > -1) and (Index < FCount) and (not FSorted) then
  begin
    Notify(Index, lnDeleted);
    PStrRecArray(FData)[Index].Key := Value;
    Notify(Index, lnAdded);
  end;
end;

procedure TNStringList.InsertInternal(Index: Long; const S: NString);
var
  Rec: TStringRecord;
begin
  // this is needed - exceptions otherwise
  // old string at Index is copied (pointer copy) to Index+1 but
  // it also remains at Index. If we call
  // StrArray[Index].Key:=New then reference count of old string
  // is decreased and old string is destroyed so
  // old string moved to Index+1 is also destroyed and this
  // causes exceptions when you access old string at Index+1.
  // So we must first set old string at Index to not
  // point at old string at Index+1, it is done in InsertItem()
  // by ZeroMemory()
  InsertItem(Index, Rec);
  with PStrRecArray(FData)[Index] do
  begin
    Key := S;
  end;
end;

procedure TNStringList.FinalizeItem(Index: Long);
begin
  Finalize(PStrRecArray(FData)[Index], 1);
end;

function TNStringList.Compare(Index1, Index2: Long): Long;
begin
  if FCaseSensitive then
    Result := Kernel.CompStrCase(PStrRecArray(FData)[Index1].Key,
      PStrRecArray(FData)[Index2].Key)
  else
    Result := Kernel.CompStr(PStrRecArray(FData)[Index1].Key,
      PStrRecArray(FData)[Index2].Key);
end;

function TNStringList.Compare(Index: Long; const S: NString): Long;
begin
  if FCaseSensitive then
    Result := Kernel.CompStrCase(PStrRecArray(FData)[Index].Key, S)
  else
    Result := Kernel.CompStr(PStrRecArray(FData)[Index].Key, S);
end;

function TNStringList.GetMemoryUsage: Long;
var
  I: Long;
begin
  Result := inherited GetMemoryUsage;
  for I := 0 to FCount - 1 do
    Result := Result + Length(PStrRecArray(FData)[I].Key) * SizeOf(Char);
end;

function TNStringList.GetLine(Index: Long): NString;
begin
  Result := PStrRecArray(FData)[Index].Key + sLineBreak;
end;

procedure TNStringList.AddFromLine(const S: NString);
begin
  Add(S);
end;

procedure TNStringList.ReadData(Reader: TNAbstractReader);
var
  Number, I: Long;
begin
  Number := Reader.ReadInt;
  for I := 0 to Number - 1 do
    Add(Reader.ReadString);
end;

procedure TNStringList.WriteData(Writer: TNAbstractWriter);
var
  I: Long;
begin
  Writer.WriteInt(FCount);
  for I := 0 to FCount - 1 do
    Writer.WriteString(PStrRecArray(FData)[I].Key);
end;

function TNStringList.Add(const AKey: NString): Long;
begin
  if not FSorted then
    Result := FCount
  else
    if Find(AKey, Result) then
      if FDuplicates = dupIgnore then
        Exit;
  InsertInternal(Result, AKey);
  Notify(Result, lnAdded);
end;

procedure TNStringList.Insert(Index: Long; const AKey: NString);
begin
  if (Index > -1) and (Index <= FCount) then
  begin
    if not FSorted then
      InsertInternal(Index, AKey)
    else
      Add(AKey);
  end;
end;

procedure TNStringList.AddStrings(List: TNCustomStringKeys);
var
  I: Long;
begin
  if List <> nil then
    for I := 0 to List.FCount - 1 do
      Add(List.GetKey(I));
end;


{*************************************************************************}
{                      TNCustomStringPointerList methods                  }
{*************************************************************************}


constructor TNCustomStringPointerList.Create;
begin
  inherited Create(SizeOf(TStrPtrRecord));
end;

function TNCustomStringPointerList.GetKey(Index: Long): NString;
begin
  if (Index > -1) and (Index < FCount) then
    Result := PStrPtrArray(FData)[Index].Key
  else
    Result := '';
end;

procedure TNCustomStringPointerList.PutKey(Index: Long; const Value: NString);
begin
  if (Index > -1) and (Index < FCount) and (not FSorted) then
  begin
    Notify(Index, lnKeyDeleted);
    PStrPtrArray(FData)[Index].Key := Value;
    Notify(Index, lnKeyAdded);
  end;
end;

function TNCustomStringPointerList.GetPointer(Index: Long): Pointer;
begin
  if (Index > -1) and (Index < FCount) then
    Result := PStrPtrArray(FData)[Index].Value
  else
    Result := nil;
end;

procedure TNCustomStringPointerList.PutPointer(Index: Long; const Value:
  Pointer);
begin
  if (Index > -1) and (Index < FCount) then
  begin
    Notify(Index, lnValueDeleted);
    PStrPtrArray(FData)[Index].Value := Value;
    Notify(Index, lnValueAdded);
  end;
end;

procedure TNCustomStringPointerList.InsertInternal(Index: Long;
  const AKey: NString; Item: Pointer);
var
  Rec: TStrPtrRecord;
begin
  InsertItem(Index, Rec);
  with PStrPtrArray(FData)[Index] do
  begin
    Value := Item;
    Key := AKey;
  end;
end;

procedure TNCustomStringPointerList.FinalizeItem(Index: Long);
begin
  Finalize(PStrPtrArray(FData)[Index], 1);
end;

function TNCustomStringPointerList.Compare(Index1, Index2: Long): Long;
begin
  if FCaseSensitive then
    Result := Kernel.CompStrCase(PStrPtrArray(FData)[Index1].Key,
      PStrPtrArray(FData)[Index2].Key)
  else
    Result := Kernel.CompStr(PStrPtrArray(FData)[Index1].Key,
      PStrPtrArray(FData)[Index2].Key);
end;

function TNCustomStringPointerList.Compare(Index: Long; const S: NString): Long;
begin
  if FCaseSensitive then
    Result := Kernel.CompStrCase(PStrPtrArray(FData)[Index].Key, S)
  else
    Result := Kernel.CompStr(PStrPtrArray(FData)[Index].Key, S);
end;

function TNCustomStringPointerList.GetMemoryUsage: Long;
var
  I: Long;
begin
  Result := inherited GetMemoryUsage;
  for I := 0 to FCount - 1 do
    Result := Result + Length(PStrPtrArray(FData)[I].Key) * SizeOf(Char);
end;

function TNCustomStringPointerList.GetLine(Index: Long): NString;
begin
  Result := PStrPtrArray(FData)[Index].Key + sLineBreak;
end;

procedure TNCustomStringPointerList.AddFromLine(const S: NString);
begin
  Add(S, nil);
end;

procedure TNCustomStringPointerList.ReadData(Reader: TNAbstractReader);
var
  Number, I: Long;
begin
  Number := Reader.ReadInt;
  for I := 0 to Number - 1 do
    Add(Reader.ReadString, nil);
end;

function TNCustomStringPointerList.IndexOfValue(Value: Pointer): Long;
var
  I: Long;
begin
  Result := -1;
  for I := 0 to FCount - 1 do
    if PStrPtrArray(FData)[I].Value = Value then
    begin
      Result := I;
      Exit;
    end;
end;

function TNCustomStringPointerList.Add(const AKey: NString; Value: Pointer):
  Long;
begin
  if not FSorted then
    Result := FCount
  else
    if Find(AKey, Result) then
      if FDuplicates = dupIgnore then
        Exit;
  InsertInternal(Result, AKey, Value);
  Notify(Result, lnAdded);
end;

procedure TNCustomStringPointerList.Insert(Index: Long; const AKey: NString;
  Value: Pointer);
begin
  if (Index > -1) and (Index <= FCount) then
  begin
    if not FSorted then
      InsertInternal(Index, AKey, Value)
    else
      Add(AKey, Value);
  end;
end;

function TNCustomStringPointerList.FindValue(const AKey: NString): Pointer;
begin
  Result := GetPointer(IndexOf(AKey));
end;


{*************************************************************************}
{                        TNStringObjectList methods                       }
{*************************************************************************}


function TNStringObjectList.GetObject(Index: Long): TObject;
begin
  if (Index > -1) and (Index < FCount) then
    Result := PStrPtrArray(FData)[Index].Value
  else
    Result := nil;
end;

procedure TNStringObjectList.PutObject(Index: Long; Value: TObject);
begin
  if (Index > -1) and (Index < FCount) then
  begin
    Notify(Index, lnValueDeleted);
    PStrPtrArray(FData)[Index].Value := Value;
    Notify(Index, lnValueAdded);
  end;
end;


{*************************************************************************}
{                         TNSortedStrings methods                         }
{*************************************************************************}


constructor TNSortedStrings.Create;
begin
  inherited Create;
  FSorted := True;
  FDuplicates := dupIgnore;
  FCaseSensitive := False;
end;


{*************************************************************************}
{                       TNSortedStringPointers methods                    }
{*************************************************************************}


constructor TNSortedStringPointers.Create;
begin
  inherited Create;
  FSorted := True;
  FDuplicates := dupIgnore;
  FCaseSensitive := False;
end;


{*************************************************************************}
{                       TNSortedStringObjects methods                     }
{*************************************************************************}


constructor TNSortedStringObjects.Create;
begin
  inherited Create;
  FSorted := True;
  FDuplicates := dupIgnore;
  FCaseSensitive := False;
end;


{*************************************************************************}
{                         TNCustomOrderedList methods                     }
{*************************************************************************}


procedure TNCustomOrderedList.Clear;
begin
  FList.Clear;
end;

function TNCustomOrderedList.Count: Long;
begin
  Result := FList.Count;
end;

constructor TNCustomOrderedList.Create(AItemSize: Long);
begin
  inherited Create;
  FList := TNCustomValueList.Create(AItemSize);
end;

destructor TNCustomOrderedList.Destroy;
begin
  FList.Free;
  inherited Destroy;
end;

procedure TNCustomOrderedList.PeekItem(var Item);
begin
  FList.GetItem(FList.FCount - 1, Item);
end;

procedure TNCustomOrderedList.PopItem(var Item);
begin
  PeekItem(Item);
  FList.Delete(FList.FCount - 1);
end;


{*************************************************************************}
{                         TNCustomStack methods                           }
{*************************************************************************}


procedure TNCustomStack.PushItem(const Item);
begin
  FList.AddItem(Item);
end;


{*************************************************************************}
{                          TNCustomQueue methods                              }
{*************************************************************************}


procedure TNCustomQueue.PushItem(const Item);
begin
  FList.InsertItem(0, Item);
end;


{*************************************************************************}
{                         TNPointerStack methods                          }
{*************************************************************************}


constructor TNPointerStack.Create;
begin
  inherited Create(SizeOf(Pointer));
end;

procedure TNPointerStack.Push(Item: Pointer);
begin
  PushItem(Item);
end;

function TNPointerStack.Pop: Pointer;
begin
  PopItem(Result);
end;

function TNPointerStack.Peek: Pointer;
begin
  PeekItem(Result);
end;


{*************************************************************************}
{                       TNPointerQueue methods                            }
{*************************************************************************}


constructor TNPointerQueue.Create;
begin
  inherited Create(SizeOf(Pointer));
end;

procedure TNPointerQueue.Push(Item: Pointer);
begin
  PushItem(Item);
end;

function TNPointerQueue.Pop: Pointer;
begin
  PopItem(Result);
end;

function TNPointerQueue.Peek: Pointer;
begin
  PeekItem(Result);
end;


{*************************************************************************}
{                         TNLog methods                                   }
{*************************************************************************}


constructor TNLog.Create(const AFileName: NString);
begin
  inherited Create;
  Name := 'Log File';
  Description := 'Stores information about program flow.';
  FMemoryLog := TNStringList.Create;
  FBuffer := TNStringList.Create;
  FBufMax := MaxLogBuffer;
  FLineLength := MaxLogLine;
{$IFDEF LOGTOFILES}
  FWriteToFile := not System.IsLibrary;
{$ELSE}
  FWriteToFile := False;
{$ENDIF}
  FFileName := ExeDir + LogDir + PathDelim +
    ExtractFileName(ChangeFileExt(AFileName, '.' + sLogFileExt));
  if FWriteToFile then
  try
    ForceDirectories(ExtractFileDir(FFileName));
    FStream := TNFileStream.Create(FFileName, fmCreate or fmShareDenyNone);
  except
    // File can not be created
    FWriteToFile := False;
  end;
  FIdent := '';
  FCheckRepeats := True;
  FStartTime := Platform.GetTimeMicroseconds;
  AddStartUp;
end;

destructor TNLog.Destroy;
begin
  AddShutDown;
  Flush;
  FStream.Free;
  FBuffer.Free;
  FMemoryLog.Free;
  inherited Destroy;
end;

procedure TNLog.AddStartup;
begin
end;

procedure TNLog.AddShutdown;
begin
end;

procedure TNLog.IncIdent;
begin
  FIdent := FIdent + sLogIdent;
end;

procedure TNLog.DecIdent;
begin
  Delete(FIdent, Length(FIdent) - Length(sLogIdent) + 1, Length(sLogIdent));
end;

procedure TNLog.WriteToBuffer(const S: NString);
begin
  if Assigned(FOnAddMsg) then
    FOnAddMsg(S);
  FBuffer.Add(S);
  FMemoryLog.Add(S);
end;

procedure TNLog.Flush;
begin
  if FStream <> nil then
  begin
    FBuffer.SaveToStream(FStream);
  end;
  FBuffer.Clear;
end;

procedure TNLog.Add(const Msg: NString);
var
  S: NString;
begin
  if FCheckRepeats and (Msg = FLastMsg) and (FLastMsg[Length(FLastMsg)] <>
    sLogLBChar) then
    Inc(FRepeats)
  else
  begin
    if FRepeats > 0 then
    begin
      S := FLastMsg + Format(' (Added %d times)', [FRepeats]);
      S := SysUtils.WrapText(FIdent + S, sLineBreak + FIdent + sLogIdent,
        [' ', '.', ',', '-', #9], FLineLength);
      WriteToBuffer(S);
      FRepeats := 0;
    end;
    FLastMsg := Msg;
    WriteToBuffer(SysUtils.WrapText(FIdent + Msg, sLineBreak + FIdent +
      sLogIdent,
      [' ', '.', ',', '-', #9], FLineLength));
  end;
  if (FBuffer.Count = FBufMax) and FWriteToFile then
    Flush;
end;

procedure TNLog.Add(const MsgFmt: NString; Args: array of const);
begin
  Add(Format(MsgFmt, Args));
end;

procedure TNLog.AddTime(const Msg: NString);
begin
  Add('Time: ' + TimeToStr(SysUtils.Time) + ' ' +
    FloatToStrF(Platform.GetTimeMicroseconds - FStartTime, ffNumber, 10, 0) +
    ' ' + Msg);
end;

procedure TNLog.AddError(const Msg: NString);
begin
  Add(sError + ': ' + Msg);
end;

procedure TNLog.AddWarning(const Msg: NString);
begin
  Add(sWarning + ': ' + Msg);
end;

procedure TNLog.BeginSubList(const Msg: NString);
begin
  Add(Msg);
  IncIdent;
  AddLineBreak;
end;

procedure TNLog.EndSubList(const Msg: NString);
begin
  AddLineBreak;
  DecIdent;
  Add(Msg);
end;

procedure TNLog.AddLineBreak;
begin
  Add(StringOfChar(sLogLBChar, FLineLength - Length(FIdent)));
end;


{*************************************************************************}
{                         TNKernelLog methods                             }
{*************************************************************************}


constructor TNKernelLog.Create(const AFileName: NString);
begin
  inherited Create(AFileName);
  Name := 'Kernel Log File';
  Description := 'Stores information about flow of engine kernel.';
end;

destructor TNKernelLog.Destroy;
begin
  LogInstance := nil;
  inherited Destroy;
end;

procedure TNKernelLog.AddStartup;
var
  Cache: NString;
begin
  inherited AddStartUp;
  AddLineBreak;
  Add('Engine StartUp');
  Add('Execution started: ' + DateToStr(Date) + ' ' + TimeToStr(SysUtils.Time));
  Add('Engine version: ' + VersionToStr(EngineVersion));
  Add('Compiler: ' + CompilerName);
  Add('Compilation mode: ' + {$IFNDEF DEBUG}sRelease{$ELSE}sDebug{$ENDIF});
  Add('Platform: ' + Platform.Name + ' ' + VersionToStr(Platform.Version));
  Cache := '';
  if Platform.CPUInfo.Cache > 0 then
    Cache := Format(' %d kB cache,', [Platform.CPUInfo.Cache]);
  Add('CPU: %s, %d MHz,%s %s', [Platform.CPUInfo.Name,
    Platform.CPUInfo.Frequency,
      Cache, CPUFeaturesToStr(Platform.CPUInfo.Features)]);
  Add('Memory (Free/Total): Physical %d/%d kB, PageFile %d/%d kB',
    [Platform.MemoryInfo.PhysicalFree, Platform.MemoryInfo.PhysicalTotal,
    Platform.MemoryInfo.PageFileFree, Platform.MemoryInfo.PageFileTotal]);
  BeginSubList('Messages:')
end;

procedure TNKernelLog.AddShutdown;
  function GetLastedTime: NString;
  var
    H, M, S, MS: Long;
    Tot: Int64;
  begin
    Tot := (Platform.GetTimeMicroseconds - FStartTime) div 1000;
    H := Tot div (3600000);
    Tot := Tot - H * 3600000;
    M := Tot div (60000);
    Tot := Tot - M * 60000;
    S := Tot div (1000);
    Tot := Tot - S * 1000;
    MS := Tot;
    Result := Format('%.2d:%.2d:%.2d:%.3d', [H, M, S, MS]);
  end;

begin
  EndSubList('Engine ShutDown');
  Add('Class instances created: %d', [GetCreatedObjects]);
  Add('Class instances still living: %d', [GetLivingObjects]);
  Add('Class instances freed: %d', [GetFreedObjects]);
  Add('Execution lasted (H:M:S:MS): ' + GetLastedTime);
  Add('Execution stopped: ' + DateToStr(Date) + ' ' + TimeToStr(SysUtils.Time));
  AddLineBreak;
  inherited AddShutDown;
end;


{*************************************************************************}
{                          TNFileModule methods                           }
{*************************************************************************}


function TNFileModule.TestFormat(Stream: TNStream): Boolean;
begin
  Result := False;
end;


{*************************************************************************}
{                         TNFileModuleList methods                        }
{*************************************************************************}


constructor TNFileModuleList.Create;
begin
  inherited Create;
  FModules := TNPointerList.Create;
end;

destructor TNFileModuleList.Destroy;
begin
  FModules.Free;
  inherited Destroy;
end;

function TNFileModuleList.GetDefault: TNFileModule;
begin
  if (FDefault = nil) and (FModules.Count > 0) then
    FDefault := TNFileModule(FModules[0]);
  Result := FDefault;
end;

procedure TNFileModuleList.Add(Module: TNFileModule);
begin
  if Module <> nil then
    if FModules.IndexOf(Module) < 0 then
      if FModules.Add(Module) > -1 then
        Log.Add('File format module registered: %s (%s)', [Module.FormatName,
          Module.Extensions]);
end;

procedure TNFileModuleList.Remove(Module: TNFileModule);
var
  I: Long;
begin
  if Module <> nil then
  begin
    I := FModules.IndexOf(Module);
    if I > -1 then
    begin
      FModules.Delete(I);
      if (FDefault = Module) and (FModules.Count > 0) then
        FDefault := TNFileModule(FModules[0]);
    end;
  end;
end;

function TNFileModuleList.Find(Module: TNFileModule): Boolean;
begin
  Result := FModules.IndexOf(Module) > -1;
end;

function TNFileModuleList.Find(const Ext: NString): TNFileModule;
var
  I, J: Long;
  List: TNStringList;
  Found: Boolean;

  procedure StrTokenToStrings(const Exts: NString);
  var
    Token: NString;
    S: NString;
  begin
    List.Clear;
    S := Exts;
    while S <> '' do
    begin
      Token := StrToken(S, [',', ';']);
      List.Add(Token);
    end;
  end;

begin
  Result := nil;
  Found := False;
  List := TNStringList.Create;
  for I := 0 to FModules.Count - 1 do
    if PosNoCase(Ext, TNFileModule(FModules[I]).Extensions, 1) > 0 then
    begin
      StrTokenToStrings(TNFileModule(FModules[I]).Extensions);
      for J := 0 to List.Count - 1 do
        if SameStr(Ext, List[J]) then
          Found := True;

      if Found then
      begin
        Result := TNFileModule(FModules[I]);
        Break;
      end;
    end;
  List.Free;
end;


{*************************************************************************}
{                        TNModuleRegister methods                         }
{*************************************************************************}


{$IFNDEF MODLIBRARY}

constructor TNModuleRegister.Create;
begin
  inherited Create;
  Name := 'Module Register';
  Description := 'Loads, unloads and manages external module libraries.';
  FList := TNSortedStringPointers.Create;
  FLoadCalls := TNPointerList.Create;
  FUnLoadCalls := TNPointerList.Create;
  FDefaultDirectory := ExeDir + ModuleDir + PathDelim;
end;

destructor TNModuleRegister.Destroy;
begin
  UnloadAll;
  FList.Free;
  FLoadCalls.Free;
  FUnLoadCalls.Free;
  ModRegInstance := nil;
  inherited Destroy;
end;

procedure TNModuleRegister.UnloadAll;
begin
  while FList.Count > 0 do
    Unload(PModuleEntry(FList.Pointers[0]).FileName);
end;

procedure TNModuleRegister.RunLoadCallBacks(Entry: PModuleEntry);
var
  I: Long;
  Done: Boolean;
begin
  Done := False;
  for I := 0 to FLoadCalls.Count - 1 do
  begin
    if Assigned(FLoadCalls[I]) then
      TModLoadCallBack(FLoadCalls[I])(Entry, Done);
    if Done then
      Break;
  end;
end;

procedure TNModuleRegister.RunUnLoadCallBacks(Entry: PModuleEntry);
var
  I: Long;
  Done: Boolean;
begin
  Done := False;
  for I := 0 to FUnLoadCalls.Count - 1 do
  begin
    if Assigned(FUnLoadCalls[I]) then
      TModUnLoadCallBack(FUnLoadCalls[I])(Entry, Done);
    if Done then
      Break;
  end;
end;

procedure TNModuleRegister.Load(const FileName: NString);
var
  Entry: PModuleEntry;
  FName: NString;
  GetID: TGetIDProc;
  GetCount: TGetCountProc;
begin
  Entry := nil;
  if FileExists(FileName) or FileExists(FDefaultDirectory + FileName) then
  begin
    FName := ExtractFileName(FileName);
    if (PosNoCase(sModulePrefix, FName, 1) = 1) then
      if not IsLoaded(FName) then
      try
        New(Entry);
        Entry^.Name := FName;

        if FileExists(FDefaultDirectory + FileName) then
          FName := FDefaultDirectory + FileName
        else
          FName := FileName;

        Entry^.Handle := LoadLibrary(FName);
        if Entry^.Handle = 0 then
          raise Exception.Create('Error when loading! Library not loaded!');
        Entry^.FileName := Platform.GetModuleFileName(Entry^.Handle);
        @GetID := GetProcAddress(Entry^.Handle, sGetIDProc);
        if not Assigned(GetID) then
          raise Exception.Create('');
        GetID(Entry^.ID);

        @GetCount := GetProcAddress(Entry^.Handle, sGetCountProc);
        if not Assigned(GetCount) then
          raise Exception.Create('');
        GetCount(Entry^.Count);

        FList.Add(Entry^.Name, Entry);
      except
        on E: Exception do
        begin
          if FList.IndexOfValue(Entry) > -1 then
            FList.Delete(FList.IndexOfValue(Entry));
          Dispose(Entry);
          // do not show errors here
          Exit;
        end;
      end;
    Log.Add('Module library loaded: ' + Entry^.Name);
    RunLoadCallBacks(Entry);
  end;
end;

procedure TNModuleRegister.LoadFromDirectory(const Dir: NString);
var
  DName: NString;
  Rec: TSearchRec;
  Res: Long;
begin
  DName := Dir + PathDelim + sModulePrefix + '*';
  ZeroMemory(@Rec, SizeOf(Rec));
  Res := FindFirst(DName, faAnyFile, Rec);
  try
    while Res = 0 do
    begin
      if IsValidModuleLibrary(Dir + PathDelim + Rec.Name) then
        Load(Dir + PathDelim + Rec.Name);
      Res := FindNext(Rec);
    end;
  finally
    FindClose(Rec);
  end;
end;

procedure TNModuleRegister.Unload(const FileName: NString);
var
  Entry: PModuleEntry;
  Idx: Long;
begin
  if FList.Find(ExtractFileName(FileName), Idx) then
  begin
    Entry := PModuleEntry(FList.Pointers[Idx]);
    RunUnLoadCallBacks(Entry);
    Log.Add('Module library unloaded: ' + Entry^.Name);
    try
      FList.Delete(Idx);
      UnloadLibrary(Entry^.Handle);
      Dispose(Entry);
    except
      on E: Exception do
      begin
        ShowError(ErrorStr(Self, 'Load', 'Error when unloading module library '
          + ' file "' + FileName + '".' + sLineBreak + sExceptMsg +
          E.Message), E.ClassName);
        Exit;
      end;
    end;
  end;
end;

procedure TNModuleRegister.AddLoadCallBack(Call: TModLoadCallBack);
begin
  FLoadCalls.Add(@Call);
end;

procedure TNModuleRegister.AddUnLoadCallBack(Call: TModUnLoadCallBack);
begin
  FUnLoadCalls.Add(@Call);
end;

function TNModuleRegister.IsLoaded(const FileName: NString): Boolean;
begin
  Result := FList.IndexOf(ExtractFileName(FileName)) > -1;
end;

function TNModuleRegister.IsValidModuleLibrary(const FileName: NString):
  Boolean;
begin
  Result := Platform.IsExecutable(FileName);
end;
{$ENDIF}


{*************************************************************************}
{                         Internal stuff                                  }
{*************************************************************************}


procedure CheckUnLoadClass(Entry: PModuleEntry; var Done: Boolean);
var
  I: Long;
begin
  I := 0;
  while I < RegClasses.Count - 1 do
  begin
    if FindHInstance(RegClasses.Objects[I]) = Entry^.Handle then
    begin
      UnRegisterClass(TNPersistentClass(RegClasses.Objects[I]));
      Dec(I);
    end;
    Inc(I);
  end;
end;


initialization

{$IFNDEF MODLIBRARY}
  {This procedure checks if some of TNPersistent classes registered
   in RegClasses are from loaded library (which is now unloading).
   If it is from library then it is unregistered automatically.}
  ModuleRegister.AddUnLoadCallBack(CheckUnLoadClass);
{$ENDIF}

finalization

  FreeAndNil(RegClasses);

end.

