{*************************************************************************}
{*   Platform                                                            *}
{*   Copyright (C) 2003 Marek Mauder, Vampyre SoftWorks                  *}
{*   Engine\Kernel\Platform.pas                                          *}
{*************************************************************************}
{                                                                         }
{ This program is free software; you can redistribute it and/or modify it }
{ under the terms of the GNU General Public License as published by the   }
{ Free Software Foundation; either version 2 of the License, or (at your  }
{ option) any later version. This program is distributed in the hope that }
{ it will be useful, but WITHOUT ANY WARRANTY; without even the implied   }
{ warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See   }
{ the GNU General Public License for more details.                        }
{                                                                         }
{ You should have received a copy of the GNU General Public License       }
{ along with this program; if not, write to the Free Software Foundation, }
{ Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.         }
{                                                                         }
{*************************************************************************}
{  Created On: 12th August 2003  (12.08.2003)                             }
{*************************************************************************}

{This unit contains abstract class which provides access to platform and OS
 specific functions (windows, threads, ...) and information. Global
 variable Platform is
 created in unit init section according to actual platform on which
 unit was compiled.}

unit Platforms;

{ TODOS:
  - Add something like TWindowManager for creating windows
      and dialogs. Each platform then can have more managers.
      There should be WindowManager(VCL,API,Num,...) and
      WindowRenderer(ASCII,D3D,OGL).
  - Add win managers for Win32: ASCII, API, VCL, CLX, Qt.
  - Add win managers for Linux: CLX, Qt, GTK.
  - Find out which older AMD CPUs has ProcNameString (K6?).
  - Get Cache Info on more CPU types.
  - Add MsgBox for floats.
}

{$I Compiler.inc}

interface

uses
  SysUtils, Kernel, Objects;

type


{*************************************************************************}
{                           Some Types                                    }
{*************************************************************************}


  {Some most known CPU features.}
  TCPUFeature = (cfFPU, cfCMOV, cfMMX, cfMMX2, cf3DNow, cf3DNowEx, cfSSE,
    cfSSE2);

  TCPUFeatures = set of TCPUFeature;

  {(*}

  {Record describing installed CPU.}
  TCPUInfo = record
              Vendor: NString;          // Vendor - AMD, Intel, Trash, ...
              Name: NString;            // Processor Name - Athlon, P4, 8051
              Frequency: Long;          // In Mhz
              Family: Long;             // CPU Family
              Model: Long;              // CPU Model
              Stepping: Long;           // CPU Stepping
              Cache: Long;              // In kB
              Features: TCPUFeatures;   // set of CPU features
              ProcessorCount: Long;     // number of installed processors
             end;

  {Record describing current memory situation.}
  TMemoryInfo = record
                 CurrentLoad: Byte;     // current memory load in percent
                 PhysicalTotal: Long;   // total physical memory in kB
                 PhysicalFree: Long;    // total physical memory in kB
                 PageFileTotal: Long;   // total pagefile memory in kB
                 PageFileFree: Long;    // free pagefile memory in kB
                 VirtualTotal: Long;    // total virtual memory in kB
                 VirtualFree: Long;     // free virtual memory in kB
                end;

  {*)}


  {Value returned by TNPlatform.ErrorMsg that application can
   continue or should be terminated.}
  TErrorMsgResult = (emContinue, emTerminate);


{*************************************************************************}
{                         Class Types                                     }
{*************************************************************************}


  {Class which describes and provides access to platform on which engine is
   running. It contains info about CPU, memory, devices, ...  and provides
   wrapper to OS specific functions - window creation, threads, display,
   input, ...}
  TNPlatform = class(TNSingleton)
  protected
    FVersion: TVersionInfo;
    FCPUInfo: TCPUInfo;
    FMemoryInfo: TMemoryInfo;
    function GetMemoryInfo: TMemoryInfo;
    {This is called only once at startup. Gathers info which will not change
     at runtime (CPU Type for example).}
    procedure GatherStaticInfo; virtual; abstract;
    {This can be called whenever you want. Gets e.g. memory situation, ...
     Index determines which info type to update. See index values at property
     getters declarations.}
    procedure GatherDynamicInfo(Index: Long); virtual; abstract;
  public
    constructor Create; override;
    destructor Destroy; override;

    {* Windows, dialogs and other methods *}

    {Simple method which shows informational message. Used mainly
     when debugging.}
    procedure MsgBox(const Msg: NString); overload;
    {Simple method which shows informational message. Used mainly
     when debugging.}
    procedure MsgBox(const Number: Int64); overload;
    {Simple method which shows informational message. Used mainly
     when debugging.}
    procedure MsgBox(const Number: Float); overload;
    {Method which shows formated message. Uses standard SysUtils.Format func.}
    procedure MsgBox(const FormatStr: NString; const Args: array of const); overload;
      virtual; abstract;
    {Method shows error message with continue/terminate option.}
    function ErrorMsg(const Msg: NString): TErrorMsgResult; virtual; abstract;

    {* Files, libraries and other methods *}

    {Returns full file name of given module.}
    function GetModuleFileName(Handle: THandle): NString; virtual; abstract;
    {Loads library and returns its handle (0 if not succesful).
     Use Kernel.LoadLibrary for loading packages.}
    function LoadLibrary(const FileName: NString): THandle; virtual; abstract;
    {Unloads library and returns True if succesful).
     Use Kernel.UnloadLibrary for unloading packages.}
    function UnloadLibrary(Handle: THandle): Boolean; virtual; abstract;
    {Returns address of given symbol in given module.
     Returns nil if unsuccessful.}
    function GetProcAddress(Handle: THandle; const ProcName: NString): Pointer;
      virtual; abstract;
    {Returns True if FileName is valid executable file for given platform.}
    function IsExecutable(const FileName: NString): Boolean; virtual; abstract;

    {* Time, Date and other methods *}

    {Returns time in microsecond resolution. Return value may differ
     on different platforms so use it only when comparing two values.}
    function GetTimeMicroseconds: Int64; virtual; abstract;
    {Delay program execution for a specified number of milliseconds.}
    procedure Sleep(Milliseconds: ULong); virtual; abstract;

    {* Properties *}

    {Platform version: e.g. 5.1.2600 for WinXP or 2.4.19 for some Linux.}
    property Version: TVersionInfo read FVersion;
    {Info about installed CPU. See TCPUInfo for details.}
    property CPUInfo: TCPUInfo read FCPUInfo;
    {Info about current memory situation. See TMemoryInfo for details.}
    property MemoryInfo: TMemoryInfo read GetMemoryInfo;
  end;


{*************************************************************************}
{                           Singleton access                              }
{*************************************************************************}


{This function provides access to TNPlatform singleton instance.
 I rather use function than global variable because it can check
 if instance is valid.}
function Platform: TNPlatform;


{*************************************************************************}
{                           Misc functions                                }
{*************************************************************************}


{Converts CPU features set to readable string.}
function CPUFeaturesToStr(const Feats: TCPUFeatures): NString;


{*************************************************************************}
{                             Implementation                              }
{*************************************************************************}


implementation

uses
{$IFDEF MSWINDOWS}
  Win32Platform
{$ENDIF}
{$IFDEF LINUX}
  LinuxPlatform
{$ENDIF}
{$IFDEF MSDOS}
  DosPlatform
{$ENDIF}
{$IFDEF DOTNET}
  DotNetPlatform
{$ENDIF};

var

  {Global instance of Platform accessible trough Platform() function
   (it check if instance is created).}
  PlatformInstance: TNPlatform = nil;


{*************************************************************************}
{                           Singleton access                              }
{*************************************************************************}


function Platform: TNPlatform;
begin
  if PlatformInstance = nil then
  begin
{$IFDEF MSWINDOWS}
    PlatformInstance := TNPlatformWin32.Create
{$ENDIF}
{$IFDEF LINUX}
    PlatformInstance := TNPlatformLinux.Create
{$ENDIF};
{$IFDEF MSDOS}
    PlatformInstance := TNPlatformDos.Create
{$ENDIF};
  end;
  Result := PlatformInstance;
  if Result = nil then
    ShowError(ErrorStr(nil, 'Platform', 'No active platform created!'
      + ' Application will crash soon!'));
end;


{*************************************************************************}
{                         Misc functions                                  }
{*************************************************************************}


function CPUFeaturesToStr(const Feats: TCPUFeatures): NString;
begin
  Result := '';
  if cfFPU in Feats then
    Result := Result + 'FPU ';
  if cfCMOV in Feats then
    Result := Result + 'CMOV ';
  if cfMMX in Feats then
    Result := Result + 'MMX ';
  if cfMMX2 in Feats then
    Result := Result + 'MMX2 ';
  if cf3DNow in Feats then
    Result := Result + '3DNow ';
  if cf3DNowEx in Feats then
    Result := Result + '3DNowEx ';
  if cfSSE in Feats then
    Result := Result + 'SSE ';
  if cfSSE2 in Feats then
    Result := Result + 'SSE2 ';
  Result := Trim(Result);
end;


{*************************************************************************}
{                         TNPlatform methods                              }
{*************************************************************************}


constructor TNPlatform.Create;
begin
  inherited Create;
  FName := 'Platform';
  FDescription := 'Provides information about and access to functions ' +
    'of running platform';
  GatherStaticInfo;
end;

destructor TNPlatform.Destroy;
begin
  PlatformInstance := nil;
  inherited Destroy;
end;

function TNPlatform.GetMemoryInfo: TMemoryInfo;
begin
  GatherDynamicInfo(1);
  Result := FMemoryInfo;
end;

procedure TNPlatform.MsgBox(const Msg: NString);
begin
  MsgBox('%s', [Msg]);
end;

procedure TNPlatform.MsgBox(const Number: Int64);
begin
  MsgBox('%d', [Number]);
end;

procedure TNPlatform.MsgBox(const Number: Float);
begin
  MsgBox('%f', [Number]);
end;


{*************************************************************************}
{                         Internal Stuff                                  }
{*************************************************************************}


initialization

finalization

end.

