{*************************************************************************}
{*   LinuxPlatform                                                       *}
{*   Copyright (C) 2003 Marek Mauder, Vampyre SoftWorks                  *}
{*   Engine\Kernel\Platforms\Linux\LinuxPlatform.pas                     *}
{*************************************************************************}
{                                                                         }
{ This program is free software; you can redistribute it and/or modify it }
{ under the terms of the GNU General Public License as published by the   }
{ Free Software Foundation; either version 2 of the License, or (at your  }
{ option) any later version. This program is distributed in the hope that }
{ it will be useful, but WITHOUT ANY WARRANTY; without even the implied   }
{ warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See   }
{ the GNU General Public License for more details.                        }
{                                                                         }
{ You should have received a copy of the GNU General Public License       }
{ along with this program; if not, write to the Free Software Foundation, }
{ Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.         }
{                                                                         }
{*************************************************************************}
{  Created On: 12th August 2003  (12.08.2003)                             }
{*************************************************************************}

{This unit contains descendant of TNPlatform designed for Linux platform.}

unit LinuxPlatform;

{$I Compiler.inc}

{Settings for module libraries' names - do not start
 with "lib" and end with "Linux".}
{$IFDEF MODLIBRARY}
{$LIBPREFIX ''}
{$LIBSUFFIX 'Linux'}
{$ENDIF}

interface

uses
  SysUtils, Kernel, Objects, Platforms, Libc, LibcElf, Xlib, LinuxCrt,
  LinuxSysInfo;

type


{*************************************************************************}
{                          Class Types                                    }
{*************************************************************************}


  {Descendant of TNPlatform designed for Linux X86 systems.
   It was tested on 2.4.19 kernel version. Details on class members seek in the
   declaration of TNPlatform class.}
  TNPlatformLinux = class(TNPlatform)
  private
    FTimeVal: TTimeVal;
  protected
    procedure GatherStaticInfo; override;
    procedure GatherDynamicInfo(Index: Long); override;
  public
    procedure MsgBox(const FormatStr: NString; const Args: array of const);
      override;
    function ErrorMsg(const Msg: NString): TErrorMsgResult; override;
    function GetModuleFileName(Handle: THandle): NString; override;
    function LoadLibrary(const FileName: NString): THandle; override;
    function UnloadLibrary(Handle: THandle): Boolean; override;
    function GetProcAddress(Handle: THandle; const ProcName: NString): Pointer;
      override;
    function IsExecutable(const FileName: NString): Boolean; override;
    function GetTimeMicroseconds: Int64; override;
    procedure Sleep(Milliseconds: ULong); override;
  end;


{*************************************************************************}
{                             IMPLEMEMTATION                              }
{*************************************************************************}


implementation


{*************************************************************************}
{                          TNPlatformLinux methods                        }
{*************************************************************************}


procedure TNPlatformLinux.GatherStaticInfo;
var
  S: NString;
begin
{$IFNDEF MODLIBRARY}
  LinuxSysInfo.GetCPUInfo(FCPUInfo);
  LinuxSysInfo.GetOSInfo(FVersion, S);
{$ENDIF}
  Name := S;
end;

procedure TNPlatformLinux.GatherDynamicInfo(Index: Long);
begin
  case Index of
    1: LinuxSysInfo.GetMemoryInfo(FMemoryInfo);
  end;
end;

procedure TNPlatformLinux.MsgBox(const FormatStr: NString;
  const Args: array of const);
var
  Msg: NString;
begin
  Msg := SysUtils.Format(FormatStr, Args);
{$IFDEF CONSOLE}
  WriteLn(#32);
  TextColor(LightCyan);
  WriteLn('Application Message');
  TextColor(LightGray);
  WriteLn(Msg + sLineBreak + 'Press [Enter] to continue');
  repeat
  until ReadKey = #13;
{$ELSE}

{$ENDIF}
end;

function TNPlatformLinux.ErrorMsg(const Msg: NString): TErrorMsgResult;
var
  S: NString;
begin
  Result := emContinue;
{$IFDEF CONSOLE}
  WriteLn(#32);
  TextColor(LightRed);
  WriteLn('Application Error!');
  TextColor(LightGray);
  WriteLn(Msg + sLineBreak +
    'Press [Enter] to continue or [Esc] to stop execution.');
  S := ReadKey;
  if S[1] = #27 then
    Result := emTerminate;
{$ELSE}

{$ENDIF}
end;

function TNPlatformLinux.GetModuleFileName(Handle: THandle): NString;
var
  FileName: array[0..FILENAME_MAX] of Char;
begin
  SetString(Result, FileName,
    System.GetModuleFileName(Handle, FileName, SizeOf(FileName)));
end;

function TNPlatformLinux.LoadLibrary(const FileName: NString): THandle;
begin
  Result := THandle(dlopen(PChar(FileName), RTLD_LAZY));
end;

function TNPlatformLinux.UnloadLibrary(Handle: THandle): Boolean;
begin
  Result := LongBool(dlclose(Pointer(Handle)));
  InvalidateModuleCache;
end;

function TNPlatformLinux.GetProcAddress(Handle: THandle;
  const ProcName: NString): Pointer;
begin
  Result := SysUtils.GetProcAddress(Handle, PChar(ProcName));
end;

function TNPlatformLinux.IsExecutable(const FileName: NString): Boolean;
var
  Stream: TNFileStream;
  Header: Elf32_EHdr;
begin
  Result := False;
  if not FileExists(FileName) then
    Exit;
  try
    Stream := TNFileStream.Create(FileName, fmOpenRead or fmShareDenyNone);
    try
      Stream.Read(Header, SizeOf(Header));
      if (PULong(@Header.e_ident)^ = ELFMAG) and (Header.e_version in [0, 1, 2])
        then
        Result := True;
    finally
      Stream.Free;
    end;
  except
    Result := False;
  end;
end;

function TNPlatformLinux.GetTimeMicroseconds: Int64;
begin
  GetTimeOfDay(FTimeVal, nil);
  Result := Int64(FTimeVal.tv_sec) * 1000000 + FTimeVal.tv_usec;
end;

procedure TNPlatformLinux.Sleep(Milliseconds: ULong);
begin
  USleep(Milliseconds * 1000);
end;

end.

