{*************************************************************************}
{*   Win32MemMgrs                                                        *}
{*   Copyright (C) 2004 Marek Mauder, Vampyre SoftWorks                  *}
{*   Engine\Kernel\Platforms\Win32\Win32Platform.pas                     *}
{*************************************************************************}
{                                                                         }
{ This program is free software; you can redistribute it and/or modify it }
{ under the terms of the GNU General Public License as published by the   }
{ Free Software Foundation; either version 2 of the License, or (at your  }
{ option) any later version. This program is distributed in the hope that }
{ it will be useful, but WITHOUT ANY WARRANTY; without even the implied   }
{ warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See   }
{ the GNU General Public License for more details.                        }
{                                                                         }
{ You should have received a copy of the GNU General Public License       }
{ along with this program; if not, write to the Free Software Foundation, }
{ Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.         }
{                                                                         }
{*************************************************************************}
{  Created On: 31st January 2004  (31.01.2004)                            }
{*************************************************************************}

{This unit contains few memory managers used on Windows systems when
 USEMEMMGR symbol is defined. Always define USEMEMMGR when you want
 to write or use DLL libraries which allocate memory freed in other module
 and vice versa. This means all long strings, dynamic arrays, classes, ...
 Important: Both EXE and DLL must use the same memory manager
 (except Global and Heap - these can be used together).
 You can find benchmarks of managers somewhere in Doc directory.}

unit Win32MemMgrs;

interface

{$I Compiler.inc}

uses
  Windows;

var
  {Standard mem mgr set by Delphi (or FPC). It is very FAST but it
   CAN NOT BE used in DLL/EXE interaction (if you are using long strings,...)}
  DelphiMemoryManager: TMemoryManager;
  {Mem mgr using WinAPI HeapAlloc/Realloc/Free functions. It is SLOWER
   but CAN BE used in EXE/DLL interaction.}
  HeapMemoryManager: TMemoryManager;
  {Mem mgr using WinAPI GlobalAlloc/Realloc/Free functions. It is SLOWER
   but CAN BE used in EXE/DLL interaction. It is limited to only 65536
   allocated blocks per process.}
  GlobalMemoryManager: TMemoryManager;
  {Uses trick to make Delphi mem mgr work in DLL/EXE . It is very FAST
   and CAN BE used in EXE/DLL interaction. This is recommended manager.}
  ShareMemoryManager: TMemoryManager;
  {Uses standard C functions malloc/realloc/free imported from MSVCRT.DLL.
   It is SLOWER and CAN BE used in EXE/DLL interaction. It can be useful when
   using some DLL written in C which returns allocated memory.}
  {$IFNDEF BCB}
  CMemoryManager: TMemoryManager;
  {$ENDIF}

{Returns pointer to memory manager that is actually used.}
function ActiveMemoryManager: PMemoryManager;
{Returns pointer to manager used in MemMgrs.pas when DEBUG is on.
 DelphiMM and ShareMM CAN NOT be used or you get stuck in loop.}
function DebugMemoryManager: PMemoryManager;


{*************************************************************************}
{                             IMPLEMEMTATION                              }
{*************************************************************************}


implementation

function ActiveMemoryManager: PMemoryManager;
begin
  Result := @ShareMemoryManager;
end;

function DebugMemoryManager: PMemoryManager;
begin
  Result := @HeapMemoryManager;
end;

{* Heap WinAPI memory manager *}

const
  HEAP_NO_SERIALIZE = $00001;
  HEAP_GROWABLE = $00002;
  HEAP_GENERATE_EXCEPTIONS = $00004;
  HEAP_ZERO_MEMORY = $00008;
  HEAP_REALLOC_IN_PLACE_ONLY = $00010;
  HEAP_TAIL_CHECKING_ENABLED = $00020;
  HEAP_FREE_CHECKING_ENABLED = $00040;
  HEAP_DISABLE_COALESCE_ON_FREE = $00080;
  HEAP_CREATE_ALIGN_16 = $10000;
  HEAP_CREATE_ENABLE_TRACING = $20000;
  HEAP_MAXIMUM_TAG = $00FFF;
  HEAP_PSEUDO_TAG_FLAG = $08000;
  HEAP_TAG_SHIFT = 16;

var
  HeapHandle: THandle;
  HeapFlags: DWord = 0;

function HeapGetMem(Size: Integer): Pointer;
begin
  Result := Windows.HeapAlloc(HeapHandle, HeapFlags, Size);
end;

function HeapFreeMem(P: Pointer): Integer;
begin
  Result := Integer(not Windows.HeapFree(HeapHandle, HeapFlags, P));
end;

function HeapReallocMem(P: Pointer; Size: Integer): Pointer;
begin
  if P <> nil then
    Result := Windows.HeapRealloc(HeapHandle,
      HeapFlags and (HEAP_GENERATE_EXCEPTIONS), P, Size)
  else
    Result := HeapGetMem(Size);
end;

{* Global WinAPI memory manager *}

function GlobalGetMem(Size: Integer): Pointer;
begin
  Result := GlobalAllocPtr(GMEM_MOVEABLE, Size);
end;

function GlobalFreeMem(P: Pointer): Integer;
begin
  // in FPC this returns Pointer
  Result := Integer(GlobalFreePtr(P));
end;

function GlobalReallocMem(P: Pointer; Size: Integer): Pointer;
begin
  if P <> nil then
    Result := GlobalReallocPtr(P, Size, 0)
  else
    Result := GlobalGetMem(Size);
end;

{* Share memory manager *}

const
  SignatureBytes1 = $BABE01234567FEED;
  SignatureBytes2 = $F00D76543210B00B;
  PagesBound = 15;

type
  TMemMgrPack = record
    RefCount: Integer;
    Signature1: Int64;
    MemMgr: TMemoryManager;
    Signature2: Int64;
  end;
  PMemMgrPack = ^TMemMgrPack;

var
  OldMemMgr: TMemoryManager;
  Requested, Allocated: PMemMgrPack;
  SI: TSystemInfo;

procedure ValidatePack(P: PMemMgrPack);
var
  Pid: DWord;
begin
  P^.RefCount := 1;
  Pid := GetCurrentProcessId;
  P^.Signature1 := SignatureBytes1 xor Pid;
  P^.Signature2 := SignatureBytes2 xor Pid;
end;

function IsPackValid(P: PMemMgrPack): Boolean;
var
  Pid: DWord;
begin
  Pid := GetCurrentProcessId;
  Result := (P^.Signature1 = SignatureBytes1 xor Pid) and
    (P^.Signature2 = SignatureBytes2 xor Pid);
end;

procedure SetupShareMem;
var
  Pages: Integer;
begin
  ShareMemoryManager := DelphiMemoryManager;
  GetSystemInfo(SI);
  Pages := 0;
  repeat
    Requested := SI.lpMaximumApplicationAddress;
    Requested := Pointer((LongWord(Requested) div $10000) * $10000);
    Dec(LongWord(Requested), Pages * $10000);
    Requested := Pointer((LongWord(Requested) div SI.dwPageSize) *
      SI.dwPageSize);
    Allocated := VirtualAlloc(Requested, SI.dwPageSize, MEM_RESERVE or
      MEM_COMMIT, PAGE_READWRITE);
    if (Allocated <> nil) then
      if (Requested = Allocated) then
        Break
      else
    else
      if IsPackValid(Requested) then Break;
    Inc(Pages);
  until Pages > PagesBound;

  if Allocated <> nil then
  begin
    if Requested <> Allocated then
    begin
      MessageBox(0, 'Shared Memory Allocator setup failed: ' +
        'Address was relocated.', 'ShareMem Error', MB_ICONERROR);
      Halt;
    end;
    GetMemoryManager(Allocated^.MemMgr);
    ValidatePack(Allocated);
  end
  else
  begin
    if not IsPackValid(Requested) then
    begin
      MessageBox(0,
        'Shared Memory Allocator setup failed: ' +
        'Address already reserved.', 'ShareMem Error', MB_ICONERROR);
      Halt;
    end;
    GetMemoryManager(OldMemMgr);
    ShareMemoryManager := Requested^.MemMgr;
    Inc(Requested^.RefCount);
  end;
end;

procedure CleanUpShareMem;
begin
  Dec(Requested^.RefCount);
  if Allocated <> nil then
    if Requested^.RefCount > 0 then TerminateProcess(GetCurrentProcess, 0);
  if Requested^.RefCount = 0 then
  begin
    VirtualFree(Allocated, SI.dwPageSize, MEM_DECOMMIT);
    VirtualFree(Allocated, 0, MEM_RELEASE);
  end;
end;

{$IFNDEF BCB}

{* C memory manager *}

const
  LibName = 'msvcrt.dll';

function Malloc(Size: DWord): Pointer; cdecl; external LibName name 'malloc';
procedure Free(P: Pointer); cdecl; external LibName name 'free';
function ReAlloc(P: Pointer; Size: DWord): Pointer; cdecl; external LibName name
  'realloc';

function CGetMem(Size: Integer): Pointer;
begin
  Result := Malloc(Size);
end;

function CFreeMem(P: Pointer): Integer;
begin
  Free(P);
  Result := 0;
end;

function CReallocMem(P: Pointer; Size: Integer): Pointer;
begin
  Result := Realloc(P, Size);
end;

{$ENDIF}


{*************************************************************************}
{                         Internal stuff                                  }
{*************************************************************************}


initialization

  GetMemoryManager(DelphiMemoryManager);

  HeapHandle := Windows.GetProcessHeap;
  HeapMemoryManager.GetMem := @HeapGetMem;
  HeapMemoryManager.FreeMem := @HeapFreeMem;
  HeapMemoryManager.ReallocMem := @HeapReallocMem;

  GlobalMemoryManager.GetMem := @GlobalGetMem;
  GlobalMemoryManager.FreeMem := @GlobalFreeMem;
  GlobalMemoryManager.ReallocMem := @GlobalReallocMem;

{$IFNDEF BCB}
  CMemoryManager.GetMem := @CGetMem;
  CMemoryManager.FreeMem := @CFreeMem;
  CMemoryManager.ReallocMem := @CReallocMem;
{$ENDIF}

  if ActiveMemoryManager = @ShareMemoryManager then SetupShareMem;

finalization

  if ActiveMemoryManager = @ShareMemoryManager then CleanUpShareMem;

end.

