{*************************************************************************}
{*   Win32Platform                                                       *}
{*   Copyright (C) 2003 Marek Mauder, Vampyre SoftWorks                  *}
{*   Engine\Kernel\Platforms\Win32\Win32Platform.pas                     *}
{*************************************************************************}
{                                                                         }
{ This program is free software; you can redistribute it and/or modify it }
{ under the terms of the GNU General Public License as published by the   }
{ Free Software Foundation; either version 2 of the License, or (at your  }
{ option) any later version. This program is distributed in the hope that }
{ it will be useful, but WITHOUT ANY WARRANTY; without even the implied   }
{ warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See   }
{ the GNU General Public License for more details.                        }
{                                                                         }
{ You should have received a copy of the GNU General Public License       }
{ along with this program; if not, write to the Free Software Foundation, }
{ Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.         }
{                                                                         }
{*************************************************************************}
{  Created On: 12th August 2003  (12.08.2003)                             }
{*************************************************************************}

{This unit contains descendant of TNPlatform designed for Win32 platform.}

unit Win32Platform;

{$I Compiler.inc}

{Settings for module libraries' names - do not start
 with "lib" and end with "Win32".}
{$IFDEF MODLIBRARY}
  {$LIBPREFIX ''}
  {$LIBSUFFIX 'Win32'}
{$ENDIF}

interface

uses
  SysUtils, Kernel, Objects, Platforms, Windows, Win32Imports, Win32Crt,
  Win32SysInfo;


{*************************************************************************}
{                          Class Types                                    }
{*************************************************************************}


type

  {Descendant of TNPlatform designed for Win32 systems. It should
  support all windows versions from Win95. However, Win2000 or WinXP
  are strongly recommended. Details on class members seek in the declaration
  of TNPlatform class.}
  TNPlatformWin32 = class(TNPlatform)
  private
    FPerfFrequency: Int64;
    FInvPerfFrequency: Double;
    FTimeInt64: Int64;
  protected
    procedure GatherStaticInfo; override;
    procedure GatherDynamicInfo(Index: Long); override;
  public
    procedure MsgBox(const FormatStr: NString; const Args: array of const); override;
    function ErrorMsg(const Msg: NString): TErrorMsgResult; override;
    function GetModuleFileName(Handle: THandle): NString; override;
    function LoadLibrary(const FileName: NString): THandle; override;
    function UnloadLibrary(Handle: THandle): Boolean; override;
    function GetProcAddress(Handle: THandle; const ProcName: NString): Pointer;
      override;
    function IsExecutable(const FileName: NString): Boolean; override;
    function GetTimeMicroseconds: Int64; override;
    procedure Sleep(Milliseconds:ULong);override;
  end;


{*************************************************************************}
{                             Implementation                              }
{*************************************************************************}


implementation


{*************************************************************************}
{                         TNPlatformWin32 methods                         }
{*************************************************************************}


procedure TNPlatformWin32.GatherStaticInfo;
var
  S: NString;
begin
{$IFNDEF MODLIBRARY}
  Win32SysInfo.GetCPUInfo(FCPUInfo);
  Win32SysInfo.GetOSInfo(FVersion, S);
{$ENDIF}
  Name := S;
  Windows.QueryPerformanceFrequency(FPerfFrequency);
  FInvPerfFrequency := 1.0 / FPerfFrequency;
end;

procedure TNPlatformWin32.GatherDynamicInfo(Index: Long);
begin
  case Index of
    1: Win32SysInfo.GetMemoryInfo(FMemoryInfo);
  end;
end;

procedure TNPlatformWin32.MsgBox(const FormatStr: NString; const Args: array of const);
var
  Msg:NString;
begin
  Msg:=SysUtils.Format(FormatStr, Args);
  if not System.IsConsole then
    MessageBox(GetForegroundWindow, PChar(Msg), 'Message',
      MB_ICONINFORMATION or MB_OK)
  else
  begin
    WriteLn;
    TextColor(LightCyan);
    WriteLn('Application Message');
    TextColor(LightGray);
    Write(Msg + sLineBreak + 'Press [Enter] to continue');
    repeat
    until ReadKey = #13;
  end;
end;

function TNPlatformWin32.ErrorMsg(const Msg: NString): TErrorMsgResult;
var
  MRes, Flags: ULong;
  S: NString;
begin
  Result := emContinue;
  if IsConsole then
  begin
    TextColor(LightRed);
    WriteLn;
    WriteLn(sLineBreak + 'Application Error!');
    TextColor(LightGray);
    Write(Msg + sLineBreak +
      'Press [Enter] to continue or [Esc] to stop execution.');
    S := ReadKey;
    if S[1] = #27 then
      Result := emTerminate;
  end
  else
  begin
    Flags := MB_ICONERROR or MB_YESNO;
    S := sLineBreak + 'Continue execution?' + sLineBreak;
    MRes := Windows.MessageBox(GetForegroundWindow, PChar(Msg + S),
      'Application Error', Flags);
    if MRes = IDNO then
      Result := emTerminate;
  end;
end;

function TNPlatformWin32.GetModuleFileName(Handle: THandle): NString;
var
  FileName: array[0..MAX_PATH] of Char;
begin
  SetString(Result, FileName,
    Windows.GetModuleFileName(Handle, FileName, SizeOf(FileName)));
end;

function TNPlatformWin32.LoadLibrary(const FileName: NString): THandle;
begin
  Result := Windows.LoadLibrary(PChar(FileName));
end;

function TNPlatformWin32.UnloadLibrary(Handle: THandle): Boolean;
begin
  Result := Windows.FreeLibrary(Handle);
end;

function TNPlatformWin32.GetProcAddress(Handle: THandle;
  const ProcName: NString): Pointer;
begin
  Result := Windows.GetProcAddress(Handle, PChar(ProcName));
end;

function TNPlatformWin32.IsExecutable(const FileName: NString): Boolean;
var
  Stream: TNFileStream;
  Image: TImageDosHeader;
begin
  Result := False;
  if FileExists(FileName) then 
  try
    Stream := TNFileStream.Create(FileName, fmOpenRead or fmShareDenyNone);
    try
      Stream.Read(Image, SizeOf(Image));
      if Image.e_magic = IMAGE_DOS_SIGNATURE then Result := True;
    finally
      Stream.Free;
    end;
  except
    Result := False;
  end;
end;

function TNPlatformWin32.GetTimeMicroseconds: Int64;
begin
  QueryPerformanceCounter(FTimeInt64);
  Result := Round(1000000 * FInvPerfFrequency * FTimeInt64);
end;

procedure TNPlatformWin32.Sleep(Milliseconds: ULong);
begin
  Windows.Sleep(Milliseconds);
end;

end.

