{*************************************************************************}
{*   X86CpuInfo                                                          *}
{*   Copyright (C) 2003 Marek Mauder, Vampyre SoftWorks                  *}
{*   Engine\Kernel\Platforms\X86CpuInfo.pas                              *}
{*************************************************************************}
{                                                                         }
{ This program is free software; you can redistribute it and/or modify it }
{ under the terms of the GNU General Public License as published by the   }
{ Free Software Foundation; either version 2 of the License, or (at your  }
{ option) any later version. This program is distributed in the hope that }
{ it will be useful, but WITHOUT ANY WARRANTY; without even the implied   }
{ warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See   }
{ the GNU General Public License for more details.                        }
{                                                                         }
{ You should have received a copy of the GNU General Public License       }
{ along with this program; if not, write to the Free Software Foundation, }
{ Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.         }
{                                                                         }
{*************************************************************************}
{  Created On: 18th September 2003  (18.09.2003)                          }
{*************************************************************************}

{This unit contains functions to retrieve information about
installed CPU. It contains functions which can be on X86
based systems.}

unit X86CpuInfo;

{$I Compiler.inc}

interface

uses
  SysUtils, Kernel, Platforms;

type

{Function which returns calue of Programmable Interrupt Timer.}
  TGetPITValue = function: Int64;

{Returns information on installed CPU.  Remember that
 it does NOT RETRIEVE ALL info - ProcessorCount
 is not filled, you must find it by platform specific things.
 Fill PITRate and PITFunction (according to current platform) if you want to
 get CPU frequency returned in Info.}
procedure GetX86CPUInfo(var Info: TCPUInfo; PITRate: Int64 = 0; PITFunction:
  TGetPITValue = nil);


{*************************************************************************}
{                             IMPLEMEMTATION                              }
{*************************************************************************}


implementation

const
  IDAMD = 'AuthenticAMD';
  IDIntel = 'GenuineIntel';
  IDCyrix = 'CyrixInstead';
  IDCentaur = 'CentaurHauls';

  sAMD = 'AMD';
  sIntel = 'Intel';
  sCyrix = 'Cyrix';
  sCentaur = 'Centaur/IDT';

  VAMD = 1;
  VIntel = 2;
  VCyrix = 3;
  VCentaur = 4;

  Level1Cache = $80000005;
  Level2Cache = $80000006;

type
  TCPUIDRet = packed record
    EAX: ULong;
    EBX: ULong;
    ECX: ULong;
    EDX: ULong;
  end;

procedure GetX86CPUInfo(var Info: TCPUInfo; PITRate: Int64; PITFunction:
  TGetPITValue);
type
  TStrBuf = array[0..11] of Char;

  function GetTSCValue: Int64;
  asm
  DW      $310F
  end;

  function GetCPUFreq: Int64;
  const
    MaxTries = 20;
    Tolerance = 0.1;
  var
    Freq: array[0..2] of Double;
    Time0, Time1, Stamp0, Stamp1, Cycles, Ticks: Int64;
    Total: Double;
    I, Tries: Long;
  begin
    Result := 0;
    if (not Assigned(PITFunction)) or (PITRate = 0) then
      Exit;
    for I := 0 to 2 do
      Freq[I] := 0.0;
    Tries := 0;
    Stamp0 := 0;
    Stamp1 := 0;
    repeat
      Tries := Tries + 1;
      Freq[2] := Freq[1];
      Freq[1] := Freq[0];
      Time0 := PITFunction;
      Time1 := Time0;
      while ((Time1 - Time0) < 50) do
      begin
        Time1 := PITFunction;
        Stamp0 := GetTSCvalue;
      end;
      Time0 := Time1;
      while ((Time1 - Time0) < 10000) do
      begin
        Time1 := PITFunction;
        Stamp1 := GetTSCvalue;
      end;
      Cycles := Stamp1 - Stamp0;
      Ticks := ((Time1 - Time0) * 1000000) div PITRate;
      Freq[0] := Cycles / Ticks;
      Total := Freq[0] + Freq[1] + Freq[2];
      if (Tries > MaxTries) then
        Break;
    until ((Abs((3 * Freq[0]) - Total) <= Tolerance) and
      (Abs((3 * Freq[1]) - Total) <= Tolerance) and
      (Abs((3 * Freq[2]) - Total) <= Tolerance));
    Result := Round(Total / 3);
  end;

  function HasCPUID: LongBool;
  asm
    PUSHFD
    POP    EAX
    MOV    ECX, EAX
    XOR    EAX, $00200000
    PUSH   EAX
    POPFD
    PUSHFD
    POP    EAX
    XOR    EAX, ECX
  end;

  function CPUID(Flag: ULong): TCPUIDRet;
  asm
    PUSH   EBX
    PUSH   EDI
    MOV    EDI, EDX
    CPUID
    STOSD
    MOV    EAX, EBX
    STOSD
    MOV    EAX, ECX
    STOSD
    MOV    EAX, EDX
    STOSD
    POP    EDI
    POP    EBX
  end;

  function GetCPUVendor: NString;

    function GetBuf: TStrBuf;
    asm
      PUSH   EBX
      PUSH   EDI
      MOV    EDI, EAX
      MOV    EAX, 0
      DW     $A20F
      MOV    EAX, EBX
      XCHG   EBX, ECX
      MOV    ECX, 4
    @1:
      STOSB
      SHR    EAX, 8
      LOOP   @1
      MOV    EAX, EDX
      MOV    ECX, 4
    @2:
      STOSB
      SHR    EAX, 8
      LOOP   @2
      MOV    EAX, EBX
      MOV    ECX, 4
    @3:
      STOSB
      SHR    EAX, 8
      LOOP   @3
      POP    EDI
      POP    EBX
    end;

  begin
    Result := GetBuf;
  end;

  function MakeVendor(const ID: NString): NString;
  begin
    if ID = IDAMD then
      Result := sAMD
    else
      if ID = IDIntel then
        Result := sIntel
      else
        if ID = IDCyrix then
          Result := sCyrix
        else
          if ID = IDCentaur then
            Result := sCentaur
          else
            Result := ID;
  end;

  procedure GetStringAMD(Flag: ULong; Str: PChar);
  var
    RetVal, PS: ULong;
  asm
    PUSH   EBX
    MOV    PS, EDX
    CPUID
    MOV    RetVaL, EAX
    MOV    EAX, PS
    MOV    DWORD PTR[EAX+4], EBX
    MOV    DWORD PTR[EAX+8], ECX
    MOV    EBX, RETVAL
    MOV    DWORD PTR[EAX+12], EDX
    MOV    DWORD PTR[EAX], EBX
    POP    EBX
  end;

  function GetVendorNum(const Vendor: NString): Long;
  begin
    if Vendor = sAMD then
      Result := VAMD
    else
      if Vendor = sIntel then
        Result := VIntel
      else
        if Vendor = sCyrix then
          Result := VCyrix
        else
          if Vendor = sCentaur then
            Result := VCentaur
          else
            Result := 0;
  end;

  function GetCPUName: NString;
  begin
    case GetVendorNum(Info.Vendor) of
      VAMD:
        case Info.Family of
          5:
            case Info.Model of
              0: Result := 'K5 PR-75/90/100)';
              1: Result := 'K5 PR-120/133';
              2: Result := 'K5 PR-166';
              3: Result := 'K5 PR-200';
              6: Result := 'K6 0.3 micron';
              7: Result := 'K6 0.25 micron';
              8: Result := 'K6-2';
              9: Result := 'K6-III';
              $D: Result := 'K6-2+/K6-III+';
            else
              Result := 'K5/K6';
            end;
          6: Result := 'K7';
        else
          Result := 'AMD CPU';
        end;
      VIntel:
        case Info.Family of
          5:
            case Info.Model of
              0: Result := 'Pentium A-Step';
              1: Result := 'Pentium';
              2: Result := 'Pentium';
              3: Result := 'Pentium P24T OverDrive';
              4: Result := 'Pentium MMX';
              5: Result := 'Pentium P54C';
              7: Result := 'Pentium P55C 0.25 micron';
              8: Result := 'Pentium MMX mobile';
            else
              Result := 'Pentium';
            end;
          6:
            case Info.Model of
              0, 1: Result := 'Pentium Pro';
              3: Result := 'Pentium II';
              5: Result := 'Pentium II/Celeron';
              6: Result := 'Celeron A';
              7: Result := 'Pentium III/Xeon';
              8: Result := 'Pentium III E/Xeon';
              10: Result := 'Xeon A';
            else
              Result := 'Pentium Pro';
            end;
          $15: Result := 'Pentium 4';
        else
          Result := 'Intel CPU';
        end;
      VCyrix:
        case Info.Family of
          5:
            case Info.Model of
              2: Result := '6x86/M1';
              4: Result := 'Media GX/GXm';
            else
              Result := '6x86/M1';
            end;
          6:
            case Info.Model of
              0: Result := '6x86MX/M2';
              5: Result := 'III';
            else
              Result := '6x86MX/M2';
            end;
        else
          Result := 'Cyrix CPU';
        end;
      VCentaur:
        case Info.Family of
          5:
            case Info.Model of
              4: Result := 'WinChip/C6';
              8: Result := 'WinChip 2/W2';
              9: Result := 'WinChip 3/W3';
            else
              Result := 'WinChip';
            end;
          6:
            case Info.Model of
              0: Result := 'C7';
            else
              Result := 'C7';
            end;
        else
          Result := 'Centaur/IDT CPU';
        end;
    else
      Result := 'Unknown CPU';
    end;
  end;

var
  L2, L1Data, L1Code: ULong;
  Ret: TCPUIDRet;
  Str: PChar;
begin
  Finalize(Info);
  ZeroMemory(@Info, SizeOf(Info));
  if HasCPUID then
  begin
    Ret := CPUID(1);
    Info.Family := (Ret.EAX shr 8) and $0F;
    Info.Model := (Ret.EAX shr 4) and $0F;
    Info.Stepping := Ret.EAX and $07;
    Info.Frequency := GetCPUFreq;

    Info.Vendor := MakeVendor(GetCPUVendor);

    // Get AMD name string (works on P4 too, but must be trimmed)
    if ((Info.Vendor = sAMD) and (Info.Family > 5)) or
      ((Info.Vendor = sIntel) and (Info.Family > $15)) then
    begin
      Str := StrAlloc(48);
      GetStringAMD($80000002, Str);
      GetStringAMD($80000003, PChar(Long(Str) + 16));
      GetStringAMD($80000004, PChar(Long(Str) + 32));
      Info.Name := Trim(Str);
      StrDispose(Str);
    end
    else
      Info.Name := GetCPUName;

    // standard features
    if LongBool(Ret.EDX and 1) then
      Info.Features := Info.Features + [cfFPU];
    if LongBool(Ret.EDX and (1 shl 15)) then
      Info.Features := Info.Features + [cfCMOV];
    if LongBool(Ret.EDX and (1 shl 23)) then
      Info.Features := Info.Features + [cfMMX];
    if LongBool(Ret.EDX and (1 shl 25)) then
      Info.Features := Info.Features + [cfSSE];
    if LongBool(Ret.EDX and (1 shl 26)) then
      Info.Features := Info.Features + [cfSSE2];

    // extended features
    Ret := CPUID($80000000);
    if Ret.EAX > $80000000 then
    begin
      Ret := CPUID($80000001);
      if LongBool(Ret.EDX and (1 shl 22)) then
        Info.Features := Info.Features + [cfMMX2];
      if LongBool(Ret.EDX and (1 shl 31)) then
        Info.Features := Info.Features + [cf3DNow];
      if LongBool(Ret.EDX and (1 shl 30)) then
        Info.Features := Info.Features + [cf3DNowEx];
      // cache info (this works on Athlons only)
      Ret := CPUID(Level2Cache);
      L2 := TLongRec(Ret.ECX).Words[1]; // L2 unified
      Ret := CPUID(Level1Cache);
      L1Data := TLongRec(Ret.ECX).Bytes[3]; // L1 data
      L1Code := TLongRec(Ret.EDX).Bytes[3]; // L1 code
      Info.Cache := L2 + L1Data + L1Code;
    end;
  end;
end;

end.

