{*************************************************************************}
{*   Strings                                                             *}
{*   Copyright (C) 2004 Marek Mauder, Vampyre SoftWorks                  *}
{*   Engine\Kernel\Strings.pas                                           *}
{*************************************************************************}
{                                                                         }
{ This program is free software; you can redistribute it and/or modify it }
{ under the terms of the GNU General Public License as published by the   }
{ Free Software Foundation; either version 2 of the License, or (at your  }
{ option) any later version. This program is distributed in the hope that }
{ it will be useful, but WITHOUT ANY WARRANTY; without even the implied   }
{ warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See   }
{ the GNU General Public License for more details.                        }
{                                                                         }
{ You should have received a copy of the GNU General Public License       }
{ along with this program; if not, write to the Free Software Foundation, }
{ Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.         }
{                                                                         }
{*************************************************************************}
{  Created On: 20th January 2004  (20.01.2004)                            }
{*************************************************************************}

{There are some additional functions for manipulation with strings.}

unit Strings;

{TODOD -oGalfar: Rewrite array of Char parameters to TCharSet.}

interface

{$I Compiler.inc}

uses
  Kernel, Objects;


{*************************************************************************}
{                         Types and constants                             }
{*************************************************************************}


type

  TCharSet = set of Char;

const

  {Some standard CharSets.}
  csAlphaUpper = ['A'..'Z'];
  csAlphaLower = ['a'..'z'];
  csAlpha = csAlphaUpper + csAlphaLower;
  csDecimalDigits = ['0'..'9'];
  csAlhaNumeric = csAlpha + csDecimalDigits;
  csBinaryDigits = ['0'..'1'];
  csOctalDigits = ['0'..'7'];
  csHexDigitsUpper = csDecimalDigits + ['A'..'F'];
  csHexDigitsLower = csDecimalDigits + ['a'..'f'];
  csHexDigits = csHexDigitsUpper + csHexDigitsLower;
  csQuotes = ['''', '"'];
  csSlash = ['\', '/'];
  csSign = ['+', '-'];
  csParentheses = ['(', ')'];
  csBrackets = ['[', ']'];
  csCurlyBrackets = ['{', '}'];
  csAllBrackets = csParentheses + csBrackets + csCurlyBrackets;
  csSymbols = csSign + csAllBrackets + csQuotes + csSlash + ['*', ':', '<', '>',
    '=', '^', '&', '|', '!', '@', '#', '$', '%', '_', '.', ',', ';', '`', '?',
    '~'];


{*************************************************************************}
{                           AnsiString functions                          }
{*************************************************************************}


{* Array/TNStringList/Token/Wrap/Format functions *}

{Returns true if S is in Arr, case-insensitive}
function StrInArray(const S: NString; const Arr: array of NString): Boolean;
{Returns true if S is in Arr, case-sensitive}
function StrInArrayCase(const S: NString; const Arr: array of NString): Boolean;
{Fills instance of TNStringList with given NStrings.}
procedure StrArrayToList(const Arr: array of NString; List: TNStringList);
{Returns string created from items of Arr array separated by Sep character.}
function ComposeString(const Arr: array of NString; Sep: Char = ';'): NString;
overload;
{Returns string created from items (Keys only) of List separated
 by Sep character.}
function ComposeString(List: TNCustomStringKeys; Sep: Char = ';'): NString;
overload;
{Fills instance of TNStringList with tokens from string S separated by
 one of Seps characters.}
procedure StrTokensToList(const S: NString; const Seps: array of Char;
  List: TNStringList);
{Converts S to string where all characters after one of Delimiters (as well as
 the first character of S) are in upper case and others are in lower case.
 Exmaple: StrProperCase('converts s TO sTRING') = 'Converts S To String'}
function StrProperCase(const S: AnsiString; const Delimiters: TCharSet): AnsiString;

{* Conversion functions *}

{Converts string representation of binary number to ULong.}
function BinToULong(const S: AnsiString): ULong;
{Converts string representation of octal number to ULong.}
function OctToULong(const S: AnsiString): ULong;
{Converts string representation of hexadecimal number to ULong.}
function HexToULong(const S: AnsiString): ULong;
{Converts string representation of decimal number to ULong.}
function DecToULong(const S: AnsiString): ULong;

{If Digits parameter in these functions is 0 then Result
 will have all valid digits with no zeroes at left.
 If Digits is nonzero then Result will have exactly Digits
 digits so there can be zeroes at left or some digits
 may be missing.}
{Converts ULong to string of binary digits.}
function ULongToBin(Num: ULong; Digits: Byte = 0): AnsiString;
{Converts ULong to string of octal digits.}
function ULongToOct(Num: ULong; Digits: Byte = 0): AnsiString;
{Converts ULong to string of hexadecimal digits.}
function ULongToHex(Num: ULong; Digits: Byte = 0): AnsiString;
{Converts ULong to string of decimal digits.}
function ULongToDec(Num: ULong; Digits: Byte = 0): AnsiString;

{Converts string representation of Roman numeral to ULong.}
function RomanToULong(const S: NString): ULong;
{Converts number 1..5000 to Roman numeral.}
function ULongToRoman(Num: ULong): NString;

{* Info/Count functions *}

{Returns number of occurences (case insensitive) of C character in string S.}
function CountCharsNoCase(const S: AnsiString; C: Char): Long; overload;
{Returns number of occurences (case sensitive) of C character in string S.}
function CountCharsCase(const S: AnsiString; C: Char): Long; overload;
{Returns number of occurences (case insensitive) of C characters in string S.}
function CountCharsNoCase(const S: AnsiString; const C: array of Char): Long;
overload;
{Returns number of occurences (case sensitive) of C characters in string S.}
function CountCharsCase(const S: AnsiString; C: array of Char): Long; overload;

{Returns actual reference count of given string.}
function StrRefCount(const S: AnsiString): Long;

{* CharSet/String type functions}

{Returns set of characters which appear in string S.}
function GetCharSet(const S: AnsiString): TCharSet;
{Converts set of characters to string.}
function CharSetToStr(const CharSet: TCharSet): AnsiString;


{*************************************************************************}
{                             IMPLEMEMTATION                              }
{*************************************************************************}


implementation


{*************************************************************************}
{                           AnsiString functions                          }
{*************************************************************************}


{* String Array/TNStringList functions *}

function StrInArray(const S: NString; const Arr: array of NString): Boolean;
var
  I: Long;
begin
  Result := True;
  for I := Low(Arr) to High(Arr) do
    if CompStr(S, Arr[I]) = 0 then
      Exit;
  Result := False;
end;

function StrInArrayCase(const S: NString; const Arr: array of NString): Boolean;
var
  I: Long;
begin
  Result := True;
  for I := Low(Arr) to High(Arr) do
    if CompStrCase(S, Arr[I]) = 0 then
      Exit;
  Result := False;
end;

procedure StrArrayToList(const Arr: array of NString; List: TNStringList);
var
  I: Long;
begin
  List.Clear;
  for I := Low(Arr) to High(Arr) do
    List.Add(Arr[I]);
end;

function ComposeString(const Arr: array of NString; Sep: Char): NString;
var
  I: Long;
begin
  Result := '';
  for I := Low(Arr) to High(Arr) do
    Result := Result + Arr[I] + Sep;
  SetLength(Result, Length(Result) - 1);
end;

function ComposeString(List: TNCustomStringKeys; Sep: Char): NString;
var
  I: Long;
begin
  Result := '';
  for I := 0 to List.Count - 1 do
    Result := Result + List.Keys[I] + Sep;
  SetLength(Result, Length(Result) - 1);
end;

procedure StrTokensToList(const S: NString; const Seps: array of Char; List:
  TNStringList);
var
  Token: NString;
  ModS: NString;
begin
  List.Clear;
  ModS := S;
  while ModS <> '' do
  begin
    Token := Kernel.StrToken(ModS, Seps);
    List.Add(Token);
  end;
end;

function StrProperCase(const S: AnsiString; const Delimiters: TCharSet): AnsiString;
var
  I: Long;
  L: ULong;
  P: PChar;
  A: Boolean;
begin
  L := Length(S);
  if L <> 0 then
  begin
    Result := StrLowerCase(S);
    P := Pointer(Result);
    A := False;
    for I := 1 to L do
    begin
      if not (P^ in Delimiters) then
      begin
        if not A then
        begin
          A := True;
          P^ := UpperChars[Byte(P^)];
        end;
      end
      else
        if A then
          A := False;
      Inc(P);
    end;
  end
  else
    Result := '';
end;

{* Conversion functions *}

{(*}      // do not process by Source Formatter
const
  HexLookup: array[Char] of Byte = (
    $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF,
    $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF,
    $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF,
    $FF, $FF, $FF,   0,   1,   2,   3,   4,   5,   6,   7,   8,   9, $FF, $FF,
    $FF, $FF, $FF, $FF, $FF,  10,  11,  12,  13,  14,  15, $FF, $FF, $FF, $FF,
    $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF,
    $FF, $FF, $FF, $FF, $FF, $FF, $FF,  10,  11,  12,  13,  14,  15, $FF, $FF,
    $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF,
    $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF,
    $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF,
    $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF,
    $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF,
    $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF,
    $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF,
    $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF,
    $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF,
    $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF,
    $FF);
{*)}

function BaseToULong(const S: AnsiString; const BaseLog2: Byte): ULong;
var
  L: ULong;
  P: Byte;
  C: Byte;
  Q: PChar;
begin
  P := Length(S);
  if P = 0 then
  begin
    Result := 0;
    Exit;
  end;
  L := 0;
  Result := 0;
  Q := Pointer(S);
  Inc(Q, P - 1);
  repeat
    C := HexLookup[Q^];
    if C <> $FF then
      Inc(Result, ULong(C) shl L);
    Inc(L, BaseLog2);
    Dec(P);
    Dec(Q);
  until (P = 0) or (L = 32);
end;

function BinToULong(const S: AnsiString): ULong;
begin
  Result := BaseToULong(S, 1);
end;

function OctToULong(const S: AnsiString): ULong;
begin
  Result := BaseToULong(S, 3);
end;

function HexToULong(const S: AnsiString): ULong;
begin
  Result := BaseToULong(S, 4);
end;

function DecToULong(const S: AnsiString): ULong;
var
  L: Long;
  P: PChar;
  C: Char;
  F: ULong;
begin
  L := Length(S);
  if L = 0 then
  begin
    Result := 0;
    Exit;
  end;
  Result := 0;
  F := 1;
  P := Pointer(S);
  Inc(P, L - 1);
  repeat
    C := P^;
    if C in ['0'..'9'] then
      Inc(Result, Int64(Ord(C) - Ord('0')) * F);
    if F = 1000000000 then
      Exit;
    F := F * 10;
    Dec(P);
    Dec(L);
  until L = 0;
end;

function ULongToBase(Num: ULong; const Digits, Base: Byte): AnsiString;
const
  HexDigits = '0123456789ABCDEF';
var
  D: ULong;
  L: Byte;
  P: PChar;
begin
  if Num = 0 then
  begin
    if Digits = 0 then
      L := 1
    else
      L := Digits;
    SetLength(Result, L);
    FillMemory(Pointer(Result), L, Ord('0'));
    Exit;
  end;
  L := 0;
  D := Num;
  while D > 0 do
  begin
    Inc(L);
    D := D div Base;
  end;
  if L < Digits then
    L := Digits;
  SetLength(Result, L);
  P := Pointer(Result);
  Inc(P, L - 1);
  D := Num;
  while D > 0 do
  begin
    P^ := HexDigits[D mod Base + 1];
    Dec(P);
    Dec(L);
    D := D div Base;
  end;
  while L > 0 do
  begin
    P^ := '0';
    Dec(P);
    Dec(L);
  end;
end;

function ULongToBin(Num: ULong; Digits: Byte): AnsiString;
begin
  Result := ULongToBase(Num, Digits, 2);
end;

function ULongToOct(Num: ULong; Digits: Byte): AnsiString;
begin
  Result := ULongToBase(Num, Digits, 8);
end;

function ULongToHex(Num: ULong; Digits: Byte): AnsiString;
begin
  Result := ULongToBase(Num, Digits, 16);
end;

function ULongToDec(Num: ULong; Digits: Byte): AnsiString;
begin
  Result := ULongToBase(Num, Digits, 10);
end;

function RomanToULong(const S: NString): ULong;
var
  I, N, PN: Long;
begin
  N := 10000;
  Result := 0;
  for I := 1 to Length(S) do
  begin
    PN := N;
    case S[I] of
      'C': N := 100;
      'D': N := 500;
      'I': N := 1;
      'L': N := 50;
      'M': N := 1000;
      'V': N := 5;
      'X': N := 10;
      'c': N := 100;
      'd': N := 500;
      'i': N := 1;
      'l': N := 50;
      'm': N := 1000;
      'v': N := 5;
      'x': N := 10;
    else
      N := 0;
    end;
    if N <= PN then
      Inc(Result, N)
    else
      Inc(Result, N - (PN shl 1));
  end;
end;

function ULongToRoman(Num: ULong): NString;
const
  TS: array[0..5] of NString = ('', 'M', 'MM', 'MMM', 'MMMM', 'MMMMM');
  SS: array[0..9] of NString = ('', 'C', 'CC', 'CCC', 'CD', 'D', 'DC', 'DCC',
    'DCCC', 'CM');
  DS: array[0..9] of NString = ('', 'X', 'XX', 'XXX', 'XL', 'L', 'LX', 'LXX',
    'LXXX', 'XC');
  ES: array[0..9] of NString = ('', 'I', 'II', 'III', 'IV', 'V', 'VI', 'VII',
    'VIII', 'IX');
var
  T, S, D: Long;
begin
  if (Num >= 1) and (Num <= 5000) then
  begin
    T := 0;
    while Num >= 1000 do
    begin
      Dec(Num, 1000);
      Inc(T);
    end;
    S := 0;
    if Num >= 500 then
    begin
      Dec(Num, 500);
      S := 5;
    end;
    while Num >= 100 do
    begin
      Dec(Num, 100);
      Inc(S);
    end;
    D := 0;
    if Num >= 50 then
    begin
      Dec(Num, 50);
      D := 5;
    end;
    while Num >= 10 do
    begin
      Dec(Num, 10);
      Inc(D);
    end;
    Result := TS[T] + SS[S] + DS[D] + ES[Num];
  end
  else
    Result := '?????';
end;

{* Info/Count functions *}

function CountCharsNoCase(const S: AnsiString; C: Char): Long;
var
  Str: AnsiString;
  P: PChar;
  L, I: Long;
begin
  Result := 0;
  Str := Kernel.StrUpperCase(S);
  C := Kernel.StrUpperCase(C)[1];
  L := Length(Str);
  P := Pointer(Str);
  for I := 0 to L - 1 do
  begin
    if P^ = C then
      Inc(Result);
    Inc(P);
  end;
end;

function CountCharsCase(const S: NString; C: Char): Long;
var
  P: PChar;
  L, I: Long;
begin
  Result := 0;
  L := Length(S);
  P := Pointer(S);
  for I := 0 to L - 1 do
  begin
    if P^ = C then
      Inc(Result);
    Inc(P);
  end;
end;

function CountCharsNoCase(const S: AnsiString; const C: array of Char): Long;
var
  I: Long;
begin
  Result := 0;
  for I := Low(C) to High(C) do
    Result := Result + CountCharsNoCase(S, C[I]);
end;

function CountCharsCase(const S: AnsiString; C: array of Char): Long;
var
  I: Long;
begin
  Result := 0;
  for I := Low(C) to High(C) do
    Result := Result + CountCharsCase(S, C[I]);
end;

function StrRefCount(const S: AnsiString): Long;
var
  P: Pointer;
begin
  Result := 0;
  if Pointer(S) <> nil then
  begin
    P := Pointer(Long(Pointer(S)) - 8);
    Result := Long(P^);
  end;
end;

{* CharSet functions *}

function GetCharSet(const S: AnsiString): TCharSet;
var
  P: PChar;
  L, I: Long;
begin
  Result := [];
  L := Length(S);
  P := Pointer(S);
  for I := 0 to L - 1 do
  begin
    if not (P^ in Result) then
      Result := Result + [P^];
    Inc(P);
  end;
end;

function CharSetToStr(const CharSet: TCharSet): AnsiString;
{$IFDEF ASMVERSIONS}
asm
  PUSH    EBX
  MOV     ECX, $100
  MOV     EBX, EAX
  PUSH    ESI
  MOV     EAX, EDX
  SUB     ESP, ECX
  XOR     ESI, ESI
  XOR     EDX, EDX
@Loop:
  BT      [EBX], EDX
  JC      @Mm
@Nx:
  INC     EDX
  DEC     ECX
  JNE     @Loop
  MOV     ECX, ESI
  MOV     EDX, ESP
  CALL    System.@LStrFromPCharLen
  ADD     ESP, $100
  POP     ESI
  POP     EBX
  RET
@Mm:
  MOV     BYTE PTR [ESP + ESI], DL
  INC     ESI
  JMP     @Nx
end;
{$ELSE}
var
  C: Char;
begin
  Result := '';
  for C := #0 to #255 do
    if C in CharSet then
      Result := Result + C;
end;
{$ENDIF}

end.

