{*************************************************************************}
{*   BsaArchive                                                          *}
{*   Copyright (C) 2003 Marek Mauder, Vampyre SoftWorks                  *}
{*   Modules\Archives\Bsa\BsaArchive.pas                                 *}
{*************************************************************************}
{                                                                         }
{ This program is free software; you can redistribute it and/or modify it }
{ under the terms of the GNU General Public License as published by the   }
{ Free Software Foundation; either version 2 of the License, or (at your  }
{ option) any later version. This program is distributed in the hope that }
{ it will be useful, but WITHOUT ANY WARRANTY; without even the implied   }
{ warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See   }
{ the GNU General Public License for more details.                        }
{                                                                         }
{ You should have received a copy of the GNU General Public License       }
{ along with this program; if not, write to the Free Software Foundation, }
{ Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.         }
{                                                                         }
{*************************************************************************}
{  Created On: 7th November 2003  (07.11.2003)                            }
{*************************************************************************}

{This unit contains class which provides access to Bethesda's BSA archives.
 Note that it supports BSA file from older games (Arena, Terminator, Daggerfall,
  ...) and NOT FROM MORROWIND.}

unit BsaArchive;

{$I Compiler.inc}

interface

uses
  SysUtils, Kernel, Objects;

type

  TBsaArchiveType = (
    baArena, // Arc from TES: Arena, Termanator: Future Shock,...
    baNamed, // Arc from TES: Daggerfall with file name entries
    baNumbered // Arc from TES: Daggerfall with file numbered entries
    );

  TBsaHeader = packed record
    Count: Word; // Number of entries in directory
    DirectoryType: Word; // Type of directoy
  end;

  TBsaNamedEntry = packed record
    Name: array[0..13] of Char;
    Size: Long;
  end;

  TBsaNumberedEntry = packed record
    ID: Long;
    Size: Long;
  end;

  TDirEntry = record
    Name: NString;
    Offset: Long;
    Size: Long;
  end;
  PDirEntry = ^TDirEntry;

  TNBsaArchive = class(TObject)
  protected
    FDirectory: TNSortedStringPointers;
    FArchiveType: TBsaArchiveType;
    FFileName: NString;
    FStream: TNFileStream;
    {Parses bsa archive's directory and adds entries to FDirectory list.}
    procedure UpdateDirectory;
  public
    {Opens archive.}
    constructor Create(const AFileName: NString);
    destructor Destroy; override;
    {Fills Strings with filenames of archive's content.}
    procedure GetFileList(Strings: TNStringList);
    {Extracts given file from archive. Note that bsa archive supports only
     8.3 dos file names. Records from numbered archives can be extracted too,
     just call ExtractFile(IntToStr(SomeID),P,I).}
    procedure ExtractFile(FileName: NString; var Data: Pointer; var Size: Long);
    {List of TDirEntry pointers. Pointers are stored in Objects list.}
    property Directory: TNSortedStringPointers read FDirectory;
    {Type of this archive.}
    property ArchiveType: TBsaArchiveType read FArchiveType;
    {FileName of this archive.}
    property FileName: NString read FFileName;
    {File stream containing this bsa archive.}
    property Stream: TNFileStream read FStream;
  end;


{*************************************************************************}
{                             IMPLEMEMTATION                              }
{*************************************************************************}


implementation


{*************************************************************************}
{                         TBsaArchive methods                             }
{*************************************************************************}


constructor TNBsaArchive.Create(const AFileName: NString);
begin
  inherited Create;
  FStream := nil;
  FFileName := AFileName;
  FDirectory := TNSortedStringPointers.Create;
  FStream := TNFileStream.Create(AFileName, fmOpenRead);
  UpdateDirectory;
end;

destructor TNBsaArchive.Destroy;
var
  I: Long;
begin
  for I := 0 to FDirectory.Count - 1 do
    Dispose(PDirEntry(FDirectory.Pointers[I]));
  FDirectory.Free;
  FStream.Free;
  inherited Destroy;
end;

procedure TNBsaArchive.UpdateDirectory;
var
  Hdr: TBsaHeader;
  NaEnt: TBsaNamedEntry;
  NuEnt: TBsaNumberedEntry;
  I: Long;
  ESize: Long;
  Off: Long;
  Entry: PDirEntry;
  Buffer: Pointer;
  BufPos: Long;
begin
  if FStream = nil then
    Exit;
  FStream.Position := 0;
  FStream.Read(Hdr, SizeOf(Hdr));
  case Hdr.DirectoryType of
    $0100: FArchiveType := baNamed;
    $0200: FArchiveType := baNumbered;
    else
      FArchiveType := baArena;
  end;
  ESize := Iff(FArchiveType = baNumbered, Long(SizeOf(NuEnt)),
    Long(SizeOf(NaEnt)));
  Off := Iff(FArchiveType = baArena, 2, 4);
  FStream.Position := FStream.Size - Hdr.Count * ESize;
  GetMem(Buffer, Hdr.Count * ESize);
  FStream.Read(Buffer^, Hdr.Count * ESize);
  BufPos := 0;
  for I := 0 to Hdr.Count - 1 do
  begin
    New(Entry);
    if FArchiveType = baNumbered then
    begin
      Move(Pointer(Long(Buffer) + BufPos)^, NuEnt, ESize);
      Entry^.Name := IntToStr(NuEnt.ID);
      Entry^.Size := NuEnt.Size;
      Entry^.Offset := Off;
    end
    else
    begin
      Move(Pointer(Long(Buffer) + BufPos)^, NaEnt, ESize);
      Entry^.Name := NaEnt.Name;
      Entry^.Size := NaEnt.Size;
      Entry^.Offset := Off;
    end;
    BufPos := BufPos + ESize;
    Off := Off + Entry^.Size;
    FDirectory.Add(Entry^.Name, Entry);
  end;
end;

procedure TNBsaArchive.GetFileList(Strings: TNStringList);
begin
  Strings.Clear;
  Strings.AddStrings(FDirectory);
end;

procedure TNBsaArchive.ExtractFile(FileName: NString; var Data: Pointer;
  var Size: Long);
var
  I: Long;
  Entry: PDirEntry;
begin
  if FDirectory.Find(FileName, I) then
  begin
    Entry := PDirEntry(FDirectory.Pointers[I]);
    FStream.Position := Entry^.Offset;
    Size := Entry^.Size;
    GetMem(Data, Size);
    FStream.Read(Data^, Size);
  end;
end;

end.

