{*************************************************************************}
{*   CifFormat                                                           *}
{*   Copyright (C) 2003 Marek Mauder, Vampyre SoftWorks                  *}
{*   Modules\ImageFileFormats\Daggerfall\CifFormat.pas                   *}
{*************************************************************************}
{                                                                         }
{ This program is free software; you can redistribute it and/or modify it }
{ under the terms of the GNU General Public License as published by the   }
{ Free Software Foundation; either version 2 of the License, or (at your  }
{ option) any later version. This program is distributed in the hope that }
{ it will be useful, but WITHOUT ANY WARRANTY; without even the implied   }
{ warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See   }
{ the GNU General Public License for more details.                        }
{                                                                         }
{ You should have received a copy of the GNU General Public License       }
{ along with this program; if not, write to the Free Software Foundation, }
{ Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.         }
{                                                                         }
{*************************************************************************}
{  Created On: 9th November 2003  (09.11.2003)                            }
{*************************************************************************}

{This unit contains image file format loader/saver for Daggerfall CIF
 animations.}

unit CifFormat;

{$I Compiler.inc}

interface

uses
  Kernel, Objects, Platforms, Graphic, Images, DaggerGraphic;

type

  {This module proveides loading/saving of images in CIF format.
  This format is simple non-compressed 8bit indexed containing more frames.
  But some files are without header and some are RLEd. So exact file size must be known.
  This is why Stream functions must contain only one CIF file.}
  TNCifFormat = class(TNDaggerFormat)
  protected
    procedure DoLoadImage(var Image: TImageData; Stream: TNStream); override;
    procedure DoSaveImage(var Image: TImageData; Stream: TNStream;
      SaveCompressed: Boolean); override;
  public
    constructor Create; override;
  end;

const

  sCifExtensions = 'cif,rci';
  sCifFormatName = 'Daggerfall Cif MultiImage';

var

  CifInstance: TNCifFormat = nil;


{*************************************************************************}
{                             IMPLEMEMTATION                              }
{*************************************************************************}


implementation


{*************************************************************************}
{                         TNCifFormat methods                             }
{*************************************************************************}


constructor TNCifFormat.Create;
begin
  inherited Create;
  FExtensions := sCifExtensions;
  FFormatName := sCifFormatName;
  FCanLoad := True;
  FCanSave := True;
end;

procedure TNCifFormat.DoLoadImage(var Image: TImageData; Stream: TNStream);
var
  Hdr: TImgHeader;
  Grp: TCifGroup;
  HasHdr: boolean;
  IsWeapon, ISW9, IsStd, IsFirst: boolean;
  Data: Pointer;
  OldSPos: Int64;
  W, H, ALev, I, S: Long;
begin
  with Stream, Image do
  begin
    Image.DataFormat := FindImageDataFormat(idfIndex8);
    GetMem(Palette, 256 * SizeOf(TARGBRec));
    Move(FPalette, Palette^, SizeOf(FPalette));
    PalSwapChannels(Image.Palette, 256);
    W := 0;
    H := 0;
    // solve cif type
    HasHdr := True;
    IsWeapon := False;
    IsW9 := False;
    IsFirst := True;
    ReadBuffer(Hdr, SizeOf(Hdr));
    if (Stream.Size = 249856) or (Stream.Size = 49152) or
      (Stream.Size = 2060295) then
    begin
      // image has no header (FACES.CIF, TFAC00I0.RCI, CHLD00I0.RCI)
      W := 64;
      H := 64;
      HasHdr := False;
    end;
    if (Stream.Size = 10752) then
    begin
      // image has no header (BUTTONS.RCI)
      W := 32;
      H := 16;
      HasHdr := False;
    end;
    if (Stream.Size = 2601) then
    begin
      // image has no header (MPOP.RCI)
      W := 17;
      H := 17;
      HasHdr := False;
    end;
    if (Stream.Size = 4356) then
    begin
      // image has no header (SPOP.RCI)
      W := 22;
      H := 22;
      HasHdr := False;
    end;
    if (Stream.Size = 3168) then
    begin
      // image has no header (NOTE.RCI)
      W := 44;
      H := 9;
      HasHdr := False;
    end;
    if Hdr.Unk = $15 then
    begin
      // this file is Weapon09.cif
      IsWeapon := True;
      IsW9 := True;
    end;
    if HasHdr then
      if Position + Hdr.ImageSize < Stream.Size then
      begin
        Position := Position + Hdr.ImageSize;
        ReadBuffer(Grp, SizeOf(Grp));
        if Grp.Offsets[0] = 76 then
          // cif is regular weapon
          IsWeapon := True;
      end;
    IsStd := HasHdr and (not IsWeapon);
    Position := 0;
    while Position < Stream.Size do
    begin
      // process standard cif file
      if IsStd then
      begin
        ReadBuffer(Hdr, SizeOf(Hdr));
        AddImageLevel(Image);
        ALev := Image.Count - 1;
        Levels[ALev].Width := Hdr.Width;
        Levels[ALev].Height := Hdr.Height;
        Levels[ALev].Size := Hdr.Width * Hdr.Height;
        GetMem(Data, Hdr.ImageSize);
        ReadBuffer(Data^, Hdr.ImageSize);
        if Hdr.Unk <> 2 then
        begin
          // uncompressed
          GetMem(Levels[ALev].Bits, Levels[ALev].Size);
          Move(Data^, Levels[ALev].Bits^, Levels[ALev].Size);
        end
        else
          //rle compressed
          DagRLEDecode(Data, Hdr.ImageSize, Levels[ALev].Size,
            Levels[ALev].Bits);
        FreeMem(Data);
      end;
      // process file without header
      if not HasHdr then
        if Position + W * H >= Stream.Size then
          Position := Stream.Size
        else
        begin
          AddImageLevel(Image);
          ALev := Image.Count - 1;
          Levels[ALev].Width := W;
          Levels[ALev].Height := H;
          Levels[ALev].Size := W * H;
          GetMem(Levels[ALev].Bits, Levels[ALev].Size);
          ReadBuffer(Levels[ALev].Bits^, Levels[ALev].Size);
        end;
      // process weapon file
      if IsWeapon then
        // first frame is std IMG file, next are not
        // but if IsW9 this first frame is missing
        if IsFirst and (not IsW9) then
        begin
          ReadBuffer(Hdr, SizeOf(Hdr));
          AddImageLevel(Image);
          ALev := Image.Count - 1;
          Levels[ALev].Width := Hdr.Width;
          Levels[ALev].Height := Hdr.Height;
          Levels[ALev].Size := Hdr.ImageSize;
          GetMem(Levels[ALev].Bits, Levels[ALev].Size);
          ReadBuffer(Levels[ALev].Bits^, Levels[ALev].Size);
          IsFirst := False;
        end
        else
        begin
          OldSPos := Position;
          // read next group
          ReadBuffer(Grp, SizeOf(Grp));
          // read images in group
          I := 0;
          while Grp.Offsets[I] <> 0 do
          begin
            S := Grp.Offsets[I + 1] - Grp.Offsets[I];
            if S < 0 then
              S := Grp.Offsets[31] - Grp.Offsets[I];
            Position := OldSPos + Grp.Offsets[I];
            AddImageLevel(Image);
            ALev := Image.Count - 1;
            Levels[ALev].Width := Grp.Width;
            Levels[ALev].Height := Grp.Height;
            Levels[ALev].Size := Grp.Width * Grp.Height;
            GetMem(Data, S);
            ReadBuffer(Data^, S);
            DagRLEDecode(Data, Grp.ImageSize, Levels[ALev].Size,
              Levels[ALev].Bits);
            FreeMem(Data);
            Inc(I);
          end;
          Position := OldSPos + Grp.Offsets[31];
        end;
    end;
  end;
end;

procedure TNCifFormat.DoSaveImage(var Image: TImageData; Stream: TNStream;
  SaveCompressed: Boolean);
var
  Hdr: TImgHeader;
  I: Long;
  Clone: TImageData;

  procedure SaveLevel(Level: TImageLevel);
  var
    I: Long;
  begin
    with Level do
    begin
      if Size > 0 then
      begin
        FillChar(Hdr, SizeOf(Hdr), 0);
        Hdr.Width := Width;
        Hdr.Height := Height;
        Hdr.ImageSize := Width * Height;
        Stream.WriteBuffer(Hdr, SizeOf(Hdr));
        Stream.WriteBuffer(Bits^, Hdr.ImageSize);
      end;
      for I := 0 to Count - 1 do
        SaveLevel(Levels[I]);
    end;
  end;

begin
  PrepareImage(Image, Clone);
  with Stream, Clone do
  try
    for I := 0 to Count - 1 do
      SaveLevel(Clone.Levels[I]);
  finally
    FreeImage(Clone);
  end;
end;


{*************************************************************************}
{                         Internal Stuff                                  }
{*************************************************************************}


initialization

  CifInstance := TNCifFormat.Create;

finalization

  FreeAndNil(CifInstance);

end.

