{*************************************************************************}
{*   ImgFormat                                                           *}
{*   Copyright (C) 2003 Marek Mauder, Vampyre SoftWorks                  *}
{*   Modules\ImageFileFormats\Daggerfall\ImgFormat.pas                   *}
{*************************************************************************}
{                                                                         }
{ This program is free software; you can redistribute it and/or modify it }
{ under the terms of the GNU General Public License as published by the   }
{ Free Software Foundation; either version 2 of the License, or (at your  }
{ option) any later version. This program is distributed in the hope that }
{ it will be useful, but WITHOUT ANY WARRANTY; without even the implied   }
{ warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See   }
{ the GNU General Public License for more details.                        }
{                                                                         }
{ You should have received a copy of the GNU General Public License       }
{ along with this program; if not, write to the Free Software Foundation, }
{ Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.         }
{                                                                         }
{*************************************************************************}
{  Created On: 9th November 2003  (09.11.2003)                            }
{*************************************************************************}

{This unit contains image file format loader/saver for Daggerfall IMG files.}

unit ImgFormat;

{$I Compiler.inc}

interface

uses
  Kernel, Objects, Platforms, Graphic, Images, DaggerGraphic;

type

  {This module proveides loading/saving of images in IMG format.
  This format is simple non-compressed 8bit indexed.
  But some files are without header!  so exact file size must be known.
  This is why Stream functions must contain only one IMG file.}
  TNImgFormat = class(TNDaggerFormat)
  protected
    procedure DoLoadImage(var Image: TImageData; Stream: TNStream); override;
    procedure DoSaveImage(var Image: TImageData; Stream: TNStream;
      SaveCompressed: Boolean); override;
  public
    constructor Create; override;
  end;

const

  sImgExtensions = 'img';
  sImgFormatName = 'Daggerfall Img Image';

var

  ImgInstance: TNImgFormat = nil;


{*************************************************************************}
{                             IMPLEMEMTATION                              }
{*************************************************************************}


implementation


{*************************************************************************}
{                         TNImgFormat methods                             }
{*************************************************************************}


constructor TNImgFormat.Create;
begin
  inherited Create;
  FExtensions := sImgExtensions;
  FFormatName := sImgFormatName;
  FCanLoad := True;
  FCanSave := True;
end;

procedure TNImgFormat.DoLoadImage(var Image: TImageData; Stream: TNStream);
var
  Hdr: TImgHeader;
  Pal: TPalRGB;
  PalA: TPalette;
  Data: Pointer;
  IsRLE: boolean;

  procedure SetDims;

    procedure SetWH(W, H: Long);
    begin
      Image.Levels[0].Width := W;
      Image.Levels[0].Height := H;
      Image.Levels[0].Size := W * H;
    end;

  begin
    // special images, without header
    case Stream.Size of
      64: SetWH(8, 8); // Arena file
      90: SetWH(9, 10); // Arena file
      128: SetWH(8, 16); // Arena file
      720: SetWH(9, 80);
      990: SetWH(45, 22);
      1720: SetWH(43, 40);
      2140: SetWH(107, 20);
      2916: SetWH(36, 81);
      3200: SetWH(40, 80);
      3938: SetWH(179, 22);
      4096: SetWH(64, 64); // textures from TES:Arena
      4280: SetWH(107, 40);
      4508: SetWH(322, 14);
      20480: SetWH(320, 64);
      26496: SetWH(207, 128);
      64000: SetWH(320, 200);
      64768: SetWH(320, 200); //this contains also palette -> parsed further
      68800: SetWH(320, 215);
      112128: SetWH(512, 219);
    end;
  end;

begin
  AddImageLevel(Image);
  // image has null header or is empty
  if Stream.Size <= 12 then
  begin
    NewImage(Image, 8, 8, FindImageDataFormat(idfIndex8));
  end
  else
    with Stream, Image, Image.Levels[0] do
    begin
      Image.DataFormat := FindImageDataFormat(idfIndex8);
      IsRLE := False;
      ReadBuffer(Hdr, SizeOf(Hdr));
      // solve many img types
      if (Hdr.Width * Hdr.Height = Hdr.ImageSize) and (
        (Hdr.Width * Hdr.Height = Stream.Size - SizeOf(Hdr)) or
        (Hdr.Width * Hdr.Height = Stream.Size - SizeOf(Hdr) - 768)) then
      begin
        // uncompressed with header
        Width := Hdr.Width;
        Height := Hdr.Height;
        Size := Width * Height;
      end
      else
        if ((Hdr.ImageSize = Stream.Size - SizeOf(Hdr)) or
          (Hdr.ImageSize = Stream.Size - SizeOf(Hdr) - 768)) and
          (Hdr.Unk <> 0) then
        begin
          // compressed with header
          Width := Hdr.Width;
          Height := Hdr.Height;
          Size := Width * Height;
          IsRLE := True;
        end
        else
        begin
          // uncompressed, no header
          SetDims;
          Position := Position - SizeOf(Hdr);
        end;
      if not IsRLE then
      begin
        // read uncompressed data
        if (Width = 0) or (Height = 0) then
          // file has no header and dims are unknown
          NewImage(Image, 8, 8, FindImageDataFormat(idfIndex8))
        else
        begin
          GetMem(Bits, Size);
          ReadBuffer(Bits^, Size);
        end;
      end
      else
        if Hdr.Unk = 2 then
        begin
          // read compressed data Daggerfall
          GetMem(Data, Hdr.ImageSize);
          ReadBuffer(Data^, Hdr.ImageSize);
          DagRLEDecode(Data, Hdr.ImageSize, Size, Bits);
        end
        else
        begin
          // read compressed data Arena
          // compress algorhitm is unknown to me now
          // if Unk = 264 then after header is word size of original data
          // if Unk = 260 no size after head
          NewImage(Image, Width, Height, FindImageDataFormat(idfIndex8));
        end;
      // palette creation
      PalA := FPalette;
      if (Stream.Size = Stream.Position + 768) then
      begin
        ReadBuffer(Pal, SizeOf(Pal));
        PalA := RGBPalToARGBPal(Pal, $FF);
      end;
      GetMem(Palette, 256 * SizeOf(TARGBRec));
      Move(PalA, Image.Palette^, SizeOf(PalA));
      PalSwapChannels(Image.Palette, 256);
    end;
end;

procedure TNImgFormat.DoSaveImage(var Image: TImageData; Stream: TNStream;
  SaveCompressed: Boolean);
var
  Hdr: TImgHeader;
  Clone: TImageData;
  Level: PImageLevel;
begin
  PrepareImage(Image, Clone);
  Level := FindFirstValidLevel(Clone);
  if Level = nil then
    Level := @Clone.Levels[0];
  with Stream, Clone, Level^ do
  try
    FillChar(Hdr, SizeOf(Hdr), 0);
    Hdr.Width := Width;
    Hdr.Height := Height;
    Hdr.ImageSize := Width * Height;
    WriteBuffer(Hdr, SizeOf(Hdr));
    WriteBuffer(Bits^, Hdr.ImageSize);
  finally
    FreeImage(Clone);
  end;
end;


{*************************************************************************}
{                         Internal Stuff                                  }
{*************************************************************************}


initialization

  ImgInstance := TNImgFormat.Create;

finalization

  FreeAndNil(ImgInstance);

end.

