{*****************************************************************************
*  ZLibEx.pas                                                                *
*                                                                            *
*  copyright (c) 2002 Roberto Della Pasqua                                   *
*  copyright (c) 2000-2002 base2 technologies                                *
*  copyright (c) 1997 Borland International                                  *
*                                                                            *
*  revision history                                                          *
*    2002.11.02  ZSendToBrowser: deflate algorithm for HTTP1.1 compression   *
*    2002.10.24  ZFastCompressString and ZFastDecompressString:300% faster   *
*    2002.10.15  recompiled zlib 1.1.4 c sources with speed optimizations    *
*                (and targeting 686+ cpu) and changes to accomodate Borland  *
*                standards  (C++ v5.6 compiler)                              *
*    2002.10.15  optimized move mem for not aligned structures  (strings,etc)*
*    2002.10.15  little changes to avoid system unique string calls          *
*                                                                            *
*    2002.03.15  updated to zlib version 1.1.4                               *
*    2001.11.27  enhanced TZDecompressionStream.Read to adjust source        *
*                  stream position upon end of compression data              *
*                fixed endless loop in TZDecompressionStream.Read when       *
*                  destination count was greater than uncompressed data      *
*    2001.10.26  renamed unit to integrate "nicely" with delphi 6            *
*    2000.11.24  added soFromEnd condition to TZDecompressionStream.Seek     *
*                added ZCompressStream and ZDecompressStream                 *
*    2000.06.13  optimized, fixed, rewrote, and enhanced the zlib.pas unit   *
*                  included on the delphi cd (zlib version 1.1.3)            *
*                                                                            *
*  acknowledgements                                                          *
*    erik turner    Z*Stream routines                                        *
*    david bennion  finding the nastly little endless loop quirk with the    *
*                     TZDecompressionStream.Read method                      *
*    burak kalayci  informing me about the zlib 1.1.4 update                 *
*****************************************************************************}


{*************************************************************************}
{*   ZLibEx                                                              *}
{*   Marek Mauder, Vampyre SoftWorks                                     *}
{*************************************************************************}
{                                                                         }
{    Some code was removed to get smaller exefile. Using only             }
{    Compress/Decompress Buff functions. Added support for Linux.         }
{                                                                         }
{*************************************************************************}
{  Modified On: 22nd September 03  (22.09.2003)                           }
{*************************************************************************}


unit ZLibEx;

interface

uses
  SysUtils;

const
  ZLIB_VERSION = '1.1.4';

type
  TZAlloc = function(opaque: Pointer; items, size: Integer): Pointer;
  {$IFDEF MSWINDOWS} register;{$ENDIF}
  {$IFDEF LINUX} cdecl;{$ENDIF}
  TZFree = procedure(opaque, block: Pointer);
  {$IFDEF MSWINDOWS} register;{$ENDIF}
  {$IFDEF LINUX} cdecl;{$ENDIF}

  TZCompressionLevel = (zcNone, zcFastest, zcDefault, zcMax);

  {** TZStreamRec ***********************************************************}

  TZStreamRec = packed record
    next_in: PChar; // next input byte
    avail_in: Longint; // number of bytes available at next_in
    total_in: Longint; // total nb of input bytes read so far

    next_out: PChar; // next output byte should be put here
    avail_out: Longint; // remaining free space at next_out
    total_out: Longint; // total nb of bytes output so far

    msg: PChar; // last error message, NULL if no error
    state: Pointer; // not visible by applications

    zalloc: TZAlloc; // used to allocate the internal state
    zfree: TZFree; // used to free the internal state
    opaque: Pointer; // private data object passed to zalloc and zfree

    data_type: Integer; // best guess about the data type: ascii or binary
    adler: Longint; // adler32 value of the uncompressed data
    reserved: Longint; // reserved for future use
  end;


{** zlib public routines ****************************************************}

{*****************************************************************************
*  CompressBuffer                                                            *
*                                                                            *
*  pre-conditions                                                            *
*    InData  = pointer to uncompressed data                                  *
*    InSize    = size of inBuffer (bytes)                                    *
*    OutData = pointer (unallocated)                                         *
*    level     = compression level                                           *
*                                                                            *
*  post-conditions                                                           *
*    OutData = pointer to compressed data (allocated)                        *
*    OutSize   = size of outBuffer (bytes)                                   *
*****************************************************************************}

procedure CompressBuffer(const InData: Pointer; InSize: LongInt;
  var OutData: Pointer; var OutSize: LongInt;
  Level: TZCompressionLevel = zcDefault);

{*****************************************************************************
*  DecompressBuffer                                                          *
*                                                                            *
*  pre-conditions                                                            *
*    InData    = pointer to compressed data                                  *
*    InSize      = size of inBuffer (bytes)                                  *
*    OutData   = pointer (unallocated)                                       *
*    OutEstimate = estimated size of uncompressed data (bytes)               *
*                                                                            *
*  post-conditions                                                           *
*    OutData = pointer to decompressed data (allocated)                      *
*    OutSize   = size of outBuffer (bytes)                                   *
*****************************************************************************}

procedure DecompressBuffer(const InData: Pointer; InSize: LongInt;
  var OutData: Pointer; var OutSize: LongInt; OutEstimate: LongInt = 0);


type
  EZLibError = class(Exception);
  EZCompressionError = class(EZLibError);
  EZDecompressionError = class(EZLibError);

{** link zlib code **********************************************************}

{$IFDEF MSWINDOWS}
{$L deflate.obj}
{$L inflate.obj}
{$L infblock.obj}
{$L inftrees.obj}
{$L infcodes.obj}
{$L infutil.obj}
{$L inffast.obj}
{$L trees.obj}
{$L adler32.obj}
{$ENDIF}

{*****************************************************************************
*  note: do not reorder the above -- doing so will result in external        *
*  functions being undefined                                                 *
*****************************************************************************}

const
  {** flush constants *******************************************************}

  Z_NO_FLUSH = 0;
  Z_PARTIAL_FLUSH = 1;
  Z_SYNC_FLUSH = 2;
  Z_FULL_FLUSH = 3;
  Z_FINISH = 4;

  {** return codes **********************************************************}

  Z_OK = 0;
  Z_STREAM_END = 1;
  Z_NEED_DICT = 2;
  Z_ERRNO = (-1);
  Z_STREAM_ERROR = (-2);
  Z_DATA_ERROR = (-3);
  Z_MEM_ERROR = (-4);
  Z_BUF_ERROR = (-5);
  Z_VERSION_ERROR = (-6);

  {** compression levels ****************************************************}

  Z_NO_COMPRESSION = 0;
  Z_BEST_SPEED = 1;
  Z_BEST_COMPRESSION = 9;
  Z_DEFAULT_COMPRESSION = (-1);

  {** compression strategies ************************************************}

  Z_FILTERED = 1;
  Z_HUFFMAN_ONLY = 2;
  Z_DEFAULT_STRATEGY = 0;

  {** data types ************************************************************}

  Z_BINARY = 0;
  Z_ASCII = 1;
  Z_UNKNOWN = 2;

  {** compression methods ***************************************************}

  Z_DEFLATED = 8;

  {** return code messages **************************************************}

  _z_errmsg: array[0..9] of PChar = (
    'need dictionary', // Z_NEED_DICT      (2)
    'stream end', // Z_STREAM_END     (1)
    '', // Z_OK             (0)
    'file error', // Z_ERRNO          (-1)
    'stream error', // Z_STREAM_ERROR   (-2)
    'data error', // Z_DATA_ERROR     (-3)
    'insufficient memory', // Z_MEM_ERROR      (-4)
    'buffer error', // Z_BUF_ERROR      (-5)
    'incompatible version', // Z_VERSION_ERROR  (-6)
    ''
    );

  ZLevels: array[TZCompressionLevel] of Shortint = (
    Z_NO_COMPRESSION,
    Z_BEST_SPEED,
    Z_DEFAULT_COMPRESSION,
    Z_BEST_COMPRESSION
    );

  SZInvalid = 'Invalid ZStream operation!';


{****************************************************************************}

{$IFDEF LINUX}
const
  zlibso = 'libz.so.1';
{$ENDIF}

{** deflate routines ********************************************************}

function deflateInit_(var strm: TZStreamRec; level: Integer; version: PChar;
  recsize: Integer): Integer;
{$IFDEF MSWINDOWS} external;{$ENDIF}
{$IFDEF LINUX} cdecl; external zlibso name 'deflateInit_';{$ENDIF}
function deflate(var strm: TZStreamRec; flush: Integer): Integer;
{$IFDEF MSWINDOWS} external;{$ENDIF}
{$IFDEF LINUX} cdecl; external zlibso name 'deflate';{$ENDIF}
function deflateEnd(var strm: TZStreamRec): Integer;
{$IFDEF MSWINDOWS} external;{$ENDIF}
{$IFDEF LINUX} cdecl; external zlibso name 'deflateEnd';{$ENDIF}

{** inflate routines ********************************************************}

function inflateInit_(var strm: TZStreamRec; version: PChar;
  recsize: Integer): Integer;
{$IFDEF MSWINDOWS}external;{$ENDIF}
{$IFDEF LINUX} cdecl; external zlibso name 'inflateInit_';{$ENDIF}
function inflateInit2_(var strm: TZStreamRec;
                       W : Integer;
                       const Version : string;
                       stream_size : Integer) : Integer;
{$IFDEF MSWINDOWS}external;{$ENDIF}
{$IFDEF LINUX} cdecl; external zlibso name 'inflateInit2_';{$ENDIF}
function inflate(var strm: TZStreamRec; flush: Integer): Integer;
{$IFDEF MSWINDOWS} external;{$ENDIF}
{$IFDEF LINUX} cdecl; external zlibso name 'inflate';{$ENDIF}
function inflateEnd(var strm: TZStreamRec): Integer;
{$IFDEF MSWINDOWS} external;{$ENDIF}
{$IFDEF LINUX} cdecl; external zlibso name 'inflateEnd';{$ENDIF}
function inflateReset(var strm: TZStreamRec): Integer;
{$IFDEF MSWINDOWS} external;{$ENDIF}
{$IFDEF LINUX} cdecl; external zlibso name 'inflateReset';{$ENDIF}

function zcAlloc(AppData: Pointer; Items, Size: Integer): Pointer;
{$IFDEF MSWINDOWS} register;{$ENDIF}
{$IFDEF LINUX} cdecl;{$ENDIF}

procedure zcFree(AppData, Block: Pointer);
{$IFDEF MSWINDOWS} register;{$ENDIF}
{$IFDEF LINUX} cdecl;{$ENDIF}


implementation


{** zlib function implementations *******************************************}

function zcAlloc(AppData: Pointer; Items, Size: Integer): Pointer;
{$IFDEF MSWINDOWS} register;{$ENDIF}
{$IFDEF LINUX} cdecl;{$ENDIF}
begin
  Result := AllocMem(Items * Size);
end;

procedure zcFree(AppData, Block: Pointer);
{$IFDEF MSWINDOWS} register;{$ENDIF}
{$IFDEF LINUX} cdecl;{$ENDIF}
begin
  FreeMem(Block);
end;

{$IFDEF MSWINDOWS}
{** c function implementations **********************************************}

procedure _memset(p: Pointer; b: Byte; count: Integer); cdecl;
begin
  FillChar(p^, count, b);
end;

procedure _memcpy(dest, source: Pointer; count: Integer); cdecl;
begin
  Move(source^, dest^, count);
end;
{$ENDIF}

{** custom zlib routines ****************************************************}

function DeflateInit(var stream: TZStreamRec; level: Integer): Integer;
begin
  result := DeflateInit_(stream, level, ZLIB_VERSION, SizeOf(TZStreamRec));
end;

function InflateInit(var stream: TZStreamRec): Integer;
begin
  result := InflateInit_(stream, ZLIB_VERSION, SizeOf(TZStreamRec));
end;

{****************************************************************************}

function ZCompressCheck(code: Integer): Integer;
begin
  result := code;
  if code < 0 then
  begin
    raise EZCompressionError.Create(_z_errmsg[2 - code]);
  end;
end;

function ZDecompressCheck(code: Integer): Integer;
begin
  Result := code;
  if code < 0 then
  begin
    raise EZDecompressionError.Create(_z_errmsg[2 - code]);
  end;
end;

procedure CompressBuffer(const InData: Pointer; InSize: LongInt;
  var OutData: Pointer; var OutSize: LongInt;
  Level: TZCompressionLevel = zcDefault);
const
  delta = 256;
var
  zstream: TZStreamRec;
begin
  FillChar(zstream, SizeOf(TZStreamRec), 0);
  outSize := ((inSize + (inSize div 10) + 12) + 255) and not 255;
  GetMem(outData, outSize);
  try
    zstream.next_in := inData;
    zstream.avail_in := inSize;
    zstream.next_out := outData;
    zstream.avail_out := outSize;
    ZCompressCheck(DeflateInit(zstream, ZLevels[level]));
    try
      while ZCompressCheck(deflate(zstream, Z_FINISH)) <> Z_STREAM_END do
      begin
        Inc(outSize, delta);
        ReallocMem(outData, outSize);
        zstream.next_out := PChar(Integer(outData) + zstream.total_out);
        zstream.avail_out := delta;
      end;
    finally
      ZCompressCheck(deflateEnd(zstream));
    end;
    ReallocMem(outData, zstream.total_out);
    outSize := zstream.total_out;
  except
    FreeMem(outData);
    raise;
  end;
end;

procedure DecompressBuffer(const InData: Pointer; InSize: LongInt;
  var OutData: Pointer; var OutSize: LongInt; OutEstimate: LongInt = 0);
var
  zstream: TZStreamRec;
  delta: Integer;
begin
  FillChar(zstream, SizeOf(TZStreamRec), 0);
  delta := (inSize + 255) and not 255;
  if outEstimate = 0 then
    outSize := delta
  else
    outSize := outEstimate;
  GetMem(outData, outSize);
  try
    zstream.next_in := inData;
    zstream.avail_in := inSize;
    zstream.next_out := outData;
    zstream.avail_out := outSize;
    ZDecompressCheck(InflateInit(zstream));
    try
      while ZDecompressCheck(inflate(zstream, Z_NO_FLUSH)) <> Z_STREAM_END do
      begin
        Inc(outSize, delta);
        ReallocMem(outData, outSize);
        zstream.next_out := PChar(Integer(outData) + zstream.total_out);
        zstream.avail_out := delta;
      end;
    finally
      ZDecompressCheck(inflateEnd(zstream));
    end;
    ReallocMem(outData, zstream.total_out);
    outSize := zstream.total_out;
  except
    FreeMem(outData);
    outData:=nil;
    raise;
  end;
end;

end.


