unit ACompilers;

interface

uses
  Windows, Kernel, SysUtils, Classes, Objects, CBase, ABase, CProjects, Lists,
  Files, ComCtrls, Forms;

type

  TCommonCompiler = class(TVideCompiler)
  private
    FCmdLine: string;
    FProg: string;
  public
    function Compile(Project: TVideProject; Build: Boolean = False): Boolean;
      override;
  end;

  TDelphi = class(TCommonCompiler)
  public
    constructor Create; override;
    function Compile(Project: TVideProject; Build: Boolean = False): Boolean;
      override;
    procedure ApplyDefaults(Settings: TVideCompilerSettings); override;
  end;

  TFreePascal = class(TCommonCompiler)
  public
    constructor Create; override;
    function Compile(Project: TVideProject; Build: Boolean = False): Boolean;
      override;
    procedure ApplyDefaults(Settings: TVideCompilerSettings); override;
  end;

  TCompilerList = class(TNObjectList)
  protected
    function GetNewItem(Index: Long): TVideCompiler;
    procedure SetNewItem(Index: Long; const Value: TVideCompiler);
    procedure Notify(Index: Long; Action: TListNotification); override;
  public
    function Find(const Name: string): TVideCompiler;
    property Items[Index: Long]: TVideCompiler read GetNewItem write SetNewItem;
    default;
  end;

function Compilers: TCompilerList;
function RunAndGetOutput(Cmd, WorkDir: string;
  ShowReturnValue: Boolean): string;
function RunFile(FileToRun: string; Params: string; Dir: string; Wait: boolean):
  DWord;
procedure LineOutput(const Line: string);  

implementation

uses
  AOptions, AMainActions, FMain;

var
  CLInstance: TCompilerList;

procedure AddDefaultCompilers; forward;

function Compilers: TCompilerList;
begin
  if CLInstance = nil then
  begin
    CLInstance := TCompilerList.Create;
    AddDefaultCompilers;
  end;
  Result := CLInstance;
end;

procedure LineOutput(const Line: string);
begin
  MainForm.ListMsg.AddItem(StrReplace(Line, #13, ''), nil);
  MainForm.ListMsg.ItemIndex := MainForm.ListMsg.Items.Count - 1;
  MainForm.ListMsg.Scroll(0, 10);
end;

function RunFile(FileToRun: string; Params: string; Dir: string; Wait: boolean):
  DWord;
const
  BufSize = $4000;
var
  StartupInfo: TStartupInfo;
  ProcessInfo: TProcessInformation;
begin
  Result := 35; // > 33 for no error
  FillChar(StartupInfo, SizeOf(StartupInfo), 0);

  with StartupInfo do
  begin
    cb := SizeOf(StartupInfo);
    dwFlags := STARTF_USESHOWWINDOW;
    wShowWindow := SW_SHOW;
  end;

  if not CreateProcess(nil, PChar(FileToRun + ' ' + Params), nil, nil,
    false, CREATE_NEW_CONSOLE or NORMAL_PRIORITY_CLASS,
    nil, pChar(Dir), StartupInfo, ProcessInfo) then
  begin
    Result := GetLastError;
    MainActions.ErrorDlg('Error when executing file: ' + FileToRun,
      'Run Error');
    Exit;
  end
  else
  begin
    if Wait then
    begin
      repeat
        Application.ProcessMessages;
      until WaitForSingleObject(ProcessInfo.hProcess, 1) = WAIT_OBJECT_0;
    end;

    CloseHandle(ProcessInfo.hProcess);
    CloseHandle(ProcessInfo.hThread);
  end;
end;


function RunAndGetOutput(Cmd, WorkDir: string;
  ShowReturnValue: Boolean): string;
var
  tsi: TStartupInfo;
  tpi: TProcessInformation;
  nRead: DWORD;
  aBuf: array[0..101] of char;
  sa: TSecurityAttributes;
  hOutputReadTmp, hOutputRead, hOutputWrite, hInputWriteTmp, hInputRead,
    hInputWrite, hErrorWrite: THandle;
  FOutput: string;
  CurrentLine: string;
begin
  FOutput := '';
  CurrentLine := '';
  sa.nLength := SizeOf(TSecurityAttributes);
  sa.lpSecurityDescriptor := nil;
  sa.bInheritHandle := True;

  CreatePipe(hOutputReadTmp, hOutputWrite, @sa, 0);
  DuplicateHandle(GetCurrentProcess(), hOutputWrite, GetCurrentProcess(),
    @hErrorWrite, 0, true, DUPLICATE_SAME_ACCESS);
  CreatePipe(hInputRead, hInputWriteTmp, @sa, 0);

  DuplicateHandle(GetCurrentProcess(), hOutputReadTmp, GetCurrentProcess(),
    @hOutputRead, 0, false, DUPLICATE_SAME_ACCESS);
  DuplicateHandle(GetCurrentProcess(), hInputWriteTmp, GetCurrentProcess(),
    @hInputWrite, 0, false, DUPLICATE_SAME_ACCESS);
  CloseHandle(hOutputReadTmp);
  CloseHandle(hInputWriteTmp);

  FillChar(tsi, SizeOf(TStartupInfo), 0);
  tsi.cb := SizeOf(TStartupInfo);
  tsi.dwFlags := STARTF_USESTDHANDLES or STARTF_USESHOWWINDOW;
  tsi.hStdInput := hInputRead;
  tsi.hStdOutput := hOutputWrite;
  tsi.hStdError := hErrorWrite;

  if not CreateProcess(nil, PChar(Cmd), @sa, @sa, true, 0, nil, PChar(WorkDir),
    tsi, tpi) then
  begin
    result := InttoStr(GetLastError) + ': unable to run program file.';
    exit;
  end;
  CloseHandle(hOutputWrite);
  CloseHandle(hInputRead);
  CloseHandle(hErrorWrite);

  repeat
    if (not ReadFile(hOutputRead, aBuf, 1, nRead, nil)) or (nRead = 0) then
    begin
      if GetLastError = ERROR_BROKEN_PIPE then
        Break;
    end;
    aBuf[nRead] := #0;
    FOutput := FOutput + PChar(@aBuf[0]);
    begin
      CurrentLine := CurrentLine + PChar(@aBuf[0]);
      if CurrentLine[Length(CurrentLine)] = #10 then
      begin
        Delete(CurrentLine, Length(CurrentLine) - 1, 2);
        LineOutput(CurrentLine);
        CurrentLine := '';
      end;
    end;
    Application.ProcessMessages;
  until False;
  GetExitCodeProcess(tpi.hProcess, nRead);
  if ShowReturnValue then
    Result := FOutput + ' ' + IntToStr(nread)
  else
    Result := FOutput;

  CloseHandle(hOutputRead);
  CloseHandle(hInputWrite);
  CloseHandle(tpi.hProcess);
  CloseHandle(tpi.hThread);
end;


{ TDelphi }

procedure TDelphi.ApplyDefaults(Settings: TVideCompilerSettings);
begin
  Settings.MinEnumSize := 1;
  Settings.RecordAlign := 8;
  Settings.BooleanEval := False;
  Settings.TypedPointers := False;
  Settings.ExtSyntax := True;
  Settings.AnsiStrings := True;
  Settings.CppExtensions := False;
  Settings.DebugInfo := False;
  Settings.Assertions := True;
  Settings.Warnings := True;
  Settings.Hints := True;
  Settings.RangeCheck := False;
  Settings.IOCheck := True;
  Settings.OverflowCheck := False;
  Settings.Mode := cmDelphi;
end;

function TDelphi.Compile(Project: TVideProject; Build: Boolean): Boolean;
var
  S: string;
  I: Long;
begin
  Result := False;
  if not (inherited Compile(Project)) then
    Exit;

  if not Vfs.FileExistsOutsideVfs(FPath + FProg) then
  begin
    MainActions.ErrorDlg('Delphi Compiler not found: ' +
      FPath + FProg, 'Compilation Error');
    Exit;
  end;

  FCmdLine := ' -$O+ -$J+ -$P+';
  FCmdLine := FCmdLine + ' "' + Project.ActiveFile.FileName + '" ';

  if Build then
    FCmdLine := FCmdLine + ' -B';

  with Project do
  begin
    FCmdLine := FCmdLine + ' -$Z' + IntToStr(CompilerSettings.MinEnumSize);
    FCmdLine := FCmdLine + ' -$A' + IntToStr(CompilerSettings.RecordAlign);
    FCmdLine := FCmdLine + ' -$B' + Iff(CompilerSettings.BooleanEval, '+', '-');
    FCmdLine := FCmdLine + ' -$T' + Iff(CompilerSettings.TypedPointers, '+',
      '-');
    FCmdLine := FCmdLine + ' -$X' + Iff(CompilerSettings.ExtSyntax, '+', '-');
    FCmdLine := FCmdLine + ' -$H' + Iff(CompilerSettings.AnsiStrings, '+', '-');
    FCmdLine := FCmdLine + ' -$D' + Iff(CompilerSettings.DebugInfo, '+', '-');
    FCmdLine := FCmdLine + ' -$C' + Iff(CompilerSettings.Assertions, '+', '-');

    FCmdLine := FCmdLine + ' -W' + Iff(CompilerSettings.Warnings, '+', '-');
    FCmdLine := FCmdLine + ' -H' + Iff(CompilerSettings.Hints, '+', '-');

    FCmdLine := FCmdLine + ' -$R' + Iff(CompilerSettings.RangeCheck, '+', '-');
    FCmdLine := FCmdLine + ' -$Q' + Iff(CompilerSettings.OverflowCheck, '+',
      '-');
    FCmdLine := FCmdLine + ' -$I' + Iff(CompilerSettings.IOCheck, '+', '-');
  end;

  for I := 0 to Options.Paths.Count - 1 do
  begin
    FCmdLine := FCmdLine + ' -I"' + SysUtils.ExcludeTrailingPathDelimiter(Vfs.GetFullDirName(Options.Paths[I])) + '"';
    FCmdLine := FCmdLine + ' -U"' + SysUtils.ExcludeTrailingPathDelimiter(Vfs.GetFullDirName(Options.Paths[I])) + '"';
    FCmdLine := FCmdLine + ' -R"' + SysUtils.ExcludeTrailingPathDelimiter(Vfs.GetFullDirName(Options.Paths[I])) + '"';
    FCmdLine := FCmdLine + ' -Q"' + SysUtils.ExcludeTrailingPathDelimiter(Vfs.GetFullDirName(Options.Paths[I])) + '"';
  end;

  for I := 0 to Project.Paths.Count - 1 do
  begin
    FCmdLine := FCmdLine + ' -I"' + SysUtils.ExcludeTrailingPathDelimiter(Vfs.GetFullDirName(Options.Paths[I])) + '"';
    FCmdLine := FCmdLine + ' -U"' + SysUtils.ExcludeTrailingPathDelimiter(Vfs.GetFullDirName(Options.Paths[I])) + '"';
    FCmdLine := FCmdLine + ' -R"' + SysUtils.ExcludeTrailingPathDelimiter(Vfs.GetFullDirName(Options.Paths[I])) + '"';
    FCmdLine := FCmdLine + ' -Q"' + SysUtils.ExcludeTrailingPathDelimiter(Vfs.GetFullDirName(Options.Paths[I])) + '"';
  end;

  S := FPath + PathDelim + 'lib';
  FCmdLine := FCmdLine + ' -I"' + S + '"';
  FCmdLine := FCmdLine + ' -U"' + S + '"';
  FCmdLine := FCmdLine + ' -R"' + S + '"';
  FCmdLine := FCmdLine + ' -Q"' + S + '"';

  case Project.ProjectType of
    ptGuiApp: FCmdLine := FCmdLine + ' -CG';
    ptConsoleApp: FCmdLine := FCmdLine + ' -CC';
  end;

  if Project.OutputDir <> '' then
  FCmdLine := FCmdLine + ' -E"' + SysUtils.ExcludeTrailingPathDelimiter(Project.OutputDir) + '"';

  LineOutput('dcc32 ' + FCmdLine);

  S := RunAndGetOutput('"' + Vfs.GetFullFileName(FPath + PathDelim + FProg) + '"'
    + FCmdLine, ExtractFileDir(Project.FileName), True);

  Result := True;

  if (PosCase('Error:', S) = 0) and (PosCase('Fatal:', S) = 0)and
   (PosCase('Target address not found', S) = 0) then
    LineOutput('Finished Successfuly.')
  else
  begin
    LineOutput('Finished With Errors.');
    Result := False;
  end;
end;

constructor TDelphi.Create;
begin
  inherited Create;
  FName := 'Delphi';
  FPath := 'C:\Program Files\Borland\Delphi7';
  FProg := '\Bin\dcc32.exe';
end;

{ TFreePascal }

procedure TFreePascal.ApplyDefaults(Settings: TVideCompilerSettings);
begin
  Settings.MinEnumSize := 4;
  Settings.RecordAlign := 4;
  Settings.BooleanEval := False;
  Settings.TypedPointers := False;
  Settings.ExtSyntax := True;
  Settings.AnsiStrings := True;
  Settings.CppExtensions := False;
  Settings.DebugInfo := False;
  Settings.Assertions := True;
  Settings.Warnings := True;
  Settings.Hints := True;
  Settings.RangeCheck := False;
  Settings.IOCheck := True;
  Settings.OverflowCheck := False;
  Settings.TargetCPU := tcP5;
  Settings.Mode := cmObjFPC;
end;

function TFreePascal.Compile(Project: TVideProject; Build: Boolean): Boolean;
var
  I: Long;
  S: string;
begin
  Result := False;
  if not (inherited Compile(Project)) then
    Exit;

  if not Vfs.FileExistsOutsideVfs(FPath + FProg) then
  begin
    MainActions.ErrorDlg('Free Pascal Compiler not found: ' +
      FPath + FProg, 'Compilation Error');
    Exit;
  end;

  FCmdLine := ' -Twin32 -Xs -XX -Rintel -OG2 -velix -n';
  FCmdLine := FCmdLine + ' "' + Project.ActiveFile.FileName + '" ';

  if Build then
    FCmdLine := FCmdLine + ' -B';

  with Project do
  begin
   //CompilerSettings.MinEnumSize
   //CompilerSettings.RecordAlign
    case CompilerSettings.TargetCPU of
      tc386: FCmdLine := FCmdLine + ' -Op1';
      tcP5: FCmdLine := FCmdLine + ' -Op2';
      tcP6: FCmdLine := FCmdLine + ' -Op3';
    end;
    case CompilerSettings.Mode of
      cmDefault: FCmdLine := FCmdLine + ' -Mfpc';
      cmDelphi: FCmdLine := FCmdLine + ' -Mdelphi';
      cmObjFPC: FCmdLine := FCmdLine + ' -Mobjfpc';
      cmTP: FCmdLine := FCmdLine + ' -Mtp';
      cmGPC: FCmdLine := FCmdLine + ' -Mgpc';
    end;
   //CompilerSettings.BooleanEval
   //CompilerSettings.TypedPointers
   //CompilerSettings.ExtSyntax
    if CompilerSettings.AnsiStrings then
      FCmdLine := FCmdLine + ' -Sh';
    if CompilerSettings.CppExtensions then
      FCmdLine := FCmdLine + ' -Sm -Si -Sc';
    if CompilerSettings.DebugInfo then
      FCmdLine := FCmdLine + ' -gg -gh -gc -gl';
    if CompilerSettings.Assertions then
      FCmdLine := FCmdLine + ' -Sa';
    if CompilerSettings.Warnings then
      FCmdLine := FCmdLine + ' -vw';
    if CompilerSettings.Hints then
      FCmdLine := FCmdLine + ' -vh';
    if CompilerSettings.RangeCheck then
      FCmdLine := FCmdLine + ' -Cr';
    if CompilerSettings.IOCheck then
      FCmdLine := FCmdLine + ' -Ci';
    if CompilerSettings.OverflowCheck then
      FCmdLine := FCmdLine + ' -Co';
  end;

  case Project.ProjectType of
    ptGuiApp: FCmdLine := FCmdLine + ' -WG';
    ptConsoleApp: FCmdLine := FCmdLine + ' -WC';
  end;

  for I := 0 to Options.Paths.Count - 1 do
  begin
    FCmdLine := FCmdLine + ' -Fi"' + SysUtils.ExcludeTrailingPathDelimiter(Vfs.GetFullDirName(Options.Paths[I])) + '"';
    FCmdLine := FCmdLine + ' -Fl"' + SysUtils.ExcludeTrailingPathDelimiter(Vfs.GetFullDirName(Options.Paths[I])) + '"';
    FCmdLine := FCmdLine + ' -Fo"' + SysUtils.ExcludeTrailingPathDelimiter(Vfs.GetFullDirName(Options.Paths[I])) + '"';
    FCmdLine := FCmdLine + ' -Fu"' + SysUtils.ExcludeTrailingPathDelimiter(Vfs.GetFullDirName(Options.Paths[I])) + '"';
  end;

  for I := 0 to Project.Paths.Count - 1 do
  begin
    FCmdLine := FCmdLine + ' -Fi"' + SysUtils.ExcludeTrailingPathDelimiter(Vfs.GetFullDirName(Project.Paths[I])) + '"';
    FCmdLine := FCmdLine + ' -Fl"' + SysUtils.ExcludeTrailingPathDelimiter(Vfs.GetFullDirName(Project.Paths[I])) + '"';
    FCmdLine := FCmdLine + ' -Fo"' + SysUtils.ExcludeTrailingPathDelimiter(Vfs.GetFullDirName(Project.Paths[I])) + '"';
    FCmdLine := FCmdLine + ' -Fu"' + SysUtils.ExcludeTrailingPathDelimiter(Vfs.GetFullDirName(Project.Paths[I])) + '"';
  end;

  S := FPath + PathDelim + 'units\$TARGET\*';
  FCmdLine := FCmdLine + ' -Fi"' + S + '"';
  FCmdLine := FCmdLine + ' -Fl"' + S + '"';
  FCmdLine := FCmdLine + ' -Fo"' + S + '"';
  FCmdLine := FCmdLine + ' -Fu"' + S + '"';

  S := FPath + PathDelim + 'units\$TARGET\*\*';
  FCmdLine := FCmdLine + ' -Fi"' + S + '"';
  FCmdLine := FCmdLine + ' -Fl"' + S + '"';
  FCmdLine := FCmdLine + ' -Fo"' + S + '"';
  FCmdLine := FCmdLine + ' -Fu"' + S + '"';

  FCmdLine := FCmdLine + ' -FD"' + FPath + '\bin\$TARGET"';
  FCmdLine := FCmdLine + ' -FE"' + SysUtils.ExcludeTrailingPathDelimiter(Project.OutputDir) + '"';

  LineOutput('ppc386 ' + FCmdLine);

  S := RunAndGetOutput('"' + Vfs.GetFullFileName(FPath + PathDelim + FProg) + '"'
    + FCmdLine, ExtractFileDir(Project.FileName), True);

  Result := True;

  if (PosCase('Error:', S) = 0) and (PosCase('Fatal:', S) = 0) then
    LineOutput('Finished Successfuly.')
  else
  begin
    LineOutput('Finished With Errors.');
    Result := False;
  end;
end;

constructor TFreePascal.Create;
begin
  inherited Create;
  FName := 'Free Pascal';
  FPath := 'C:\Program Files\FPC';
  FProg := '\bin\win32\ppc386.exe';
end;

{ TCompilerList }

function TCompilerList.Find(const Name: string): TVideCompiler;
var
  I: Long;
begin
  Result := nil;
  for I := 0 to Count - 1 do
    if SameStr(Name, Items[I].Name) then
    begin
      Result := Items[I];
      Exit;
    end;
end;

function TCompilerList.GetNewItem(Index: Long): TVideCompiler;
begin
  if (Index > -1) and (Index < FCount) then
    Result := TVideCompiler(PPointerArray(FData)[Index])
  else
    Result := nil;
end;

procedure TCompilerList.Notify(Index: Long; Action: TListNotification);
begin
  if Action = lnDeleted then
    TObject(PPointerArray(FData)[Index]).Free;
end;

procedure TCompilerList.SetNewItem(Index: Long;
  const Value: TVideCompiler);
begin
  if (Index > -1) and (Index < FCount) then
  begin
    Notify(Index, lnDeleted);
    PPointerArray(FData)[Index] := Value;
    Notify(Index, lnAdded);
  end;
end;

procedure AddDefaultCompilers;
begin
  Compilers.Add(TDelphi.Create);
  Compilers.Add(TFreePascal.Create);
end;

{ TCommonCompiler }

function TCommonCompiler.Compile(Project: TVideProject; Build: Boolean = False):
  Boolean;
begin
  Result := False;
  if Options.AutoSave and (not MainActions.SaveProject(Project)) then
    Exit;

  if Project.ActiveFile = nil  then
  begin
    MainActions.ErrorDlg(Format('Project ''%s'' has no active file which can be send to compiler.',
    [Project.Name]), 'Compilation Error');
    Exit;
  end;

  LineOutput('');
  LineOutput(Format('%s - compilig project using %s compiler.', [Project.Name,
  FName]));

  Result := True;
end;

initialization

  RegisterClasses([TDelphi, TFreePascal]);


end.

