unit AEditorActions;

interface

uses
  Windows, SysUtils, Kernel, CSingletons, CBase, CEditor, Objects,
  Forms, SynEdit, ComCtrls, ABase, FData, SynEditTypes, Lists, AOptions;

type

  TEditorActions = class(TObject)
  protected
    FSearch: TSearchOptions;
    FFind: TSearchOptions;
    FReplace: TSearchOptions;

    function GetEditor: TSynEdit;
    function GetTabs: TTabControl;
    function GetFrame: TFrame;

    procedure ActivateFile(const FileName: string);
    procedure SetSyntax(Syntax: TTextType);
    function CanFocusEditor: Boolean;
    procedure UnselectBlock(MoveToEnd:Boolean);
    function GetDefSearchWord:string;
  public
    constructor Create;
    procedure Idle;

    function CreateEdFile(const AFileName:string; FileType:TVideFileType;
    ADirty:Boolean):TVideEditorFile;

    procedure OpenFile(const FileName: string; FileType: TVideFileType);
    procedure AddFile(AFile: TVideEditorFile);
    procedure CloseIndex(Index: Long);
    procedure CloseActiveFile;

    procedure ActivateIndex(NewIndex: Long; SaveOldPos: Boolean = True);
    procedure UpdateIndex(Index: Long);
    procedure UpdateActive;

    procedure Exchange(Index: Long);

    procedure Search(Mode: TSearchMode);
    procedure GoToLine(Line:Long);

    property EditorVisible: Boolean read CanFocusEditor;
  end;

function EditorActions: TEditorActions;

implementation

uses
  FMain, FEditor, AMainActions, FSaveChanges, FSearch, FReplace;

var
  EditorActionsInstance: TEditorActions = nil;

function EditorActions: TEditorActions;
begin
  if EditorActionsInstance = nil then
    EditorActionsInstance := TEditorActions.Create;
  Result := EditorActionsInstance;
end;

{ TEditorActions }

procedure TEditorActions.ActivateFile(const FileName: string);
begin
  ActivateIndex(Editor.IndexOfFileName(FileName));
end;

procedure TEditorActions.ActivateIndex(NewIndex: Long; SaveOldPos: Boolean);
var
  Hook:TEditorHook;
begin
  if (NewIndex > -1) and (NewIndex<Editor.Count) then
  begin
    if (Editor.ActiveIndex > -1) and SaveOldPos then
      UpdateActive;
    Editor.ActiveIndex := NewIndex;
    GetTabs.TabIndex := NewIndex;
    with GetEditor, Editor do
    begin
      Modified := ActiveFile.Modified;
      Hook:=TEditorHook(Editor.ActiveFile.Tag);
      GetEditor.UnHookTextBuffer;
      GetEditor.HookTextBuffer(Hook.Buffer, Hook.UndoList, Hook.RedoList);
      CaretX := ActiveFile.CaretX;
      CaretY := ActiveFile.CaretY;
      TopLine := ActiveFile.TopLine;
      UpdateIndex(Editor.ActiveIndex);
      if ActiveFile.FileType <> nil then
        SetSyntax(ActiveFile.FileType.TextType)
      else
        SetSyntax(ttNone);
      if CanFocusEditor then
        SetFocus;
    end;
  end;
end;

procedure TEditorActions.AddFile(AFile: TVideEditorFile);
begin
  if Editor.FilesByName[AFile.FileName] = nil then
  begin
    TEditorHook(AFile.Tag).Buffer.Text:=AFile.Buffer.Text;
    Editor.AddFile(AFile);
    GetTabs.Tabs.Add(AFile.Name);
    if Editor.Count > 0 then
      GetFrame.Visible := True;
  end;
  ActivateFile(AFile.FileName);
end;

function TEditorActions.CanFocusEditor: Boolean;
begin
  with MainForm do
    Result := EditorFrame.Visible and PanelEditor.Visible;
end;

procedure TEditorActions.CloseActiveFile;
begin
  UpdateActive;
  CloseIndex(Editor.ActiveIndex);
end;

procedure TEditorActions.CloseIndex(Index: Long);
begin
  if (Index > -1) and (Index < Editor.Count) then
  begin
    if (Editor.Files[Index].Dirty) and
      (Editor.Files[Index].Project <> nil) then
      Editor.Files[Index].Project.RemoveFile(Editor.Files[Index].FileName);
    GetTabs.Tabs.Delete(Index);
    GetEditor.UnHookTextBuffer;
    Editor.Delete(Index);
    if Index = Editor.Count then
      ActivateIndex(Editor.Count - 1, False)
    else
      ActivateIndex(Index, False);
    if Editor.Count = 0 then
      GetFrame.Visible := False
  end;
end;

constructor TEditorActions.Create;
begin
  Log.Add('EditorActions created.');
end;

function TEditorActions.CreateEdFile(const AFileName: string;
  FileType: TVideFileType;ADirty:Boolean): TVideEditorFile;
var
  Hook:TEditorHook;
begin
  Result :=TVideEditorFile.Create(AFileName, FileType, ADirty);
  Hook:=TEditorHook.Create;
  Hook.Buffer.Text := Result.Buffer.Text;
  Result.Tag:=Hook;
end;

procedure TEditorActions.Exchange(Index: Long);
var
  S: string;
begin
  if (Index > -1) and (Index < Editor.Count) then
  begin
    Editor.Exchange(Editor.ActiveIndex, Index);
    S := GetTabs.Tabs[Editor.ActiveIndex];
    GetTabs.Tabs[Editor.ActiveIndex] := GetTabs.Tabs[Index];
    GetTabs.Tabs[Index] := S;
    Editor.ActiveIndex := Index;
    GetTabs.TabIndex := Index;
  end;
end;

function TEditorActions.GetDefSearchWord: string;
begin
  with GetEditor do
  if (SelText = '') then
    Result:= GetWordAtRowCol(CaretXY)
   else
    Result:= SelText;
end;

function TEditorActions.GetEditor: TSynEdit;
begin
  Result := MainForm.EditorFrame.SynEdit;
end;

function TEditorActions.GetFrame: TFrame;
begin
  Result := FMain.MainForm.EditorFrame;
end;

function TEditorActions.GetTabs: TTabControl;
begin
  Result := MainForm.EditorFrame.TabEditor;
end;

procedure TEditorActions.GoToLine(Line: Long);
begin
  with GetEditor do
  begin
    CaretY:=Line;
    TopLine:=Line - LinesInWindow div 2;
  end;
end;

procedure TEditorActions.Idle;
var
  I: Long;
  S: string;
  EdFile: TVideEditorFile;
begin
  S := Format('%s%d: %d / %d', [StringOfChar(#32,
      Length(IntToStr(GetEditor.CaretY)) - Length(IntToStr(GetEditor.CaretX))),
    GetEditor.CaretX, GetEditor.CaretY, GetEditor.Lines.Count]);
  if MainForm.EditorFrame.Status.Panels[0].Text <> S then
    MainForm.EditorFrame.Status.Panels[0].Text := S;

  S := Iff(GetEditor.Modified, 'Modified', '');
  if Editor.ActiveFile <> nil then
    if Editor.ActiveFile.Dirty and (S = '') then
      S := 'Modified';
  if MainForm.EditorFrame.Status.Panels[1].Text <> S then
    MainForm.EditorFrame.Status.Panels[1].Text := S;

  S := Iff(GetEditor.InsertMode, 'Insert', 'Overwrite');
  if MainForm.EditorFrame.Status.Panels[2].Text <> S then
    MainForm.EditorFrame.Status.Panels[2].Text := S;

  if Editor.ActiveFile <> nil then
    S := Editor.ActiveFile.FileName;
  if MainForm.EditorFrame.Status.Panels[3].Text <> S then
    MainForm.EditorFrame.Status.Panels[3].Text := S;

  if Editor.ActiveFile <> nil then
    Editor.ActiveFile.Modified := GetEditor.Modified;

  for I := 0 to Editor.Count - 1 do
  begin
    EdFile := Editor.Files[I];
    S := EdFile.Name;
    if Editor.Files[I].Modified or Editor.Files[I].Dirty then
      S := S + '*';
    if GetTabs.Tabs[I] <> S then
      GetTabs.Tabs[I] := S;
  end;
end;

procedure TEditorActions.OpenFile(const FileName: string;
  FileType: TVideFileType);
var
  EdFile: TVideEditorFile;
begin
  EdFile := Editor.FilesByName[FileName];
  if EdFile = nil then
  begin
    EdFile := CreateEdFile(FileName, FileType, False);
    Editor.AddFile(EdFile);
    GetTabs.Tabs.Add(EdFile.Name);
    if Editor.Count > 0 then
      GetFrame.Visible := True;
  end;
  ActivateFile(EdFile.FileName);
end;

procedure TEditorActions.Search(Mode: TSearchMode);

  procedure NotFound;
  begin
    UnselectBlock(True);
    MainActions.InfoDlg(Format('Search string ''%s'' not found.',
      [FSearch.Text]), 'Find');
  end;

begin
  case Mode of
    smFindNew:
      begin
      if Options.Editor.FindAtCursor then
        FFind.Text:=GetDefSearchWord;
      if SearchForm.Find(FFind) then
      begin
        FSearch := SearchForm.GetOptions;
        FFind:=FSearch;
        if GetEditor.SearchReplace(FSearch.Text, '', FSearch.SynOptions) = 0
          then
          NotFound;
      end;
      end;
    smReplaceNew:
    begin
      FReplace.Text:=GetDefSearchWord;
      if ReplaceForm.Replace(FReplace) then
      begin
        FSearch := ReplaceForm.GetOptions;
        FReplace:=FSearch;
        if GetEditor.SearchReplace(FSearch.Text, FSearch.ReplaceText,
          FSearch.SynOptions) = 0 then
          NotFound;
        UnselectBlock(True);
      end;
    end;
    smNext:
      if FSearch.Text <> '' then
      begin
        if GetEditor.SearchReplace(FSearch.Text, FSearch.ReplaceText,
          FSearch.SynOptions) = 0 then
          NotFound;
        if ssoReplace in FSearch.SynOptions then
          UnselectBlock(True);
      end
      else
        Search(smFindNew);
  end;
end;

procedure TEditorActions.SetSyntax(Syntax: TTextType);
begin
  with GetEditor, Data do
    case Syntax of
      ttNone: Highlighter := nil;
      ttPascal: Highlighter := PasSyn;
      ttCpp: Highlighter := CppSyn;
      ttAsm: Highlighter := AsmSyn;
      ttJava: Highlighter := JavaSyn;
      ttCSharp: Highlighter := CSharpSyn;
      ttBasic: Highlighter := BasicSyn;
      ttDelphiForm: Highlighter := DfmSyn;
      ttXML: Highlighter := XMLSyn;
      ttHTML: Highlighter := HTMLSyn;
      ttINI: Highlighter := INISyn;
    end;
end;

procedure TEditorActions.UnselectBlock(MoveToEnd: Boolean);
begin
  if MoveToEnd then
  GetEditor.SelStart := GetEditor.SelEnd
  else
  GetEditor.SelLength:=0;
end;

procedure TEditorActions.UpdateActive;
begin
  if Editor.ActiveFile <> nil then
    with Editor.ActiveFile do
    begin
      CaretX := GetEditor.CaretX;
      CaretY := GetEditor.CaretY;
      TopLine := GetEditor.TopLine;
      Buffer.Text := GetEditor.Lines.Text;
      Modified := GetEditor.Modified;
    end;
end;

procedure TEditorActions.UpdateIndex(Index: Long);
begin
  if (Index > -1) and (Index < Editor.Count) then
    with Editor.Files[Index] do
    begin
      GetEditor.Modified := Modified;
    end;
end;

initialization
finalization
  FreeAndNil(EditorActionsInstance);

end.

