unit AMainActions;

interface

uses
  Windows, SysUtils, Classes, Kernel, Files, CSingletons, CBase, ComCtrls,
  CEditor, CProjects, Dialogs, Objects, ABase, ANewItems, Platforms,
  Lists, FData, Configs, AOptions, ACompilers, Forms;

type

  TMainActions = class(TObject)
  protected
    FProject: TVideProject;
    FCanAsk: Boolean;

    function GetSave: TSaveDialog;
    function GetOpen: TOpenDialog;

    procedure OnIndexChange(NewIndex: Long);
    procedure OnSolutionLoading(Cfg: TNConfig);
    procedure OnSolutionSaving(Cfg: TNConfig);
    function OnGetCompiler(const Name: NString): TVideCompiler;
  public
    constructor Create;
    procedure Initialize;
    procedure FirstShow;
    procedure Finalize;
    procedure Idle;

    procedure OpenFile(const FileName: string);
    procedure OpenFiles(AskForFiles: Boolean; Files: TStrings = nil);
    procedure OpenSolution(AskForName: Boolean; const AFileName: string = '');

    procedure CloseFile;
    function CloseAll: Boolean;
    function CloseApp: Boolean;

    procedure AddFile(Project: TVideProject; const FileName: string);
    procedure AddFiles(Project: TVideProject; AskForFiles: Boolean; Files:
      TStrings = nil);
    function GetFileFilter(AllFiles: Boolean; FixedPos: TFixedTypePos;
      LastIndex: PLong = nil): string;
    procedure AddProject(const FileName: string);
    procedure AddProjects(AskForFiles: Boolean; Files: TStrings = nil);

    procedure ActivateProject(Index: Long);
    procedure ActivateFile(AFile: TVideFile);

    procedure InitSaveFileDialog(Dialog: TSaveDialog; const Name: string);
    procedure InitSaveProjectDialog(Dialog: TSaveDialog; const Name: string);
    procedure InitSaveSolutionDialog(Dialog: TSaveDialog; const Name: string);
    procedure InitAddFileDialog(Dialog: TOpenDialog);
    procedure InitAddProjectDialog(Dialog: TOpenDialog);
    procedure InitOpenFileDialog(Dialog: TOpenDialog);
    procedure InitOpenSolutionDialog(Dialog: TOpenDialog);

    function SaveFile(EdFile: TVideEditorFile; AskForName: Boolean = False):
      Boolean;
    function SaveProject(AProject: TVideProject; AskForName: Boolean = False):
      Boolean;
    function SaveSolution(AskForName: Boolean = False): Boolean;
    procedure SaveAll;
    procedure SaveList(List: TNObjectList);
    procedure SaveItems;

    procedure NewItem(AllowedItems: TNewItems; Project: TVideProject = nil);
      overload;
    procedure NewItem(const Name: string; Project: TVideProject = nil);
      overload;
    procedure FillNewItems(NewType: TNewItemCategory; View: TListView);

    function CheckName(const Name: string; AType: TNewItemCategory;
      Project: TVideProject = nil; const OldName: string = ''): Boolean;
    function CheckIntName(const Name: string; AType: TNewItemCategory; Project:
      TVideProject = nil): string;

    procedure SolViewSource;
    function SolRename(const NewName: string): Boolean;
    procedure SolRemove(AProject: TVideProject; Ask: Boolean);

    procedure ProjectViewSource(Project: TVideProject);
    function ProjectRename(Project: TVideProject; const NewName: string):
      Boolean;
    procedure ProjectRemove(AFile: TVideFile; Ask: Boolean);
    function ProjectCompile(Project: TVideProject; Echo: Boolean; Build:
      Boolean;
      DeleteMsg: Boolean): Boolean;
    procedure ProjectRun(Project: TVideProject);
    procedure ProjectBuild(Project: TVideProject);

    function FileRename(AFile: TVideFile; const NewName: string): Boolean;

    procedure ErrorDlg(const Msg, Caption: string);
    procedure InfoDlg(const Msg, Caption: string);
    function AskDlg(const Msg, Caption: string): Boolean;

    property ProjectToAdd: TVideProject read FProject;
  end;

function MainActions: TMainActions;

implementation

uses
  FMain, FFileNew, AEditorActions, AExpActions, FSaveChanges;

var
  MainActionsInstance: TMainActions = nil;

function MainActions: TMainActions;
begin
  if MainActionsInstance = nil then
    MainActionsInstance := TMainActions.Create;
  Result := MainActionsInstance;
end;

{ TMainActions }

procedure TMainActions.ActivateProject(Index: Long);
begin
  MainForm.TreeExp.Invalidate;
end;

procedure TMainActions.AddFile(Project: TVideProject; const FileName: string);
var
  F: TVideFile;
begin
  F := Project.AddFile(FileName);
  if F <> nil then
    OpenFile(FileName);
end;

procedure TMainActions.AddFiles(Project: TVideProject; AskForFiles: Boolean;
  Files: TStrings);
var
  I: Long;
begin
  if AskForFiles then
  begin
    InitAddFileDialog(GetOpen);
    if GetOpen.Execute then
      Files := GetOpen.Files;
  end;
  if Files <> nil then
    for I := 0 to Files.Count - 1 do
      AddFile(Project, Files[I]);
end;

procedure TMainActions.AddProject(const FileName: string);
begin
  Solution.AddProject(FileName);
end;

procedure TMainActions.AddProjects(AskForFiles: Boolean; Files: TStrings);
var
  I: Long;
begin
  if AskForFiles then
  begin
    InitAddProjectDialog(GetOpen);
    if GetOpen.Execute then
      Files := GetOpen.Files;
  end;
  if Files <> nil then
    for I := 0 to Files.Count - 1 do
      AddProject(Files[I]);
end;

function TMainActions.AskDlg(const Msg, Caption: string): Boolean;
begin
  Result := MessageBox(GetForegroundWindow, PChar(Msg), PChar(Caption),
    MB_YESNO or MB_ICONQUESTION) = IDYES;
end;

function TMainActions.CheckIntName(const Name: string;
  AType: TNewItemCategory; Project: TVideProject): string;
var
  I: Long;
  Ext: string;
  FName: string;
begin
  I := 1;
  Ext := GetFileExt(Name);
  FName := ChangeFileExt(Name, '');
  while not CheckName(FName + IntToStr(I) + '.' + Ext, AType, Project) do
    Inc(I);
  Result := FName + IntToStr(I) + '.' + Ext;
end;

function TMainActions.CheckName(const Name: string;
  AType: TNewItemCategory; Project: TVideProject; const OldName: string):
  Boolean;
var
  I: Long;
begin
  if Project = nil then
    Project := Solution.ActiveProject;
  Result := True;
  case AType of
    ntProjects:
      for I := 0 to Solution.Count - 1 do
        if SameStr(ExtractFileName(Solution.Projects[I].FileName), Name) and
          (Solution.Projects[I].FileName <> OldName) then
        begin
          Result := False;
          Break;
        end;
    ntFiles:
      if Project <> nil then
        for I := 0 to Project.Count - 1 do
          if SameStr(ExtractFileName(Project.Files[I].FileName), Name) and
            (Project.Files[I].FileName <> OldName) then
          begin
            Result := False;
            Break;
          end;
  end;
end;

function TMainActions.CloseAll: Boolean;
begin
  Result := True;
  if FCanAsk then
    Result := SaveForm.SaveModifiedItems(True, AllSaveItems);
  if Result then
  begin
    MainActions.SaveList(SaveForm.Items);
    while Editor.Count > 0 do
      EditorActions.CloseActiveFile;
    Solution.NewSolution;
  end;
end;

procedure TMainActions.CloseFile;
begin
  if SaveForm.SaveModifiedItems(True, [scActiveFile]) then
  begin
    MainActions.SaveList(SaveForm.Items);
    EditorActions.CloseActiveFile;
  end;
end;

constructor TMainActions.Create;
begin
  Log.Add('MainActions created.');
end;

procedure TMainActions.ErrorDlg(const Msg, Caption: string);
begin
  MessageBox(GetForegroundWindow, PChar(Msg), PChar(Caption),
    MB_OK or MB_ICONERROR);
end;

function TMainActions.FileRename(AFile: TVideFile;
  const NewName: string): Boolean;
var
  FullName: string;
  EdFile: TVideEditorFile;
begin
  Result := False;
  if not CheckName(NewName, ntFiles, AFile.Project, AFile.FileName) then
    ErrorDlg(Format('File %s is already contained in project %s',
      [NewName, AFile.Project.Name]), 'Duplicate File Name')
  else
  begin
    EdFile := Editor.FilesByName[AFile.FileName];
    Result := True;
    if AFile.Dirty then
    begin
      AFile.Rename(NewName);
      if EdFile <> nil then
        EdFile.FileName := NewName;
    end
    else
    begin
      FullName := ExtractFilePath(AFile.FileName) + NewName;
      if EdFile <> nil then
        SaveFile(EdFile);

      if Vfs.FileExists(FullName) then
      begin
        if AskDlg(Format('File named %s already exists. Overwrite it?',
          [NewName]), 'Overwrite File') then
          AFile.Rename(FullName)
        else
          Result := False;
      end
      else
        AFile.Rename(FullName);

      if EdFile <> nil then
        EdFile.FileName := FullName;
    end;
  end;
end;

procedure TMainActions.FillNewItems(NewType: TNewItemCategory;
  View: TListView);
var
  I: Long;
  Item: TListItem;
begin
  View.Clear;
  for I := 0 to NewItems.Count - 1 do
    if NewItems[I].Category = NewType then
    begin
//      View.AddItem(NewItems[I].Name, NewItems[I]);
      Item := View.Items.Add;
      Item.Caption := NewItems[I].Name;
      Item.Data := NewItems[I];
      Item.ImageIndex := NewItems[I].Image;
    end;
  if View.Items.Count > 0 then
    View.Selected := View.Items[0];
end;

function TMainActions.GetFileFilter(AllFiles: Boolean; FixedPos: TFixedTypePos;
  LastIndex: PLong):
  string;
var
  I: Long;
  Cnt: Long;
begin
  Result := '';
  Cnt := 0;
  for I := 0 to FileTypes.Count - 1 do
    if FileTypes.Types[I].Fixed and (FixedPos = fpTop) then
    begin
      Result := Result + FileTypes.Types[I].Filter;
      Inc(Cnt);
    end;
  for I := 0 to FileTypes.Count - 1 do
    if not FileTypes.Types[I].Fixed then
    begin
      Result := Result + FileTypes.Types[I].Filter;
      Inc(Cnt);
    end;
  for I := 0 to FileTypes.Count - 1 do
    if FileTypes.Types[I].Fixed and (FixedPos = fpBottom) then
    begin
      Result := Result + FileTypes.Types[I].Filter;
      Inc(Cnt);
    end;
  if AllFiles then
  begin
    Result := Result + 'All files (*.*)|*.*';
    Inc(Cnt);
  end;
  if LastIndex <> nil then
    LastIndex^ := Cnt;
end;

function TMainActions.GetOpen: TOpenDialog;
begin
  Result := Data.OpenD;
end;

function TMainActions.GetSave: TSaveDialog;
begin
  Result := Data.SaveD;
end;

procedure TMainActions.Idle;
var
  S: string;
begin
  if Solution.ActiveProject <> nil then
  begin
    S := Solution.ActiveProject.Name + ' - ' + sAppName;
    if MainForm.Caption <> S then
      MainForm.Caption := S;
  end;
end;

procedure TMainActions.OnIndexChange(NewIndex: Long);
begin
  ActivateProject(NewIndex);
end;

procedure TMainActions.InfoDlg(const Msg, Caption: string);
begin
  MessageBox(GetForegroundWindow, PChar(Msg), PChar(Caption),
    MB_OK or MB_ICONINFORMATION);
end;

procedure TMainActions.InitAddFileDialog(Dialog: TOpenDialog);
begin
  Dialog.Filter := GetFileFilter(True, fpNone);
  Dialog.Options := Dialog.Options + [ofAllowMultiSelect];
end;

procedure TMainActions.InitAddProjectDialog(Dialog: TOpenDialog);
begin
  Dialog.Filter := FileTypes.TypesName[sVideProject].Filter;
  Dialog.Options := Dialog.Options + [ofAllowMultiSelect];
end;

procedure TMainActions.Initialize;
begin
  NewItems;
  Solution;
  Solution.OnIndexChange := OnIndexChange;
  Solution.OnLoading := OnSolutionLoading;
  Solution.OnSaving := OnSolutionSaving;
  Solution.OnGetCompiler := OnGetCompiler;
  FileTypes;
  Editor;
  EditorActions;
  ExpActions;
end;

procedure TMainActions.InitOpenFileDialog(Dialog: TOpenDialog);
begin
  Dialog.Filter := GetFileFilter(True, fpBottom);
  Dialog.Options := Dialog.Options + [ofAllowMultiSelect];
end;

procedure TMainActions.InitOpenSolutionDialog(Dialog: TOpenDialog);
begin
  Dialog.Filter := FileTypes.TypesName[sVideSolution].Filter;
  Dialog.Options := Dialog.Options - [ofAllowMultiSelect];
end;

procedure TMainActions.InitSaveFileDialog(Dialog: TSaveDialog; const Name:
  string);
var
  I: Long;
  Ext: string;
  FType: TVideFileType;
begin
  Dialog.FileName := Name;
  Ext := GetFileExt(Name);
  Dialog.DefaultExt := FileTypes.Types[0].Extensions[0];
  Dialog.Filter := GetFileFilter(True, fpNone, @I);
  Dialog.FilterIndex := I;
  FType := FileTypes.TypesExt[Ext];
  if FType <> nil then
    Dialog.FilterIndex := FileTypes.IndexOfType(FType) + 1;
end;

procedure TMainActions.InitSaveProjectDialog(Dialog: TSaveDialog; const Name:
  string);
begin
  Dialog.FileName := Name;
  Dialog.DefaultExt := sVideProjectExt;
  Dialog.Filter := FileTypes.TypesName[sVideProject].Filter;
end;

procedure TMainActions.InitSaveSolutionDialog(Dialog: TSaveDialog; const Name:
  string);
begin
  Dialog.FileName := Name;
  Dialog.DefaultExt := sVideSolutionExt;
  Dialog.Filter := FileTypes.TypesName[sVideSolution].Filter;
end;

procedure TMainActions.NewItem(AllowedItems: TNewItems; Project: TVideProject);
var
  Item: TNewItem;
begin
  FProject := Project;
  if FProject = nil then
    FProject := Solution.ActiveProject;
  Item := NewForm.ShowNewItems(AllowedItems);
  if Item <> nil then
    Item.Execute;
end;

procedure TMainActions.NewItem(const Name: string; Project: TVideProject);
var
  Item: TNewItem;
begin
  FProject := Project;
  if FProject = nil then
    FProject := Solution.ActiveProject;
  Item := NewItems.Find(Name);
  if Item <> nil then
    Item.Execute;
end;

procedure TMainActions.OpenFile(const FileName: string);
var
  Ext, FName: string;
  FType: TVideFileType;
begin
  FName:=Vfs.GetFullFileName(FileName);
  if Vfs.FileExists(FName) or (Editor.FilesByName[FName] <> nil) then
  begin
    Ext := GetFileExt(FName);
    FType := FileTypes.TypesExt[Ext];
    if FType <> nil then
    begin
      if FType.Name = sVideProject then
        AddProject(FName)
      else
        if FType.Name = sVideSolution then
          OpenSolution(False, FName)
        else
        begin
          if FType.CanOpenToEditor then
            EditorActions.OpenFile(FName, FType);
        end;
    end
    else
      EditorActions.OpenFile(FName, FType);
  end;
end;

procedure TMainActions.OpenFiles(AskForFiles: Boolean; Files: TStrings = nil);
var
  I: Long;
begin
  if AskForFiles then
  begin
    InitOpenFileDialog(GetOpen);
    if GetOpen.Execute then
      Files := GetOpen.Files;
  end;
  if Files <> nil then
    for I := 0 to Files.Count - 1 do
      OpenFile(Files[I]);
end;

procedure TMainActions.OpenSolution(AskForName: Boolean;
  const AFileName: string);
var
  Name: string;
begin
  if AskForName then
  begin
    InitOpenSolutionDialog(GetOpen);
    if GetOpen.Execute then
      Name := GetOpen.FileName
    else
      Exit;
  end
  else
    Name := AFileName;
  if Vfs.FileExists(Name) and CloseAll then
  begin
    Solution.LoadFromFile(Name);
  end;
end;

procedure TMainActions.ProjectRemove(AFile: TVideFile; Ask: Boolean);
var
  Ed: TVideEditorFile;
begin
  if Ask then
    if
      AskDlg(Format('Do you really wan to remove file ''%s'' from project ''%s''?',
      [ExtractFileName(AFile.FileName), AFile.Project.Name]),
      'Remove File From Project') then
    begin
      Ed := Editor.FilesByName[AFile.FileName];
      if Ed <> nil then
      begin
        Ed.FileInProject := nil;
        // EditorActions.CloseIndex(Editor.IndexOfFileName(AFile.FileName));
      end;
      AFile.Project.RemoveFile(AFile.FileName);
    end;
end;

function TMainActions.ProjectRename(Project: TVideProject;
  const NewName: string): Boolean;
var
  FullName, ExtName: string;
begin
  Result := False;
  ExtName := NewName + '.' + sVideProjectExt;
  if not CheckName(ExtName, ntProjects, Project, Project.FileName) then
    ErrorDlg(Format('Project %s is already contained in solution %s',
      [NewName, Solution.Name]), 'Duplicate Project Name')
  else
  begin
    Result := True;
    if Project.Dirty then
      Project.Rename(ExtName)
    else
    begin
      FullName := ExtractFilePath(Project.FileName) + ExtName;
      if Vfs.FileExists(FullName) then
      begin
        if AskDlg(Format('Project named %s already exists. Overwrite it?',
          [NewName]), 'Overwrite Project') then
          Project.Rename(FullName)
        else
          Result := False;
      end
      else
        Project.Rename(FullName)
    end;
  end;
end;

procedure TMainActions.ProjectViewSource(Project: TVideProject);
begin
  if (not Vfs.FileExists(Project.FileName)) or Project.Dirty then
    if not SaveProject(Project) then
    begin
      InfoDlg('You must save project first to view its source.',
        'View Project Source');
      Exit;
    end;
  EditorActions.OpenFile(Project.FileName, FileTypes.TypesName[sVideProject]);
end;

procedure TMainActions.SaveAll;
var
  I: Long;
begin
  for I := 0 to Editor.Count - 1 do
    SaveFile(Editor.Files[I]);
  SaveSolution;
end;

function TMainActions.SaveFile(EdFile: TVideEditorFile; AskForName: Boolean):
  Boolean;
begin
  Result := False;
  if EdFile <> nil then
  begin
    InitSaveFileDialog(GetSave, EdFile.FileName);
    if AskForName or (not Vfs.FileExists(EdFile.FileName)) or EdFile.Dirty then
      if GetSave.Execute then
      begin
        if CheckName(ExtractFileName(GetSave.FileName), ntFiles, EdFile.Project,
          EdFile.FileName) then
          EdFile.FileName := GetSave.FileName
        else
        begin
          ErrorDlg(Format('File %s is already contained in project %s',
            [ExtractFileName(GetSave.FileName), EdFile.Project.Name]),
            'Duplicate File Name');
          Exit;
        end;
      end
      else
        Exit;

    EditorActions.UpdateActive;
    EdFile.SaveToFile(EdFile.FileName);
    EditorActions.UpdateIndex(Editor.IndexOfFileName(EdFile.FileName));
    Result := True;
  end;
end;

procedure TMainActions.SaveItems;
begin
  if SaveForm.SaveModifiedItems(False, AllSaveItems) then
    SaveList(SaveForm.Items);
end;

procedure TMainActions.SaveList(List: TNObjectList);
var
  I: Long;
  Obj: TVideItem;
begin
  if List <> nil then
    for I := 0 to List.Count - 1 do
    begin
      Obj := TVideItem(List[I]);
      if Obj <> nil then
      begin
        if Obj.ClassNameIs(TVideSolution.ClassName) then
          SaveSolution
        else
          if Obj.ClassNameIs(TVideProject.ClassName) then
            SaveProject(TVideProject(Obj))
          else
            if Obj.ClassNameIs(TVideEditorFile.ClassName) then
              SaveFile(TVideEditorFile(Obj));
      end;
    end;
end;

function TMainActions.SaveProject(AProject: TVideProject; AskForName: Boolean):
  Boolean;
var
  I: Long;
  Name: string;
begin
  Result := False;
  InitSaveProjectDialog(GetSave, AProject.FileName);
  if AskForName or (not Vfs.FileExists(AProject.FileName)) or AProject.Dirty
    then
    if GetSave.Execute then
    begin
      Name := GetSave.FileName;
      if not CheckName(ExtractFileName(GetSave.FileName), ntProjects, AProject,
        AProject.FileName) then
      begin
        ErrorDlg(Format('Project %s is already contained in solution %s',
          [ChangeFileExt(ExtractFileName(GetSave.FileName), ''),
          Solution.Name]),
            'Duplicate Project Name');
        Exit;
      end;
      AProject.FileName := Name;
    end
    else
      Exit;

  for I := 0 to AProject.Count - 1 do
    if Editor.FilesByName[AProject.Files[I].FileName] <> nil then
      if not SaveFile(Editor.FilesByName[AProject.Files[I].FileName]) then
        Exit;

  //AProject.FileName := Name;
  AProject.SaveToFile(AProject.FileName);
  Result := True;
end;

function TMainActions.SaveSolution(AskForName: Boolean): Boolean;
var
  I: Long;
begin
  Result := False;
  InitSaveSolutionDialog(GetSave, Solution.FileName);
  if AskForName or (not Vfs.FileExists(Solution.FileName)) or Solution.Dirty
    then
    if GetSave.Execute then
      Solution.FileName := GetSave.FileName
    else
      Exit;

  for I := 0 to Solution.Count - 1 do
    if not SaveProject(Solution.Projects[I]) then
      Exit;

  Solution.SaveToFile(Solution.FileName);
  Result := True;
end;

procedure TMainActions.SolRemove(AProject: TVideProject; Ask: Boolean);
var
  Ed: TVideEditorFile;
  I: Long;
begin
  if Ask then
    if
      AskDlg(Format('Do you really wan to remove project ''%s'' from solution ''%s''?',
      [AProject.Name, Solution.Name]), 'Remove Project From Solution') then
    begin
      for I := 0 to Editor.Count - 1 do
      begin
        Ed := Editor.Files[I];
        if (Ed.FileInProject <> nil) then
          if (Ed.FileInProject.Project = AProject) then
            Ed.FileInProject := nil;
      end;
      Solution.RemoveProject(AProject.FileName);
    end;
end;

function TMainActions.SolRename(const NewName: string): Boolean;
var
  FullName: string;
begin
  Result := True;
  if Solution.Dirty then
    Solution.Rename(NewName + '.' + sVideSolutionExt)
  else
  begin
    FullName := ExtractFilePath(Solution.FileName) + NewName + '.' +
      sVideSolutionExt;
    if Vfs.FileExists(FullName) then
    begin
      if AskDlg(Format('Solution named %s already exists. Overwrite it?',
        [NewName]), 'Overwrite Solution') then
        Solution.Rename(FullName)
      else
        Result := False;
    end
    else
      Solution.Rename(FullName)
  end;
end;

procedure TMainActions.SolViewSource;
begin
  if (not Vfs.FileExists(Solution.FileName)) or Solution.Dirty then
    if not SaveSolution then
    begin
      InfoDlg('You must save solution first to view its source.',
        'View Solution Source');
      Exit;
    end;
  EditorActions.OpenFile(Solution.FileName, FileTypes.TypesName[sVideSolution]);
end;

procedure TMainActions.OnSolutionLoading(Cfg: TNConfig);
begin

end;

procedure TMainActions.OnSolutionSaving(Cfg: TNConfig);
begin

end;

function TMainActions.OnGetCompiler(const Name: NString): TVideCompiler;
begin
  Result := Compilers.Find(Name);
end;

procedure TMainActions.ActivateFile(AFile: TVideFile);
begin
  MainForm.TreeExp.Invalidate;
end;

function TMainActions.ProjectCompile(Project: TVideProject; Echo: Boolean;
  Build: Boolean; DeleteMsg: Boolean):
  Boolean;
begin
  Result := False;
  if DeleteMsg then
    MainForm.ListMsg.Items.Clear;
  if Project.CompilerSettings.Compiler = nil then
  begin
    if Echo then
      ErrorDlg(Format('Project %s can not be compiled. No compiler is asigned to it.',
        [Project.Name]), 'Compilation Error');
    Exit;
  end;
  if Project.ProjectType = ptCustom then
  begin
    if Echo then
      InfoDlg(Format('Project %s is of custom type which can not be compiled.',
        [Project.Name]), 'Compilation Info');
    Exit;
  end;
  Result := Project.CompilerSettings.Compiler.Compile(Project, Build);
end;

procedure TMainActions.Finalize;
begin
end;

function TMainActions.CloseApp: Boolean;
begin
  Options.SaveOptions;
  Result := CloseAll;
end;

procedure TMainActions.FirstShow;
begin
  Options.LoadOptions;
  FCanAsk := True;
end;

procedure TMainActions.ProjectRun(Project: TVideProject);
var
  S, Name: string;
begin
  if ProjectCompile(Project, True, False, True) then
  begin
    if Project.ProjectType = ptLibrary then
    begin
      ErrorDlg(Format('Project ''%s'' is of type DLL which can not be run.',
        [Project.Name]), 'Run Project');
      Exit;
    end;
    Name := ChangeFileExt(ExtractFileName(Project.ActiveFile.FileName), '.exe');
    if Project.OutputDir <> '' then
    begin
      S := Project.OutputDir + PathDelim + S;
      if not Vfs.FileExistsOutsideVfs(S) then
        S := ExtractFileDir(Project.FileName)+PathDelim+Project.OutputDir + PathDelim + Name;
    end
    else
      S := ExtractFileDir(Project.ActiveFile.FileName) + PathDelim + Name;
    if Vfs.FileExistsOutsideVfs(S) then
    begin
        if Options.MinimizeOnRun then
          Application.Minimize;
        RunFile(S, '', ExtractFileDir(S), True);
        if Options.MinimizeOnRun then
          Application.Restore;
    end
    else
      ErrorDlg(Format('File ''%s'' was not found.', [S]), 'Run Project');
  end;
end;

procedure TMainActions.ProjectBuild(Project: TVideProject);
begin
  ProjectCompile(Project, True, True, True);
end;

initialization
finalization
  FreeAndNil(MainActionsInstance);

end.

