unit ANewItems;

interface

uses
  SysUtils, Kernel, ABase, Objects, Lists, CSingletons, CEditor, AConsts,
  CProjects, CBase;

type

  TNewItem = class(TObject)
  protected
    FName: string;
    FCategory: TNewItemCategory;
    FImage: Long;
  public
    constructor Create(const AName:string; AImage:Long = 0);virtual;
    procedure Execute; virtual; abstract;
    property Name: string read FName write FName;
    property Category: TNewItemCategory read FCategory write FCategory;
    property Image:Long read FImage write FImage;
  end;

  TNewItemList = class(TNObjectList)
  protected
    function GetNewItem(Index: Long): TNewItem;
    procedure SetNewItem(Index: Long; const Value: TNewItem);
    procedure Notify(Index: Long; Action: TListNotification); override;
  public
    function Find(const Name:string):TNewItem;
    property Items[Index: Long]: TNewItem read GetNewItem write SetNewItem;
      default;
  end;

  TNewSolution = class(TNewItem)
  public
    constructor Create(const AName:string; AImage:Long = -1);override;
  end;

  TNewProject = class(TNewItem)
  public
    constructor Create(const AName:string; AImage:Long = -1);override;
  end;

  TNewFile = class(TNewItem)
  public
    constructor Create(const AName:string; AImage:Long = -1);override;
  end;

  TBlankSolution = class(TNewSolution)
  public
    procedure Execute;override;
  end;

  TGuiApplication = class(TNewProject)
  public
    procedure Execute;override;
  end;

  TConsoleApplication = class(TNewProject)
  public
    procedure Execute;override;
  end;

  TWin32Application = class(TNewProject)
  public
    procedure Execute;override;
  end;

  TLibrary = class(TNewProject)
  public
    procedure Execute;override;
  end;

  TPascalUnit = class(TNewFile)
  public
    procedure Execute;override;
  end;

  TDelphiDPR = class(TNewFile)
  public
    procedure Execute;override;
  end;

  TTextFile = class(TNewFile)
  public
    procedure Execute;override;
  end;

  TXMLDoc = class(TNewFile)
  public
    procedure Execute;override;
  end;

  THTMLPage = class(TNewFile)
  public
    procedure Execute;override;
  end;


function NewItems: TNewItemList;

implementation

uses
  AMainActions, AEditorActions, FFileNew;

var
  NewItemsInstance: TNewItemList;

procedure FillNewItems;forward;  

function NewItems: TNewItemList;
begin
  if NewItemsInstance = nil then
  begin
    NewItemsInstance := TNewItemList.Create;
    FillNewItems;
  end;
  Result := NewItemsInstance;
end;

{ TNewItemList }

function TNewItemList.Find(const Name: string): TNewItem;
var
  I:Long;
begin
  Result:=nil;
  for I:=0 to Count-1 do
    if SameStr(Name, Items[I].Name) then
    begin
      Result:=Items[I];
      Exit;
    end;
end;

function TNewItemList.GetNewItem(Index: Long): TNewItem;
begin
  if (Index > -1) and (Index < FCount) then
    Result := TNewItem(PPointerArray(FData)[Index])
  else
    Result := nil;
end;

procedure TNewItemList.Notify(Index: Long; Action: TListNotification);
begin
  if Action = lnDeleted then
    TNewItem(PPointerArray(FData)[Index]).Free;
end;

procedure TNewItemList.SetNewItem(Index: Long; const Value: TNewItem);
begin
  if (Index > -1) and (Index < FCount) then
  begin
    Notify(Index, lnDeleted);
    PPointerArray(FData)[Index] := Value;
    Notify(Index, lnAdded);
  end;
end;

procedure FillNewItems;
begin
  with NewItems do
  begin
    Add(TBlankSolution.Create(sBlankSolution, 9));
    Add(TGuiApplication.Create(sGUIApplication, 1));
    Add(TConsoleApplication.Create(sConsoleApplication, 6));
    Add(TWin32Application.Create(sWin32Application, 1));
    Add(TLibrary.Create(sLibrary, 5));
    Add(TPascalUnit.Create(sPascalUnit, 2));
    Add(TDelphiDPR.Create(sDelphiDPR, 3));
    Add(TXMLDoc.Create(sXMLDoc, 4));
    Add(THTMLPage.Create(sHTMLPage, 7));
    Add(TTextFile.Create(sTextFile, 8));
  end;
end;

{ TNewItem }

constructor TNewItem.Create(const AName:string; AImage:Long);
begin
  inherited Create;
  FName:=AName;
  FImage:=AImage;
end;

{ TNewSolution }

constructor TNewSolution.Create(const AName: string; AImage:Long);
begin
  inherited Create(AName, AImage);
  FCategory:=ntSolutions;
end;

{ TNewProject }

constructor TNewProject.Create(const AName: string; AImage:Long);
begin
  inherited Create(AName, AImage);
  FCategory:=ntProjects;
end;

{ TNewFile }

constructor TNewFile.Create(const AName: string; AImage:Long);
begin
  inherited Create(AName, AImage);
  FCategory:=ntFiles;
end;

{ TBlankSolution }

procedure TBlankSolution.Execute;
begin
  MainActions.CloseAll;
end;

{ TGuiApplication }

procedure TGuiApplication.Execute;
var
  Name:string;
  Prj:TVideProject;
  Ed:TVideEditorFile;
begin
  Name:=MainActions.CheckIntName('Project.'+sVideProjectExt, ntProjects);
  Prj:=Solution.AddNewProject(Name);
  Prj.ProjectType:=ptGuiApp;

  Name:='Main.pas';
  Prj.AddNewFile(Name);
  Ed:=EditorActions.CreateEdFile(Name, FileTypes.TypesExt[GetFileExt(Name)], True);

  Ed.Buffer.Add(Format(sGuiAppMainText, [ChangeFileExt(Prj.Name, '')]));
  EditorActions.AddFile(Ed);
end;

{ TConsoleApplication }

procedure TConsoleApplication.Execute;
var
  Name:string;
  Prj:TVideProject;
  Ed:TVideEditorFile;
begin
  Name:=MainActions.CheckIntName('Project.'+sVideProjectExt, ntProjects);
  Prj:=Solution.AddNewProject(Name);
  Prj.ProjectType:=ptConsoleApp;

  Name:='Main.pas';
  Prj.AddNewFile(Name);
  Ed:=EditorActions.CreateEdFile(Name, FileTypes.TypesExt[GetFileExt(Name)], True);

  Ed.Buffer.Add(Format(sConsoleAppMainText, [ChangeFileExt(Prj.Name, '')]));
  EditorActions.AddFile(Ed);
end;

{ TWin32Application }

procedure TWin32Application.Execute;
var
  Name:string;
  Prj:TVideProject;
  Ed:TVideEditorFile;
begin
  Name:=MainActions.CheckIntName('Project.'+sVideProjectExt, ntProjects);
  Prj:=Solution.AddNewProject(Name);
  Prj.ProjectType:=ptGuiApp;

  Name:='Main.pas';
  Prj.AddNewFile(Name);
  Ed:=EditorActions.CreateEdFile(Name, FileTypes.TypesExt[GetFileExt(Name)], True);

  Ed.Buffer.Add(Format(NewForm.LabWin32.Caption, [ChangeFileExt(Prj.Name, ''),
  ChangeFileExt(Prj.Name, ''), ChangeFileExt(Prj.Name, '')]));
  EditorActions.AddFile(Ed);
end;

{ TLibrary }

procedure TLibrary.Execute;
var
  Name:string;
  Prj:TVideProject;
  Ed:TVideEditorFile;
begin
  Name:=MainActions.CheckIntName('Project.'+sVideProjectExt, ntProjects);
  Prj:=Solution.AddNewProject(Name);
  Prj.ProjectType:=ptLibrary;

  Name:='Main.pas';
  Prj.AddNewFile(Name);
  Ed:=EditorActions.CreateEdFile(Name, FileTypes.TypesExt[GetFileExt(Name)], True);

  Ed.Buffer.Add(Format(sLibraryMainText, [ChangeFileExt(Prj.Name, '')]));
  EditorActions.AddFile(Ed);
end;


{ TPascalUnit }

procedure TPascalUnit.Execute;
var
  Ed:TVideEditorFile;
  Name:string;
begin
  Name:=MainActions.CheckIntName('Unit.pas', ntFiles, MainActions.ProjectToAdd);
  MainActions.ProjectToAdd.AddNewFile(Name);
  Ed:=EditorActions.CreateEdFile(Name, FileTypes.TypesExt[GetFileExt(Name)], True);

  Ed.Buffer.Add(Format(sPascalUnitText, [ChangeFileExt(Name, '')]));
  EditorActions.AddFile(Ed);
end;

{ TTextFile }

procedure TTextFile.Execute;
var
  Ed:TVideEditorFile;
  Name:string;
begin
  Name:=MainActions.CheckIntName('File.txt', ntFiles, MainActions.ProjectToAdd);
  MainActions.ProjectToAdd.AddNewFile(Name);
  Ed:=EditorActions.CreateEdFile(Name, FileTypes.TypesExt[GetFileExt(Name)], True);

  EditorActions.AddFile(Ed);
end;

{ TXMLDoc }

procedure TXMLDoc.Execute;
var
  Ed:TVideEditorFile;
  Name:string;
begin
  Name:=MainActions.CheckIntName('Doc.xml', ntFiles, MainActions.ProjectToAdd);
  MainActions.ProjectToAdd.AddNewFile(Name);
  Ed:=EditorActions.CreateEdFile(Name, FileTypes.TypesExt[GetFileExt(Name)], True);

  Ed.Buffer.Add('<?xml version="1.0" encoding="utf-8"?>');
  EditorActions.AddFile(Ed);
end;

{ TDelphiDPR }

procedure TDelphiDPR.Execute;
var
  Ed:TVideEditorFile;
  Name:string;
begin
  Name:=MainActions.CheckIntName('Program.dpr', ntFiles, MainActions.ProjectToAdd);
  MainActions.ProjectToAdd.AddNewFile(Name);
  Ed:=EditorActions.CreateEdFile(Name, FileTypes.TypesExt[GetFileExt(Name)], True);

  Ed.Buffer.Add(Format(sDelphiDPRText, [ChangeFileExt(Name, '')]));
  EditorActions.AddFile(Ed);
end;

{ THTMLPage }

procedure THTMLPage.Execute;
var
  Ed:TVideEditorFile;
  Name:string;
begin
  Name:=MainActions.CheckIntName('Page.html', ntFiles, MainActions.ProjectToAdd);
  MainActions.ProjectToAdd.AddNewFile(Name);
  Ed:=EditorActions.CreateEdFile(Name, FileTypes.TypesExt[GetFileExt(Name)], True);

  Ed.Buffer.Add('<html>');
  Ed.Buffer.Add('  <head>');
  Ed.Buffer.Add(Format('    <title>%s</title>',[ChangeFileExt(Name, '')]));
  Ed.Buffer.Add('  </head>');
  Ed.Buffer.Add('  <body>'+sLineBreak);
  Ed.Buffer.Add('  </body>');
  Ed.Buffer.Add('</html>');

  EditorActions.AddFile(Ed);
end;

initialization

finalization

  FreeAndNil(NewItemsInstance);

end.

