unit AOptions;

interface

uses
  Kernel, Classes, SysUtils, Objects, ActnMan, Lists, Configs, Files,
  AConsts, ActnPopupCtrl, Forms, CBase, ACompilers, CProjects, CSingletons,
  SynEdit, TypInfo, SynEditHighlighter, Streaming, SynHighlighterPas;

type

  TColorMapItem = class
  public
    Name: string;
    Map: TCustomActionBarColorMap;
    constructor Create(const AName: string; AMap: TCustomActionBarColorMap);
  end;

  TColorMapList = class(TNObjectList)
  protected
    function GetNewItem(Index: Long): TColorMapItem;
    procedure SetNewItem(Index: Long; const Value: TColorMapItem);
    procedure Notify(Index: Long; Action: TListNotification); override;
  public
    function Find(const Name: string): TColorMapItem;
    property Items[Index: Long]: TColorMapItem read GetNewItem write SetNewItem;
    default;
  end;

  TEditorOptions = class(TVidePersistent)
  protected
    FFindAtCursor: Boolean;
    FDblClickLine: Boolean;
    FBackSpaceUnindents: Boolean;
    FSyntaxColors: TSynPasSyn;
  public
    constructor Create;override;
    destructor Destroy;override;
    procedure ApplySyntaxColors(Syn:TSynPasSyn);
  published
    property FindAtCursor: Boolean read FFindAtCursor write FFindAtCursor;
    property DblClickLine: Boolean read FDblClickLine write FDblClickLine;
    property BackSpaceUnindents: Boolean read FBackSpaceUnindents write
      FBackSpaceUnindents;
    property SyntaxColors:TSynPasSyn read FSyntaxColors;  
  end;

  TOptions = class
  protected
    FColorMap: TColorMapItem;
    FStyle: TActionBarStyle;
    FPopupList: TNObjectList;
    FEditor: TEditorOptions;
    FAutoSave: Boolean;
    FRememberWindow: Boolean;
    FRememberFiles: Boolean;
    FShowProgress: Boolean;
    FMinimizeOnRun: Boolean;
    FPaths: TNStringList;
    FCompiler: TVideCompiler;
    procedure SetColorMap(const Value: TColorMapItem);
    procedure SetStyle(const Value: TActionBarStyle);
    procedure SetCompiler(const Value: TVideCompiler);

    procedure LoadEditor;
    procedure SaveEditor;
    procedure LoadSession;
    procedure SaveSession;
    procedure LoadWindow;
    procedure SaveWindow;
  public
    constructor Create;
    destructor Destroy; override;
    procedure LoadOptions;
    procedure SaveOptions;
    procedure RegisterAllPopups;
    procedure RegisterPopup(Menu: TPopupActionBarEx);

    property ColorMap: TColorMapItem read FColorMap write SetColorMap;
    property ActionBarStyle: TActionBarStyle read FStyle write SetStyle;
    property AutoSave: Boolean read FAutoSave write FAutoSave;
    property RememberWindow: Boolean read FRememberWindow write FRememberWindow;
    property RememberFiles: Boolean read FRememberWindow write FRememberFiles;
    property ShowProgress: Boolean read FShowProgress write FShowProgress;
    property MinimizeOnRun: Boolean read FMinimizeOnRun write FMinimizeOnRun;

    property Compiler: TVideCompiler read FCompiler write SetCompiler;

    property Paths: TNStringList read FPaths write FPaths;

    property Editor: TEditorOptions read FEditor write FEditor;
  end;

function ColorMaps: TColorMapList;
procedure RegisterColorMap(const AName: string; AMap: TCustomActionBarColorMap);
function Options: TOptions;

implementation

uses
  FMain, AMainActions, AEditorActions,  FData;

var
  OptionsInst: TOptions;
  ColorMapsInst: TColorMapList;

function ColorMaps: TColorMapList;
begin
  if ColorMapsInst = nil then
    ColorMapsInst := TColorMapList.Create;
  Result := ColorMapsInst;
end;

function Options: TOptions;
begin
  if OptionsInst = nil then
    OptionsInst := TOptions.Create;
  Result := OptionsInst;
end;

procedure RegisterColorMap(const AName: string; AMap: TCustomActionBarColorMap);
var
  Item: tColorMapItem;
begin
  Item := TColorMapItem.Create(AName, AMap);
  ColorMaps.Add(Item);
end;

{ TColorMapList }

function TColorMapList.Find(const Name: string): TColorMapItem;
var
  I: Long;
begin
  Result := nil;
  for I := 0 to Count - 1 do
    if SameStr(Name, Items[I].Name) then
    begin
      Result := Items[I];
      Exit;
    end;
end;

function TColorMapList.GetNewItem(Index: Long): TColorMapItem;
begin
  if (Index > -1) and (Index < FCount) then
    Result := TColorMapItem(PPointerArray(FData)[Index])
  else
    Result := nil;
end;

procedure TColorMapList.Notify(Index: Long; Action: TListNotification);
begin
  if Action = lnDeleted then
    TObject(PPointerArray(FData)[Index]).Free;
end;

procedure TColorMapList.SetNewItem(Index: Long;
  const Value: TColorMapItem);
begin
  if (Index > -1) and (Index < FCount) then
  begin
    Notify(Index, lnDeleted);
    PPointerArray(FData)[Index] := Value;
    Notify(Index, lnAdded);
  end;
end;

{ TColorMapItem }

constructor TColorMapItem.Create(const AName: string;
  AMap: TCustomActionBarColorMap);
begin
  Name := AName;
  Map := AMap;
end;

{ TOptions }

constructor TOptions.Create;
begin
  FPopupList := TNObjectList.Create;
  FEditor := TEditorOptions.Create;
  FPaths := TNStringList.Create;
  RegisterAllPopups;
end;

destructor TOptions.Destroy;
begin
  FPopupList.Free;
  FPaths.Free;
  FEditor.Free;
  inherited Destroy;
end;

procedure TOptions.LoadEditor;
var
  Cfg: TNConfig;
begin
  Cfg := TNConfig.Create(SettingsDir + PathDelim + sEditorCfg, csXML);

  Editor.BackSpaceUnindents:=Cfg.ReadBool('BackSpaceUnindents');
  Editor.DblClickLine:=Cfg.ReadBool('DblClickLine');
  Editor.FindAtCursor:=Cfg.ReadBool('FindAtCursor');

  with MainForm.EditorFrame do
  begin
    TypInfo.SetSetProp(SynEdit, 'Options', Cfg.ReadString('SynEditOptions'));
    SynEdit.WordWrap:=Cfg.ReadBool('WordWrap');
    SynEdit.InsertMode:=Cfg.ReadBool('InsertMode');
    SynEdit.TabWidth:=Cfg.ReadInt('TabWidth');
   // SynEdit.MaxUndo:=Cfg.ReadInt('MaxUndo');
    SynEdit.InsertCaret:=TSynEditCaretType(Cfg.ReadInt('InsertCaret'));
    SynEdit.OverWriteCaret:=TSynEditCaretType(Cfg.ReadInt('OverwriteCaret'));

    SynEdit.Font.Name:=Cfg.ReadString('Font.Name');
    SynEdit.Font.Size:=Cfg.ReadInt('Font.Size');
    SynEdit.Gutter.Visible:=Cfg.ReadBool('Gutter.Visible');
    SynEdit.Gutter.Font.Name:=Cfg.ReadString('Gutter.Font.Name');
    SynEdit.Gutter.Font.Size:=Cfg.ReadInt('Gutter.Font.Size');
    SynEdit.Gutter.ShowLineNumbers:=Cfg.ReadBool('Gutter.LineNumbers');
    SynEdit.Gutter.LeadingZeros:=Cfg.ReadBool('Gutter.LeadingZeros');
  end;

  Editor.FSyntaxColors.LoadFromFile(Vfs.GetFullFileName(SettingsDir + PathDelim + sSyntaxColorsCfg));
  Editor.ApplySyntaxColors(Editor.FSyntaxColors);

  Cfg.Free;
end;

procedure TOptions.LoadOptions;
var
  Cfg: TNConfig;
  S: string;
  Num, I: Long;
begin
  Cfg := TNConfig.Create(SettingsDir + PathDelim + sEnvironmentCfg, csXML);

  FileTypes.LoadTypes;

  FAutoSave := Cfg.ReadBool('Preferences.AutoSave');
  FRememberWindow := Cfg.ReadBool('Preferences.RememberWindow');
  FRememberFiles := Cfg.ReadBool('Preferences.RememberFiles');
  FShowProgress := Cfg.ReadBool('Preferences.ShowProgress');
  FMinimizeOnRun := Cfg.ReadBool('Preferences.MinimizeOnRun');

  S := Cfg.ReadString('Appearance.ColorMap');
  if ColorMaps.Find(S) <> nil then
    ColorMap := ColorMaps.Find(S)
  else
    ColorMap := ColorMaps.Find(sXPStyleColors);

  S := Cfg.ReadString('Appearance.ActionBarStyle');
  if ActionBarStyles.IndexOf(S) >= 0 then
    FStyle := ActionBarStyles.Style[ActionBarStyles.IndexOf(S)]
  else
    FStyle := MainForm.ActionManager.Style;

  S := Cfg.ReadString('Compilers.DefaultOne');
  if Compilers.Find(S) <> nil then
    Compiler := Compilers.Find(S)
  else
    Compiler := Compilers[0];

  Num := Cfg.ReadInt('Compilers.Count');
  Num := Min(Num, Compilers.Count);
  for I := 0 to Num - 1 do
    Compilers[I].Path := Cfg.ReadString(Format('Compilers.Item%d', [I]));

  Num := Cfg.ReadInt('Paths.Count');
  for I := 0 to Num - 1 do
    FPaths.Add(Cfg.ReadString('Paths.Item' + IntToStr(I)));

  Cfg.Free;

  LoadEditor;

  if RememberWindow then
    LoadWindow;
  if RememberFiles then
    LoadSession;
end;

procedure TOptions.LoadSession;
var
  Cfg: TNConfig;
  Num, I: Long;
  S: string;
begin
  Cfg := TNConfig.Create(SettingsDir + PathDelim + sSessionCfg, csXML);

  S := Cfg.ReadString('Solution');
  if Vfs.FileExists(S) then
    MainActions.OpenSolution(False, S);

  Num := Cfg.ReadInt('Editor.Count');
  for I := 0 to Num - 1 do
    MainActions.OpenFile(Cfg.ReadString('Editor.File' + IntToStr(I)));
  EditorActions.ActivateIndex(Cfg.ReadInt('Editor.Active'));

  Cfg.Free;
end;

procedure TOptions.LoadWindow;
var
  Cfg: TNConfig;
begin
  Cfg := TNConfig.Create(SettingsDir + PathDelim + sWindowCfg, csXML);

  MainForm.Left := Cfg.ReadInt('Window.Left');
  MainForm.Top := Cfg.ReadInt('Window.Top');
  MainForm.Width := Cfg.ReadInt('Window.Width');
  MainForm.Height := Cfg.ReadInt('Window.Height');
  MainForm.WindowState := TWindowState(Cfg.ReadInt('Window.State'));

  MainForm.PanelExp.Width := Cfg.ReadInt('Explorer.Width');
  MainForm.PanelExp.Visible := Cfg.ReadBool('Explorer.Visible');

  MainForm.PanelMsg.Height := Cfg.ReadInt('Messages.Height');
  MainForm.PanelMsg.Visible := Cfg.ReadBool('Messages.Visible');

  MainForm.PanelEditor.Visible := Cfg.ReadBool('Editor.Visible');

  Cfg.Free;
end;

procedure TOptions.RegisterAllPopups;
var
  I, J, K: Long;
begin
  for I := 0 to Application.ComponentCount - 1 do
    for J := 0 to Application.Components[I].ComponentCount - 1 do
      if Application.Components[I].Components[J] is TPopupActionBarEx then
        RegisterPopup(TPopupActionBarEx(Application.Components[I].Components[J]))
      else
        if Application.Components[I].Components[J] is TFrame then
          for K := 0 to Application.Components[I].Components[J].ComponentCount -
            1 do
            if Application.Components[I].Components[J].Components[K] is
              TPopupActionBarEx then
              RegisterPopup(TPopupActionBarEx(Application.Components[I].Components[J].Components[K]))
end;

procedure TOptions.RegisterPopup(Menu: TPopupActionBarEx);
begin
  FPopupList.Add(Menu);
end;

procedure TOptions.SaveEditor;
var
  Cfg: TNConfig;
begin
  Cfg := TNConfig.Create(csXML);

  Cfg.WriteBool('BackSpaceUnindents', Editor.BackSpaceUnindents);
  Cfg.WriteBool('DblClickLine', Editor.DblClickLine);
  Cfg.WriteBool('FindAtCursor', Editor.FindAtCursor);

  with MainForm.EditorFrame do
  begin
    Cfg.WriteString('SynEditOptions', TypInfo.GetSetProp(SynEdit,'Options'));
    Cfg.WriteBool('WordWrap', SynEdit.WordWrap);
    Cfg.WriteBool('InsertMode', SynEdit.InsertMode);
    Cfg.WriteInt('TabWidth', SynEdit.TabWidth);
    Cfg.WriteInt('MaxUndo', SynEdit.MaxUndo);
    Cfg.WriteInt('InsertCaret', Long(SynEdit.InsertCaret));
    Cfg.WriteInt('OverwriteCaret', Long(SynEdit.OverwriteCaret));

    Cfg.WriteString('Font.Name', SynEdit.Font.Name);
    Cfg.WriteInt('Font.Size', SynEdit.Font.Size);
    Cfg.WriteBool('Gutter.Visible', SynEdit.Gutter.Visible);
    Cfg.WriteString('Gutter.Font.Name', SynEdit.Gutter.Font.Name);
    Cfg.WriteInt('Gutter.Font.Size', SynEdit.Gutter.Font.Size);
    Cfg.WriteBool('Gutter.LineNumbers', SynEdit.Gutter.ShowLineNumbers);
    Cfg.WriteBool('Gutter.LeadingZeros', SynEdit.Gutter.LeadingZeros);
  end;

  Cfg.SaveToFile(SettingsDir + PathDelim + sEditorCfg);
  Cfg.Free;

  Editor.FSyntaxColors.SaveToFile(Vfs.GetFullFileName(SettingsDir + PathDelim + sSyntaxColorsCfg));
end;

procedure TOptions.SaveOptions;
var
  Cfg: TNConfig;
  I: Long;
begin
  FileTypes.SaveTypes;

  SaveEditor;

  Cfg := TNConfig.Create(csXML);

  Cfg.WriteBool('Preferences.AutoSave', FAutoSave);
  Cfg.WriteBool('Preferences.RememberWindow', FRememberWindow);
  Cfg.WriteBool('Preferences.RememberFiles', FRememberFiles);
  Cfg.WriteBool('Preferences.ShowProgress', FShowProgress);
  Cfg.WriteBool('Preferences.MinimizeOnRun', FMinimizeOnRun);

  Cfg.WriteString('Appearance.ColorMap', FColorMap.Name);
  Cfg.WriteString('Appearance.ActionBarStyle', FStyle.GetStyleName);

  Cfg.WriteString('Compilers.DefaultOne', FCompiler.Name);
  Cfg.WriteInt('Compilers.Count', Compilers.Count);
  for I := 0 to Compilers.Count - 1 do
    Cfg.WriteString(Format('Compilers.Item%d', [I]), Compilers[I].Path);

  Cfg.WriteInt('Paths.Count', FPaths.Count);
  for I := 0 to FPaths.Count - 1 do
    Cfg.WriteString('Paths.Item' + IntToStr(I), FPaths[I]);

  Cfg.SaveToFile(SettingsDir + PathDelim + sEnvironmentCfg);
  Cfg.Free;

  if RememberFiles then
    SaveSession;
  if RememberWindow then
    SaveWindow;
end;

procedure TOptions.SaveSession;
var
  Cfg: TNConfig;
  I: Long;
begin
  Cfg := TNConfig.Create(csXML);

  Cfg.WriteString('Solution', Solution.FileName);
  Cfg.WriteInt('Editor.Count', CSingletons.Editor.Count);
  Cfg.WriteInt('Editor.Active', CSingletons.Editor.ActiveIndex);
  for I := 0 to CSingletons.Editor.Count - 1 do
    if Vfs.FileExists(CSingletons.Editor.Files[I].FileName) then
      Cfg.WriteString('Editor.File' + IntToStr(I),
        CSingletons.Editor.Files[I].FileName);

  Cfg.SaveToFile(SettingsDir + PathDelim + sSessionCfg);
  Cfg.Free;
end;

procedure TOptions.SaveWindow;
var
  Cfg: TNConfig;
begin
  Cfg := TNConfig.Create(csXML);

  Cfg.WriteInt('Window.Left', MainForm.Left);
  Cfg.WriteInt('Window.Top', MainForm.Top);
  Cfg.WriteInt('Window.Width', MainForm.Width);
  Cfg.WriteInt('Window.Height', MainForm.Height);
  Cfg.WriteInt('Window.State', Long(MainForm.WindowState));

  Cfg.WriteInt('Explorer.Width', MainForm.PanelExp.Width);
  Cfg.WriteBool('Explorer.Visible', MainForm.PanelExp.Visible);

  Cfg.WriteInt('Messages.Height', MainForm.PanelMsg.Height);
  Cfg.WriteBool('Messages.Visible', MainForm.PanelMsg.Visible);

  Cfg.WriteBool('Editor.Visible', MainForm.PanelEditor.Visible);

  Cfg.SaveToFile(SettingsDir + PathDelim + sWindowCfg);
  Cfg.Free;
end;

procedure TOptions.SetColorMap(const Value: TColorMapItem);
var
  I: Long;
begin
  FColorMap := Value;
  for I := 0 to MainForm.ActionManager.ActionBars.Count - 1 do
    if MainForm.ActionManager.ActionBars[I].ActionBar <> nil then
      MainForm.ActionManager.ActionBars[I].ActionBar.ColorMap := FColorMap.Map;
  for I := 0 to FPopupList.Count - 1 do
    TPopupActionBarEx(FPopupList[I]).ColorMap := FColorMap.Map;
end;

procedure TOptions.SetCompiler(const Value: TVideCompiler);
begin
  FCompiler := Value;
  Solution.DefaultCompiler := FCompiler;
end;

procedure TOptions.SetStyle(const Value: TActionBarStyle);
begin
  FStyle := Value;
  MainForm.ActionManager.Style := FStyle;
end;

{ TEditorOptions }

procedure TEditorOptions.ApplySyntaxColors(Syn: TSynPasSyn);
begin
  FSyntaxColors.Assign(Syn);
  Data.PasSyn.Assign(Syn);
  Data.CppSyn.Assign(Syn);
  Data.XMLSyn.Assign(Syn);
  Data.JavaSyn.Assign(Syn);
  Data.BasicSyn.Assign(Syn);
  Data.CSharpSyn.Assign(Syn);
  Data.HTMLSyn.Assign(Syn);
  Data.AsmSyn.Assign(Syn);
  Data.DfmSyn.Assign(Syn);
  Data.INISyn.Assign(Syn);
end;

constructor TEditorOptions.Create;
begin
  inherited Create;
  FSyntaxColors:= TSynPasSyn.Create(MainForm);
end;

destructor TEditorOptions.Destroy;
begin
 // FSyntaxColors.Free;
  inherited Destroy;
end;

initialization

  RegisterClasses([TEditorOptions]);

finalization

  FreeAndNil(OptionsInst);
  FreeAndNil(ColorMapsInst);

end.

